/*
 * Endpoints EPCAM USB Camera Driver
 *
 * Copyright (c) 2003, 2004 Jeroen B. Vreeken (pe1rxq@amsat.org)
 *
 * Based on the se401 driver, which in turn is based on the ov511 driver.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Thanks to Endpoints Inc. (www.endpoints.com) for making documentation on
 * their chips available and sending me two reference boards.
 * 	- Jeroen Vreeken
 */

static const char version[] = "0.07";

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/pagemap.h>
#include <linux/usb.h>
#include <asm/io.h>
#include <asm/semaphore.h>


#include "epcam.h"

static int video_nr=-1;

static struct usb_device_id device_table [] = {
	{ USB_DEVICE(0x03e8, 0x1005), driver_info: (unsigned long)"Endpoints EP800"	},/* Reference model */
	{ USB_DEVICE(0x03e8, 0x1003), driver_info: (unsigned long)"Endpoints SE402"	},/* Reference model */
	{ USB_DEVICE(0x03e8, 0x1000), driver_info: (unsigned long)"Endpoints SE401"	},/* Reference model */
	{ USB_DEVICE(0x03e8, 0x2112), driver_info: (unsigned long)"SpyPen Actor" 	},
	{ USB_DEVICE(0x03e8, 0x2040), driver_info: (unsigned long)"Rimax Slim Multicam"	},
	{ USB_DEVICE(0x03e8, 0x1010), driver_info: (unsigned long)"Concord Eye-Q Easy"	},
	{ USB_DEVICE(0x041e, 0x400d), driver_info: (unsigned long)"Creative PD1001"	},
	{ USB_DEVICE(0x04f2, 0xa001), driver_info: (unsigned long)"Chicony DC-100"	},
	{ USB_DEVICE(0x08ca, 0x0102), driver_info: (unsigned long)"Aiptek Pencam 400"	},
	{ USB_DEVICE(0x03e8, 0x2182), driver_info: (unsigned long)"Concord EyeQ Mini"	},
	{ USB_DEVICE(0x03e8, 0x2123), driver_info: (unsigned long)"Sipix StyleCam"	},
	{}
};

MODULE_DEVICE_TABLE(usb, device_table);

MODULE_AUTHOR("Jeroen Vreeken <pe1rxq@amsat.org>");
MODULE_DESCRIPTION("EPcam USB Camera Driver");
MODULE_LICENSE("GPL");
MODULE_PARM(video_nr, "i");
MODULE_DEVICE_TABLE(usb, device_table);


static struct usb_driver epcam_driver;


/**********************************************************************
 *
 * Memory management
 *
 **********************************************************************/

/* Here we want the physical address of the memory.
 * This is used when initializing the contents of the area.
 */
static inline unsigned long kvirt_to_pa(unsigned long adr)
{
	unsigned long kva, ret;

	kva = (unsigned long) page_address(vmalloc_to_page((void *)adr));
	kva |= adr & (PAGE_SIZE-1); /* restore the offset */
	ret = __pa(kva);
	return ret;
}

static void *rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr;

	size = PAGE_ALIGN(size);
	mem = vmalloc_32(size);
	if (!mem)
		return NULL;

	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr = (unsigned long) mem;
	while (size > 0) {
		SetPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}

	return mem;
}

static void rvfree(void *mem, unsigned long size)
{
	unsigned long adr;

	if (!mem)
		return;

	adr = (unsigned long) mem;
	while ((long) size > 0) {
		ClearPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}



/****************************************************************************
 *
 * epcam register read/write functions
 *
 ***************************************************************************/

static int epcam_sndctrl(int set, struct usb_epcam *epcam, unsigned short req,
			 unsigned short value, unsigned char *cp, int size)
{
	return usb_control_msg (
                epcam->dev,
                set ? usb_sndctrlpipe(epcam->dev, 0) : usb_rcvctrlpipe(epcam->dev, 0),
                req,
                (set ? USB_DIR_OUT : USB_DIR_IN) | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                value,
                0,
                cp,
                size,
                HZ
        );
}

static int epcam_set_feature(struct usb_epcam *epcam, unsigned short reg,
			     unsigned short value)
{
	unsigned char cp[2];
	INT2QT(value, cp);
	return usb_control_msg (
	epcam->dev,
		usb_sndctrlpipe(epcam->dev, 0),
		VENDOR_REQ_EXT_FEATURE,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		reg,
		0,
		cp,
		2,
		HZ
	);
}

static unsigned short epcam_get_feature(struct usb_epcam *epcam, 
				        unsigned short reg)
{
	unsigned char cp[2];
	usb_control_msg (
		epcam->dev,
		usb_rcvctrlpipe(epcam->dev, 0),
		VENDOR_REQ_EXT_FEATURE,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		reg,
		0,
		cp,
		2,
		HZ
		);
	return QT2INT(cp);
}

static unsigned short epcam_read_bios(struct usb_epcam *epcam,
					unsigned short address)
{
	unsigned char cp[2];
	usb_control_msg (
		epcam->dev,
		usb_rcvctrlpipe(epcam->dev, 0),
		VENDOR_REQ_BIOS,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		VENDOR_CMD_BIOS_READ,
		address,
		cp,
		2,
		HZ
	);
	return QT2INT(cp);
}

/****************************************************************************
 *
 * Camera control
 *
 ***************************************************************************/


static int epcam_send_pict(struct usb_epcam *epcam)
{
	unsigned char cp[40];
	int size;

	if (epcam->brightness<4096)
		epcam->brightness=4096;
	
//	epcam_set_feature(epcam, HV7131_REG_TITU, epcam->brightness>>8);
//	epcam_set_feature(epcam, HV7131_REG_TITM, epcam->brightness&0xff);
//	epcam_set_feature(epcam, HV7131_REG_TITL, 0);

	/* the SE402 does this itself, don't bother it... */
	if (epcam->camid!=0x402) {
		epcam_sndctrl(0, epcam, VENDOR_REQ_IMAGE_INFO, 0, cp, sizeof(cp));
		INT2QT(epcam->brightness>>5, cp+2);
		size=QT2INT(cp);
		if (size>40)
			size=40;
		epcam_sndctrl(1, epcam, VENDOR_REQ_IMAGE_INFO, 0, cp, size);
	}

//	info("TIT: %d", 256*256*epcam_get_feature(epcam, HV7131_REG_TITU)+
//	    256*epcam_get_feature(epcam, HV7131_REG_TITM)+
//	    epcam_get_feature(epcam, HV7131_REG_TITL));

	return 0;
}

static int epcam_recv_pict(struct usb_epcam *epcam)
{
	unsigned char cp[40];
	    	
	if (epcam->camid!=0x402) {
		epcam_sndctrl(0, epcam, VENDOR_REQ_IMAGE_INFO, 0, cp, sizeof(cp));
		epcam->brightness=QT2INT(cp+2)<<5;
	}

//	epcam->brightness=
//	    epcam_get_feature(epcam, HV7131_REG_TITU)*256+
//	    epcam_get_feature(epcam, HV7131_REG_TITM);
//	epcam_get_feature(epcam, HV7131_REG_TITL);

	return 0;
}

static int epcam_get_pict(struct usb_epcam *epcam, struct video_picture *p)
{
	epcam_recv_pict(epcam);
	p->brightness=epcam->brightness;
	p->whiteness=32768;
	p->colour=65535;
	p->contrast=65535;
	p->hue=65535;
	p->palette=epcam->palette;
	p->depth=24; /* rgb24 */

	return 0;
}


static int epcam_set_pict(struct usb_epcam *epcam, struct video_picture *p)
{
	if (p->palette != VIDEO_PALETTE_RGB24) {
		epcam->palette=p->palette;
		return 1;
	}

	p->whiteness=32768;
	p->colour=65535;
	p->contrast=65535;
	p->hue=65535;
	epcam->brightness=p->brightness;
	epcam_send_pict(epcam);
	return 0;
}

/*
	Hyundai have some really nice docs about this and other sensor related
	stuff on their homepage: www.hei.co.kr
*/
static void epcam_auto_resetlevel(struct usb_epcam *epcam)
{
	unsigned int ahrc, alrc;
	int oldreset=epcam->resetlevel;

	/* For some reason these normally read-only registers don't get reset
	   to zero after reading them just once...
	 */
	epcam_get_feature(epcam, HV7131_REG_HIREFNOH); 
	epcam_get_feature(epcam, HV7131_REG_HIREFNOL);
	epcam_get_feature(epcam, HV7131_REG_LOREFNOH);
	epcam_get_feature(epcam, HV7131_REG_LOREFNOL);
	ahrc=256*epcam_get_feature(epcam, HV7131_REG_HIREFNOH) + 
	    epcam_get_feature(epcam, HV7131_REG_HIREFNOL);
	alrc=256*epcam_get_feature(epcam, HV7131_REG_LOREFNOH) +
	    epcam_get_feature(epcam, HV7131_REG_LOREFNOL);

	/* Not an exact science, but it seems to work pretty well... */
	if (alrc > 10) {
		while (alrc>=10 && epcam->resetlevel < 63) {
			epcam->resetlevel++;
			alrc /=2;
		}
	} else if (ahrc > 20) {
		while (ahrc>=20 && epcam->resetlevel > 0) {
			epcam->resetlevel--;
			ahrc /=2;
		}
	}
	if (epcam->resetlevel!=oldreset)
		epcam_set_feature(epcam, HV7131_REG_ARLV, epcam->resetlevel);

	return;
}

static void epcam_video_irq(struct urb *urb, struct pt_regs *regs)
{
	struct usb_epcam *epcam=urb->context;
	int length=0, i;

	/* ohoh... */
	if (!epcam->streaming) {
		return;
	}
	if (!urb) {
		info ("ohoh: null urb");
		return;
	}
	if (!epcam->dev) {
		info ("ohoh: device vapourished");
		return;
	}

	switch(epcam->scratch[epcam->scratch_next].state) {
		case BUFFER_READY:
		case BUFFER_BUSY: {
			epcam->dropped++;
			break;
		}
		case BUFFER_UNUSED: {
			for (i=0; i<EPCAM_PACKETBUFS; i++) {
				if (urb->iso_frame_desc[i].actual_length) {
					epcam->nullpackets=0;
					memcpy(
					 epcam->scratch[epcam->scratch_next].data+length,
					 (unsigned char *)urb->transfer_buffer+urb->iso_frame_desc[i].offset,
					 urb->iso_frame_desc[i].actual_length);
					length+=urb->iso_frame_desc[i].actual_length;
				}
			}
			if (length) {
				epcam->scratch[epcam->scratch_next].state=BUFFER_READY;
				epcam->scratch[epcam->scratch_next].offset=epcam->scratch_offset;
				epcam->scratch[epcam->scratch_next].length=length;
//				if (waitqueue_active(&epcam->wq))
					wake_up_interruptible(&epcam->wq);
				epcam->scratch_overflow=0;
				epcam->scratch_next++;
				if (epcam->scratch_next>=EPCAM_NUMSCRATCH)
					epcam->scratch_next=0;
			}
		}
	}

	if (length)
		epcam->scratch_offset++;
	else
		epcam->scratch_offset=0;
	urb->dev = epcam->dev;
	if ((i = usb_submit_urb(urb, GFP_ATOMIC)) != 0)
		info("usb_submit_urb() returnd %d", i);

	return;
}

static void epcam_send_size(struct usb_epcam *epcam, int width, int height)
{
	unsigned char cp[40];
	int compress=0;
	int size;

	/* Only set auto controls on cameras that support it... */
	if (epcam->camid==0x402)
		epcam_sndctrl(0, epcam, VENDOR_REQ_AUTO_CONTROL, 1, NULL, 0);
	
	epcam->format=FMT_BAYER;
	if (epcam->camid==0x800) {
		/* Don't use compression on small images: camera goes crazy */
		if (width*height>=38400) {
			compress=0x02;
			epcam->format=FMT_EPLITE;
			if (width <= epcam->maxwidth/2 &&
			    height <= epcam->maxheight/2) {
				width*=2;
				height*=2;
				compress|=0x80;
			}
		}
	}

	info("vendor_req_capture_info: %d", epcam_sndctrl(0, epcam, VENDOR_REQ_CAPTURE_INFO, 0, cp, sizeof(cp)));
	INT2QT(1, cp+2);
	INT2QT(0, cp+4);
	INT2QT(0, cp+6);
	INT2QT((epcam->maxwidth-width)/2, cp+4);
	INT2QT((epcam->maxheight-height)/2, cp+6);
	INT2QT(width, cp+8);
	INT2QT(height, cp+10);
	size=QT2INT(cp);
	if (size>40)
		size=40;
	epcam_sndctrl(1, epcam, VENDOR_REQ_CAPTURE_INFO, 0, cp, size);
	info("vendor_req_capture_info: %d", epcam_sndctrl(0, epcam, VENDOR_REQ_CAPTURE_INFO, 0, cp, sizeof(cp)));
	info("capture info size: %d", QT2INT(cp));
	info("mode     : %d", QT2INT(cp+2));
	info("xstart   : %d", QT2INT(cp+4));
	info("ystart   : %d", QT2INT(cp+6));
	info("width    : %d", QT2INT(cp+8));
	info("height   : %d", QT2INT(cp+10));
	info("framerate: %d", QT2INT(cp+12));
	info("zoom     : %d", QT2INT(cp+14));

	info("vendor_req_compression: %d", epcam_sndctrl(0, epcam, VENDOR_REQ_COMPRESSION, 0, cp, sizeof(cp)));
	INT2QT(compress, cp);
	epcam_sndctrl(1, epcam, VENDOR_REQ_COMPRESSION, 0, cp, 2);
	info("compression: %d", compress);

	return;
}

/*
	In this function epcam_send_pict is called several times,
	for some reason (depending on the state of the sensor and the phase of
	the moon :) doing this only in either place doesn't always work...
*/
static int epcam_start_stream(struct usb_epcam *epcam)
{
	struct urb *urb;
	int err=0, i, fx;
	struct usb_interface_descriptor *interface;
	struct usb_endpoint_descriptor *endpoint;

info("starting stream");

	/* Set the camera to Iso transfers */
	if (usb_set_interface(epcam->dev, epcam->iface, 4)< 0) {
		info("interface set failed");
		return -1;
	}

	epcam->streaming=1;

info("interface set");
//	interface=&epcam->dev->actconfig->interface[epcam->iface].altsetting[4];
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
	epcam->packetsize=le16_to_cpu(epcam->dev->config[0].interface[0]->altsetting[4].endpoint[0].desc.wMaxPacketSize);
#else
	epcam->packetsize=epcam->dev->config[0].interface[0]->altsetting[4].endpoint[0].desc.wMaxPacketSize;
#endif
	usb_set_interface(epcam->dev, epcam->iface, 4);
info("packetsize: %d", epcam->packetsize);

	epcam_sndctrl(1, epcam, VENDOR_REQ_CAM_POWER, 1, NULL, 0);
        epcam_sndctrl(1, epcam, VENDOR_REQ_LED_CONTROL, 1, NULL, 0);
info("led and power on");
	/* Set picture settings */

	epcam_send_pict(epcam);

	epcam_send_size(epcam, epcam->cwidth, epcam->cheight);
	

	epcam_sndctrl(1, epcam, VENDOR_REQ_CONT_CAPTURE, 1, NULL, 0);
info("capture on");
	/* Do some memory allocation */
	for (i=0; i<EPCAM_NUMFRAMES; i++) {
		epcam->frame[i].data=epcam->fbuf + i * epcam->maxframesize;
		epcam->frame[i].curpix=0;
	}
	for (i=0; i<EPCAM_NUMSBUF; i++) {
		epcam->sbuf[i].data=kmalloc(epcam->packetsize*EPCAM_PACKETBUFS, GFP_KERNEL);
	}

	epcam->scratch_offset=0;
	epcam->lastoffset=-1;
	epcam->scratch_next=0;
	epcam->scratch_use=0;
	epcam->scratch_overflow=0;
	for (i=0; i<EPCAM_NUMSCRATCH; i++) {
		epcam->scratch[i].data=kmalloc(epcam->packetsize*EPCAM_PACKETBUFS, GFP_KERNEL);
		epcam->scratch[i].state=BUFFER_UNUSED;
	}

	for (i=0; i<EPCAM_NUMSBUF; i++) {
		urb=usb_alloc_urb(EPCAM_PACKETBUFS, GFP_KERNEL);
		if(!urb)
			return ENOMEM;

		urb->dev=epcam->dev;
		urb->context=epcam;
		urb->pipe=usb_rcvisocpipe(epcam->dev, EPCAM_VIDEO_ENDPOINT);
		urb->transfer_flags=URB_ISO_ASAP;
		urb->transfer_buffer=epcam->sbuf[i].data;
		urb->complete=epcam_video_irq;
		urb->number_of_packets=EPCAM_PACKETBUFS;
		urb->transfer_buffer_length=epcam->packetsize*EPCAM_PACKETBUFS;
		urb->interval = 1;
		for (fx=0; fx<EPCAM_PACKETBUFS; fx++) {
			urb->iso_frame_desc[fx].offset=epcam->packetsize*fx;
			urb->iso_frame_desc[fx].length=epcam->packetsize;
		}
		epcam->urb[i]=urb;
	}
	for (i=0; i<EPCAM_NUMSBUF; i++) {
		err=usb_submit_urb(epcam->urb[i], GFP_KERNEL);
		if(err)
			err("urb burned down");
	}
info("urbs flying!");
	epcam_set_feature(epcam, HV7131_REG_ARLV, epcam->resetlevel);

	return 0;
}

static int epcam_stop_stream(struct usb_epcam *epcam)
{
	int i;

	if (!epcam->streaming || !epcam->dev)
		return 1;

	epcam->streaming=0;

	for (i=0; i<EPCAM_NUMSBUF; i++) if (epcam->urb[i]) {
		usb_unlink_urb(epcam->urb[i]);
		usb_free_urb(epcam->urb[i]);
		epcam->urb[i]=NULL;
		kfree(epcam->sbuf[i].data);
	}
	for (i=0; i<EPCAM_NUMSCRATCH; i++) {
		kfree(epcam->scratch[i].data);
		epcam->scratch[i].data=NULL;
	}
	epcam_sndctrl(1, epcam, VENDOR_REQ_CONT_CAPTURE, 0, NULL, 0);
        epcam_sndctrl(1, epcam, VENDOR_REQ_LED_CONTROL, 0, NULL, 0);
	epcam_sndctrl(1, epcam, VENDOR_REQ_CAM_POWER, 0, NULL, 0);

	return 0;
}

static int epcam_set_size(struct usb_epcam *epcam, int width, int height)
{
	int wasstreaming=epcam->streaming;

	/* Check to see if we need to change */
	if (epcam->cwidth==width && epcam->cheight==height)
		return 0;

	/* Check for a valid mode */
	if (width<=0 || height<=0)
		return 1;
	if ((width & 1) || (height & 1))
		return 1;
	if (width>epcam->maxwidth)
		return 1;
	if (height>epcam->maxheight)
		return 1;

	/* Stop a current stream and start it again at the new size */
	if (wasstreaming)
		epcam_stop_stream(epcam);
	epcam->cwidth=width;
	epcam->cheight=height;
	if (wasstreaming)
		epcam_start_stream(epcam);

	return 0;
}


/****************************************************************************
 *
 * Video Decoding
 *
 ***************************************************************************/

static inline void decode_bayer (struct usb_epcam *epcam, unsigned char *data, int len)
{
	int datasize=epcam->cwidth*epcam->cheight;
	struct epcam_frame *frame=&epcam->frame[epcam->curframe];

	unsigned char *framedata=frame->data, *curline, *nextline;
	int width=epcam->cwidth;
	int blineoffset=0, bline;
	int linelength=width*3, i;
	

	/* Check if we have to much data */
	if (frame->curpix+len > datasize) {
		info("to much! %d %d %d %d", datasize, frame->curpix, len, frame->curpix+len-datasize);
//		frame->curpix=0;
//		return;
		len=datasize-frame->curpix;
	}

	if (frame->curpix==0) {
		if (frame->grabstate==FRAME_READY) {
			frame->grabstate=FRAME_GRABBING;
		}
		/* Clear stuff the user might have put on this line */
		memset(framedata+datasize*3-linelength, 0, linelength);
		frame->curline=framedata+linelength*2;
		frame->curlinepix=0;
	}

	if (epcam->cheight%4)
		blineoffset=1;
	bline=frame->curpix/epcam->cwidth+blineoffset;

	curline=frame->curline;
	nextline=curline+linelength;
	if (nextline >= framedata+datasize*3)
		nextline=curline;
	while (len) {
		if (frame->curlinepix>=width) {
			frame->curlinepix-=width;
			bline=frame->curpix/width+blineoffset;
			curline+=linelength*2;
			nextline+=linelength*2;
			if (curline >= framedata+datasize*3) {
				frame->curlinepix++;
				curline-=3;
				nextline-=3;
				len--;
				data++;
				frame->curpix++;
			}
			if (nextline >= framedata+datasize*3)
				nextline=curline;
		}
		if ((bline&1)) {
			if ((frame->curlinepix&1)) {
				*(curline+2)=*data;
				*(curline-1)=*data;
				*(nextline+2)=*data;
				*(nextline-1)=*data;
			} else {
				*(curline+1)=
					(*(curline+1)+*data)/2;
				*(curline-2)=
					(*(curline-2)+*data)/2;
				*(nextline+1)=*data;
				*(nextline-2)=*data;
			}
		} else {
			if ((frame->curlinepix&1)) {
				*(curline+1)=
					(*(curline+1)+*data)/2;
				*(curline-2)=
					(*(curline-2)+*data)/2;
				*(nextline+1)=*data;
				*(nextline-2)=*data;
			} else {
				*curline=*data;
				*(curline-3)=*data;
				*nextline=*data;
				*(nextline-3)=*data;
			}
		}
		frame->curlinepix++;
		curline-=3;
		nextline-=3;
		len--;
		data++;
		frame->curpix++;
	}
	frame->curline=curline;

	if (frame->curpix>=datasize-epcam->packetsize) {
		/* Fix the top line */
		framedata+=linelength*2;
		for (i=0; i<linelength*2; i++) {
			framedata--;
			*(framedata)=*(framedata+linelength);
		}
		/* Fix the left side (green is already present) */
		for (i=0; i<epcam->cheight; i++) {
			*framedata=*(framedata+3);
			*(framedata+1)=*(framedata+4);
			*(framedata+2)=*(framedata+5);
			framedata+=linelength;
		}
		framedata-=linelength*2;
		/* Bottom */
		for (i=0; i<linelength*2; i++) {
			framedata++;
			*(framedata)=*(framedata-linelength);
		}
		frame->curpix=0;
		frame->grabstate=FRAME_DONE;
		epcam->framecount++;
		epcam->readcount++;
		if (epcam->frame[(epcam->curframe+1)&(EPCAM_NUMFRAMES-1)].grabstate==FRAME_READY) {
			epcam->curframe=(epcam->curframe+1) & (EPCAM_NUMFRAMES-1);
		}
	}
}

static inline void decode_eplite_integrate(struct usb_epcam *epcam, int data)
{
	int linelength=epcam->cwidth;
	int i;

	/* First two are absolute, all others relative.
	 */
	if (epcam->eplite_curpix < 2) {
		*(epcam->eplite_data+epcam->eplite_curpix)=1+data*4;
	} else {
		*(epcam->eplite_data+epcam->eplite_curpix)=i=
		    *(epcam->eplite_data+epcam->eplite_curpix-2)+data*4;
		if (i>255 || i<0) {
			epcam->lastoffset=-1;
			epcam->datacorrupt++;
		}
	}

	epcam->eplite_curpix++;

	if (epcam->eplite_curpix>=linelength) {
		decode_bayer(epcam, epcam->eplite_data, linelength);

		epcam->eplite_curpix=0;
		epcam->eplite_curline+=linelength;
		if (epcam->eplite_curline>=epcam->cheight*linelength)
			epcam->eplite_curline=0;
	}
}

static inline void decode_eplite (struct usb_epcam *epcam, struct epcam_scratch *buffer)
{
	int i;
	
	unsigned char *data=buffer->data;
	int len=buffer->length;
	struct epcam_frame *frame=&epcam->frame[epcam->curframe];

	int pos=0;
	int vlc_cod;
	int vlc_size;
	int vlc_data;
	int bit_cur;

	int bit;

	/* Check for cancelled frames: */
	for (i=0; i<len-3; i++) {
		if (data[i]==0xff && data[i+1]==0xff && data[i+2]==0xff && data[i+3]==0xff) {
			epcam->cancel++;
			frame->curpix=0;
			return;
		}
	}

	/* New image? */
	if (!buffer->offset) {
		frame->curpix=0;
	}
	if (!frame->curpix) {
		epcam->eplite_curline=0;
		epcam->eplite_curpix=0;
		epcam->vlc_cod=0;
		epcam->vlc_data=0;
		epcam->vlc_size=0;
		if (frame->grabstate==FRAME_READY)
			frame->grabstate=FRAME_GRABBING;
	}

	vlc_cod=epcam->vlc_cod;
	vlc_size=epcam->vlc_size;
	vlc_data=epcam->vlc_data;
	
	while (pos < len && frame->grabstate==FRAME_GRABBING) {
		bit_cur=8;
		while (bit_cur) {
			bit=((*data)>>(bit_cur-1))&1;
			if (!vlc_cod) {
				if (bit) {
					vlc_size++;
				} else {
					if (!vlc_size) {
						decode_eplite_integrate(epcam, 0);
					} else {
						vlc_cod=2;
						vlc_data=0;
					}
				}
			} else {
				if (vlc_size > 7) {
					epcam->datacorrupt++;
					epcam->lastoffset=-1;
					return;
				}
				if (vlc_cod==2) {
					if (!bit) vlc_data=-(1<<vlc_size)+1;
					vlc_cod--;
				}
				vlc_size--;
				vlc_data+=bit<<vlc_size;
				if (!vlc_size) {
					decode_eplite_integrate(epcam, vlc_data);
					vlc_cod=0;
				}
			}
			bit_cur--;
		}
		pos++;
		data++;
		
	}
	
	/* Store variable-lengt-decoder state if in middle of frame */
	if (frame->grabstate==FRAME_GRABBING) {
		epcam->vlc_size=vlc_size;
		epcam->vlc_cod=vlc_cod;
		epcam->vlc_data=vlc_data;
	} else {
		/* If there is data left regard image as corrupt */
		if (len-pos > epcam->packetsize) {
			info("len-pos: %d\n", len-pos);
			frame->grabstate=FRAME_GRABBING;
			epcam->datacorrupt++;
		}
		epcam->lastoffset=-1;
	}

}

static int epcam_newframe(struct usb_epcam *epcam, int framenr)
{
	DECLARE_WAITQUEUE(wait, current);
	int errors=0;
	int process;

	if (epcam->framecount%64==8)
		epcam_auto_resetlevel(epcam);

	while (epcam->streaming &&
	    (epcam->frame[framenr].grabstate==FRAME_READY ||
	     epcam->frame[framenr].grabstate==FRAME_GRABBING) ) {
		if(!epcam->frame[framenr].curpix) {
			errors++;
		}
		wait_interruptible(
		    epcam->scratch[epcam->scratch_use].state!=BUFFER_READY,
		    0,
		    &epcam->wq,
		    &wait
		);
		if (epcam->nullpackets > EPCAM_MAX_NULLPACKETS) {
			epcam->nullpackets=0;
			info("to many null length packets, restarting capture");
			epcam_stop_stream(epcam);
			epcam_start_stream(epcam);			
		} else {
			struct epcam_scratch *buffer=&epcam->scratch[epcam->scratch_use];
			unsigned char *data=epcam->scratch[epcam->scratch_use].data;
			int len=epcam->scratch[epcam->scratch_use].length;
			struct epcam_frame *frame=&epcam->frame[epcam->curframe];

			if (epcam->scratch[epcam->scratch_use].state!=BUFFER_READY) {
				epcam->frame[framenr].grabstate=FRAME_ERROR;
				return -EIO;
			}
			epcam->scratch[epcam->scratch_use].state=BUFFER_BUSY;

			process=1;
			/* Check if we lost packets in the que */
			if ( buffer->offset!=epcam->lastoffset+1) {
				epcam->underrun++;
				epcam->lastoffset=-1;
				process=0;
			} else {
				epcam->lastoffset=buffer->offset;
			}

			/* New image? */
			if (!buffer->offset) {
				frame->curpix=0;
			}

			if (process) {
				if (epcam->format==FMT_EPLITE) {
					decode_eplite(epcam, &epcam->scratch[epcam->scratch_use]);
				} else {
					decode_bayer(epcam, data, len);
				}
			}
			
			epcam->scratch[epcam->scratch_use].state=BUFFER_UNUSED;
			epcam->scratch_use++;
			if (epcam->scratch_use>=EPCAM_NUMSCRATCH)
				epcam->scratch_use=0;
			if (errors > EPCAM_MAX_ERRORS) {
				errors=0;
				info("to much errors, restarting capture");
				epcam_stop_stream(epcam);
				epcam_start_stream(epcam);
			}
		}
	}
	if (epcam->frame[framenr].grabstate==FRAME_GRABBING)
		return -EAGAIN;
	
	return 0;
}

static void epcam_remove_disconnected(struct usb_epcam *epcam)
{
	int i;

        epcam->dev = NULL;

	for (i=0; i<EPCAM_NUMSBUF; i++) if (epcam->urb[i]) {
		usb_unlink_urb(epcam->urb[i]);
		usb_free_urb(epcam->urb[i]);
		epcam->urb[i] = NULL;
		kfree(epcam->sbuf[i].data);
	}
	for (i=0; i<EPCAM_NUMSCRATCH; i++) if (epcam->scratch[i].data) {
		kfree(epcam->scratch[i].data);
	}

	info("%s disconnected", epcam->camera_name);

        /* Free the memory */
        if (!epcam->user) {
                kfree(epcam);
                epcam = NULL;
        }
}



/****************************************************************************
 *
 * Video4Linux
 *
 ***************************************************************************/


static int epcam_open(struct inode *inode, struct file *file)
{
	struct video_device *dev = video_devdata(file);
	struct usb_epcam *epcam = (struct usb_epcam *)dev;

	info("epcam_open()");
	if (epcam->user)
		return -EBUSY;
	epcam->fbuf=rvmalloc(epcam->maxframesize * EPCAM_NUMFRAMES);
	if(epcam->fbuf) 
		file->private_data = dev;
	else
		return -ENOMEM;

	epcam->user=1;
	info("epcam_open() done");

	return 0;
}

static int epcam_close(struct inode *inode, struct file *file)
{
	struct video_device *dev = video_devdata(file);
        struct usb_epcam *epcam = (struct usb_epcam *)dev;
	int i;

	rvfree(epcam->fbuf, epcam->maxframesize * EPCAM_NUMFRAMES);
	if (epcam->removed) {
		epcam_remove_disconnected(epcam);
		info("device unregistered");
	} else {
		for (i=0; i<EPCAM_NUMFRAMES; i++)
			epcam->frame[i].grabstate=FRAME_UNUSED;
		if (epcam->streaming)
			epcam_stop_stream(epcam);
		epcam->user=0;
	}
	file->private_data = NULL;
	return 0;
}

static int epcam_do_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, void *arg)
{
	struct video_device *vdev = file->private_data;
        struct usb_epcam *epcam = (struct usb_epcam *)vdev;

        if (!epcam->dev)
                return -EIO;
        switch (cmd) {
	case VIDIOCGCAP:
	{
		struct video_capability *b=arg;
		strcpy(b->name, epcam->camera_name);
		b->type = VID_TYPE_CAPTURE;
		b->channels = 1;
		b->audios = 0;
		b->maxwidth = epcam->maxwidth;
		b->maxheight = epcam->maxheight;
		b->minwidth = 2;
		b->minheight = 2;

		return 0;
	}
	case VIDIOCGCHAN:
	{
		struct video_channel *v=arg;

		if (v->channel != 0)
			return -EINVAL;

		v->flags = 0;
		v->tuners = 0;
		v->type = VIDEO_TYPE_CAMERA;
		strcpy(v->name, "Camera");

		return 0;
	}
	case VIDIOCSCHAN:
	{
		struct video_channel *v=arg;

		if (v->channel != 0)
			return -EINVAL;

		return 0;
	}
        case VIDIOCGPICT:
        {
		struct video_picture *p=arg;

		epcam_get_pict(epcam, p);

		return 0;
	}
	case VIDIOCSPICT:
	{
		struct video_picture *p=arg;

		if (epcam_set_pict(epcam, p))
			return -EINVAL;
		return 0;
	}
	case VIDIOCSWIN:
	{
		struct video_window *vw=arg;

		if (vw->flags)
			return -EINVAL;
		if (vw->clipcount)
			return -EINVAL;
		if (epcam_set_size(epcam, vw->width, vw->height))
			return -EINVAL;
		
		return 0;
        }
	case VIDIOCGWIN:
	{
		struct video_window *vw=arg;

		vw->x = 0;
		vw->y = 0;
		vw->chromakey = 0;
		vw->flags = 0;
		vw->clipcount = 0;
		vw->width = epcam->cwidth;
		vw->height = epcam->cheight;

		return 0;
	}
	case VIDIOCGMBUF:
	{
		struct video_mbuf *vm=arg;
		int i;

		memset(vm, 0, sizeof(struct video_mbuf));
		vm->size = EPCAM_NUMFRAMES * epcam->maxframesize;
		vm->frames = EPCAM_NUMFRAMES;
		for (i=0; i<EPCAM_NUMFRAMES; i++)
			vm->offsets[i] = epcam->maxframesize * i;

		return 0;
	}
	case VIDIOCMCAPTURE:
	{
		struct video_mmap *vm=arg;

		if (vm->format != VIDEO_PALETTE_RGB24)
			return -EINVAL;
		if (vm->frame < 0 || vm->frame >= EPCAM_NUMFRAMES)
			return -EINVAL;
		if (epcam->frame[vm->frame].grabstate != FRAME_UNUSED)
			return -EBUSY;

		if (epcam_set_size(epcam, vm->width, vm->height))
			return -EINVAL;
		epcam->frame[vm->frame].grabstate=FRAME_READY;

		if (!epcam->streaming)
			epcam_start_stream(epcam);

		return 0;
	}
	case VIDIOCSYNC:
	{
		int *frame=arg, ret=0;

		if (*frame < 0 || *frame >= EPCAM_NUMFRAMES)
			return -EINVAL;

		ret=epcam_newframe(epcam, *frame);
		epcam->frame[*frame].grabstate=FRAME_UNUSED;
		return ret;
	}
	case VIDIOCGFBUF:
	{
		struct video_buffer *vb=arg;

		memset(vb, 0, sizeof(*vb));
		vb->base = NULL; /* frame buffer not supported, not used */

		return 0;
	}
	case VIDIOCKEY:
		return 0;
	case VIDIOCCAPTURE:
		return -EINVAL;
	case VIDIOCSFBUF:
		return -EINVAL;
	case VIDIOCGTUNER:
	case VIDIOCSTUNER:
		return -EINVAL;
	case VIDIOCGFREQ:
	case VIDIOCSFREQ:
		return -EINVAL;
	case VIDIOCGAUDIO:
	case VIDIOCSAUDIO:
		return -EINVAL;
        default:
                return -ENOIOCTLCMD;
        } /* end switch */

        return 0;
}

static int epcam_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	return video_usercopy(inode, file, cmd, arg, epcam_do_ioctl);
}

static ssize_t epcam_read(struct file *file, char *buf,
	size_t count, loff_t *ppos)
{
	int realcount=count, ret=0;
	struct video_device *dev = file->private_data;
	struct usb_epcam *epcam = (struct usb_epcam *)dev;


	if (epcam->dev == NULL)
		return -EIO;
	if (realcount > epcam->cwidth*epcam->cheight*3)
		realcount=epcam->cwidth*epcam->cheight*3;

	if (!(epcam->frame[0].grabstate==FRAME_GRABBING)) {
		epcam->frame[0].grabstate=FRAME_READY;
		epcam->frame[1].grabstate=FRAME_UNUSED;
		epcam->curframe=0;
	}

	if (!epcam->streaming)
		epcam_start_stream(epcam);

	ret=epcam_newframe(epcam, 0);

	if (ret == -EAGAIN)
		return ret;
	if (!ret) {
		copy_to_user(buf, epcam->frame[0].data, realcount);
	} else {
		realcount=ret;
	}

	epcam->frame[0].grabstate=FRAME_UNUSED;
	return realcount;
}

static int epcam_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct video_device *dev = file->private_data;
	struct usb_epcam *epcam = (struct usb_epcam *)dev;
	unsigned long start = vma->vm_start;
	unsigned long size  = vma->vm_end-vma->vm_start;
	unsigned long page, pos;

	down(&epcam->lock);

	if (epcam->dev == NULL) {
		up(&epcam->lock);
		return -EIO;
	}
	if (size > (((EPCAM_NUMFRAMES * epcam->maxframesize) + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1))) {
		up(&epcam->lock);
		return -EINVAL;
	}
	pos = (unsigned long)epcam->fbuf;
	while (size > 0) {
		page = kvirt_to_pa(pos);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
		if (remap_page_range(vma, start, page, PAGE_SIZE, PAGE_SHARED)) {
#else
		if (remap_pfn_range(vma, start, page >> PAGE_SHIFT, PAGE_SIZE, PAGE_SHARED)) {
#endif
			up(&epcam->lock);
			return -EAGAIN;
		}
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}
	up(&epcam->lock);

        return 0;
}

static struct file_operations epcam_fops = {
	.owner =	THIS_MODULE,
	.open =		epcam_open,
	.release =	epcam_close,
	.read =		epcam_read,
	.mmap =		epcam_mmap,
	.ioctl =	epcam_ioctl,
	.llseek = 	no_llseek,
};

static struct video_device epcam_template = {
	.owner =	THIS_MODULE,
        .name =		"EPCAM USB camera",
    	.type = 	VID_TYPE_CAPTURE,
        .hardware =	VID_HARDWARE_SE401,
	.fops =		&epcam_fops,
	.release =	video_device_release,
	.minor =	-1,
};



/***************************/
static int epcam_init(struct usb_epcam *epcam)
{
        int i=0, rc;
        unsigned char cp[0x80];
	/* led on */
        epcam_sndctrl(1, epcam, VENDOR_REQ_LED_CONTROL, 1, NULL, 0);

	/* get camera descriptor */
	memset(cp, 0, 0x80);
	rc=epcam_sndctrl(0, epcam, VENDOR_REQ_CAMERA_INFO, 0, cp, sizeof(cp));
	info("vendor_req_camera_info: %d", rc);
	if (rc<0) {
		info("Error reading camera descriptor");
		return 1;
	}
	info("size     :%d  %x %x", QT2INT(cp   ), cp[0], cp[1]);
	info("rev      :%d  %x %x", QT2INT(cp+4 ), cp[4], cp[5]);
	info("maxwidth :%d  %x %x", QT2INT(cp+6 ), cp[6], cp[7]);
	info("maxheight:%d  %x %x", QT2INT(cp+8 ), cp[8], cp[9]);
	info("zoomcaps :%d  %x %x", QT2INT(cp+10), cp[10], cp[11]);
	info("ISPCaps  :%d  %x %x", QT2INT(cp+12), cp[12], cp[13]);
	info("Formats  :%d  %x %x", QT2INT(cp+14), cp[14], cp[15]);
	for (i=0; i<QT2INT(cp+14); i++) {
		if (QT2INT(cp+16+i*2)==EPCAM_FORMAT_BAYER)
			info("Bayer format supported");
		else
			info("Unknown format: %d", QT2INT(cp+16+i*2));
	}
	info("bios version: %d", epcam_read_bios(epcam, 0xfffc));

	epcam->maxwidth=QT2INT(cp+6);
	epcam->maxheight=QT2INT(cp+8);
	epcam->camid=QT2INT(cp+2);

	info("camid: %x", epcam->camid);
	if (epcam->camid!=0x800 &&
	    epcam->camid!=0x402 &&
	    epcam->camid!=0x401 ) {
		err("Not a supported camid: %d!", epcam->camid);
		return 1;
	}
	if (epcam->camid==0x401) {
		info("Your camera might work with this driver...");
		info("Please send your results to: pe1rxq@amsat.org");
	}

	epcam->cwidth=epcam->maxwidth/2;
	epcam->cheight=epcam->maxheight/2;
	epcam->maxframesize=epcam->maxwidth*epcam->maxheight*3;

	/* some default values */
	epcam->brightness=32768;
	epcam_send_pict(epcam);
	epcam->resetlevel=0x2d;
	epcam->palette=VIDEO_PALETTE_RGB24;
	epcam->dropped=0;
	epcam->framecount=0;
	epcam->readcount=0;
	epcam_recv_pict(epcam);
	
        /* Flash the led */
        epcam_sndctrl(1, epcam, VENDOR_REQ_CAM_POWER, 1, NULL, 0);
	epcam_sndctrl(1, epcam, VENDOR_REQ_LED_CONTROL, 1, NULL, 0);
        epcam_sndctrl(1, epcam, VENDOR_REQ_CAM_POWER, 0, NULL, 0);
	epcam_sndctrl(1, epcam, VENDOR_REQ_LED_CONTROL, 0, NULL, 0);


        return 0;
}

static int epcam_probe(struct usb_interface *intf,
	const struct usb_device_id *id)
{
	struct usb_device *dev = interface_to_usbdev(intf);
        struct usb_interface_descriptor *interface;
        struct usb_epcam *epcam;
        char *camera_name=NULL;

	if (!id)
		return -ENODEV;
        if (dev->descriptor.bNumConfigurations != 1)
                return -ENODEV;

	interface = &intf->altsetting[0].desc;
	
	if (interface->bInterfaceNumber!=0)
		return -ENODEV;

	camera_name=(char *)id->driver_info;

        /* We found one */
        info("epcam camera found: %s", camera_name);

        if ((epcam = kmalloc(sizeof(*epcam), GFP_KERNEL)) == NULL) {
                err("couldn't kmalloc epcam struct");
                return -ENODEV;
        }

        memset(epcam, 0, sizeof(*epcam));

        epcam->dev = dev;
        epcam->iface = interface->bInterfaceNumber;
        epcam->camera_name = camera_name;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
	info("firmware version: %02x", le16_to_cpu(dev->descriptor.bcdDevice) & 255);
#else
	info("firmware version: %02x", dev->descriptor.bcdDevice & 255);
#endif

        if (epcam_init(epcam)) {
		kfree(epcam);
		return -ENODEV;
	}

	init_waitqueue_head(&epcam->wq);
	init_MUTEX(&epcam->lock);
	wmb();

	memcpy(&epcam->vdev, &epcam_template, sizeof(epcam_template));
	memcpy(epcam->vdev.name, epcam->camera_name, strlen(epcam->camera_name));
	if (video_register_device(&epcam->vdev, VFL_TYPE_GRABBER, video_nr) == -1) {
		err("video_register_device failed");
		kfree(epcam);
		return -ENODEV;
	}
	info("registered new video device: video%d", epcam->vdev.minor);
	usb_set_intfdata(intf, epcam);

        return 0;
}

static void epcam_disconnect(struct usb_interface *intf)
{
	struct usb_epcam *epcam = usb_get_intfdata(intf);

	info("disconnected");
	usb_set_intfdata(intf, NULL);
	if (epcam) {
		info("unregistering video device");
        	video_unregister_device(&epcam->vdev);
		if (!epcam->user)
			epcam_remove_disconnected(epcam);
		else {
		        epcam->frame[0].grabstate = FRAME_ERROR;
		        epcam->frame[1].grabstate = FRAME_ERROR;

			epcam->streaming = 0;

	                wake_up_interruptible(&epcam->wq);
			epcam->removed = 1;
		}
	}
}


static struct usb_driver epcam_driver = {
//	.owner		= THIS_MODULE,
	.name		= "epcam",
	.id_table	= device_table,
	.probe		= epcam_probe,
	.disconnect	= epcam_disconnect,
};



/****************************************************************************
 *
 *  Module routines
 *
 ***************************************************************************/

static int __init usb_epcam_init(void)
{
	info("EPcam usb camera driver version %s registering", version);
	return usb_register(&epcam_driver);
}

static void __exit usb_epcam_exit(void)
{
	usb_deregister(&epcam_driver);
	info("EPcam driver deregistered");
}

module_init(usb_epcam_init);
module_exit(usb_epcam_exit);
