/* $Id: exac3.c,v 1.13 2003/07/09 11:19:13 cochran Exp $ 
 * Copyright (c) 2003 Richard Cochran (cochran@cs.umass.edu)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 */

/*
 *
 *  NB this source file is only here for testing and for my reference.
 *  All of the wisdom from here has been put into audiofunc.c.
 *
 */

#include <asm/uaccess.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/sound.h>
#include <linux/soundcard.h>
#include <linux/usb.h>

MODULE_AUTHOR("Richard Cochran");
MODULE_DESCRIPTION("AC3 usb audio driver");
MODULE_LICENSE("GPL");

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,19)
typedef struct iso_packet_descriptor iso_packet_descriptor_t;
#endif

/* prototypes */
static void    exac3_disconnect(struct usb_device *udev, void *context);
static int     exac3_ioctl(struct inode *inode,struct file *file,
			   unsigned int cmd,unsigned long arg);
static int     exac3_open(struct inode *inode, struct file *file);
static void*   exac3_probe(struct usb_device *dev,unsigned int ifnum,
			   const struct usb_device_id *id);
static int     exac3_release(struct inode *inode, struct file *file);
static ssize_t exac3_write(struct file *file,const char *buffer,
			   size_t count,loff_t *ppos);

/* globals */
static DECLARE_MUTEX(alldevmux);
static LIST_HEAD(alldevices);

static DECLARE_WAIT_QUEUE_HEAD(open_wq);
static DECLARE_WAIT_QUEUE_HEAD(write_wq);

#define AC3_IFNUM          3
#define AC3_DATA_ENDPOINT  4
#define AC3_SYNCH_ENDPOINT 5

#define AC3_FRAME_LEN 6144
#define BUFLEN (AC3_FRAME_LEN*10)
#define FRMFACTOR 1
#define NFRAMES   (FRMFACTOR * 5)
#define MAXPKTLEN 192 //0xC0
#define XFERSIZE  (NFRAMES * MAXPKTLEN)

#define N_OUTURBS  2
#define SYNCFRAMES 5

/* structs */

struct outbuf {
  spinlock_t lock;
  int running;
  char buf[XFERSIZE * 2];
  struct isourb {
    struct urb urb;
    iso_packet_descriptor_t frame[NFRAMES];
  } xferurb;
};

struct syncbuf {
  int running;
  struct syncurb {
    struct urb urb;
    iso_packet_descriptor_t frame[SYNCFRAMES];
  } syncurb;
};

struct exac3_device {
  struct list_head devices; /* Points to next (!) extigy */
  struct usb_device *dev;
  int in_use;
  int ossdev; /* Remembers which minors registered */
  int afmt;   /* Selected OSS audio format */
  struct outbuf out[N_OUTURBS];
  struct syncbuf sync[N_OUTURBS];;
  unsigned int freq;
  unsigned int phase;
  spinlock_t buflock;
  int head;
  int tail;
  int used;
  char inbuf[BUFLEN];
};

static struct usb_device_id exac3_id_table [] = {
  { USB_DEVICE(0x041E,0x3000),
    bInterfaceClass: USB_CLASS_AUDIO,
    bInterfaceSubClass: 1
  },
  {} /* Terminating entry */
};

static struct usb_driver exac3_driver = {
  name:       "exac3",
  disconnect: exac3_disconnect,
  id_table:   exac3_id_table,
  probe:      exac3_probe,
};

static struct file_operations exac3_fops = {
  owner:   THIS_MODULE,
  ioctl:   exac3_ioctl,
  open:    exac3_open,
  release: exac3_release,
  write:   exac3_write,
};

/* module boiler plate */

int exac3_init(void)
{
  info("$Revision: 1.13 $");
  return(usb_register(&exac3_driver));
}

void exac3_exit(void)
{
  usb_deregister(&exac3_driver);
}

module_init(exac3_init);
module_exit(exac3_exit);

/* methods */
static int start_xfer(struct exac3_device* exac3,int index,int size);
static void unprobe(struct exac3_device *exac3);

static void SET(struct usb_device* dev,
		u8 req,u16 wValue,u16 wIndex,u16 d1,u16 d2,u16 len)
{
  int err;
  unsigned char buf[2];
  buf[0] = d1;
  buf[1] = d2;
  if (len < 3) {
    err = usb_control_msg(dev, usb_sndctrlpipe(dev,0), req,
			  USB_RECIP_INTERFACE|USB_TYPE_CLASS|USB_DIR_OUT,
			  wValue, wIndex, buf, len, HZ);
    if (err < 0) err("cannot SET wIndex 0x%04hX : %d",wIndex,err);
  }
}

static void sync_complete(struct urb *urb)
{
  static int lastsf[N_OUTURBS] = {0};
  int delta, index;

  struct exac3_device* exac3 = (struct exac3_device*)urb->context;
  unsigned char *cp = urb->transfer_buffer;
  unsigned int freq, i;

  for (index = 0; index < N_OUTURBS; index++) {
    if (urb == &exac3->sync[index].syncurb.urb) break;
  }
  if (index > N_OUTURBS - 1) {
    err("urb unknown");
    return;
  }
  delta = urb->start_frame - lastsf[index];
  if (delta < 0) delta += 1024;
  lastsf[index] = urb->start_frame;

  if (delta != SYNCFRAMES * N_OUTURBS)
    err("sync_cmp: urb[%d] expected=%d delta=%d",index,SYNCFRAMES*N_OUTURBS,delta);

  if (urb->status) {
    err("sync_complete: bad status=%d",urb->status);
    //exac3->sync[index].running = 0;
  }
  for (i = 0; i < SYNCFRAMES; i++, cp += 3) {
    if (urb->iso_frame_desc[i].status) {
      err("sync_cmp: frame %u bad status %d",i,urb->iso_frame_desc[i].status);
      continue;
    }
    if (urb->iso_frame_desc[i].actual_length < 3) continue;
    freq = cp[0] | (cp[1] << 8) | (cp[2] << 16);
    if (exac3->freq != freq) {
      err("freq=%08X or %u . 0x%04X", freq, freq>>14, freq&0x3fff);
      exac3->freq = freq;
    }
  }

#if 1
  exac3->sync[index].running = 1;
#else
  if (exac3->used >= XFERSIZE) { /* Queue not empty, so restart xfer. */
    if (start_sync(exac3))
      exac3->sync.running = 0;
    else
      exac3->sync.running = 1;
  } else { /* Queue empty, no need to restart. */
    exac3->sync.running = 0;
  }
#endif
}

static int start_sync(struct exac3_device* exac3,int index)
{
  int err, i;
  struct urb* urb;
  if (!exac3->dev) return(-ENODEV);
  urb = &exac3->sync[index].syncurb.urb;
  for (i = 0; i < SYNCFRAMES; i++) {
    urb->iso_frame_desc[i].length = 3;
    urb->iso_frame_desc[i].offset = i*3;
  }
  urb->next = urb; //NULL;
  urb->dev = exac3->dev;
  if (!urb->pipe)
    urb->pipe = usb_rcvisocpipe(exac3->dev,AC3_SYNCH_ENDPOINT);
  urb->transfer_flags = USB_ISO_ASAP;
  if (!urb->transfer_buffer) 
    urb->transfer_buffer = kmalloc(3*SYNCFRAMES,GFP_KERNEL);
  urb->transfer_buffer_length = 3*SYNCFRAMES;
  urb->number_of_packets = SYNCFRAMES;
  urb->context = exac3;
  urb->complete = sync_complete;
  err = usb_submit_urb(urb);
  if (err) err("cannot submit sync urb: %d",err);
  return(err);
}

#define TIMING

static void xfer_complete(struct urb *urb)
{
  int index;
  struct exac3_device* exac3 = (struct exac3_device*)urb->context;
  unsigned long flags;
#ifdef TIMING
  static int lastsf[N_OUTURBS] = {0};
  int delta;
#endif

  for (index = 0; index < N_OUTURBS; index++) {
    if (urb == &exac3->out[index].xferurb.urb) break;
  }
  if (index > N_OUTURBS - 1) {
    err("urb unknown");
    return;
  }

#ifdef TIMING
  delta = urb->start_frame - lastsf[index]; 
  if (delta < 0) delta += 1024;
  if (delta != NFRAMES * N_OUTURBS)
    err("xfer_cmp: urb[%d] expected=%d delta=%d",index,NFRAMES*N_OUTURBS,delta);
  lastsf[index] = urb->start_frame;
#endif

  if (urb->status) {
    err("xfer_complete: bad status=%d",urb->status);
    spin_lock_irqsave(&exac3->out[index].lock,flags);
    exac3->out[index].running = 0;
    spin_unlock_irqrestore(&exac3->out[index].lock,flags);
    return;
  }

  /* Dequeue the block just completed. */
  spin_lock_irqsave(&exac3->buflock,flags);
  exac3->used -= urb->transfer_buffer_length;
  spin_unlock_irqrestore(&exac3->buflock,flags);

  exac3->head = (exac3->head + urb->transfer_buffer_length) % BUFLEN;

  spin_lock_irqsave(&exac3->out[index].lock,flags);
  if (exac3->used >= XFERSIZE) { /* Queue not empty, so restart xfer. */
    
    if (start_xfer(exac3,index,XFERSIZE))
      exac3->out[index].running = 0;
    else
      exac3->out[index].running = 1;

  } else { /* Queue empty, no need to restart. */
    exac3->out[index].running = 0;
    wake_up_interruptible(&write_wq);
  }
  spin_unlock_irqrestore(&exac3->out[index].lock,flags);
}

static int start_xfer(struct exac3_device* exac3,int index,int size)
{
  int err, i;
  int offset=0;
  int wholeframes, remainder;
  struct urb* urb;
  int pktlen=MAXPKTLEN;
  int samples;
  int len, rest;

  if (!exac3->dev) return(-ENODEV);

  urb = &exac3->out[index].xferurb.urb;
  wholeframes = size / pktlen;
  remainder = size % pktlen;
  if (remainder) {
    err("remainder=%d",remainder);
    return(-1);
  }
  if (wholeframes > NFRAMES) {
    err("too many whole frames needed");
    return(-1);
  }

  for (i = 0; i < wholeframes; i++) {
    exac3->phase = (exac3->phase & 0x3fff) + exac3->freq;
    samples = exac3->phase >> 14;
    pktlen = samples * 4;
    urb->iso_frame_desc[i].offset = offset;
    urb->iso_frame_desc[i].length = pktlen;
    offset += pktlen;
    //err("%d phase %u.0x%04X pktlen %03X", i, exac3->phase>>14, exac3->phase&0x3fff, pktlen);
  }

  //if (offset != size) err("short %d of %d",offset,size);

  urb->dev = exac3->dev;
  urb->pipe = usb_sndisocpipe(exac3->dev,AC3_DATA_ENDPOINT);
  urb->transfer_flags = USB_ISO_ASAP;

  if (exac3->head + offset > BUFLEN) {
    rest = exac3->head + offset - BUFLEN;
  } else {
    rest = 0;
  }
  len = offset - rest;

  memcpy(exac3->out[index].buf, &exac3->inbuf[exac3->head], len);

  if (rest) {
    //err("size=%d, %d - %d = %d",offset,exac3->head + offset,BUFLEN,rest);
    memcpy(&exac3->out[index].buf[len], &exac3->inbuf[0], rest);
  }

  urb->transfer_buffer = exac3->out[index].buf;

  urb->transfer_buffer_length = offset;
  urb->number_of_packets = wholeframes;
  urb->context = exac3;
  urb->complete = xfer_complete;
  err = usb_submit_urb(urb);
  if (err) {
    err("cannot submit iso urb: %d",err);
    return(err);
  }

  if (!exac3->sync[index].running)
    start_sync(exac3,index);

  return(0);
}

static void exac3_disconnect(struct usb_device *udev, void *context)
{
  int i, err;
  struct exac3_device *exac3 = context;
  unsigned long flags;

  info("disconnect() called");

  for (i = 0; i < N_OUTURBS; i++) {

    spin_lock_irqsave(&exac3->out[i].lock,flags);
    if (exac3->out[i].running) {
      err = usb_unlink_urb(&exac3->out[i].xferurb.urb);
      if (err) err("cannot unlink data urb%d: %d",i,err);
      else err("unlinked data urb%d: %d",i,err);
    }
    spin_unlock_irqrestore(&exac3->out[i].lock,flags);

    if (exac3->sync[i].running) {
      err = usb_unlink_urb(&exac3->sync[i].syncurb.urb);
      if (err) err("cannot unlink sync urb: %d",err);
      else err("unlinked sync urb: %d",err);
    }

  }

  down(&alldevmux);
  exac3->dev = NULL;
  if (!exac3->in_use) {
    unprobe(exac3);
  } else {
    wake_up_interruptible(&write_wq);
  }
  up(&alldevmux);
  return;
}

static int exac3_ioctl(struct inode *inode,struct file *file,
		       unsigned int cmd,unsigned long arg)
{
  int val = 0;
  int srate = 48000;
  struct exac3_device *exac3 = (struct exac3_device*)file->private_data;

  switch (cmd) {
  case OSS_GETVERSION:
    return put_user(SOUND_VERSION,(int*)arg);

  case SNDCTL_DSP_CHANNELS:
    if (get_user(val,(int*)arg))
      return(-EFAULT);
    err("ioctl SNDCTL_DSP_CHANNELS to %d",val);
    return put_user(val,(int*)arg);

  case SNDCTL_DSP_GETBLKSIZE: /* returns the fragment size */
    err("ioctl SNDCTL_DSP_GETBLKSIZE");
    return put_user(6144,(int*)arg);
    
  case SNDCTL_DSP_GETCAPS:
    err("ioctl SNDCTL_DSP_GETCAPS");
    val = DSP_CAP_BATCH;
    return put_user(val,(int*)arg);
    
  case SNDCTL_DSP_GETFMTS:
    err("ioctl SNDCTL_DSP_GETFMTS");
    return put_user(AFMT_AC3,(int*)arg);

  case SNDCTL_DSP_GETODELAY: /* returns number of unplayed bytes */
    //err("ioctl SNDCTL_DSP_GETODELAY");
    return put_user(exac3->used,(int*)arg);

  case SNDCTL_DSP_NONBLOCK:
    err("ioctl SNDCTL_DSP_NONBLOCK");
    file->f_flags |= O_NONBLOCK;
    return(0);
    
  case SNDCTL_DSP_SETFMT:
    if (get_user(val,(int*)arg)) return(-EFAULT);

    err("ioctl SNDCTL_DSP_SETFMT %d (AC3=%d QUERY=%d)",val,AFMT_AC3,AFMT_QUERY);
    
    if (val == AFMT_QUERY) return put_user(AFMT_AC3,(int*)arg);

    exac3->afmt = val;

    return put_user(val/*AFMT_AC3*/,(int*)arg);

  case SNDCTL_DSP_SPEED:
    if (get_user(val,(int*)arg))
      return(-EFAULT);
    err("ioctl SNDCTL_DSP_SPEED %d Hz",val);
    if (44100 == val) {
      srate = 44100;
    } else {
      srate = 48000;
    }
    return put_user(srate,(int*)arg);

  case SNDCTL_DSP_SETFRAGMENT: /* packed 0xMMMMSSSS max frags and frag size */
    err("ioctl SNDCTL_DSP_SETFRAGMENT");
    if (get_user(val,(int*)arg))
      return(-EFAULT);
    return 0;
    
  case SNDCTL_DSP_POST:
  case SNDCTL_DSP_RESET:
  case SNDCTL_DSP_STEREO:
  case SNDCTL_DSP_SYNC: /* return when last byte written to device */
    //err("ioctl ignored");
    return(0);

  case SNDCTL_DSP_GETIPTR:   /* returns count_info for input buffer */
  case SNDCTL_DSP_GETISPACE: /* returns audio_info_block for input buffer */
  case SNDCTL_DSP_GETOPTR:   /* returns count_info for output buffer */
  case SNDCTL_DSP_GETOSPACE: /* returns audio_info_block for output buffer */
  case SNDCTL_DSP_GETTRIGGER:
  case SNDCTL_DSP_SETDUPLEX:
  case SNDCTL_DSP_SETSYNCRO:
  case SNDCTL_DSP_SETTRIGGER:
  case SNDCTL_DSP_SUBDIVIDE:
  case SOUND_PCM_READ_BITS:
  case SOUND_PCM_READ_CHANNELS:
  case SOUND_PCM_READ_FILTER:
  case SOUND_PCM_READ_RATE:
  case SOUND_PCM_WRITE_FILTER:
    err("ioctl INVAL");
    return(-EINVAL);
  }
  return(-ENOIOCTLCMD);
}

static int exac3_open(struct inode *inode, struct file *file)
{
  DECLARE_WAITQUEUE(wait,current);
  int device_found=0, minor=MINOR(inode->i_rdev);
  struct exac3_device *exac3=NULL;
  struct list_head *dev;

  if (file->f_mode & FMODE_READ) return(-EINVAL);

  while (1) {
    if (down_interruptible(&alldevmux)) return(-ERESTARTSYS);
    for (dev = alldevices.next; dev != &alldevices; dev = dev->next) {
      exac3 = list_entry(dev,struct exac3_device,devices);
      if ( !((exac3->ossdev ^ minor) & ~0xf) ) {
	device_found = 1;
	break;
      }
    }
    if (!device_found) { /* Requested device not found. */
      up(&alldevmux);
      return(-ENODEV);
    }
    add_wait_queue(&open_wq,&wait);
    set_current_state(TASK_INTERRUPTIBLE);
    if (!exac3->in_use) break; /* Device is free to open... */
    if (file->f_flags & O_NONBLOCK) { /* Prefers not to wait */
      set_current_state(TASK_RUNNING);
      remove_wait_queue(&open_wq,&wait);
      up(&alldevmux);
      return(-EBUSY);
    }
    up(&alldevmux); /* Sorry, go to sleep. */
    schedule();
    if (signal_pending(current)) return(-ERESTARTSYS);
  }
  set_current_state(TASK_RUNNING);
  remove_wait_queue(&open_wq,&wait);
  file->private_data = exac3;
  exac3->in_use = 1;
  exac3->phase = 48<<14;
  up(&alldevmux);
  return(0);
}

static void *exac3_probe(struct usb_device *dev,unsigned int ifnum,
			 const struct usb_device_id *id)
{
  int i;
  struct exac3_device *exac3;
  struct usb_config_descriptor *config = dev->actconfig;
  
  if (0x041E != dev->descriptor.idVendor ||
      0x3000 != dev->descriptor.idProduct) {
    err("not an extigy");
    return(NULL);
  }
  if (ifnum != AC3_IFNUM) return(NULL);

  if (usb_set_interface(dev,AC3_IFNUM,1) < 0)
    err("usb_set_interface failed, device %d interface %d altsetting 1",
	dev->devnum,AC3_IFNUM);

  /* 1. allocate a device structure */
  exac3 = kmalloc(sizeof(struct exac3_device),GFP_KERNEL);
  if (!exac3) {
    err("low memory");
    return(NULL);
  }
  memset(exac3,0,sizeof(struct exac3_device));
  exac3->dev = dev;
  usb_driver_claim_interface(&exac3_driver,&config->interface[ifnum],exac3);
  info("claimed DEV%d IF%d",dev->devnum,ifnum);
  exac3->freq = 48 << 14;
  for (i = 0; i < N_OUTURBS; i++)
    spin_lock_init(&exac3->out[i].lock);
  spin_lock_init(&exac3->buflock);

  /* 2. register an OSS device */
  exac3->ossdev = register_sound_dsp(&exac3_fops,-1);
  if (exac3->ossdev < 0) {
    err("cannot register dsp device");
    kfree(exac3);
    return(NULL);
  } else {
    info("registered dsp (major 14 minor %d)",exac3->ossdev);
  }

  /* 3. add into alldevices */
  down(&alldevmux);
  list_add_tail(&exac3->devices,&alldevices);
  up(&alldevmux);

  SET(dev,0x01,0x0000,0x1d00,0x01,0x0,1);//selector
  SET(dev,0x01,0x0100,0x1100,0x01,0x0,1);//enable up MIX
  SET(dev,0x01,0x0200,0x1100,0x05,0x0,1);//mode   up MIX

  return(exac3);
}

static int exac3_release(struct inode *inode, struct file *file)
{
  int i;
  struct exac3_device *exac3 = (struct exac3_device*)file->private_data;
  //lock_kernel();
  down(&alldevmux);
  exac3->in_use = 0;
  if (!exac3->dev) {
    unprobe(exac3);
  } else {
    for (i = 0; i < N_OUTURBS; i++) {
      if (exac3->sync[i].running) {
	int err = usb_unlink_urb(&exac3->sync[i].syncurb.urb);
	if (err) err("release: cannot unlink sync urb: %d",err);
	else err("release: unlinked sync urb: %d",err);
      }
    }
  }
  up(&alldevmux);
  wake_up_interruptible(&open_wq);
  //unlock_kernel();
  return(0);
}

//#define TIMING_WRITES

#ifdef TIMING_WRITES
static long usdiff(struct timeval* t1,struct timeval* t2)
{
  return( 1000000*(t2->tv_sec - t1->tv_sec) + t2->tv_usec - t1->tv_usec );
}
#endif

static ssize_t exac3_write(struct file *file,const char *buffer,
			   size_t count,loff_t *ppos)
{
  struct exac3_device *exac3 = (struct exac3_device*)file->private_data;
  unsigned long flags;
  int i, used, len, rest, urbs=0;

#ifdef TIMING_WRITES
  static struct timeval t1;
  struct timeval t2;
  long delta;
#endif

  if (ppos != &file->f_pos) return(-ESPIPE);
  if (!access_ok(VERIFY_READ,buffer,count)) return(-EFAULT);

#ifdef TIMING_WRITES
  do_gettimeofday(&t2);
  delta = usdiff(&t1,&t2);
  if (t1.tv_sec) err("write: count=%d delta=%ld",count,delta);
  memcpy(&t1,&t2,sizeof(struct timeval));
#endif

  if (count != AC3_FRAME_LEN) {
    err("write: bad count=%d",count);
    return(-EINVAL);
  }
  if (AFMT_AC3 != exac3->afmt) {
    err("write: count=%d format ignored",count);
    return(count);
  }

  if (wait_event_interruptible(write_wq,(BUFLEN - exac3->used - 1) >= count))
    return(-ERESTARTSYS);

  if (!exac3->dev) return(-ENODEV);

  /* Enqueue the block. */
  if (exac3->tail + count > BUFLEN) {
    rest = exac3->tail + count - BUFLEN;
  } else {
    rest = 0;
  }
  len = count - rest;
  
  __copy_from_user(&exac3->inbuf[exac3->tail], buffer, len);
  
  if (rest) {
    __copy_from_user(&exac3->inbuf[0], &buffer[len], rest);
  }

  exac3->tail = (exac3->tail + count) % BUFLEN;

  spin_lock_irqsave(&exac3->buflock,flags);
  exac3->used += count;
  spin_unlock_irqrestore(&exac3->buflock,flags);
  
  used = exac3->used;

  for (i = 0; i < N_OUTURBS && used >= XFERSIZE; i++) {
    /* Queue not empty, so restart xfer. */
    if (!exac3->out[i].running) {
      spin_lock_irqsave(&exac3->out[i].lock,flags);
      if (start_xfer(exac3,i,XFERSIZE)) {
	exac3->out[i].running = 0;
      } else {
	exac3->out[i].running = 1;
	used -= XFERSIZE;
      }
      spin_unlock_irqrestore(&exac3->out[i].lock,flags);
    } else {
      urbs++;
    }
  }

  if (!urbs) err("urbs stopped");

  /*if (!exac3->sync.running) {
    if (start_sync(exac3)) exac3->sync.running = 0;
    else exac3->sync.running = 1;
    }*/

  return(count);
}

static void unprobe(struct exac3_device *exac3)
{
  struct list_head *dev;
  /* Undo what we did in probe() */
  /* 3. add into alldevices */
  for (dev = alldevices.next; dev != &alldevices; dev = dev->next) {
    if (exac3 == list_entry(dev,struct exac3_device,devices)) {
      list_del(dev);
      info("unprobe: list_del()");
      break;
    }
  }
  /* 2. register an OSS device */
  unregister_sound_dsp(exac3->ossdev);
  /* 1. allocate a device structure */
  kfree(exac3);
}
