/* $Id: irxine.c,v 1.2 2004/04/12 02:05:06 rcochran Exp $
 *
 *  Copyright (c) 2004 Richard Cochran
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 */
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include "ircodes.h"

#define MAXLINE 1024

static int verbose = 0;

static void quit(char *fmt, ...);
static int sendstr(int fd,const char *str);

static int get_mute(int fd)
{
  char buf[MAXLINE]={0};
  int on;
  if (!sendstr(fd,"get audio mute\n")) quit("cannot send command");
  if (read(fd,buf,MAXLINE) < 0) quit("Cannot read from xine socket");
  if (verbose) fprintf(stderr,buf);
  sscanf(buf,"Current audio mute: %d",&on);
  return(on);
}

static int get_volume(int fd)
{
  char buf[MAXLINE]={0};
  int level;
  if (!sendstr(fd,"get audio volume\n")) quit("cannot send command");
  if (read(fd,buf,MAXLINE) < 0) quit("Cannot read from xine socket");
  if (verbose) fprintf(stderr,buf);
  sscanf(buf,"Current audio volume: %d",&level);
  return(level);
}

static void quit(char *fmt, ...)
{
  va_list ap;
  char msg[MAXLINE];
  va_start(ap,fmt);
  vsnprintf(msg,MAXLINE,fmt,ap);
  va_end(ap);
  fprintf(stderr,"%s\n",msg);
  _exit(-1);
}

static int open_socket(const char *server,int port)
{
  int fd;
  struct sockaddr_in servaddr;
  if ((fd = socket(AF_INET,SOCK_STREAM,0)) < 0) {
    //fprintf(stderr,"socket error\n");
    return(-1);
  }
  memset(&servaddr,0,sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_port = htons(port);
  if (!inet_aton(server,&servaddr.sin_addr)) {
    //fprintf(stderr,"IP address error\n");
    return(-1);
  }
  if (connect(fd,(struct sockaddr*)&servaddr,sizeof(servaddr)) < 0) {
    //fprintf(stderr,"connect error\n");
    return(-1);
  }
  return(fd);
}

static int sendstr(int fd,const char *str)
{
  if (verbose) fprintf(stderr,str);
  if (strlen(str) != write(fd,str,strlen(str))) {
    return(0);
  } else {
    return(1);
  }
}

static void set_volume(int fd,int level)
{
  char buf[MAXLINE]={0};
  if (level < 0) level = 0;
  else if (level > 100) level = 100;
  snprintf(buf,MAXLINE,"set audio volume %d\n",level);
  if (!sendstr(fd,buf)) quit("cannot send command");
}

static int status_playing(int fd)
{
  char buf[MAXLINE]={0};
  if (!sendstr(fd,"get status\n")) quit("cannot send command");
  if (read(fd,buf,MAXLINE) < 0) quit("Cannot read from xine socket");
  if (verbose) fprintf(stderr,buf);
  if (strncmp(buf,"Current status: XINE_STATUS_PLAY",
	      strlen("Current status: XINE_STATUS_PLAY"))) {
    return(0);
  } else {
    return(1);
  }
}

int main(int argc,char* argv[])
{  
  char buf[MAXLINE]={0}, server[MAXLINE]="127.0.0.1";
  int fd, port=6789, xine;
  pid_t pid;
  unsigned int tmp;
  unsigned short code;

  xine = open_socket(server,port);
  if (xine < 0) {
    pid = fork();
    if (pid) { /* parent */
      sleep(3);
      xine = open_socket(server,port);
      if (xine < 0) {
	fprintf(stderr,"Cannot open connection to %s port %d\n",server,port);
	fprintf(stderr,"Cannot launch xine\n");
	kill(pid,SIGKILL);
	return(-1);
      }
    } else { /* child */
      execlp("xine","xine","-n",NULL);
    }
  }
  if (read(xine,buf,MAXLINE) < 0) quit("Cannot read from xine socket");
  if (verbose) fprintf(stderr,buf);
  if (!sendstr(xine,"gui hide\n")) quit("cannot send command");

  fd = open("/dev/ir",O_RDONLY);
  if (fd < 0) {
    perror("Cannot open IR device");
    return(-1);
  }

  while (1) {
    if (sizeof(int) != read(fd,&tmp,sizeof(int))) {
      perror("Cannot read IR device");
      return(-1);
    } else {
      code = tmp & 0xFF;
      switch (code) {
      case EXTIGY_NULL:
	break;
      case EXTIGY_POWER_ON:
	break;
      case EXTIGY_NUMERIC_1:
	if (!sendstr(xine,"event 1\n")) quit("cannot send command");
	break;
      case EXTIGY_NUMERIC_2:
	if (!sendstr(xine,"event 2\n")) quit("cannot send command");
	break;
      case EXTIGY_NUMERIC_3:
	if (!sendstr(xine,"event 3\n")) quit("cannot send command");
	break;
      case EXTIGY_NUMERIC_4:
	if (!sendstr(xine,"event 4\n")) quit("cannot send command");
	break;
      case EXTIGY_NUMERIC_5:
	if (!sendstr(xine,"event 5\n")) quit("cannot send command");
	break;
      case EXTIGY_NUMERIC_6:
	if (!sendstr(xine,"event 6\n")) quit("cannot send command");
	break;
      case EXTIGY_NUMERIC_7:
	if (!sendstr(xine,"event 7\n")) quit("cannot send command");
	break;
      case EXTIGY_NUMERIC_8:
	if (!sendstr(xine,"event 8\n")) quit("cannot send command");
	break;
      case EXTIGY_NUMERIC_9:
	if (!sendstr(xine,"event 9\n")) quit("cannot send command");
	break;
      case EXTIGY_NUMERIC_0:
	if (!sendstr(xine,"event 0\n")) quit("cannot send command");
	break;
      case EXTIGY_CMSS:
	break;
      case EXTIGY_MUTE:
	if (get_mute(xine)) {
	  if (!sendstr(xine,"set mute 0\n")) quit("cannot send command");
	} else {
	  if (!sendstr(xine,"set mute 1\n")) quit("cannot send command");
	}
	break;
      case EXTIGY_REC:
	break;
      case EXTIGY_VOL_MINUS:
	set_volume(xine,get_volume(xine)-5);
	break;
      case EXTIGY_VOL_PLUS:
	set_volume(xine,get_volume(xine)+5);
	break;
      case EXTIGY_STOP_EJECT:
	if (!sendstr(xine,"stop\n")) quit("cannot send command");
	break;
      case EXTIGY_PLAY_PAUSE:
	if (status_playing(xine)) {
	  if (!sendstr(xine,"pause\n")) quit("cannot send command");
	} else {
	  if (!sendstr(xine,"play\n")) quit("cannot send command");
	}
	break;
      case EXTIGY_SLOW:
	if (!sendstr(xine,"set speed *2\n")) quit("cannot send command");
	break;
      case EXTIGY_PREVIOUS:
	if (!sendstr(xine,"event previous\n")) quit("cannot send command");
	break;
      case EXTIGY_NEXT:
	if (!sendstr(xine,"event next\n")) quit("cannot send command");
	break;
      case EXTIGY_STEP:
	if (!sendstr(xine,"set speed *4\n")) quit("cannot send command");
	break;
      case EXTIGY_EAX:
	break;
      case EXTIGY_OPTIONS:
	if (!sendstr(xine,"mrl play dvd:/\n")) quit("cannot send command");
	break;
      case EXTIGY_DISPLAY:
	if (!sendstr(xine,"fullscreen\n")) quit("cannot send command");
	break;
      case EXTIGY_RETURN:
	break;
      case EXTIGY_START:
	if (!sendstr(xine,"event menu1\n")) quit("cannot send command");
	break;
      case EXTIGY_CANCEL:
	break;
      case EXTIGY_UP_ARROW:
	if (!sendstr(xine,"event up\n")) quit("cannot send command");
	break;
      case EXTIGY_LEFT_ARROWS:
	if (!sendstr(xine,"event left\n")) quit("cannot send command");
	break;
      case EXTIGY_SELECT_OK:
	if (!sendstr(xine,"event select\n")) quit("cannot send command");
	break;
      case EXTIGY_RIGHT_ARROWS:
	if (!sendstr(xine,"event right\n")) quit("cannot send command");
	break;
      case EXTIGY_DOWN_ARROW:
	if (!sendstr(xine,"event down\n")) quit("cannot send command");
	break;
      }

    }
  }

  return(0);
}
