/* $Id: remote.c,v 1.21 2004/11/29 01:21:08 rcochran Exp $
 *
 *  Copyright (c) 2003, 2004 Richard Cochran (cochran@cs.umass.edu)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 */
#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/usb.h>
#include <linux/version.h>
#include <linux/vmalloc.h>
#ifdef USE_LIRC
#include "../lirc/drivers/lirc.h"
#define DRIVER_LIRC "exaudio_remote_lirc"
#endif
#include "exaudio_remote.h"

#define DRIVER "exaudio_remote"
MODULE_AUTHOR("Richard Cochran");
MODULE_DESCRIPTION("IR remote driver for extigy");
MODULE_LICENSE("GPL");

#ifdef DEBUG
#define DBG(fmt, arg...) printk(KERN_DEBUG "%s: " fmt "\n" , DRIVER , ## arg)
#else
#define DBG(fmt, arg...) do {} while (0)
#endif
#define ERR(fmt, arg...) printk(KERN_ERR "%s: " fmt "\n" , DRIVER , ## arg)
#define INFO(fmt, arg...) printk(KERN_INFO "%s: " fmt "\n" , DRIVER , ## arg)

#define SET_CUR    0x01   /* Audio Class specific Request Codes */
#define GET_CUR    0x81
#define SET_MIN    0x02
#define GET_MIN    0x82
#define SET_MAX    0x03
#define GET_MAX    0x83
#define SET_RES    0x04
#define GET_RES    0x84
#define SET_MEM    0x05
#define GET_MEM    0x85
#define GET_STAT   0xff

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,19)
typedef struct devrequest {
	__u8 requesttype;
	__u8 request;
	__u16 value;
	__u16 index;
	__u16 length;
} __attribute__ ((packed)) devrequest;
#endif

/* private prototypes */
static void ir_disconnect(struct usb_interface *intf);
static int ir_ioctl(struct inode *node,struct file *filep,unsigned int cmd,
		    unsigned long arg);
static int ir_open(struct inode *inode, struct file *filp);
static int ir_probe(struct usb_interface *intf,
		    const struct usb_device_id *id);
static ssize_t ir_read(struct file *filp,char *buf,size_t count,loff_t *f_pos);
static int ir_release(struct inode *inode, struct file *filp);

/* globals */
static DECLARE_WAIT_QUEUE_HEAD(readwq);
static int irmajor;

static struct ir_device {
	struct usb_device *dev;
	struct urb *urb;
	unsigned char *buf;
	int present;
	u32 irval;
} remote;

static struct usb_device_id ir_id_table [] = {
	{ USB_DEVICE(0x041E,0x3000),
	  bInterfaceClass: USB_CLASS_AUDIO,
	  bInterfaceSubClass: 1
	},
	{} /* Terminating entry */
};

static struct usb_driver ir_driver = {
	name:       DRIVER,
	disconnect: ir_disconnect,
	id_table:   ir_id_table,
	probe:      ir_probe,
};

static struct file_operations ir_fops = {
	owner:   THIS_MODULE,
	ioctl:   ir_ioctl,
	open:    ir_open,
	read:    ir_read,
	release: ir_release,
};

#ifdef USE_LIRC
#define LIRC_MAJOR 61

static int lirc_ioctl(struct inode *node,struct file *filep,unsigned int cmd,
		      unsigned long arg)
{
	int result;
	unsigned long features=LIRC_CAN_REC_CODE, mode;
	switch(cmd) {
	case LIRC_GET_FEATURES:
		result = put_user(features,(unsigned long*)arg);
		if (result) return(result); 
		break;
	case LIRC_GET_REC_MODE:
		result = put_user(LIRC_MODE_CODE,(unsigned long*)arg);
		if (result) return(result); 
		break;
	case LIRC_SET_REC_MODE:
		result = get_user(mode,(unsigned long*)arg);
		if (result) return(result);
		if (mode!=LIRC_MODE_CODE) return(-ENOSYS);
		break;
	default:
		return(-ENOIOCTLCMD);
	}
	return(0);
}

static int lirc_open(struct inode *node,struct file *filep)
{
	return(ir_open(node,filep));
}

static int lirc_release(struct inode *node,struct file *filep)
{
	return(0);
}

static struct file_operations lirc_fops = 
{
	owner:   THIS_MODULE,
	ioctl:   lirc_ioctl,
	open:    lirc_open,
	read:    ir_read,
	release: lirc_release,
};
#endif

/* module boiler plate */

int ir_init(void)
{
	INFO("$Revision: 1.21 $");
	irmajor = register_chrdev(0,DRIVER,&ir_fops);
	if (irmajor < 0) {
		ERR("cannot get a major number");
		return(-1);
	} else {
		INFO("registered /dev/ir (major %d)",irmajor);
	}
#ifdef USE_LIRC
	if (register_chrdev(LIRC_MAJOR,DRIVER_LIRC,&lirc_fops) < 0) {
		ERR("register_chrdev() failed for lirc");
		unregister_chrdev(irmajor,DRIVER);
		return(-1);
	}
	INFO("registered lirc device, major %d",LIRC_MAJOR);
#endif
	return(usb_register(&ir_driver));
}

void ir_exit(void)
{
	int err = usb_unlink_urb(remote.urb);
	if (err) DBG("ir_exit: cannot unlink data urb: %d",err);
	else DBG("ir_exit: unlinked data urb: %d",err);
	if (remote.urb) {
		usb_free_urb(remote.urb);
	}
	if (remote.buf) {
		kfree(remote.buf);
	}
#ifdef USE_LIRC
	if (unregister_chrdev(LIRC_MAJOR,DRIVER_LIRC))
		ERR("cannot unregister lirc device");
#endif
	if (unregister_chrdev(irmajor,DRIVER))
		ERR("cannot unregister ir device");
	usb_deregister(&ir_driver);
}

module_init(ir_init);
module_exit(ir_exit);

/* private functions */

static int ctlmsg(struct usb_device *dev,u8 request,u8 requesttype,u16 value,
		  u16 index,u16 len,usb_complete_t complete,void *context)
{
	struct urb *urb = NULL;
	devrequest *dr = NULL;
	unsigned char *buf = NULL;
	urb = usb_alloc_urb(0,GFP_ATOMIC);
	dr  = kmalloc(sizeof(devrequest),GFP_ATOMIC);
	buf = kmalloc(len,GFP_ATOMIC);
	if (!urb || !dr || !buf) {
		ERR("low memory");
		if (urb) kfree(urb);
		if (dr)  kfree(dr);
		if (buf) kfree(buf);
		return(-ENOMEM);
	}
	dr->request     = request;
	dr->requesttype = requesttype;
	dr->value       = cpu_to_le16p(&value);
	dr->index       = cpu_to_le16p(&index);
	dr->length      = cpu_to_le16p(&len);
	usb_fill_control_urb(urb,dev,usb_rcvctrlpipe(dev,0),
			     (char*)dr,buf,len,complete,context);
	return(usb_submit_urb(urb,GFP_ATOMIC));
}

static void fetch_ir_code(struct urb *u,struct pt_regs *regs)
{
	struct ir_device *ir = u->context;
	unsigned char *buf = u->transfer_buffer;
	if (u->status < 0) {
		ERR("control request completes with bad status=%d",u->status);
	} else {
		ir->irval = (buf[1]<<8) | buf[0];
	}
	kfree(u->setup_packet);
	kfree(u->transfer_buffer);
	usb_free_urb(u);
	wake_up_interruptible(&readwq);
}

static void handle_interrupt(struct urb *u,struct pt_regs *regs)
{
	static unsigned int last = 0;
	struct ir_device *ir = u->context;
	int err;
	if (u->status) {
		DBG("interrupt status=%d, device no longer present",u->status);
		ir->present = 0;
		usb_free_urb(ir->urb);
		ir->urb = NULL;
		kfree(ir->buf);
		ir->buf = NULL;
		wake_up_interruptible(&readwq);
		return;
	}
	err = usb_submit_urb(ir->urb,GFP_ATOMIC);
	if (err) {
		ERR("cannot submit interrupt urb: %d",err);
	}

	/* Fix the toggle bit, perhaps. */
	DBG("IN BITs %08X last %d",ir->dev->toggle[0],last);
	if (last == usb_gettoggle(ir->dev,1,0)) {
		DBG("missed a toggle - fix it");
		usb_dotoggle(ir->dev,1,0);
	}
	last = usb_gettoggle(ir->dev,1,0);

	if (ir->buf[1]) return; /* A jack (un)plugged event. */

	/* Fetch the code from memory. */
	if (ctlmsg(ir->dev,GET_MEM,0xA1,0,0,2,fetch_ir_code,ir)) {
		ERR("cannot send control message to get IR code");
	}
}

static int interface_number(struct usb_interface *intf)
{
	struct usb_device *dev = interface_to_usbdev(intf);
	struct usb_host_config *config = dev->actconfig;
	int i;
	for (i = 0; i < USB_MAXINTERFACES; i++)
		if (intf == config->interface[i])
			return(i);
	return(i-1);
}

static int get(struct usb_device *dev,u8 req,u16 wValue,u16 wIndex,
	       char *buf,u16 len)
{
	int cnt = usb_control_msg(dev, usb_rcvctrlpipe(dev,0), req,
				  USB_RECIP_INTERFACE|
				  USB_TYPE_CLASS|USB_DIR_IN,
				  wValue, wIndex, buf, len, HZ);
	if (cnt < 0) ERR("cannot GET wIndex 0x%04hX : %d",wIndex,cnt);
	return(cnt);
}

static int set(struct usb_device *dev,u8 req,u16 wValue,u16 wIndex,
	       char *buf,u16 len)
{
	int cnt = usb_control_msg(dev, usb_sndctrlpipe(dev,0), req,
				  USB_RECIP_INTERFACE|
				  USB_TYPE_CLASS|USB_DIR_OUT,
				  wValue, wIndex, buf, len, HZ);
	if (cnt < 0) ERR("cannot SET wIndex 0x%04hX : %d",wIndex,cnt);
	return(cnt);
}

/* usb driver functions */

static void ir_disconnect(struct usb_interface *intf)
{
	/* This never gets called, since probe() returns null. */
	ERR("disconnect");
	return;
}

static int ir_probe(struct usb_interface *intf,
		    const struct usb_device_id *id)
{
	int err, ifnum = interface_number(intf);
	unsigned int pipe;
	struct usb_device *dev = interface_to_usbdev(intf);
  
	if (0 != ifnum) return(-ENODEV);

	if (remote.present) {
		/* Allow only one device, for now. */
		DBG("already initialized, during probe of interface %d",ifnum);
		return(-ENODEV);
	}
	/* Allocate an URB */
	remote.urb = usb_alloc_urb(0,GFP_ATOMIC);
	if (!remote.urb) {
		return(-ENOMEM);
	}
	remote.buf = kmalloc(8,GFP_ATOMIC);
	if (!remote.buf) {
		usb_free_urb(remote.urb);
		return(-ENOMEM);
	}
	remote.present = 1;
	INFO("initializing driver on probe of interface %d",ifnum);
	remote.dev = dev;

	/* Setup interrupt. */
	pipe = usb_rcvintpipe(dev,1);
	usb_fill_int_urb(remote.urb,dev,pipe,remote.buf,8,
			 handle_interrupt,&remote,1);
	err = usb_submit_urb(remote.urb,GFP_ATOMIC);
	if (err) {
		ERR("cannot submit interrupt urb: %d",err);
		return(-ENODEV);
	}
	/* Don't bother to claim an interface. */
	return(-ENODEV);
}

/* file operations */

static int ir_ioctl(struct inode *node,struct file *file,unsigned int cmd,
		    unsigned long arg)
{
	char mode;
	int err, val;
	struct ir_device *ir = file->private_data;
	struct usb_device *dev = ir->dev;
	switch (cmd) {
	case EXAUDIO_SET_AC3_MODE:
		err = get_user(val,(unsigned long*)arg);
		if (err) return(err);
		if (val < 1) return(-EINVAL);
		else if (val > 7) return(-EINVAL);
		mode = val;
		if (1 != set(dev,SET_CUR,0x0200,0x1100,&mode,1)) return(-EIO);
		break;
	case EXAUDIO_GET_AC3_MODE:
		if (1 != get(dev,GET_CUR,0x0200,0x1100,&mode,1)) return(-EIO);
		val = mode;
		err = put_user(val,(unsigned long*)arg);
		if (err) return(err);
		break;
	case EXAUDIO_SET_SU29:
		err = get_user(val,(unsigned long*)arg);
		if (err) return(err);
		if (val < 1) return(-EINVAL);
		else if (val > 2) return(-EINVAL);
		mode = val;
		if (1 != set(dev,SET_CUR,0x0000,0x1d00,&mode,1)) return(-EIO);
		break;
	case EXAUDIO_GET_SU29:
		if (1 != get(dev,GET_CUR,0x0000,0x1d00,&mode,1)) return(-EIO);
		val = mode;
		err = put_user(val,(unsigned long*)arg);
		if (err) return(err);
		break;
	default:
		return(-ENOTTY);
	}
	return(0);
}

static int ir_open(struct inode *inode, struct file *filp)
{
	if (!remote.present) { /* Requested device not found. */
		return(-ENODEV);
	}
	filp->private_data = &remote;
	return(0);
}

static ssize_t ir_read(struct file *filp,char *buf,size_t count,loff_t *f_pos)
{
	struct ir_device *ir = filp->private_data;
	u8 code;
	u32 result;
	/* Go to sleep until an IR event occurs. */
	interruptible_sleep_on(&readwq);
	if (signal_pending(current)) return(-ERESTARTSYS);
	if (!ir->present) { /* Device has disappeared. */
		return(-ENODEV);
	}
	code = result = ir->irval;
	switch (count) {
	case 1: if (copy_to_user(buf,&code,count))   return(-EFAULT); break;
	case 4: if (copy_to_user(buf,&result,count)) return(-EFAULT); break;
	default: return(-EINVAL);
	}
	*f_pos += count;
	return(count);
}

static int ir_release(struct inode *inode, struct file *file)
{
	return(0);
}
