/*
 * V4L2 USB driver for Fujifilm FinePix cameras
 *
 * Copyright (C) 2004 Frank Zago
 * Copyright (C) 2004 Luc Willems
 *
 * This driver is based on the USB skeleton driver
 * Copyright (C) 2001-2003 Greg Kroah-Hartman (greg@kroah.com)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation, version 2.
 *
 */

#include "finepix.h"

#define DRIVER_DESC "Fujifilm FinePix USB V4L/V4L2 driver"
#define DRIVER_VERSION "0.90.1"	/* also change in KERNEL_VERSION(...) */
#define DRIVER_AUTHOR "Frank Zago"
#define DRIVER_NAME "finepix"

/* Table of supported devices */
static struct usb_device_id fpix_table[] = {
	{FPIX_DEV(USB_FINEPIX_4800_PID, "Fujifilm FinePix 4800")},
	{FPIX_DEV(USB_FINEPIX_A202_PID, "Fujifilm FinePix A202")},
	{FPIX_DEV(USB_FINEPIX_A203_PID, "Fujifilm FinePix A203")},
	{FPIX_DEV(USB_FINEPIX_A204_PID, "Fujifilm FinePix A204")},
	{FPIX_DEV(USB_FINEPIX_A205_PID, "Fujifilm FinePix A205")},
	{FPIX_DEV(USB_FINEPIX_A210_PID, "Fujifilm FinePix A210")},
	{FPIX_DEV(USB_FINEPIX_A303_PID, "Fujifilm FinePix A303")},
	{FPIX_DEV(USB_FINEPIX_A310_PID, "Fujifilm FinePix A310")},
	{FPIX_DEV(USB_FINEPIX_F401_PID, "Fujifilm FinePix F401")},
	{FPIX_DEV(USB_FINEPIX_F402_PID, "Fujifilm FinePix F402")},
	{FPIX_DEV(USB_FINEPIX_F410_PID, "Fujifilm FinePix F410")},
	{FPIX_DEV(USB_FINEPIX_F601_PID, "Fujifilm FinePix F601")},
	{FPIX_DEV(USB_FINEPIX_F700_PID, "Fujifilm FinePix F700")},
	{FPIX_DEV(USB_FINEPIX_M603_PID, "Fujifilm FinePix M603")},
	{FPIX_DEV(USB_FINEPIX_S3000_PID, "Fujifilm FinePix S3000")},
	{FPIX_DEV(USB_FINEPIX_S304_PID, "Fujifilm FinePix S304")},
	{FPIX_DEV(USB_FINEPIX_S5000_PID, "Fujifilm FinePix S5000")},
	{FPIX_DEV(USB_FINEPIX_S602_PID, "Fujifilm FinePix S602")},
	{FPIX_DEV(USB_FINEPIX_S7000_PID, "Fujifilm FinePix S7000")},
	{FPIX_DEV(USB_FINEPIX_X1_PID, "Fujifilm FinePix unknown model")},
	{FPIX_DEV(USB_FINEPIX_X2_PID, "Fujifilm FinePix unknown model")},
	{FPIX_DEV(USB_FINEPIX_X3_PID, "Fujifilm FinePix unknown model")},
	{FPIX_DEV(USB_FINEPIX_F420_PID, "Fujifilm FinePix F420")},

	/* Terminating entry */
	{}
};

MODULE_DEVICE_TABLE(usb, fpix_table);

/* prevent races between open() and disconnect() */
static DECLARE_MUTEX(disconnect_sem);

/*--------------------------------------------------------------------------*/

/* Initialize the device */
static int fpix_initialize_device(struct usb_fpix *dev)
{
	int ret;
	int size_ret;

	/* Reset bulk in endpoint */
	usb_clear_halt(dev->udev, dev->bulk_in_endpointAddr);

	/* Init the device */
	memset(dev->control_buffer, 0, 12);
	dev->control_buffer[0] = 0xc6;
	dev->control_buffer[8] = 0x20;

	ret = usb_control_msg(dev->udev,
			      usb_sndctrlpipe(dev->udev, 0),
			      USB_REQ_GET_STATUS,
			      USB_DIR_OUT | USB_TYPE_CLASS |
			      USB_RECIP_INTERFACE, 0, 0, dev->control_buffer,
			      12, USB_TIMEOUT(HZ));
	if (ret != 12) {
		dev_err(&dev->udev->dev, "usb_control_msg failed (%d)\n", ret);
		return -EIO;
	}

	/* Read the result of the command. Ignore the result, for it varies with the device. */
	ret = usb_bulk_msg(dev->udev,
			   usb_rcvbulkpipe(dev->udev,
					   dev->bulk_in_endpointAddr),
			   dev->control_buffer, FPIX_MAX_TRANSFER, &size_ret,
			   USB_TIMEOUT(HZ));
	if (ret != 0) {
		dev_err(&dev->udev->dev, "usb_bulk_msg failed (%d)\n", ret);
		return -EIO;
	}

	/* Request a frame, but don't read it */
	memset(dev->control_buffer, 0, 12);
	dev->control_buffer[0] = 0xd3;
	dev->control_buffer[7] = 0x01;

	ret = usb_control_msg(dev->udev,
			      usb_sndctrlpipe(dev->udev, 0),
			      USB_REQ_GET_STATUS,
			      USB_DIR_OUT | USB_TYPE_CLASS |
			      USB_RECIP_INTERFACE, 0, 0, dev->control_buffer,
			      12, USB_TIMEOUT(HZ));
	if (ret != 12) {
		dev_err(&dev->udev->dev, "usb_control_msg failed (%d)\n", ret);
		return -EIO;
	}

	/* Again, reset bulk in endpoints */
	usb_clear_halt(dev->udev, dev->bulk_in_endpointAddr);

	return 0;
}

#if 0
/* Release (?) the device */
static int fpix_release_device(struct usb_fpix *dev)
{
	int ret;
	int size_ret;

	/* Prepare and send the command */
	memset(dev->control_buffer, 0, 12);
	dev->control_buffer[0] = 0xc8;
	dev->control_buffer[8] = 0x01;

	ret = usb_control_msg(dev->udev,
			      usb_sndctrlpipe(dev->udev, 0),
			      USB_REQ_GET_STATUS,
			      USB_DIR_OUT | USB_TYPE_CLASS |
			      USB_RECIP_INTERFACE, 0, 0, dev->control_buffer,
			      12, USB_TIMEOUT(HZ));
	if (ret != 12) {
		err("usb_control_msg failed (%d)", ret);
		return -EIO;
	}

	/* Read the result of the command. Ignore the result, for it varies with the device. */
	ret = usb_bulk_msg(dev->udev,
			   usb_rcvbulkpipe(dev->udev,
					   dev->bulk_in_endpointAddr),
			   dev->control_buffer, 0x100, &size_ret, USB_TIMEOUT(HZ));
	if (ret != 0) {
		err("usb_bulk_msg failed (%d)", ret);
		return -EIO;
	}

	return 0;
}
#endif

/*--------------------------------------------------------------------------*/

/**
 *  fpix_delete
 */
static void fpix_delete(struct usb_fpix *dev)
{
	if (dev == NULL)
		return;

	if (dev->control_urb)
		usb_free_urb(dev->control_urb);

	if (dev->bulk_in_urb)
		usb_free_urb(dev->bulk_in_urb);

	if (dev->control_buffer)
		kfree(dev->control_buffer);

#ifdef CONFIG_DEBUG_FS
	debugfs_remove(dev->debug_state);
	debugfs_remove(dev->debug_dir);
#endif

	kfree(dev);
}

/*--------------------------------------------------------------------------*/

static void timeout_kill(unsigned long data)
{
    struct urb  *urb = (struct urb *) data;

    usb_unlink_urb(urb);
}

static void read_frame_part_callback(struct urb *urb, struct pt_regs *regs);

/* Reads part of a frame */
static void read_frame_part(struct usb_fpix *dev)
{
	int ret;
	struct fpix_frame *frame = dev->current_frame;
	unsigned long flags;

	if (dev->state != FPIX_READ_FRAME_1 && dev->state != FPIX_READ_FRAME_2) {
		dev_err(&dev->udev->dev, "invalid state %d, should be %d or %d\n",
		    dev->state, FPIX_READ_FRAME_1, FPIX_READ_FRAME_2);
	}

	/* Reads part of a frame */
	usb_fill_bulk_urb(dev->bulk_in_urb,
			  dev->udev,
			  usb_rcvbulkpipe(dev->udev, dev->bulk_in_endpointAddr),
			  frame->buffer + frame->offset,
			  FPIX_MAX_TRANSFER, read_frame_part_callback, dev);
#if URB_HAS_TIMEOUT
	dev->bulk_in_urb->timeout = FPIX_TIMEOUT;
#endif

	dev->bulk_in_urb->transfer_flags = URB_ASYNC_UNLINK;

	spin_lock_irqsave(&dev->state_lock, flags);

	ret = usb_submit_urb(dev->bulk_in_urb, GFP_ATOMIC);

	if (ret) {
		dev_new_state(FPIX_RESET);
		schedule_delayed_work(&dev->wqe, 1);
		dev_err(&dev->udev->dev, "usb_submit_urb failed with %d\n", ret);
	} else {
		dev_new_state(FPIX_READ_FRAME_2);

		/* This p.o.s. of hardware does sometimes never come back.
		 * Is this masking a bug somewhere else? */
		dev->bulk_timer.expires = jiffies + msecs_to_jiffies(150);
		add_timer(&dev->bulk_timer);
	}

	spin_unlock_irqrestore(&dev->state_lock, flags);
}

static void decodeframe(struct fpix_frame *frame, struct usb_fpix *dev)
{
	int n, i, temp;
	unsigned char *pos;

	if (jpeg_check_size(&frame->buffer[0],
			    FPIX_RGB24_WIDTH, FPIX_RGB24_HEIGHT)) {
		n = jpeg_decode(&frame->buffer[0],
				dev->rgb24buf +
				(frame->index * FPIX_RGB24_FRAME_SIZE),
				FPIX_RGB24_WIDTH, FPIX_RGB24_HEIGHT, 24,
				&frame->decdata);
		if (n) {
			dev_err(&dev->udev->dev, "ooeps decode jpg result : %d\n", n);
		}
	} else {
		dev_err(&dev->udev->dev, "ooops check_size failed\n");
	}

	/* software brightness  & contrast */
	pos = dev->rgb24buf + (frame->index * FPIX_RGB24_FRAME_SIZE);
	for (i = 0; i < FPIX_RGB24_FRAME_SIZE; i++, pos++) {
		temp = *pos;
		//fix brightness
		temp = temp + dev->brightness * 2 - 256;
		//fix contrast
		temp = (temp - 128) * dev->contrast / 128 + 128;
		if (temp < 0)
			temp = 0;
		if (temp > 0xff)
			temp = 0xff;
		*pos = (char)temp;
	}
}

static void read_frame_part_callback(struct urb *urb, struct pt_regs *regs)
{
	struct usb_fpix *dev = urb->context;
	unsigned long flags_state;

	dev_dbg(&dev->udev->dev, "enter read_frame_part_callback\n");

	del_timer_sync(&dev->bulk_timer);

	spin_lock_irqsave(&dev->state_lock, flags_state);

	if (dev->state != FPIX_READ_FRAME_2) {
		dev_err(&dev->udev->dev, "invalid state %d, should be %d\n",
		    dev->state, FPIX_READ_FRAME_2);
	}

	if (urb->status != 0) {

		/* We kill a stuck urb every 50 frames on average, so don't
		 * display a log message for that. */
		if (urb->status != -ECONNRESET)
			dev_err(&dev->udev->dev, "bad URB status %d\n", urb->status);
		dev_new_state(FPIX_RESET);
		schedule_delayed_work(&dev->wqe, 1);
		spin_unlock_irqrestore(&dev->state_lock, flags_state);

	} else {

		struct fpix_frame *frame = dev->current_frame;

		/* Adjust frame pointers */
		frame->offset += urb->actual_length;
		frame->size_left -= urb->actual_length;

		if (frame->size_left == 0) {
			/* Apparently the frame is too big for our buffers. Never
			 * seen that case. */
			dev_err(&dev->udev->dev, "Destination buffer too small for frame. Please, report to driver maintainer.\n");
			dev_new_state(FPIX_READ_FRAME_4);
			schedule_delayed_work(&dev->wqe, 1);
			spin_unlock_irqrestore(&dev->state_lock, flags_state);
		} else if (urb->actual_length < FPIX_MAX_TRANSFER) {
			/* If the result is less than what was asked for, then it's
			 * the end of the frame. */
			unsigned long flags_lists;

			spin_lock_irqsave(&dev->lists_lock, flags_lists);

			list_del(&frame->item);
			list_add_tail(&frame->item, &dev->done_buffers);

			frame->flags =
			    V4L2_BUF_FLAG_MAPPED | V4L2_BUF_FLAG_DONE;

			/* decode JPEG if mode is RGB24 (V4L) */
			if (dev->mode == MODE_V4L1) {
				decodeframe(frame, dev);
			}

			spin_unlock_irqrestore(&dev->lists_lock, flags_lists);

			complete(&dev->frame_done);

			dev_new_state(FPIX_READ_FRAME_4);
			schedule_delayed_work(&dev->wqe, 1);
			spin_unlock_irqrestore(&dev->state_lock, flags_state);
		} else {
			/* got a partial image */
			spin_unlock_irqrestore(&dev->state_lock, flags_state);

			read_frame_part(dev);
		}
	}
}

/* Request frame callback */
static void request_frame_callback(struct urb *urb, struct pt_regs *regs)
{
	struct usb_fpix *dev = urb->context;
	unsigned long flags;

	spin_lock_irqsave(&dev->state_lock, flags);

	if (dev->state != FPIX_REQ_FRAME_2) {
		dev_err(&dev->udev->dev, "invalid state %d, should be %d\n",
		    dev->state, FPIX_REQ_FRAME_2);
	}

	if (urb->status != 0) {

		if (urb->status != -ETIMEDOUT)
			dev_err(&dev->udev->dev, "bad URB status %d\n", urb->status);

		dev_new_state(FPIX_RESET);
		schedule_delayed_work(&dev->wqe, 1);
		spin_unlock_irqrestore(&dev->state_lock, flags);

	} else {
		dev_new_state(FPIX_READ_FRAME_1);
		spin_unlock_irqrestore(&dev->state_lock, flags);

		read_frame_part(dev);
	}
}

/* Request a new frame */
static void request_frame(struct usb_fpix *dev)
{
	int ret;
	struct fpix_frame *frame;
	unsigned long flags;

	if (dev->state != FPIX_REQ_FRAME_1) {
		dev_err(&dev->udev->dev, "invalid state %d, should be %d\n",
		    dev->state, FPIX_REQ_FRAME_1);
	}

	/* Ensure there is a frame queued. */
	spin_lock_irqsave(&dev->lists_lock, flags);

	if (list_empty(&dev->queued_buffers)) {
		/* Skip that frame */
		spin_unlock_irqrestore(&dev->lists_lock, flags);
		schedule_delayed_work(&dev->wqe, NEXT_FRAME_DELAY);
		return;
	} else {
		frame = dev->current_frame =
		    list_entry(dev->queued_buffers.next, struct fpix_frame,
			       item);
	}

	spin_unlock_irqrestore(&dev->lists_lock, flags);

	/* Setup command packet */
	memset(dev->control_buffer, 0, 12);
	dev->control_buffer[0] = 0xd3;
	dev->control_buffer[7] = 0x01;

	/* Request a frame */
	frame->size_left = FPIX_FRAME_SIZE;
	frame->offset = 0;

	dev->ctrlreq.bRequestType =
	    USB_DIR_OUT | USB_TYPE_CLASS | USB_RECIP_INTERFACE;
	dev->ctrlreq.bRequest = USB_REQ_GET_STATUS;
	dev->ctrlreq.wValue = 0;
	dev->ctrlreq.wIndex = 0;
	dev->ctrlreq.wLength = cpu_to_le16(12);
#if URB_HAS_TIMEOUT
	dev->control_urb->timeout = FPIX_TIMEOUT;
#endif

	usb_fill_control_urb(dev->control_urb,
			     dev->udev,
			     usb_sndctrlpipe(dev->udev, 0),
			     (unsigned char *)&dev->ctrlreq,
			     dev->control_buffer,
			     12, request_frame_callback, dev);

	spin_lock_irqsave(&dev->state_lock, flags);

	ret = usb_submit_urb(dev->control_urb, GFP_ATOMIC);
	if (ret) {
		dev_new_state(FPIX_RESET);
		dev_err(&dev->udev->dev, "usb_submit_urb failed with %d\n", ret);
		schedule_delayed_work(&dev->wqe, 1);
	} else {
		dev_new_state(FPIX_REQ_FRAME_2);
	}

	spin_unlock_irqrestore(&dev->state_lock, flags);
}

/*--------------------------------------------------------------------------*/

/* Fallback state machine. */
static void fpix_sm(void *context)
{
	struct usb_fpix *dev = context;
	int ret;

	down(&dev->sem);

	/* verify that the device wasn't unplugged */
	if (!dev->present) {
		dev_err(&dev->udev->dev, "cannot recover - device is gone\n");
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,6)
		/* complete_all is not exported before kernel 2.6.6. */
		complete(&dev->frame_done);
#else
		complete_all(&dev->frame_done);
#endif
		complete(&dev->can_close);
		up(&dev->sem);
		return;
	}

	switch (dev->state) {
	case FPIX_RESET:
		dev_dbg(&dev->udev->dev, "reseting device\n");

		/* Init the device */
		memset(dev->control_buffer, 0, 12);
		dev->control_buffer[0] = 0xc6;
		dev->control_buffer[8] = 0x20;

		ret = usb_control_msg(dev->udev,
				      usb_sndctrlpipe(dev->udev, 0),
				      USB_REQ_GET_STATUS,
				      USB_DIR_OUT | USB_TYPE_CLASS |
				      USB_RECIP_INTERFACE, 0, 0,
				      dev->control_buffer, 12, USB_TIMEOUT(HZ));
		if (ret != 12) {
			dev_err(&dev->udev->dev, "usb_control_msg failed (%d)\n", ret);
		}

		usb_clear_halt(dev->udev, dev->bulk_in_endpointAddr);

		dev_new_state(FPIX_REQ_FRAME_1);
		schedule_delayed_work(&dev->wqe, HZ / 10);
		break;

	case FPIX_READ_FRAME_4:
		/* TODO: let's say the result is good */
		if (list_empty(&dev->queued_buffers)) {
			dev_new_state(FPIX_NOP);

			if (dev->must_stop) {
				complete(&dev->can_close);
			}

		} else {
			dev_new_state(FPIX_REQ_FRAME_1);
			schedule_delayed_work(&dev->wqe, NEXT_FRAME_DELAY);
		}
		break;

	case FPIX_REQ_FRAME_1:
		if (dev->must_stop) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,6)
			/* complete_all is not exported before kernel 2.6.6. */
			complete(&dev->frame_done);
#else
			complete_all(&dev->frame_done);
#endif
			dev_new_state(FPIX_NOP);
			complete(&dev->can_close);
			up(&dev->sem);
			return;
		}

		request_frame(dev);
		break;

	case FPIX_NOP:
	case FPIX_REQ_FRAME_2:
	case FPIX_READ_FRAME_1:
	case FPIX_READ_FRAME_2:
		dev_err(&dev->udev->dev, "invalid state %d\n", dev->state);
	}

	up(&dev->sem);
}

/*--------------------------------------------------------------------------*/

#ifdef DEBUG
static const char *ioctl_name(unsigned int cmd)
{
	switch (cmd) {
	case VIDIOC_QUERYCAP:
		return "VIDIOC_QUERYCAP";

	case VIDIOC_G_STD:
		return "VIDIOC_G_STD";

	case VIDIOC_CROPCAP:
		return "VIDIOC_CROPCAP";

	case VIDIOC_ENUM_FMT:
		return "VIDIOC_ENUM_FMT";

	case VIDIOC_ENUMSTD:
		return "VIDIOC_ENUMSTD";

	case VIDIOC_ENUMINPUT:
		return "VIDIOC_ENUMINPUT";

	case VIDIOC_QUERYCTRL:
		return "VIDIOC_QUERYCTRL";

	case VIDIOC_G_PARM:
		return "VIDIOC_G_PARM";

	case VIDIOC_G_INPUT:
		return "VIDIOC_G_INPUT";

	case VIDIOC_S_FMT:
		return "VIDIOC_S_FMT";

	case VIDIOC_G_FMT:
		return "VIDIOC_G_FMT";

	case VIDIOC_S_INPUT:
		return "VIDIOC_S_INPUT";

	case VIDIOC_S_STD:
		return "VIDIOC_S_STD";

	case VIDIOC_TRY_FMT:
		return "VIDIOC_TRY_FMT";

	case VIDIOC_REQBUFS:
		return "VIDIOC_REQBUFS";

	case VIDIOC_QUERYBUF:
		return "VIDIOC_QUERYBUF";

	case VIDIOC_QBUF:
		return "VIDIOC_QBUF";

	case VIDIOC_DQBUF:
		return "VIDIOC_DQBUF";

	case VIDIOC_STREAMON:
		return "VIDIOC_STREAMON";

	case VIDIOC_STREAMOFF:
		return "VIDIOC_STREAMOFF";

		/*
		 * V4L 
		 */

	case VIDIOCGCAP:
		return "VIDIOCGCAP";

	case VIDIOCGCHAN:
		return "VIDIOCGCHAN";

	case VIDIOCSCHAN:
		return "VIDIOCSCHAN";

	case VIDIOCGPICT:
		return "VIDIOCGPICT";

	case VIDIOCSPICT:
		return "VIDIOCSPICT";

	case VIDIOCGWIN:
		return "VIDIOCGWIN";

	case VIDIOCSWIN:
		return "VIDIOCSWIN";

	case VIDIOCGMBUF:
		return "VIDIOCGMBUF";

	case VIDIOCMCAPTURE:
		return "VIDIOCMCAPTURE";

	case VIDIOCSYNC:
		return "VIDIOCSYNC";

	case VIDIOCCAPTURE:
		return "VIDIOCCAPTURE";

	case VIDIOCGFBUF:
		return "VIDIOCGFBUF";

	case VIDIOCSFBUF:
		return "VIDIOCSFBUF";

	case VIDIOCKEY:
		return "VIDIOCKEY";

	case VIDIOCGTUNER:
		return "VIDIOCGTUNER";

	case VIDIOCSTUNER:
		return "VIDIOCSTUNER";

	case VIDIOCGFREQ:
		return "VIDIOCGFREQ";

	case VIDIOCSFREQ:
		return "VIDIOCSFREQ";

	case VIDIOCGAUDIO:
		return "VIDIOCGAUDIO";

	case VIDIOCSAUDIO:
		return "VIDIOCSAUDIO";

	default:
		return "UNKNOWN";
	}
}
#endif /* FPIX_DEBUG */

/**
 * fpix_do_ioctl
 *
 */
static int fpix_do_ioctl(struct inode *inode, struct file *file,
			 unsigned int cmd, void *arg)
{
	struct usb_fpix *dev = file->private_data;
	int ret;

	dev_dbg(&dev->udev->dev, "ioctl command %s (%x)\n", ioctl_name(cmd), cmd);

	/* The driver supports V4L and V4L2. Always allow VIDIOCGCAP and
	 * VIDIOC_QUERYCAP. Once the application decides between V4L and
	 * V4l2 by using one of the API ioctl, do not call the other set
	 * of ioctls. */

	if (cmd == VIDIOCGCAP) {
		struct video_capability *b = arg;

		strlcpy(b->name, dev->name, sizeof(b->name));
		b->type = VID_TYPE_CAPTURE;
		b->channels = 1;
		b->audios = 0;
		b->maxwidth = FPIX_RGB24_WIDTH;	/* VIDEOSIZE_CIF */
		b->maxheight = FPIX_RGB24_HEIGHT;
		b->minwidth = FPIX_RGB24_WIDTH;	/* VIDEOSIZE_48_48 */
		b->minheight = FPIX_RGB24_HEIGHT;

		ret = 0;
	} else if (cmd == VIDIOC_QUERYCAP) {

		struct v4l2_capability *cap = arg;

		memset(cap, 0, sizeof(struct v4l2_capability));

		strlcpy(cap->driver, DRIVER_NAME, sizeof(cap->driver));
		strlcpy(cap->card, dev->name, sizeof(cap->card));
		memset(cap->bus_info, 0, sizeof(cap->bus_info));
		cap->version = KERNEL_VERSION(0, 90, 1);
		cap->capabilities = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_READWRITE;

		ret = 0;
	} else if (_IOC_TYPE(cmd) == 'v') {
		/* V4L ioctl command */
		if (dev->mode == MODE_V4L1 || dev->mode == MODE_UNKNOWN) {

			dev->mode = MODE_V4L1;
			ret = fpix_do_ioctl_v4l(dev, inode, file, cmd, arg);
		} else {
			ret = -EINVAL;
		}
	} else if (_IOC_TYPE(cmd) == 'V') {
		/* V4L2 ioctl command */
		if (dev->mode == MODE_V4L2 || dev->mode == MODE_UNKNOWN) {

			dev->mode = MODE_V4L2;
			ret = fpix_do_ioctl_v4l2(dev, inode, file, cmd, arg);
		} else {
			ret = -EINVAL;
		}
	} else {
		/* Let's say everything else is not supported. */
		ret = -EINVAL;
	}

	if (ret != 0) {
		dev_dbg(&dev->udev->dev, "ioctl returns %s:[%x] %d\n", ioctl_name(cmd), cmd, ret);
	}

	return ret;
}

/**
 * fpix_do_ioctl
 *
 */
static int fpix_ioctl(struct inode *inode, struct file *file,
		      unsigned int cmd, unsigned long arg)
{
	return video_usercopy(inode, file, cmd, arg, fpix_do_ioctl);
}

/**
 * fpix_mmap
 */
static int fpix_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct usb_fpix *dev = file->private_data;

	switch (dev->mode) {
	case MODE_V4L2:
		return fpix_mmap_jpeg(file, vma);
	case MODE_V4L1:
		return fpix_mmap_rgb(file, vma);
	default:
		return -EIO;
	}
	return -EIO;
}

/* Free the resources allocated by open. */
static void deallocate_res(struct usb_fpix *dev)
{
	struct page *page;
	int i;

	for (i = 0; i < FPIX_MAX_FRAMES; i++) {

		struct fpix_frame *frame = &dev->frame[i];

		if (frame->buffer) {
			memset(frame->buffer, 0, FPIX_FRAME_SIZE);

			for (page = virt_to_page(frame->buffer);
			     page <
			     virt_to_page(frame->buffer + FPIX_FRAME_SIZE);
			     page++) {
				ClearPageReserved(page);
			}

			kfree(frame->buffer);
			frame->buffer = NULL;
		}
	}

	/*  v4l resources */
	if (dev->rgb24buf) {
		rvfree(dev->rgb24buf, FPIX_RGB24_BUFFER_SIZE);
		dev->rgb24buf = NULL;
	}
}

/* Allocate the ressources for open. */
static int allocate_res(struct usb_fpix *dev)
{
	int i;
	int ret;
	struct page *page;

	memset(dev->frame, 0, sizeof(dev->frame));

	/* Allocate the buffers to hold the frames. */
	for (i = 0; i < FPIX_MAX_FRAMES; i++) {

		struct fpix_frame *frame = &dev->frame[i];

		frame->index = i;

		frame->buffer = kmalloc(FPIX_FRAME_SIZE, GFP_KERNEL);

		if (frame->buffer == NULL) {
			dev_err(&dev->udev->dev, "Couldn't allocate a buffer for a frame\n");

			deallocate_res(dev);
			ret = -ENOMEM;
			goto done;
		}

		/* Since this memory is going to be user mmaped, erase its
		 * current content. */
		memset(frame->buffer, 0, FPIX_FRAME_SIZE);

		/* Mark it as reserved, else user-space will not see it. */
		for (page = virt_to_page(frame->buffer);
		     page < virt_to_page(frame->buffer + FPIX_FRAME_SIZE);
		     page++) {
			SetPageReserved(page);
		}

		/* Link the frame to the free list */
		list_add_tail(&frame->item, &dev->free_buffers);
	}

	/*  v4l resources */
	//allocation buffer for jpeg decoder
	//dbg("allocate %d frames = %d bytes for rgb24 buffer (",FPIX_RGB24_MAX_FRAMES,FPIX_RGB24_BUFFER_SIZE);
	dev->rgb24buf = rvmalloc(FPIX_RGB24_BUFFER_SIZE);
	if (dev->rgb24buf == NULL) {
		dev_err(&dev->udev->dev, "Couldn't allocate a buffer for a RGB24 buffer\n");
		deallocate_res(dev);
		ret = -ENOMEM;
		goto done;
	}
	ret = 0;

      done:
	if (ret)
		deallocate_res(dev);

	return ret;
}

/**
 *  fpix_open
 */
static int fpix_open(struct inode *inode, struct file *file)
{
	struct video_device *vdev = video_devdata(file);
	struct usb_fpix *dev = container_of(vdev, struct usb_fpix, vdev);
	int subminor;
	int retval;

	subminor = iminor(inode);

	/* prevent disconnects */
	down(&disconnect_sem);

	/* lock this device */
	down(&dev->sem);

	if (dev->open) {
		/* Only one process can open it at any given time. */
		up(&dev->sem);
		retval = -EBUSY;
		goto exit_no_device;
	}

	/* increment our usage count for the driver */
	++dev->open;

	dev->nonblocking = (file->f_flags & O_NONBLOCK);

	/* save our object in the file's private structure */
	file->private_data = dev;

	/* Initialize the device structure */
	INIT_WORK(&dev->wqe, fpix_sm, dev);
	init_completion(&dev->frame_done);
	dev->must_stop = 0;
	dev->mode = MODE_UNKNOWN;

	INIT_LIST_HEAD(&dev->free_buffers);
	INIT_LIST_HEAD(&dev->queued_buffers);
	INIT_LIST_HEAD(&dev->done_buffers);

	spin_lock_init(&dev->lists_lock);
	spin_lock_init(&dev->state_lock);

	retval = fpix_initialize_device(dev);
	if (retval) {
		dev->open--;
		up(&dev->sem);
		dev_err(&dev->udev->dev, "fpix_initialize_device failed with %d.\n", retval);
		goto exit_no_device;
	}

	retval = allocate_res(dev);
	if (retval) {
		dev_err(&dev->udev->dev, "couldn't allocate ressources!\n");

		dev->open--;
		up(&dev->sem);
		goto exit_no_device;
	}

	dev->brightness = 128;
	dev->contrast = 128;

	/* unlock this device */
	up(&dev->sem);

	retval = 0;

      exit_no_device:
	up(&disconnect_sem);
	return retval;
}

/**
 *  fpix_release
 */
static int fpix_release(struct inode *inode, struct file *file)
{
	struct usb_fpix *dev;
	int retval = 0;

	dev = (struct usb_fpix *)file->private_data;
	if (dev == NULL) {
		dev_err(&dev->udev->dev, "%s - object is NULL\n", __FUNCTION__);
		return -ENODEV;
	}

	/* Stop the camera if the application didn't do it */
	stop_streaming(dev);

	/* fpix_release_device(dev); */

	/* lock our device */
	down(&dev->sem);

	if (dev->open <= 0) {
		dev_err(&dev->udev->dev, "%s - device not opened\n", __FUNCTION__);
		retval = -ENODEV;
		goto exit_not_opened;
	}

	deallocate_res(dev);

	--dev->open;

	if (!dev->present && !dev->open) {
		/* the device was unplugged before the file was released */
		up(&dev->sem);
		fpix_delete(dev);
		return 0;
	}

      exit_not_opened:
	up(&dev->sem);

	return retval;
}

#if 0
static ssize_t fpix_read(struct tty_struct *tty, struct file *file,
						 unsigned char __user *buf, size_t count)
{
	struct video_device *vdev = video_devdata(file);
	struct usb_fpix *dev = container_of(vdev, struct usb_fpix, vdev);

	if (dev->mmapped) return -EINVAL;

	if (list_empty(dev->done_buffers.next)) {
		start_streaming(dev);
	} else {
		
	}
}
#endif

/*--------------------------------------------------------------------------*/

static struct file_operations fpix_fops = {
	.owner = THIS_MODULE,

	.ioctl = fpix_ioctl,
	.open = fpix_open,
	.release = fpix_release,
	.mmap = fpix_mmap,
	//.read = fpix_read
};

/* TODO: what's up with that function? */
static void fpix_vdev_release(struct video_device *vdev)
{
	struct usb_fpix *dev = container_of(vdev, struct usb_fpix, vdev);

	dev_dbg(&dev->udev->dev, "fpix_vdev_release - TODO\n");
}

/**
 *  fpix_probe
 *
 *  Called by the usb core when a new device is connected that it thinks
 *  this driver might be interested in.
 */
static int fpix_probe(struct usb_interface *interface,
		      const struct usb_device_id *id)
{
	struct usb_fpix *dev;
	struct usb_device *udev = interface_to_usbdev(interface);
	int retval = -ENOMEM;
	struct usb_host_interface *iface_desc;
	int i;
	char buf[64];

	dev_dbg(&udev->dev, "fpix_probe for device %04x:%04x\n",
	    udev->descriptor.idVendor, udev->descriptor.idProduct);

	/* allocate memory for our device state and initialize it */
	dev = kmalloc(sizeof(struct usb_fpix), GFP_KERNEL);
	if (dev == NULL) {
		dev_err(&dev->udev->dev, "Out of memory\n");
		return -ENOMEM;
	}
	memset(dev, 0x00, sizeof(*dev));

	init_MUTEX(&dev->sem);
	dev->udev = udev;
	dev->name = (void *)id->driver_info;
	
	dev->control_buffer = kmalloc(FPIX_MAX_TRANSFER, GFP_KERNEL);
	if (dev->control_buffer == NULL) {
		dev_err(&dev->udev->dev, "Out of memory\n");
		retval = -ENOMEM;
		goto error;
	}

	/* set up the endpoint information */
	/* check out the endpoints */
	/* use only the first bulk-in, bulk-out and bulk-int endpoints */
	iface_desc = &interface->altsetting[0];

	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		struct usb_endpoint_descriptor *endpoint =
		    &iface_desc->endpoint[i].desc;
		if (!dev->bulk_in_endpointAddr
		    && (endpoint->bEndpointAddress & USB_DIR_IN)
		    && ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) ==
			USB_ENDPOINT_XFER_BULK)) {

			dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;

			dev->bulk_in_urb = usb_alloc_urb(0, GFP_KERNEL);
			if (!dev->bulk_in_urb) {
				dev_err(&dev->udev->dev, "No free urbs available\n");
				goto error;
			}
		}
	}

	if (dev->bulk_in_endpointAddr == 0) {
		dev_err(&dev->udev->dev, "Couldn't find bulk-in endpoint\n");
		goto error;
	}

	init_timer(&dev->bulk_timer);
	dev->bulk_timer.data = (unsigned long)dev->bulk_in_urb;
	dev->bulk_timer.function = timeout_kill;

	dev->control_urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!dev->control_urb) {
		dev_err(&dev->udev->dev, "No free urbs available\n");
		goto error;
	}

	/* allow device read, write and ioctl */
	dev->present = 1;

	/* we can register the device now, as it is ready */
	usb_set_intfdata(interface, dev);

	dev->vdev.owner = THIS_MODULE;
	strlcpy(dev->vdev.name, dev->name, sizeof(dev->vdev.name));
	dev->vdev.type = VID_TYPE_CAPTURE;
	dev->vdev.type2 = V4L2_CAP_VIDEO_CAPTURE;
	dev->vdev.hardware = VID_HARDWARE_FPIX;
	dev->vdev.fops = &fpix_fops;
	dev->vdev.release = fpix_vdev_release;

	if (video_register_device(&dev->vdev, VFL_TYPE_GRABBER, -1) == -1) {
		dev_err(&dev->udev->dev, "video_register_device failed\n");
		goto error;
	}

	/* let the user know what node this device is now attached to */
	dev_info(&dev->udev->dev, "registered new video device: video%d (%s)\n", dev->vdev.minor,
	     dev->vdev.name);

#ifdef CONFIG_DEBUG_FS
	/* Create a sub-directory per device. The key is unique at any
	 * given time but may be reused. */
	sprintf(buf, "finepix-%p", dev);
	dev->debug_dir = debugfs_create_dir(buf, NULL);
	dev->debug_state = debugfs_create_u32("state", 0444, dev->debug_dir, &dev->state);
#endif

	return 0;

 error:
	fpix_delete(dev);
	return retval;
}

/**
 *  fpix_disconnect
 *
 *  Called by the usb core when the device is removed from the system.
 *
 *  This routine guarantees that the driver will not submit any more urbs
 *  by clearing dev->udev.  It is also supposed to terminate any currently
 *  active urbs.  Unfortunately, usb_bulk_msg(), used in fpix_read(), does
 *  not provide any way to do this.  But at least we can cancel an active
 *  write.
 */
static void fpix_disconnect(struct usb_interface *interface)
{
	struct usb_fpix *dev;
	int minor;

	/* prevent races with open() */
	down(&disconnect_sem);

	dev = usb_get_intfdata(interface);
	usb_set_intfdata(interface, NULL);

	down(&dev->sem);

	minor = dev->vdev.minor;

	video_unregister_device(&dev->vdev);

	/* prevent device read, write and ioctl */
	dev->present = 0;

	up(&dev->sem);

	/* if the device is opened, fpix_release will clean this up */
	if (!dev->open) {
		fpix_delete(dev);
	}

	up(&disconnect_sem);

	dev_info(&dev->udev->dev, "USB FinePix #%d now disconnected\n", minor);
}

static struct usb_driver fpix_driver = {
//	.owner = THIS_MODULE,
	.name = "finepix",
	.probe = fpix_probe,
	.disconnect = fpix_disconnect,
	.id_table = fpix_table
};

static int __init fpix_init(void)
{
	int result;

	/* register this driver with the USB subsystem */
	result = usb_register(&fpix_driver);
	if (result) {
		printk(KERN_ERR "usb_register failed. Error number %d\n", result);
		return result;
	}

	printk(KERN_INFO DRIVER_DESC " " DRIVER_VERSION "\n");
	return 0;
}

static void __exit fpix_exit(void)
{
	/* deregister this driver with the USB subsystem */
	usb_deregister(&fpix_driver);
}

module_init(fpix_init);
module_exit(fpix_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
