/*
 * V4L USB driver for Fujifilm FinePix cameras
 *
 * Copyright (C) 2004 Frank Zago
 * Copyright (C) 2004 Luc Willems
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation, version 2.
 *
 */

#include "finepix.h"

int fpix_do_ioctl_v4l(struct usb_fpix *dev,
		      struct inode *inode, struct file *file,
		      unsigned int cmd, void *arg)
{
	int ret;

      dqbuf_again:

	if (dev->present == 0) {
		/* TODO: correct return code? */
		return -ENODEV;
	}

	switch (cmd) {

		/* get/set video source - we are a camera and nothing else */
	case VIDIOCGCHAN:
		{
			struct video_channel *v = arg;
			if (v->channel != 0) {
				ret = -EINVAL;
				break;
			}
			v->channel = 0;
			strcpy(v->name, dev->name);
			v->tuners = 0;
			v->flags = 0;
			v->type = VIDEO_TYPE_CAMERA;
			v->norm = 0;

			ret = 0;
			break;
		}
	case VIDIOCSCHAN:
		{
			ret = 0;
			break;
		}
		/* image properties */
	case VIDIOCGPICT:
		{
			struct video_picture *vp = arg;
			vp->hue = 0;
			vp->colour = 0;
			vp->whiteness = 0;
			/* set software brightness and contrast */
			vp->brightness = dev->brightness << 8;
			vp->contrast = dev->contrast << 8;
			vp->depth = 24;
			vp->palette = VIDEO_PALETTE_RGB24;

			ret = 0;
			break;
		}
	case VIDIOCSPICT:
		{
			struct video_picture *vp = arg;
			if (vp->depth != 24
			    || (vp->palette != VIDEO_PALETTE_RGB24))
				ret = -EINVAL;
			else {
				/* set software brightness and contrast */
				dev->brightness = vp->brightness >> 8;
				dev->contrast = vp->contrast >> 8;
				ret = 0;
			}
			break;
		}
		/* get/set capture window */
	case VIDIOCGWIN:
		{
			struct video_window *vw = arg;
			vw->x = 0;
			vw->y = 0;
			vw->width = FPIX_RGB24_WIDTH;
			vw->height = FPIX_RGB24_HEIGHT;
			vw->chromakey = 0;
			vw->flags = 0;
			vw->clips = NULL;
			vw->clipcount = 0;

			ret = 0;
			break;
		}
	case VIDIOCSWIN:
		{
			struct video_window *vw = (struct video_window *)arg;
			if (vw->width != FPIX_RGB24_WIDTH
			    || vw->height != FPIX_RGB24_HEIGHT)
				ret = -EFAULT;
			else
				ret = 0;
			break;
		}

		/* mmap interface */
	case VIDIOCGMBUF:
		{
			struct video_mbuf *vm = arg;
			int i;
			vm->size = FPIX_RGB24_BUFFER_SIZE;
			vm->frames = FPIX_RGB24_MAX_FRAMES;
			for (i = 0; i < FPIX_RGB24_MAX_FRAMES; i++)
				vm->offsets[i] = FPIX_RGB24_FRAME_SIZE * i;

			ret = 0;
			break;
		}
	case VIDIOCMCAPTURE:
		{
			struct video_mmap *vm = arg;

			if (vm->frame < FPIX_RGB24_MAX_FRAMES) {				
				ret = queue_frame(dev, vm->frame);
			} else {
				dev_err(&dev->udev->dev, "request frame %d not supported\n", vm->frame);
				ret = -EINVAL;
			}
			break;
		}
	case VIDIOCSYNC:
		{
			unsigned long flags;

			spin_lock_irqsave(&dev->lists_lock, flags);

			if (list_empty(&dev->done_buffers)) {
				if (dev->nonblocking || dev->state == FPIX_NOP) {
					ret = -EAGAIN;
				} else {
					spin_unlock_irqrestore(&dev->lists_lock,
							       flags);
					wait_for_completion(&dev->frame_done);
					goto dqbuf_again;
				}
			} else {
				/* Remove the oldest frame */
				struct fpix_frame *frame =
				    list_entry(dev->done_buffers.next,
					       struct fpix_frame, item);

				list_del(dev->done_buffers.next);
				list_add_tail(&frame->item, &dev->free_buffers);

				frame->flags = V4L2_BUF_FLAG_MAPPED;

				ret = 0;
			}

			spin_unlock_irqrestore(&dev->lists_lock, flags);
			ret = 0;
			break;
		}

		/* pointless to implement overlay with this camera */
	case VIDIOCCAPTURE:
	case VIDIOCGFBUF:
	case VIDIOCSFBUF:
	case VIDIOCKEY:
		ret = -EINVAL;
		break;

		/* tuner interface - we have none */
	case VIDIOCGTUNER:
	case VIDIOCSTUNER:
	case VIDIOCGFREQ:
	case VIDIOCSFREQ:
		ret = -EINVAL;
		break;

		/* audio interface - we have none */
	case VIDIOCGAUDIO:
	case VIDIOCSAUDIO:
		//ret = -EINVAL;
		ret = 0;
		break;

	default:
		/* Let's say everything else is not supported. */
		ret = -EINVAL;
	}

	return ret;
}
