/*
 * V4L2 USB driver for Fujifilm cameras
 * This driver potentially support most of the FinePix USB cameras family
 *
 * Copyright (C) 2004   Frank Zago
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation, version 2.
 *
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/usb.h>
#include <linux/videodev.h>
#include <media/v4l2-common.h>
#include <linux/device.h>

#ifdef CONFIG_DEBUG_FS
#include <linux/debugfs.h>
#endif

#include <asm/uaccess.h>
#include <asm/io.h>

#include "finepix-jpeg.h"

/* V4L                  */
#define FPIX_RGB24_MAX_FRAMES 4
#define FPIX_RGB24_WIDTH 320
#define FPIX_RGB24_HEIGHT 240
#define FPIX_RGB24_FRAME_SIZE FPIX_RGB24_WIDTH * FPIX_RGB24_HEIGHT * 3
#define FPIX_RGB24_BUFFER_SIZE FPIX_RGB24_FRAME_SIZE * FPIX_RGB24_MAX_FRAMES

/* Mode in which the driver is opened */
#define MODE_UNKNOWN  0
#define MODE_V4L1     1
#define MODE_V4L2     2

/* IDs of cameras the driver supports. Some different cameras have the
 * same USB ids, so we just keep one here.  */
#define USB_FUJIFILM_VENDOR_ID  0x04cb

#define USB_FINEPIX_4800_PID	0x0104
#define USB_FINEPIX_F601_PID	0x0109
#define USB_FINEPIX_S602_PID	0x010b
#define USB_FINEPIX_F402_PID	0x010f
#define USB_FINEPIX_M603_PID	0x0111
#define USB_FINEPIX_A202_PID	0x0113
#define USB_FINEPIX_F401_PID	0x0115
#define USB_FINEPIX_A203_PID	0x0117
#define USB_FINEPIX_A303_PID	0x0119
#define USB_FINEPIX_S304_PID	0x011b
#define USB_FINEPIX_A204_PID	0x011d
#define USB_FINEPIX_F700_PID	0x0121
#define USB_FINEPIX_F410_PID	0x0123
#define USB_FINEPIX_A310_PID	0x0125
#define USB_FINEPIX_A210_PID	0x0127
#define USB_FINEPIX_A205_PID	0x0129
#define USB_FINEPIX_X1_PID		0x012B
#define USB_FINEPIX_S7000_PID	0x012d
#define USB_FINEPIX_X2_PID		0x012F
#define USB_FINEPIX_S5000_PID	0x0131
#define USB_FINEPIX_X3_PID		0x013B
#define USB_FINEPIX_S3000_PID	0x013d
#define USB_FINEPIX_F420_PID	0x013f


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
/* Default timeout. Should this be overriden by what the device says? */
#define FPIX_TIMEOUT (HZ / 10)
#endif

/* V4L2 hardware index. TODO- get an official one in videodev.h. */
#define VID_HARDWARE_FPIX  37	/* Fujifilm FinePix in PC-CAM mode */

/* Maximum number of frames to keep in memory. Each frame reserves
 * FPIX_MAX_TRANSFER of kmalloc'ed memory, so keep as low as
 * possible. */
#define FPIX_MAX_FRAMES 4

/* Maximum transfer size to use. The windows driver reads by chunks of
 * 0x2000 bytes, so do the same. Note: reading more seems to work
 * too. */
#define FPIX_MAX_TRANSFER 0x2000

/* Maximum size of a buffer that holds a frame. 16KB is sometimes not
 * enough. */
#define FPIX_FRAME_SIZE (3*FPIX_MAX_TRANSFER)

/* Holds a frame informations */
struct fpix_frame {
	unsigned char *buffer;	/* The buffer to receive data. This buffer
							 * must be DMA capable (ie not
							 * vmalloc'ed). */
	size_t size;		/* total the size of the receive buffer */

	size_t size_left;
	size_t offset;

	/* List the buffer in their respective queue (free, queued, done) */
	struct list_head item;

	/* V4L2 buffer flags (V4L2_BUF_FLAG_xxx) */
	u32 index;
	u32 flags;

	/* V4l : jpeg decoder data */
	struct jpeg_decdata decdata;
};

/* Structure to hold all of our device specific stuff */
struct usb_fpix {

	/* Pointer to the name stored in fpix_table */
	char *name;

	/* 
	 * USB stuff 
	 */

	struct usb_ctrlrequest ctrlreq;
	struct urb *control_urb;
	unsigned char *control_buffer;

	__u8 bulk_in_endpointAddr;	/* the address of the bulk in endpoint */
	struct urb *bulk_in_urb;
    struct timer_list  bulk_timer;

	enum {
		FPIX_NOP,	/* inactive, else streaming */
		FPIX_RESET,	/* must reset */
		FPIX_REQ_FRAME_1,	/* init frame request */
		FPIX_REQ_FRAME_2,	/* wait for frame request completion */
		FPIX_READ_FRAME_1,	/* int posted, must read frame */
		FPIX_READ_FRAME_2,	/* reading frame */
		FPIX_READ_FRAME_4	/* read complete */
	} state;
	spinlock_t state_lock;	/* protects the state */

	int must_stop;		/* must stop streaming and go back to the FPIX_NOP state */

	/* 
	 * Video stuff
	 */
	struct video_device vdev;	/* V4L2 device handler */

	/* List of buffers. Queue at tail, remove at head. */
	struct list_head free_buffers;
	struct list_head queued_buffers;
	struct list_head done_buffers;

	spinlock_t lists_lock;

	struct fpix_frame frame[FPIX_MAX_FRAMES];
	struct fpix_frame *current_frame;

	/*
	 * Driver stuff 
	 */
	struct usb_device *udev;	/* save off the usb device pointer */
	int open;					/* if the port is open or not */
	int present;			   /* if the device is not disconnected */
	struct semaphore sem;		/* locks this structure */

	struct work_struct wqe;

	struct completion can_close;
	struct completion frame_done;

	int nonblocking;	/* open can be blocking or non blocking */

	unsigned int mmapped :1;	/* buffers are mapped. */

#ifdef CONFIG_DEBUG_FS
	struct dentry *debug_dir;
	struct dentry *debug_state;
#endif

	/* V4L entries */
	unsigned char *rgb24buf;	/* buffer to keep jpeg decoded data */
	int mode;
	int brightness;
	int contrast;

};

/* Delay after which claim the next frame. If the delay is too small,
 * the camera will return old frames. On the 4800Z, 20ms is bad, 25ms
 * will fail every 4 or 5 frames, but 30ms is perfect. */
#define NEXT_FRAME_DELAY  (((HZ * 30) + 999) / 1000)

/*
 * FPIX_DEV
 * Defines our specific usb device
 *
 * This macro is used to create a struct usb_device_id that matches a
 * FujiFilm FinePix digital camera in PC-CAM mode (i.e. webcam).
 */
#define FPIX_DEV(prod,name) \
    .match_flags = USB_DEVICE_ID_MATCH_DEVICE | USB_DEVICE_ID_MATCH_INT_CLASS, .idVendor = USB_FUJIFILM_VENDOR_ID, .idProduct = (prod), .bInterfaceClass = 255, .driver_info=(unsigned long)name

#define dev_new_state(new_state) { \
	dev_dbg(&dev->udev->dev, "new state from %d to %d at %s:%d\n", dev->state, new_state, __func__, __LINE__); \
	dev->state = new_state; \
}

/* 
 * Prototypes 
 */

int fpix_do_ioctl_v4l(struct usb_fpix *dev, struct inode *inode,
		      struct file *file, unsigned int cmd, void *arg);
int fpix_do_ioctl_v4l2(struct usb_fpix *dev, struct inode *inode,
		       struct file *file, unsigned int cmd, void *arg);
int fpix_mmap_jpeg(struct file *file, struct vm_area_struct *vma);
int fpix_mmap_rgb(struct file *file, struct vm_area_struct *vma);
void start_streaming(struct usb_fpix *dev);
void stop_streaming(struct usb_fpix *dev);
void rvfree(void *mem, unsigned long size);
void *rvmalloc(unsigned long size);
int queue_frame(struct usb_fpix *dev, int frame_num);

/*
 * Emulation layer for various kernels.
 * 2.6 sucks for out of tree drivers :(
 */

/* 2.6.9: The timeout field in urb was removed. */
#define URB_HAS_TIMEOUT ((LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)) && !(defined (CONFIG_SUSE_KERNEL) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,8))))

/* 2.6.11: remap_page_range is being replaced by remap_pfn_range */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
#define remap_page_range(a,b,c,d,e) remap_pfn_range(a,b,(c)>>PAGE_SHIFT,d,e)
#endif

/* 2.6.12: The timeout value passed to usb_bulk_msg() and
 * usb_control_msg() is now expressed in milliseconds instead of
 * jiffies. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
#define USB_TIMEOUT(j) (((j)/HZ) * 1000)
#else
#define USB_TIMEOUT(j) (j)
#endif

/* 2.6.14: URB_ASYNC_UNLINK is gone */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
#define URB_ASYNC_UNLINK 0
#endif


