/*
 * fan.c -- fan status/control
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Written by Sos Pter <sp@osb.hu>, 2002-2004
 */


#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/proc_fs.h>

#include <asm/system.h>
#include <asm/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#ifdef OMNIBOOK_STANDALONE
#include "omnibook.h"
#else
#include <linux/omnibook.h>
#endif

#include "ec.h"
#include "util.h"

static struct proc_dir_entry *proc_fan;

int omnibook_get_fan(void)
{
	u8 fan;
	int retval;
	
	switch (omnibook_ectype) {
	case XE3GF:
	case TSP10:
		if ((retval = omnibook_ec_read(XE3GF_FSRD, &fan)))
			return retval;
		retval = fan;
		break;
	case OB500:
		if ((retval = omnibook_io_read(OB500_GPO1, &fan)))
			return retval;
		retval = (fan & OB500_FAN_OFF_MASK) ? 0 : 1;
		break;
	case OB510:
		if ((retval = omnibook_io_read(OB510_GPIO, &fan)))
			return retval;
		retval = (fan & OB510_FAN_OFF_MASK) ? 0 : 1;
		break;
	case OB6000:
	case OB6100:
		if ((retval = omnibook_ec_read(OB6000_STA1, &fan)))
			return retval;
		retval = (fan & OB6000_FAN_MASK) ? 1 : 0;
		break;
	case OB4150:
	case AMILOD:
		if ((retval = omnibook_ec_read(OB4150_STA1, &fan)))
			return retval;
		retval = (fan & OB4150_FAN_MASK) ? 1 : 0;
		break;
	case XE2:
		if ((retval = omnibook_io_read(OB500_GPO1, &fan)))
			return retval;
		retval = (fan & XE2_FAN_MASK) ? 0 : 1;
		break;
	default:
		printk(KERN_INFO "%s: Fan status monitoring is unsupported on this machie.\n", OMNIBOOK_MODULE_NAME);
		retval = -ENODEV;
	}

	return retval;
}
	
int omnibook_fan_on(void)
{
	u8 fan;
	int retval;

	switch (omnibook_ectype) {
	case XE3GF:
	case TSP10:
		if ((retval = omnibook_ec_read(XE3GF_FSRD, &fan)))
			return retval;
		if ((retval = omnibook_ec_write(XE3GF_FSRD, fan | XE3GF_FAN_ON_MASK)))
			return retval;
		break;
	case OB500:
		if ((retval = omnibook_io_read(OB500_GPO1, &fan)))
			return retval;
		if ((retval = omnibook_io_write(OB500_GPO1, fan & ~OB500_FAN_ON_MASK)))
			return retval;
		break;
	case OB510:
		if ((retval = omnibook_io_read(OB510_GPIO, &fan)))
			return retval;
		if ((retval = omnibook_io_write(OB510_GPIO, fan & ~OB510_FAN_ON_MASK)))
			return retval;
		break;
	case OB6000:
	case OB6100:
		if ((retval = omnibook_ec_read(OB6000_STA1, &fan)))
			return retval;
		if ((retval = omnibook_ec_write(OB6000_STA1, fan | OB6000_FAN_MASK)))
			return retval;
		break;
	case OB4150:
	case AMILOD:
		if ((retval = omnibook_ec_read(OB4150_STA1, &fan)))
			return retval;
		if ((retval = omnibook_ec_write(OB4150_STA1, fan | OB4150_FAN_MASK)))
			return retval;
		break;
	case XE2:
		if ((retval = omnibook_io_read(OB500_GPO1, &fan)))
			return retval;
		if ((retval = omnibook_io_write(OB500_GPO1, fan & ~XE2_FAN_MASK)))
			return retval;
		break;
	default:
		printk(KERN_INFO "%s: Direct fan control is unsupported on this machie.\n", OMNIBOOK_MODULE_NAME);
		retval = -ENODEV;
	}

	return retval;
}

int omnibook_fan_off(void)
{
	u8 fan;
	int retval;

	switch (omnibook_ectype) {
	case XE3GF:
	case TSP10:
		{
			u8 fot, temp;
			if ((retval = omnibook_ec_read(XE3GF_FSRD, &fan)))
				return retval;

			/* fan is already off */
			if (!fan)
				return 0;

			/* now we set FOT to current temp, then reset to initial value */
			if ((retval = omnibook_ec_read(XE3GF_FOT, &fot)))
				return retval;
			if ((retval = omnibook_ec_read(XE3GF_CTMP, &temp)))
				return retval;
			
			do {
				omnibook_ec_write(XE3GF_FOT, temp);
				mdelay(1);
			} while (omnibook_get_fan() != 0);

			omnibook_ec_write(XE3GF_FOT, fot);
		}
		break;
	case OB500:
		if ((retval = omnibook_io_read(OB500_GPO1, &fan)))
			return retval;
		if ((retval = omnibook_io_write(OB500_GPO1, fan | OB500_FAN_OFF_MASK)))
			return retval;
		break;
	case OB510:
		if ((retval = omnibook_io_read(OB510_GPIO, &fan)))
			return retval;
		if ((retval = omnibook_io_write(OB510_GPIO, fan | OB510_FAN_OFF_MASK)))
			return retval;
		break;
	case OB6000:
	case OB6100:
		if ((retval = omnibook_ec_read(OB6000_STA1, &fan)))
			return retval;
		if ((retval = omnibook_ec_write(OB6000_STA1, fan & ~OB6000_FAN_MASK)))
			return retval;
		break;
	case OB4150:
	case AMILOD:
		if ((retval = omnibook_ec_read(OB4150_STA1, &fan)))
			return retval;
		if ((retval = omnibook_ec_write(OB4150_STA1, fan & ~OB4150_FAN_MASK)))
			return retval;
		break;
	case XE2:
		if ((retval = omnibook_io_read(OB500_GPO1, &fan)))
			return retval;
		if ((retval = omnibook_io_write(OB500_GPO1, fan | XE2_FAN_MASK)))
			return retval;
		break;
	default:
		printk(KERN_INFO "%s: Direct fan control is unsupported on this machie.\n", OMNIBOOK_MODULE_NAME);
		retval = -ENODEV;
	}

	return retval;
}

static int omnibook_fan_status(char *buffer, char **start, off_t off, int count, int *eof, void *data)
{
	int fan;
	char *b = buffer;
	char *str;

	fan = omnibook_get_fan();
	if (fan < 0)
		return fan;
	str = (fan) ? "on" : "off";

	if (fan > 1)
		b += sprintf(b, "Fan is %s (level %d)\n", str, fan);
	else
		b += sprintf(b, "Fan is %s\n", str);

	return omnibook_proc_len(buffer, start, off, count, eof, b);;
}

static int omnibook_fan_set(struct file *file, const char *buffer, unsigned long count, void *data)
{
	char status[1] = {'\0'};

	if (copy_from_user(status, buffer, 1))
		return -EFAULT;
	switch (*status) {
	case '0':
		omnibook_fan_off();
		break;
	case '1':
		omnibook_fan_on();
		break;
	default:
		count = -EINVAL;
	}
	
	return count;
}

int __init omnibook_fan_init(void)
{
	mode_t pmode;

	switch (omnibook_ectype) {
	case XE3GF:
	case OB500:
	case OB510:
	case OB6000:
	case OB6100:
	case TSP10:
		pmode = S_IFREG | S_IWUSR | S_IRUGO;
		if (omnibook_userset)
			pmode = pmode | S_IWUGO;
		proc_fan = create_proc_entry("fan", pmode, omnibook_proc_root);
		if (proc_fan) {
			proc_fan->read_proc = omnibook_fan_status;
			proc_fan->write_proc = omnibook_fan_set;
			printk(KERN_INFO "%s: Fan status monitor and control is enabled.\n", OMNIBOOK_MODULE_NAME);
		}
		break;
	case OB4150:
	case XE2:
	case AMILOD:
		pmode = S_IFREG | S_IRUGO;
		proc_fan = create_proc_read_entry("fan", pmode, omnibook_proc_root, omnibook_fan_status, NULL);
		if (proc_fan) {
			printk(KERN_INFO "%s: Fan status monitor is enabled.\n", OMNIBOOK_MODULE_NAME);
			printk(KERN_INFO "%s: Fan control is unsupported on this machine.\n", OMNIBOOK_MODULE_NAME);
		}
		break;
	default:
		printk(KERN_INFO "%s: Fan status monitor and control is unsupported on this machine.\n", OMNIBOOK_MODULE_NAME);
		return 0;
	}
	if (! proc_fan) {
		printk(KERN_ERR "%s: Unable to create /proc/%s/fan.\n", OMNIBOOK_MODULE_NAME, OMNIBOOK_MODULE_NAME);
		return -ENOENT;
	}
	
	return 0;
}

void __exit omnibook_fan_cleanup(void)
{
	if (proc_fan)
		remove_proc_entry("fan", omnibook_proc_root);
}

EXPORT_SYMBOL(omnibook_get_fan);
EXPORT_SYMBOL(omnibook_fan_on);
EXPORT_SYMBOL(omnibook_fan_off);

/* End of file */
