/*
 * onetouch.c -- code to handling OneTouch buttons
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Written by Sos Pter <sp@osb.hu>, 2002-2004
 */


#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/proc_fs.h>

#include <asm/system.h>
#include <asm/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#ifdef OMNIBOOK_STANDALONE
#include "omnibook.h"
#else
#include <linux/omnibook.h>
#endif

#include "ec.h"
#include "util.h"

static struct proc_dir_entry *proc_onetouch;

/* There is no information about reading OneTouch status */
int omnibook_onetouch_enabled = 0;

int omnibook_onetouch_on(void)
{
	if (omnibook_kbc_command(OMNIBOOK_KBC_CONTROL_CMD, OMNIBOOK_KBC_CMD_ONETOUCH_ENABLE)) {
		printk(KERN_ERR "%s: failed OneTouch enable command.\n", OMNIBOOK_MODULE_NAME);
		return -EIO;
	}
	return 0;
}

int omnibook_onetouch_off(void)
{
	if (omnibook_kbc_command(OMNIBOOK_KBC_CONTROL_CMD, OMNIBOOK_KBC_CMD_ONETOUCH_DISABLE)) {
		printk(KERN_ERR "%s: failed OneTouch disable command.\n", OMNIBOOK_MODULE_NAME);
		return -EIO;
	}
	return 0;
}

int omnibook_onetouch_enable(void)
{
#ifdef CONFIG_VT
	switch (omnibook_ectype) {
	case XE3GF:
	case XE3GC:
	case OB500:
	case OB510:
	case OB6000:
	case OB6100:
	case XE4500:
	case AMILOD:
	case TSP10:
	case TSM30X:
		if (! omnibook_onetouch_enabled) {
			if (omnibook_onetouch_on())
				return -EIO;
			omnibook_onetouch_enabled = 1;
			printk(KERN_INFO "%s: OneTouch buttons (if any) are enabled.\n", OMNIBOOK_MODULE_NAME);
		}
		break;
	default:
		omnibook_onetouch_enabled = 0;
		return -ENODEV;
	}
#endif
	return 0;
}

int omnibook_onetouch_disable(void)
{
#ifdef CONFIG_VT
	switch (omnibook_ectype) {
	case XE3GF:
	case XE3GC:
	case OB500:
	case OB510:
	case OB6000:
	case OB6100:
	case XE4500:
	case AMILOD:
	case TSP10:
	case TSM30X:
		if (omnibook_onetouch_enabled) {
			if (omnibook_onetouch_off()) {
				return -EIO;
			}
			omnibook_onetouch_enabled = 0;
			printk(KERN_INFO "%s: OneTouch buttons (if any) are disabled.\n", OMNIBOOK_MODULE_NAME);
		}
		break;
	default:
		omnibook_onetouch_enabled = 0;
		return -ENODEV;
	}
#endif
	return 0;
}

static int omnibook_onetouch_status(char *buffer, char **start, off_t off, int count, int *eof, void *data)
{
	char *b = buffer;

	b += sprintf(b, "OneTouch buttons are %s\n", (omnibook_onetouch_enabled) ? "enabled" : "disabled");

	return omnibook_proc_len(buffer, start, off, count, eof, b);
}

static int omnibook_onetouch_set(struct file *file, const char *buffer, unsigned long count, void *data)
{
	char status[1] = {'\0'};

	if (copy_from_user(status, buffer, 1))
		return -EFAULT;
	switch (*status) {
	case '0':
		omnibook_onetouch_disable();
		break;
	case '1':
		omnibook_onetouch_enable();
		break;
	default:
		count = -EINVAL;
	}
	return count;
}

int __init omnibook_onetouch_init(void)
{
#ifdef CONFIG_VT
	int retval;
	mode_t pmode;

	switch (omnibook_ectype) {
	case XE3GF:
	case XE3GC:
	case OB500:
	case OB510:
	case OB6000:
	case OB6100:
	case XE4500:
	case AMILOD:
	case TSP10:
	case TSM30X:
		pmode = S_IFREG | S_IWUSR | S_IRUGO;
		if (omnibook_userset)
			pmode = pmode | S_IWUGO;
		proc_onetouch = create_proc_entry("onetouch", pmode, omnibook_proc_root);
		break;
	default:
		printk(KERN_INFO "%s: OneTouch button handling is unsupported on this machine.\n", OMNIBOOK_MODULE_NAME);
		omnibook_onetouch_enabled = 0;
		return 0;
	}
	if (proc_onetouch) {
		proc_onetouch->read_proc = omnibook_onetouch_status;
		proc_onetouch->write_proc = omnibook_onetouch_set;
	} else {
		printk(KERN_ERR "%s: Unable to create /proc/%s/onetouch.\n", OMNIBOOK_MODULE_NAME, OMNIBOOK_MODULE_NAME);
		return -ENOENT;
	}
	if ((retval = omnibook_onetouch_enable()))
		return retval;

	return 0;
#else
	printk(KERN_ERR "%s: Virtual terminal support is not compiled into your kernel.\n", OMNIBOOK_MODULE_NAME);
//	return -ENODEV;
	return 0;
#endif
}

void __exit omnibook_onetouch_cleanup(void)
{
#ifdef CONFIG_VT
	if (proc_onetouch)
		remove_proc_entry("onetouch", omnibook_proc_root);
	
	omnibook_onetouch_disable();
#endif
}

EXPORT_SYMBOL(omnibook_onetouch_enabled);
EXPORT_SYMBOL(omnibook_onetouch_enable);
EXPORT_SYMBOL(omnibook_onetouch_disable);

/* End of file */
