/*
 * touchpad.c -- enable/disable touchpad
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Written by Sos Pter <sp@osb.hu>, 2002-2004
 */


#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/proc_fs.h>

#include <asm/system.h>
#include <asm/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#ifdef OMNIBOOK_STANDALONE
#include "omnibook.h"
#else
#include <linux/omnibook.h>
#endif

#include "ec.h"
#include "util.h"

static struct proc_dir_entry *proc_touchpad;

#ifdef CONFIG_MOUSE_PS2
#define OMNIBOOK_TOUCHPAD
#endif

#ifdef CONFIG_MOUSE_PS2_SYNAPTICS
#define OMNIBOOK_TOUCHPAD
#endif

#ifdef CONFIG_PSMOUSE
#define OMNIBOOK_TOUCHPAD
#endif

#ifdef OMNIBOOK_TOUCHPAD
/* Touchpad is enabled by default */
int omnibook_touchpad_enabled = 1;
#else
int omnibook_touchpad_enabled = 0;
#endif

static int omnibook_touchpad_on(void)
{
	if (omnibook_kbc_command(OMNIBOOK_KBC_CONTROL_CMD, OMNIBOOK_KBC_CMD_TOUCHPAD_ENABLE)) {
		printk(KERN_ERR "%s: failed touchpad enable command.\n", OMNIBOOK_MODULE_NAME);
		return -EIO;
	}
	return 0;
}

int omnibook_touchpad_off(void)
{
	if (omnibook_kbc_command(OMNIBOOK_KBC_CONTROL_CMD, OMNIBOOK_KBC_CMD_TOUCHPAD_DISABLE)) {
		printk(KERN_ERR "%s: failed touchpad disable command.\n", OMNIBOOK_MODULE_NAME);
		return -EIO;
	}
	return 0;
}

int omnibook_touchpad_enable(void)
{
#ifdef OMNIBOOK_TOUCHPAD
	switch (omnibook_ectype) {
	case XE3GF:
	case XE3GC:
	case TSP10:
		if (! omnibook_touchpad_enabled) {
			if (omnibook_touchpad_on())
				return -EIO;
			omnibook_touchpad_enabled = 1;
			printk(KERN_INFO "%s: Touchpad is enabled.\n", OMNIBOOK_MODULE_NAME);
		}
		break;
        /* These models have stickpointer, not touchpad */
	case OB500:
	case OB510:
		omnibook_touchpad_enabled = 0;
		return -ENODEV;
		break;
	default:
		omnibook_touchpad_enabled = 1;
		return -ENODEV;
	}
#endif
	return 0;
}

int omnibook_touchpad_disable(void)
{
#ifdef OMNIBOOK_TOUCHPAD
	switch (omnibook_ectype) {
	case XE3GF:
	case XE3GC:
	case TSP10:
		if (omnibook_touchpad_enabled) {
			if (omnibook_touchpad_off()) {
				return -EIO;
			}
			omnibook_touchpad_enabled = 0;
			printk(KERN_INFO "%s: Touchpad is disabled.\n", OMNIBOOK_MODULE_NAME);
		}
		break;
        /* These models have stickpointer, not touchpad */
	case OB500:
	case OB510:
		omnibook_touchpad_enabled = 0;
		return -ENODEV;
		break;
	default:
		omnibook_touchpad_enabled = 1;
		return -ENODEV;
	}
#endif
	return 0;
}

static int omnibook_touchpad_status(char *buffer, char **start, off_t off, int count, int *eof, void *data)
{
	char *b = buffer;

        b += sprintf(b, "Touchpad is %s\n", (omnibook_touchpad_enabled) ? "enabled" : "disabled");
	
	return omnibook_proc_len(buffer, start, off, count, eof, b);
}

static int omnibook_touchpad_set(struct file *file, const char *buffer, unsigned long count, void *data)
{
	char status[1] = {'\0'};

	if (copy_from_user(status, buffer, 1))
		return -EFAULT;
	switch (*status) {
	case '0':
		omnibook_touchpad_disable();
		break;
	case '1':
		omnibook_touchpad_enable();
		break;
	default:
		count = -EINVAL;
	}
	return count;
}

int __init omnibook_touchpad_init(void)
{
#ifdef OMNIBOOK_TOUCHPAD
	int retval;
	mode_t pmode;
	
	switch (omnibook_ectype) {
	case XE3GF:
	case XE3GC:
	case TSP10:
		pmode = S_IFREG | S_IWUSR | S_IRUGO;
		if (omnibook_userset)
			pmode = pmode | S_IWUGO;
		proc_touchpad = create_proc_entry("touchpad", pmode, omnibook_proc_root);
		break;
	default:
		printk(KERN_INFO "%s: Touchpad handling is unsupported on this machine.\n", OMNIBOOK_MODULE_NAME);
		omnibook_touchpad_enabled = 1;
		return 0;
	}
	if (proc_touchpad) {
		proc_touchpad->read_proc = omnibook_touchpad_status;
		proc_touchpad->write_proc = omnibook_touchpad_set;
	} else {
		printk(KERN_ERR "%s: Unable to create /proc/%s/touchpad.\n", OMNIBOOK_MODULE_NAME, OMNIBOOK_MODULE_NAME);
		return -ENOENT;
	}

	return 0;
#else
	printk(KERN_ERR "%s: PS/2 mouse support is not compiled into your kernel.\n", OMNIBOOK_MODULE_NAME);
//	return -ENODEV;
	return 0;
#endif
}

void __exit omnibook_touchpad_cleanup(void)
{
#ifdef OMNIBOOK_TOUCHPAD
	if (proc_touchpad)
		remove_proc_entry("touchpad", omnibook_proc_root);
	omnibook_touchpad_enable();
#endif
}

EXPORT_SYMBOL(omnibook_touchpad_enabled);
EXPORT_SYMBOL(omnibook_touchpad_enable);
EXPORT_SYMBOL(omnibook_touchpad_disable);

/* End of file */
