/*
 *
 *    Connectix USB QuickCam VC Video Camera driver
 *
 *    (C) Copyright 2001 De Marchi Daniele
 *    (C) Copyright 2004 Terry Mohan
 *
 *    This program is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU General Public License as
 *    published by the Free Software Foundation; either version 2 of the
 *    License, or (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    Jan 2001 This driver develop started on the linux 
 *            kernel 2.4.0.
 *
 */

#include <linux/usb.h>
#include <linux/smp_lock.h>
#include <linux/spinlock.h>

#include "qcamvc.h"

MODULE_AUTHOR("De Marchi Daniele <demarchidaniele@libero.it>");
MODULE_DESCRIPTION("Connectix QuickCam VC(USB) Driver");

/*****************************************************************************/
/*                                                                           */
/* Da portare in qcamvc.h                                                    */
/*                                                                           */
/*****************************************************************************/
/* USS 720 requests type */
#define REQT_GET_DEVICE_ID    0xA1
#define REQT_GET_PORT_STATUS  0xA1
#define REQT_SOFT_RESET       0x23
#define REQT_GET_1284_REG     0xC0
#define REQT_SET_1284_REG     0x40

/* USS 720 bRequest      */
#define BREQ_GET_DEVICE_ID    0x00
#define BREQ_GET_PORT_STATUS  0x01
#define BREQ_SOFT_RESET       0x02
#define BREQ_GET_1284_REG     0x03
#define BREQ_SET_1284_REG     0x04

/* USS 720 registers */
#define SET_USS720_DATA    0x00
#define SET_USS720_STATUS  0x01
#define SET_USS720_CONTROL 0x02
#define SET_USS720_EPPADDR 0x03
#define SET_USS720_EPPDATA 0x04
#define SET_USS720_ECPCMD  0x05
#define SET_USS720_EXTCTRL 0x06
#define SET_USS720_USSCTRL 0x07

#define GET_USS720_STATUS  0x00
#define GET_USS720_CONTROL 0x01
#define GET_USS720_EXTCTRL 0x02
#define GET_USS720_USSCTRL 0x03
#define GET_USS720_DATA    0x04
#define GET_USS720_EPP     0x05
#define GET_USS720_SETUP   0x06

/* USS 720 register 1 STATUS */
#define EPP_TIMEOUT    0x01 /*Timeout 10us during EPP read/write */
#define PLH            0x02
#define RESERVED       0x04
#define NFAULT         0x08
#define SELECT         0x10
#define PERROR         0x20
#define NACK           0x40
#define NBUSY          0x80

/* USS 720 register 2 CONTROL */
#define STROBE         0x01
#define AUTO_FD        0x02
#define NINIT          0x04
#define SELECT_IN      0x08
#define INT_ENABLE     0x10
#define DIRECTION      0x20
#define EPP_MASK       0x40
#define HLH            0x80

/* USS 720 register 6 EXTCTRL */
#define BULK_OUT_EMPTY 0x01
#define BULK_IN_EMPTY  0x02
#define BULK_IN_INT    0x04
#define NFAULT_INT     0x08
#define NACK_INT       0x10
#define MODE_MASK      0xe0
#define ECR_SPP        0x00
#define ECR_PS2        0x20
#define ECR_PPF        0x40
#define ECR_ECP        0x60
#define ECR_EPP        0x80

/* USS 720 register 7 USSCTRL */
#define AUTO_MODE         0x01
#define COMPRESS_ENABLE   0x02
#define NFAULT_INT_MASK   0x08
#define BULK_OUT_INT_MASK 0x10
#define BULK_IN_INT_MASK  0x20
#define CHANGE_INT_MASK   0x40
#define DISCONN_INT_MASK  0x80
#define ALL_INT_MASK      0xf8


struct usb_qcamvc
{
	struct usb_device *udev;
	unsigned char iface;
	int open;
	int present;
};

/*****************************************************************************/
/*                                                                           */
/* Lowlevel driver for Connectix QuickCam VC(USB)                            */
/*                                                                           */
/*****************************************************************************/
static int		qcamvc_open(void *privdata);
static int		qcamvc_close(void *privdata);
static size_t	qcamvc_get_reg(void *privdata, unsigned char reg, void *buffer, size_t len);
static int		qcamvc_set_reg(void *privdata, unsigned char reg, void *buffer, size_t len);
static size_t	qcamvc_read(void *privdata, void *buffer, size_t len);
static int		qcamvc_write(void *privdata, void *buffer, size_t len);
static int		qcamvc_stream_start(void *privdata);
static int		qcamvc_stream_stop(void *privdata);
static size_t	qcamvc_stream_read(void *privdata, void *buffer, size_t len);

static int		get_1284_register(struct usb_device *usbdev, unsigned char reg, unsigned char *val);
static int		set_1284_register(struct usb_device *usbdev, unsigned char reg, unsigned char val);
static size_t	qcamvc_bulk_read(struct usb_device *usbdev, void *buffer, size_t len);
static size_t	qcamvc_bulk_write(struct usb_device *usbdev, void *buffer, size_t len );
static int		reset_uss(struct usb_device *usbdev);


static struct qcamvc_camera_ops qcamvc_usb_ops =
{
	qcamvc_open,
	qcamvc_close,
	qcamvc_write,
	qcamvc_read,
	qcamvc_set_reg,
	qcamvc_get_reg,
	qcamvc_stream_start,
	qcamvc_stream_stop,
	qcamvc_stream_read
};

static int probe_qcamvc(struct usb_interface *intf, const struct usb_device_id *id);
static void disconnect_qcamvc(struct usb_interface *intf);

static LIST_HEAD(cam_list);
static spinlock_t cam_list_lock_usb;

#if LINUX_VERSION_CODE >= 0x020400
static struct usb_device_id qcamvc_device_ids [] =
{
	{ USB_DEVICE(0x0478, 0x0001) },
	{ }
};
MODULE_DEVICE_TABLE(usb,qcamvc_device_ids);
MODULE_LICENSE("GPL");
#endif /* LINUX_VERSION_CODE >= 0x020400 */


static struct usb_driver qcamvc_driver =
{
//	.owner		= THIS_MODULE,
	.name		= "Connectix QuickCam VC",
	.probe		= probe_qcamvc,
	.disconnect	= disconnect_qcamvc,
#if LINUX_VERSION_CODE >= 0x020400
	.id_table	= qcamvc_device_ids,
#else
	{ NULL, NULL}
#endif
};

static int qcamvc_stream_start(void *privdata)
{
	int ret = 0;
	return ret;
}

static int qcamvc_stream_stop(void *privdata)
{
	int ret = 0;
	return ret;
}

static int qcamvc_open(void *privdata)
{
	/*struct usb_qcamvc *usb_qcamvc =  (struct usb_qcamvc*)privdata;

	if (!usb_qcamvc)
		return -EINVAL;

	usb_qcamvc->open = 1;*/
	
	return 0;
}

static int qcamvc_close(void *privdata)
{
	/*struct usb_qcamvc *usb_qcamvc =  (struct usb_qcamvc*)privdata;

	if (!usb_qcamvc)
		return -EINVAL;

	usb_qcamvc->open = 0;*/
	
	return 0;
}

static int qcamvc_write(void *privdata, void *buffer, size_t len)
{
	struct usb_qcamvc *usb_qcamvc =  (struct usb_qcamvc*)privdata;
	struct usb_device *udev = usb_qcamvc->udev;
	return qcamvc_bulk_write(udev, buffer, len);
}

static size_t qcamvc_read(void *privdata, void *buffer, size_t len)
{
	struct usb_qcamvc *usb_qcamvc =  (struct usb_qcamvc*)privdata;
	struct usb_device *udev = usb_qcamvc->udev;
	return qcamvc_bulk_read(udev, buffer, len);
}

static int qcamvc_set_reg(void *privdata, unsigned char cmd, void *buffer, size_t len)
{
	struct usb_qcamvc *usb_qcamvc =  (struct usb_qcamvc*)privdata;
	struct usb_device *udev = usb_qcamvc->udev;
	int	written = 0;
	
	reset_uss(udev);
	set_1284_register(udev, SET_USS720_USSCTRL, ALL_INT_MASK);
	set_1284_register(udev, SET_USS720_CONTROL, SELECT_IN | NINIT);
	set_1284_register(udev, SET_USS720_DATA, 0x10);
	set_1284_register(udev, SET_USS720_CONTROL, AUTO_FD | NINIT);
	set_1284_register(udev, SET_USS720_CONTROL, AUTO_FD | NINIT | STROBE);
	set_1284_register(udev, SET_USS720_CONTROL, NINIT);
	set_1284_register(udev, SET_USS720_CONTROL, AUTO_FD | NINIT);
	set_1284_register(udev, SET_USS720_DATA, cmd);
	set_1284_register(udev, SET_USS720_CONTROL, AUTO_FD | NINIT | STROBE);
	set_1284_register(udev, SET_USS720_CONTROL, AUTO_FD | NINIT);
	reset_uss(udev);
	
	if ((len > 0)&(buffer!=NULL))
		written = qcamvc_bulk_write(udev, buffer, len);
	
	USB_DBG("Send CMD reg=0x%02x,\n", cmd);
	
	return written;
}

static size_t qcamvc_get_reg(void *privdata, unsigned char reg, void *buffer, size_t len)
{
	struct usb_qcamvc *usb_qcamvc =  (struct usb_qcamvc*)privdata;
	struct usb_device *udev = usb_qcamvc->udev;
	
	unsigned char regs [7];
	int retval = 0;
	
	if ((udev==NULL)||(buffer==NULL))
		return -1;
		
	reset_uss(udev);
	get_1284_register(udev, GET_USS720_CONTROL, regs);
	set_1284_register(udev, SET_USS720_USSCTRL, ALL_INT_MASK);
	get_1284_register(udev, GET_USS720_CONTROL, regs);
	set_1284_register(udev, SET_USS720_USSCTRL, ALL_INT_MASK);
	set_1284_register(udev, SET_USS720_CONTROL, EPP_MASK | NINIT | HLH | SELECT_IN);
	set_1284_register(udev, SET_USS720_DATA, 0x10);
	set_1284_register(udev, SET_USS720_CONTROL, EPP_MASK | NINIT | HLH | AUTO_FD);
	get_1284_register(udev, GET_USS720_CONTROL, regs);
	set_1284_register(udev, SET_USS720_CONTROL, EPP_MASK | NINIT | HLH | AUTO_FD |STROBE );
	set_1284_register(udev, SET_USS720_CONTROL, EPP_MASK | NINIT | HLH);
	get_1284_register(udev, GET_USS720_CONTROL, regs);
	set_1284_register(udev, SET_USS720_CONTROL, EPP_MASK | NINIT | HLH | AUTO_FD);
	get_1284_register(udev, GET_USS720_CONTROL, regs);
	get_1284_register(udev, GET_USS720_STATUS,  regs);
	set_1284_register(udev, SET_USS720_EXTCTRL, ECR_ECP | BULK_IN_EMPTY| BULK_OUT_EMPTY );
	set_1284_register(udev, SET_USS720_CONTROL, EPP_MASK | NINIT | HLH);
	set_1284_register(udev, SET_USS720_ECPCMD,  reg);
	reset_uss(udev);
	
	if ((len > 0)&(buffer!=NULL))
	{
		retval = qcamvc_bulk_read(udev, buffer, len);
		reset_uss(udev);
	}
	
	USB_DBG("Read 0x%02x, len=%02x, \n", reg, sizeof(buffer));
	
	return retval;
}

static size_t qcamvc_stream_read(void *privdata, void *buffer, size_t len)
{
	struct usb_qcamvc *usb_qcamvc =  (struct usb_qcamvc*)privdata;
	struct usb_device *udev = usb_qcamvc->udev;
	//unsigned char buf = 0;
	size_t ret;
	
	reset_uss(udev);
	//qcamvc_bulk_write(udev, &buf, 1);
	ret = qcamvc_bulk_read(udev, buffer, len);
	
	return ret;
}

static int get_1284_register(struct usb_device *usbdev, unsigned char reg, unsigned char *val) 
{
	int ret;
	
	if (usbdev==NULL)
	{
		printk("QuickCam VC(USB):get_1284_reg not usbdev!\n");
		return -1;
	}
	
	ret = usb_control_msg(usbdev, usb_rcvctrlpipe(usbdev,0), 3, 0xc0, 
		((unsigned int)reg) << 8, 0, val, 7, HZ);
	if (ret < 0)
	{
		printk("QuickCam VC(USB): get_1284_register(%d) failed, status 0x%02x\n",
			(unsigned int)reg, ret);
	}
	else
	{
	}     
	
	return ret;
}


static int set_1284_register(struct usb_device *usbdev, unsigned char reg, unsigned char val)
{
	int ret;
		
	if (usbdev==NULL)
	{
		printk("QuickCam VC(USB):set_1284_reg not usbdev!\n");
		return -1;
	}
		
	ret = usb_control_msg(usbdev, usb_sndctrlpipe(usbdev,0), 4, 0x40, 
		(((unsigned int)reg) << 8) | val, 0, NULL, 0, HZ);      
	if (ret)
	{
		printk("QuickCam VC(USB): set_1284_register(%u,0x%02x) failed, status 0x%02x\n", 
			(unsigned int)reg, (unsigned int)val, ret);
	}
	else
	{
	}
	return ret;
}

static int reset_uss(struct usb_device *usbdev)
{
	int ret;
	if (usbdev==NULL)
		return -1;
	ret = usb_control_msg(usbdev, usb_sndctrlpipe(usbdev,0), 2, 0x23, 
		0, 0, NULL, 0, HZ);
	if (ret)
	{
		printk("QuickCam VC(USB): reset_uss failed, status 0x%02x\n", ret);
	}
	else
	{
	}
	
	return ret;
}

static size_t qcamvc_bulk_write(struct usb_device *usbdev, void *buffer, size_t len )
{
	int rlen;
	int i;
	
	if (usbdev==NULL)
	{
		printk("QuickCam VC(USB):Bulk_write not usbdev!");
		return -1;
	}
	if (buffer==NULL)
	{
		printk("QuickCam VC(USB):Bulk_write NULL pointer buffer!");
		return -1;
	}
	USB_DBG("bulk write len=%d\n",(int) len);           
	i = usb_bulk_msg(usbdev, usb_sndbulkpipe(usbdev, 1), (void *)buffer, len, &rlen, HZ*20);
	if (i)
		printk("QuickCam VC(USB): bulkwrite buf %p len %u rlen %u\n", buffer, (int)len, rlen);
	
	return rlen;
}

static size_t qcamvc_bulk_read(struct usb_device *usbdev, void *buffer, size_t len)
{
	int rlen;
	int i;
	
	if (usbdev==NULL)
	{
		printk("QuickCam VC(USB):Bulk_read not usbdev!");
		return -1;
	}
	if (buffer==NULL)
	{
		printk("QuickCam VC(USB):Bulk_read NULL pointer buffer!");
		return -1;
	}
	USB_DBG("bulk read len=0x%x ",(int) len);         
	
	i = usb_bulk_msg(usbdev, usb_rcvbulkpipe(usbdev, 2), buffer, len, &rlen, HZ*20);
	if (len!=rlen)
		printk("QuickCam VC(USB): bulkread buf %p len 0x%x rlen 0x%x\n", buffer, (int)len, rlen);
		
	USB_DBG("rlen=%d\n",(int) rlen);
	
	return rlen;
}

static int probe_qcamvc(struct usb_interface *intf, const struct usb_device_id *id)
{
	struct usb_device *udev = interface_to_usbdev(intf);
	struct usb_host_interface *interface = NULL;
	struct usb_qcamvc *usb_qcamvc = NULL;
	struct qcamvc_data *qcamvc = NULL;
	int ret;
  
	printk(KERN_INFO "QuickCam VC(USB): Found Vendor ID: 0x%x, Product ID: %d\n", 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
		le16_to_cpu(udev->descriptor.idVendor), 
		le16_to_cpu(udev->descriptor.idProduct));
#else
		udev->descriptor.idVendor, udev->descriptor.idProduct);
#endif
	
	/* See if the device offered us matches what we can accept */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
	if ((le16_to_cpu(udev->descriptor.idVendor) != 0x0478) || 
		(le16_to_cpu(udev->descriptor.idProduct) != 0x0001))
	{
		printk(KERN_INFO "QuickCam VC(USB): That device is not a QuickCam VC (0x%x,%d)\n", 
			le16_to_cpu(udev->descriptor.idVendor), 
			le16_to_cpu(udev->descriptor.idProduct));
#else
	if ((udev->descriptor.idVendor != 0x0478) || (udev->descriptor.idProduct != 0x0001))
	{
		printk(KERN_INFO "QuickCam VC(USB): That device is not a QuickCam VC (0x%x,%d)\n", 
			udev->descriptor.idVendor, udev->descriptor.idProduct);
#endif
		return -ENODEV;
	}

	usb_qcamvc = kmalloc(sizeof(*usb_qcamvc), GFP_KERNEL);
	if (!usb_qcamvc) {
		printk(KERN_ERR "couldn't kmalloc usb_qcamvc struct\n");
		return -ENOMEM;
	}

	interface = &intf->altsetting[0];
	usb_qcamvc->iface = interface->desc.bInterfaceNumber;
	usb_qcamvc->udev = udev;

	ret = usb_set_interface(udev, usb_qcamvc->iface, 2);
	if (ret < 0)
	{
		printk(KERN_ERR "qcamvc_probe: usb_set_interface error (returned %d)\n", ret);
	}
	USB_DBG(" USS720: set inteface result %d\n", i);

	if ( (qcamvc = qcamvc_register_camera(&qcamvc_usb_ops, usb_qcamvc, "QuickCam VC(USB)")) == NULL )
	{
		printk(KERN_ERR "Failed to qcamvc_register_camera (USB)\n");
		kfree(usb_qcamvc);
		return -EIO;
	}

	spin_lock( &cam_list_lock_usb );
	list_add( &qcamvc->cam_data_list, &cam_list );
	spin_unlock( &cam_list_lock_usb );

	usb_set_intfdata(intf, qcamvc);
	
	printk(KERN_INFO "QuickCam VC(USB): sucessfully registered.\n");
	
	return 0;
}

static void disconnect_qcamvc(struct usb_interface *intf)
{
	struct qcamvc_data *qcamvc = usb_get_intfdata(intf);
	struct usb_qcamvc *usb_qcamvc;
	struct usb_device *udev;
	
	usb_set_intfdata(intf, NULL);
	
	if (!qcamvc)
	{
		//printk(KERN_ERR "usb_get_intfdata() returned NULL!\n");
		return;
	}

	printk(KERN_INFO "QuickCam VC(USB): unregistering video/USB...\n");
	
	usb_qcamvc = (struct usb_qcamvc *) qcamvc->lowlevel_data;
	
	spin_lock( &cam_list_lock_usb );
	list_del(&qcamvc->cam_data_list);
	spin_unlock( &cam_list_lock_usb );

	qcamvc_unregister_camera(qcamvc);
	
	udev = interface_to_usbdev(intf);
	usb_driver_release_interface(&qcamvc_driver, udev->actconfig->interface[0]);
	
	kfree(usb_qcamvc);

	printk(KERN_INFO "QuickCam VC(USB): USB/video devices unregistered.\n");
}

static int __init qcamvc_usb_init (void)
{
	printk("QuickCam VC(USB): (C) 2001 by De Marchi Daniele, <demarchidaniele@libero.it>\n");
	printk("QuickCam VC(USB): lowlevel driver version %s registered.\n",QCAMVCVERSION);

	spin_lock_init(&cam_list_lock_usb);
	
	return usb_register(&qcamvc_driver);
}

static void __exit qcamvc_usb_exit(void)
{
	usb_deregister(&qcamvc_driver);
}

module_init(qcamvc_usb_init);
module_exit(qcamvc_usb_exit);
