/*	Copyright (C) 2004 Bertrik Sikken, Thomas Cougnard
	Parts of this file were adapted from the sn9c102 driver by Luca Risolia.
	
	This file is part of the sonix-cam driver.

    The sonix-cam driver is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    The sonix-cam driver is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the sonix-cam driver; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "sonix_sensor.h"

/*	The following defines the number of lines per frame.
	A smaller number means a higher framerate.
	Minimum value is about 290 (any smaller and frames will be incomplete)
	A higher number results in a brighter picture.
	Maximum value is 511 (any bigger and auto-exposure will stop working)
*/
#define FRAME_LINES		300

#define MAX(a,b) ((a)>(b)?(a):(b))
#define MIN(a,b) ((a)<(b)?(a):(b))

static int pas106b_probe(struct uvd *uvd)
{
	struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;
	int r0 = 0, r1 = 0, err = 0;
	unsigned int pid = 0;

	/*
	   Minimal initialization to enable the I2C communication
	   NOTE: do NOT change the values!
	*/
	err += sn9c102_write_reg(cam, 0x01, 0x01); /* sensor power down */
	err += sn9c102_write_reg(cam, 0x01, 0x00); /* sensor power on */
	err += sn9c102_write_reg(cam, 0x17, 0x28); /* sensor clock at 24 MHz */
	if (err) {
		return -EIO;
	}

	r0 = sn9c102_i2c_read(cam, 0x00);
	r1 = sn9c102_i2c_read(cam, 0x01);

	if (r0 < 0 || r1 < 0) {
		return -EIO;
	}

	pid = (r0 << 11) | ((r1 & 0xf0) >> 4);
	if (pid != 0x007) {
		return -ENODEV;
	}

	return 0;
}


static int pas106b_init(struct uvd *uvd)
{
	struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;
	int err = 0;
	
	/* pixel clock divider */
	err += sn9c102_i2c_write(cam,0x02,0x0A);
	/* frame length */
	err += sn9c102_i2c_write(cam,0x03,FRAME_LINES / 16);
	err += sn9c102_i2c_write(cam,0x04,FRAME_LINES % 16);
	/* shutter */
	err += sn9c102_i2c_write(cam,0x05,0x00);
	err += sn9c102_i2c_write(cam,0x06,0x00);
	/* DAC sign */
	err += sn9c102_i2c_write(cam,0x07,0x01);
	/* DAC magnitude (= black level) */
	err += sn9c102_i2c_write(cam,0x08,0x06);
	/* color gains */
	err += sn9c102_i2c_write(cam,0x09,0x01);
	err += sn9c102_i2c_write(cam,0x0A,0x00);
	err += sn9c102_i2c_write(cam,0x0B,0x00);
	err += sn9c102_i2c_write(cam,0x0C,0x01);
	/* ??? */
	err += sn9c102_i2c_write(cam,0x0D,0x00);
	/* global gain */
	err += sn9c102_i2c_write(cam,0x0E,0x10);
	/* ??? */
	err += sn9c102_i2c_write(cam,0x0F,0x00);
	err += sn9c102_i2c_write(cam,0x10,0x06);
	err += sn9c102_i2c_write(cam,0x11,0x06);
	err += sn9c102_i2c_write(cam,0x12,0x10);
	/* ? */
	err += sn9c102_i2c_write(cam,0x14,0x02);
	/* validate settings */	
	err += sn9c102_i2c_write(cam,0x13,0x01);
	
	return err;
}


static int pas106b_adjust_picture(struct uvd *uvd)
{
  struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;
  struct sn9c102_sensor *sensor = cam->sensor;
  int err = 0;
  
  /* We should change brightness,colour, contrast and hue here */
  u16 amplitude;
  u16 red=0,blue=0;

  DEBUG(2,"Adjust Picture requested");
  DEBUG(2,"Brightness: %u,Hue= %u, Coulour= %u, Contrast= %u",uvd->vpic.brightness,uvd->vpic.hue,uvd->vpic.colour,uvd->vpic.contrast);

  /* Adjust "Brightness" */
  err += sn9c102_i2c_write(cam,0x07,(uvd->vpic.brightness>0x7fff?0:1));

  if(uvd->vpic.brightness>0x7fff)
    amplitude=uvd->vpic.brightness-0x7fff;
  else
    amplitude=0x7fff-uvd->vpic.brightness;

  err += sn9c102_i2c_write(cam,0x08,amplitude>>8);

  /* Adjust color blue <-> Red */
  if(uvd->vpic.colour > 0x7fff)
      red=uvd->vpic.colour-0x7fff;
  else
      blue=0x7fff-uvd->vpic.colour;

  err += sn9c102_i2c_write(cam,0x0C,red>>10);
  err += sn9c102_i2c_write(cam,0x09,blue>>10);

  /* Update Hue ? */
  err += sn9c102_i2c_write(cam,0x0E,uvd->vpic.hue>>11);
  
  /* Validate Change */
  err += sn9c102_i2c_write(cam,0x13,0x01);
  
  /* contrast controls AE target value */
  sensor->exp_target = uvd->vpic.contrast;
  
  return err;
}


static int pas106b_post_process(struct uvd *uvd)
{
	struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;
	struct sn9c102_sensor *sensor = cam->sensor;
	int exp_old, exp_new, exp_measured, exp_target, exp_max;
	int err = 0;
	
	/* calculate exposure from AE measured, previous value and target value */
	exp_max = MIN(255, FRAME_LINES / 2);
	exp_measured = MAX(1, cam->ae_data1 + cam->ae_data2);
	exp_target = (sensor->exp_target * 21750) >> 16;
	exp_old = MAX(1, sensor->exposure);
	exp_new = MIN(exp_max, (exp_old * exp_target) / exp_measured);
	/* simple averaging filter to make dynamics smoother */
	exp_new = (exp_old + exp_new) / 2;
	if (exp_new != sensor->exposure) {
		sensor->exposure = exp_new;
		err += sn9c102_i2c_write(cam, 0x05, exp_max - sensor->exposure);
		err += sn9c102_i2c_write(cam, 0x13, 0x01);
	}
	
	return err;
}


struct sn9c102_sensor sensor_pas106b = {
	.name = "PAS106B",
	
	.frequency = SN9C102_I2C_100KHZ,
	.interface = SN9C102_I2C_2WIRES,
	.slave_id = 0x40,
	
	.probe = pas106b_probe,
	.init = pas106b_init,
	.adjust_picture = pas106b_adjust_picture,
	.post_process = pas106b_post_process
};


