/***************************************************************************
 * Plug-in for PAS202BCB image sensor connected to the SN9C10x PC Camera   *
 * Controllers                                                             *
 *                                                                         *
 * Copyright (C) 2004 by Carlos Eduardo Medaglia Dyonisio                  *
 *                       <medaglia@undl.org.br>                            *
 *                       http://cadu.homelinux.com:8080/                   *
 *                                                                         *
 * DAC Magnitude, DAC sign, exposure and green gain controls added by      *
 * Luca Risolia <luca.risolia@studio.unibo.it>                             *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify    *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation; either version 2 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with this program; if not, write to the Free Software             *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               *
 ***************************************************************************/

#include "sonix_sensor.h"


struct cmos_regs {
	int enabled;
	int val;
};

static struct cmos_regs pas202b_regs[22] = {
	{0, 0x00}, /* 0x00 - Part ID */
	{0, 0x00}, /* 0x01 - Part ID / Version ID */
	{1, 0x02}, /* 0x02 - Np[5:0] */
	{1, 0x40}, /* 0x03 - Nov_by2[7:0] */
	{1, 0x12}, /* 0x04 - lpf[13:6] - Line per frame */
	{1, 0x04}, /* 0x05 - lpf[5:0] - Line per frame */
	{0, 0x04}, /* 0x06 - cgn_dk[3:0] - color gaim for dark pixel */
	{0, 0x00}, /* 0x07 - cgn_B[3:0] - color gaim for Blue pixel */
	{0, 0x00}, /* 0x08 - cgn_G[3:0] - color gaim for Green pixel */
	{0, 0x00}, /* 0x09 - cgn_R[3:0] - color gaim for Red pixel */
	{0, 0x00}, /* 0x0a - Reserved */
	{1, 0x01}, /* 0x0b - dac[8] - sign bit for DAC */
	{1, 0x2f}, /* 0x0c - dac[0:7] - Magnitude for DAC */
	{1, 0x7e}, /* 0x0d - offset_ne_by4[0:7] - Exposure time pixel offset */
	{1, 0x01}, /* 0x0e - offset_ny[0] - Exposure time: line offset */
	{1, 0xdc}, /* 0x0f - offset_ny[8:1] - Exposure time: line offset */
	{0, 0x00}, /* 0x10 - global[0:4] - PGA global gaim */
	{0, 0x00}, /* 0x11 - i2c_update_sync_flag[0] */
	{0, 0x00}, /* 0x12 - Reserved */
	{1, 0x63}, /* 0x13 - Reserved */
	{0, 0x00}, /* 0x14 - Reserved */
	{1, 0x70}  /* 0x15 - Reserved */
};


static int pas202b_probe(struct uvd *uvd)
{       
	struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;
	int r0 = 0, r1 = 0, err = 0;
	unsigned int pid = 0;

	/*
	 *  Minimal initialization to enable the I2C communication
	 *  NOTE: do NOT change the values!
	 */
	err += sn9c102_write_reg(cam, 0x01, 0x01); /* sensor power down */
	err += sn9c102_write_reg(cam, 0x01, 0x40); /* sensor power on */
	err += sn9c102_write_reg(cam, 0x17, 0x28); /* sensor clock at 24 MHz */
	if (err)
		return -EIO;

	r0 = sn9c102_i2c_read(cam, 0x00);
	r1 = sn9c102_i2c_read(cam, 0x01);
	
	if (r0 < 0 || r1 < 0)
		return -EIO;

	pid = (r0 << 4) | ((r1 & 0xf0) >> 4);
	if (pid != 0x017)
		return -ENODEV;

	return 0;
}


static int pas202b_init(struct uvd *uvd)
{
	struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;
	int i;
	int err = 0;

	err += sn9c102_write_reg(cam, 0x10, 0x00);		/* R_GAIN/B_GAIN */
	err += sn9c102_write_reg(cam, 0x11, 0x00);		/* G_GAIN */
	err += sn9c102_write_reg(cam, 0x14, 0x00);		/* OFFSET */
	err += sn9c102_write_reg(cam, 0x17, 0x20);		/* enable SEN_CLK, set SEN_RATE */
	err += sn9c102_write_reg(cam, 0x19, 0x30);		/* MCK_SIZE = 3 */
	err += sn9c102_write_reg(cam, 0x18, 0x09);		/* PCK_RIS | VSYNC_HIGH */
	/* sensor image size */
	err += sn9c102_write_reg(cam, 0x15, 0x28);		/* H_SIZE */
	err += sn9c102_write_reg(cam, 0x16, 0x1E);		/* V_SIZE */
	err += sn9c102_write_reg(cam, 0x1A, 0x14);		/* HO_SIZE */
	err += sn9c102_write_reg(cam, 0x1B, 0x0F);		/* VO_SIZE */
	/* AE window */
	err += sn9c102_write_reg(cam, 0x1C, 5);			/* AE_STRX */
	err += sn9c102_write_reg(cam, 0x1D, 15);		/* AE_STRY */
	err += sn9c102_write_reg(cam, 0x1E, 4);			/* AE_ENDX */
	err += sn9c102_write_reg(cam, 0x1F, 11);		/* AE_ENDY */

#if 0	/* code from sn9c102 driver */
	err += sn9c102_i2c_write(cam, 0x02, 0x14);
	err += sn9c102_i2c_write(cam, 0x03, 0x40);
	err += sn9c102_i2c_write(cam, 0x0d, 0x2c);
	err += sn9c102_i2c_write(cam, 0x0e, 0x01);
	err += sn9c102_i2c_write(cam, 0x0f, 0xa9);
	err += sn9c102_i2c_write(cam, 0x10, 0x08);
	err += sn9c102_i2c_write(cam, 0x13, 0x63);
	err += sn9c102_i2c_write(cam, 0x15, 0x70);
	err += sn9c102_i2c_write(cam, 0x11, 0x01);
#else	/* code from Carlos Eduardo Medaglia Dyonisio */
	for (i = 0; i < 22; i++) {
		if (pas202b_regs[i].enabled == 1) {
			err += sn9c102_i2c_write(cam, i, pas202b_regs[i].val);
		}
	}
#endif
	
	return err;
}


/* We should change brightness,colour, contrast and hue here */
static int pas202b_adjust_picture(struct uvd *uvd)
{
	struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;
	int err = 0;
	u16 amplitude;
	u16 red=0,blue=0;
	
	DEBUG(2,"Adjust Picture requested");
	DEBUG(2,"Brightness: %u,Hue= %u, Coulour= %u, Contrast= %u",uvd->vpic.brightness,uvd->vpic.hue,uvd->vpic.colour,uvd->vpic.contrast);
	
	/* Adjust "Brightness" */
	err += sn9c102_i2c_write(cam,0x0b,(uvd->vpic.brightness>0x7fff?0:1));
	
	if(uvd->vpic.brightness>0x7fff)
	amplitude=uvd->vpic.brightness-0x7fff;
	else
	amplitude=0x7fff-uvd->vpic.brightness;
	
	err += sn9c102_i2c_write(cam,0x0c,amplitude>>8);
	
	/* Adjust color blue <-> Red */
	if(uvd->vpic.colour > 0x7fff)
		red=uvd->vpic.colour-0x7fff;
	else
		blue=0x7fff-uvd->vpic.colour;
	
	err += sn9c102_i2c_write(cam,0x09,red>>10);
	err += sn9c102_i2c_write(cam,0x07,blue>>10);
	
	/* Update Hue ? */
	err += sn9c102_i2c_write(cam,0x10,uvd->vpic.hue>>11);
	
	/* Validate Change */
	err += sn9c102_i2c_write(cam,0x11,0x01);
	
	return err;
}


struct sn9c102_sensor sensor_pas202b = {
	.name = "PAS202B",
	
	.frequency = SN9C102_I2C_100KHZ,
	.interface = SN9C102_I2C_2WIRES,
	.slave_id = 0x40,
	
	.probe = pas202b_probe,
	.init = pas202b_init,
	.adjust_picture = pas202b_adjust_picture
};


