/*
 * sonixcam.c - sonixcam webcam driver
 *
 * Author: Thomas Cougnard <tgnard@free.fr>
 *
 * Copyright (C) 2004 Thomas Cougnard
 *
 * Licence: GPL
 *
 * Driver for USB webcams based on a Sonix sn9c101 chipset.
 * Currently, only one CMOS chipset is supported
 * 
 * "working" webcam are: 
 *     trust spacec@m 120 (usb vid/pid 0C45:600D) 
 *                         usb vid/pid 0C45:6009
 *
 *
 * Based on the Konica driver by Simon Evans and the usbvideo module 
 * 
 *   -- back-port for 2.4.x kernels and support for usb vid/pid 0C45:6029 by Dennis Gerasimov <denis@gerasimov.net>, 6/21/2004
 */

#include "sonix.h"
#include "sonix_sensor.h"

#define MAX_CAMERAS	2

#define DRIVER_VERSION	"v0.1"
#define DRIVER_DESC	"Sonix Webcam driver"

#define SONIX_HEADER_LENGTH 12

#define SN9C102_CTRL_TIMEOUT      10*HZ


enum frame_sizes {
	SIZE_352X288	= 0,
	
};

static enum frame_sizes size;

struct cam_size {
	u16	width;
	u16	height;
};

static struct cam_size camera_sizes[] = { { 352, 288} };

static struct usbvideo *cams;

static struct sn9c102_sensor *sensor_table[] = {
	&sensor_pas106b,
	&sensor_pas202b,
	&sensor_tas5110c1b,
	NULL					/* terminating entry */
};

int debug=5;


int sn9c102_write_regs(struct sn9c102_device* cam, u16 index, u8 *data, int len)
{
	struct usb_device* udev = cam->usbdev;

	return usb_control_msg(udev, usb_sndctrlpipe(udev, 0), 0x08, 0x41,
	                       index, 0, data, len, SN9C102_CTRL_TIMEOUT);
}


int sn9c102_write_reg(struct sn9c102_device* cam, u16 index, u8 value)
{
	int res;
	
	res = sn9c102_write_regs(cam, index, &value, 1);
	if (res < 0) {
		DEBUG(3, "Failed to write a register (value 0x%02X, index "
		       "0x%02X, error %d)", value, index, res)
		return -1;
	}
	return 0;
}


/* NOTE: reading some registers always returns 0 */
static int sn9c102_read_regs(struct sn9c102_device* cam, u16 index, u8 *data, int len)
{
	struct usb_device* udev = cam->usbdev;

	return usb_control_msg(udev, usb_rcvctrlpipe(udev, 0), 0x00, 0xc1,
	                       index, 0, data, len, SN9C102_CTRL_TIMEOUT);
}


int sn9c102_read_reg(struct sn9c102_device* cam, u16 index)
{
	int res;
	u8 value;
	
	res = sn9c102_read_regs(cam, index, &value, 1);
	if (res < 0) {
		DEBUG(3, "Failed to read a register (index 0x%02X, error %d)",
		    index, res)
		return -1;
	}
	
	return value;
}

/*****************************************************************************/
/*	i2c stuff, to talk to the sensor
 */

static int
sn9c102_i2c_wait(struct sn9c102_device* cam, struct sn9c102_sensor* sensor)
{
	int i, r;

	for (i = 1; i <= 5; i++) {
		r = sn9c102_read_reg(cam, 0x08);
		if (r < 0)
			return -EIO;
		if (r & 0x04)
			return 0;
		if (sensor->frequency & SN9C102_I2C_400KHZ)
			udelay(5*8);
		else
			udelay(16*8);
	}
	return -EBUSY;
}


static int
sn9c102_i2c_detect_error(struct sn9c102_device* cam, 
                         struct sn9c102_sensor* sensor)
{
	int r;
	r = sn9c102_read_reg(cam, 0x08);
	return (r < 0 || (r >= 0 && !(r & 0x08))) ? -EIO : 0;
}


int sn9c102_i2c_read(struct sn9c102_device* cam, u8 address)
{
	struct sn9c102_sensor* sensor = cam->sensor;
	u8 data[8];
	int err = 0, res;

	/* Write cycle - address */
	data[0] = ((sensor->interface == SN9C102_I2C_2WIRES) ? 0x80 : 0) |
	          ((sensor->frequency & SN9C102_I2C_400KHZ) ? 0x01 : 0) | 0x10;
	data[1] = sensor->slave_id;
	data[2] = address;
	data[7] = 0x10;
	res = sn9c102_write_regs(cam, 0x08, data, 8);
	if (res < 0)
		err += res;

	err += sn9c102_i2c_wait(cam, sensor);

	/* Read cycle - 1 byte */
	data[0] = ((sensor->interface == SN9C102_I2C_2WIRES) ? 0x80 : 0) |
	          ((sensor->frequency & SN9C102_I2C_400KHZ) ? 0x01 : 0) |
	          0x10 | 0x02;
	data[1] = sensor->slave_id;
	data[7] = 0x10;
	res = sn9c102_write_regs(cam, 0x08, data, 8);
	if (res < 0)
		err += res;

	err += sn9c102_i2c_wait(cam, sensor);

	/* The read byte will be placed in data[4] */
	res = sn9c102_read_regs(cam, 0x0A, data, 5);
	if (res < 0)
		err += res;

	err += sn9c102_i2c_detect_error(cam, sensor);

	if (err)
		DEBUG(3, "I2C read failed for %s image sensor", sensor->name)

	return err ? -1 : (int)data[4];
}


static int 
sn9c102_i2c_try_raw_write(struct sn9c102_device* cam,
                          struct sn9c102_sensor* sensor, u8 n, u8 data0,
                          u8 data1, u8 data2, u8 data3, u8 data4, u8 data5)
{
	u8 data[8];
	int err = 0, res;

	/* Write cycle. It usually is address + value */
	data[0] = ((sensor->interface == SN9C102_I2C_2WIRES) ? 0x80 : 0) |
	          ((sensor->frequency & SN9C102_I2C_400KHZ) ? 0x01 : 0)
	          | ((n - 1) << 4);
	data[1] = data0;
	data[2] = data1;
	data[3] = data2;
	data[4] = data3;
	data[5] = data4;
	data[6] = data5;
	data[7] = 0x14;
	res = sn9c102_write_regs(cam, 0x08, data, 8);
	if (res < 0)
		err += res;

	err += sn9c102_i2c_wait(cam, sensor);
	err += sn9c102_i2c_detect_error(cam, sensor);

	if (err)
		DEBUG(3, "I2C write [%02X]<=%02X failed for %s image sensor", data1, data2, sensor->name)

	return err ? -1 : 0;
}


int sn9c102_i2c_write(struct sn9c102_device* cam, u8 address, u8 value)
{
	struct sn9c102_sensor* sensor = cam->sensor;
	return sn9c102_i2c_try_raw_write(cam, sensor, 3, 
	                                 sensor->slave_id, address,
	                                 value, 0, 0, 0);
}


/*****************************************************************************/

static inline void sonixcam_camera_on(struct uvd *uvd)
{
		struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;

        DEBUG(0, "camera on");
        sn9c102_write_reg(cam, 0x1,0x00);
}


static inline void sonixcam_camera_off(struct uvd *uvd)
{
		struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;
        
		DEBUG(0, "camera off");
        sn9c102_write_reg(cam, 0x1,0x00);
}


static void sonixcam_video_start(struct uvd *uvd)
{
  struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;
  
  u8 buf[8];

  /* Turn on the led */
  sn9c102_write_reg(cam, 0x1,0x00);
  DEBUG(0, "Video start\n");


  /* Sending the magic init sequence */
  
  // pixel number for sensor
  buf[0]=0x16;buf[1]=0x12;
  sn9c102_write_regs(cam,0x15,buf,2);
  
  // MCK_SIZE=2
  buf[0]=0x20;
  buf[0]=0x24;
  sn9c102_write_regs(cam,0x19,buf,1);
  
  // Compression & Synchro
  buf[0]=0x09;
  //buf[0]=0x89;
  //buf[0]=0x89;
  sn9c102_write_regs(cam,0x18,buf,1);
   

  // clock
//  buf[0]=0x28;
  buf[0]=0x24;
  sn9c102_write_regs(cam,0x17,buf,1);

  //H_start
  buf[0]=0x04;buf[1]=0x03;buf[2]=0x00;
  sn9c102_write_regs(cam,0x12,buf,3);

  //Active window
  buf[0]=0x03;buf[1]=0x02;buf[2]=0x08;buf[3]=0x07;
  sn9c102_write_regs(cam,0x1c,buf,4);

  // Greenify
  //buf[0]=0x07;
  //sn9c102_write_regs(cam,0x11,buf,1);
  
	if (cam->sensor->init != NULL) {
		cam->sensor->init(uvd);
	}
    
  //enable transfert
  buf[0]=0x04;
  sn9c102_write_regs(cam,0x01,buf,1);
}


static void sonixcam_video_stop(struct uvd *uvd)
{
  struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;

  /* Turn off the led and disable transfert*/
  sn9c102_write_reg(cam, 0x1,0x08);
  DEBUG(0, "Video stop\n");

}


static void sonixcam_adjust_picture(struct uvd *uvd)
{
	struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;
	
	/* pass it to the sensor */
	if (cam->sensor->adjust_picture != NULL) {
		cam->sensor->adjust_picture(uvd);
	}
}


static int sonixcam_set_video_mode(struct uvd *uvd, struct video_window *vw)
{
  DEBUG(3,"requested x: %i, request y: %i, requested width= %i, requested height = %i",vw->x,vw->y,vw->width,vw->height);
  DEBUG(3,"requested flag = %x, chrom= %x",vw->flags,vw->chromakey);
  return 0;
}


static int sonixcam_search_header(struct uvd *uvd,int length)
{
  int offset=0;
  

  length = (length < RingQueue_GetLength(&uvd->dp)) ? length : RingQueue_GetLength(&uvd->dp);
 
  while(offset<length-6)
    {
      if((RING_QUEUE_PEEK(&uvd->dp,offset  ) == 0xff) &&
	 (RING_QUEUE_PEEK(&uvd->dp,offset+1) == 0xff) &&
	 (RING_QUEUE_PEEK(&uvd->dp,offset+2) == 0x00) &&
	 (RING_QUEUE_PEEK(&uvd->dp,offset+3) == 0xc4) &&
	 (RING_QUEUE_PEEK(&uvd->dp,offset+4) == 0xc4) &&
	 (RING_QUEUE_PEEK(&uvd->dp,offset+5) == 0x96) &&
	 (RING_QUEUE_PEEK(&uvd->dp,offset+6) == 0x00))
	{
	  return offset;
	}
      offset++;

    }
  return -1;
}

static void sonixcam_process_isoc(struct uvd *uvd, struct usbvideo_frame *frame)
{	
	struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;
	int maxline = cam->height;
	int header_offset=-1;
	int count=0;
	assert(frame != NULL);
	
	DEBUG(2, "Frame Scan State = %d, frame State = %d, queue length= %d", frame->scanstate,frame->frameState,RingQueue_GetLength(&uvd->dp));

	if(frame->scanstate == ScanState_Scanning)
	  {
	    /* Assuming the headers are at the beginning of the packet */

	    header_offset=sonixcam_search_header(uvd,RingQueue_GetLength(&uvd->dp));
	    DEBUG(2,"header at %d",header_offset);
	    if(header_offset>0)
	      {
		/* Parse until header_offset */
		frame->scanstate=ScanState_Lines;
		cam->ae_data1 = RING_QUEUE_PEEK(&uvd->dp,header_offset+8) + (RING_QUEUE_PEEK(&uvd->dp,header_offset+9) << 8);
		cam->ae_data2 = RING_QUEUE_PEEK(&uvd->dp,header_offset+10) + (RING_QUEUE_PEEK(&uvd->dp,header_offset+11) << 8);
	      }
	    else if(header_offset ==0)
	      {
		/* Drop this */
		RING_QUEUE_DEQUEUE_BYTES(&uvd->dp,SONIX_HEADER_LENGTH);
		
	      }
	    else
	      {
		/* Not found, continue scanning */
	      }
	  }
	
	if(frame->scanstate == ScanState_Scanning)
	  {
	    DEBUG(2,"keep scanning");
	    return;
	  }

	/* 
	 * Ok, we've got some data to parse
	 * the data is:
	 *
	 *  BGBGBGB....BG  \ height lines
	 *  GRGRGRG....GR  /
	 *  <-- width -->
	 *
	 *  so for one pixel, we only have the "real" value for only one color
	 *  We must "interpolate" the two others. (This is done in the post_process function)
	 */
	DEBUG(2,"Parsing data");
	while(frame->curline < maxline && (RingQueue_GetLength(&uvd->dp) >= cam->width*2))
	  {
	    /* We decode 2 lines by 2 lines */
	    int pixel=0;
	    
	    while(pixel<cam->width)
	      {
		char red=RING_QUEUE_PEEK(&uvd->dp,cam->width+2*(pixel/2)+1);
		char blue=RING_QUEUE_PEEK(&uvd->dp,2*(pixel/2));
		char green1=RING_QUEUE_PEEK(&uvd->dp,2*(pixel/2)+1);
		char green2=RING_QUEUE_PEEK(&uvd->dp,cam->width+2*(pixel/2));

		/* Do the red of the first line  */
		*(frame->data + (frame->curline * cam->width*3)+pixel*3) = 0;//red;
		*(frame->data + (frame->curline * cam->width*3)+(pixel+1)*3) = 0;//red;
		/* Do the red of the second line  */
		*(frame->data + ((frame->curline+1) * cam->width*3)+pixel*3) = 0;//red;
		*(frame->data + ((frame->curline+1) * cam->width*3)+(pixel+1)*3) = red;

		/* Do the green of the first line  */
		*(frame->data + (frame->curline * cam->width*3)+pixel*3+1) = 0;//green1;
		*(frame->data + (frame->curline * cam->width*3)+(pixel+1)*3+1) = green1;//green1;
		/* Do the green of the second line  */
		*(frame->data + ((frame->curline+1) * cam->width*3)+pixel*3+1) =green2;// green2;
		*(frame->data + ((frame->curline+1) * cam->width*3)+(pixel+1)*3+1) =0;// green2;


		/* Do the blue of the first line  */
		*(frame->data + (frame->curline * cam->width*3)+pixel*3+2) = blue;
		*(frame->data + (frame->curline * cam->width*3)+(pixel+1)*3+2) =0;// blue;
		/* Do the blue of the second line  */
		*(frame->data + ((frame->curline+1) * cam->width*3)+pixel*3+2) = 0;//blue;
		*(frame->data + ((frame->curline+1) * cam->width*3)+(pixel+1)*3+2) = 0;//blue;

		pixel+=2;
	      }
	    frame->curline+=2;
	    //	    frame->seqRead_Length += 3*2*cam->width;
	    count += 2*cam->width;
	    RING_QUEUE_DEQUEUE_BYTES(&uvd->dp,2*cam->width);
	  }
	DEBUG(2,"header: %i, length: %i",header_offset,RingQueue_GetLength(&uvd->dp));
	DEBUG(2,"Diff= %i",header_offset - count/3);

	if(frame->curline == maxline)
	  {
	    DEBUG(2,"got whole frame");
	    frame->frameState = FrameState_Done;
	    frame->curline = 0;
	    uvd->curframe = -1;
	    uvd->stats.frame_num++;
	    frame->seqRead_Length += 3*count;
	  }
	else
	  {
	    DEBUG(0,"Missing Lines");
	    frame->curline = 0;
	    frame->frameState = FrameState_Done;
	    uvd->curframe = -1;
	    uvd->stats.frame_num++;
	    frame->seqRead_Length += 3*count;

	    RING_QUEUE_DEQUEUE_BYTES(&uvd->dp,header_offset-count);
	    
	  }

	RING_QUEUE_DEQUEUE_BYTES(&uvd->dp,SONIX_HEADER_LENGTH);

}

#define R_(_source,_w,_h) (*(_source + 3*width*_h + 3*_w + 0)) 
#define G_(_source,_w,_h) (*(_source + 3*width*_h + 3*_w + 1)) 
#define B_(_source,_w,_h) (*(_source + 3*width*_h + 3*_w + 2))

static void sonixcam_post_process(struct uvd *uvd, struct usbvideo_frame *frame)
{
  struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;
  u8 *source;
  int height = cam->height;
  int width = cam->width;
  int x,y,xp,xn,yp,yn;
  int R,G,B;

  DEBUG(2,"Postprocessing the frame (%ix%i)\n",width,height);
  /*De-interlace frame->data */
  
  source = (char *)vmalloc(3*width*height);

  if(!source)
    {
      DEBUG(0,"Failed kmalloc");
      return;
    }
  DEBUG(2,"Source: %p",source);
  memcpy(source,frame->data,3*width*height);
  memset(frame->data,0,3*width*height);

  for(y=0;y<height;y++)
    {
      yp = y -1;yn=y+1;
      if(yp<0) yp = -yp;
      if(yn>=height) yn = 2*height -1 - yn;


      for(x=0;x<width;x++)
	{
	  xp= x -1;xn=x+1;
	  if(xp<0) xp = -xp;
	  if(xn>=width) xn = 2*width -1 - xn;

	  /* It seems we need to output BGR....*/

	  B = (R_(source,x, y ) << 2) +
	      (R_(source,xp,yp )) +
	      (R_(source,xp,yn)) +
	      (R_(source,xn,yn)) +
	      (R_(source,xn,yp)) +
	      (R_(source,xp, y) << 1) +
	      (R_(source,xn, y) << 1) +
	      (R_(source,x, yp) << 1) +
	      (R_(source,x, yp) << 1);
	  B_(frame->data,x,y) = (B + 2) / 4;

	  R = (B_(source,x, y ) << 2) +
	      (B_(source,xp,yp )) +
	      (B_(source,xp,yn)) +
	      (B_(source,xn,yn)) +
	      (B_(source,xn,yp)) +
	      (B_(source,xp, y) << 1) +
	      (B_(source,xn, y) << 1) +
	      (B_(source,x, yp) << 1) +
	      (B_(source,x, yp) << 1);
	  R_(frame->data,x,y) = (R + 2) / 4;
	  
	  G = (G_(source,x, y ) << 2) +
	      (G_(source,xp,y )) +
	      (G_(source,xn,y )) +
	      (G_(source,x, yp)) +
	      (G_(source,x, yn));
	  G_(frame->data,x,y) = (G + 2) / 4;
	}
      
      
    }

  vfree(source);
  
 	/* post process sensor */
	if (cam->sensor->post_process != NULL) {
		cam->sensor->post_process(uvd);
	}
}

static void sonixcam_configure_video(struct uvd *uvd)
{
	struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;

	memset(&uvd->vpic, 0, sizeof(uvd->vpic));
	memset(&uvd->vpic_old, 0x55, sizeof(uvd->vpic_old));

	uvd->debug=debug;
	uvd->vpic.colour = 0x7fff;
	uvd->vpic.hue = 0x7fff;
	uvd->vpic.brightness = 0x7fff; 
	uvd->vpic.contrast = 0x7fff;
	uvd->vpic.whiteness = 0x7fff;
	uvd->vpic.depth = 8;
	uvd->vpic.palette = VIDEO_PALETTE_RGB24;

	memset(&uvd->vcap, 0, sizeof(uvd->vcap));
	strcpy(uvd->vcap.name, "Sonix Webcam");
	uvd->vcap.type = VID_TYPE_CAPTURE;
	uvd->vcap.channels = 1;
	uvd->vcap.audios = 0;
	uvd->vcap.minwidth = camera_sizes[SIZE_352X288].width;
	uvd->vcap.minheight = camera_sizes[SIZE_352X288].height;
	uvd->vcap.maxwidth = camera_sizes[SIZE_352X288].width;
	uvd->vcap.maxheight = camera_sizes[SIZE_352X288].height;

	memset(&uvd->vchan, 0, sizeof(uvd->vchan));
	uvd->vchan.flags = 0 ;
	uvd->vchan.tuners = 0;
	uvd->vchan.channel = 0;
	uvd->vchan.type = VIDEO_TYPE_CAMERA;
	strcpy(uvd->vchan.name, "Sonix webcam");
	


	/* Turn off the led */
        DEBUG(1,"Turn off the led\n");
	sn9c102_write_reg(cam, 0x1,0x08);
}


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
static void *sonixcam_probe(struct usb_device *dev, unsigned int ifnum,
                        const struct usb_device_id *id)
{
#else
static int sonixcam_probe(struct usb_interface *intf, const struct usb_device_id *devid)
{
	struct usb_device *dev = interface_to_usbdev(intf);
#endif
	struct uvd *uvd = NULL;
	int i, nas;
	int actInterface=-1, inactInterface=-1, maxPS=0;
	unsigned char video_ep = 0;
	u8 id;
	struct sn9c102_sensor **sensor;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
	DEBUG(1, "sonixcam_probe(%p)", dev);
#else
	DEBUG(1, "sonixcam_probe(%p)", intf);
#endif

	/* We don't handle multi-config cameras */
	if (dev->descriptor.bNumConfigurations != 1)
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
                return NULL;
#else
                return -ENODEV;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
	DEBUG(1, "Sonixcam (device= 0x%04x)", le16_to_cpu(dev->descriptor.bcdDevice));
#else
	DEBUG(1, "Sonixcam (device= 0x%04x)", dev->descriptor.bcdDevice);
#endif

	/* Validate found interface: must have one ISO endpoint */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
	nas = dev->actconfig->interface[ifnum].num_altsetting;
#else
	nas = intf->num_altsetting;
#endif
	if (nas != 9) {
		err("Incorrect number of alternate settings (%d) for this camera!", nas);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
		return NULL;
#else
		return -ENODEV;
#endif
	}
	/* Validate all alternate settings */
	for (i=0; i < nas; i++) {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
		const struct usb_interface_descriptor *interface;
		interface = &dev->actconfig->interface[ifnum].altsetting[i];
		const struct usb_endpoint_descriptor *endpoint;
		if (interface->bNumEndpoints != 3) {
			err("Interface %d. has %u. endpoints!",
			    interface->bInterfaceNumber,
			    (unsigned)(interface->bNumEndpoints));
			return NULL;
		}
		endpoint = &interface->endpoint[0];
#else
		const struct usb_host_interface *interface;
		interface = &intf->altsetting[i];
		const struct usb_endpoint_descriptor *endpoint;
		if (interface->desc.bNumEndpoints != 3) {
			err("Interface %d. has %u. endpoints!",
			    interface->desc.bInterfaceNumber,
			    (unsigned)(interface->desc.bNumEndpoints));
			return -ENODEV;
		}
		endpoint = &interface->endpoint[0].desc;
#endif

		DEBUG(1, "found endpoint: addr: 0x%2.2x maxps = 0x%4.4x",
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
		    endpoint->bEndpointAddress, le16_to_cpu(endpoint->wMaxPacketSize));
#else
		    endpoint->bEndpointAddress, endpoint->wMaxPacketSize);
#endif
		if (video_ep == 0)
			video_ep = endpoint->bEndpointAddress;
		else if (video_ep != endpoint->bEndpointAddress) {
			err("Alternate settings have different endpoint addresses!");
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
			return NULL;
#else
			return -ENODEV;
#endif
		}
		if ((endpoint->bmAttributes & 0x03) != 0x01) {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
			err("Interface %d. has non-ISO endpoint!",
			    interface->bInterfaceNumber);
			return NULL;
#else
			err("Interface %d. has non-ISO endpoint!",
			    interface->desc.bInterfaceNumber);
			return -ENODEV;
#endif
		}
		if ((endpoint->bEndpointAddress & 0x80) == 0) {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
			err("Interface %d. has ISO OUT endpoint!",
			    interface->bInterfaceNumber);
			return NULL;
#else
			err("Interface %d. has ISO OUT endpoint!",
			    interface->desc.bInterfaceNumber);
			return -ENODEV;
#endif
		}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
		DEBUG(1, "alternate setting: 0x%x, PacketSIze 0x%x\n",i,le16_to_cpu(endpoint->wMaxPacketSize));
		if (le16_to_cpu(endpoint->wMaxPacketSize) == 0) {
#else
		DEBUG(1, "alternate setting: 0x%x, PacketSIze 0x%x\n",i,endpoint->wMaxPacketSize);
		if (endpoint->wMaxPacketSize == 0) {
#endif
			if (inactInterface < 0)
				inactInterface = i;
			else {
				err("More than one inactive alt. setting!");
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
				//return NULL;
#else
				return -ENODEV;
#endif
			}
		} else {
			if (i == 8) {
				/* This one is the requested one */
				actInterface = i;
			}
		}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
		if(le16_to_cpu(endpoint->wMaxPacketSize) > maxPS)
			maxPS = le16_to_cpu(endpoint->wMaxPacketSize);
#else
		if(endpoint->wMaxPacketSize > maxPS)
			maxPS = endpoint->wMaxPacketSize;
#endif
	}
	if(actInterface == -1) {
		err("Cant find required endpoint");
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
		return NULL;
#else
		return -ENODEV;
#endif
	}

	DEBUG(1, "Selecting requested active setting=%d. maxPS=%d.", actInterface, maxPS);

	uvd = usbvideo_AllocateDevice(cams);
	if (uvd != NULL) {
		struct sn9c102_device *cam = (struct sn9c102_device *)(uvd->user_data);
		/* Here uvd is a fully allocated uvd object */
		for(i = 0; i < USBVIDEO_NUMSBUF; i++) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
			cam->sts_urb[i] = usb_alloc_urb(FRAMES_PER_DESC);
#else
			cam->sts_urb[i] = usb_alloc_urb(FRAMES_PER_DESC, GFP_KERNEL);
#endif
			if(cam->sts_urb[i] == NULL) {
				while(i--) {
					usb_free_urb(cam->sts_urb[i]);
				}
				err("can't allocate urbs");
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
				return NULL;
#else
				return -ENOMEM;
#endif
			}
		}
		RESTRICT_TO_RANGE(size, SIZE_352X288, SIZE_352X288);
		cam->width = camera_sizes[size].width;
		cam->height = camera_sizes[size].height;
		cam->size = size;

		uvd->flags = 0;
		uvd->debug = debug;
		uvd->dev = dev;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
		uvd->iface = dev->actconfig->interface[ifnum].altsetting[0].bInterfaceNumber;
#else
		uvd->iface = intf->altsetting->desc.bInterfaceNumber;
#endif
		uvd->ifaceAltInactive = inactInterface;
		uvd->ifaceAltActive = actInterface;
		uvd->video_endp = video_ep;
		uvd->iso_packet_len = maxPS;
		uvd->paletteBits = 1L << VIDEO_PALETTE_RGB24;
		uvd->defaultPalette = VIDEO_PALETTE_RGB24;
		uvd->canvas = VIDEOSIZE(352, 288);
		uvd->videosize = VIDEOSIZE(cam->width, cam->height);

		/* check for SN9C10x */
		cam->usbdev = uvd->dev;
		id = sn9c102_read_reg(cam, 0x00);
		DEBUG(0, "SN9C10x camera controller ID=%d", id);
		
		/* init sensor/i2c sub-system */
		for (sensor = sensor_table; *sensor != NULL; sensor++) {
			cam->sensor = *sensor;
			if (cam->sensor->probe(uvd) >= 0) {
				DEBUG(0, "Detected '%s' image sensor", cam->sensor->name);
				break;
			}
		}
		if (*sensor == NULL) {
			DEBUG(0, "No image sensor found!");
		}
		
		/* Initialize sonixcam specific data */
		sonixcam_configure_video(uvd);

		i = usbvideo_RegisterVideoDevice(uvd);
		uvd->max_frame_size = (352 * 288 * 3);
		if (i != 0) {
			err("usbvideo_RegisterVideoDevice() failed.");
			uvd = NULL;
		}
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
	if (uvd) {
		return uvd;
	}
			return NULL;
#else
	if (uvd) {
		usb_set_intfdata (intf, uvd);
		return 0;
	}
			return -EIO;
#endif
}


static void sonixcam_free_uvd(struct uvd *uvd)
{
	int i;
	struct sn9c102_device *cam = (struct sn9c102_device *)uvd->user_data;

	for (i=0; i < USBVIDEO_NUMSBUF; i++) {
		usb_free_urb(cam->sts_urb[i]);
		cam->sts_urb[i] = NULL;
	}
}


static struct usb_device_id id_table[] = {
	{ USB_DEVICE(0x0c45, 0x6009) }, /* trust spacec@m 120 */
	{ USB_DEVICE(0x0c45, 0x600D) }, /* trust spacec@m 120 */
	{ USB_DEVICE(0x0c45, 0x6029) }, /* trust spacec@m 120 */
	{ USB_DEVICE(0x0c45, 0x6001) }, /* ? */
	{ USB_DEVICE(0x0c45, 0x6005) }, /* Sweex webcam mini */
	{ USB_DEVICE(0x0c45, 0x60ab) }, /* ? */
	{ USB_DEVICE(0x0c45, 0x6028) }, /* ? = sn9c10x + PAS202B */
	{ }  /* Terminating entry */
};

static int __init sonixcam_init(void)
{
	struct usbvideo_cb cbTbl;
	info(DRIVER_DESC " " DRIVER_VERSION);
	memset(&cbTbl, 0, sizeof(cbTbl));
	cbTbl.probe = sonixcam_probe;

	cbTbl.processData = sonixcam_process_isoc;
	cbTbl.videoStart = sonixcam_video_start;
	cbTbl.videoStop = sonixcam_video_stop;
	cbTbl.userFree = sonixcam_free_uvd;
	cbTbl.postProcess = sonixcam_post_process;
	cbTbl.setVideoMode=sonixcam_set_video_mode;
	cbTbl.adjustPicture = sonixcam_adjust_picture;
	return usbvideo_register(
		&cams,
		MAX_CAMERAS,
		sizeof(struct sn9c102_device),
		"sonixcam",
		&cbTbl,
		THIS_MODULE,
		id_table);
}


static void __exit sonixcam_cleanup(void)
{
	usbvideo_Deregister(&cams);
}


MODULE_DEVICE_TABLE(usb, id_table);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Thomas Cougnard <tgnard@free.fr>");
MODULE_DESCRIPTION(DRIVER_DESC);


MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug level: 0-3 (default=0)");


module_init(sonixcam_init);
module_exit(sonixcam_cleanup);
