/*	common/initiator_negotiate.c
 *
 *	This defines the functions used in the Login phase
 *	by the iSCSI initiator for parameter negotiation
 *
 *	vi: set autoindent tabstop=8 shiftwidth=4 :
 *
 *	This file contains auxilliary functions for iscsi initiator 
 *	code that are responsible for dealing with error recovery.
 *
 *	Copyright (C) 2001-2003 InterOperability Lab (IOL)
 *	University of New Hampshier (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote 
 *	products derived from this software without specific prior 
 * 	written permission.
*/

/*	
 * The code for redirection during login is based on code first written
 * by Scott Cranston of Equal Logic (scranston@equallogic.com).
*/

/*
 * Converted to:  indent -kr -i8 -ts8 -sob -l80 -ss -bs -psl on 10/8/2003 .
 * 
 */

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/socket.h>
#include <linux/ip.h>

#include "iscsi_common.h"
#include "debug.h"
#include "my_memory.h"

/* chap support - CHONG */
#include "../security/chap/chap.h"
#include "../security/srp/srp.h"
#include "../security/misc/misc_func.h"

#include "text_param.h"
#include "initiator_negotiate.h"

/*	check the correctness of target ip address which should be in format of
 *	TargetAddress=domainname[:port][,portal-group-tag]
 */
int
check_target_address(char *value, int when)
{
	int retval = 0;

	return retval;
}

/*	when get here, we already know that opcode is login response.
 *	returns 1 if valid redirection, 0 if other success, -1 if any error
 */
static int
check_login_response(struct generic_pdu *inputpdu,
		     struct generic_pdu *outputpdu)
{
	int retval = 0;
	char *tsix;
	struct iscsi_targ_login_rsp *login_rsp;
	struct iscsi_init_login_cmnd *login_req;

	TRACE(TRACE_ENTER_LEAVE, "Enter check_login_response\n");

	tsix = "TSIH";

	login_rsp = (struct iscsi_targ_login_rsp *) inputpdu;
	login_req = (struct iscsi_init_login_cmnd *) outputpdu;

	/* check that reserved fields in the login response are all zero */
	if ((login_rsp->opcode & OLD_X_BIT)
	    || (login_rsp->flags & (W_BIT | BRO_BIT))
	    || login_rsp->rsvd1 || login_rsp->rsvd2 || login_rsp->rsvd3) {
		TRACE_WARNING
		    ("reserved fields in Login Response header not all 0\n");
	}

	if ((inputpdu->status_class != STAT_CLASS_SUCCESS)) {
		/* the target sent a login reject */

		TRACE_ERROR
		      ("Login Response with non-zero StatusClass 0x%02x, "
		      "StatusDetail 0x%02x\n", inputpdu->status_class,
		      inputpdu->status_detail);

		/*  Draft 20, Section 5.3.1 Login Phase Start
		 */

		if ((inputpdu->flags & T_BIT)
		    || (inputpdu->flags & CSG) != 0
		    || (inputpdu->flags & NSG) != 0) {
			/* protocol error */
			TRACE_ERROR
			    ("non-zero Status-Class MUST have T=0, CSG=0 & NSG=0\n");
			retval = -1;
		}

		if (inputpdu->status_class == STAT_CLASS_REDIRECTION) {
			/* target redirecting login to different IP 
				address/Port number */
			if (inputpdu->status_detail <
			    STAT_DETAIL_TARG_MOVED_TEMP
			    && inputpdu->status_detail >
			    STAT_DETAIL_TARG_MOVED_PERM) {
				TRACE_ERROR
				    ("invalid Status Detail 0x%02x for Redirection\n",
				     inputpdu->status_detail);
				retval = -1;
			} else {
				/* valid redirection detail, there 
				 * should be a TargetAddress
				 * key attached to this PDU containing 
			 	 * the redirection info 
				 */
				retval = 1;
			}
		} else if (inputpdu->status_class == STAT_CLASS_INITIATOR) {
			/* Initiator Error (not a format error) */
			if (inputpdu->status_detail >
			    STAT_DETAIL_INVALID_DURING_LOGIN) {
				TRACE_ERROR
				    ("invalid Status Detail 0x%02x for Initiator Error\n",
				     inputpdu->status_detail);
			}
			retval = -1;
		} else if (inputpdu->status_class == STAT_CLASS_TARGET) {
			/* Target Error */
			if (inputpdu->status_detail >
			    STAT_DETAIL_OUT_OF_RESOURCE) {
				TRACE_ERROR
				    ("invalid Status Detail 0x%02x for Target Error\n",
				     inputpdu->status_detail);
			}
			retval = -1;
		} else {
			/* Bad status class */
			TRACE_ERROR("invalid Status Class 0x%02x\n",
				    inputpdu->status_class);
			retval = -1;
		}
	}

	/* status_class is 0, status_detail should also be 0 */
	else if ((inputpdu->status_detail != 0x0)) {
		TRACE_ERROR("Login Response with zero StatusClass but non-zero "
			    "StatusDetail 0x%02x\n", inputpdu->status_detail);
		retval = -1;
	}

	else if ((inputpdu->flags & CSG) != (outputpdu->flags & CSG)) {
		/* target's state not the same as initiator's state */
		/*  protocol error */
		TRACE_ERROR
		    ("inconsistant state! Target in state %d, initiator in %d\n",
		     (inputpdu->flags & CSG) >> CSG_SHIFT,
		     (outputpdu->flags & CSG) >> CSG_SHIFT);
		retval = -1;
	}

	else if (memcmp(login_rsp->isid, login_req->isid, 6) != 0) {
		/* target changed the isid */
		/*  protocol error */
		TRACE_ERROR("target changed the isid\n");
		retval = -1;
	}

	else if (!(inputpdu->flags & T_BIT)) {
		/* target does not want to change stage */
		if (inputpdu->flags & NSG) {
			/* check reserved field, but only give 
			 * warning if we find errors 
			 */
			if (TRACE_TEST(TRACE_ISCSI)) {
			      TRACE_WARNING("NSG %d should be 0 when T bit is 0\n",
			      inputpdu->flags & NSG);
			}
		}
	}

	/* input T bit is set to 1, target wants to change stage */
	else if (!(outputpdu->flags & T_BIT)) {
		/* but initiator did not want to change stage yet */
		/*  protocol error */
		TRACE_ERROR
		    ("Login Response with T bit set but Initiator's T bit "
		     "not set \n");
		retval = -1;
	}

	else if ((inputpdu->flags & NSG) <=
		 ((inputpdu->flags & CSG) >> CSG_SHIFT)
		 || (inputpdu->flags & NSG) == NSG2) {
		/* target wants to go to an illegal stage */
		/*  protocol error */
		TRACE_ERROR("target trying to change to illegal stage %d\n",
			    inputpdu->flags & NSG);
		retval = -1;
	}

	else if ((inputpdu->flags & NSG) == NSG3) {
		/* target going to FFP, can only do this if initiator 
		 * asked for it 
		 *  Draft 20, Section 5.3 Login Phase
		 */
		if ((outputpdu->flags & NSG) != NSG3) {
			TRACE_ERROR
			    ("target trying to change to FFP, initiator wanted "
			     "to change to Operational stage only\n");
			retval = -1;
		} else if (login_rsp->tsih == 0) {	/* protocol error */
			TRACE_ERROR
			    ("target's final login response has %s = 0\n",
			     tsix);
			retval = -1;
		}
	} else if (login_req->tsih != login_rsp->tsih) {
		TRACE_ERROR("target changed initiator's %s from %u to %u\n",
			    tsix, be16_to_cpu(login_req->tsih),
			    be16_to_cpu(login_rsp->tsih));
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave check_login_response retval = %d\n",
	      retval);
	return retval;
}

/*	Called to send a Login PDU already set up in outputpdu
 *	and then to read back a login response into inputpdu
 *	Returns 1 on redirection, 0 on other success, -1 on any error
 */
static int
write_and_read(struct socket *sock,
	struct generic_pdu *inputpdu,
	struct generic_pdu *outputpdu,
	int *target_version_max, 
	int *target_version_active, 
	__u32 flags)
{
	int padding = 0;
	int retval = 0;
	__u32 size;

	TRACE(TRACE_ENTER_LEAVE, "Enter write_and_read\n");

	/* send output */
	if (iscsi_send_msg(sock, outputpdu, flags) < 0) {
		TRACE(TRACE_DEBUG, "iscsi_send_msg failed\n");
		return -1;
	}

	/* After sending the outputpdu, reset it's text len to 0 */
	outputpdu->text_length = 0;

	/* wait for input from target */
	size = ISCSI_HDR_LEN;

	/* NO packets sent or received during login will have 
	 * digests of any kind 
	 */

	if (iscsi_recv_msg(sock, size, (char *) inputpdu, flags) < 0) {
		TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
		return -1;
	}

	/* Also get the text associated with it */
	inputpdu->text_length = be32_to_cpu(inputpdu->length);

	/* Include the padding bytes too */
	padding = inputpdu->text_length + ((-inputpdu->text_length) & 3);

	if (inputpdu->text_length > MAX_TEXT_LEN) {
		TRACE_ERROR
		    ("DSL %u greater than default MaxRecvDataSegmentLength %d\n",
		     inputpdu->text_length, MAX_TEXT_LEN);
		return -1;
	}

	if (padding > 0
	    && iscsi_recv_msg(sock, padding, inputpdu->text, flags) < 0) {
		TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
		return -1;
	}

	if ((inputpdu->opcode & ISCSI_OPCODE) == ISCSI_TARG_LOGIN_RSP) {
		/* as expected, this is a login response, check it */
		TRACE(TRACE_ISCSI, "Got Login Rsp, ITT %u\n",
		      ntohl(inputpdu->init_task_tag));
		if (TRACE_TEST(TRACE_ISCSI_FULL))
			print_targ_login_rsp((struct iscsi_targ_login_rsp *)
					     inputpdu);

		/* Check the initiator task tag */
		if (inputpdu->init_task_tag != outputpdu->init_task_tag) {
			if (TRACE_TEST(TRACE_ISCSI)) {
			      TRACE_WARNING("received init_task_tag = %d, "
					"expected %d\n",
					be32_to_cpu(inputpdu->init_task_tag),
					be32_to_cpu(outputpdu->init_task_tag));
			}
		}

		/* Check recved version_max and version_active fields 
		 * for consistency 
		 */
		if (inputpdu->version_max < inputpdu->version_active) {
			if (TRACE_TEST(TRACE_ISCSI)) {
			      TRACE_WARNING("received version_max = 0x%02x "
			      "less than received version_active 0x%02x\n",
			      inputpdu->version_max, inputpdu->version_active);
			}
		}

		/* The received version_max and version_active fields 
		 * must not change 
		 */

		/*  Draft 20, Section 10.13.1 Version-max (for Login Response)
		 *  "All Login responses within the Login Phase MUST 
		 *  carry the same Version-max.
		 *
		 *  Draft 20, Section 10.13.2 Version-active (for 
		 *  Login Response)
		 *  "All Login responses within the Login Phase 
		 *  MUST carry the same Version-active.
		 *  The initiator MUST use the value presented as 
		 *  response to the first login request."
		 */
		if (*target_version_max < 0) {
			/* first login response, just remember its values */
			*target_version_max = inputpdu->version_max;
			*target_version_active = inputpdu->version_active;
		} else if (*target_version_max != inputpdu->version_max) {
			if (TRACE_TEST(TRACE_ISCSI)) {
			      TRACE_WARNING("received version_max = 0x%02x "
			      "differs from received version_max 0x%02x\n",
			      inputpdu->version_max, *target_version_max);
			}
		} else if (*target_version_active != inputpdu->version_active) {
			if (TRACE_TEST(TRACE_ISCSI)) {
			      TRACE_WARNING("received version_active = 0x%02x"
			      " differs from received version_active 0x%02x\n",
			      inputpdu->version_active, *target_version_active);
			}
		}

		/*  Check the version_active field to see if we can use it
		 *  For now, we send version_min == version_max, so 
		 *  target has no choice
		 */
		if (inputpdu->version_active != outputpdu->version_active) {
			TRACE_ERROR
			    ("received version_active = 0x%02x, expected 0x%02x\n",
			     inputpdu->version_active,
			     outputpdu->version_active);
			retval = -1;
		}

		/* Check the C bit, which we do not use yet */
		else if (inputpdu->flags & C_BIT) {	
			/* C bit is set in Login Response */
			/*  Draft 20, Section 10.13.7 C (Continue) Bit
			 *  "A Login Response with the C bit set must have the
			 *  T bit set to 0."
			 */
			if (inputpdu->flags & T_BIT) {
				TRACE_ERROR
				    ("C=1 in Login Response requires T=0, not T=1\n");
			} else {	/* for now */
				TRACE_ERROR
				    ("C=1 in Login Response not implemented yet\n");
			}
			retval = -1;
		} else
			retval = check_login_response(inputpdu, outputpdu);
	}

	else if ((inputpdu->opcode & ISCSI_OPCODE) == ISCSI_TARG_RJT) {
		/* protocol error */
		TRACE_ERROR("got Reject PDU\n");
		print_targ_rjt((struct iscsi_targ_rjt *) inputpdu);
		retval = -1;
	}

	else {
		TRACE_ERROR("received opcode = 0x%02x, expected 0x%02x\n",
			    inputpdu->opcode & ISCSI_OPCODE,
			    ISCSI_TARG_LOGIN_RSP);
		retval = -1;	/* this is fatal */
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave write_and_read, retval %d\n", retval);

	return retval;
}

/*	states to drive authentication steps during security phase */
enum security_steps {
	ss_initial,             // 0
	ss_find_chap_a_i_c,	// 1
	ss_find_chap_n_r,	// 2
	ss_find_srp_g_s,	// 3
	ss_find_srp_b,		// 4
	ss_find_srp_h,		// 5
	ss_done,		// 6
	ss_error		// 7
};



/* forward defs */


int
ss_initial_func(enum security_steps *security_step, 
	struct unknown_key **	unknown_key_list,
	struct parameter_type *p_param_tbl,
	struct auth_parameter_type p_auth_param,
	int noperational,
	struct parameter_type *auth_p,
	struct generic_pdu *inputpdu,
	struct generic_pdu *outputpdu);

int
ss_find_chap_a_i_c_func(enum security_steps *security_step, 
	struct unknown_key **	unknown_key_list,
	struct parameter_type *p_param_tbl,
	struct auth_parameter_type p_auth_param,
	int noperational,
	struct parameter_type *auth_p,
	struct generic_pdu *inputpdu,
	struct generic_pdu *outputpdu,
	__u32 * got_keys,
	__u32 * got_value);

int
ss_find_chap_n_r_func(enum security_steps *security_step,
        struct unknown_key **   unknown_key_list,
        struct parameter_type *p_param_tbl,
        struct auth_parameter_type p_auth_param,
        int noperational,
        struct generic_pdu *inputpdu,
        struct generic_pdu *outputpdu,
        __u32 * got_keys,
        __u32 * got_value,
	int *count);

int
ss_find_srp_g_s_func( enum security_steps *security_step,
        struct unknown_key **   unknown_key_list,
        struct parameter_type *p_param_tbl,
        struct auth_parameter_type p_auth_param,
        int noperational,
        struct generic_pdu *inputpdu,
        struct generic_pdu *outputpdu,
        __u32 * got_keys );

int
ss_find_srp_b_func( enum security_steps *security_step,
        struct unknown_key **   unknown_key_list,
        struct parameter_type *p_param_tbl,
        struct auth_parameter_type p_auth_param,
        int noperational,
        struct generic_pdu *inputpdu,
        struct generic_pdu *outputpdu,
        __u32 * got_keys );

int
ss_find_srp_h_func( enum security_steps *security_step,
        struct unknown_key **   unknown_key_list,
        struct parameter_type *p_param_tbl,
        struct auth_parameter_type p_auth_param,
        int noperational,
        struct generic_pdu *inputpdu,
        struct generic_pdu *outputpdu,
        __u32 * got_keys,
	int *count);
int
ss_done_func( enum security_steps *security_step,
        struct unknown_key **   unknown_key_list,
        struct parameter_type *p_param_tbl,
        struct auth_parameter_type p_auth_param,
        int noperational,
        struct generic_pdu *inputpdu,
        struct generic_pdu *outputpdu,
        __u32 * got_keys,
	int *count);




/*	When called, 1st login PDU is set up in outputpdu except for T, 
* 	NSG, CSG.  Returns 1 for redirection, 0 for all other successes, 
* 	-1 for any error
*/
static int
initiator_security_phase(struct socket *sock,
	struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
	struct auth_parameter_type p_auth_param,
	__u32 * cur_cmd_sn,
	int *exp_stat_sn,
	int *max_send_length,
	int noperational,
	__u32 who_called,
	struct generic_pdu *inputpdu,
	struct generic_pdu *outputpdu,
	int *target_version_max,
	int *target_version_active,
	__u32 connection_flags,
	__u64 * login_flags,
	struct unknown_key **unknown_key_list)
{
	int add_length = 0;
	int retval = -1;	/* assume we will fail */
	struct parameter_type *auth_p = NULL;
	int count;
	__u32 got_value ;
	__u32 got_keys = 0;
	enum security_steps security_step = ss_initial;
	int result ;

	TRACE(TRACE_ENTER_LEAVE, "Enter initiator_security_phase\n");

	/*  offer no transition on first login pdu sent to target */
	outputpdu->flags &= ~(CSG | NSG | T_BIT);

	add_length = scan_table_and_process(sock,
					    p_param_tbl,
					    SECURITY_PARAM |
					    INFORMATIONAL_PARAM, 
					    0,
					    INITIATOR,
					    inputpdu, 
					    outputpdu,
					    connection_flags, 
					    login_flags);
	if (add_length < 0) {
		retval = add_length;
		goto out;
	}
	outputpdu->text_length = add_length;

	for (count = 0; count < LOOP_TIMES; count++) {
		TRACE(TRACE_DEBUG, "Send security login, key length %d\n",
		      add_length);
		outputpdu->cmd_sn = cpu_to_be32(*cur_cmd_sn);
		TRACE(TRACE_ISCSI, "Send Login, ITT %u\n",
		      ntohl(outputpdu->init_task_tag));

		if (write_and_read (sock, inputpdu, outputpdu, 
					target_version_max,
		     			target_version_active, 
					connection_flags) < 0) {
			goto out;
		}

		if (inputpdu->status_class != 0x00) {
			TRACE_ERROR
			    ("Login Reject from target, Class 0x%02x, Detail 0x%02x\n",
			     inputpdu->status_class, inputpdu->status_detail);
			goto out;
		}

		*exp_stat_sn = be32_to_cpu(((struct iscsi_targ_login_rsp *)
					    inputpdu)->stat_sn) + 1;
		outputpdu->exp_stat_sn = cpu_to_be32(*exp_stat_sn);

		/*  assume we will want no transition on our next output pdu */
		outputpdu->flags &= ~(CSG | NSG | T_BIT);

		/* Build next transmission PDU */
		add_length = scan_input_and_process(sock,
						    p_param_tbl,
						    SECURITY_PARAM |
						    INFORMATIONAL_PARAM,
						    TARGETPORTALGROUPTAG_FLAG,
						    INITIATOR, 
						    max_send_length,
						    who_called, 
						    inputpdu,
						    outputpdu, 
						    connection_flags,
						    login_flags,
						    unknown_key_list);

		if (add_length < 0) {
			retval = add_length;
			goto out;
		}
		outputpdu->text_length = add_length;

		TRACE(TRACE_ISCSI_FULL,
		      "Initiator switch on security_step %d\n", security_step);

		switch (security_step) {
		case ss_initial:
			{
			TRACE(TRACE_ISCSI_FULL,
				      "security_step %d\n", security_step);

			if ((result = ss_initial_func(&security_step, 
					unknown_key_list,
					p_param_tbl,
					p_auth_param,
					noperational,
					auth_p,
					inputpdu,
					outputpdu )) < 0 )
	
				{ TRACE_ERROR("ss_initial_func -1\n");
					goto out;
				}  else if (result == 1)
						continue;
					else break;
			}
			break;

		case ss_find_chap_a_i_c:
			{	// 718

			/* previously sent CHAP_A list, 
			 * expecting CHAP_A,
			 * CHAP_I, CHAP_C
			 */

			TRACE(TRACE_ISCSI_FULL,
				      "security_step %d\n", security_step);

				if(ss_find_chap_a_i_c_func(&security_step, 
							unknown_key_list,
							p_param_tbl,
							p_auth_param,
							noperational,
							auth_p,
							inputpdu,
							outputpdu,
							&got_keys,
							&got_value) < 0 ){
					goto out;
				}				
			}
			break ;

		case ss_find_chap_n_r:
			{ 
			/* expecting target to reply to 
			 * our previously sent challenge 
 			 */
			TRACE(TRACE_ISCSI_FULL,
				      "security_step %d\n", security_step);

				if(ss_find_chap_n_r_func(&security_step,
  						unknown_key_list,
					        p_param_tbl,
					        p_auth_param,
					        noperational,
					        inputpdu,
					        outputpdu,
						&got_keys,
        					&got_value,
						&count) < 0 )
				{ goto out;
				}
			}
			break;


		case ss_find_srp_g_s:
			{
			/* previously sent SRP_U, 
			 * expecting SRP_GROUP, SRP_s 
			 */
			TRACE(TRACE_ISCSI_FULL,
				      "security_step %d\n", security_step);


				if(ss_find_srp_g_s_func(&security_step,
  						unknown_key_list,
					        p_param_tbl,
					        p_auth_param,
					        noperational,
					        inputpdu,
					        outputpdu,
						&got_keys) < 0 )
				{ goto out;
				}
			
			}
			break;

		case ss_find_srp_b:
			{ 
			/* previously sent SRP_A and SRP_GROUP, 
			 * expecting SRP_B 
			 */

			TRACE(TRACE_ISCSI_FULL,
				      "security_step %d\n", security_step);
				if (ss_find_srp_b_func( &security_step,
        					unknown_key_list,
        					p_param_tbl,
        					p_auth_param,
        					noperational,
        					inputpdu,
        					outputpdu,
        					&got_keys) < 0 )
				{ goto out;
				}
			
			}
					
			break;

		case ss_find_srp_h:
			{ 
			
			/* expecting target to send SRP_HM */

			TRACE(TRACE_ISCSI_FULL,
				      "security_step %d\n", security_step);

				if (ss_find_srp_h_func( &security_step,
        					unknown_key_list,
        					p_param_tbl,
        					p_auth_param,
        					noperational,
        					inputpdu,
        					outputpdu,
        					&got_keys,
						&count) < 0 )
				{ goto out;
				}
			}
			break;

		case ss_done:
			{ 
				/* done with authentication 
				 * (may never have done any) 
				 */

			TRACE(TRACE_ISCSI_FULL,
				      "security_step %d\n", security_step);
				if (ss_done_func( &security_step,
        					unknown_key_list,
        					p_param_tbl,
        					p_auth_param,
        					noperational,
        					inputpdu,
        					outputpdu,
        					&got_keys,
						&count) < 0 )
				{ goto out;
				}
			}
			break;

		default:
			/* should NEVER happen */
			TRACE_ERROR("unknown security step %d\n",
				    security_step);
			goto out;

		}		/* switch */
	}			/* loop */

	if (count == LOOP_TIMES) {
		TRACE_ERROR("infinite loop in security phase negotiations\n");
	} else
		retval = 0;	/* our pessimism was unjustified, we succeeded ! */

      out:

	TRACE(TRACE_ENTER_LEAVE, "Leave initiator_security_phase, retval %d\n",
	      retval);
	return retval;
}

/*
 * Setup up outgoing PDU to send AuthMethod 
 */
int
ss_initial_func(enum security_steps *security_step, 
	struct unknown_key **	unknown_key_list,
	struct parameter_type *p_param_tbl,
	struct auth_parameter_type p_auth_param,
	int noperational,
	struct parameter_type *auth_p,
	struct generic_pdu *inputpdu,
	struct generic_pdu *outputpdu)
{
    	char *dummy_string;
	int got_bitmask;
	struct unknown_key *key;
	int temp_len;
	char *srp_u = NULL;
    	struct SRP_Context *srp_ctx = NULL;
	struct CHAP_Context *init_ctx = NULL;
    	struct CHAP_Context *targ_ctx = NULL;


	TRACE(TRACE_ENTER_LEAVE, "Enter ss_initial_func() \n");

	ALLOCATE_MAX_TEXT_LEN(dummy_string);


	/* looking to find AuthMethod=CHAP or 
	 * AuthMethod=SRP 
	 * Also deal with unknown keys: CHAP, SRP 
	 * keys illegal in this state 
	 */
	for (key = *unknown_key_list; key != NULL; key = key->next) {
		if (!key->processed) {
			key->processed = 1;
			if ((got_bitmask = print_bad_security_key(key))) {
				print_not_allowed_security_key(key);
			}
			goto out;
		}
	}

	if ((auth_p = find_flag_parameter(AUTHMETHOD_FLAG, p_param_tbl))
	    == NULL) {		/* should NEVER happen */
		TRACE_ERROR("AuthMethod parameter not found\n");
		goto out;
	}

	if (!IS_KEY_GOT_FROM_OTHER_SIDE(auth_p->neg_info)) {
		/* Target did not send us an AuthMethod key, find out why */
		if (!IS_KEY_SENT_TO_OTHER_SIDE(auth_p->neg_info)) {
			/* AuthMethod key not going to be negotiated */
			*security_step = ss_done;
			if (noperational == 0)
				outputpdu->flags |= (NSG3 | T_BIT);
			else
				outputpdu->flags |= (NSG1 | T_BIT);
		} else {
			/* AuthMethod key sent but reply not received yet,
			   stay in ss_initial state */
			goto out_cont;
		}
	} else if (!strcmp(auth_p->str_value, NONE)) {
		/* not going to do authentication, offer transition */
		*security_step = ss_done;
		if (noperational == 0)
			outputpdu->flags |= (NSG3 | T_BIT);
		else
			outputpdu->flags |= (NSG1 | T_BIT);
	} else if (!strcmp(auth_p->str_value, CHAP)) {
		/* start CHAP authentication */
		TRACE(TRACE_DEBUG, "Start initiator chap authentication\n");
		init_ctx = p_auth_param.chap_local_ctx;
		targ_ctx = p_auth_param.chap_peer_ctx;
		if (init_ctx == NULL || targ_ctx == NULL) {
			TRACE_ERROR("CHAP context not initialized yet\n");
			goto out;
		}
		CHAP_SetAlgorithm(MD5_ALGORITHM, targ_ctx);
		CHAP_SetAlgorithm(MD5_ALGORITHM, init_ctx);

		temp_len = sprintf(dummy_string, "CHAP_A=");
		temp_len += CHAP_GetAlgorithmList(dummy_string + temp_len);
		TRACE(TRACE_ISCSI, "Attach key: %s\n", dummy_string);
		strcpy(outputpdu->text + outputpdu->text_length, dummy_string);
		outputpdu->text_length += temp_len + 1;

		*security_step = ss_find_chap_a_i_c;
	} else if (!strcmp(auth_p->str_value, SRP)) {
		/* start SRP authentication */
		TRACE(TRACE_DEBUG, "Start initiator srp authentication\n");
		srp_ctx = p_auth_param.srp_ctx;
		if (srp_ctx == NULL) {
			TRACE_ERROR("SRP context not initialized yet\n");
			goto out;
		}

		/* look up SRP user name, error if none */
		if ((srp_u = SRP_Initiator_GetUsername(srp_ctx))
		    == NULL) {
			TRACE_ERROR("uo User Name in SRP database\n");
			goto out;
		}

		/* generate and attach key SRP_U=value */
		temp_len = sprintf(dummy_string, "%s=%s", SRP_U, srp_u);
		TRACE(TRACE_ISCSI, "Attach key: %s\n", dummy_string);
		strcpy(outputpdu->text + outputpdu->text_length, dummy_string);
		outputpdu->text_length += temp_len + 1;
		my_kfree((void **) &srp_u, "SRP_U");

		/* generate and attach key SRP_TargetAuth=value */
		temp_len =
		    sprintf(dummy_string, "%s=%s",
			    SRP_TARGETAUTH,
			    p_auth_param.
			    auth_flags & USE_TARGET_CONFIRMATION ? YES : NO);
		TRACE(TRACE_ISCSI, "Attach key: %s\n", dummy_string);
		strcpy(outputpdu->text + outputpdu->text_length, dummy_string);
		outputpdu->text_length += temp_len + 1;

		*security_step = ss_find_srp_g_s;
	} else {		/* should NEVER happen */
		TRACE_ERROR
		    ("unknown AuthMethod value \"%s\"\n", auth_p->str_value);
		goto out;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_initial_func() Success Next %d\n",
				*security_step);
	FREE_STRING(dummy_string);
	return 1;

out_cont:
	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_initial_func() Continue Next %d\n",
				*security_step);
	FREE_STRING(dummy_string);
	return 0;

out:
	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_initial_func() Failure Next %d\n",
				*security_step);
	FREE_STRING(dummy_string);
	return -1;

}
int
ss_find_chap_a_i_c_func(enum security_steps *security_step, 
	struct unknown_key **	unknown_key_list,
	struct parameter_type *p_param_tbl,
	struct auth_parameter_type p_auth_param,
	int noperational,
	struct parameter_type *auth_p,
	struct generic_pdu *inputpdu,
	struct generic_pdu *outputpdu,
	__u32 * got_keys,
	__u32 * got_value)

{
    	char *dummy_string;
	int got_bitmask;
	int add_length;
	struct unknown_key *key;
	int temp_len;
	char *our_chap_n = NULL;
	char *our_chap_c = NULL;
	char *our_chap_r = NULL;
	char *his_chap_c = NULL;
	__u8 chap_i = 0;
	struct CHAP_Context *init_ctx ;
    	struct CHAP_Context *targ_ctx ;

        init_ctx = p_auth_param.chap_local_ctx;
        targ_ctx = p_auth_param.chap_peer_ctx;

	TRACE(TRACE_ENTER_LEAVE, "Enter ss_find_chap_a_i_c() \n");
	ALLOCATE_MAX_TEXT_LEN(dummy_string);

	/* previously sent CHAP_A list, expecting CHAP_A, 
	 * CHAP_I, CHAP_C 
	 */
	for (key = *unknown_key_list; key != NULL; key = key->next) {
		if (!key->processed) {
			key->processed = 1;
			got_bitmask = print_bad_security_key(key);
			if (got_bitmask == GOT_CHAP_A) {
				if (check_step_key_number
				    (key, got_keys, GOT_CHAP_A, UINT_MAX,
				     got_value))
					goto out;
				if (!CHAP_SetAlgorithm(*got_value, targ_ctx)
				    || !CHAP_SetAlgorithm(*got_value, init_ctx)) {
					TRACE_ERROR
					    ("illegal %s algorithm \"%s\"\n",
					     key->keyname, key->keyvalue);
					goto out;
				}
			} else if (got_bitmask == GOT_CHAP_I) {
				if (check_step_key_number
				    (key, got_keys, GOT_CHAP_I, 
					255, got_value))
					goto out;
				chap_i = *got_value;
			} else if (got_bitmask == GOT_CHAP_C) {
				if (check_step_key(key, got_keys, GOT_CHAP_C))
					goto out;
				his_chap_c = key->keyvalue;
			} else {
				if (got_bitmask) {
					print_not_allowed_security_key(key);
				}
				goto out;
			}
		}
	}
	if ((*got_keys & (GOT_CHAP_A | GOT_CHAP_I | GOT_CHAP_C))
	    == (GOT_CHAP_A | GOT_CHAP_I | GOT_CHAP_C)) {
		/* have all target keys needed to respond 
		 * to his challenge 
		 */
		our_chap_n = CHAP_GetName(targ_ctx);
		if (our_chap_n == NULL) {
			TRACE_ERROR("CHAP_N to Target not configured\n");
			goto out;
		}
		our_chap_r =
		    CHAP_GetResponse(chap_i, his_chap_c,
				     MAX_CHAP_BINARY_LENGTH, targ_ctx);
		if (our_chap_r == NULL) {
			TRACE_ERROR("CHAP_R to Target cannot be generated\n");
			my_kfree((void **) &our_chap_n, "CHAP Get Name");
			goto out;
		}
		temp_len = sprintf(dummy_string, "CHAP_N=%s", our_chap_n);
		TRACE(TRACE_ISCSI, "Attach key: %s\n", dummy_string);
		my_kfree((void **) &our_chap_n, "CHAP Get Name");
		add_length = (temp_len + 1);
		temp_len =
		    sprintf(dummy_string + add_length, "CHAP_R=%s", our_chap_r);
		TRACE(TRACE_ISCSI, "Attach key: %s\n",
		      dummy_string + add_length);
		my_kfree((void **) &our_chap_r, "CHAP Response");
		add_length += (temp_len + 1);
		if (p_auth_param.auth_flags & USE_TARGET_CONFIRMATION) {
			chap_i = CHAP_GetIdentifier(init_ctx);
			our_chap_c = CHAP_GetChallenge(init_ctx);
			if (our_chap_c == NULL) {
				TRACE_ERROR("second exchange of chap failed\n");
				goto out;
			}
			temp_len =
			    sprintf(dummy_string +
				    add_length, "CHAP_I=%d", chap_i);
			TRACE(TRACE_ISCSI,
			      "Attach key: %s\n", dummy_string + add_length);
			add_length += (temp_len + 1);
			temp_len =
			    sprintf(dummy_string +
				    add_length, "CHAP_C=%s", our_chap_c);
			TRACE(TRACE_ISCSI,
			      "Attach key: %s\n", dummy_string + add_length);
			add_length += (temp_len + 1);
			my_kfree((void **) &our_chap_c, "CHAP Challenge");

			/* expect target reply to challenge */
			*security_step = ss_find_chap_n_r;
		} else {
			/* we are not challenging target */
			*security_step = ss_done;
		}

		memcpy(outputpdu->text +
		       outputpdu->text_length, dummy_string, add_length);
		outputpdu->text_length += add_length;

		/* regardless of whether we challenge target or
		 * not, this is the last pdu we need to send, so 
		 * offer transition 
		 */
		if (noperational == 0)
			outputpdu->flags |= (NSG3 | T_BIT);
		else
			outputpdu->flags |= (NSG1 | T_BIT);
	}


	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_find_chap_a_i_c()OK Next %d\n",
				*security_step);
	FREE_STRING(dummy_string);
	return 0;
out:

	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_find_chap_a_i_c() Failure Next %d\n",
				*security_step);
	FREE_STRING(dummy_string);
	return -1;

}				

int
ss_find_chap_n_r_func(enum security_steps *security_step,
        struct unknown_key **   unknown_key_list,
        struct parameter_type *p_param_tbl,
        struct auth_parameter_type p_auth_param,
        int noperational,
        struct generic_pdu *inputpdu,
        struct generic_pdu *outputpdu,
        __u32 * got_keys,
        __u32 * got_value,
	int *count)

{

	int got_bitmask;
	struct unknown_key *key;
	char *his_chap_r = NULL;
        struct CHAP_Context *init_ctx ;
        struct CHAP_Context *targ_ctx ;

	init_ctx = p_auth_param.chap_local_ctx;
	targ_ctx = p_auth_param.chap_peer_ctx;

	TRACE(TRACE_ENTER_LEAVE, "Enter ss_find_chap_n_r_func()  %d\n",
				*security_step);

	/* expecting target to reply to our previously sent challenge */
	for (key = *unknown_key_list; key != NULL; key = key->next) {
		if (!key->processed) {
			key->processed = 1;
			got_bitmask = print_bad_security_key(key);
			if (got_bitmask == GOT_CHAP_N) {
				if (check_step_key(key, got_keys, GOT_CHAP_N))
					goto out;
				if (strcmp(key->keyvalue, init_ctx->name)) {
					TRACE_ERROR
					    ("got %s = \"%s\", expected \"%s\"\n",
					     key->
					     keyname,
					     key->keyvalue, init_ctx->name);
					goto out;
				}
			} else if (got_bitmask == GOT_CHAP_R) {
				if (check_step_key(key, got_keys, GOT_CHAP_R))
					goto out;
				his_chap_r = key->keyvalue;
			} else {
				if (got_bitmask) {
					print_not_allowed_security_key(key);
				}
				goto out;
			}
		}
	}
	if ((*got_keys & (GOT_CHAP_A | GOT_CHAP_I | GOT_CHAP_C |
		GOT_CHAP_N | GOT_CHAP_R)) ==
	    (GOT_CHAP_A | GOT_CHAP_I | GOT_CHAP_C | GOT_CHAP_N | GOT_CHAP_R)) {

		/* target sent complete reply to our 
		 * challenge, check it now 
		 */
		if (CHAP_CheckResponse
		    (his_chap_r, MAX_CHAP_BINARY_LENGTH, init_ctx) <= 0) {
			TRACE_ERROR("CHAP failed to authenticate target\n");
			goto out;
		}

		TRACE(TRACE_ISCSI, "CHAP successfully authenticated target\n");

		if (inputpdu->flags & T_BIT) {
			/* target agreed to transition out of security stage */
			*count = LOOP_TIMES;
			goto break_out;
		}

		TRACE(TRACE_ISCSI, "end up the security phase\n");
		*security_step = ss_done;
		if (noperational == 0)
			outputpdu->flags |= (NSG3 | T_BIT);
		else
			outputpdu->flags |= (NSG1 | T_BIT);
	}

break_out:
	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_find_chap_n_r.c_func Ok Next %d\n",
				*security_step);
	return 0;

out:
	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_find_chap_n_r.c_func Failure Next %d\n",
				*security_step);
	return -1;

}

int
ss_find_srp_g_s_func( enum security_steps *security_step,
        struct unknown_key **   unknown_key_list,
        struct parameter_type *p_param_tbl,
        struct auth_parameter_type p_auth_param,
        int noperational,
        struct generic_pdu *inputpdu,
        struct generic_pdu *outputpdu,
        __u32 * got_keys)

{		

    	char *dummy_string;
	int got_bitmask;
	struct unknown_key *key;
	int temp_len,i;
	char *ptr = NULL;
	char **srp_argv = NULL;
	char *srp_group =NULL;
	char *srp_a =NULL;

	struct SRP_Context *srp_ctx = p_auth_param.srp_ctx;

	TRACE(TRACE_ENTER_LEAVE," Enter ss_find_srp_g_s_func %d srp_ctx %p\n",
				*security_step,srp_ctx );

	ALLOCATE_MAX_TEXT_LEN(dummy_string);

	/* previously sent SRP_U, expecting SRP_GROUP, SRP_s */
	for (key = *unknown_key_list; key != NULL; key = key->next) {
		if (!key->processed) {
			key->processed = 1;
			got_bitmask = print_bad_security_key(key);
			if (got_bitmask == GOT_SRP_GROUP) {
				if (check_step_key
				    (key, got_keys, GOT_SRP_GROUP))
					goto out;
				/* process the SRP_GROUP key value */
				/* first see how many items are in the list */
				/* always at least 1 item plus NULL at end */
				i = 2;	
				ptr = key->keyvalue;
				do {
					if (*ptr == '\0')
						break;
					if (*ptr++ == ',')
						i++;
				}
				while (1);

				if ((srp_argv = my_kmalloc(i * sizeof(void *),
						 "SRP argv")) == NULL)
					goto out;

				/* tokenize the value list into 
				 * an argv[] list 
				 */
				i = 0;
				ptr = key->keyvalue;
				srp_argv[0] = ptr;
				do {
					if (*ptr == '\0') {
						/* end of list */
						srp_argv[++i] = NULL;	
						break;
					}
					if (*ptr == ',') {
						/* nul-terminate string */
						*ptr = '\0';	
						/* start of next str */
						srp_argv[++i]
						    = ptr + 1;
					}
					ptr++;
				}
				while (1);

				if ((srp_group = SRP_Initiator_GetGroup(
					srp_argv, srp_ctx)) == NULL) {
					/* to give error message, fix 
					 *      tokenizing damage 
					 */
					for (i -= 2; i >= 0; i--) {
						ptr = srp_argv[i];
						*(ptr + strlen(ptr))
						    = ',';
					}
					TRACE_ERROR
					    ("desired group choice not in list %s\n",
					     key->keyvalue);
					goto out;
				} else {
					TRACE
					    (TRACE_ISCSI,
					     "Selected Group %s\n", srp_group);
				}
			} else if (got_bitmask == GOT_SRP_S) {
				if (check_step_key(key, got_keys, GOT_SRP_S))
					goto out;
					TRACE(TRACE_DEBUG,
					    "key %s\n",key->keyvalue);
				if (SRP_Initiator_SetSalt
				    (key->keyvalue, MAX_SRP_BINARY_LENGTH,
				     srp_ctx) == 0) {
					TRACE_ERROR("illegal SRP_s\n");
					goto out;
				}
			} else {
				if (got_bitmask) {
					print_not_allowed_security_key(key);
				}
				goto out;
			}
		}
	}
	if ((*got_keys & (GOT_SRP_GROUP | GOT_SRP_S))
	    == (GOT_SRP_GROUP | GOT_SRP_S)) {
		/* have all target keys needed to send him a challenge */

		if ((srp_a = SRP_Initiator_GetA(srp_ctx)) == NULL) {
			TRACE_ERROR("unable to generate SRP_A\n");
			goto out;
		}
		/* have srp_a and srp_group, send them */
		*security_step = ss_find_srp_b;

		/* generate and attach key SRP_A=value */
		temp_len = sprintf(dummy_string, "%s=%s", SRP_A, srp_a);
		TRACE(TRACE_ISCSI, "Attach key: %s\n", dummy_string);
		strcpy(outputpdu->text + outputpdu->text_length, dummy_string);
		outputpdu->text_length += temp_len + 1;

		/* generate and attach key SRP_GROUP=value */
		temp_len = sprintf(dummy_string, "%s=%s", SRP_GROUP, srp_group);
		TRACE(TRACE_ISCSI, "Attach key: %s\n", dummy_string);
		strcpy(outputpdu->text + outputpdu->text_length, dummy_string);
		outputpdu->text_length += temp_len + 1;

	}

	my_kfree((void **) &srp_group, "selected SRP_GROUP");
	my_kfree((void **) &srp_argv, "SRP argv");
	my_kfree((void **) &srp_a, "SRP_A");
	FREE_STRING(dummy_string);
	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_find_srp_g_s_func Ok Next %d\n",
				*security_step);
	return 0;

out:
	my_kfree((void **) &srp_group, "selected SRP_GROUP");
	my_kfree((void **) &srp_argv, "SRP argv");
	my_kfree((void **) &srp_a, "SRP_A");
	FREE_STRING(dummy_string);
	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_find_srp_g_s_func Failure Next %d\n",
				*security_step);

	return -1;
}
int
ss_find_srp_b_func( enum security_steps *security_step,
        struct unknown_key **   unknown_key_list,
        struct parameter_type *p_param_tbl,
        struct auth_parameter_type p_auth_param,
        int noperational,
        struct generic_pdu *inputpdu,
        struct generic_pdu *outputpdu,
        __u32 * got_keys)


{		

    	char *dummy_string;
	int got_bitmask;
	struct unknown_key *key;
	int temp_len;
	char *srp_m =NULL;

	struct SRP_Context *srp_ctx = p_auth_param.srp_ctx;

	TRACE(TRACE_ENTER_LEAVE, "Enter ss_find_srp_b Ok Next %d\n",
				*security_step);

	ALLOCATE_MAX_TEXT_LEN(dummy_string);

	/* previously sent SRP_A and SRP_GROUP, expecting SRP_B */
	for (key = *unknown_key_list; key != NULL; key = key->next) {
		if (!key->processed) {
			key->processed = 1;
			got_bitmask = print_bad_security_key(key);
			if (got_bitmask == GOT_SRP_B) {
				if (check_step_key(key, got_keys, GOT_SRP_B))
					goto out;
				if (SRP_Initiator_SetB
				    (key->keyvalue,
				     MAX_SRP_BINARY_LENGTH, srp_ctx) == 0) {
					TRACE_ERROR
					    ("unacceptable value for SRP_B\n");
					goto out;
				}
			} else {
				if (got_bitmask) {
					print_not_allowed_security_key(key);
				}
				goto out;
			}
		}
	}
	if ((*got_keys & (GOT_SRP_GROUP | GOT_SRP_S | GOT_SRP_B))
	    == (GOT_SRP_GROUP | GOT_SRP_S | GOT_SRP_B)) {
		if ((srp_m = SRP_Initiator_GetM(srp_ctx)) == NULL) {
			TRACE_ERROR("unable to generate SRP_M\n");
			goto out;
		}

		/* generate and attach key SRP_M=value */
		temp_len = sprintf(dummy_string, "%s=%s", SRP_M, srp_m);
		TRACE(TRACE_ISCSI, "Attach key: %s\n", dummy_string);
		strcpy(outputpdu->text + outputpdu->text_length, dummy_string);
		outputpdu->text_length += temp_len + 1;

		if (p_auth_param.auth_flags & USE_TARGET_CONFIRMATION) {
			/* we previously sent TargetAuth=Yes, so we 
			 * expect the target to reply with SRP_HM next 
			 */
			*security_step = ss_find_srp_h;
		} else
			*security_step = ss_done;

		/*  regardless of whether we expect target to 
		 * send back SRP_HM or not, this is the last 
		 * pdu we need to send, so offer transition 
		 */
		if (noperational == 0)
			outputpdu->flags |= (NSG3 | T_BIT);
		else
			outputpdu->flags |= (NSG1 | T_BIT);
	}

	my_kfree((void **) &srp_m, "SRP_M");
	FREE_STRING(dummy_string);
	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_find_srp_b Ok Next %d\n",
				*security_step);
	return 0;

out:
	my_kfree((void **) &srp_m, "SRP_M");
	FREE_STRING(dummy_string);
	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_find_srp_b Failure Next %d\n",
				*security_step);
	return -1;
}

/* expecting target to send SRP_HM */

int
ss_find_srp_h_func( enum security_steps *security_step,
        struct unknown_key **   unknown_key_list,
        struct parameter_type *p_param_tbl,
        struct auth_parameter_type p_auth_param,
        int noperational,
        struct generic_pdu *inputpdu,
        struct generic_pdu *outputpdu,
        __u32 * got_keys,
	int *count )


{		

	int got_bitmask;
	struct unknown_key *key;
	struct SRP_Context *srp_ctx = p_auth_param.srp_ctx;

	TRACE(TRACE_ENTER_LEAVE, "Enter ss_find_srp_h %d\n",
				*security_step);

	/* expecting target to send SRP_HM */
	for (key = *unknown_key_list; key != NULL; key = key->next) {
		if (!key->processed) {
			key->processed = 1;
			got_bitmask = print_bad_security_key(key);
			if (got_bitmask == GOT_SRP_HM) {
				if (check_step_key(key, got_keys, GOT_SRP_HM))
					goto out;
				if (SRP_Initiator_SetHM
				    (key->keyvalue,
				     MAX_SRP_BINARY_LENGTH, srp_ctx) == 0) {
					TRACE_ERROR
					    ("SRP authentication failed\n");
					goto out;
				}
			} else {
				if (got_bitmask) {
					print_not_allowed_security_key(key);
				}
				goto out;
			}
		}
	}
	if ((*got_keys & (GOT_SRP_GROUP | GOT_SRP_S | GOT_SRP_B | GOT_SRP_HM))
	    == (GOT_SRP_GROUP | GOT_SRP_S | GOT_SRP_B | GOT_SRP_HM)) {
		/* target sent complete reply to our 
		 * challenge, check it now 
		 */

		TRACE(TRACE_ISCSI, "SRP successfully authenticated target\n");

		if (inputpdu->flags & T_BIT) {
			/* target agreed to transition out of security stage */
			*count = LOOP_TIMES;
			goto break_out;
		}

		TRACE(TRACE_ISCSI, "end up the security phase\n");
		*security_step = ss_done;
		if (noperational == 0)
			outputpdu->flags |= (NSG3 | T_BIT);
		else
			outputpdu->flags |= (NSG1 | T_BIT);
	}

break_out:

	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_find_srp_h %d\n",
				*security_step);
	return 0;

out:
	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_find_srp_h Failure Next %d\n",
				*security_step);
	return -1;
}


/* 
 * done with authentication (may never have done any) 
 */
int
ss_done_func( enum security_steps *security_step,
        struct unknown_key **   unknown_key_list,
        struct parameter_type *p_param_tbl,
        struct auth_parameter_type p_auth_param,
        int noperational,
        struct generic_pdu *inputpdu,
        struct generic_pdu *outputpdu,
        __u32 * got_keys,
	int *count)

{		

	int got_bitmask;
	struct unknown_key *key;

	TRACE(TRACE_ENTER_LEAVE, "Enter ss_done %d\n",
				*security_step);

	/* deal with unknown keys: CHAP, SRP keys illegal 
	   in this state */
	for (key = *unknown_key_list; key != NULL; key = key->next) {
		if (!key->processed) {
			key->processed = 1;
			if ((got_bitmask = print_bad_security_key(key))) {
				print_not_allowed_security_key(key);
			}
			goto out;
		}
	}

	if (inputpdu->flags & T_BIT) {
		/* target agreed to transition out of security stage */
		*count = LOOP_TIMES;
		goto break_out; 
	}

	TRACE(TRACE_ISCSI, "end up the security phase\n");
	if (noperational == 0)
		outputpdu->flags |= (NSG3 | T_BIT);
	else
		outputpdu->flags |= (NSG1 | T_BIT);

break_out:

	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_find_srp_h %d\n",
				*security_step);
	return 0;

out:
	TRACE(TRACE_ENTER_LEAVE, "Leaving ss_find_srp_h Failure Next %d\n",
				*security_step);

	return -1;
}


/*	Returns 1 for redirection, 
 *	0 for all other successes, -1 for any error 
*/
static int
initiator_operational_phase(struct socket *sock, 
	struct parameter_type
	p_param_tbl[MAX_CONFIG_PARAMS], 
	__u32 * cur_cmd_sn,
	int *exp_stat_sn, 
	int *max_send_length,
	__u32 when_called, 
	struct generic_pdu *inputpdu,
	struct generic_pdu *outputpdu,
	int *target_version_max, 
	int *target_version_active,
	__u32 connection_flags, 
	__u64 * login_flags,
	struct unknown_key **unknown_key_list)
{
	int add_length = 0;
	int count, retval = 0;

	TRACE(TRACE_ENTER_LEAVE, "Entering initiator_operational_phase\n");

	count = 0;
	do {
		TRACE(TRACE_DEBUG, "count %d, inputpdu->T_BIT 0x%02x\n",
		      count, inputpdu->flags & T_BIT);

		/* assume there is only one message is to be 
		 * sent in this phase 
		 */
		outputpdu->flags &= (~CSG | NSG);
		outputpdu->flags |= CSG1 | T_BIT | NSG3;	

		if ((*login_flags & MORE_TO_SEND_FLAG) && count < 3) {
			/* set output.T = 0 */
			outputpdu->flags &= ~(T_BIT);
		}

		*login_flags &= ~MORE_TO_SEND_FLAG;

		/* scan through our table and process any 
		 * unsent O or I keys 
		 *      -- if any unsent S keys then protocol error 
  		 */

		add_length = scan_table_and_process(sock,
						    p_param_tbl,
						    OPERATIONAL_PARAM |
						    INFORMATIONAL_PARAM,
						    TARGETNAME_FLAG,
						    INITIATOR, inputpdu,
						    outputpdu, connection_flags,
						    login_flags);
		if (add_length < 0) {
			retval = add_length;
			goto out;
		}

		outputpdu->text_length += add_length;

		/*  Draft 20, Section 10.12.8 CmdSN (for Login Request):
		   "CmdSN is either the initial command sequence number 
		    of a session (for the first Login request of a 
		    session - the "leading" login), or the command 
		    sequence number in the command stream if the login
		   is for a new connection in an existing session."
		 */

		/*  This means we always have to pick up the session's 
		    CmdSN and put it into the pdu just as we are about 
		    to send it on the wire.  However, we do not 
		    advance it during login phase.
		 */
		outputpdu->cmd_sn = cpu_to_be32(*cur_cmd_sn);

		TRACE(TRACE_ISCSI, "Send Login, ITT %u\n",
		      ntohl(outputpdu->init_task_tag));
		if ((retval = write_and_read(sock, inputpdu, outputpdu,
					     target_version_max,
					     target_version_active,
					     connection_flags)) < 0) {
			goto out;
		}

		/* pass back the exp_stat_sn from the last login 
		 * response we get 
		 */

		/* exp_stat_sn is always 1 greater than last statsn received */
		*exp_stat_sn = 1 +
		    be32_to_cpu(((struct iscsi_targ_login_rsp *) inputpdu)->
				stat_sn);

		/* set up the reply (if any) to have the correct ExpStatSN */
		outputpdu->exp_stat_sn = cpu_to_be32(*exp_stat_sn);

		/* process each key attached to the input packet */
		add_length = scan_input_and_process(sock,
						    p_param_tbl,
						    OPERATIONAL_PARAM |
						    INFORMATIONAL_PARAM,
						    TARGETPORTALGROUPTAG_FLAG,
						    INITIATOR, max_send_length,
						    when_called, inputpdu,
						    outputpdu, connection_flags,
						    login_flags,
						    unknown_key_list);
		if (add_length < 0) {
			retval = add_length;
			goto out;
		}
		outputpdu->text_length = add_length;

		/* see if Login Request was a redirect */
		if (retval == STAT_CLASS_REDIRECTION) {
			goto out;
		}

		if (*login_flags & FIRST_FLAG) {
			/* this was first Login Request/Login Response pair */
			if (!(*login_flags & TARGETPORTALGROUPTAG_FLAG)
				&& *login_flags & TARGETNAME_FLAG) {

				/* RFC 3720 Section 5.3.1 Login Phase Start
				 * During the Login Phase the iSCSI target 
				 * MUST return the TargetPortalGroupTag 
				 * key with the first Login Response
				 * PDU with which it is allowed to do 
				 * so (i.e., the first Login Response 
				 * issued after the first Login Request with
				 * the C bit set to 0) for all session types
				 * when TargetName is given and the response
				 * is not a redirection."
				 */
				TRACE(TRACE_ISCSI, "\"TargetPortalGroupTag\" "
				      "not in first Login Response from target\n");
				retval = -1;
				goto out;
			}
			*login_flags &= ~FIRST_FLAG;
		}

		count++;	/* to prevent infinite loops */
		if (count >= LOOP_TIMES) {
			TRACE_ERROR
			    ("infinite loop in operational phase negotiations\n");
			retval = -1;
			goto out;
		}

		if ((inputpdu->flags & T_BIT) && add_length > 0) {
			/* protocol error  */
			TRACE_ERROR
			    ("target has T bit set, but initiator still needs to send more\n");
			retval = -1;
			goto out;
		}

	} while (add_length > 0 || !(inputpdu->flags & T_BIT));

	/* When loop ends, operational parameter negotiation stage 
	 * is finished 
	 *
	 * Check whether all parameters that were sent in this stage 
	 * have received their responses 
 	 */
	if (check_neg_responses(p_param_tbl, 1) < 0) {
		/* for now, ignore this (message already given) */
	}

	if (*login_flags & MORE_TO_SEND_FLAG) {	/* protocol error */
		TRACE_ERROR("not all parameters negotiated\n");
	}

      out:

	TRACE(TRACE_ENTER_LEAVE,
	      "Leaving initiator_operational_phase, retval %d\n", retval);

	return retval;
}

/*	This local function performs parameter negotiation on initiator side.
 *	Called only from initiator_parameter_negotiate() after all local
 *	working storage has been allocated.
 *	Returns 1 if Redirection, 0 if otherwise successful, -1 on error.
 *	If successful:
 *		login phase is complete,
 *		digests have been enabled as necessary,
 *		all parameters passed by pointer have been updated,
 *		we are now in FFP.
 */
static int
init_parameter_negotiate(struct socket *sock,
	struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
	struct auth_parameter_type auth_param,
	__u32 * cur_cmd_sn,
	__u32 * connection_flags,
	__u32 when_called,
	int *exp_stat_sn,
	int *max_send_length,
	int *max_recv_length,
	struct generic_pdu *inputpdu,
	struct generic_pdu *outputpdu,
	struct unknown_key **unknown_key_list)
{
	int retval = 0;
	int nsecurity, ninformational, noperational;
	int target_version_max = -1, target_version_active = -1;
	__u64 login_flags;

	TRACE(TRACE_ENTER_LEAVE, "Enter init_parameter_negotiate\n");

	/* about to process the first Login Request/Response */
	login_flags = FIRST_FLAG;

	/* count the number of each kind of parameter still to be negotiated */
	scan_table_and_count(p_param_tbl, &nsecurity, &ninformational,
			     &noperational);

	/* to force a login phase, make it look like there is a 
	 * param of that type 
	 */
	if (*connection_flags & USE_SECURITY_PHASE)
		nsecurity++;
	if (*connection_flags & USE_OPERATIONAL_PHASE)
		noperational++;

	if (nsecurity + noperational + ninformational == 0) {
		/* no parameters at all to negotiate, and no phases forced! */
		nsecurity++;	/* might as well start in security phase */
	}

	if (nsecurity > 0) {
		/* have some security parameters to negotiate
		 * or want to force security phase in any case
		 * chap support - CHONG
		 */
		if ((retval =
		     initiator_security_phase(sock, p_param_tbl, 
						auth_param,
						cur_cmd_sn, 
						exp_stat_sn,
						max_send_length, 
						noperational,
						when_called, 
						inputpdu, 
						outputpdu,
						&target_version_max,
						&target_version_active,
						*connection_flags, 
						&login_flags,
						unknown_key_list)) < 0) {
			goto out;
		}

		ninformational = 0;	/* these have been sent in security phase */
	}

	if ((noperational + ninformational > 0 ||
	     (inputpdu->flags & NSG) == NSG1) && retval == 0) {
		/* enter the operational negotiation phase */
		if ((retval = initiator_operational_phase(sock, p_param_tbl,
							cur_cmd_sn,
							exp_stat_sn,
							max_send_length,
							when_called, 
							inputpdu,
							outputpdu,
							&target_version_max,
							&target_version_active,
							*connection_flags,
							&login_flags,
							unknown_key_list)) < 0)
							 { goto out;
		}
	}

      out:

	TRACE(TRACE_ENTER_LEAVE, "Leave init_parameter_negotiate, retval %d\n",
	      retval);
	return retval;
}

/*	This function is the entry point to this file to perform parameter
 *	negotiation on initiator side.
 *	Called only from iscsi_initiator_login() in iscsi_initiator.c.
 *	Returns 1 on Redirection, 0 on all other successes, -1 on any error.
 */
int
initiator_parameter_negotiate(struct socket *sock,
	struct parameter_type
	p_param_tbl[MAX_CONFIG_PARAMS],
	struct auth_parameter_type auth_param,
	__u32 *cur_cmd_sn, 
	__u32 login_task_tag,
	void *isid, 
	__u16 *tsih, 
	__u16 cid,
	__u32 *connection_flags, 
	__u32 version_number,
	__u32 when_called, 
	int *exp_stat_sn,
	int *max_send_length, 
	int *max_recv_length)
{
	int retval = 0;
	struct generic_pdu *inputpdu, *outputpdu;
	struct iscsi_init_login_cmnd *loginpdu;
	struct unknown_key *uptr, *unknown_key_list = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter initiator_parameter_negotiate\n");

	if ((inputpdu =
	     my_kmalloc(sizeof (struct generic_pdu), "Input pdu")) == NULL) {
		retval = -1;
		goto out;
	}

	if ((outputpdu =
	     my_kmalloc(sizeof (struct generic_pdu), "Output pdu")) == NULL) {
		my_kfree((void **) &inputpdu, "Input pdu");
		retval = -1;
		goto out;
	}

	/* Get memory for text portion of inputpdu and outputpdu */
	if ((inputpdu->text = my_kmalloc(MAX_TEXT_LEN, "Input text")) == NULL) {
		my_kfree((void **) &outputpdu, "Output pdu");
		my_kfree((void **) &inputpdu, "Input pdu");
		retval = -1;
		goto out;
	}
	inputpdu->text_length = 0;

	if ((outputpdu->text = my_kmalloc(MAX_TEXT_LEN, 
		"Output text")) == NULL) {
		my_kfree((void **) &inputpdu->text, "Input text");
		my_kfree((void **) &outputpdu, "Output pdu");
		my_kfree((void **) &inputpdu, "Input pdu");
		retval = -1;
		goto out;
	}
	outputpdu->text_length = 0;

	/* Set up the output pdu */
	loginpdu = (struct iscsi_init_login_cmnd *) outputpdu;

	/* initialize everything to 0 so reserved fields will be 0 */
	memset(loginpdu, 0, ISCSI_HDR_LEN);

	/* then set up those fields in the Login Request 
	 * PDU that will not change 
	 */
	loginpdu->opcode = ISCSI_INIT_LOGIN_CMND;	
	/* Login command always marked immediate */
	loginpdu->opcode |= I_BIT;
	loginpdu->version_max = version_number;
	loginpdu->version_min = version_number;
	memcpy(loginpdu->isid, isid, 6);
	loginpdu->tsih = cpu_to_be16(*tsih);
	loginpdu->init_task_tag = cpu_to_be32(login_task_tag);
	loginpdu->cid = cpu_to_be16(cid);

	TRACE(TRACE_DEBUG, "initiator_parameter_negotiate version_number %u\n",
	      version_number);

	/* chap support - CHONG */
	retval = init_parameter_negotiate(sock, 
					p_param_tbl, 
					auth_param, 
					cur_cmd_sn,
					connection_flags, 
					when_called, 
					exp_stat_sn,
					max_send_length, 
					max_recv_length, 
					inputpdu,
					outputpdu, 
					&unknown_key_list);

	/*  Draft 20, Section 5.2 Text Mode Negotiation
	 *  "Some parameters may be subject to integrity rules 
	 *  (e.g., parameter-x must not exceed parameter-y or 
	 *  parameter-u not 1 implies parameter-v  be Yes).  
	 * Whenever required, integrity rules are specified with the
	 *  keys.  Checking for compliance with the integrity rule must only be
	 *  performed after all the parameters are available (the existent and
	 *  newly negotiated).  An iSCSI target MUST perform integrity checking
	 *  before the new parameters take effect.  An initiator MAY perform
	 *  integrity checking."
	 */
	if (retval == 0) {	/* not a redirection */
		check_integrity_rules(p_param_tbl, *tsih);

		/*  Draft 20 Section 12.1 HeaderDigest and DataDigest
		 *  "When Initiator and Target agree on a digest, 
		 *  this digest MUST be used for every PDU in 
		 *  Full Feature Phase."
		 */
		set_digestflags(p_param_tbl, connection_flags);

		/*  if we (the initiator) sent a 
		 * MaxRecvDataSegmentLength (draft 18)
		 *  value to the target, we have to now start using it in FFP
		 */
		set_connection_recv_length(p_param_tbl, max_recv_length);

		if (*tsih == 0) {
			/* first login phase of a new session 
			 * completed successfully,
			 * pass back the tsih assigned to this session 
			 * by the target 
			 */
			*tsih = be16_to_cpu(((struct iscsi_targ_login_rsp *)
					    inputpdu)->tsih);
		}
	}

	/* free up memory allocated for any unknown keys we received */
	while (unknown_key_list != NULL) {
		if (!unknown_key_list->processed) {
			TRACE_ERROR("unknown key \"%s=%s\"\n",
				    unknown_key_list->keyname,
				    unknown_key_list->keyvalue);
		}
		uptr = unknown_key_list->next;
		my_kfree((void **) &unknown_key_list->keyvalue, "Key Value");
		my_kfree((void **) &unknown_key_list->keyname, "Key Name");
		my_kfree((void **) &unknown_key_list, "Unknown key");
		unknown_key_list = uptr;
	}

	/* Free all allocated memory */
	my_kfree((void **) &outputpdu->text, "Output text");
	my_kfree((void **) &inputpdu->text, "Input text");
	my_kfree((void **) &outputpdu, "Output pdu");
	my_kfree((void **) &inputpdu, "Input pdu");

      out:

	TRACE(TRACE_ENTER_LEAVE,
	      "Leave initiator_parameter_negotiate, retval %d\n", retval);

	return retval;
}
