/*	initiator/initiator_error_rec.c
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This file contains auxilliary functions for iscsi initiator
 *      code that are responsible for dealing with error recovery.
 *
 *	Copyright (C) 2001-2004 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 * 	written permission.
*/
#include "iscsi_initiator.h"
#include "initiator_utilities.h"
#include "initiator_error_rec.h"

static DECLARE_MUTEX_LOCKED(init_sess_recovery_sem);
static struct task_struct *init_rec_thread = NULL;

/************************************************************************
* The Main Error Recovery Handler Called from iscsi_initiator when it
* finds any error.
* When called, session lock must be locked
* Returns >0 on success
*	  =0 on failure, session lock is locked
*	  <0 on failure, session lock is NOT locked
************************************************************************/
int
init_do_error_recovery(struct init_error_rec *err_rec)
{

	int retval = 0;		/* assume this will fail with lock held */
	struct connection *curr_conn;

	TRACE(TRACE_ENTER_LEAVE, "Enter init_do_error_recovery \n");

	if (err_rec == NULL) {
		TRACE_ERROR("Error Recovery structure is NULL\n");
		goto out;
	}

	if (err_rec->curr_conn == NULL) {
		TRACE_ERROR
		    ("NULL Connection Pointer in Error Recovery struct\n");
		goto out;
	}

	curr_conn = err_rec->curr_conn;

	switch (curr_conn->my_session->oper_param->ErrorRecoveryLevel) {
	case SESSION_RECOVERY:
		retval = init_session_recovery(err_rec);
		break;
	case DIGEST_RECOVERY:
	case CONNECTION_RECOVERY:
		retval = init_digest_recovery(err_rec);
		break;
	}

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave init_do_error_recovery, retval %d\n",
	      retval);

	return retval;
}

/************************************************************************
 * executed only by rx thread.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * This function handles the session recovery. It creates a recovery
 * thread that will do session recovery.
 * While the recovery thread is running the rx thread is blocked waiting for it.
 * Returns <0 forced failure, session lock is NOT locked
 * In all cases, current thread should not continue after the return!
 ************************************************************************/
int
init_session_recovery(struct init_error_rec *err_rec)
{
	struct connection *current_connection;
	struct session *current_session;

	TRACE(TRACE_ENTER_LEAVE, "Enter init_session_recovery \n");

	if (!global_hostdata->init_sessrec_flg) {

		/* globally mark that session recovery is in progress, so
		 * attempts by SCSI to queue new commands are caught
		 */
		global_hostdata->init_sessrec_flg = 1;

		current_connection = err_rec->curr_conn;
		current_session = current_connection->my_session;

		init_MUTEX_LOCKED(&init_sess_recovery_sem);

		/* Release any resources held by this thread */
		UNH_UNLOCK(&current_session->sess_lock,
			   current_connection->rx_lock_flags);

		TRACE(TRACE_ISCSI, "Starting initiator session recovery\n");

		/* Create a recovery thread */
		kernel_thread((int (*)(void *)) init_recovery_thread,
				(void *) err_rec, 0);

		/* Wait till the recovery thread completes processing */
		down_interruptible(&init_sess_recovery_sem);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave init_session_recovery\n");

	return -ESRCH;	/* any error will do, so give "No such process" */
}

/* Create a new recovery session for the failing session before cleaning-up
 * the failing session
 */
struct sess_rec_group *
set_rec_group(struct scsi_cmnd * Cmnd)
{
	struct sess_rec_group *sess_grp;
	struct sess_rec_group *prev_sess_grp;
	struct sess_rec_cmnd *cmnd;

	TRACE(TRACE_ENTER_LEAVE, "Enter set_rec_group\n");

	prev_sess_grp = sess_grp = global_hostdata->sess_rec_head;
	for (; sess_grp != NULL; sess_grp = sess_grp->next) {
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
		if (sess_grp->cmnd_list_head->Cmnd->device->id == Cmnd->device->id)
#else
		if (sess_grp->cmnd_list_head->Cmnd->target == Cmnd->target)
#endif
			break;
		prev_sess_grp = sess_grp;
	}

	if (sess_grp == NULL) {
		sess_grp = (struct sess_rec_group *)
				    my_kmalloc(sizeof(struct sess_rec_group),
					       "sess_rec_group");

		if (!sess_grp) {
			goto out;
		}

		memset(sess_grp, 0, sizeof(struct sess_rec_group));

		if (global_hostdata->sess_rec_head == NULL) {
			TRACE(TRACE_DEBUG, "Previous Recovery list empty - "
			      "creating new list\n");
			global_hostdata->sess_rec_head = sess_grp;
		} else {
			prev_sess_grp->next = sess_grp;
		}
	}

	cmnd =
	    (struct sess_rec_cmnd *)my_kmalloc(sizeof(struct sess_rec_cmnd),
						"sess_rec_cmnd");

	if (unlikely(cmnd == NULL)) {
		my_kfree((void **)&sess_grp, "sess_rec_group");
		goto out;
	}

	memset(cmnd, 0, sizeof(struct sess_rec_cmnd));

	cmnd->Cmnd = Cmnd;

	if (sess_grp->cmnd_list_head == NULL) {
		sess_grp->cmnd_list_head = cmnd;
	} else {
		sess_grp->cmnd_list_tail->next = cmnd;
	}
	sess_grp->cmnd_list_tail = cmnd;

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave set_rec_group sess_grp %p\n", sess_grp);
	return sess_grp;
}

void
clean_rec_group(struct sess_rec_group *sess)
{
	struct sess_rec_group *sess_grp = global_hostdata->sess_rec_head;
	struct sess_rec_group *prev_sess_grp = NULL;
	struct sess_rec_cmnd *cmnd = NULL, *tmp_cmnd = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter clean_rec_group \n");

	for (; sess_grp != NULL; sess_grp = sess_grp->next) {
		if (sess_grp == sess)
			break;
		prev_sess_grp = sess;
	}

	cmnd = sess_grp->cmnd_list_head;
	while (cmnd != NULL) {
		tmp_cmnd = cmnd;
		cmnd = cmnd->next;

		my_kfree((void **) &tmp_cmnd, "sess_rec_cmnd");
	}

	if (sess_grp == global_hostdata->sess_rec_head) {
		global_hostdata->sess_rec_head = sess_grp->next;
	} else
		prev_sess_grp->next = sess_grp->next;

	if (sess_grp)
		my_kfree((void **) &sess_grp, "sess_rec_group");

	TRACE(TRACE_ENTER_LEAVE, "Leave clean_rec_group \n");

}

/************************************************************************
* This thread handles the session recovery. Session Recovery will       *
* abort all pending tasks and close all transport connections           *
* Section (6.12.4)                                                      *
************************************************************************/
void
init_recovery_thread(void *arg)
{
	unsigned long flags;
	int target, lun, cid;
	int received = 1;
	__u32 size;
	struct response_pdu *header = NULL;
	struct init_error_rec *err_rec;
	struct connection *conn;
	struct session *session;
	struct sess_rec_cmnd *cmnd = NULL;
	struct sess_rec_group *sess_grp = NULL;
	struct sockaddr *old_ip_address, *new_ip_address;
	int new_ip_length;
#if defined(CONFIG_ISCSI_DEBUG)
	__u32 save_trace;
#endif

	/* Lock kernel to daemonize thread and update init_rec_thread pointer */
	lock_kernel();

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	daemonize(
#else
	daemonize();

	/* NOTE: in struct task_struct have declaration char comm[16] */
	strcpy(current->comm,
#endif
		"ini_recovery");

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 18)
	/*
	 * Arne Redlich, agr1@users.sourceforge.net:
	 * This prevents the recovery thread from becoming a zombie after exit.
	 */
	reparent_to_init();
#endif

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGBITS);

	/* mark that the recovery thread is alive */
	init_rec_thread = current;

	unlock_kernel();

	err_rec = (struct init_error_rec *)arg;
	conn = err_rec->curr_conn;
	session = conn->my_session;

	printk("%s Starting pid %d conn %p\n", current->comm, current->pid,
		conn);

	/* turn on some tracing */
	TRACE_GET(save_trace);
	TRACE_SET(save_trace |(TRACE_ISCSI_FULL|TRACE_ISCSI));

	/* Like the rx thread it is replacing,
	 * this thread normally holds exclusive access to the iscsi structures.
	 */

	UNH_LOCK(&session->sess_lock, conn->rx_lock_flags);

	target = session->scsi_target_id;
	old_ip_address = conn->ip_address;
	cid = conn->connection_id;

	/* Abort all queued scsi commands with an appropriate
	 * scsi service response Section - 6.12.4
	 */
	if (!(conn->connection_flags & CONN_HIT_EOF)) {
		/* the connection is still open, try for a graceful shutdown */
		drive_logout(session, conn, LOGOUT_CLOSE_SESSION);

		if (conn->rx_buf == NULL) {
			TRACE_ERROR("scsi Response Header is NULL\n");
			/* release our exclusive access to iscsi structures */
			UNH_UNLOCK(&session->sess_lock, conn->rx_lock_flags);
			goto rec_thread_out;
		}

		header = (struct response_pdu *)conn->rx_buf;
		size = ntohl(header->length);

		if (size > 0) {
			/* target response has data attached, toss it */
			received = recv_ignore_data(conn, size);
		}

		while (received > 0) {
			if ((received = recv_pdu_header(conn)) <= 0) {
				/* connection died or we got a kill signal,
				 * stop trying to read more input
				 */
				break;
			}

			/* ignore header digests while waiting for logout
			 * during recovery
			 */
			header = (struct response_pdu *)conn->rx_buf;

			size = ntohl(header->length);

			if (size > MASK_24_BITS) {
				/* bigger than 2^24 - 1 */
				TRACE_ERROR("TotalAHSLength 0x%02x, "
					    "expected 0x00\n", size >> 24);
				break;
			}

			/* mark that we got some activity from the target */
			conn->connection_flags |= GOT_ACTIVITY;

			/* Print the received iSCSI header */
			TRACE_BUFFER(TRACE_BUF, conn->rx_buf,
				     ISCSI_HDR_LEN, "%s got message\n",
				     current->comm);

			if (conn->connection_flags & NEED_TX_WAKEUP) {
				/* get the tx thread started again */
				TRACE(TRACE_ISCSI_FULL,
				      "%s wake up tx_thread\n", current->comm);
				up(&conn->tx_sem);
			}

			if (size > 0) {
				/* target response has data attached, toss it */
				received = recv_ignore_data(conn, size);
			}

			if (header->opcode == ISCSI_TARG_LOGOUT_RSP) {
				TRACE(TRACE_DEBUG, "Got LogOut Response\n");
				break;
			}
		}
	}

	/* release exclusive access to iscsi structures if we already have it */
	if (received >= 0) {
		UNH_UNLOCK(&session->sess_lock, conn->rx_lock_flags);
	}

	/* remove all the luns in this session, thereby removing the session */
	if ((lun = remove_luns(session, global_host, global_hostdata))
		< 0) {
		lun = 0;
	}

	/* at this point all structures and threads for original session
	 * and all its connections should be gone!
	 */

	dup_inet_struct(old_ip_address, &new_ip_address, &new_ip_length);
	if (create_session(target, new_ip_address, new_ip_length, lun, cid,
							global_host) != 1) {
		TRACE_ERROR("Unable to create recovery session for target %d "
			    "lun %d cid %d\n", target, lun, cid);
		goto rec_thread_out;
	}

	UNH_LOCK(&host_data_lock, flags);

	session = find_session_by_id(target, global_hostdata);

	if (session == NULL || (conn = session->connection_head) == NULL) {
		TRACE_ERROR("Unable to find recovery session for target %d "
			    "lun %d cid %d\n", target, lun, cid);
		goto rec_thread_out_locked;
	}

	sess_grp = global_hostdata->sess_rec_head;
	for (; sess_grp != NULL; sess_grp = sess_grp->next) {
		if (sess_grp->cmnd_list_head) {
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
			if (sess_grp->cmnd_list_head->Cmnd->device->id
#else
			if (sess_grp->cmnd_list_head->Cmnd->target
#endif
				== session->scsi_target_id)
				break;
		}
	}

	if (sess_grp) {
		cmnd = sess_grp->cmnd_list_head;
		for (; cmnd != NULL; cmnd = cmnd->next) {
			UNH_LOCK(&session->sess_lock, conn->rx_lock_flags);
			scsi_to_iscsi(cmnd->Cmnd, session);
			UNH_UNLOCK(&session->sess_lock, conn->rx_lock_flags);
		}

		clean_rec_group(sess_grp);
	}

rec_thread_out_locked:
	
	UNH_UNLOCK(&host_data_lock, flags);

rec_thread_out:

	global_hostdata->init_sessrec_flg = 0;

	init_rec_thread = NULL;

	up(&init_sess_recovery_sem);

	TRACE_SET(save_trace);

	printk("%s Exiting pid %d\n", current->comm, current->pid);

	return;
}

/************************************************************************
* This function handles the digest errors and sequence errors that may
* have caused because of previous digest errors.
* When called, session lock must be locked
* Returns >0 on success
*	  =0 on failure, session lock is held
*	  <0 on failure, session lock is NOT locked
************************************************************************/
int
init_digest_recovery(struct init_error_rec *err_rec)
{
	int retval = 1;		/* assume success */
	int size = 0, err_type = 0, delta;
	int serial_no = 0;
	__u32 opcode = 0;
	struct response_pdu *header;
	struct iscsi_targ_scsi_rsp *rsp_hdr;
	struct connection *curr_conn;
	struct session *current_session;
	struct command *related_cmd;
	struct iscsi_targ_scsi_data_in *data_in_hdr;

	curr_conn = err_rec->curr_conn;
	current_session = curr_conn->my_session;
	related_cmd = err_rec->related_cmd;
	err_type = err_rec->err_type;

	TRACE(TRACE_ENTER_LEAVE, "Enter init_digest_recovery \n");

	if (curr_conn->rx_buf == NULL) {
		TRACE_ERROR("scsi Response Header is NULL\n");
		retval = 0;
		goto out;
	} else {
		header = (struct response_pdu *)curr_conn->rx_buf;
		opcode = header->opcode;
		size = ntohl(header->length);
	}

	if (err_type == HEADER_DIGERR) {
		TRACE(TRACE_DEBUG, "Handling Header Digest Error \n");

		if (size > 0) {
			/* this size could be bad, but try using it anyway
			 * to read in and ignore this PDU's data payload
			 */
			retval = recv_ignore_data(curr_conn, size);
		}
		goto out;
	}

	if (opcode == ISCSI_TARG_SCSI_DATA_IN) {
		/* here on SEQUENCE_ERR or PAYLOAD_DIGERR only */
		if (err_rec->related_cmd == NULL) {
			TRACE_ERROR("Related SCSI Command is NULL - "
				    "Snack Request not sent\n");
			goto out;
		}

		data_in_hdr =
		    (struct iscsi_targ_scsi_data_in *)curr_conn->rx_buf;

		if (related_cmd->cmd_error_state == CMD_OK_SO_FAR) {
			/* this is the first error for this command */
			related_cmd->cmd_error_state = CMD_NEEDS_SNACK;
			related_cmd->cmd_error_size = 0;
			if (err_type == SEQUENCE_ERR) {
				/* keep this pdu, missing pdus are in gap
				 * before this one
				 */
				related_cmd->cmd_error_offset =
				    related_cmd->data_offset;
				related_cmd->cmd_error_begrun =
				    related_cmd->data_in_sn;
				related_cmd->data_in_sn = data_in_hdr->data_sn;
				related_cmd->cmd_error_runlen =
				    related_cmd->data_in_sn -
				    related_cmd->cmd_error_begrun;
			} else {
				/* this pdu itself was bad,
				 * need to ask for it again
				 */
				related_cmd->cmd_error_offset =
				    data_in_hdr->offset;
				related_cmd->cmd_error_begrun =
				    data_in_hdr->data_sn;
				related_cmd->cmd_error_runlen = 1;
			}
		} else if (related_cmd->cmd_error_state == CMD_NEEDS_SNACK) {
			/* already saw a previous error for this command */
			if (err_type == SEQUENCE_ERR) {
				/* keep DataSN of the out-of-order DataIn PDU */
				related_cmd->data_in_sn = data_in_hdr->data_sn;
			}
			/* effectively "throw away" any good pdus received since
			 * previous bad pdu, as they will all be resent later
			 */
			related_cmd->r2t_range_list.limit -=
				related_cmd->cmd_error_size;
			related_cmd->cmd_error_size = 0;
			related_cmd->cmd_error_runlen = related_cmd->data_in_sn
			    - related_cmd->cmd_error_begrun;
		} else {
			/* already sent a SNACK, now have new errors in reply */
			TRACE_ERROR("Error in replies to SNACK, "
				    "escalate to session recovery\n");
			retval = init_session_recovery(err_rec);
		}
	} else if (opcode == ISCSI_TARG_R2T) {
		TRACE(TRACE_DEBUG, "Handling R2T Sequence Error\n");
		serial_no = related_cmd->r2t_sn;
		retval = init_send_snack(err_rec, serial_no, 0, DATA_R2T_SNACK);
	} else if (opcode == ISCSI_TARG_SCSI_RSP) {

		TRACE(TRACE_DEBUG, "Handling Response Sequence Error\n");

		rsp_hdr = (struct iscsi_targ_scsi_rsp *)curr_conn->rx_buf;

		if (related_cmd == NULL) {
			TRACE_ERROR("Related SCSI Command is NULL - "
				    "Snack Request not sent\n");
			goto out;
		}

		if (related_cmd->data_in_sn != rsp_hdr->exp_data_sn) {
			serial_no = related_cmd->data_in_sn;
			retval = init_send_snack(err_rec, serial_no, 0,
						    DATA_R2T_SNACK);
		}

		delta = ntohl(rsp_hdr->stat_sn) - curr_conn->exp_stat_sn;
		if (delta < 0) {
			TRACE(TRACE_ISCSI,
			      "Discarding already received status PDU\n");
		} else {
			related_cmd->snack_flags |= STATUS_BIT;
			serial_no = curr_conn->exp_stat_sn;
			retval = init_send_snack(err_rec, serial_no, 1,
						    STATUS_SNACK);
		}
	}

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave init_digest_recovery %d\n", retval);

	return retval;
}

/*******************************************************************************
*       Called to send a SNACK Request to target for retransmission of
*	numbered-responses, data, or R2T PDU.
*	Returns	>0 on success
*		=0 on failure, session lock is locked
*		<0 on failure, session lock is NOT locked
*******************************************************************************/
int
init_send_snack(struct init_error_rec *err_rec, int begrun, int runlen,
		__u8 type)
{
	int retval = 0;		/* assume we will hit an error */
	struct connection *current_connection;
	struct session *current_session;
	struct command *new_command;
	struct iscsi_init_snack *snack_cmd;
	struct response_pdu *header;

	TRACE(TRACE_ENTER_LEAVE, "Enter init_send_snack\n");

	current_connection = err_rec->curr_conn;
	current_session = current_connection->my_session;
	header = (struct response_pdu *)current_connection->rx_buf;

	if (type == DATA_R2T_SNACK) {
		if (!(current_session->init_snack_flag & DATA_SNACK_ENABLE)) {
			TRACE_ERROR("Initiator Data Snack Is Not Enabled - "
				    "SNACK request Not Sent\n");
			goto out;
		}
	}

	if (type == STATUS_SNACK) {
		if (!(current_session->init_snack_flag & STATUS_SNACK_ENABLE)) {
			TRACE_ERROR("Initiator Status Snack Is Not Enabled - "
				    "SNACK request Not Sent\n");
			goto out;
		}
	}

	new_command = setup_command(NULL, current_connection);
	if (new_command == NULL) {
		/* will not be able to send the snack */
		goto out;
	}

	/* set up the SnackReq header (section 10.16) to send to the target */
	snack_cmd = (struct iscsi_init_snack *)&new_command->iscsi_cmd;
	snack_cmd->opcode = ISCSI_INIT_SNACK;
	snack_cmd->flags = F_BIT | type;     /* F_BIT always set on SnackReq */

	/* RFC 3720 Section 10.16.4 Initiator Task Tag (for SNACK Request)
	   "For Status SNACK and DataACK, the Initiator Task Tag MUST be set to
	   the reserved value 0xffffffff.  In all other cases, the Initiator
	   Task Tag field MUST be set to the Initiator Task Tag of the
	   referenced command."
	 */
	if (type != STATUS_SNACK && type != DATACK_SNACK) {
		new_command->init_task_tag =
		    err_rec->related_cmd->init_task_tag;
		snack_cmd->init_task_tag =
		    htonl(err_rec->related_cmd->init_task_tag);
	} else {
		new_command->init_task_tag = ALL_ONES;
		snack_cmd->init_task_tag = ALL_ONES;
		if (type == STATUS_SNACK)
			err_rec->related_cmd->snack_flags |= STATUS_BIT;
	}

	/* Draft 20, Section 10.16.5 Target Transfer Tag or SNACK Tag
	 * "For DataACK, the Target Transfer Tag MUST contain a copy of the
	 * Target Transfer Tag and LUN provided with the SCSI Data-In PDU
	 * with the A bit set to 1."
	 * In all other cases, the Target Transfer Tag field MUST be se to the
	 * reserved value of 0xffffffff.
	 */
	if (type == DATACK_SNACK) {
		snack_cmd->target_xfer_tag = header->target_xfer_tag;
		snack_cmd->lun = header->lun;
	} else {
		snack_cmd->target_xfer_tag = ALL_ONES;
	}

	snack_cmd->begrun = htonl(begrun);
	snack_cmd->runlen = htonl(runlen);

	TRACE(TRACE_ERROR_RECOVERY,
	      "Send SNACK, ITT %u, type %u, begrun %d, runlen %d\n",
	      ntohl(snack_cmd->init_task_tag), type, begrun, runlen);

	/* can now attach the new SNACK to end of pending commands list
	 * (ITT already set and SNACK has no CmdSN value)
	 */
	attach_pending_command(new_command, current_connection, -1);

	retval = 1;		/* pessimistic assumption was wrong */
out:
	TRACE(TRACE_ENTER_LEAVE, "Leave init_send_snack, retval %d\n", retval);

	return retval;
}

/******************************************************************
 * executed only by rx thread.
 * the session->sess_lock MUST NOT be held by the calling process/thread.
 * Deal with the Connection Recovery after Time2Wait expires.
 * Called with no locks held.
 * Returns >0 on success, session lock not held
 *	   =0 on failure, session lock is held
 *	   <0 on failure, session lock is NOT locked
 ******************************************************************/
int
deal_with_connrec(struct connection *fail_connection)
{
	struct session *current_session;
	struct connection *new_connection = NULL;
	struct command *task_mgt_command = NULL, *cmd;
	struct iscsi_init_task_mgt_command *task_mgt_command_hdr = NULL;
	struct init_error_rec err_rec;
	int lun, retval = 1;
	struct list_head *lptr, *next;

	TRACE(TRACE_ENTER_LEAVE, "Enter deal_with_connrec\n");

	TRACE(TRACE_ERROR_RECOVERY, "Conn Recovery Timer expired\n");

	current_session = fail_connection->my_session;

	fail_connection->connection_state = CONNECTION_RECOVERING;

	if (current_session->nconnections <= 1) {
		/* This is last connection available, create new connection */
		if (!create_rec_conn(fail_connection)) {
			retval = -1;
			goto out;
		}
	}

	UNH_LOCK(&current_session->sess_lock, fail_connection->rx_lock_flags);

	current_session->n_rec_tasks = 0;

	list_for_each_safe(lptr, next, &fail_connection->pending_commands) {
		cmd = list_entry(lptr, struct command, link);
		cmd->activity_flg = 0;
		if (cmd->iscsi_cmd.opcode != ISCSI_INIT_SCSI_CMND)
			continue;

		new_connection = get_connection(cmd->SCpnt, current_session);

		if (new_connection == fail_connection) {
			TRACE_ERROR("get_connection returned the failed "
				    "connection - can't re-assign commands\n");
			goto out_escalate;
		}

		if (!new_connection) {
			TRACE_ERROR("get_connection returned NULL\n");
			goto out_escalate;
		}

		if ((task_mgt_command =
		     setup_command(NULL, new_connection)) == NULL) {
			goto out_escalate;
		}

		TRACE(TRACE_DEBUG, "Populating task mgt command for connection "
		      "recovery\n");
		task_mgt_command->task_mgt_function = TMF_TASK_REASSIGN;

		task_mgt_command->task_mgt_response = FUNCTION_REJECTED;
		task_mgt_command->ref_task_tag = cmd->init_task_tag;

		/* fill up the task_mgt_header info. */
		task_mgt_command_hdr = (struct iscsi_init_task_mgt_command *)
		    &task_mgt_command->iscsi_cmd;
		task_mgt_command_hdr->opcode = ISCSI_INIT_TASK_MGMT_CMND;
		task_mgt_command_hdr->opcode |= I_BIT;
		task_mgt_command_hdr->function = (TMF_TASK_REASSIGN | F_BIT);
		task_mgt_command_hdr->lun = cmd->iscsi_cmd.lun;
		task_mgt_command_hdr->ref_task_tag = htonl(cmd->init_task_tag);

		task_mgt_command_hdr->ref_cmd_sn = cmd->iscsi_cmd.cmd_sn;

		TRACE(TRACE_ERROR_RECOVERY,
		      "Send TASK MGT Cmnd, ITT %u, %u bytes\n",
		      task_mgt_command->init_task_tag,
		      task_mgt_command->tx_size);

		/* attach the new task mgt command to end of pending commands
		 * this also assigns a new ITT and sets the CmdSN.
		 */
		attach_pending_command(task_mgt_command, new_connection, 1);

		current_session->n_rec_tasks++;
		TRACE(TRACE_DEBUG,
		      "Waiting to get task mgt. rsp from the target\n");

		/* move command from old connection to new connection */
		list_del(&cmd->link);
		list_add_tail(&cmd->link, &new_connection->pending_commands);
	}

	if (current_session->n_rec_tasks > 0) {
		TRACE(TRACE_DEBUG, "%d commands recovered\n",
		      current_session->n_rec_tasks);

	} else {
		/* no tasks transferred */
	}

	UNH_UNLOCK(&current_session->sess_lock, fail_connection->rx_lock_flags);

	/* find a lun that is already in use in this session */
	if ((lun = find_used_lun(current_session)) >= 0) {
		/* got the lun, remove failed connection from this session */
		remove_session_from_hostdata(current_session->scsi_target_id,
					     lun,
					     fail_connection->connection_id,
					     global_hostdata);
	}

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave deal_with_connrec, retval %d\n",
		retval);
	return retval;

out_escalate:
	UNH_UNLOCK(&current_session->sess_lock, fail_connection->rx_lock_flags);
	err_rec.curr_conn = fail_connection;
	retval = init_session_recovery(&err_rec);
	goto out;
}

static void
time2wait_function(unsigned long wait_sem)
{
	struct semaphore *rec_wait_sem = (struct semaphore *)wait_sem;
	up(rec_wait_sem);
}

/***************************************************************
	Start Time2Wait Timer for Connection Recovery
***************************************************************/
void
start_connrec_timer(struct session *current_session,
		    struct semaphore *rec_wait_sem, __u32 period)
{

	TRACE(TRACE_ENTER_LEAVE, "Enter start_connrec_timer, period %u\n",
	      period);

	if (period > 0 && !timer_pending(&current_session->connrec_timer)) {
			/* we can start the timer for a non-zero
			 * period of time (in ticks)
			 */
		TRACE(TRACE_ERROR_RECOVERY, "Start timer for %u ticks\n",
		      period);
		current_session->connrec_timer.expires = jiffies + period;
		current_session->connrec_timer.data =
		    (unsigned long) rec_wait_sem;
		current_session->connrec_timer.function = time2wait_function;
		add_timer(&current_session->connrec_timer);
	} else {
		/* else period is 0 or no timer or timer already going,
		 * don't start it now
		 */
		TRACE_ERROR("Conn Recovery Timer not started for %u ticks\n",
			    period);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave start_connrec_timer\n");
}

void
restore_global_data(struct iscsi_targetdata *targ_data)
{
	char *targ_name, *init_name;
	struct parameter_type *dst, *param;
	/* Uncopy the old values from the parameter table */

	TRACE(TRACE_ENTER_LEAVE, "Entering restore_global_data\n");

	dst = targ_data->param_tbl;

	param = find_parameter("TargetName", dst);
	targ_name = strdup(param->str_value);

	param = find_parameter("InitiatorName", dst);
	init_name = strdup(param->str_value);

	param_tbl_uncpy(dst);
	param_tbl_init(dst);

	param = find_parameter("TargetName", dst);
	strreplace(&param->str_value, targ_name);
	param->neg_info = KEY_TO_BE_NEGOTIATED;
	param = find_parameter("InitiatorName", dst);
	strreplace(&param->str_value, init_name);
	param->neg_info = KEY_TO_BE_NEGOTIATED;

	TRACE(TRACE_ENTER_LEAVE, "Leaving restore_global_data\n");
}
