/*	initiator/initiator_proc_iface.c
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This file contains the functions for iscsi initiator code that are
 *	responsible for interfacing with the /proc filesystem.
 *
 *	Copyright (C) 2001-2004 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 * 	written permission.
*/


#include "iscsi_initiator.h"
#include "initiator_utilities.h"
#include "initiator_error_rec.h"
#include "initiator_proc_iface.h"
#include <linux/proc_fs.h>

#ifdef ISCSI_STATS
static int __attribute__ ((no_instrument_function))
dump_conn_info(struct connection *, char *);
static int __attribute__ ((no_instrument_function))
dump_sess_info(struct session *, char *);
static int __attribute__ ((no_instrument_function))
dump_instance_stats(char *);
#endif

/*
 * executed only by command process.
 * Called only by scan_test_stuff()
 * returns	>= 0 on success,
 * 		 < 0 on error
 */
static int __attribute__ ((no_instrument_function))
do_test(__u32 target, __u32 cid, __u32 test_no)
{
	unsigned long flags, lock_flags;
	int result;
	struct connection *conn;
	struct session *sess;

	/*  When using session list, be sure nobody else accesses it */
	UNH_LOCK(&host_data_lock, flags);

	/* use target and cid to find session and connection */
	result = find_connection(target, cid, global_hostdata, &conn, &sess);

	if (result >= 0) {
		/* perform the requested test on this connection */

		UNH_LOCK(&sess->sess_lock, lock_flags);
		UNH_UNLOCK(&host_data_lock, flags);

		TRACE(TRACE_ISCSI, "Do test number %u\n", test_no);

		switch (test_no) {
		case 0:
			/* send text request */
			result = drive_text_negotiate(conn, sess, 0, 1);
			break;
		case 1:
			/* send text request as immediate cmd */
			result = drive_text_negotiate(conn, sess, I_BIT, 1);
			break;
		case 2:
			/* send nop (not immediate, no data) */
			result = drive_nopout(conn, sess, 0, 0);
			break;
		case 3:
			/* send nop (immediate, no data) */
			result = drive_nopout(conn, sess, I_BIT, 0);
			break;
		case 4:
			/* send nop ping (not immediate, data) */
			result = drive_nopout(conn, sess, 0, 1);
			break;
		case 5:
			/* send nop ping (immediate, data) */
			result = drive_nopout(conn, sess, I_BIT, 1);
			break;
		case 6:
			/* stop/start responding to nop ping from target */
			conn->connection_flags ^= SEND_NO_REPLY_TO_NOP_PING;
			break;
		case 7:
			/* stop/start responding to async logout
			 * request from target
			 */
			conn->connection_flags ^= SEND_NO_REPLY_TO_ASYNC_LOGOUT;
			break;
		case 8:
			/* perform connection re-assignment - SAI */
			conn->rec_tests = test_no;
			break;
		case 9:
			/* send session logout request */
			result = !drive_logout(sess, conn,
					       LOGOUT_CLOSE_SESSION);
			break;
		case 10:
			/* send connection logout request */
			result = !drive_logout(sess, conn,
					       LOGOUT_CLOSE_CONNECTION);
			break;
		default:
			TRACE_ERROR("Test number %u not implemented yet\n",
				    test_no);
			result = -EINVAL;
			break;
		} /* switch */
		UNH_UNLOCK(&sess->sess_lock, lock_flags);
	} else {
		/* release our exclusive access to the iscsi structures */
		UNH_UNLOCK(&host_data_lock, flags);
	}

	return result;
}

/*
 * executed only by command process.
 * host_data_lock MUST be held by the calling process/thread
 * Called by dump_conn_info() and print_session_info()
 */
static int __attribute__ ((no_instrument_function))
print_connection_info( struct connection *conn, char *buffer, int pos )
{
	const char *ptr;
	int lun;
	struct session *sess;
	char extra[20];

	pos += sprintf(buffer + pos, "%16s: %u\n", "Connection CID",
		       conn->connection_id);

	pos += sprintf(buffer + pos, "%16s: ", "Connection State");
	switch (conn->connection_state) {
	case CONNECTION_FULL_FEATURE_PHASE:
		ptr = "FULL FEATURE PHASE";
		break;
	case CONNECTION_LOGGED_IN:
	case CONNECTION_CONNECTED:
		ptr = "LOGIN PHASE";
		break;
	case CONNECTION_RECOVERING:
		ptr = "ERROR RECOVERY";
		break;
	case CONNECTION_LOGGED_OUT:
	case CONNECTION_DISCONNECTED:
		ptr = "LOGOUT PHASE";
		break;
	case CONNECTION_NOT_PRESENT:
		ptr = "NOT CONNECTED";
		break;
	default:
		pos += sprintf(buffer + pos, "0x%x", conn->connection_state);
		ptr = "";
		break;
	}
	pos += sprintf(buffer + pos, "%s\n", ptr);

	if (conn->connection_state == CONNECTION_FULL_FEATURE_PHASE
		|| conn->connection_state == CONNECTION_CONNECTED) {
		char ip_string[INET6_ADDRSTRLEN+2], port_string[8];

		if (cnv_inet_to_string(conn->ip_address, ip_string, port_string)
									> 0) {
			pos += sprintf(buffer + pos, "%16s: %s:%s\n",
				       "Target Address",
				       ip_string, port_string);
		}
		if (cnv_inet_to_string(conn->local_ip_address, ip_string,
				       port_string) > 0) {
			pos += sprintf(buffer + pos, "%16s: %s:%s\n",
				       "Local Address",
				       ip_string, port_string);
		}

		sess = conn->my_session;
		if (sess->sched_scheme == CONN_SCHED_LUN ) {
			/* LUN assigned to connection scheduling is in effect,
			 * print numbers of all luns assigned to this connection
			 * (if any)
			 */
			lun = conn->assigned_lun_count;
			sprintf(extra, "%d assigned LUN%s", lun,
				lun == 1 ? "" : "s");
			pos += sprintf(buffer + pos, "%16s:", extra);
			for (lun = 0; lun < MAX_LUNS; lun++) {
				if (conn == sess->lun_assignments[lun]) {
					pos += sprintf(buffer + pos," %d",lun);
				}
			}
			pos += sprintf(buffer + pos, "\n");
		}
	}
	return pos;
}

static int __attribute__ ((no_instrument_function))
print_ascii(char *buffer, __u8 *ptr, int len, char *mess)
{
	int pos;
	pos = sprintf(buffer, "%24s: ", mess);
	memcpy(buffer + pos, ptr, len);
	pos += len;
	pos += sprintf(buffer + pos, "\n");
	return pos;
}

static int __attribute__ ((no_instrument_function))
print_ver_descr(char *buffer, __u8 *ptr, __u8 *base)
{
	int pos = 0, i;

	for (i = 0; i < 8; i++) {
		if (ptr > base)
			break;
		if (*ptr || *(ptr+1)) {
			pos += sprintf(buffer + pos, "%24s: 0x%02x%02x\n",
				       "version descriptor", *ptr, *(ptr+1));
		}
		ptr += 2;
	}
	return pos;
}

static int __attribute__ ((no_instrument_function))
print_capacity_stuff(char *buffer, struct session *sess, __u32 lun)
{
	int pos = 0;
	__u32 lba = sess->capacity_lba[lun];
	__u32 len = sess->capacity_len[lun];

	if (lba && len) {
		if (lba != ALL_ONES)
			lba++;
		pos = sprintf(buffer + pos,
				"%16s: %u %u %u-byte hardware sectors\n",
				"capacity for lun", lun, lba, len);
	}
	return pos;
}


static int __attribute__ ((no_instrument_function))
print_limits_stuff(char *buffer, struct session *sess, __u32 lun)
{
	int pos = 0;
	__u32 max = sess->limit_max[lun];
	__u32 min = sess->limit_min[lun];

	if (max || min) {
		pos = sprintf(buffer + pos,
				"%16s: %u [%u, %u] bytes per block\n",
				"limits for lun", lun, min, max);
	}
	return pos;
}

static const char * __attribute__ ((no_instrument_function))
printable_dev_type( __u32 dev_type )
{
	const char *mess;

	if (dev_type < MAX_SCSI_DEVICE_CODE) {
		mess = scsi_device_types[dev_type];
	} else {
		mess = "reserved";
		switch (dev_type) {
		case 0x0e:
			mess = "simple-direct-access";
			break;
		case 0x0f:
			mess = "optical card reader/writer";
			break;
		case 0x11:
			mess = "object-based-storage";
			break;
		case 0x1f:
			mess = "unknown";
			break;
		}	/* switch */
	}
	return mess;
}

static int __attribute__ ((no_instrument_function))
print_inquiry_stuff(char *buffer, struct session *sess, __u32 lun)
{
	int pos = 0, size = sess->inquiry_size[lun], len;
	__u8 *base = sess->inquiry_buf[lun], *ptr;

	if ((ptr = base) && size) {
		pos += sprintf(buffer + pos, "%16s: %u (bytes 0..%u)\n",
				"inquiry for lun", lun, *(ptr + 4) + 4);
		if ((*ptr >> 5) == 0)
			pos += sprintf(buffer + pos, "%24s: %s\n", "device",
				"connected");
		else if ((*ptr >> 5) == 1)
			pos += sprintf(buffer + pos, "%24s: %s\n", "device",
				"not connected");
		else
			pos += sprintf(buffer + pos, "%24s: %u\n", "qualifier",
					*ptr >> 5);
		pos += sprintf(buffer + pos, "%24s: %s\n", "device type",
				printable_dev_type(*ptr & 0x1f));
		ptr += 1;
		pos += sprintf(buffer + pos, "%24s: %sremovable\n", "medium",
				*ptr & 0x80 ? "" : "not ");
		ptr += 1;
		pos += sprintf(buffer + pos, "%24s: 0x%02x\n", "version", *ptr);
		ptr += 1;
		if (*ptr != 2)
			pos += sprintf(buffer + pos, "%24s: 0x%02x\n",
					"byte 3", *ptr);
		ptr += 1;
		base += 4 + *ptr++;
		if (ptr > base)
			goto done;
		if (*ptr & 0x80)
			pos =+ sprintf(buffer + pos, "%24s: %s\n",
					"device contains",
					"embedded storage array controller");
		ptr += 1;		/* skip byte 5 for now */
		if (ptr > base)
			goto done;
		pos += sprintf(buffer + pos, "%24s: 0x%02x\n", "byte 6", *ptr);
		ptr += 1;
		if (ptr > base)
			goto done;
		pos += sprintf(buffer + pos, "%24s: 0x%02x\n", "byte 7", *ptr);
		ptr += 1;
		if (ptr > base)
			goto done;
		pos += print_ascii(buffer + pos, ptr, 8, "vendor id");
		ptr += 8;
		if (ptr > base)
			goto done;
		pos += print_ascii(buffer + pos, ptr, 16, "product id");
		ptr += 16;
		if (ptr > base)
			goto done;
		pos += print_ascii(buffer + pos, ptr, 4, "product revision");
		ptr += 4;
		if (ptr > base)
			goto done;
		len = strlen(ptr);
		if (len > 0  && len < 20)
			pos += sprintf(buffer + pos, "%24s: %s\n",
					"vendor specific", ptr);
		ptr += 20;
		if (ptr > base)
			goto done;
		pos += sprintf(buffer + pos, "%24s: 0x%02x\n", "byte 56", *ptr);
		ptr += 1 + 1;
		pos += print_ver_descr(buffer + pos, ptr, base);
	}
done:
	return pos;
}

/*
 * executed only by command process.
 * host_data_lock MUST be held by the calling process/thread
 * Called by dump_sess_info() and iscsi_initiator_proc_info()
 * this is for configuring the login parameters
 */
static int __attribute__ ((no_instrument_function))
print_session_info( struct session *sess, char *buffer, int pos )
{
	int i;
	struct connection *conn;
	const char *ptr;
	char extra[12];

	pos += sprintf(buffer + pos, "%16s: 0x%p\n", "pointer", sess);
	pos += sprintf(buffer + pos, "%16s: %u\n", "target id",
		       sess->scsi_target_id);
	sprintf(extra, "%u LUN%s", sess->n_luns_in_session,
		sess->n_luns_in_session == 1 ? "" : "s");
	pos += sprintf(buffer + pos, "%16s:", extra);
	for (i = 0; i < global_host->max_lun; i++) {
		if (sess->lun_bits & (1 << i)) {
			pos += sprintf(buffer + pos, " %d", i);
		}
	}
	pos += sprintf(buffer + pos, "\n");
	for (i = 0; i < global_host->max_lun; i++) {
		if (sess->lun_bits & (1 << i)) {
			pos += print_capacity_stuff(buffer + pos, sess, i);
			pos += print_limits_stuff(buffer + pos, sess, i);
			pos += print_inquiry_stuff(buffer + pos, sess, i);
		}
	}

	switch (sess->sched_scheme) {
	case CONN_SCHED_RR:	ptr = "round-robin"; break;
	case CONN_SCHED_LUN:	ptr = "LUN to connection"; break;
	default:		ptr = "first connection only"; break;
	}
	pos += sprintf(buffer + pos, "%16s: %s\n", "conn. scheduling",
		       ptr);
	pos += sprintf(buffer + pos, "%16s: %u\n", "nop interval",
		       sess->nop_period);

	pos += sprintf(buffer + pos, "%16s: %u\n", "connections",
		       sess->nconnections);
	for (conn = sess->connection_head; conn != NULL; conn = conn->next) {
		pos = print_connection_info(conn, buffer, pos);
	}
	return pos;
}

/*
 * executed only by command process.
 * host_data_lock MUST be held by the calling process/thread
 * Called by iscsi_initiator_proc_info()
 */
static int __attribute__ ((no_instrument_function))
print_host_info( struct Scsi_Host *host, char *buffer, int pos )
{
	pos += sprintf(buffer + pos, "%24s  %u\n",
			"host number", host->host_no);
	pos += sprintf(buffer + pos, "%24s  %u\n",
			"max targets", host->max_id);
	pos += sprintf(buffer + pos, "%24s  %u\n",
			"max luns per target", host->max_lun);
	pos += sprintf(buffer + pos, "%24s  %u\n",
			"max bytes per cdb", host->max_cmd_len);
	pos += sprintf(buffer + pos, "%24s  %d\n",
			"max outstanding commands", host->can_queue);
	pos += sprintf(buffer + pos, "%24s  %d\n",
			"max commands per lun", host->cmd_per_lun);
	pos += sprintf(buffer + pos, "%24s  %u\n",
			"max sg size per command", host->sg_tablesize);
	pos += sprintf(buffer + pos, "%24s  %u\n",
			"max sectors per command", host->max_sectors);
	pos += sprintf(buffer + pos, "\n");
	return pos;
}

/* scans the unsigned integer starting at ptr.
 * if result is <= max and only white space follows, return the result
 * otherwise set err
 */
static void __attribute__ ((no_instrument_function))
scan_numeric(char *ptr, __u32 max, __u32 *result, int *err)
{
	__u32 temp;

	temp = simple_strtoul(ptr, &ptr, 0);
	if (temp > max || strspn(ptr, WHITE_SPACE) != strlen(ptr))
		*err = -EINVAL;
	else
		*result = temp;
}

/*
 * executed only by command process.
 * Called only by scan_manage_stuff()
 */
static int __attribute__ ((no_instrument_function))
scan_force_stuff(char *ptr, struct iscsi_targetdata *targ_data, int result)
{
	__u32 c_len, trace_info;

	if (strcmp(ptr, "s") == 0)
		targ_data->init_conn_flags |= USE_SECURITY_PHASE;
	else if (strcmp(ptr, "o") == 0)
		targ_data->init_conn_flags |= USE_OPERATIONAL_PHASE;
	else if (strcmp(ptr, "f") == 0)
		targ_data->init_conn_flags |= USE_FLAT_SPACE_LUN;
	else if (strcmp(ptr, "r") == 0)
		targ_data->init_conn_flags |= USE_FULL_REPLIES;
	else if (strcmp(ptr, "xok") == 0)
		targ_data->init_conn_flags |= USE_REFLECT_XKEYS;
	else if (strncmp(ptr, "n=", 2) == 0) {
		/* set periodic timer for sending nop-out to target */
		scan_numeric(ptr+2, UINT_MAX, &targ_data->nop_period, &result);
	} else if (strncmp(ptr, "v=", 2) == 0) {
		/* ignore any draft number */
		TRACE(TRACE_ISCSI, "Draft number \"%s\" ignored\n", ptr);
	} else if (strncmp(ptr, "p=", 2) == 0) {
		/* Store the retran_interval - SAI */
		scan_numeric(ptr+2, UINT_MAX, &targ_data->retran_period, &result);
		TRACE(TRACE_DEBUG, "Set Initiator to enforce retransmit"
			  " wait period as %u\n", targ_data->retran_period);
	} else if (strncmp(ptr, "sch=", 4) == 0) {
		/* Store the connection scheduling scheme - SAI */
		scan_numeric(ptr+4, UINT_MAX, &targ_data->sched_scheme, &result);
		TRACE(TRACE_DEBUG, "Set Initiator to enforce Connection"
			  " Scheduling Scheme as %u\n",
			  targ_data->sched_scheme);
	} else if (strcmp(ptr, "t") == 0) {
		/* chap and srp support - CHONG */
		TRACE(TRACE_ISCSI, "target authentication enabled\n");
		targ_data->auth_parameter.auth_flags
			|= USE_TARGET_CONFIRMATION;
	} else if (strcmp(ptr, "b") == 0) {
		TRACE(TRACE_ISCSI, "base64 format enabled\n");
		targ_data->auth_parameter.auth_flags |= USE_BASE64;
		targ_data->auth_parameter.chap_local_ctx->number_format
			= BASE64_FORMAT;
		targ_data->auth_parameter.chap_peer_ctx->number_format
			= BASE64_FORMAT;
		targ_data->auth_parameter.srp_ctx->format
			= BASE64_FORMAT;
	} else if (strncmp(ptr, "px=", 3) == 0) {
		TRACE(TRACE_ISCSI, "CHAP peer secret set to \"%s\"\n",
			  ptr + 3);
		CHAP_SetSecret(ptr + 3,
			       targ_data->auth_parameter.chap_peer_ctx);
	} else if (strncmp(ptr, "pn=", 3) == 0) {
		TRACE(TRACE_ISCSI, "CHAP peer name set to \"%s\"\n",
			  ptr + 3);
		CHAP_SetName(ptr + 3,
			     targ_data->auth_parameter.chap_peer_ctx);
	} else if (strncmp(ptr, "lx=", 3) == 0) {
		TRACE(TRACE_ISCSI, "CHAP local secret set to \"%s\"\n",
			  ptr + 3);
		CHAP_SetSecret(ptr + 3,
			      targ_data->auth_parameter.chap_local_ctx);
	} else if (strncmp(ptr, "ln=", 3) == 0) {
		TRACE(TRACE_ISCSI, "CHAP local name set to \"%s\"\n",
			  ptr + 3);
		CHAP_SetName(ptr + 3,
			     targ_data->auth_parameter.chap_local_ctx);
	} else if (strncmp(ptr, "cl=", 3) == 0) {
		scan_numeric(ptr+3, MAX_CHAP_BINARY_LENGTH, &c_len, &result);
		if (c_len == 0 || result < 0) {
			TRACE_ERROR("invalid CHAP challenge length: \"%s\"\n",
				    ptr + 3);
		} else {
			TRACE(TRACE_ISCSI,
				  "CHAP challenge length set to %u\n", c_len);
			CHAP_SetChallengeLength(c_len,
						targ_data->auth_parameter.
							chap_local_ctx);
		}
	} else if (strncmp(ptr, "sx=", 3) == 0) {
		TRACE(TRACE_ISCSI, "SRP secret set to \"%s\"\n",
			  ptr + 3);
		SRP_SetSecret(ptr + 3,
				  targ_data->auth_parameter.srp_ctx);
	} else if (strncmp(ptr, "sn=", 3) == 0) {
		TRACE(TRACE_ISCSI, "SRP name set to \"%s\"\n",
			  ptr + 3);
		SRP_SetName(ptr + 3,
					targ_data->auth_parameter.srp_ctx);
	} else if (strncmp(ptr, "sg=", 3) == 0) {
		if (SRP_SetSRPGroup(ptr + 3,
				targ_data->auth_parameter.srp_ctx) == 0) {
			TRACE_ERROR("Unable to set SRP group %s\n",
						ptr + 3);
		} else {
			TRACE(TRACE_ISCSI, "SRP group number set to %s\n",
				  ptr + 3);
		}
	} else if (strncmp(ptr, "V=", 2) == 0) {
		scan_numeric(ptr+2, UINT_MAX, &trace_info, &result);
		printk("trace_info = 0x%04x\n", trace_info);
		TRACE_SET(trace_info);
	} else if (strncmp(ptr, "isidt=", 6) == 0) {
		scan_numeric(ptr+6, 3, &targ_data->isid_type, &result);
	} else if (strncmp(ptr, "isidn=", 6) == 0) {
		scan_numeric(ptr+6, MASK_24_BITS, &targ_data->isid_number,
			     &result);
	} else if (strncmp(ptr, "isidq=", 6) == 0) {
		scan_numeric(ptr+6, MASK_24_BITS, &targ_data->isid_qualifier,
			     &result);
	}else {
		TRACE_ERROR("Unknown force \"%s\"\n", ptr);
		result = -EINVAL;
	}
	return result;
}

/*
 * executed only by command process.
 * Called only by scan_manage_stuff()
 * Scan Error Recovery Options to enable or disable snack - SAI
 */
static void __attribute__ ((no_instrument_function))
scan_snack_stuff(char *ptr, struct iscsi_targetdata *targ_data)
{
	if (strncmp(ptr, "d=", 2) == 0) {
		/* enable or disable data snack */
		if (strcmp(ptr + 2, "y") == 0)
			targ_data->init_snack_flag |= DATA_SNACK_ENABLE;
		else if (strcmp(ptr + 2, "Y") == 0)
			targ_data->init_snack_flag |= DATA_SNACK_ENABLE;
	} else if (strncmp(ptr, "s=", 2) == 0) {
		/* enable or disable status snack */
		if (strcmp(ptr + 2, "y") == 0)
			targ_data->init_snack_flag |= STATUS_SNACK_ENABLE;
		else if (strcmp(ptr + 2, "Y") == 0)
			targ_data->init_snack_flag |= STATUS_SNACK_ENABLE;
	} else {
		TRACE_ERROR("unknown snack \"%s\"\n", ptr);
	}
	TRACE(TRACE_DEBUG,
		  "SNACK Flags: datasnack %d statsnack %d\n",
		  (targ_data->init_snack_flag & DATA_SNACK_ENABLE),
		  (targ_data-> init_snack_flag & STATUS_SNACK_ENABLE) >> 1);
}

/*
 * executed only by command process.
 * Called only by scan_manage_stuff()
 * Restore all settings to default values
 */
static void __attribute__ ((no_instrument_function))
scan_restore_stuff(struct iscsi_targetdata *targ_data)
{
	uninit_targetdata(targ_data);
	init_targetdata(targ_data);

	global_hostdata->init_sessrec_flg = 0;
}

/*
 * executed only by command process.
 * Called only by iscsi_initiator_proc_info()
 * this is for configuring the login parameters
 */
static int __attribute__ ((no_instrument_function))
scan_manage_stuff(char *ptr, int result)
{
	__u32 target_no, param_neg_info;
	struct iscsi_targetdata *targ_data;

	/* Set the target number */
	target_no = simple_strtoul(ptr, &ptr, 0);

	if (target_no >= global_host->max_id) {
		result = -EINVAL;
		goto out;
	}
	targ_data = &global_hostdata->target_data[target_no];

	/* scan number for "set/setr/unset/setp/restore/force" */
	param_neg_info = simple_strtoul(ptr+1, &ptr, 0);

	/* Move a position to ignore space */
	ptr++;

	/* Now configure the parameters */
	if (param_neg_info == FORCE) {
		result = scan_force_stuff(ptr, targ_data, result);
	} else if (param_neg_info == SNACK) {
		scan_snack_stuff(ptr, targ_data);
	} else if (param_neg_info == RESTORE) {
		/* restore all settings to default values */
		scan_restore_stuff(targ_data);
	} else {
		/* a "normal" key=value type of thing */
		configure_parameter(param_neg_info, ptr, targ_data->param_tbl);
	}
out:
	return result;
}

/*
 * executed only by command process.
 * Called only by iscsi_initiator_proc_info()
 * This is for triggering test actions in the initiator
 */
static int __attribute__ ((no_instrument_function))
scan_test_stuff(char * ptr, int result)
{
	int i;
	__u32 target_no, cid, test_no;

	TRACE(TRACE_DEBUG, "trigger test action\n");

	if (strncmp(ptr, " target ", 8) != 0) {
		result = -EINVAL;
		goto out;
	}

	/* this buffer has expected preface, scan target number */
	target_no = simple_strtoul(ptr + 8, &ptr, 0);

	TRACE(TRACE_DEBUG, "got target number %u\n", target_no);

	if (target_no >= global_host->max_id || strncmp(ptr, " cid ", 5) != 0) {
		result = -EINVAL;
		goto out;
	}

	/* this buffer has expected preface, scan cid number */
	cid = simple_strtoul(ptr + 5, &ptr, 0);

	TRACE(TRACE_DEBUG, "got cid %u\n", cid);

	if (cid > 0xff || strncmp(ptr, " test ", 6) != 0) {
		result = -EINVAL;
		goto out;
	}

	/* this buffer has expected preface, scan test number */
	test_no = simple_strtoul(ptr + 6, &ptr, 0);

	TRACE(TRACE_DEBUG, "got test_no %u\n", test_no);

	if (strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
		result = -EINVAL;
		goto out;
	}

	if ((i = do_test(target_no, cid, test_no)) < 0)
		result = i;
out:
	return result;
}

/*
 * executed only by command process.
 * Called only by iscsi_initiator_proc_info()
 */
static int __attribute__ ((no_instrument_function))
scan_ip_stuff(char * ip_string, int result)
{
	char *ptr, *port_string;
	__u32 port_no, target, lun, cid;
	int i;
	struct sockaddr *ip_address;
	int ip_length;

	ptr = strchr(ip_string, ' ');

	if (ptr == NULL || strncmp(ptr, " port ", 6) != 0) {
		result = -EINVAL;
		goto out;
	}

	/* this buffer has expected preface, scan port number */
	port_string = ptr + 6;
	port_no = simple_strtoul(port_string, &ptr, 0);

	if (strncmp(ptr, " target ", 8) != 0) {
		result = -EINVAL;
		goto out;
	}

	/* this buffer has expected preface, scan target number */
	target = simple_strtoul(ptr + 8, &ptr, 0);

	if (target >= global_host->max_id || strncmp(ptr, " lun ", 5) != 0) {
		result = -EINVAL;
		goto out;
	}

	/* this buffer has expected preface, scan lun number */
	lun = simple_strtoul(ptr + 5, &ptr, 0);

	if (lun >= global_host->max_lun || strncmp(ptr, " cid ", 5) != 0) {
		result = -EINVAL;
		goto out;
	}

	/* this buffer has expected preface, scan cid number */
	cid = simple_strtoul(ptr + 5, &ptr, 0);

	if (cid > 0xff || strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
		result = -EINVAL;
		goto out;
	}

	if (port_no == 0) {
		/* "down target lun cid" on this host */
		if ((i = remove_session_from_hostdata(target, lun, cid,
						      global_hostdata)) <= 0) {
			/* error (<0) or only old connection removed (==0)
			 * pass this back as the result
			 */
			result = i;
		}
		/*  else >0, removed old session or lun, result==length */
	} else {
		if ((i = cnv_string_to_inet(ip_string, port_string,
						&ip_address, &ip_length)) < 0) {
			result = i;
		} else {
#if defined(CONFIG_ISCSI_DEBUG)
			__u32 save_trace;
#endif
			/* "up target lun cid" to ip:port on this host.
			 * Returns:
			 * < 0: error
			 * = 0: added new connection (only) to existing session
			 * = 1: added complete new session, lun and connection
			 * = 2: added a new lun (only) to existing session
			 */
			TRACE_GET(save_trace);
			if ((i = create_session(target, ip_address, ip_length,
						lun, cid, global_host)) <= 0) {
				/* error (<0) or new connection (only) (==0)
				 * pass this back as the result
				 */
				result = i;
			}
			/*  else >0, added new session or lun, result==length */
			TRACE_SET(save_trace);
		}
	}
out:
	return result;
}

/******************************************************************************
*	Called when the host file "0" or "1" or ...
*	in the /proc/scsi/iscsi_initiator/ directory is read or written.
*	The parameter "inout" is 0 for a read, 1 for a write operation.
*	A write expects the buffer to contain something in the format:
*		"iscsi_initiator ip 1234abcd port 4000 target 6 lun 0"
*	or
*		"iscsi_initiator manage 0 5 key=value"
*	or
		"iscsi_iniitator manage 1 32 o"
*	or
*		"iscsi_initiator test 7"
*	Return value:	0, on success
*					-Exxxx, on failure
******************************************************************************/

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
int
iscsi_initiator_proc_info(struct Scsi_Host *sh, char *buffer, char **start,
			  off_t offset, int length, int inout)
{
	unsigned long flags;
	int pos, begin, result;
	struct session *sess;

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_proc_info\n");
	TRACE(TRACE_DEBUG, "host %p, priv %p, buffer %p, start %p\n",
	      sh, &sh->hostdata[0], buffer, start);
	TRACE(TRACE_DEBUG, "offset %ld, length %d, inout %d\n",
	      offset, length, inout);
#else
int
iscsi_initiator_proc_info(char *buffer, char **start, off_t offset,
			  int length, int inode, int inout)
{
	unsigned long flags;
	int pos, begin, result;
	struct session *sess;

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_proc_info\n");
	TRACE(TRACE_DEBUG, "buffer %p, offset %ld, length %d, inout %d\n",
	      buffer, offset, length, inout);
#endif

	if (inout) {
		/* write to this file, buffer should be:
		 * "iscsi_initiator ip 1234abcd port 4000 target 1 lun 2 cid 3"
		 * or
		 * "iscsi_initiator manage 0
		 * <neg_info number for "set/setr/unset/setp/restore/force">
		 * <parameter for "set/setr/unset/setp/restore/force"> "
		*/


		if (global_host == NULL) {
			/* this should never happen */
			TRACE_ERROR("global_host is NULL\n");
			result = -ENODEV;
		} else if (global_hostdata == NULL) {
			/* this should never happen */
			TRACE_ERROR("global_hostdata is NULL\n");
			result = -ENODEV;
		} else if (buffer == NULL || length > PAGE_SIZE)
			result = -EINVAL;
		else {
			result = length;
			if (length == PAGE_SIZE)
				length--;
			/* make the buffer into a string */
			buffer[length] = '\0';

			TRACE(TRACE_ISCSI,
			      "iSCSI write to proc_info (%d): %s\n",
			      length, buffer);

			if (length >= 23
			       && strncmp(buffer, "iscsi_initiator manage ", 23)
									== 0) {
				result = scan_manage_stuff(buffer + 23, result);
			} else if (length >= 21
				 && strncmp(buffer, "iscsi_initiator test ", 21)
									== 0) {
				/* Trigger test actions in the initiator */
				result = scan_test_stuff(buffer + 20, result);
			} else if (length >= 19
				&& strncmp(buffer, "iscsi_initiator ip ", 19)
									== 0) {
				result = scan_ip_stuff(buffer + 19, result);
			} else {
				TRACE(TRACE_ISCSI,
				      "Illegal write to proc_info\n");
				result = -EINVAL;
			}
		}
	} else {
		/* read from this file
		 * this is what gets printed
		 * by "cat /proc/scsi/iscsi_initiator/0"
		 */
		pos = sprintf(buffer, "%s\n", OUR_NAME);

		/*  When using global data, be sure nobody else accesses it */
		UNH_LOCK(&host_data_lock, flags);

		if (global_host != NULL && global_hostdata != NULL) {
			pos = print_host_info(global_host, buffer, pos);
			pos += sprintf(buffer + pos, "host has %u session%s\n",
				      global_hostdata->nsessions,
				      global_hostdata->nsessions==1 ? "" : "s");
			if (global_hostdata->nsessions == 0 ) {
				pos += print_config_info(
						global_hostdata->target_data[0]
								.param_tbl,
						buffer + pos);
			}
			list_for_each_entry(sess,
					    &global_hostdata->session_list,
					    link) {
				pos = print_session_info(sess, buffer, pos);
				/* if initiator has lots of sessions,
				 * printing the session parameters for all of
				 * them here may overflow the proc 4K buffer.
				 */
				/********** RDR
				pos += print_session_params(sess->oper_param,
							  *sess->session_params,
							  buffer + pos);
				**********/
				pos += sprintf(buffer + pos, "\n");
			}
#ifdef ISCSI_STATS
			pos += dump_instance_stats(buffer + pos);
#endif
		} else {
			pos += sprintf(buffer + pos, "\n");
		}

		/* release our exclusive access to the global data */
		UNH_UNLOCK(&host_data_lock, flags);

		pos += print_my_kmemory(buffer + pos);
		pos += print_iovs(buffer + pos);
		if (pos < offset) {
			/* reader already read everything we have to say,
			 * send back EOF
			 */
			result = 0;
			begin = pos;
		} else {
			/* reader had not read everything yet,
			 * send back something
			 */
			result = pos;
			begin = 0;
		}
		/* Start of wanted data */
		*start = buffer + (offset - begin);
		result -= (offset - begin);
		if (result > length)
			result = length;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_initiator_proc_info, result %d\n",
	      result);

	return result;
}

#ifdef ISCSI_STATS

static int __attribute__ ((no_instrument_function))
dump_instance_stats(char *page)
{
	int len=0;
	struct iscsi_instance_stats *istats;
	char ip_string[INET6_ADDRSTRLEN+2], port_string[8];

	istats = &(global_hostdata->instance_stats);
	len += sprintf(page + len, "Session Login Failure = %d\n",
			istats->iscsi_intr_login_failure);

	len += sprintf(page + len, "Last Login Failure Target Name= %s\n",
			(istats->iscsi_intr_last_tgt_failure_name) ?
			istats->iscsi_intr_last_tgt_failure_name: "<NULL>");

	len+=sprintf(page+len, "Last Target Failure Address = ");
	if (istats->iscsi_intr_last_tgt_failure_address == NULL)
		len += sprintf(page + len, "<NULL>\n");
	else if (cnv_inet_to_string(istats->iscsi_intr_last_tgt_failure_address,
												ip_string, port_string) > 0)
		len += sprintf(page +len, "%s:%s\n", ip_string, port_string);

	len += sprintf( page + len, "Last Login Failure Time = %ld:%ld\n",
			istats->iscsi_intr_last_failure_time.tv_sec,
			istats->iscsi_intr_last_failure_time.tv_usec);
#if USE_PRINT_DATE
	len += print_date(page + len, &(istats->iscsi_intr_last_failure_time));
#endif

	len += sprintf(page + len, "Last Login Failure Type = ");
	switch (istats->iscsi_intr_last_failure_type) {
	case ACCEPT_RESPONSE:
		len += sprintf(page + len, "ACCEPT_RESPONSE\n");
		break;
	case REDIRECT_RESPONSE:
		len += sprintf(page + len, "REDIRECT_RESPONSE\n");
		break;
	case AUTH_FAIL_RESPONSE:
		len += sprintf(page + len, "AUTH_FAIL_RESPONSE\n");
		break;
	case AUTHENTICATE_FAIL:
		len += sprintf(page + len, "AUTHENTICATE_FAIL\n");
		break;
	case NEGOTIATE_FAIL:
		len += sprintf(page + len, "NEGOTIATE_FAIL\n");
		break;
	case OTHER_FAIL:
		len += sprintf(page + len, "OTHER_FAIL\n");
		break;
	case UNKNOWN_FAILURE:
	default:
		if(istats->iscsi_intr_login_failure)
			len += sprintf(page +len, "<UNKNOWN FAILURE>\n");
		else
			len += sprintf(page +len, "<NULL>\n");
		break;
	}

	len += sprintf(page +len, "Login Accept Response = %u\n",
		       istats->iscsi_intr_login_accept_rsp);
	len += sprintf(page +len, "Login Redirect Response = %u\n",
		       istats->iscsi_intr_login_redirect_rsp);
	len += sprintf(page +len, "Login Authenticate Fail Response = %u\n",
		       istats->iscsi_intr_login_auth_fail_rsp);
	len += sprintf(page +len, "Login Authentication Failure = %u\n",
		       istats->iscsi_intr_login_auth_fail_rsp);
	len += sprintf(page +len, "Login Negotiate Failure = %u\n",
		       istats->iscsi_intr_login_negotiate_fails);
	len += sprintf(page +len, "Login Other Failure = %u\n",
		       istats->iscsi_intr_login_other_fail_rsp);
	len += sprintf(page +len, "Normal Logouts = %u\n",
		       istats->iscsi_intr_logout_normals);
	len += sprintf(page +len, "Other Logouts = %u\n",
		       istats->iscsi_intr_logout_others);
	return len;
}

static int __attribute__ ((no_instrument_function))
read_proc_file(char *page, char **start, off_t off, int count, int *eof,
			   void *data)
{
	unsigned long flags;
	int len = 0;
	struct proc_data *proc_data = (struct proc_data *) data;
	struct session *sess;
	struct connection *conn;

	/*  When using global session list, be sure nobody else accesses it */
	UNH_LOCK(&host_data_lock, flags);

	if (proc_data == NULL || proc_data->file != STAT)
		goto out;

	len = sprintf(page, "%s\n", OUR_NAME);

	/* find the session for this target_id */
	sess = find_session_by_id(proc_data->sess_id, global_hostdata);
	if (sess != NULL) {
		/* found the session */
		if (proc_data->type & SESSION) {
			len += dump_sess_info(sess, page+len);
		} else if (proc_data->type & CONNECTION) {
			/* find the connection within this session */
			for (conn = sess->connection_head;
			     conn != NULL;
			     conn = conn->next) {
				if (conn->connection_id
						== proc_data->conn_id) {
					len += dump_conn_info(conn, page+len);
					break;
				}
			}
		}
	}
out:
	/* release our exclusive access to the global structures */
	UNH_UNLOCK(&host_data_lock, flags);
	return len;
}

/*
 * executed only by command process.
 * host_data_lock MUST be held by the calling process/thread
 * Called only by read_proc_file().
 * dump_sess_info() dumps the session specific info.
 * It dumps the session negotiated values and statistics
 * such as cmd pdus txd, response pdus txd, data bytes txd,
 * data bytes rxd, error count etc...
 */
static int __attribute__ ((no_instrument_function))
dump_sess_info(struct session *sess, char *page)
{
	int len;
	struct iscsi_session_stats *sess_stats;

	if (sess == NULL)
		return 0;

	len = print_session_info(sess, page, 0);

	sess_stats = sess->sess_stats;
	len += print_session_params(sess->oper_param,
				    *sess->session_params,
				    page + len);
	len += sprintf(page +len,
			"\nPDUs Transmitted\n------------------\n");
	len += sprintf(page +len,
			"%12s %12s %12s %12s %12s %12s\n"
			"%12u %12u %12u %12u %12u %12u\n",
			"NopOut", "SCSICommand", "TaskMgmtReq", "TextReq",
			"DataOut", "SNACK",
			sess_stats->iscsi_ssn_noptx_pdus,
			sess_stats->iscsi_ssn_scsicmd_pdus,
			sess_stats->iscsi_ssn_tskmgmtcmd_pdus,
			sess_stats->iscsi_ssn_text_pdus,
			sess_stats->iscsi_ssn_dataout_pdus,
			sess_stats->iscsi_ssn_snack_pdus);

	len += sprintf(page +len, "PDUs Received\n------------------\n");
	len += sprintf(page +len,
			"%12s %12s %12s %12s %12s %12s %12s %12s\n"
			"%12u %12u %12u %12u %12u %12u %12u %12u\n",
			"NopIn",  "SCSIResp", "TaskMgmtResp", "TextResp",
			"DataIn", "R2T", "AsyncMess", "Reject",
			sess_stats->iscsi_ssn_noprx_pdus,
			sess_stats->iscsi_ssn_rsp_pdus,
			sess_stats->iscsi_ssn_tskmgmtrsp_pdus,
			sess_stats->iscsi_ssn_textrsp_pdus,
			sess_stats->iscsi_ssn_datain_pdus,
			sess_stats->iscsi_ssn_r2t_pdus,
			sess_stats->iscsi_ssn_async_pdus,
			sess_stats->iscsi_ssn_rjt_pdus);

	len += sprintf(page +len,
			"Data Bytes Txd = %lld\nData Bytes Rxd = %llu\n",
			sess_stats->iscsi_ssn_txdata_octets,
			sess_stats->iscsi_ssn_rxdata_octets);

	len += sprintf(page +len, "Input DataDigest Errors = %u\n",
			sess->sess_stats->iscsi_ssn_digest_err);
#if USE_CONTROL_FILE
	len += sprintf(page +len, "Connection Timeout Error = %u\n",
			sess->sess_stats->iscsi_ssn_timeout_err);
#endif
	return len;
}

/*
 * executed only by command process.
 * host_data_lock MUST be held by the calling process/thread
 * Called only by read_proc_file().
 * dump_conn_info () prints the connection stats
 * CID, Connection STATE, Protocol Type, Local Address, Port
 * Remote Address, Port, Remote Address Type, Max Receive Data
 * Max Send Data, Receive Marker, Send Marker, Header Digest
 * Data Digest etc..
 */
static int __attribute__ ((no_instrument_function))
dump_conn_info(struct connection *conn, char *buf)
{
	int len;

	if (conn == NULL)
		return 0 ;

	len = print_connection_info(conn, buf, 0);

	if(conn->connection_state == CONNECTION_FULL_FEATURE_PHASE ||
			conn->connection_state == CONNECTION_CONNECTED) {
		len += sprintf(buf + len, "%16s: %s\n", "LUN addressing",
			       conn->connection_flags & USE_FLAT_SPACE_LUN
			       ? "flat space" : "peripheral device");
		len += sprintf(buf +len,
				"%16s: %s\n", "Header Digest",
				(conn->connection_flags & USE_HEADERDIGEST)
					? "CRC32C" : "None");
		len += sprintf(buf +len,
				"%16s: %s\n", "Data Digest",
				(conn->connection_flags & USE_DATADIGEST)
					? "CRC32C" : "None");
		len += sprintf(buf +len, "%16s: %u Bytes\n",
				"Max Send DSL", conn->max_send_length);
		len += sprintf(buf +len, "%16s: %u Bytes\n",
				"Max Recv DSL", conn->max_recv_length);
#if 0
		len += sprintf(buf +len, "%16s: %s\n", "Send Marker", "TODO");
		len += sprintf(buf +len, "%16s: %s\n", "Receive Marker",
				"TODO");
#endif
		len += sprintf(buf +len, "%16s: %lld\n", "Data Bytes Sent",
				conn->iscsi_conn_txdata_octets);
		len += sprintf(buf +len, "%16s: %lld\n", "Data Bytes Rcvd",
				conn->iscsi_conn_rxdata_octets);
	}
	return len;
}

#if USE_CONTROL_FILE
static int __attribute__ ((no_instrument_function))
write_proc_file(struct file *file, const char *buffer, unsigned long count,
		void *data)
{
	/* Nothing to do here */
	return count;
}
#endif

/*
 * add_iscsi_proc_entry() routine accepts a void pointer,
 * the void ptr should be either a session ptr or a connection
 * ptr. The void ptr is typecast to either of them depending upon
 * "type".
 * The session proc entries include a "target#"<num> dir and
 * stat and ctrl file in it.
 * The connection proc entries include a "conn#"<cid> dir and
 * stat and ctrl file in it.
 * Same read and write routines are used for all files, the operation
 * is distinguished based on the "data" value in proc_data.
 */
int
add_iscsi_proc_entry(void *ptr, int type)
{
	int ret = -1;
	struct proc_dir_entry *dir = NULL, *stat_file = NULL;
#if USE_CONTROL_FILE
	struct proc_dir_entry *ctrl_file = NULL;
#endif
	struct session *sess = (struct session *) ptr;
	struct connection *conn = (struct connection *) ptr;
	char dir_name[32];
	struct proc_data *stat_data
		= (struct proc_data *) kmalloc(sizeof (struct proc_data),
							     GFP_KERNEL);
#if USE_CONTROL_FILE
	struct proc_data *ctrl_data
		= (struct proc_data *) kmalloc(sizeof (struct proc_data),
							     GFP_KERNEL);
#endif

	sprintf(dir_name, "%s-%d",
			(type & SESSION) ? TARGET_FNAME : CONNECTION_FNAME,
			(type & SESSION) ? sess->scsi_target_id
					  : conn->connection_id);
	dir = proc_mkdir(dir_name, (type & SESSION)
		  ? global_host->hostt->proc_dir : conn->my_session-> proc_dir);
	if (dir == NULL) {
		TRACE(TRACE_DEBUG, "add_proc: creation of proc dir %s failed\n",
		      dir_name);
		ret = -ENOMEM;
		goto error_out1;
	}
	/* set the proc dir entry in session structure */
	if (type & SESSION) {
		sess->proc_dir = dir;
		stat_data->sess_id = sess->scsi_target_id;
	} else {
		/* Create connection specific stat and control files */
		conn->proc_dir = dir;
		stat_data->sess_id = conn->my_session->scsi_target_id;
		stat_data->conn_id = conn->connection_id;
	}

	stat_data->type = type;
	stat_data->file = STAT;
	stat_file
		= create_proc_read_entry(STAT_FNAME, 0444, dir, read_proc_file,
				         (void *) stat_data);
	if (stat_file == NULL) {
		TRACE(TRACE_DEBUG,
		      "add_iscsi_proc_entry: creation of proc file %s failed\n",
		      STAT_FNAME);
		ret = -ENOMEM;
		goto error_out2;
	}
#if USE_CONTROL_FILE
	/* For time being, there is nothing that can be written
	 * into a ctrl for connection. For sess level ctrl file,
	 * we can only set the scheduling algo.
	 */
	ctrl_file = create_proc_entry(CONTROL_FNAME, 0644, dir);
	if (ctrl_file == NULL) {
		TRACE(TRACE_DEBUG,
		      "add_iscsi_proc_entry: creation of proc file %s failed\n",
		      CONTROL_FNAME);
		ret = -ENOMEM;
		goto error_out3;
	}
	ctrl_data->type = type;
	ctrl_data->file = CONTROL;
	if (type & SESSION) {
		ctrl_data->sess_id = sess->iscsi_target_id;
	} else {
		ctrl_data->sess_id = conn->my_session->iscsi_target_id;
		ctrl_data->conn_id = conn->connection_id;
	}
	ctrl_file->data = (void *) ctrl_data;
	ctrl_file->read_proc = read_proc_file;
	ctrl_file->write_proc = write_proc_file;
	ctrl_file->owner = THIS_MODULE;
#endif
	ret = 0;
	return ret;
#if USE_CONTROL_FILE
error_out3:
	remove_proc_entry(STAT_FNAME, dir);
#endif

error_out2:
	remove_proc_entry(dir_name, global_host->hostt->proc_dir);

error_out1:
	kfree(stat_data);
#if USE_CONTROL_FILE
	kfree(ctrl_data);
#endif
	return ret;
}

/*
 * remove_iscsi_proc_entry() accepts a void ptr and type,
 * The void ptr is typecast to either session or connection
 * based on the type.
 */
void
remove_iscsi_proc_entry(void *ptr, int type)
{
	struct session *sess = (struct session *) ptr;
	struct connection *conn = (struct connection *) ptr;
	char dir_name[32];

	remove_proc_entry(STAT_FNAME,
			  (type & SESSION) ? sess->proc_dir : conn->proc_dir);
	/*not sure if I have to free the "data" mem. */
#if USE_CONTROL_FILE
	remove_proc_entry(CONTROL_FNAME,
			  (type & SESSION) ? sess->proc_dir : conn->proc_dir);
#endif
	sprintf(dir_name, "%s-%d",
		(type & SESSION) ? TARGET_FNAME : CONNECTION_FNAME,
		(type & SESSION) ? sess->scsi_target_id : conn->connection_id);
	remove_proc_entry(dir_name, (type & SESSION)
		? global_host->hostt->proc_dir : conn->my_session->proc_dir);
	if (type & SESSION)
		sess->proc_dir = NULL;
	else
		conn->proc_dir = NULL;
	return;
}

#if USE_PRINT_DATE

#define SECS_PER_HOUR (60 * 60)
#define SECS_PER_DAY (60 * 60 * 24)
#define isleap(year) \
	((year) % 4 == 0 && ((year) % 100 != 0 || (year) % 400 == 0))
#define DIV(a, b) ((a) / (b) - ((a) % (b) < 0))
#define LEAPS_THRU_END_OF(y) (DIV (y, 4) - DIV (y, 100) + DIV (y, 400))
#define TIMEZONE (-5 * 60 * 60)
static int __attribute__ ((no_instrument_function))
print_date (char *page, const struct timeval *t)
{
	int tm_sec, tm_min, tm_hour, tm_mday, tm_mon, tm_year;

	static const __u16 mon_yday[2][13] = {
           /* Normal years.  */
           { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 },
           /* Leap years.  */
           { 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366 }
        };

        long int days, rem, y;
        const __u16 *ip;

        days = t->tv_sec / SECS_PER_DAY;
        rem = t->tv_sec % SECS_PER_DAY;
        rem += TIMEZONE;
        while (rem < 0) {
                rem += SECS_PER_DAY;
                --days;
        }
        while (rem >= SECS_PER_DAY) {
                rem -= SECS_PER_DAY;
                ++days;
        }
        tm_hour = rem / SECS_PER_HOUR;
        rem %= SECS_PER_HOUR;
        tm_min = rem / 60;
        tm_sec = rem % 60;
        y = 1970;

        while (days < 0 || days >= (isleap (y) ? 366 : 365)) {
                long int yg = y + days / 365 - (days % 365 < 0);
                days -= ((yg - y) * 365
                        + LEAPS_THRU_END_OF (yg - 1)
                        - LEAPS_THRU_END_OF (y - 1));
                y = yg;
        }
        tm_year = y - 1900;
        if (tm_year != y - 1900)
                return ;
        ip = mon_yday[isleap(y)];
        for (y = 11; days < (long int) ip[y]; --y)
                continue;
        days -= ip[y];
        tm_mon = y;
        tm_mday = days + 1;
	return sprintf(page,
		       "DATE %.2d/%.2d/%d- %.2d:%.2d:%.2d", tm_mday, tm_mon + 1,
		       tm_year + 1900, tm_hour, tm_min, tm_sec);
}
#endif

#endif
