/*	initiator/initiator_rx.c
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This file contains the functions for iscsi initiator code that are
 *	responsible for reading and processing PDUs sent by the target.
 *
 *	Copyright (C) 2001-2004 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name IOL and/or UNH may not be used to endorse or promote 
 *	products derived from this software without specific prior 
 * 	written permission.
*/
#include "iscsi_initiator.h"
#include "initiator_utilities.h"
#include "initiator_error_rec.h"
#include "../common/lun_packing.h"

/* attributes for each opcode that can be received from target */
#define ITT_OPTIONAL	0x0000
#define ITT_REQUIRED	0x0100
#define ITT_RESERVED	0x0200
#define ITT_MASK	0x0300
#define F_MUST_BE_1	0x0400
#define ILLEGAL_REPLY	0x0800
#define DSL_MUST_BE_0	0x1000
#define LUN_MUST_BE_0	0x2000

/* table of attributes (in byte 1) and related opcode (in byte 0)
 * for each opcode that can be received from target
 */
static __u16 opcode_attributes[32] = {
/* 0x20	NopIn ==> NopOut */
	F_MUST_BE_1 | ITT_OPTIONAL | ISCSI_INIT_NOP_OUT,
/* 0x21	SCSI Response ==> SCSI Command */
	LUN_MUST_BE_0 | F_MUST_BE_1 | ITT_REQUIRED | ISCSI_INIT_SCSI_CMND,
/* 0x22	Task Management Response ==> Task Management Request */
	LUN_MUST_BE_0 | DSL_MUST_BE_0 | F_MUST_BE_1 | ITT_REQUIRED
						| ISCSI_INIT_TASK_MGMT_CMND,
/* 0x23	Login Response ==>Login Request */
	ITT_REQUIRED | ISCSI_INIT_LOGIN_CMND,
/* 0x24	Text Response ==>Text Request */
	ITT_REQUIRED | ISCSI_INIT_TEXT_CMND,
/* 0x25	DataIn ==> SCSI Command */
	ITT_REQUIRED | ISCSI_INIT_SCSI_CMND,
/* 0x26	Logout Response ==> Logout Request */
	LUN_MUST_BE_0 | DSL_MUST_BE_0 | F_MUST_BE_1 | ITT_REQUIRED
						| ISCSI_INIT_LOGOUT_CMND,
/* 0x27 */ ILLEGAL_REPLY,
/* 0x28 */ ILLEGAL_REPLY,
/* 0x29 */ ILLEGAL_REPLY,
/* 0x2a */ ILLEGAL_REPLY,
/* 0x2b */ ILLEGAL_REPLY,
/* 0x2c */ ILLEGAL_REPLY,
/* 0x2d */ ILLEGAL_REPLY,
/* 0x2e */ ILLEGAL_REPLY,
/* 0x2f */ ILLEGAL_REPLY,
/* 0x30 */ ILLEGAL_REPLY,
/* 0x31	R2T ==> SCSI Command */
	DSL_MUST_BE_0 | F_MUST_BE_1 | ITT_REQUIRED | ISCSI_INIT_SCSI_CMND,
/* 0x32	Asynch ==> -none- */
	F_MUST_BE_1 | ITT_RESERVED,
/* 0x33 */ ILLEGAL_REPLY,
/* 0x34 */ ILLEGAL_REPLY,
/* 0x35 */ ILLEGAL_REPLY,
/* 0x36 */ ILLEGAL_REPLY,
/* 0x37 */ ILLEGAL_REPLY,
/* 0x38 */ ILLEGAL_REPLY,
/* 0x39 */ ILLEGAL_REPLY,
/* 0x3a */ ILLEGAL_REPLY,
/* 0x3b */ ILLEGAL_REPLY,
/* 0x3c */ ILLEGAL_REPLY,
/* 0x3d */ ILLEGAL_REPLY,
/* 0x3e */ ILLEGAL_REPLY,
/* 0x3f	Reject ==> -none- */
	LUN_MUST_BE_0 | F_MUST_BE_1 | ITT_RESERVED
};


#ifdef ISCSI_CHECK_PDU_FORMAT
/* utility function called by all rx___() functions when they
 * determine that their reserved fields are not all 0
 */
static void __attribute__ ((no_instrument_function))
reserved_not_0(char *pdu_name)
{
	if (TRACE_TEST(TRACE_ISCSI)) {
		TRACE_WARNING("reserved fields in %s header not all 0\n",
			      pdu_name);
	}
}
#endif


static void __attribute__ ((no_instrument_function))
debug_print_ip(struct connection *conn, char *message)
{
	char ip_string[INET6_ADDRSTRLEN+2], port_string[8];

	if (cnv_inet_to_string(conn->ip_address, ip_string, port_string) > 0) {
		TRACE(TRACE_DEBUG, "%s from %s:%s\n", message,
							ip_string, port_string);
	}
}

/*
 * searches conn's pending_commands list for a command with init_task_tag that
 * matches this_itt.
 * returns pointer to that command if found,
 *	   NULL if not found
 */
static inline struct command * __attribute__ ((no_instrument_function))
find_command_by_itt(__u32 this_itt, struct connection *conn)
{
	struct command *cmd;

	list_for_each_entry(cmd, &conn->pending_commands, link) {
		if (cmd->init_task_tag == this_itt) {
			return cmd;
		}
	}
	return NULL;
}

/****************************************************************************
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_rx_thread().
 * This function processes the DataIn payload of a DataIn PDU. It fills the
 * scatter-gather buffers provided by the SCSI Mid-level. The data digest
 * is also checked (if such checking has been enabled).
 *
 * Parameters:	- current_connection which is a ptr to the struct where the SCSI
 *				DataIn Header was received.
 *		- related_command which is a ptr to the struct for the SCSI READ
 *				command to which this DataIn is related.
 * Return Value: >0 success, session lock is locked
 *		 =0 failure, session lock is locked
 *		 <0 failure, session lock is NOT locked
 *
 * While reading, the following variables in the related_command maintain state:
 *	r2t_range_list.limit	sum of the size of all finally accepted DataIn
 *				pdus in command -- incremented by size of
 *				every good recvd data-in pdu that will not be
 *				retransmitted later.
 *	r2t_range_list.offset	offset to first DataIN pdu in current sequence.
 *				set at start of an out-of-order seq if
 *				in-order pdus. set at end of sequence
 *				of out-of-order pdus
 *	data_offset		expected offset in next in-order DataIn pdu
 *				incremented by size of every recvd DataIn pdu.
 *				set at start of an out-of-order sequence
 *	recvd_length		sum of size of recvd data-in pdus in this
 *				sequence incremented by size of every recvd
 *				data-in pdu reset to 0 at end of each sequence .
 *				used to test for start of a new sequence when 0.
 *	pdu_range_list.limit	largest limit of any out-of-order pdu
 *				in current seq .
 *	pdu_range_list.offset	smallest offset of any out-of-order pdu in
 *				cur seq used to set r2t_range_list.offset at
 *				end of seq.
 ****************************************************************/
static int
rx_data(struct connection *current_connection,
	__u32 local_itt, struct command *related_command,
	__u32 size, char *string)
{
	struct iscsi_targ_scsi_data_in *data_in_header;
	int received = 1, delta;
	int over_flow_data_size = 0;
	struct session *related_session;
	struct init_error_rec err_rec;
	char *over_flow_buffer = NULL;
	__u32 lun;
	__u32 sg_size;

	TRACE(TRACE_ENTER_LEAVE, "Enter rx_data, current_connection %p\n",
	      current_connection);

	data_in_header = (struct iscsi_targ_scsi_data_in *)
						current_connection->rx_buf;

	sg_size = 0;
	if (likely(related_command != NULL && related_command->SCpnt != NULL))
		sg_size = related_command->SCpnt->use_sg;

	related_session = current_connection->my_session;

#ifdef ISCSI_CHECK_PDU_FORMAT
	/* Check if reserved fields are zero or not */
	if (unlikely(!((data_in_header->rsvd1 == 0)
	      && ((data_in_header->flags & (W_BIT|BRO_BIT|BRU_BIT)) == 0)))) {
		reserved_not_0("SCSI DataIn");
	}
#endif

	if (unlikely(data_in_header->flags & A_BIT)) {
		/* A bit is 1 -- MUST have ErrorRecoveryLevel > 0,
				and TTT and LUN must be valid */
		/* Draft 20, Section 10.7.2 A (Acknowledge) bit
		 * "The Target MUST NOT set to 1 the A bit for sessions with
		 * ErrorRecoveryLevel=0.  The initiator MUST ignore the A bit
		 * set to 1 for sessions with ErrorRecoveryLevel=0.
		 */
		if (related_session->oper_param->ErrorRecoveryLevel == 0) {
			TRACE_WARNING("Ignoring A_Bit=1 in DataIn, ITT %u, with"
					" ErrorRecoveryLevel=0\n", local_itt);
		}

		/* Draft 20, Section 10.7.4 Target Transfer Tag and LUN
		 * "On incoming data, the Target Transfer Tag and LUN MUST be
		 * provided by the target if the A bit is set to 1; otherwise
		 * they are reserved. ...
		 *
		 * The Target Transfer Tag values are not specified by this
		 * protocol except that the value 0xffffffff is reserved and
		 * means that the Target Transfer Tag is not supplied.  If the
		 * Target Transfer Tag is provided, then the LUN field MUST hold
		 * a valid value and be consistent with whatever was specified
		 * with the command; otherwise, the LUN field is reserved."
		 */
		if (data_in_header->target_xfer_tag == ALL_ONES) {
			/* TTT is reserved, this is an error */
			TRACE_ERROR
			    ("DataIn with A bit = 1 has reserved TTT 0x%08x, "
			     "expected valid value\n", ALL_ONES);
		} else if (related_command != NULL &&
				related_command->SCpnt != NULL) {
			/* TTT is valid, LUN must be same as in related Cmnd */
			lun = unpack_lun((__u8 *) & data_in_header->lun);

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
			if (lun != related_command->SCpnt->device->lun) {
				TRACE_ERROR(
				  "DataIn for ITT %u has LUN %u, expected %u\n",
				  local_itt, lun,
				  related_command->SCpnt->device->lun);
			}
#else
			if (lun != related_command->SCpnt->lun) {
				TRACE_ERROR(
				  "DataIn for ITT %u has LUN %u, expected %u\n",
				  local_itt, lun,
				  related_command->SCpnt->lun);
			}
			/* else LUN matches the command's LUN, all ok */
#endif
		}
	} else {
		/* A bit is 0 -- TTT can be reserved, doesn't have to be */
		if (data_in_header->target_xfer_tag == ALL_ONES) {
			/* TTT is reserved */
			if (unlikely(data_in_header->lun != 0LL)) {
				/* TTT is reserved, LUN is not, that's bad */
				TRACE_ERROR("DataIn with reserved TTT 0x%08x, "
					    "non-reserved LUN %s\n",
					    ALL_ONES,
					    string_llx(data_in_header->lun,
					    			string));
			}
			/* else LUN is also reserved, all ok */
		} else if (related_command != NULL &&
				related_command->SCpnt != NULL) {
					/* TTT has valid value, LUN must
						be same as in related Cmnd */
			lun = unpack_lun((__u8 *) & data_in_header->lun);

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
			if (unlikely(lun != related_command
							->SCpnt->device->lun)) {
				TRACE_ERROR
				    ("DataIn with ITT %u has invalid LUN %u, "
				     "expected %u\n", local_itt, lun,
				     related_command->SCpnt->device->lun);
			}
#else
			if (unlikely(lun != related_command->SCpnt->lun)) {
				TRACE_ERROR
				    ("DataIn with ITT %u has invalid LUN %u, "
				     "expected %u\n", local_itt, lun,
				     related_command->SCpnt->lun);
			}
#endif
		}
	}

#ifdef ISCSI_CHECK_PDU_FORMAT
	if (data_in_header->flags & S_BIT) {
		/* this command is a "phase collapse" with "good" status */
		if (unlikely(!(data_in_header->flags & F_BIT))) {
			/* F bit not set, but it should be when S bit is set */

			/* Draft 20, Section 10.7.3 Flags (byte 1)
			 * "bit 7 S (status) - set to indicate that the Command
			 * Status field contains status.  If this bit is set
			 * to 1, the F bit MUST also be set to 1."
			 */
			TRACE_ERROR(
			    "DataIn with ITT %u, S bit = 1 has F Bit = 0, "
			    "expected 1\n", local_itt);
		}
	}
#endif

	data_in_header->offset = ntohl(data_in_header->offset);
	data_in_header->data_sn = ntohl(data_in_header->data_sn);

	if (unlikely(related_command==NULL || related_command->SCpnt==NULL)) {
		/* We lost the ptr to the command or to the SCSI
		   Midlevel struct, can't do anything */
		TRACE(TRACE_ERROR_RECOVERY,
		      "No related command/SCpnt: ITT %u, size %u, offset %u, "
		      "DataSN %u\n", local_itt, size, data_in_header->offset,
		      data_in_header->data_sn);
		if (size > 0) {
			/* read in and ignore this PDU's data payload */
			received = recv_ignore_data(current_connection, size);
		}
	} else {
		/* have matched the DataIn to a legitimate SCSI READ request */
		/* Check if DataSN field in header is what we expected */
		delta = data_in_header->data_sn - related_command->data_in_sn;
		if (unlikely(delta != 0)) {
			TRACE_ERROR("DataIn with ITT %u, got DataSN %u, "
				    "expected %u\n", local_itt,
				    data_in_header->data_sn,
				    related_command->data_in_sn);
			if (delta < 0) {
				/* duplicate DataSn, just silently ignore
					this PDU */
				if (size > 0) {
					/* read in and ignore this PDU's
						data payload */
					received = recv_ignore_data(
						current_connection,
						size);
				}
				goto out;
			} else {
				/* DataIn PDU Sequence Error - SAI */
				err_rec.curr_conn = current_connection;
				err_rec.related_cmd = related_command;
				err_rec.err_type = SEQUENCE_ERR;
				if ((received
				       = init_do_error_recovery(&err_rec)) <= 0)
					goto out;
				/* Fall thru and use this out-of-order data.
				 * Force the expected data offset to the
				 * one received in order to avoid a
				 * cascade of error messages.
				 */
				related_command->data_offset
						= data_in_header->offset;
				/* for the same reason, force the start
					of a new sequence */
				related_command->recvd_length = 0;
			}
		}

		/* Draft 20, Section 10.7.1 F (Final) Bit
		 * "Splitting the data stream into sequences does not affect
		 * DataSN counting on Data-In PDUs."
		 *
		 * Draft 20, Section 10.7.5 DataSN
		 * "For input (read) or bidirectional Data-IN PDUs, the DataSN
		 * is the input PDU number within the data transfer for the
		 * command identified by the Initiator Task Tag."
		 *
		 * Therefore, we do not reset data_in_sn at the end of a data-in
		 * sequence -- it continues incrementing over all data-in
		 * sequence boundaries.
		 */
		related_command->data_in_sn++;

		/* Check if we are overflowing the SCSI Mid-level buffers */
		if (unlikely((over_flow_data_size = data_in_header->offset
			+ size - related_command->SCpnt->request_bufflen) > 0)){
			TRACE_ERROR("DataIn with ITT %u, Received buffer size "
				   "%u bytes will overflow SCSI buffer by %d\n",
				   local_itt, size, over_flow_data_size);
			if (related_command->SCpnt->request_bufflen
						> data_in_header->offset)
				size = related_command->SCpnt->request_bufflen
					- data_in_header->offset;
			else
				size = 0;
		} else {
			/* be sure this is not negative for use later on! */
			over_flow_data_size = 0;
		}

		if (unlikely(TRACE_TEST(TRACE_DEBUG))) {
			TRACE(TRACE_DEBUG, "Local data offset %u, "
			      "request length %d\n",
			      related_command->data_offset,
			      related_command->SCpnt->request_bufflen);
			TRACE(TRACE_DEBUG,
			      "DataIn hdr data offset %u, payload len %u\n",
			      data_in_header->offset, size);
		}

		if (related_command->recvd_length == 0) {
			/* this is first PDU in a sequence */
			related_command->r2t_range_list.offset
						= data_in_header->offset;
			/* can sequences be out-of-order? */
			if (unlikely(!related_session->oper_param
							->DataSequenceInOrder)){
				/* DataSequenceInOrder==No, so seqs can be
				 * out-of-order
				 */
				related_command->data_offset
						= data_in_header->offset;
			}
			/* can the PDU's within a sequence be out-of-order? */
			if (unlikely(!related_session->oper_param
							->DataPDUInOrder)) {
				/* DataPDUInOrder==No, so PDUs can be
				 * out-of-order
				 */
				related_command->pdu_range_list.limit = 0;
				related_command->pdu_range_list.offset
								= ALL_ONES;
			}
		}

		/* can the PDU's within a sequence be out-of-order? */
		if (unlikely(!related_session->oper_param->DataPDUInOrder)) {
			/* DataPDUInOrder==No, so PDU can be out-of-order */
			__u32 limit;

			if (related_command->pdu_range_list.offset
						> data_in_header->offset) {
				/* this is smallest offset seen so far */
				related_command->pdu_range_list.offset
						= data_in_header->offset;
			}
			limit = data_in_header->offset + size;
			if (related_command->pdu_range_list.limit < limit) {
				/* this is largest limit seen so far */
				related_command->pdu_range_list.limit = limit;
			}
			/* each DataIn PDU is its own range */
			merge_offset_length(&related_command->pdu_range_list,
					    data_in_header->offset, size);
		} else {
			/* DataPDUInOrder==Yes, so PDU must be
			 * in-order within seq.
			 * Does buffer offset in DataIn header have
			 * expected value?
			 */
			if (unlikely(related_command->data_offset
						!= data_in_header->offset)) {
				TRACE_ERROR(
					"DataIn with ITT %u, DataSN %u, size "
					"%u, has offset %u, expected %u\n",
					local_itt, data_in_header->data_sn,
					size, data_in_header->offset,
					related_command->data_offset);
			}
		}

		if (unlikely(size == 0)) {
			TRACE_WARNING("DataIn with ITT %u, DataSN %u, "
				      "offset %u, has no data\n", local_itt,
				      data_in_header->data_sn,
				      data_in_header->offset);
		} else {
			received = recv_data_in_data(current_connection,
						     related_command,
						     data_in_header,
						     &size,
						     over_flow_data_size,
						     &over_flow_buffer);

			if (unlikely(received <= 0))
				goto out;

			/* data in the DataIn was received correctly,
			 * update amount of data transfered so far in this seq
			 */
			related_command->data_offset += size;
			related_command->recvd_length += size;

#ifdef ISCSI_CHECK_PDU_FORMAT
			/* Check that we have not received more data
			 * than allowed in one input sequence
			 */
			if (unlikely(related_command->recvd_length
				> related_session->oper_param->MaxBurstLength)){
				TRACE_ERROR(
					"DataIn with ITT %u gives sequence with"
					" %u bytes, more than %u allowed by "
					"%s\n", local_itt,
					related_command->recvd_length,
					related_session->oper_param
						->MaxBurstLength,
					"MaxBurstLength");
			}
#endif

			/* update amount of data transfered so far in this cmd*/
			related_command->r2t_range_list.limit += size;
			related_command->cmd_error_size += size;

			/* if this is the last DataIn PDU expected for the
			 * corresponding SCSI Command, check that F_BIT is set
			 */
			if (related_command->r2t_range_list.limit
				>= related_command->SCpnt->request_bufflen) {
				/* last rexmitted DataIn PDU may not have
				 * F bit set
				 */
				if (unlikely(!(data_in_header->flags & F_BIT)
					&& related_command->cmd_error_state
							!= CMD_USING_SNACK)) {
					TRACE_ERROR(
				 		"F_BIT not set on last expected"
						" DataIn with ITT %u\n",
						local_itt);
					/* turn it on now */
					data_in_header->flags |= F_BIT;
				}
			}
		}

		if ((data_in_header->flags & F_BIT)) {
			/* this is last DataIn PDU in a sequence, set up for
			 * any new sequence that may follow this one
			 */
			if (unlikely(!related_session->oper_param
							->DataPDUInOrder)) {
				/* DataPDUInOrder==No, pdus can be
				 * out-of-order within a sequence
				 */
				check_range_list_complete(&related_command->
								pdu_range_list);
				free_range_list(&related_command->
								pdu_range_list);
				/* set up offset to first pdu in this sequence*/
				related_command->r2t_range_list.offset
				    = related_command->pdu_range_list.offset;
			}
			if (unlikely(!related_session->oper_param
							->DataSequenceInOrder)){
				/* DataSequenceInOrder==No, sequences can
				 * be out-of-order
				 */
				merge_offset_length(&related_command->
							r2t_range_list,
						    related_command->
							r2t_range_list.offset,
						    related_command->
							recvd_length);
			} else if (unlikely(related_command
					->r2t_range_list.offset
						+ related_command->recvd_length
				    != related_command->r2t_range_list.limit)) {
				/* in-order sequence not in-order */
				if (related_command->cmd_error_state
							== CMD_OK_SO_FAR) {
					/* avoid cascading error messages
					 * during error recovery
					 */
					TRACE_ERROR(
					    "In-order sequence started at "
					    "offset %u, expected to start at "
					    "%u\n",
					     related_command->r2t_range_list.
						offset,
					     related_command->r2t_range_list.
					       limit
					       - related_command->recvd_length);
				}
			}

			/* mark that next DataIn PDU (if any) starts
			 * a new sequence
			 */
			related_command->recvd_length = 0;
		}

		if (unlikely(related_command->cmd_error_state
							== CMD_USING_SNACK)) {
			/* this DataIn is a reply to previous SNACK from us */
			related_command->cmd_error_runlen -= 1;
			if (related_command->cmd_error_runlen == 0) {
				/* this is last DataIn from previous SNACK */
				related_command->cmd_error_state=CMD_OK_SO_FAR;

				/* now we can acknowledge the stat_sn if a
				 * DataIn S=1 or SCSI Response originally
				 * caused us to send the SNACK
				 */
				if (current_connection->increment) {
					if (!(data_in_header->flags & S_BIT)) {
					      current_connection->exp_stat_sn++;
					      data_in_header->flags |= S_BIT;
					}
				} else {
					/* make it look like this pdu was the
					 * one that had the original A bit set
					 * by restoring fields saved from the
					 * DataIn pdu that started the rexmit
					 */
					data_in_header->flags |= A_BIT;
					data_in_header->target_xfer_tag =
						related_command->abit_ttt;
					data_in_header->lun =
						related_command->abit_lun;
					related_command->data_in_sn =
						related_command->abit_in_sn;
					related_command->data_offset =
						related_command->abit_offset;
				}
			}
		}

		if (unlikely((data_in_header->flags & A_BIT)
			&& related_session->oper_param->ErrorRecoveryLevel>0)) {
			if (related_command->cmd_error_state == CMD_OK_SO_FAR) {
				/* send target the DataACK SNACK he requested */
				err_rec.curr_conn = current_connection;
				err_rec.related_cmd = related_command;
				err_rec.err_type = -1;
				received = init_send_snack(&err_rec,
						related_command->data_in_sn,
						0, DATACK_SNACK);
				if (received <= 0)
					goto out;
			} else if (related_command->cmd_error_state
							== CMD_NEEDS_SNACK) {
				/* time to send a snack to get erroneous
				 * DataIn pdus retransmitted
				 */
				received=send_data_ack_snack(current_connection,
							    related_command, 0);
				if (received <= 0)
					goto out;
			} else {/* should never happen */
				TRACE_ERROR("got DataIN with A_BIT set but "
					    "cmd_error_state %d\n",
					    related_command->cmd_error_state);
			}
		}

		if (data_in_header->flags & S_BIT) {
			/* this command is a "phase collapse" with "good"
			 * status or is the last DataIn pdu retransmitted
			 * as part of a recovery triggered by S_BIT or
			 * scsi response
			 */
			TRACE(TRACE_DEBUG, "Phase collapse, ITT %u, "
			      "recvd_length %u, data_offset %u, "
			      "request_bufflen %u, recv_so_far %u, "
			      "data_in_sn %u\n", local_itt, 
			      related_command->recvd_length,
			      related_command->data_offset,
			      related_command->SCpnt->request_bufflen,
			      related_command->r2t_range_list.limit,
			      related_command->data_in_sn);
			received = do_scsi_response(current_connection,
						related_command,
						(struct iscsi_targ_scsi_rsp *)
						data_in_header, 0, NULL);
		}
	}

out:
	if (unlikely(over_flow_buffer != NULL))
		my_kfree((void **) &over_flow_buffer, "local buf");

	TRACE(TRACE_ENTER_LEAVE, "Leave rx_data, retval %d\n", received);
	return received;
}

static int __attribute__ ((no_instrument_function))
rx_pack_cookie_from_pdu( struct session *related_session,
			struct connection *current_connection,
	        	struct command *related_command ,
			struct iscsi_targ_r2t   *r2t_header,
        		__u32 local_itt,
			__u32 offset,
			__u32  xfer_len);

/****************************************************************************
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_rx_thread().
 * This function processes the R2T PDU and sets up the TCP buffers for the
 * first iSCSI DataOut PDU to be sent out to the target. It wakes up the
 * tx thread so that it can send DataOut PDU to the target.
 * Parameters: current_connection which is a ptr to the connection struct
 *			  where the R2T header was received.
 * Return Value: >0 success
 *		 =0 on failure, session lock is locked
 *		 <0 on failure, session lock is NOT locked
 ****************************************************************************/
static int
rx_r2t(struct connection *current_connection,
       __u32 local_itt, struct command *related_command)
{
	struct iscsi_targ_r2t *r2t_header;
	struct session *related_session;
	struct init_error_rec err_rec;
	int delta, delta2, retval = 1;
	__u32 r2t_sn, offset, xfer_len;
	struct r2t_cookie *cookie;

	TRACE(TRACE_ENTER_LEAVE, "Enter rx_r2t, current_connection %p\n",
	      current_connection);

	r2t_header = (struct iscsi_targ_r2t *) current_connection->rx_buf;
	r2t_sn = ntohl(r2t_header->r2t_sn);
	offset = ntohl(r2t_header->offset);
	xfer_len = ntohl(r2t_header->xfer_len);

	if (unlikely(TRACE_TEST(TRACE_DEBUG))) {
		debug_print_ip(current_connection, "R2T");
	}

	related_session = current_connection->my_session;

#ifdef ISCSI_CHECK_PDU_FORMAT
	/* Check if reserved fields are zero or not */
	if (unlikely(!((r2t_header->flags & ~F_BIT) == 0
						&& r2t_header->rsvd2 == 0))) {
		reserved_not_0("R2T");
	}
#endif

	if (likely(related_command != NULL)) {
		/* target is responding to this command, so it does not
		 * need to be sent again on a retransmission
		 */
		related_command->cmd_state &= ~CMD_STATE_REXMIT_ORIG;

		/* is this R2T the correct sequence of R2Ts for this command? */

		/* Draft 20, Section 10.8.2 R2TSN
		 * "R2TSN is the R2T PDU input PDU number within the command
		 * identified by the Initiator Task Tag."
		 */

		/* new - SAI */
		delta = related_command->r2t_sn - r2t_sn;
		delta2 = related_command->data_offset - offset;
		if (unlikely(delta != 0)) {
			if (related_session->oper_param->ErrorRecoveryLevel > 0
				&& delta == 1 && delta2 > 0) {
				/* only one outstanding R2T supported
				 * for error recovery levels > 0.
				 * If an R2T seq. repeats due to
				 * inherent timeout from Target
				 * either due to lost or delayed
				 * DataOuts, remove the old R2T
				 * cookie and recreate a new one.
				 * Remember to re-transmit all
				 * DataOuts requested by the R2T.
				 * - new 18_04 - SAI
				 */
				TRACE_WARNING("Recovery R2T for ITT %u has "
					      "R2TSN %u, expected %u\n",
					      local_itt, r2t_sn,
					      related_command->r2t_sn);
				if (!list_empty(&related_command->r2t_cookies)){
					/* we did not send the cookie's
					 * remaining xfer length
					 */
					cookie = list_entry(related_command
							    ->r2t_cookies.next,
							    struct r2t_cookie,
							    link);
					related_command->data_offset
						-= cookie-> r2t_xfer_length;
					free_r2t_cookie(current_connection,
							related_command,
							cookie);
				}
				related_command->r2t_sn--;
				TRACE(TRACE_ERROR_RECOVERY,
				      "Got Recovery R2T, ITT %u, "
				      "R2TSN %u, offset %u, DDTL %u\n",
				      local_itt, r2t_sn, offset, xfer_len);
			} else {
				TRACE_ERROR
				    ("R2T for ITT %u has R2TSN %u, expected "
				     "%u\n",
				     local_itt, r2t_sn,
				     related_command->r2t_sn);
				TRACE(TRACE_ERROR_RECOVERY,
				      "expected offset %u, tot len %u\n",
				      related_command->data_offset,
				      related_command->SCpnt->request_bufflen);
				if (unlikely(TRACE_TEST(TRACE_ERROR_RECOVERY)))
					print_targ_r2t(r2t_header);

				/* R2TSN Sequence Error Recovery - SAI */
				err_rec.curr_conn = current_connection;
				err_rec.related_cmd = related_command;
				err_rec.err_type = SEQUENCE_ERR;
				retval = init_do_error_recovery(&err_rec);
				goto rx_r2t_out;
			}
		}

		related_command->r2t_sn++;

		/* Draft 20, Section 10.7.4 Target Transfer Tag and LUN
		 * "The Target Transfer Tag values are not specified by
		 * this protocol except that the value 0xffffffff is
		 * reserved and means that the Target Transfer Tag is
		 * not supplied."
		 */
		if (unlikely(r2t_header->target_xfer_tag == ALL_ONES)) {
			TRACE_ERROR("R2T has reserved TTT value 0x%08x\n",
				    r2t_header->target_xfer_tag);
		}

		if (unlikely(related_command->SCpnt == NULL)) {
			TRACE_ERROR("No SCSI command found for ITT %u\n",
				    local_itt);
		} else if (unlikely(rx_pack_cookie_from_pdu(related_session,
							   current_connection,
							   related_command,
							   r2t_header,
							   local_itt,offset,
							   xfer_len) < 0)) {
			TRACE_ERROR("Pack Cookie failed\n");
		}
	} else {
		TRACE_ERROR("R2T for ITT %u, R2TSN %u, offset %u, DDTL %u, "
			    "no command!\n", local_itt, r2t_sn, offset,
			    xfer_len);
	}

	/* r2t exit added for Error Recovery - SAI */
rx_r2t_out:
	TRACE(TRACE_ENTER_LEAVE,
		"Leave rx_r2t, current_connection %p, retval %d\n",
		current_connection, retval);
	return retval;
}

/*
 * Called only from rx_r2t().
 * Populate a new cookie from an R2T PDU. Cookies are attached to the
 * command struct. Moved from rx_r2t to condense inline code
 * It extracts useful data from the R2T PDU and fills a new cookie.
 * Return 1 on success.
 * 	  0 on failure from create_r2t_cookie()
 */
static int __attribute__ ((no_instrument_function))
rx_pack_cookie_from_pdu( struct session *related_session,
			struct connection *current_connection,
	        	struct command *related_command ,
			struct iscsi_targ_r2t   *r2t_header,
        		__u32 local_itt,
			__u32 offset,
			__u32  xfer_len)
{
	struct r2t_cookie *cookie;
	int over_flow_data_size, r2t_sn, delta;

	TRACE(TRACE_ENTER_LEAVE, "Enter rx_pack_cookie\n");

	r2t_sn = ntohl(r2t_header->r2t_sn);
	cookie = create_r2t_cookie(related_session);
	if (unlikely(cookie == NULL)) {
		TRACE_ERROR("Cookie for ITT %u, R2TSN %u FAILED\n",
			                    local_itt, r2t_sn);
		return 0;
	}


	/* copy info from R2T into fields in the cookie */
	cookie->offset = offset;
	cookie->r2t_xfer_length = xfer_len;

	if (likely(related_session->oper_param->DataSequenceInOrder)) {
		/* DataSequenceInOrder=Yes */

		/* Draft 20, Section 10.8 Ready to Transfer (R2T)
		 * "DataSequenceInOrder governs the buffer offset ordering
		 * in consecutive R2Ts. If DataSequenceInOrder is Yes,
		 * then consecutive R2Ts MUST refer to continuous
		 * non-overlapping ranges except for Recovery-R2Ts."
		 */

		/* Consider an out of seq R2T to be a recovery R2T - SAI */
		delta = related_command->data_offset - offset;
		if (unlikely(delta != 0)) {
			/* Handle Recovery R2T without flagging errors - SAI */
			if (likely(delta > 0)) {
				/* recovery R2T, just backup command offset */
				TRACE_WARNING("Recovery R2T for ITT %u, "
				      "R2TSN %u, reset offset from %u "
				      "to %u, still_to_send from %u to %u\n",
				      local_itt, r2t_sn,
				      related_command->data_offset, offset,
				      related_command->still_to_send,
				      related_command->still_to_send + delta);
				related_command->still_to_send += delta;
				related_command->data_offset = offset;
			} else {
				TRACE_ERROR("R2T for ITT %u, R2TSN %u, has "
					    "buffer offset %u, expected %u\n",
					    local_itt, r2t_sn, offset,
					    related_command->data_offset);
				/* put back this unneeded cookie structure */
				uncreate_r2t_cookie(cookie, related_session);
				return -1;
			}
		}
	}

#ifdef ISCSI_CHECK_PDU_FORMAT
	/* Check that R2T not asking to xfer more than MaxBurstLength */
	if (unlikely(xfer_len > related_session->oper_param->MaxBurstLength)) {
		TRACE_ERROR ("R2T for ITT %u, R2TSN %u, has DDTL of %u bytes, "
			     "exceeding allowed %s %u\n", local_itt, r2t_sn,
			     xfer_len, "MaxBurstLength",
			     related_session->oper_param->MaxBurstLength);

		/* for now, just ignore this error -- we can
		 * handle data beyond MaxBurstLength without problem
		 */
	}
#endif

	/* Check if this R2T would overflow the SCSI Mid-level buffers */
	over_flow_data_size = offset + xfer_len
				- related_command->SCpnt->request_bufflen;

	if (unlikely(over_flow_data_size > 0)) {
		TRACE_ERROR("R2T for ITT %u, R2TSN %u, has DDTL of %u bytes "
			    "that overflows SCSI buffer by %d\n",
			    local_itt, r2t_sn, xfer_len, over_flow_data_size);

		/* accept only the remaining amount of buffer space */
		if (over_flow_data_size >= xfer_len) {
			/* overflow exceeds amount to be transfered */
			/* put back this unneeded cookie structure */
			uncreate_r2t_cookie(cookie, related_session);
			return -1;
		}
		cookie->r2t_xfer_length -= over_flow_data_size;
	}

	TRACE(TRACE_DEBUG, "recvd r2t for ITT %u, R2TSN %u, offset %u, "
	      "xfer length %u\n", local_itt, r2t_sn,
	      related_command->data_offset, cookie->r2t_xfer_length);

	hookup_r2t_cookie(cookie, related_session, current_connection,
			  related_command, r2t_header->target_xfer_tag);

	if (atomic_read(&current_connection->tx_sem.count) <= 0) {
		/* tell xmit thread something may be ready to send */
		TRACE(TRACE_ISCSI_FULL, "rx_r2t: wake up tx thread\n");
		up(&current_connection->tx_sem);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave rx_pack_cookie\n");
	return 1;
}

/****************************************************************************
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_rx_thread().
 * This function processes the Asynchronous Message PDU
 * Parameters: current_connection which is a ptr to the connection struct
 *			  where the PDU header was received.
 * Return Value: >0 success, =0 failure.
 ****************************************************************************/
static int
rx_async_msg(struct connection *current_connection, __u32 size, char *buffer)
{
	struct iscsi_targ_async_msg *am_header;
	struct session *related_session;
	struct connection *conn;
	__u32 lun;
	int retval = 0;		/* assume this fails */
	__u16 param1, param2, param3;

	TRACE(TRACE_ENTER_LEAVE, "Enter rx_async_msg, current_connection %p\n",
	      current_connection);

	am_header = (struct iscsi_targ_async_msg *) current_connection->rx_buf;

	if (unlikely(TRACE_TEST(TRACE_DEBUG))) {
		debug_print_ip(current_connection, "Async Msg");
	}

	related_session = current_connection->my_session;

#ifdef ISCSI_CHECK_PDU_FORMAT
	/* Check if reserved fields are zero or not */
	if (unlikely(!(((am_header->flags & ~F_BIT) == 0)
		&& (am_header->rsvd2 == 0)
		&& (am_header->rsvd3 == 0)
		&& (am_header->rsvd5 == 0)))) {
		reserved_not_0("Async Msg");
	}

	/* Draft 20, section 10.9.2 AsyncVCode
	 * "AsyncVcode is a vendor specific detail code that is only
	 * valid if the AsyncEvent field indicates a vendor specific
	 * event.  Otherwise, it is reserved."
	 */
	if (am_header->async_vcode != 0 && am_header->async_event != 255) {
		TRACE_ERROR
		    ("AsyncVCode 0x%02x should be 0 when AsyncEvent is "
		     "0x%02x\n",
		     am_header->async_vcode, am_header->async_event);
	}

	/* Draft 20, section 10.9.3 LUN "The LUN field MUST be valid
	 * if AsyncEvent is 0.  Otherwise, this field is reserved."
	 */
	if (am_header->lun != 0 && am_header->async_event != 0) {
		TRACE_ERROR
		    ("LUN 0x%016llx should be 0 when AsyncEvent is 0x%02x\n",
		     am_header->lun, am_header->async_event);
	}
#endif

	param1 = be16_to_cpu(am_header->parameter1);
	param2 = be16_to_cpu(am_header->parameter2);
	param3 = be16_to_cpu(am_header->parameter3);

	switch (am_header->async_event) {

	case 0:
		/* SCSI Asynchronous Event */
		/* check for valid LUN, per section 9.9.3 quoted above */
		lun = unpack_lun((__u8 *) & am_header->lun);
		if (!(related_session->lun_bits & (1 << lun))) {
			TRACE_ERROR("Invalid SCSI Asynch Event LUN %u\n", lun);
		}

		TRACE_ERROR("Unimplemented SCSI Asynchronous Event\n");
		break;

	case 1:
		/* target requests Logout */
#ifdef ISCSI_CHECK_PDU_FORMAT
		if (param1 != 0 || param2 != 0) {
			TRACE_ERROR
			("Parameter1 0x%04x and Parameter2 0x%04x both should "
			"be 0 when AsyncEvent is 0x%02x\n",
			param1, param2, am_header->async_event);
		}
#endif

		if (current_connection->connection_flags
			& SEND_NO_REPLY_TO_ASYNC_LOGOUT) {
			TRACE(TRACE_ISCSI, "Not replying to Async Logout "
						"request from target\n");
			retval = 1;	/* success */
		} else if (!drive_logout(related_session, current_connection,
				 	 LOGOUT_CLOSE_SESSION)) {
			TRACE_ERROR
			("Unable to send Logout PDU requested by target\n");
		} else	{
			retval = 1;	/* success */
		}
		break;

	case 2:
		/* target dropping a specified connection */
		/* NOTE we cannot just kill the connection,
		 * since that might require this thread to kill
		 * itself and then wait for itself to die,
		 * which will not work.
		 */
		/* check that this CID actually matches an existing
		 * connection id
		 */
		for (conn = related_session->connection_head;
					conn != NULL; conn = conn->next) {
			if (conn->connection_id == param1) {
				break;
			}
		}
		if (conn == NULL) {
			TRACE_ERROR
			("Target dropping connection with nonexistant CID %u\n",
			param1);
		} else {
			TRACE(TRACE_ISCSI,
			"Target dropping connection with CID %u\n", param1);
		}
		retval = 1;	/* success */
		break;

	case 3:
		/* target dropping all connections in a session */
		/* NOTE we cannot just kill the session, since
		 * that might require this thread to kill itself
		 * and then wait for itself to die, which will not work.
		 */
#ifdef ISCSI_CHECK_PDU_FORMAT
		if (param1 != 0) {
			TRACE_ERROR("Parameter1 0x%04x should be 0 "
				"when AsyncEvent is 0x%02x\n",
				param1, am_header->async_event);
		}
#endif

		TRACE(TRACE_ISCSI,
		      "Target dropping all connections in session\n");
		retval = 1;	/* success */
		break;

	case 4:
		/* target requests Text negotiation */
		if (unlikely(drive_text_negotiate(current_connection,
						related_session, 0, 0))) {
			TRACE_ERROR
			("Unable to send Text PDU requested by target\n");
		} else {
			retval = 1;	/* success */
		}
		break;

	case 255:
		/* vendor specific iSCSI Event */
		TRACE(TRACE_ISCSI,
			"Ignoring vendor specific iSCSI Event, "
			"AsyncVcode 0x%02x\n", am_header->async_vcode);
		retval = 1;	/* success */
		break;

	default:
		TRACE_ERROR("Illegal AsyncEvent 0x%02x\n",
			am_header->async_event);
		retval = 1;	/* success */
		break;
	}			/* switch */

	TRACE(TRACE_ENTER_LEAVE, "Leave rx_async_msg, retval %d\n", retval);
	return retval;
}

/****************************************************************************
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_rx_thread().
 * This function processes the task mgt response header, sets up the
 * task_mgt_response field in the command struct and wakes up the
 * initiator_abort function (called by SCSI Midlevel) waiting on the
 * task_mgt_sem semaphore.
 * Parameters: current_connection which is a ptr to the connection struct
 *			  where the Task Mgt. Header was received.
 * Return Value: >=0, on success (always)
 ****************************************************************************/
static int
rx_task_mgt_rsp(struct connection *current_connection,
		__u32 local_itt, struct command *related_command)
{
	struct session *current_session;
	struct iscsi_targ_task_mgt_response *header;
	struct command *cmd;

	TRACE(TRACE_ENTER_LEAVE, "Enter rx_task_mgt_rsp, current_connection"
	      " %p\n", current_connection);

	header = (struct iscsi_targ_task_mgt_response *)
						current_connection->rx_buf;

	TRACE(TRACE_DEBUG, "Got StatSN %u\n", ntohl(header->stat_sn));

	/* Store the session pointer for future use - SAI */
	current_session = current_connection->my_session;

#ifdef ISCSI_CHECK_PDU_FORMAT
	/* Check if reserved fields are zero or not */
	if (unlikely(!(((header->flags & ~F_BIT) == 0)
	      && (header->rsvd1 == 0)
	      && (header->rsvd2 == 0)
	      && (header->rsvd4 == 0)
	      && (header->rsvd5 == 0)))) {
		reserved_not_0("TM Response");
	}
#endif

	/* leave if command struct corresponding to task mgt rsp not found */
	if (unlikely(related_command == NULL)) {
		goto out;
	}

	if (likely((header->response == FUNCTION_COMPLETE)
	    || (header->response == TASK_DOES_NOT_EXIST)
	    || (header->response == LUN_DOES_NOT_EXIST)
	    || (header->response == FUNCTION_REJECTED))) {
		related_command->task_mgt_response = header->response;
		TRACE(TRACE_DEBUG, "Got task mgt response 0x%02x\n",
				    header->response);
	} else {
		TRACE_ERROR("Invalid task mgt response from target 0x%02x,"
			    " returning reponse FUNCTION_REJECTED to SCSI\n",
			    header->response);
		related_command->task_mgt_response = FUNCTION_REJECTED;
	}

	/* Added the TASK REASSIGN handling for connection recovery - SAI */
	if (related_command->task_mgt_function == TMF_TASK_REASSIGN) {
		cmd = find_command_by_itt(related_command->ref_task_tag,
					  current_connection);

		if (likely(cmd != NULL)) {
			/* Free the corresponding task mgt command */
			if (related_command->task_mgt_response
							== FUNCTION_REJECTED) {
				/* TASK REASSIGN probably not supported.
				 * Free pending command
				 */
				TRACE_ERROR
				    ("Task Reassignment for task ITT %u, "
				     "returned response %d\n",
				     related_command->ref_task_tag,
				     header->response);
				TRACE(TRACE_DEBUG,
				      "Freeing related command, ITT %u\n",
				      related_command->ref_task_tag);

				if (cmd->SCpnt != NULL) {
					/* pass the final status back up
					 * to SCSI Mid-level
					 */
					cmd->SCpnt->result = DID_ERROR << 16;
				}
				cmd->cmd_state |= CMD_STATE_RXDONE;
				free_pending_command(cmd, current_connection);
			}
		} else {
			/* could not find the corresponding command --
			 * reject task mgt. command
			 */
			TRACE_ERROR("No task with ITT %u found\n",
				    related_command->ref_task_tag);
		}

		TRACE(TRACE_DEBUG, "Freeing Task Mgt. request, ITT %u\n",
		      related_command->init_task_tag);
		related_command->cmd_state |= CMD_STATE_RXDONE;
		free_pending_command(related_command, current_connection);

		current_session->n_rec_tasks--;
	}

	/* wake up the process waiting on task_mgt_sem semaphore */
	up(&current_connection->task_mgt_sem);

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave rx_task_mgt_rsp ,current_connection"
	      " %p\n", current_connection);
	return 1;		/* nothing but success here */
}

/*RFC 3720 Section 10.17.1 Reason (field in Reject Response) */
static char *reject_reasons[] = {
	/* 0x02 */	"Data Digest Error",
	/* 0x03 */	"SNACK Reject",
	/* 0x04 */	"Protocol Error",
	/* 0x05 */	"Command not supported",
	/* 0x06 */	"Too many immediate commands",
	/* 0x07 */	"Task in progress",
	/* 0x08 */	"Invalid Data ACK",
	/* 0x09 */	"Invalid PDU field",
	/* 0x0a */	"Long Operation Reject",
	/* 0x0b */	"Negotiation Reset",
	/* 0x0c */	"Waiting for Logout"
	};
/***************************************************************************
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_rx_thread().
 * This function processes the Reject header.
 * Parameters: current_connection which is a ptr to the connection struct
 *			  where the Reject header was received.
 * Return Value: 0 to terminate the rx thread gracefully, 1 to keep going
 ***************************************************************************/
static int
rx_rjt(struct connection *conn, __u32 size, char *buffer)
{
	int retval = 1, found;
	__u32 local_itt;
	struct iscsi_targ_rjt *rjt_header;
	struct generic_pdu *bad_header;
	struct iscsi_init_snack *header;
	struct iscsi_init_scsi_cmnd *iscsi_cmd;
	struct command *related_command;
	struct session *sess;
	struct scsi_sense_data *senseinfo;
	char local_message[32], *local_ptr;

	TRACE(TRACE_ENTER_LEAVE, "Enter rx_rjt, conn %p\n",
	      conn);

	sess = conn->my_session;
	rjt_header = (struct iscsi_targ_rjt *)conn->rx_buf;
	bad_header = (struct generic_pdu *)buffer;

	if (unlikely(rjt_header->reason < REASON_DATA_DIGEST_ERR
	    || rjt_header->reason > REASON_WAITING_FOR_LOGOUT)) {
		sprintf(local_message, "out of range [0x%02x..0x%02x]",
			REASON_DATA_DIGEST_ERR, REASON_WAITING_FOR_LOGOUT);
		retval = 0;	/* fatal */
		local_ptr = local_message;
	} else {
		local_ptr = reject_reasons[rjt_header->reason
						- REASON_DATA_DIGEST_ERR];
		if (rjt_header->reason >= REASON_PROTOCOL_ERR)
			retval = 0;	/* fatal -- can't handle these yet */
	}

	/* Reject does NOT have an ITT value (i.e., it is 0xffffffff) */
	if (likely(bad_header != NULL && size >= ISCSI_HDR_LEN)) {
		local_itt = htonl(bad_header->init_task_tag);
		related_command = find_command_by_itt(local_itt, conn);
		TRACE_ERROR("Got Reject for ITT %u, reason 0x%02x: %s\n",
			    local_itt, rjt_header->reason, local_ptr);
		printk("header of rejected pdu:\n");
		dump_buffer((__u8 *)buffer, size);
	} else {
		local_itt = ALL_ONES;
		related_command = NULL;
		TRACE_ERROR("Got Reject with bad data, reason 0x%02x: %s\n",
			    rjt_header->reason, local_ptr);
	}

#ifdef ISCSI_CHECK_PDU_FORMAT
	/* Check that reserved fields are zero */
	if (unlikely(!((rjt_header->flags & ~F_BIT) == 0
	      && rjt_header->rsvd2 == 0
	      && rjt_header->rsvd4 == 0 && rjt_header->rsvd5 == 0))) {
		reserved_not_0("Reject");
	}
#endif

	if (retval == 0) {
		/* look for reject of R2T SNACK previously sent on a write */
		if (related_command != NULL && related_command->SCpnt != NULL
			&& (bad_header->opcode & ISCSI_OPCODE)==ISCSI_INIT_SNACK
			&& related_command->activity_flg == 2
			&& sess->oper_param->ErrorRecoveryLevel > 0
			&& sess->retran_period > 0
			&& related_command->SCpnt->sc_data_direction
					== DMA_TO_DEVICE) {
			/* found it, don't let this Reject advance the StatSN,
			 * then wake up tx thread to send Status SNACK */

			conn->exp_stat_sn = ntohl(bad_header->exp_stat_sn);
			related_command->time_stamp -= sess->retran_period * HZ;
			if (atomic_read(&conn->tx_sem.count) <= 0) {
				/* get xmit thread going */
				TRACE(TRACE_ISCSI_FULL,
					"rx_r2t: wake up tx thread\n");
				up(&conn->tx_sem);
			}
			retval = 1;	/* not fatal */
		}
		goto rjt_out;
	}

	/* ignore target digest errors and wait for target to send out any
	 * recovery R2Ts new 18_04 - SAI
	 */
	if (rjt_header->reason == REASON_DATA_DIGEST_ERR) {
		if ((iscsi_cmd=(struct iscsi_init_scsi_cmnd *)buffer) == NULL) {
			retval = 0;	/* fatal */
			goto rjt_out;
		}
		if (sess->retran_period > 0
				&& (iscsi_cmd->opcode & ISCSI_OPCODE)
						== ISCSI_INIT_SCSI_CMND) {
			/* pdu was a SCSI Command, so there must have been data
			 * attached (to generate digest error), so it must have
			 * been a write. Therefore, if this is the first such
			 * reject, set timestamp so it will be reissued at once
			 */
			if (related_command) {
				related_command->time_stamp -=
						sess->retran_period * HZ;
				TRACE_ERROR("Reject due to Data Digest Error, "
					 "retransmit command\n");
				if (atomic_read(&conn->tx_sem.count) <= 0) {
					/* get xmit thread going */
					TRACE(TRACE_ISCSI_FULL,
						"rx_r2t: wake up tx thread\n");
					up(&conn->tx_sem);
				}
				goto rjt_out;
			}
		}
		TRACE_ERROR("Reject due to Data Digest Error, do nothing\n");
		goto rjt_out;
	}

	/* for all other type of rejects, abort the related command - SAI */
	if (unlikely(bad_header == NULL)) {
		TRACE_ERROR("Reject data buffer empty, ignore\n");
		retval = 0;	/* fatal */
		goto rjt_out;
	}

	/* in case the rejected command was a SNACK */
	header = (struct iscsi_init_snack *)buffer;

	if (unlikely(size < ISCSI_HDR_LEN)) {
		TRACE_ERROR
		    ("Reject DSL = %u less than header length %u, ignore\n",
		     size, ISCSI_HDR_LEN);
		retval = 0;	/* fatal */
		goto rjt_out;
	}

	if (rjt_header->reason == REASON_DATA_SNACK) {

		/* If the rejected SNACK is any other than Data/R2T
		 * or Status, it's an error - SAI
		 */
		if (((header->flags & SNACK_TYPE) != DATA_R2T_SNACK)
		    && ((header->flags & SNACK_TYPE) != STATUS_SNACK)) {
			TRACE_ERROR
			    ("Expected Data/R2T or Status header in Reject PDU,"
			     " got type 0x%02x", (header->flags & SNACK_TYPE));
			retval = 0;	/* fatal */
			goto rjt_out;
		}
	}

	/* if this ITT is not reserved, we should find a matching command */
	found = 0;
	list_for_each_entry(related_command,
			    &conn->pending_commands,
			    link) {
		/* Status SNACK always has the ITT as ALL_ONES.
		 * So for status snack, do not check for ITT --
		 * new 18_04 SAI
		 */
		if (rjt_header->reason == REASON_DATA_SNACK
		    && (header->flags & SNACK_TYPE) == STATUS_SNACK) {
			if (related_command->snack_flags & STATUS_BIT) {
				/* Do not terminate on SNACK reject - SAI */
				TRACE(TRACE_DEBUG,
				      "Found corresponding Status SNACK request"
				      " - ITT %u\n",
				      related_command->init_task_tag);
				found = 1;
				break;
			}
		} else if (related_command->init_task_tag == local_itt) {
			found = 1;
			break;
		}
	}

	if (unlikely(found == 0)) {
		TRACE_ERROR("No command has matching ITT %u\n", local_itt);
		/* silently discard any snack rejects that appear delayed-SAI */
		goto rjt_out;
	}

	/* all rejects should terminate the related command - SAI */
	if (unlikely(related_command->SCpnt == NULL)) {
		if (related_command->header_opcode == ISCSI_INIT_SCSI_CMND) {
			/* No pointer to the SCSI Mid-level struct, bail out */
			TRACE_ERROR("ITT %u command struct has SCpnt=NULL\n",
				    local_itt);
			retval = 0;	/* fatal */
		}
		if (sess->oper_param->SessionType) {
			/* this is a discovery session, send a
			 * logout to shut it down
			 */
			drive_logout(sess, conn, LOGOUT_CLOSE_SESSION);
		}
	} else {
		/* We need to increment the exp_stat_sn as we will
		 * not receive the missing status anymore
		 */
		conn->exp_stat_sn++;

		related_command->SCpnt->result |= (DID_ERROR << 16);
		senseinfo = (struct scsi_sense_data *)
		    		related_command->SCpnt->sense_buffer;
		memset(senseinfo, 0, sizeof(struct scsi_sense_data));

		/* Draft 20 Section 10.4.7.2 Sense Data
		 * Shows table of sense values
		 */
		/* Valid Bit set, Response = 0x70 */
		senseinfo->response = 0xf0;
		senseinfo->sense_key = ABORTED_COMMAND;	/* Protocol CRC Error */
		senseinfo->asl = 10;	/* n-7 = 17-7 = 10 */
		senseinfo->asc = 0x47;	/* Protocol CRC Error
					   Additional Sense Code */
		senseinfo->ascq = 0x05;	/* Protocol CRC Error
					   Additional Sense Code qualifier */
		related_command->SCpnt->result |= (DRIVER_SENSE << 24);
	}

	/* always free up the related command */
	related_command->cmd_state |= CMD_STATE_RXDONE;
	free_pending_command(related_command, conn);

rjt_out:
	TRACE(TRACE_ENTER_LEAVE, "Leave rx_rjt, retval %d\n", retval);
	return retval;
}

/***************************************************************************
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_rx_thread().
 * This function processes the NopIn header.
 * 4 combinations of ITT and TTT to look for (resvd value is 0xffffffff):
 *TTT		ITT	StatSN	Respond
 *resvd		resvd	?	simple 	update to command window
 *resvd		!=resvd +1	no	response to NopOut sent by us
 *!=resvd	resvd 	+0	yes	provoke "ping" response from us
 *!=resvd	!=resvd	NA	NA	illegal
 *
 * Parameters: current_connection which is a ptr to the connection struct
 *			  where the NopIn header was received.
 * Return Value: >0 success (always)
 ***************************************************************************/
static int
rx_nopin(struct connection *current_connection, __u32 local_itt,
	 struct command *related_command, __u32 size, char **buffer)
{
	struct iscsi_targ_nopin *header;
	struct session *related_session;
	struct iscsi_init_nopout *nopout_cmd;
	__u32 lun;

	TRACE(TRACE_ENTER_LEAVE, "Enter rx_nopin, current_connection %p\n",
	      current_connection);

	header = (struct iscsi_targ_nopin *) current_connection->rx_buf;

#ifdef ISCSI_CHECK_PDU_FORMAT
	/* Check that reserved fields are zero */
	if (unlikely(!((header->flags & ~F_BIT) == 0
	      && header->rsvd1 == 0
	      && header->rsvd2 == 0 && header->rsvd3 == 0))) {
		reserved_not_0("NopIn");
	}
#endif

	related_session = current_connection->my_session;

	if (unlikely(size > current_connection->max_send_length)) {
		/* don't send back more than target wants to get */
		size = current_connection->max_send_length;
	}

	if (header->target_xfer_tag == ALL_ONES) {
		/* sent by target either as a response to a NopOut
		 * sent earlier by us or as a means to update the
		 * command window.  In both cases, no response is
		 * needed from the initiator.
		 */
		if (local_itt == ALL_ONES) {
			/* this is not a response to a NopOut sent
			 * earlier by us, but is just a simple update
			 * of the command window and this has already
			 * been done, so we are finished.
			 */
		} else {
			/* this NopIn is a response to a NopOut sent
			 * earlier by us
			 */
			if (related_command != NULL) {
				/* found the corresponding NopOut command,
					check it */
				nopout_cmd = (struct iscsi_init_nopout *)
				    &related_command->iscsi_cmd;
				if ((nopout_cmd->opcode & ISCSI_OPCODE)
							!= ISCSI_INIT_NOP_OUT) {
					TRACE_ERROR
					    ("Command matching NopIn has opcode"
					     " 0x%02x, expected 0x%02x\n",
					     nopout_cmd->opcode & ISCSI_OPCODE,
					     ISCSI_INIT_NOP_OUT);
				}

				if (size != ntohl(nopout_cmd->length)) {
					TRACE_ERROR
					    ("NopIn DSL is %u, expected %u\n",
					     size, ntohl(nopout_cmd->length));
				} else if (size > 0
					   && memcmp(*buffer,
						     related_command->tx_iov[1].
						     iov_base, size) != 0) {
					TRACE_ERROR
					    ("NopIn Data \"%s\", "
					     "expected \"%s\"\n", *buffer,
					     (char *) related_command->
							tx_iov[1].iov_base);
				} else {
					TRACE(TRACE_ISCSI_FULL,
					      "%u byte NopIn matches NopOut\n",
					      size);
				}

				/* in any case, free up the related command */
				related_command->cmd_state |= CMD_STATE_RXDONE;
				free_pending_command(related_command,
						     current_connection);
			}
		}
	} else {
		/* sent by target on its own to provoke a "ping"
		 * response from us
		 */
		if (unlikely(local_itt != ALL_ONES)) {
			TRACE_ERROR("Invalid NopIn ITT %u when TTT is %u\n",
				    local_itt, ntohl(header->target_xfer_tag));
		}

		/* Draft 20, Section 10.19.3 Lun
		 * "A LUN MUST be set to a correct value when the
		 * Target Transfer Tag is valid (not the reserved
		 * value 0xffffffff)."
		 */
		lun = unpack_lun((__u8 *) & header->lun);
		if (unlikely(!(related_session->lun_bits & (1 << lun)))) {
			TRACE(TRACE_ISCSI, "Invalid NopIn LUN %u\n", lun);
		}

		/* Now send a NopOut PDU in response to this NopIn */
		if (unlikely(current_connection->connection_flags
						& SEND_NO_REPLY_TO_NOP_PING)) {
			TRACE(TRACE_ISCSI,
			      "Not replying to NopIn ping from target\n");
		} else {
			tx_nopout(current_connection, 0, header, buffer, size);
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave rx_nopin\n");
	return 1;		/* only success here */
}

/*
 * print out all keys sent by target in a text response
 * returns 1 on success, 0 on error (unterminated key)
 */
static int
text_key_dump(char *buffer, __u32 size, char *message)
{
	int retval = 1;
	char *ptr, *ptr2;

	if (likely(buffer != NULL && size > 0)) {
		/* print out the info sent by target */
		for (ptr = buffer; size > 0; ptr = ptr2) {
			for (ptr2 = ptr; *ptr2 != '\0' && size > 0;
				ptr2++, size--) {
				/* skip over key itself */
			}
			if (*ptr2 != '\0') {
				*ptr2 = '\0';
				TRACE_ERROR("Unterminated key %s\n", ptr);
				retval = 0;
			}
			printk("%s %s %s\n", current->comm, message, ptr);
			for (; *ptr2 == '\0' && size > 0; ptr2++, size--) {
				/* skip over nulls between keys */
			}
		}
	} else {
		TRACE(TRACE_ISCSI_FULL, "%s empty text response\n",
		      current->comm);
	}
	return retval;
}

/* called only by rx_text_rsp()
 * Returns 1 if need reset due to error, 0 if all ok
 */
static int __attribute__ ((no_instrument_function))
accumulate_text(struct command *command, char *buffer, __u32 size)
{
	int retval, total;

	if ((total = size) > 0) {
		/* buffer contains some data, must accumulate it */
		if (command->buffer_to_free == NULL) {
			command->buffer_to_free = my_kmalloc(MAX_TEXT_LEN,
							     "Text data");
			if (command->buffer_to_free == NULL) {
				retval = 1;
				goto out;
			}
		}

		total += command->data_offset;
		if (total > MAX_TEXT_LEN) {
			TRACE_ERROR("%s too many total bytes %u in text data, "
				    " max %u\n", current->comm, total,
				    MAX_TEXT_LEN);
			retval = 1;
			goto out;
		}

		memcpy(command->buffer_to_free + command->data_offset, buffer,
		       size);
		command->data_offset = total;
	}
	retval = 0;
out:
	return retval;
}

/***************************************************************************
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_rx_thread().
 * This function processes the TextResponse header.
 *
 * Parameters: current_connection which is a ptr to the connection struct
 *			  where the TextResponse header was received.
 * Return Value: >0 success; ==0 on fatal error
 ***************************************************************************/
static int
rx_text_rsp(struct connection *current_connection,
	    __u32 local_itt,
	    struct command *related_command, __u32 size, char *buffer)
{
	struct iscsi_targ_text_rsp *header;
	struct session *related_session;
	struct iscsi_init_text_cmnd *text_cmd;
	int retval = 1, need_reset = 0;

	TRACE(TRACE_ENTER_LEAVE, "Enter rx_text_rsp, current_connection %p\n",
	      current_connection);

	header = (struct iscsi_targ_text_rsp *)current_connection->rx_buf;

	if (related_command == NULL) {
		TRACE_ERROR("%s ignoring Text Response that matches no "
			    "command\n", current->comm);
		print_targ_text_rsp(header);
		goto out;
	}

	related_session = current_connection->my_session;

#ifdef ISCSI_CHECK_PDU_FORMAT
	/* Check if reserved fields are zero or not */
	if (unlikely(!((header->flags & ~(F_BIT | C_BIT)) == 0
	      && header->rsvd2 == 0
	      && header->rsvd4 == 0 && header->rsvd5 == 0))) {
		reserved_not_0("Text Response");
	}
#endif

	if (unlikely(header->flags & C_BIT)) {
		/* C bit is set in Text Response
		 *
		 * RFC 3720 Section 10.11.2 C (Continue) Bit (in Text Response)
		 * "A Text Response with the C bit set to 1 MUST have the
		 * F bit set to 0."
		 */
		if (unlikely(header->flags & F_BIT)) {
			/* have C=1 and F=1 -- try to guess which is wrong */
			
			if (header->target_xfer_tag == ALL_ONES) {
				/* this is ok when F=1, bad when C=1 */
				TRACE_ERROR("Text Response with ITT %u has F=1 "
					    "and C=1, expected C=0\n",
					    local_itt);
				header->flags &= ~C_BIT;
			} else {
				TRACE_ERROR("Text Response with ITT %u has C=1 "
					    "and F=1, expected F=0\n",
					    local_itt);
				need_reset = 1;
				header->flags &= ~F_BIT;
			}
		}
	}
	if (!(header->flags & C_BIT)) {
		/* this pdu completes one full data payload */
		if (related_command->data_offset) {
			/* previous response pdus were continued to this one.
			 * append this new buffer, then dump the whole payload.
			 */
			need_reset = accumulate_text(related_command, buffer,
						     size);
			buffer = related_command->buffer_to_free;
			size = related_command->data_offset;
		}
		if (!need_reset) {
			if (related_session->oper_param->SessionType) {
				text_key_dump(buffer, size, "discovery");
			} else {
				text_key_dump(buffer, size, "text negotiation");
			}
		}
		/* finished with any accumulated data payload */
		related_command->data_offset = 0;
	} else {
		/* this pdu contains only part of the complete data payload */
		if (related_command->data_offset == 0) {
			/* this is first continued pdu, free original request */
			my_kfree((void **)&related_command->buffer_to_free,
				 "Text data");
		}
		need_reset = accumulate_text(related_command, buffer, size);
	}

	if (header->flags & F_BIT) {
		/* F Bit is 1, Target is finished sending everything */
		if (unlikely(header->target_xfer_tag != ALL_ONES)) {
			/* RFC 3720 Section 10.11.1 F (Final) Bit in Text Resp
			 * "A Text Response with the F bit set to 1
			 * MUST have a Target Transfer Tag field set to
			 * the reserved value of 0xffffffff."
			 */
			TRACE_ERROR("Text Response with ITT %u has F=1 TTT=%u, "
				    "expected 0xffffffff\n", local_itt,
				    ntohl(header->target_xfer_tag));
			/* ignore this since F=1 so we close it down anyway */
		}

		/* in any case, free up the related command */
		related_command->cmd_state |= CMD_STATE_RXDONE;
		free_pending_command(related_command, current_connection);

		if (related_session->oper_param->SessionType) {
			/* in a discovery session, shut it down with a logout */
			drive_logout(related_session, current_connection,
				     LOGOUT_CLOSE_SESSION);
		}
	} else {
		/* F bit is 0, target has more to send after this
		 * RFC 3720 Section 10.11.4 Target Transfer Tag
		 * "When a target has more work to do (e.g., cannot transfer all
		 * the remaining text data in a single Text Response or has to
		 * continue the negotiation) and has enough resources to
		 * proceed, it MUST set the Target Transfer Tag to a value other
		 * than the reserved value of 0xffffffff.  Otherwise, the Target
		 * Transfer Tag MUST be set to 0xffffffff.
		 * ...
		 * The initiator MUST copy the Target Transfer Tag and LUN in
		 * its next request to indicate that it wants the rest of the
		 * data."
		 */
		if (unlikely(header->target_xfer_tag == ALL_ONES)) {
			/* RFC 3720 Section 10.11.1 F (Final) Bit in Text Resp
			 * "A text response with the F bit set to 0 MUST
			 * have a Target Transfer Tag field set to a value
			 * other than the reserved 0xffffffff."
			 */
			TRACE_ERROR("Text Response with ITT %u F=0 has TTT="
				    "0x%8x, expected unreserved TTT\n",
				    local_itt, ALL_ONES);
			need_reset = 1;
		}
		/* queue up an empty TextRequest to keep Target going */
		text_cmd = (struct iscsi_init_text_cmnd *)
						&related_command->iscsi_cmd;
		if (need_reset) {
			text_cmd->target_xfer_tag = ALL_ONES;
			text_cmd->lun = 0;
		} else {
			text_cmd->target_xfer_tag = header->target_xfer_tag;
			text_cmd->lun = header->lun;
		}

		/* this follow-on TextRequest pdu has no data attached,
		 * reset iov and msghdr so only header is sent
		 */
		text_cmd->length = 0;
		related_command->tx_iov->iov_len = related_command->tx_size
					= current_connection->basic_hdr_len;
		related_command->tx_iovlen = 1;

		/* reattach the existing Text Request command to end of
		 * pending commands list.  this also sets the CmdSN
		 */
		attach_pending_command(related_command, current_connection, 2);
	}
out:
	TRACE(TRACE_ENTER_LEAVE, "Leave rx_text_rsp, retval %d\n", retval);
	return retval;
}

/***************************************************************************
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_rx_thread().
 * This function processes the LogoutResponse header.
 *
 * Parameters: current_connection which is a ptr to the connection struct
 *			  where the LogoutResponse header was received.
 * Return Value: =0 terminate connection on return, (session lock still locked)
 *		 <0 terminate connection on return, (session lock not locked)
 *		 >0 do not terminate connection (session lock still locked)
 ***************************************************************************/
static int
rx_logout_rsp(struct connection *conn,
	      __u32 local_itt, struct command *related_command)
{
	int reason, response = 1;
	struct iscsi_init_logout_cmnd *logout_cmd;
	struct iscsi_targ_logout_rsp *header;
	struct session *sess = NULL;
	struct semaphore rec_wait_sem;

	TRACE(TRACE_ENTER_LEAVE, "Enter rx_logout_rsp, current_connection %p\n",
	      conn);

	header = (struct iscsi_targ_logout_rsp *)conn->rx_buf;

	/* store the session pointer - SAI */
	sess = conn->my_session;

#ifdef ISCSI_CHECK_PDU_FORMAT
	/* Check if reserved fields are zero or not */
	if (unlikely(!((header->flags & ~F_BIT) == 0
	      && header->rsvd1 == 0
	      && header->rsvd3 == 0
	      && header->rsvd4 == 0 && header->rsvd5 == 0))) {
		reserved_not_0("Logout Response");
	}
#endif

	/* this PDU should be a response to a LogoutRequest sent
	 * earlier by us.
	 */
	if (likely(related_command != NULL)) {
		/* found the corresponding LogoutRequest command */

		response = header->response;

		/* If connection recovery, start a Wait timer - SAI */
		logout_cmd = (struct iscsi_init_logout_cmnd *)
						& (related_command->iscsi_cmd);

		reason = logout_cmd->flags & LOGOUT_REASON;

		related_command->cmd_state |= CMD_STATE_RXDONE;
		free_pending_command(related_command, conn);

		switch (reason) {
		case LOGOUT_CLOSE_SESSION:
			if (response == 0) {
				sess->session_state = SESSION_LOGGED_OUT;
			} else {
				TRACE_ERROR("Logout to close session failed"
					    ", response %u\n", response);
			}
			break;

		case LOGOUT_CLOSE_CONNECTION:
			if (response == 0) {
				/* stop using logged-out connection */
				conn->connection_state = CONNECTION_LOGGED_OUT;
			} else {
				TRACE_ERROR("Logout to close connection failed"
					    ", response %u\n", response);
			}
			break;

		case LOGOUT_REMOVE_CONN_FOR_RECOVERY:
			/* If the response is zero and ErrorRecoveryLevel > 1,
			 * then wait until Time2Wait and then recover - SAI
			 */
			if (response == 0
			    && sess->oper_param->ErrorRecoveryLevel > 1) {
				UNH_UNLOCK(&sess->sess_lock,
					   conn->rx_lock_flags);
				if (be16_to_cpu(header->time2wait) > 0) {
					init_MUTEX_LOCKED(&rec_wait_sem);

					start_connrec_timer(sess,
							    &rec_wait_sem,
							    be16_to_cpu(header->
								time2wait)
								* HZ);
					down_interruptible(&rec_wait_sem);
					/* delete the conn recovery timer SAI */
					del_timer_sync(&sess->connrec_timer);
				}
				deal_with_connrec(conn);
				response = -1;	/* lock no longer held */
			}
			break;

		default:
			/* should never happen */
			TRACE_ERROR("Illegal logout reason %u\n", reason);
			break;
		}	/* switch */
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave rx_logout_rsp, response %d\n",response);
	return response;
}

/***************************************************************************
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_rx_thread().
 * This function processes the SCSI response header. The status is passed onto
 * the SCSI Mid-level. The sense data, if present, is received but not passed to
 * the SCSI Midlevel.
 * Parameters: current_connection which is a ptr to the connection struct
 *			  where the SCSI Response header was received.
 * Return Value: >0 success, session lock is locked
 *		 =0 failure, session lock is locked
 *		 <0 failure, session lock is NOT locked
 ***************************************************************************/
static int
rx_rsp(struct connection *current_connection,
       __u32 local_itt,
       struct command *related_command, __u32 size, char *buffer)
{
	struct iscsi_targ_scsi_rsp *scsi_rsp_header;
	struct init_error_rec err_rec;
	int retval = 1;

	TRACE(TRACE_ENTER_LEAVE, "Enter rx_rsp, current_connection %p\n",
	      current_connection);

	scsi_rsp_header
		= (struct iscsi_targ_scsi_rsp *) current_connection->rx_buf;

#ifdef ISCSI_CHECK_PDU_FORMAT
	/* Check if reserved fields are zero or not */
	if (unlikely(!(((scsi_rsp_header->flags & (R_BIT | W_BIT | S_BIT)) == 0)
	      && (scsi_rsp_header->target_xfer_tag == 0)))) {
		reserved_not_0("SCSI Response");
	}

	/* Draft 20, Section 10.4.2 Status
	 * "The Status field is used to report the SCSI status of the command
	 * (as specified in [SAM2]) and is valid only if the Response Code is
	 * Command Completed at target."
	 *
	 * Draft 20, Section 10.4.3 Response
	 * "0x00 - Command Completed at Target"
	 */
	if (unlikely(scsi_rsp_header->response != 0
					&& scsi_rsp_header->status != 0)) {
		TRACE_ERROR("Status %u should be 0 when Response is %u\n",
			    scsi_rsp_header->status, scsi_rsp_header->response);
	}
#endif

	TRACE(TRACE_DEBUG, "Got StatSN %u\n", ntohl(scsi_rsp_header->stat_sn));

	if (likely(related_command != NULL)) {
		/* target is responding to this command,
		 * so it does not need to be sent again on a retransmission
		 */
		related_command->cmd_state &= ~CMD_STATE_REXMIT_ORIG;

		/* Check if ExpDataSN field in header is what we expected */
		/* NOTE: we give warning and do not try recovery if the
		 * command is not actually a read
		 */
		/* Draft 20, Section 10.4.8 ExpDataSN
		 * "The number of Data-In (read) PDUs the target has
		 * sent for the command.
		 *
		 * This field is reserved if the response code is not Command
		 * Completed at Target or the command is a write command."
		 *
		 * Draft 20, Section 10.4.3 Response
		 * "0x00 - Command Completed at Target"
		 */
		if (scsi_rsp_header->response != 0x00
		    || ((related_command->header_opcode == ISCSI_INIT_SCSI_CMND
			 && ((related_command->header_flags & W_BIT)
			|| !(related_command->header_flags & R_BIT)))
			|| related_command->header_opcode
					== ISCSI_INIT_SCSI_DATA_OUT)) {
			/* response not Command Completed or command is write */
#ifdef ISCSI_CHECK_PDU_FORMAT
			if (unlikely(scsi_rsp_header->exp_data_sn)) {
				/* reserved field not 0 but should be */
			      TRACE_WARNING("SCSI Response for ITT %u "
			      "has ExpDataSN %u, should be 0!\n",
			      local_itt,
			      ntohl(scsi_rsp_header->exp_data_sn));
			}
#endif
		} else if (unlikely(ntohl(scsi_rsp_header->exp_data_sn)
					   != related_command->data_in_sn)) {
			TRACE_ERROR
			    	("SCSI Response for ITT %u got ExpDataSN %u, "
			     	"expected %u\n", local_itt,
			     	ntohl(scsi_rsp_header->exp_data_sn),
			     	related_command->data_in_sn);
			/* DataSN Sequence Error Recovery - SAI
			 * decrement the exp_stat_sn as it will
			 * be requested for re-transmission using
			 * STATUS_SNACK - new 18_04 SAI
			 */
			current_connection->exp_stat_sn--;
			err_rec.curr_conn = current_connection;
			err_rec.related_cmd = related_command;
			err_rec.err_type = SEQUENCE_ERR;
			init_do_error_recovery(&err_rec);
			goto rx_rsp_out;
		}

		if (unlikely(related_command->SCpnt == NULL)) {
			/* If we don't have the pointer to the
			 * SCSI Mid-level struct, bail out
			 */
			TRACE_ERROR("No SCSI command found for ITT %u\n",
				    local_itt);
		} else {
			/* Do all the rest of the work in a common function
			 * shared by rx_data() when it detects a
			 * "phase collapse"
			 */
			TRACE(TRACE_DEBUG, "SCSI response, ITT %u, "
			      "recvd_length %u, data_offset %u, "
			      "request_bufflen %u, recv_so_far %u, "
			      "data_in_sn %u\n", local_itt, 
			      related_command->recvd_length,
			      related_command->data_offset,
			      related_command->SCpnt->request_bufflen,
			      related_command->r2t_range_list.limit,
			      related_command->data_in_sn);
			retval = do_scsi_response(current_connection,
						  related_command,
					 	  scsi_rsp_header,size,buffer);
		}
	} else {
		TRACE_ERROR
			("ISCSI Response for ITT %u, response %u, status %u, "
			" ExpDataSN %u, but no command!\n", local_itt,
			scsi_rsp_header->response, scsi_rsp_header->status,
			ntohl(scsi_rsp_header->exp_data_sn));
	}

rx_rsp_out:
	TRACE(TRACE_ENTER_LEAVE, "Leave rx_rsp, retval %d\n", retval);
	return retval;
}

static inline struct command * __attribute__ ((no_instrument_function))
itt_required(__u32 attributes, __u32 local_exp_cmd_sn, __u32 local_itt,
		__u32 opcode, struct connection *conn)
{
	__u32 i, delta;
	struct command *related_command;

	if (unlikely(local_itt == ALL_ONES)) {
		/* ITT has reserved value when it should not */
		TRACE_ERROR("Response has ITT reserved value 0x%08x, "
			     "but should not!\n", local_itt);
		return NULL;
	}
	/* this ITT is not reserved, we should find a matching command */
	related_command = find_command_by_itt(local_itt, conn);
	if (unlikely(related_command == NULL)) {
		TRACE(TRACE_ERROR_RECOVERY, "CID %u, no command has ITT %u\n",
		      conn->connection_id, local_itt);
		i = 0;
		list_for_each_entry(related_command,
				    &conn->pending_commands,
				    link) {
			TRACE(TRACE_ERROR_RECOVERY,"%d: Pending opcode 0x%02x, "
			      "ITT %u, tx_size %u, DataSN %u\n", ++i,
			      related_command->iscsi_cmd.opcode & ISCSI_OPCODE,
			      related_command->init_task_tag,
			      related_command->tx_size,
			      related_command->data_in_sn);
		}
		return NULL;
	}

	related_command->time_stamp = jiffies;
	related_command->activity_flg = 1;
	if (unlikely((attributes & ISCSI_OPCODE)
			!= (related_command->iscsi_cmd.opcode & ISCSI_OPCODE))){
		TRACE_ERROR("Command matching ITT %u has opcode 0x%02x, "
			    "expected 0x%02x\n", local_itt,
			    related_command->iscsi_cmd.opcode & ISCSI_OPCODE,
			    attributes & ISCSI_OPCODE);
	} else if (likely(!(related_command->iscsi_cmd.opcode & I_BIT))) {
		delta = local_exp_cmd_sn
				    - ntohl(related_command->iscsi_cmd.cmd_sn);
		if (unlikely(delta <= 0)) {
			TRACE_ERROR("opcode 0x%02x ITT %u has ExpCmdSN %u not "
				    "bigger than CmdSN %u of related command\n",
				    opcode & ISCSI_OPCODE, local_itt,
				    local_exp_cmd_sn,
				    ntohl(related_command->iscsi_cmd.cmd_sn));
		}
	}
	return related_command;
}

/****************************************************************************
 * The rx thread exists for each connection.  the thread sits
 * blocked on recvmsg until it can successfully read an iscsi header.  at
 * this point it passes off everything to the appropriate function based
 * on the type of header it received.  this function processes the rest of
 * the frame.  this way data can be read right into the correct buffers.
 ****************************************************************************/
void
iscsi_initiator_rx_thread(void *arg)
{
	struct connection *current_connection;
	struct session *current_session;
	__u32 size, i;
	int received, delta;
	__u32 checksum, local_itt, local_exp_cmd_sn;
	__u32 opcode, attributes;
	struct command *related_command;
	struct init_error_rec err_rec;
	char *buffer;
	struct response_pdu *header;
	char string[20];
	struct list_head *lptr, *next;

	current_connection = (struct connection *) arg;
	current_session = current_connection->my_session;

	/* Lock kernel to daemonize thread and update rx thread pointer */
	lock_kernel();

	/* NOTE: struct task_struct has declared char comm[16] */

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	daemonize(
#else
	daemonize();
	snprintf(current->comm, sizeof(current->comm),
#endif
		 "rx-%u-%02u", current_session->scsi_target_id,
		 current_connection->connection_id);

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 18)
	/*
	 * Arne Redlich, agr1@users.sourceforge.net:
	 * This prevents the rx_thread from becoming a zombie after it exits.
	 */
	reparent_to_init();
#endif

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGBITS);

	/* mark that this rx thread is alive */
	current_connection->rx_thread = current;

	unlock_kernel();

	/* signal anyone waiting that the rx thread is up and running */
	up(&current_connection->rx_done_sem);

	printk("%s Starting pid %d conn %p\n", current->comm, current->pid,
	       current_connection);

	/* The rx thread normally holds exclusive access to the
	 * iscsi structure associated with this session
	 */
	UNH_LOCK(&current_session->sess_lock,
		 current_connection->rx_lock_flags);

	/* no local buffer allocated yet for data attached to a pdu */
	buffer = NULL;

	/* each time around this loop we read and process one complete pdu */
	do {
		received = recv_pdu_header(current_connection);
		if (unlikely(received <= 0)) {
			/* Connection died or we got a kill signal,
			 * terminate this thread after checking if connection
			 * recovery is supported - SAI
			 * Side effect: session lock is lost since received < 0
			 */
			if (current_connection->my_session->oper_param->
			    			ErrorRecoveryLevel > 1) {
				/* Perform connection recovery */
				deal_with_connrec(current_connection);
			}
			break;
		}

		if (current_connection->connection_state !=
						CONNECTION_FULL_FEATURE_PHASE) {
			if (current_connection->connection_state ==
						CONNECTION_LOGGED_OUT) {
				printk("%s logged out\n", current->comm);
			}
			break;
		}

		/* mark that we got some activity from the target */
		current_connection->connection_flags |= GOT_ACTIVITY;

		if (current_connection->connection_flags & NEED_TX_WAKEUP) {
			/* get the tx thread started again */
			if (atomic_read(&current_connection->tx_sem.count)<=0) {
				TRACE(TRACE_ISCSI_FULL,
				      "%s wake up tx thread\n", current->comm);
				up(&current_connection->tx_sem);
			}
		}

		header = (struct response_pdu *)current_connection->rx_buf;

		/* pick up number of bytes of data attached to
		 * this PDU (if any) plus the AHS in the high-order byte
		 */
		size = ntohl(header->length);

		/* we do not yet handle AHS, so appearance of one is fatal */
		if (unlikely(size > MASK_24_BITS)) {
			/* bigger than 2^24 - 1 */
			TRACE_ERROR("TotalAHSLength 0x%02x, expected 0x00\n",
				    size >> 24);
			break;
		}

		/* ITT, ExpCmdSN in same position in all response pdus */
		local_itt = ntohl(header->init_task_tag);
		local_exp_cmd_sn = ntohl(header->exp_cmd_sn);

		TRACE(TRACE_ISCSI, "%s Recv %s, ITT %u, DSL %u, flags 0x%02x\n",
		      current->comm, printable_iscsi_op(header, string),
		      local_itt, size, header->flags);

		/* Dump and/or print the received iSCSI header */
		TRACE_BUFFER(TRACE_BUF, current_connection->rx_buf,
			     ISCSI_HDR_LEN, "%s got message\n", current->comm);
		if (unlikely(TRACE_TEST(TRACE_ISCSI_FULL))) {
			print_iscsi_command(header);
		}

		/* pick up the opcode byte, with bits that are not in
		 * opcode itself
		 */
		opcode = header->opcode;

		if (current_connection->connection_flags & USE_HEADERDIGEST) {
			/* header digests are in use, compute and check crc */
			checksum = 0;
			do_crc((__u8 *) current_connection->rx_buf,
			       ISCSI_HDR_LEN, &checksum);
			if (unlikely(checksum != header->header_digest)) {
				/* crc check should have been == and was not */
				TRACE_ERROR
					("Header CRC wrong: got 0x%08x, "
					"expected 0x%08x\n"
					"nominal opcode byte 0x%02x\n",
				    	 ntohl(header->header_digest),
				    	 ntohl(checksum), opcode);
				/* Header Digest Error Recovery - SAI */
				err_rec.curr_conn = current_connection;
				err_rec.related_cmd = NULL;
				err_rec.err_type = HEADER_DIGERR;
				init_do_error_recovery(&err_rec);
				continue;
			}
		}

#ifdef ISCSI_CHECK_PDU_FORMAT
		/* Check that the old X bit (bit 7 in byte at offset 0) is 0 */
		if (unlikely(opcode & OLD_X_BIT)) {
			/* bit is set, this is bad in later drafts */
			if (TRACE_TEST(TRACE_ISCSI)) {
			      TRACE_WARNING("response has bit 0 of "
					    "byte 0 = 1, expected 0\n");
			}
			/* turn it off, so it has expected value */
			opcode &= ~OLD_X_BIT;
		}

		/* Check the I bit (bit 6 in byte at offset 0):
		 * -- should NOT be set in later drafts > draft9
		 * -- must be set in earlier drafts <= draft9
		 */
		if (unlikely(opcode & I_BIT)) {
			/* bit is set, this is bad in later drafts */
			if (TRACE_TEST(TRACE_ISCSI)) {
			      TRACE_WARNING("response has I bit = 1, "
					    "expected 0\n");
			}
			/* turn it off, leaving opcode only */
			opcode &= ~I_BIT;
		}
#endif

		/* get attributes for this opcode from table */
		attributes = opcode_attributes[opcode
					& ~(OLD_X_BIT | I_BIT | REPLY_BIT)];

		/* Check that this response opcode is legal, fatal if not */
		if (unlikely(!(opcode & REPLY_BIT)
			|| (attributes & ILLEGAL_REPLY))) {
			TRACE_ERROR("Illegal response opcode 0x%x, %d bytes\n",
				    opcode, received);
			dump_buffer(current_connection->rx_buf, received);
			received = 0;
			break;
		}

#ifdef ISCSI_CHECK_PDU_FORMAT
		/* Check F bit (bit 7 in byte 1), most responses
		 * require it to be 1
		 */
		if (!(header->flags & F_BIT)) {
			/* F bit not set, this may still be ok but
			 * needs further check
			 */
			if (unlikely(attributes & F_MUST_BE_1)) {
				if (TRACE_TEST(TRACE_ISCSI)) {
				      TRACE_WARNING("response has F bit = 0, "
						    "expected 1\n");
				}
			}
		}
#endif

		/* Find the matching command to which this is a response */
		related_command = NULL;

		switch (attributes & ITT_MASK) {

		case ITT_RESERVED:
			/* reserved value should be 0xffffffff */
#ifdef ISCSI_CHECK_PDU_FORMAT
			if (unlikely(local_itt != ALL_ONES)) {
				if (TRACE_TEST(TRACE_ISCSI)) {
				      TRACE_WARNING("response has ITT = "
						    "0x%08x, expected 0x%x\n",
						    local_itt, ALL_ONES);
				}
			}
#endif
			break;

		case ITT_OPTIONAL:
			if (local_itt == ALL_ONES) {
				/* ITT has this reserved value in all drafts */
				break;
			}
			/* fall thru to required */

		case ITT_REQUIRED:
			related_command = itt_required(attributes,
							local_exp_cmd_sn,
							local_itt, opcode,
							current_connection);
			break;
		}	/* switch */

#ifdef ISCSI_CHECK_PDU_FORMAT
		if (header->lun != 0LL) {
			/* the LUN is not 0, which may be ok -- need to check */
			if (unlikely(attributes & LUN_MUST_BE_0)) {
				/* LUN in this response should be reserved */
				if (TRACE_TEST(TRACE_ISCSI)) {
				      TRACE_WARNING("response LUN is %s"
						    ", expected 0\n",
						    string_llx(header->lun,
								string));
				}
			}
		}
#endif

		TRACE(TRACE_DEBUG, "Response DSL %u, ITT %u\n", size,
		      local_itt);

		if (size) {
			/* a data segment is attached, is this ok? */
			if (unlikely(attributes & DSL_MUST_BE_0)) {
				/* this response cannot have any data
				 * attached -- fatal
				 */
				TRACE_ERROR
				    ("Response has DSL = %u, expected 0\n",
				     size);
				break;
			}

			/* If payload is greater than initiator's
			 * MaxRecvDataSegmentLength it is an error
			 */
			if (unlikely(size
				> current_connection->max_recv_length)) {
				TRACE_ERROR("DSL %u greater than initiator's "
					    "MaxRecvDataSegmentLength %u\n",
					    size,
					    current_connection->
							max_recv_length);
			}

			/* for all commands except DataIn, read all data
			 * attached to the pdu into a local buffer.  This
			 * includes, if necessary, reading any pad bytes and
			 * the data digest, and checking the data data digest.
			 */
			if (opcode != ISCSI_TARG_SCSI_DATA_IN) {
				received = recv_local_data(current_connection,
						related_command, size, &buffer);
				if (unlikely(received <= 0)) {
					/* got error or killed by a signal */
					break;
				}
			}
		}

		/* Check StatSN field in all responses except DataIn with S=0 */
		if (opcode != ISCSI_TARG_SCSI_DATA_IN
						|| (header->flags & S_BIT)) {
			/* check exp_stat_sn with stat_sn value received
			 * in the header
			 */
			i = ntohl(header->stat_sn);
			delta = i - current_connection->exp_stat_sn;
			if (unlikely(delta != 0)) {
				TRACE_ERROR("Opcode 0x%02x, got StatSN %u, "
					     "expected %u\n", opcode, i,
					     current_connection->exp_stat_sn);

				/* for iscsi responses, trigger error
				 * recovery if the statsn
				 * does not match - new 18_04 - SAI
				 */
				if (opcode == ISCSI_TARG_SCSI_RSP) {
					if (delta == 1) {
					    current_connection->exp_stat_sn = i;
					} else {
					/* DataSN Sequence Error Recovery */
					err_rec.curr_conn = current_connection;
					err_rec.related_cmd = related_command;
					err_rec.err_type = SEQUENCE_ERR;
					if ((received = init_do_error_recovery(
							&err_rec)) <= 0)
						break;
					}
				} else if (opcode == ISCSI_TARG_TASK_MGMT_RSP) {
				    	/* assume Task Management aborted the
					 * command that caused the missing
					 * StatSN, force our ExpStatSN to match
					 * that in the Task Management Response
					 * PDU */
					if (delta > 0) {
					    current_connection->exp_stat_sn = i;
					}
				} else {
					/* resynchronize with target!?! */
					current_connection->exp_stat_sn = i;
				}
			}

			/* the StatSN field advances for all responses in
			 * which it is checked, except R2T and NopIn with
			 * TTT not reserved
			 */

			/* Draft 20, Section 10.8.3 StatSN (under R2T)
			 * "The StatSN field will contain the next StatSN.
			 * The StatSN for this connection is not advanced
			 * after this PDU is sent."
			 *
			 * Draft 20, Section 10.19.2 StatSN (under NOP-In)
			 * "The StatSN field will contain always the next
			 * StatSN.  However, when the Initiator Task Tag is
			 * set to 0xfffffff StatSN for the connection is
			 * not advanced after this PDU is sent."
			 */

			if (opcode != ISCSI_TARG_R2T
			    && (opcode != ISCSI_TARG_NOP_IN
				|| header->init_task_tag != ALL_ONES)) {
				current_connection->exp_stat_sn++;
			}
			TRACE(TRACE_DEBUG, "Got StatSN %u, new ExpStatSN %u\n",
			      i, current_connection->exp_stat_sn);
		}

		/* Check ExpCmdSN and MaxCmdSN fields in all responses */
		check_and_update_cmd_sns(local_exp_cmd_sn,
					 ntohl(header->max_cmd_sn),
					 current_session);

		switch (opcode) {
		case ISCSI_TARG_NOP_IN:	/* NopIn PDU */
			received = rx_nopin(current_connection, local_itt,
					    related_command, size, &buffer);
			break;
		case ISCSI_TARG_SCSI_RSP:	/* SCSIResponse PDU */
			received = rx_rsp(current_connection, local_itt,
					  related_command, size, buffer);
			break;
		case ISCSI_TARG_TASK_MGMT_RSP:	/* TaskManagementResponse PDU */
			received = rx_task_mgt_rsp(current_connection,
						local_itt, related_command);
			break;
		case ISCSI_TARG_LOGIN_RSP:	/* LoginResponse PDU */
			TRACE_ERROR("Unexpected Login Response, status 0x%02x, "
				    "%d bytes\n", header->status_class,
				    received);
			received = 0;	/* fatal error */
			break;
		case ISCSI_TARG_TEXT_RSP:	/* TextResponse PDU */
			received = rx_text_rsp(current_connection, local_itt,
					       related_command, size, buffer);
			break;
		case ISCSI_TARG_SCSI_DATA_IN:	/* DataIn PDU */
			received = rx_data(current_connection, local_itt,
					   related_command, size, string);
			break;
		case ISCSI_TARG_LOGOUT_RSP:	/* LogoutResponse PDU */
			received = rx_logout_rsp(current_connection, local_itt,
						 related_command);
			break;
		case ISCSI_TARG_R2T:	/* R2T PDU */
			received = rx_r2t(current_connection, local_itt,
						 related_command);
			break;
		case ISCSI_TARG_ASYNC_MSG:	/* Asynchronous Message PDU */
			received = rx_async_msg(current_connection, size,
									buffer);
			break;
		case ISCSI_TARG_RJT:	/* Reject PDU */
			received = rx_rjt(current_connection, size, buffer);
			if (received == 0) {
				/* fatal error in handling a reject */
				err_rec.curr_conn = current_connection;
				err_rec.related_cmd = NULL;
				err_rec.err_type = HEADER_DIGERR;
				init_session_recovery(&err_rec);
			}
			break;
		default:		/* illegal opcode */
			TRACE_ERROR("Illegal response opcode 0x%x\n", opcode);
			received = 0;
			break;
		}		/* switch */

		/* we have completely finished reading and processing this pdu*/
#ifdef ISCSI_STATS
		if (likely(received > 0)) {
		   UPDATE_RX_SESS_STATS(opcode, size,
					current_session->sess_stats,
					current_connection);
		}
#endif

		/* We are now finished with any locally allocated data buffer
		 * that is not still in use (i.e., being sent in NopOut).
		 */
		if (unlikely(buffer != NULL))
			my_kfree((void **) &buffer, "local buf");
	} while (received > 0);

	/* The rx thread is finished, tell the world
	 * Note that if received < 0 we lost the session lock
	 */
	if (received < 0) {
		TRACE(TRACE_ISCSI, "%s Received error %d\n", current->comm,
		      -received);
		/* get back exclusive access to iscsi structures */
		UNH_LOCK(&current_session->sess_lock,
			 current_connection->rx_lock_flags);
	} else {
		TRACE(TRACE_ISCSI, "%s Received %d\n", current->comm,
		      received);
	}

	/* We are now finished with any locally allocated data buffer
	 * that was not cleaned up due to an error
	 */
	my_kfree((void **) &buffer, "local buf");

	/* If connection recovery active, do not free pending commands - SAI */
	i = 0;
	if (current_session->n_rec_tasks <= 0) {
		TRACE(TRACE_DEBUG, "Freeing pending commands\n");

		list_for_each_safe(lptr, next,
				   &current_connection->pending_commands) {
			related_command = list_entry(lptr,struct command,link);
			i++;
			if (related_command->SCpnt != NULL) {
				/* set up the final status to pass back
				 * up to SCSI Mid-level
				 */
				related_command->SCpnt->result
							= DID_ERROR << 16;
			}
			related_command->cmd_state
					|= CMD_STATE_RXDONE | CMD_STATE_TXDONE;
			/* free pending will call scsi_done() to pass
			 * back final status
			 */
			free_pending_command(related_command,
					     current_connection);
		}
	}

	TRACE(TRACE_DEBUG, "%d pending commands freed\n", i);

	/* save an invalid lun number from this connection in case we need it
	 * below to shut down this session
	 */
	for (i = 0; i < 32 && (current_session->lun_bits & (1 << i)); i++) {
		/* i is a valid lun number in this session */
	}

	/* NULL the rx thread before exit - SAI */
	current_connection->rx_thread = NULL;
	current_connection->connection_state = CONNECTION_DISCONNECTED;

	/* signal anyone waiting that the rx thread is finished */
	up(&current_connection->rx_done_sem);

	printk("%s Exiting pid %d conn %p\n", current->comm, current->pid,
	       current_connection);

	if (current_session->session_state == SESSION_LOGGED_OUT
	    || (current_session->session_state == SESSION_FULL_FEATURE_PHASE &&
		current_session->oper_param->SessionType)) {

		current_session->session_state = SESSION_CLOSED;

		/* release our exclusive access to the iscsi structures */
		UNH_UNLOCK(&current_session->sess_lock,
			   current_connection->rx_lock_flags);

		remove_session_from_hostdata(current_session->scsi_target_id, i,
					     current_connection->connection_id,
					     global_hostdata);

	} else if (current_connection->connection_flags & CONN_HIT_EOF
		    && current_connection->my_session->oper_param->
			    			ErrorRecoveryLevel <= 1) {
		err_rec.curr_conn = current_connection;
		err_rec.related_cmd = NULL;
		err_rec.err_type = HEADER_DIGERR;
		init_session_recovery(&err_rec);
	} else {
		/* release our exclusive access to the iscsi structures */
		UNH_UNLOCK(&current_session->sess_lock,
			   current_connection->rx_lock_flags);
	}
}
