/*	initiator/initiator_utilities.c
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This file contains auxilliary functions for iscsi initiator code
 *	that are responsible for connecting and maintaining sessions
 *	between initiator and target.
 *
 *	Copyright (C) 2001-2004 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote 
 *	products derived from this software without specific prior 
 * 	written permission.
*/
#include "iscsi_initiator.h"
#include "initiator_utilities.h"
#include "initiator_error_rec.h"
#include "../common/lun_packing.h"

/* holds bytes of zeros that are sent as pad bytes when needed */
static __u32 pad_bytes = 0;

/* allocates a new r2t cookie for this session
 * but only initializes the order_link thread in this structure
 */
struct r2t_cookie * __attribute__ ((no_instrument_function))
create_r2t_cookie(struct session *this_session)
{
	struct r2t_cookie *cookie;

	if (unlikely(list_empty(&this_session->free_r2t_cookies))) {
		/* allocate a new r2t cookie */
		cookie = (struct r2t_cookie *)
			my_kmalloc(sizeof(struct r2t_cookie), "r2t cookie");
		if (unlikely(cookie == NULL))
			goto out;
	} else {
		/* remove an old r2t cookie from the free list and reuse it */
		cookie = list_entry(this_session->free_r2t_cookies.next,
				    struct r2t_cookie, link);
		list_del(&cookie->link);
	}
	INIT_LIST_HEAD(&cookie->order_link);
out:
	return cookie;
}

/* finishes initializing an r2t cookie and adds it to end of command's
 * cookie list
 */
void __attribute__ ((no_instrument_function))
hookup_r2t_cookie(struct r2t_cookie *cookie,
		  struct session *this_session,
		  struct connection *this_connection,
		  struct command *this_command,
		  __u32 local_ttt)
{
	cookie->target_xfer_tag = local_ttt;
	if (local_ttt != ALL_ONES) {
		/* this is a "real" (not implied) r2t.
		 * count it and check number outstanding.
		 */
		this_command->n_r2t_cookies++;
		if (unlikely(this_command->n_r2t_cookies >
				this_session->oper_param->MaxOutstandingR2T)) {
			/* too many outstanding R2Ts */
			TRACE_ERROR("Target sent %u outstanding R2Ts, "
				    "but MaxOutstandingR2T is %u\n",
				    this_command->n_r2t_cookies,
				    this_session->oper_param->
				    MaxOutstandingR2T);
			/* now ignore this, since we can handle it ok */
		}
	}

	/* Draft 20, Section 10.7.5 DataSN
	 * "For output (write) data PDUs, the DataSN is the Data-Out PDU number
	 * within the current output sequence.  The current output sequence is
	 * either identified by the Initiator Task Tag (for unsolicited data) or
	 * is a data sequence generated for one R2T (for data solicited through
	 * R2T)."
	 *
	 * Therefore, each R2T resets data_out_sn to 0.
	 */
	cookie->data_out_sn = 0;

	/* Draft 20, Section 10.8 Ready to Transfer (R2T)
	 * "The target may send several R2T PDUs.  It, therefore, can have a
	 * number of pending data transfers.  The number of outstanding R2T
	 * PDUs [per task] are limited by the value of the negotiated key
	 * MaxOutstandingR2T.  Within a connection, outstanding R2Ts MUST be
	 * fulfilled by the initiator in the order in which they were received."
	 *
	 * Therefore, each R2T gets put on ordered list within its connection.
	 */
	list_add_tail(&cookie->order_link, &this_connection->r2t_list);

	/* Update command offset to that expected in next in-order R2T */
	this_command->data_offset += cookie->r2t_xfer_length;

	if (unlikely(this_session->oper_param->DataSequenceInOrder == 0)) {
		/* DataSequenceInOrder==No */
		merge_offset_length(&this_command->r2t_range_list,
				    cookie->offset, cookie->r2t_xfer_length);
	}

	if (unlikely(this_session->oper_param->DataPDUInOrder == 0)) {
		/* DataPDUInOrder==No, send Data Out PDUs backwards */
		cookie->offset += cookie->r2t_xfer_length;      /* set limit */
	}

	/* add this to end of cookie list for this command */
	list_add_tail(&cookie->link, &this_command->r2t_cookies);
}

/* put a no longer used cookie structure on available list for session */
void __attribute__ ((no_instrument_function))
uncreate_r2t_cookie(struct r2t_cookie *cookie, struct session *this_session)
{
	/* remove this cookie from the ordered list of cookies
	 * remaining to process on this connection
	 */
	list_del(&cookie->order_link);

	list_add_tail(&cookie->link, &this_session->free_r2t_cookies);
}

/* removes cookie from this command's and this connection's cookie lists */
void __attribute__ ((no_instrument_function))
free_r2t_cookie(struct connection *this_connection,
		struct command *this_command,
		struct r2t_cookie *cookie)
{

	if (cookie->target_xfer_tag != ALL_ONES) {
		/* this is a "real" (not implied) r2t.
			reduce number still outstanding */
		/*   Draft 20 Section 12.17 MaxOutstandingR2T
		 * "An R2T is considered outstanding until the last data
		 * PDU (with the F bit set to 1) is transferred, or a
		 * sequence reception timeout (Section 6.1.4.1 Recovery
		 * Within-command) is encountered for that data
		 * sequence."
		 */
		this_command->n_r2t_cookies--;
	}

	/* remove cookie from list of r2t cookies for this command */
	list_del(&cookie->link);

	/* then add it to front of list of available cookies */
	uncreate_r2t_cookie(cookie, this_connection->my_session);
}

/* info - debug
 * keeps track of maximum  number of io vector slots actually used
 */
int max_iov_count = 0;

/*
 * executed by init_recovery thread, tx thread, scsi midlevel process.
 * the session->sess_lock MUST be held by the calling process/thread.
 * set up the transmit i/o vector for the iscsi header
 */
static void __attribute__ ((no_instrument_function))
setup_header_iovec(struct connection *conn, struct command *this_command)
{
	this_command->data_offset = 0;
	this_command->data_checksum = 0;
	this_command->tx_size = conn->basic_hdr_len;
	this_command->tx_iov->iov_base = &this_command->iscsi_cmd;
	this_command->tx_iov->iov_len = conn->basic_hdr_len;
	this_command->tx_iovlen = 1;
}

/*
 * executed only by tx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called by iscsi_initiator_tx_thread() and tx_retransmit_state1().
 * build a SCSI Command with write data to be sent
 */
void
build_write_command(struct session *current_session,
		    struct connection *current_connection,
		    struct command *new_command)
{
	struct iovec *iov;
	struct iscsi_init_scsi_cmnd *iscsi_cmd;
	struct scsi_cmnd *Cmnd;
	struct r2t_cookie *cookie;
	int len, count = 1;
	int negotiated_immediate_data_length;
	int actual_immediate_data_length = 0;
	int space_left;

	Cmnd = new_command->SCpnt;
	if (unlikely(Cmnd == NULL)) {
		/* should never happen */
		TRACE_ERROR("No SCpnt for command %p\n", new_command);
		goto out;
	}

	/* first, set up the transmit i/o vector for the iscsi header */
	setup_header_iovec(current_connection, new_command);

	iscsi_cmd = &new_command->iscsi_cmd;
	iscsi_cmd->flags = W_BIT;
	if (Cmnd->request_bufflen
		&& current_session->oper_param->ImmediateData) {
		/* We are allowed to send immediate data.
		 * get the smaller of target's MaxRecvDataSegmentLength or
		 * FirstBurstLength
		 */
		negotiated_immediate_data_length
			= current_connection->max_send_length;
		if (negotiated_immediate_data_length
			> current_session->oper_param->FirstBurstLength) {
			negotiated_immediate_data_length
			    = current_session->oper_param->FirstBurstLength;
		}

		/* Actual datalength to be sent as immediate data is minimum of
		 * what command wants to send and the negotiated mimimum allowed
		 * for immediate data
		 */
		actual_immediate_data_length = Cmnd->request_bufflen;
		if (actual_immediate_data_length
			> negotiated_immediate_data_length)
			actual_immediate_data_length
				= negotiated_immediate_data_length;

		TRACE(TRACE_DEBUG, "Sending immediate data %d\n",
		      actual_immediate_data_length);

		/* get pointer to tx i/o vector (which is already set up
		 * for header)
		 */
		iov = new_command->tx_iov;

		if (Cmnd->use_sg > 0) {
			/* we have a scatter gather list to be sent */

			int immediate_data_to_be_filled
				= actual_immediate_data_length;
			struct scatterlist *sg
				= (struct scatterlist *) Cmnd->request_buffer;

#ifdef K26
			if (use_sendpage(current_connection, new_command)) {
				new_command->pindex = 0;
				new_command->padding = 0;
			}
#endif

			while (immediate_data_to_be_filled > 0) {
				/* set up iovector to send immediate_data */
				count++;
				iov++;

#ifdef K26
				/* Convert page struct passed in SG
				 * to kernel vaddr. We can sleep on this
				 * call so have to release our lock.
				 */

				UNH_UNLOCK(&current_session->sess_lock, 
					current_connection->tx_lock_flags);
				sg_dma_address(sg) = (dma_addr_t)kmap(sg->page);

				UNH_LOCK(&current_session->sess_lock, 
					current_connection->tx_lock_flags);
							
				sg_dma_address(sg) += sg->offset;

				TRACE(TRACE_DEBUG,
				      "sg %p sg_address %x sg_length %d "
				      "immediate_data_filled %d\n",
				      sg, sg_dma_address(sg), sg_dma_len(sg),
				      immediate_data_to_be_filled);
				iov->iov_base
					 = (unsigned long *)sg_dma_address(sg);
				iov->iov_len = sg_dma_len(sg);
				immediate_data_to_be_filled -= sg_dma_len(sg);

				if (use_sendpage(current_connection, new_command)) {
					new_command->sg[new_command->pindex] = *sg;
					new_command->pindex++;
				}
#else
				TRACE(TRACE_DEBUG,
				      "sg_address %p sg_length %d "
				      "immediate_data_filled %d\n",
				      sg->address, sg->length,
				      immediate_data_to_be_filled);
				iov->iov_base = sg->address;
				iov->iov_len = sg->length;
				immediate_data_to_be_filled -= sg->length;
#endif
				if (immediate_data_to_be_filled < 0) {
					/* went too far, last element
					 * needs to be shortened
					 */
					iov->iov_len
						+= immediate_data_to_be_filled;
				}
				/* compute the partial data digest if we
				 * are using them
				 */
				if (current_connection->connection_flags
					& USE_DATADIGEST) {
					do_crc(iov->iov_base, iov->iov_len,
					       &new_command->data_checksum);
				}
				sg++;
			}
		} else {
			/* we have single buffer passed from SCSI Mid-level */
			count++;
			iov++;
			iov->iov_len = actual_immediate_data_length;
			iov->iov_base = Cmnd->request_buffer;
			/* compute the data digest if we are using them */
			if (current_connection->connection_flags
				& USE_DATADIGEST) {
				do_crc(iov->iov_base, iov->iov_len,
				       &new_command->data_checksum);
			}
		}

		iscsi_cmd->length = htonl(actual_immediate_data_length);
		new_command->tx_data_length = actual_immediate_data_length;
		new_command->tx_size += actual_immediate_data_length;
		new_command->data_offset = actual_immediate_data_length;

		if (unlikely(current_session->oper_param->DataSequenceInOrder
									== 0)) {
			/* DataSequenceInOrder==No */
			merge_offset_length(&new_command->r2t_range_list,
					    0, actual_immediate_data_length);
		}

		/* now deal with any padding that is needed
		 *
		 * Draft 20, Section 10.1 iSCSI PDU Length and Padding
		 * "iSCSI PDUs are padded to the closest integer number of
		 * four byte words.  The padding bytes SHOULD be sent as 0."
		 */
		len = (-actual_immediate_data_length) & 3;
		if (unlikely(len != 0)) {
			/* send len pad bytes in another element in iovector */
			count++;
			iov++;
			iov->iov_base = &pad_bytes;
			iov->iov_len = len;
			new_command->tx_size += len;	/* now a multiple of 4*/
			TRACE(TRACE_ISCSI_FULL, "Attach %d pad bytes\n", len);
			/* compute partial data digest if we are using them */
			if (current_connection->connection_flags
				& USE_DATADIGEST) {
				do_crc(iov->iov_base, iov->iov_len,
				       &new_command->data_checksum);
			}
#ifdef K26
			if (use_sendpage(current_connection, new_command))
				new_command->padding = len;
#endif
		}

		/* if data digests are in use, be sure to set it up now */
		if (current_connection->connection_flags & USE_DATADIGEST) {
			count++;
			iov++;
			iov->iov_base = &new_command->data_checksum;
			iov->iov_len = CRC_LEN;
			new_command->tx_size += CRC_LEN;
			TRACE(TRACE_ISCSI_FULL, "Attach DataDigest\n");
		}
	}

	/* total number of data bytes left to send by this write command */
	new_command->still_to_send = Cmnd->request_bufflen
						- actual_immediate_data_length;

	/* number of bytes that could go into first burst as unsolicited */
	space_left = current_session->oper_param->FirstBurstLength
						- actual_immediate_data_length;

	/* assume no unsolicited DataOut pdus will follow this one */
	new_command->cmd_msg_flags = 0;

	if (current_session->oper_param->InitialR2T
		|| new_command->still_to_send == 0
		|| space_left <= 0) {
		/* set F bit as no separate unsolicted data PDUs can follow */
		iscsi_cmd->flags |= F_BIT;
	} else if (!(current_session->oper_param->InitialR2T)) {
		/* we are allowed to send unsolicited data PDUs
		 * after this command, but do we need to and
		 * is there any burst space left to do so?
		 */
		if (new_command->still_to_send > 0 && space_left > 0) {
			/* still data left to send, and there is room in first
			 * burst, so set up to send unsolicited data PDUs next
			 */
			if ((cookie = create_r2t_cookie(current_session))
								!= NULL) {
				cookie->offset = new_command->data_offset;
				cookie->r2t_xfer_length
						= new_command->still_to_send;
				if (new_command->still_to_send > space_left)
					cookie->r2t_xfer_length = space_left;
				hookup_r2t_cookie(cookie, current_session,
		  				  current_connection,
						  new_command, ALL_ONES);

				/* tell tcp DataOut pdus follow */
				new_command->cmd_msg_flags = MSG_MORE;

				/* need to mark this in case of retransmission,
				 * because xmitting the cookie will overwrite
				 * the command iovector and both will have
				 * to be set up again.
				 */
				new_command->cmd_state = CMD_STATE_REXMIT_ORIG;
			}
		}
	}

	/* fill in fields for tx thread */
	new_command->tx_sent_so_far = 0;	/* nothing sent so far */
	new_command->retransmit_tx_size = new_command->tx_size;
	new_command->header_complete = 0;	/* need to finish header */

	/* These are used in do_tx_send() and rx_rsp(), both of which will not
	 * see this until after we set them here.
	 */
	new_command->header_opcode = ISCSI_INIT_SCSI_CMND;
	new_command->header_flags = iscsi_cmd->flags;

out:
	/* set the number of iov slots actually used */
	new_command->tx_iovlen = count;

	if (unlikely(max_iov_count < count))
		max_iov_count = count;	/* keep track of max no. ever used */
}

/*
 * executed only by tx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 *
 * Called only from ready_to_xmit().
 * At the time of the call, cookie is known to be first on related_command's
 * r2t_cookies list, it is known to be non-NULL, it is known to be at the
 * beginning of related_connection's order_link, and it is known to
 * have a positive xfer length.
 *
 * Called to build a DataOut pdu header in the "data" area of the
 * "related_command" if that area is currently unused (else does nothing).
 * The amount of data to send is controlled by the first "cookie" for
 * that command, but the location(s) and layout of the data in memory
 * are controlled by the SCpnt field in the "related_command" when
 * setting up the msghdr and io vector in the "related_command".
 * Returns number of bytes newly set up for transmission (0 if none).
 *
 * While writing, the following variables in the related_command maintain state:
 *	tx_size			number of bytes left to send in current xmission
 *				(updated as each pdu is sent)
 *	tx_sent_so_far		number of bytes sent so far in current xmission
 *				(updated as each pdu is sent)
 *	retransmit_size		number of original bytes in current xmission
 *	still_to_send		total number of bytes of data not yet sent for
 *				this write command
 *				(updated as each pdu is sent)
 *	r2t_sn			R2TSN expected in next R2T from target
 *	data_offset		offset expected in next in-order R2T from target
 *				(updated as each R2T is received)
 *	n_r2t_cookies		number of elements in list r2t_cookies that were
 *				created by R2Ts (excludes cookie created for
 *				unsolicited DataOut pdus)
 *				(1 cookie created as each R2T is received)
 *	r2t_cookes		list with 1 r2t_cookie for each unsatisfied R2T
 *				(includes 1 r2t_cookie for any unsolicited pdus)
 *				(cookies removed from list as satisfied)
 *	header_opcode		opcode of last command sent carrying data
 *				(0x01 on WRITE, 0x05 for DataOut)
 *	data			area where DataOut pdu is built
 */
int
setup_dataoutpdu(struct connection *related_connection,
		 struct command *related_command,
		 struct r2t_cookie *cookie)
{
	int len, prev_offset, cur_offset, count, size;
	char *cur_ptr;
	struct iovec *iov;
	struct scatterlist *sg;
	struct session *related_session;
	int sglength;

	TRACE(TRACE_ENTER_LEAVE, "Enter setup_dataoutpdu\n");

	size = 0;
	if (related_command->tx_size != 0) {
		/* command's single data slot is busy, don't overwrite it */
		TRACE(TRACE_ISCSI_FULL,
		      "Busy DataOut slot, ITT %u,  tx_size %u\n",
		      related_command->init_task_tag, related_command->tx_size);
		goto out;
	}

	/* set up the DataOut PDU in command's single data slot */
	memset(&related_command->data, 0, ISCSI_HDR_LEN);

	related_command->data.opcode = ISCSI_INIT_SCSI_DATA_OUT;
	related_command->data.target_xfer_tag = cookie->target_xfer_tag;
	related_command->data.init_task_tag
				= htonl(related_command->init_task_tag);
	related_command->data.exp_stat_sn
				= htonl(related_connection->exp_stat_sn);
	related_command->data.data_sn = htonl(cookie->data_out_sn);

	/* assume this is not the last DataOut PDU to be
	 * sent in this sequence
	 */
	related_session = related_connection->my_session;
	len = related_connection->max_send_length;

	if (cookie->r2t_xfer_length <= len) {
		/* bad assumption, this is last DataOut PDU in this seq */
		len = cookie->r2t_xfer_length;
		related_command->data.flags |= F_BIT;
	}

	/* assume no DataOut pdus follow this one */
	related_command->cmd_msg_flags = 0;

	related_command->tx_data_length = len;		/* for tx stats */
	related_command->data.length = htonl(len);
	related_command->still_to_send -= len;

	/* get offset into original SCSI CMD buffer of next byte to be sent */
	if (unlikely(related_session->oper_param->DataPDUInOrder == 0)) {
		/* DataPDUInOrder==No, send Data Out PDUs backwards */
		cookie->offset -= len;		/* next Data Out PDU limit */
		prev_offset = cookie->offset;	/* this Data Out PDU offset */
	} else {
		/* DataPDUInOrder==Yes, send Data Out PDUs in order */
		prev_offset = cookie->offset;	/* this Data Out PDU offset */
		cookie->offset += len;		/* next Data Out PDU offset */
	}

	related_command->data.offset = htonl(prev_offset);

	TRACE(TRACE_DEBUG,
	      "Setup DataOut PDU, offset %d, DSL %d, still_to_send %u\n",
	      prev_offset, len, related_command->still_to_send);

	/* update values in related command cookie to account for this PDU */
	cookie->data_out_sn++;			/* Seqno of next Data Out PDU */
	cookie->r2t_xfer_length -= len;		/* total amount left to send */
	if (((int)cookie->r2t_xfer_length) <= 0) {
		/* finished with this cookie */
		free_r2t_cookie(related_connection, related_command, cookie);
	}

	/* set up the TCP io vector, beginning with no. of bytes to xfer */
	size = related_connection->basic_hdr_len + len;
	iov = related_command->tx_iov;		/* point at empty tx_iov[] */

	/* DataOut pdu header built in separate "data" area of struct command */
	iov->iov_len = related_connection->basic_hdr_len;
	iov->iov_base = &related_command->data;

	related_command->header_complete = 0;
	related_command->header_opcode = related_command->data.opcode;
	related_command->header_flags = related_command->data.flags;

#ifdef K26
	if (use_sendpage(related_connection, related_command)) {
		related_command->pindex = 0;
		related_command->padding = 0;
	}
#endif
	/* now fill in the iovector to point to data to be sent */
	iov++;
	count = 2;		/* always at least 2 items in iovector */
	related_command->data_checksum = 0;	/* initialize data checksum */
	if (related_command->SCpnt->use_sg > 0) {
		/* SCSI gave us a scatter-gather list */
		sg = (struct scatterlist *) related_command->SCpnt
							->request_buffer;
		cur_offset = 0;

#ifdef K26
		/* Convert page struct passed in SG
		 * to kernel vaddr. We can sleep on this
		 * call so have to release our lock.
		 */

		UNH_UNLOCK(&related_session->sess_lock,
				related_connection->tx_lock_flags);
		sg_dma_address(sg) = (dma_addr_t)kmap(sg->page);
		UNH_LOCK(&related_session->sess_lock,
				related_connection->tx_lock_flags);
		cur_ptr = (void *)sg_dma_address(sg);
#else
		cur_ptr = sg->address;
#endif
		/* skip thru scatterlist past data already sent */
		while (cur_offset < prev_offset) {
#ifdef K26
			sglength = sg_dma_len(sg);
#else
			sglength = sg->length;
#endif
			if (cur_offset + sglength > prev_offset) {
				cur_ptr = cur_ptr + (prev_offset - cur_offset);
				cur_offset = prev_offset;
			} else {
				cur_offset += sglength;
				sg++;
#ifdef K26
				cur_ptr =  (void *)sg_dma_address(sg);
#else
				cur_ptr = sg->address;
#endif
			}
		}

		/* sg now points at first list item to xfer */

#ifdef K26
		iov->iov_len = ((char *)(sg_dma_address(sg) + sg_dma_len(sg)))
				- cur_ptr;
		if (use_sendpage(related_connection, related_command)) {
			int offset = cur_ptr - (char *) sg_dma_address(sg);
			related_command->sg[related_command->pindex] = *sg;
			related_command->sg[related_command->pindex].length -= offset;
			related_command->sg[related_command->pindex].offset += offset;
			related_command->pindex++;
		}
#else
		iov->iov_len = (sg->address + sg->length) - cur_ptr;
#endif
		iov->iov_base = cur_ptr;
		sg++;

		len -= iov->iov_len;	/* no. of bytes left to fill */
		while (len > 0) {
			/* need to add another element in iovector, but first
			 * compute the partial data digest if we are using them
			 */
			if (related_connection->connection_flags
				& USE_DATADIGEST) {
				do_crc(iov->iov_base, iov->iov_len,
				       &related_command->data_checksum);
			}
			count++;
			iov++;

#ifdef K26
			iov->iov_base = (unsigned long *)sg_dma_address(sg);
			sglength = sg_dma_len(sg);

			if (use_sendpage(related_connection, related_command)) {
				related_command->sg[related_command->pindex]
						= *sg;
				related_command->pindex++;
			}
#else
			iov->iov_base = sg->address;
			sglength = sg->length;
#endif
			iov->iov_len = sglength;
			len -= sglength;
			sg++;
		}

		if (len < 0) {
			/* only need to send part of the last list item */
			iov->iov_len += len;
		}
	} else {
		/* all the data is in one buffer (no scatter gather list) */
		iov->iov_base
			= related_command->SCpnt->request_buffer + prev_offset;
		iov->iov_len = len;
	}

	/* now deal with any padding that is needed.
	 *
	 * Draft 20, Section 10.1 iSCSI PDU Length and Padding
	 * "iSCSI PDUs are padded to the closest integer number of four byte
	 * words.  The padding bytes SHOULD be 0."
	 */
	len = (-size) & 3;
	if (unlikely(len != 0)) {
		/* need to send len pad bytes in another element in iovector
		 * but first compute the partial data digest if we are
		 * using them
		 */
		if (related_connection->connection_flags & USE_DATADIGEST) {
			do_crc(iov->iov_base, iov->iov_len,
			       &related_command->data_checksum);
		}
		count++;
		iov++;
		iov->iov_base = &pad_bytes;
		iov->iov_len = len;
		size += len;	/* is now a multiple of 4 */
		TRACE(TRACE_ISCSI_FULL, "Attach %d pad bytes\n", len);

#ifdef K26
		if (use_sendpage(related_connection, related_command))
			related_command->padding = len;
#endif
	}

	/* now compute and add the final data digest if we are using them */
	if (related_connection->connection_flags & USE_DATADIGEST) {
		do_crc(iov->iov_base, iov->iov_len,
		       &related_command->data_checksum);
		count++;
		iov++;
		iov->iov_base = &related_command->data_checksum;
		iov->iov_len = CRC_LEN;
		size += CRC_LEN;
		TRACE(TRACE_ISCSI_FULL, "Attach DataDigest\n");
	}

	/* finally save size and number of iovector elements in message header*/
	related_command->tx_sent_so_far = 0;	/* nothing sent yet */
	related_command->tx_size = size;	/* marks this "in use" */
	/* in case of rexmit only */
	related_command->retransmit_tx_size = size;
	related_command->tx_iovlen = count;	/* no. of iovs used */
	if (unlikely(max_iov_count < count)) {
		max_iov_count = count;
	}
	related_command->tx_wait_to_send = jiffies;	/* can send it now */

	if (!list_empty(&related_connection->r2t_list)) {
		/* tell tcp more DataOut pdus are ready to follow this one */
		related_command->cmd_msg_flags = MSG_MORE;
	}
out:
	TRACE(TRACE_ENTER_LEAVE, "Leave setup_dataoutpdu, size %u\n", size);
	return size;
}

/* "prints" statistics on iov usage into buf,
 * returns no. of chars printed.
 */
int
print_iovs(char *buf)
{
	if (buf == NULL)
		return printk("iscsi Maximum of %d iov slots used out of %d\n",
			      max_iov_count, MAX_IOV_SLOTS);
	else
		return sprintf(buf, "Maximum of %d iov slots used out of %d\n",
			       max_iov_count, MAX_IOV_SLOTS);
}

void
endup_iovs(void)
{
	print_iovs(NULL);
}

/*
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called from recv_data_in_data(), recv_pdu_header().
 * Reads "orig_size" bytes into the iovector already set up in conn->rx_iov
 * from the socket attached to "conn".  Uses the rx_msg and rx_iov
 * structures attached to "conn" to control the read.
 * "orig_size" MUST be positive.
 * Returns > 0 ("orig_size") on success with session lock held,
 *	   < 0 on failure with session lock not held.
*/
static int __attribute__ ((no_instrument_function))
recv_iovector(struct connection *conn, int orig_size, int iov_count)
{
	int retval;
	struct msghdr *msg;
	mm_segment_t oldfs;
	struct session *session = conn->my_session;
	char ip_string[INET6_ADDRSTRLEN+2], port_string[8];
#if defined(CONFIG_ISCSI_DEBUG)
	struct iovec *iov, *iov_copy = NULL;
	int index;
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter recv_iovector, orig_size %d, "
		"iov_count %d\n", orig_size, iov_count);

	if (conn->connection_flags & CONN_HIT_EOF) {
		/* previously hit EOF on this connection, don't read again */
		retval = -ECONNRESET;
		goto out_lock_held;
	}

	/* set up this connection's reusable message header.
	 * NOTE: sock_recvmsg() destroys some part of this header,
	 *	 so it MUST be set up EACH TIME through here.
	 */
	msg = &conn->rx_msg;

	memset(msg, 0, sizeof(struct msghdr));	/* most fields are 0/NULL */

	msg->msg_iovlen = iov_count;
	msg->msg_iov = conn->rx_iov;

	if (unlikely(max_iov_count < iov_count)) {
		/* keep track of max no. of iov slots ever used */
		max_iov_count = iov_count;
	}

#if defined(CONFIG_ISCSI_DEBUG)
	if (unlikely(TRACE_TEST(TRACE_BUF))) {
		TRACE(TRACE_BUF, "sock_recvmsg waiting for %d bytes\n",
			orig_size);
		iov = conn->rx_iov;
		for (index = 0; index < iov_count; index++ ) {
			TRACE(TRACE_DEBUG,
				"index %x iov base %p iov size %d \n", 
				index, iov->iov_base, iov->iov_len); 
			iov++;
		}
		index *= sizeof(struct iovec);
		iov_copy = my_kmalloc(index, "iov_copy");
		if (likely(iov_copy != NULL)) {
			memcpy(iov_copy, conn->rx_iov, index);
		}
	}
#endif

	/* allow other threads to run in case we block in the receive */
	UNH_UNLOCK(&session->sess_lock, conn->rx_lock_flags);

	oldfs = get_fs();
	set_fs(get_ds());
	retval = sock_recvmsg(conn->sock, msg, orig_size, MSG_WAITALL);
	set_fs(oldfs);

	UNH_LOCK(&session->sess_lock, conn->rx_lock_flags);

	TRACE(TRACE_DEBUG, "sock_recvmsg got %d bytes\n", retval);

	if (unlikely(retval != orig_size)) {
		/* sock_recvmsg hit an error or end of file */
		if (retval == 0 || retval == -ECONNRESET) {
			/* got an EOF indication, which means remote
				peer is gone */
			if (cnv_inet_to_string(conn->ip_address, ip_string,
					port_string) > 0) {
				TRACE_ERROR("%s end-of-file from %s:%s\n",
					    current->comm, ip_string,
					    port_string);
			}
			conn->connection_flags |= CONN_HIT_EOF;
			retval = -ECONNRESET;
			goto out_lock_held;
		}

		if (retval != -ERESTARTSYS) {
			TRACE_ERROR("sock_recvmsg expected %d bytes, got %d\n",
				    orig_size, retval);
		}
		if (retval > 0) {
			/* to tell caller "lock not held" */
			retval = -ECONNRESET;
		}
		goto out_lock_held;
	}

#if defined(CONFIG_ISCSI_DEBUG)
	if (unlikely(TRACE_TEST(TRACE_BUF))) {
		if ((iov = iov_copy)) {
			for (index = 0; index < iov_count; index++) {
				dump_buffer(iov->iov_base, iov->iov_len);
				iov++;
			}
		}
	}
#endif

out:

#if defined(CONFIG_ISCSI_DEBUG)
	my_kfree((void **)&iov_copy, "iov_copy");
#endif

	TRACE(TRACE_ENTER_LEAVE, "Leave recv_iovector, retval %d\n", retval);
	return retval;

out_lock_held:
	UNH_UNLOCK(&session->sess_lock, conn->rx_lock_flags);
	goto out;
}

/*
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called from init_recovery_thread(), iscsi_initiator_rx_thread().
 * Reads an iscsi pdu header into conn->rx_buf from conn->sock.
 * The header length is 48 without header digests, 52 bytes with.
 * Uses rx_msg and rx_iov structures in "conn" to control the read.
 * Returns > 0 (header length 48 or 52) on success with session lock held,
 * 	   < 0 on failure with session lock not held.
 */
int
recv_pdu_header(struct connection *conn)
{
	int size;
	struct iovec *iov;

	size = conn->basic_hdr_len;

	/* set up this connection's reusable i/o vector to point at base.
	 * NOTE: sock_recvmsg() destroys some part of this header,
	 *	 so it MUST be set up EACH TIME through here.
	 */
	iov = conn->rx_iov;
	iov->iov_len = size;
	iov->iov_base = conn->rx_buf;

	return recv_iovector(conn, size, 1);
}

/****************************************************************************
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_rx_thread.
 * This function is responsible for making the checks for exp_cmd_sn and
 * max_cmd_sn values in the received iSCSI header from the target.
 * Parameters:	exp_cmd_sn from the PDU (in host byte order)
 *		max_cmd_sn from the PDU (in host byte order)
 *		cur_session, pointing to the struct session
 * Return value: 0, if the exp_cmd_sn and max_cmd_sn are valid
 *		-1, if either exp_cmd_sn or max_cmd_sn is invalid.
 ****************************************************************************/
int __attribute__ ((no_instrument_function))
check_and_update_cmd_sns(__u32 exp_cmd_sn, __u32 max_cmd_sn,
			 struct session *cur_session)
{
	int retval = 0, delta;
	struct connection *conn;

	TRACE(TRACE_ENTER_LEAVE, "Enter check_and_update_cmd_sns\n");

	/* RFC 3720 Section 3.2.2.1 Command Numbering and Acknowledging
	 *
	 * "Comparisons and arithmetic on ExpCmdSN and MaxCmdSN MUST use
	 * Serial Number Arithmetic as defined in [RFC1982] where
	 * SERIAL_BITS = 32.
	 *
	 * The target MUST NOT transmit a MaxCmdSN that is less than ExpCmdSN-1.
	 * ...
	 *
	 * MaxCmdSN and ExpCmdSN fields are processed by the initiator as
	 * follows:
	 *
	 * - If the PDU MaxCmdSN is less than the PDU ExpCmdSN-1 (in Serial
	 *   Arithmetic Sense), they are both ignored.
	 *
	 * - If the PDU MaxCmdSN is greater than the local MaxCmdSN (in Serial
	 *   Arithmetic Sense), it updates the local MaxCmdSN; otherwise,
	 *   it is ignored.
	 *
	 * - If the PDU ExpCmdSN is greater than the local ExpCmdSN (in Serial
	 *   Arithmetic Sense), it updates the local ExpCmdSN; otherwise,
	 *   it is ignored.
	 *
	 * This sequence is required because updates may arrive out of order
	 * (e.g., the updates are sent on different TCP connections).
	 *
	 * iSCSI initiators and targets MUST support the command numbering
	 * scheme."
	 */

	delta = max_cmd_sn - (exp_cmd_sn - 1);
	if (delta < 0) {
		/* MaxCmdSN < ExpCmdSN - 1, ignore both PDU values */
		TRACE(TRACE_ISCSI_FULL,
		      "Ignoring PDU's ExpCmdSN %u and MaxCmdSN %u\n",
		      exp_cmd_sn, max_cmd_sn);
	} else {
		/* values in the PDU are ok relative to each other */
		delta = max_cmd_sn - cur_session->max_cmd_sn;
		if (delta < 0) {
			/* PDU.MaxCmdSN < local.MaxCmdSN, ignore
				PDU's MaxCmdSN */
			TRACE(TRACE_ISCSI_FULL, "Ignoring PDU's MaxCmdSN %u\n",
			      max_cmd_sn);
		} else if (delta > 0) {
			/* MaxCmdSN just increased, update the local MaxCmdSN */
			cur_session->max_cmd_sn = max_cmd_sn;
			/* find tx_threads waiting for this limit to increase */
			for (conn = cur_session->connection_head;
			     conn != NULL;
			     conn = conn->next) {
				if (conn->connection_flags & NEED_TX_WAKEUP) {
					/* yes, so wake it up!! */
					if (atomic_read(&conn->tx_sem.count)
						<= 0) {
						TRACE(TRACE_ISCSI_FULL,
					      "rx_thread: wake up tx_thread\n");
						up(&conn->tx_sem);
					}
				}
			}
		}

		delta = exp_cmd_sn - cur_session->exp_cmd_sn;
		if (delta < 0) {
			/* PDU.ExpCmdSN < local.ExpCmdSN,
				ignore PDU's ExpCmdSN */
			TRACE(TRACE_ISCSI_FULL, "Ignoring PDU's ExpCmdSN %u\n",
			      exp_cmd_sn);
		} else if (delta > 0) {
			/* greater than, check it against local CmdSN */

			delta = cur_session->cur_cmd_sn - exp_cmd_sn;
			if (delta < 0) {
				TRACE_ERROR
				    ("Got ExpCmdSN %u, exceeds next CmdSN %u\n",
				     exp_cmd_sn, cur_session->cur_cmd_sn);
				retval = -1;
			} else {
				/* not less than, update the local ExpCmdSN */
				cur_session->exp_cmd_sn = exp_cmd_sn;
			}
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave check_and_update_cmd_sns, retval %d\n",
	      retval);

	return retval;
}

/*
 * the session->sess_lock MUST be held by the calling process/thread.
 * Calls SCSI Mid-level's scsi_done callback with result,
 * (and with session->lock held).
 * Then frees up storage allocated to old command after
 * both tx_thread and tx_thread have completed using the command.
 *
 * Note - The use of RXDONE and TXDONE flags syncronizes the freeing
 * to the command structure in an SMP environment under the
 * protection of session->lock.
 */
void
free_pending_command(struct command *old_command,
		     struct connection *current_connection)
{
	struct scsi_cmnd *SCpnt;
#if defined(CONFIG_ISCSI_DEBUG)
	__u32 local_itt = old_command->init_task_tag;
#endif
	struct r2t_cookie *cookie;
	struct list_head *lptr, *next;

	TRACE(TRACE_ENTER_LEAVE, "Enter free_pending_command %p, ITT %u\n",
	      old_command, local_itt);

	if ((old_command->cmd_state & (CMD_STATE_RXDONE | CMD_STATE_TXDONE))
	    != (CMD_STATE_RXDONE | CMD_STATE_TXDONE)) {
		/* both rx and tx threads have not yet signed
		 * off on this command
		 */
		goto out;
	}

	/* remove this command from this connection's pending command list */
	list_del(&old_command->link);

	/* release storage occupied by any range list used by this command */
	free_range_list(&old_command->r2t_range_list);

	/* free up any r2t cookies still pending for this command */
	list_for_each_safe(lptr, next, &old_command->r2t_cookies) {
		list_del(lptr);
		cookie = list_entry(lptr, struct r2t_cookie, link);
		uncreate_r2t_cookie(cookie, current_connection->my_session);
	}

	/* free up any buffer associated with this command */
	my_kfree(&old_command->buffer_to_free, "NopOut/Text data");

	/* get pointer to SCSI midlevel command */
	SCpnt = old_command->SCpnt;

	/* put old command structure on end of session's free list */
	endup_command(old_command, current_connection->my_session);
#ifdef K26
	/* free up kmappings we did 
	 * if we had scatter-gather list filled 
	 */
	if (SCpnt != NULL && SCpnt->use_sg != 0) {
		/* We filled a scatter-gather list with kmap'ings */
		int i =  SCpnt->use_sg;
		struct scatterlist *sg  = (struct scatterlist *)
		     SCpnt->request_buffer;
		while (i --) {
			TRACE(TRACE_DEBUG,
				"Unmapping %d  page %p for sg_dma_address(sg) %x\n",
				i, sg->page,sg_dma_address(sg));
			kunmap(sg->page);
			sg ++;
		}

	}
#endif /* K26 */

	if (SCpnt != NULL && SCpnt->scsi_done != NULL) {
		/* post to SCSI midlevel that this command completed */
		call_back_scsi_done(SCpnt);
		TRACE(TRACE_DEBUG, "Posted completion, ITT %u\n",
		      local_itt);
	}
out:

	TRACE(TRACE_ENTER_LEAVE, "Leave free_pending_command\n");
}

void
alloc_inquiry_stuff(struct session *sess, __u32 lun, int size)
{
	if ((sess->inquiry_buf[lun] = my_kmalloc(size, "inquiry_buf"))) {
		sess->inquiry_size[lun] = size;
	}
}

void
free_inquiry_stuff(struct session *sess, __u32 lun)
{
	my_kfree((void **)&sess->inquiry_buf[lun], "inquiry_buf");
	sess->inquiry_size[lun] = 0;
}

/* we just successfully completed an inquiry command */
void
save_inquiry(struct session *sess, struct command *related_command, int missing)
{
	int slength;
	__u32 lun;

	slength = related_command->SCpnt->request_bufflen;
	TRACE(TRACE_DEBUG, "completed inquiry, slength %d, missing %d\n",
		slength, missing);
	if (missing > 0)
		slength -= missing;
	if (slength > 0) {
#ifdef K26
		lun = related_command->SCpnt->device->lun;
#else
		lun = related_command->SCpnt->lun;
#endif
		if (sess->inquiry_size[lun] < slength) {
			free_inquiry_stuff(sess, lun);
		}
		if (!sess->inquiry_buf[lun]) {
			alloc_inquiry_stuff(sess, lun, slength);
		}
		if (sess->inquiry_buf[lun]) {
			memcpy(sess->inquiry_buf[lun],
				related_command->SCpnt->request_buffer,
				slength);
			if (unlikely(TRACE_TEST(TRACE_BUF))) {
				TRACE(TRACE_BUF, "inquiry_buf for lun %u, %d "
					"bytes\n", lun, slength);
				dump_buffer(sess->inquiry_buf[lun], slength);
			}
		}
	}
}

/* we just successfully completed a read capacity command */
void
save_capacity(struct session *sess, struct command *related_command, int missing)
{
	int slength;
	__u32 lun;

	slength = related_command->SCpnt->request_bufflen;
	TRACE(TRACE_DEBUG, "completed read capacity, slength %d, missing %d\n",
		slength, missing);
	if (missing > 0)
		slength -= missing;
	if (slength >= 8) {
#ifdef K26
		lun = related_command->SCpnt->device->lun;
#else
		lun = related_command->SCpnt->lun;
#endif
		sess->capacity_lba[lun] = ntohl(
			*(__u32 *)related_command->SCpnt->request_buffer);
		sess->capacity_len[lun] = ntohl(
			*(__u32 *)(related_command->SCpnt->request_buffer + 4));
		printk("capacity for lun %u [%u, %u]\n", lun,
			sess->capacity_lba[lun],
			sess->capacity_len[lun]);
	}
}

/* we just successfully completed a read block limits command */
void
save_limits(struct session *sess, struct command *related_command, int missing)
{
	int slength;
	__u32 lun;

	slength = related_command->SCpnt->request_bufflen;
	TRACE(TRACE_DEBUG, "completed read block limits, slength %d, "
		"missing %d\n", slength, missing);
	if (missing > 0)
		slength -= missing;
	if (slength >= 6) {
#ifdef K26
		lun = related_command->SCpnt->device->lun;
#else
		lun = related_command->SCpnt->lun;
#endif
		sess->limit_max[lun] = ntohl(
			(*(__u32 *)related_command->SCpnt->request_buffer))
			& MASK_24_BITS;
		sess->limit_min[lun] = ntohs(
		       (*(__u16 *)(related_command->SCpnt->request_buffer + 4)))
		       & MASK_24_BITS;
		printk("block limits for lun %u [%u, %u]\n", lun,
			sess->limit_min[lun],
			sess->limit_max[lun]);
	}
}

int
send_data_ack_snack(struct connection *current_connection,
		    struct command *related_command, __u32 increment)
{
	int retval = 1;
	struct init_error_rec err_rec;
	struct iscsi_targ_scsi_data_in *header;

	/* time to send a snack because of previous error */
	TRACE(TRACE_ISCSI,
	     "Sending SNACK type %u, ITT %u, offset %u, begrun %u, "
	     "runlen %u\n", DATA_R2T_SNACK,
	     related_command->init_task_tag,
	     related_command->cmd_error_offset,
	     related_command->cmd_error_begrun,
	     related_command->cmd_error_runlen);
	related_command->cmd_error_state = CMD_USING_SNACK;

	/* save fields from current command so they can be restored after
	 * retransmission (only when current command is DataIn with A bit set)
	 */
	if (increment == 0) {
		/* cause by a DataIn with A bit set, save stuff */
		related_command->abit_offset = related_command->data_offset;
		related_command->abit_in_sn = related_command->data_in_sn;

		header = (struct iscsi_targ_scsi_data_in *)
						current_connection->rx_buf;
		related_command->abit_ttt = header->target_xfer_tag;
		related_command->abit_lun = header->lun;
	}

	/* reset "normal" fields in current command to handle retransmission */
	related_command->data_offset = related_command->cmd_error_offset;
	related_command->data_in_sn = related_command->cmd_error_begrun;

	/* MAYBE undo previous increment of StatSN as we do not want to
	 * acknowledge correct receipt of all the DataIns yet.
	 * Save the increment so we can apply it again when all DataIns are ok
	 */
	current_connection->exp_stat_sn -= increment;
	current_connection->increment = increment;

	err_rec.curr_conn = current_connection;
	err_rec.related_cmd = related_command;
	err_rec.err_type = SEQUENCE_ERR;
	if (!init_send_snack(&err_rec, related_command->cmd_error_begrun,
			   related_command->cmd_error_runlen, DATA_R2T_SNACK)) {
		/* SNACK not sent for some reason, escalate to session recovry*/
		retval = init_session_recovery(&err_rec);
	}
	return retval;
}

/*
 * executed only by rx thread.
 * process the common parts of a SCSIResponse PDU and a DataIn
 * PDU with the S-bit set ("phase collapse")
 * the current_session->sess_lock MUST be held by the calling thread.
 * Returns >0 on success, session lock is locked
 *	   =0 on failure, session lock is locked
 *	   <0 on failure, session lock is NOT locked
 */
int
do_scsi_response(struct connection *current_connection,
		 struct command *related_command,
		 struct iscsi_targ_scsi_rsp *header,
		 int received,
		 __u8 * buffer)
{
	int missing, slength, retval = 1;
	__u32 residual;

	TRACE(TRACE_ENTER_LEAVE, "Enter do_scsi_response\n");

	if (unlikely(related_command->cmd_error_state == CMD_NEEDS_SNACK)) {
		/* time to send a snack because of previous error */
		retval = send_data_ack_snack(current_connection,
					     related_command, 1);
		goto out;
	}

	residual = ntohl(header->resid);

	/* see if initiator and target agree on amount of data transfered */
	if (unlikely(related_command->r2t_range_list.next != NULL)) {
		/* Command with DataSequenceInOrder==No */
		/* expected range of any command is: offset=0, limit=bufflen */
		related_command->r2t_range_list.offset = 0;
		related_command->r2t_range_list.limit
		    = related_command->SCpnt->request_bufflen;
		missing =
		    check_range_list_complete(&related_command->r2t_range_list);
	} else {
		/* Command with DataSequenceInOrder==Yes */
		missing = related_command->SCpnt->request_bufflen
				- related_command->data_offset;
	}

	if (unlikely(header->flags & U_BIT)) {
		/* U bit is set, indicating underflow, check residual count */
		TRACE(TRACE_ISCSI, "Underflow: Residual count %u\n", residual);
		if (missing != residual && !(header->status & STATUS_MASK)) {
			TRACE_ERROR("Underflow Residual count %u doesn't match "
				    "expected %d\n", residual, missing);
		}
	} else if (unlikely(header->flags & O_BIT)) {
		/* O bit is set, indicating overflow, check residual count */
		TRACE(TRACE_ISCSI, "Overflow: Residual count %u\n", residual);
		if (!(header->status & STATUS_MASK)) {
				if (missing != 0) {
				TRACE_ERROR
			    	("Overflow bit set, but still missing %d "
				"bytes\n", missing);
			} else if (residual == 0) {
				TRACE_ERROR
			    	("Overflow bit set, but Residual count is 0\n");
			}
		}
	} else if (unlikely(residual != 0)) {
		TRACE_ERROR
		    ("Neither Overflow nor Underflow set, but Residual count is"
		     " %u\n", residual);
	}

	/* pass the final status back up to the SCSI Mid-level
	 * the "result" field of the command contains 4 independent bytes:
	 * byte 0: status byte returned from the device itself in bits 1 thru 5
	 * byte 1: status returned from the host adapter (always 0 for us)
	 * byte 2: status returned by the low-level driver
	 * byte 3: "simply the message byte that comes back" whatever that is.
	 *
	 * The byte 2 errors are defined in /usr/src/linux/drivers/scsi/scsi.h
	 * and are symbols of the form DID_xxx.  DID_OK is conveniently 0.
	 */
	related_command->SCpnt->result = header->status;

	if (unlikely(header->status & STATUS_MASK)) {
		TRACE_ERROR("%s scsi response error from target, "
			    "status 0x%02x\n",
			    current->comm, header->status);

		/* mark this as an internal error of some sort */
		related_command->SCpnt->result |= (DID_ERROR << 16);

		/* we will never transmit anything else for this command */
		related_command->cmd_state |= CMD_STATE_TXDONE;

		/* Linux defines CHECK_CONDITION as 0x01,
		 * and SAM-2 requires it to be 0x02 (see Draft 20,
		 * Section 10.4.2).  This is because Linux does not
		 * consider the low order bit to be part of the status,
		 * since SAM-2 requires that bit to be 0.
		 * The status_byte() macro in linux/drivers/scsi/scsi.h
		 * handles this.  See also linux/include/scsi/scsi.h.
		 */

		if (likely(status_byte(header->status) == CHECK_CONDITION)) {
			/* Draft 20, Section 10.4.7 Data Segment -
			 * Sense and Response Data Segment "iSCSI
			 * targets MUST support and enable autosense.
			 * If Status is CHECK CONDITION (0x02),
			 * then the Data Segment MUST contain sense
			 * data for the failed command."
			 */
			slength = 0;
			if (received <= 0 || buffer == NULL) {
				TRACE_ERROR
				    ("CHECK CONDITION but no sense data\n");
			} else {
				if (received < 2) {
					TRACE_ERROR("%d data bytes not enough "
						    "to contain sense length\n",
						    received);
				} else {
					/* big-endian 16-bits */
					slength = (buffer[0] << 8) + buffer[1];
					if (slength > received - 2) {
						TRACE_ERROR("%d sense data "
						"bytes but only %d actually "
						"received\n", slength,
						received - 2);
						slength = received - 2;
					}
					if (slength > SCSI_SENSE_BUFFERSIZE) {
						if (TRACE_TEST(TRACE_ISCSI)) {
						      TRACE_WARNING("received "
						      "%d sense data bytes but "
						      "midlevel buffer holds "
						      "%d\n",
						      slength,
						      SCSI_SENSE_BUFFERSIZE);
						}
						slength = SCSI_SENSE_BUFFERSIZE;
					}
					if (slength > 0) {
						/* have some sense data,
						 * copy it into buffer
						 */
						memcpy(related_command->SCpnt->
						       sense_buffer, &buffer[2],
						       slength);
						/* Note -- we expect the
						 * midlevel to have zeroed the
						 * sense buffer before calling
						 * queuecommand().  Therefore,
						 * we do not have to zero the
						 * unfilled part of the buf
						 * (and do not have to do
						 * anything if sense data is
						 * missing completely).
						 */
						/* mark that driver sense
						 * data is present
						 */
						related_command->SCpnt-> result
							|= (DRIVER_SENSE << 24);

						/* print the sense data */
						printk("%d bytes of sense "
						       "data\n", slength);
						dump_buffer(buffer + 2,
							    slength);
					}
				}
			}
		}
	} else if (related_command->header_opcode == ISCSI_INIT_SCSI_CMND
			&& related_command->SCpnt != NULL
			&& related_command->SCpnt->use_sg == 0) {
		if (related_command->iscsi_cmd.cdb[0] == INQUIRY)
			save_inquiry(current_connection->my_session,
					related_command, missing);
		else if (related_command->iscsi_cmd.cdb[0] == READ_CAPACITY)
			save_capacity(current_connection->my_session,
					related_command, missing);
		else if (related_command->iscsi_cmd.cdb[0] == READ_BLOCK_LIMITS)
			save_limits(current_connection->my_session,
					related_command, missing);
	}

	/* no errors on this command or done with recovery */
	related_command->cmd_state |= CMD_STATE_RXDONE;
	free_pending_command(related_command, current_connection);

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave do_scsi_response, retval %d\n", retval);
	return retval;
}

/*
 * executed by init_recovery thread, scsi midlevel process.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called to allocate space for a struct command and then initialize it.
 * Returns pointer to newly allocated and initialized command if all ok,
 * else NULL.
 */
struct command *
setup_command(struct scsi_cmnd * Cmnd, struct connection *conn)
{
	struct command *new_command;
	struct session *sess = conn->my_session;
	struct list_head *lptr;

	TRACE(TRACE_ENTER_LEAVE, "Enter setup_command\n");

	if (unlikely(list_empty(&sess->free_commands))) {
		TRACE_WARNING("setup_command allocating %d bytes for "
			      "%uth command in session\n",
			      sizeof(struct command),
			      sess->n_commands_alloced+1);
		new_command = (struct command *)
				my_kmalloc(sizeof(struct command), "command");
		if (unlikely(new_command == NULL))
			goto out;
		sess->n_commands_alloced++;
	} else {
		lptr = sess->free_commands.next;
		list_del(lptr);
		new_command = list_entry(lptr, struct command, link);
	}

	/* setup the new command struct (most fields 0/NULL)
	 *
	 * Draft 20, Section 10.7.5 DataSN "For input (read) or
	 * bidirectional Data-In PDUs, the DataSN is the
	 * input PDU number within the data transfer for the
	 * command identified by the Initiator Task Tag."
	 *
	 * Draft 20, Section 10.8.2 R2TSN
	 * "R2TSN is the R2T PDU input PDU number within
	 * the command identified by the Initiator Task Tag."
	 */

	/* zero out just the first part of the struct command */
	memset(new_command, 0, (long)&((struct command *)NULL)->SCpnt);

	new_command->SCpnt = Cmnd;
	INIT_LIST_HEAD(&new_command->r2t_cookies);

	/* set up the transmit i/o vector for the iscsi header */
	setup_header_iovec(conn, new_command);
#ifdef K26
	new_command->pindex = 0;
	new_command->padding = 0;
#endif

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave setup_command, command %p\n",
	      new_command);
	return new_command;
}

/*
 * executed only by rx thread.
 * called only from recv_data_in_data()
 * There is a Data Digest err, handle appropriately - SAI
 */
static int __attribute__ ((no_instrument_function))
recv_data_in_error (struct connection *current_connection,
		    struct command *related_command,
		    struct iscsi_targ_scsi_data_in *data_in_header,
		    __u32 *size)
{
	int received;
	struct init_error_rec err_rec;

#ifdef ISCSI_STATS
	current_connection->my_session->sess_stats->iscsi_ssn_digest_err++;
#endif
	err_rec.curr_conn = current_connection;
	err_rec.related_cmd = related_command;
	err_rec.err_type = PAYLOAD_DIGERR;
	received = init_do_error_recovery(&err_rec);
	if (unlikely(received <= 0))
		goto out;
	/* force the expected data offset to the one received in order 
	 * to avoid a cascade of error messages 
	 */
	if (likely(related_command != NULL)) {
		related_command->data_offset += *size;
		TRACE(TRACE_DEBUG, "Bump offset by %d, received %d\n", *size,
			received);
		/* do not count this data as having been received yet */
		*size = 0;
	}
out:
	return received;
}

/*
 * executed only by rx thread.
 * the session->sess_lock MUST be held by the calling process/thread.
 * Called by rx_data(), recv_ignore_data().
 * Called to receive *size bytes of data (can be 0) into a scsi buffer
 * pointed to by the related_command (can be NULL),
 * plus over_flow_data_size bytes of data into a locally allocated
 * buffer.  All the data is padded and checksummed as necessary.
 * On entry MUST have (*size + over_flow_data_size) > 0.
 * Called under protection of session->lock.
 * Returns	> 0 if ok,
 * 		= 0 on err and session->sess_lock still held,
 * 		< 0 on err and session->sess_lock not held.
 * ptr to local buffer is set on return if allocated, else unchanged.
 * if returned not NULL, this buffer must be freed by caller!
*/
int
recv_data_in_data(struct connection *current_connection,
		  struct command *related_command,
		  struct iscsi_targ_scsi_data_in *data_in_header,
		  __u32 *size,
		  int over_flow_data_size,
		 char **over_flow_buffer)
{
	struct scatterlist *sg;
	__u8 *cur_ptr, extra_data_buffer[16];
	__u32 cur_offset, recv_checksum, checksum;
	int received = 1, count, cur_size, padbytes, nbytes;
	struct iovec *iov;
#ifdef K26
	struct session *current_session = current_connection->my_session;
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter recv_data_in_data, size %u, "
	      "over_flow_data_size %d\n", *size, over_flow_data_size);

	iov = current_connection->rx_iov;	/* point at empty rx_iov[] */
	count = 0;				/* no. of slots used in rx_iov*/

	if ((nbytes = *size) > 0) {
		/* Check if we have scatter-gather list to be filled */
		if (related_command->SCpnt->use_sg == 0) {
			/* No scatter-gather list, just 1 big buffer */
			TRACE(TRACE_DEBUG, "No Sc-Gather list\n");
			iov->iov_base = related_command->SCpnt->request_buffer
					+ data_in_header->offset;
			iov->iov_len = nbytes;
			iov++;
			count++;
		} else {
			/* We have to fill scatter-gather list */

			TRACE(TRACE_DEBUG, "Gather size %d\n",
					related_command->SCpnt->use_sg);
			sg = (struct scatterlist *)
			     related_command->SCpnt->request_buffer;
			cur_offset = 0;
/* donnelly */
#ifdef K26
/*	Notes on kernel 2.6 changes.
 *      Prior to 2.6, a device driver had to know if the memory
 *      model they existed under was either hi-mem or small
 *      and use the appropriate vaddr mapping of kmap() ( for large )
 *      or  page_address() ( for small) . The only recommended way 
 *	now is to allow kmap() to set up the appropriate mappings 
 * 	avoid page_adress all together. 
 */

			TRACE(TRACE_DEBUG,
				"sg %p page %p off %x addr %x len %x\n",
				sg,sg->page,sg->offset, sg->dma_address,
				sg->length);

			/* Convert page struct passed in SG instead of using
			 * pci_map_sg(NULL,sg,SCpnt->use_sg,
			 *      SCpnt->sc_data_direction)
			 * or page_address for portability between memory 
			 * models to create a kernel vaddr. 
			 * We can sleep on this
			 * call so have have to release our lock.
			 * kmap can sleep 
       			 * kmap_atomic() won't sleep but can fail 
			 */

			UNH_UNLOCK(&current_session->sess_lock,
				   current_connection->rx_lock_flags);

			cur_ptr = kmap(sg->page);
			sg_dma_address(sg) = (dma_addr_t)cur_ptr;

			UNH_LOCK(&current_session->sess_lock,
				 current_connection->rx_lock_flags);
			sg_dma_address(sg) += sg->offset;
			cur_ptr += sg->offset;

#else
			cur_ptr = sg->address;
#endif

			/* first find where to put data in the scatterlist */

#ifdef K26
			TRACE(TRACE_DEBUG, "cur_offset %d data_in_off %d sg len %d \n cur_ptr %p dma_addr %x\n",
				cur_offset, data_in_header->offset,
				sg_dma_len(sg), cur_ptr, sg_dma_address(sg));

			while (cur_offset < data_in_header->offset) {
				if (cur_offset + sg_dma_len(sg) >
					data_in_header->offset) {
					cur_ptr = (char *)sg_dma_address(sg) +
							data_in_header->offset -
							cur_offset;
					cur_offset = data_in_header->offset;
			TRACE(TRACE_DEBUG, "In while: cur_offset %d data_in_off %d sg len %d \n cur_ptr %p dma_addr %x\n",
				cur_offset, data_in_header->offset,
				sg_dma_len(sg), cur_ptr, sg_dma_address(sg));

				} else {
					cur_offset += sg_dma_len(sg);
					sg++;
					cur_ptr = (void *)sg_dma_address(sg);

			TRACE(TRACE_DEBUG, "In else: cur_offset %d data_in_off %d sg len %d \n cur_ptr %p dma_addr %x\n",
				cur_offset, data_in_header->offset,
				sg_dma_len(sg), cur_ptr, sg_dma_address(sg));

				}
			} /* while */
#else
			TRACE(TRACE_DEBUG, "cur_offset %d data_in_off %d sg len %d \n cur_ptr %p dma_addr %p\n",
				cur_offset, data_in_header->offset,
				sg->length, cur_ptr, sg->address);

			while (cur_offset < data_in_header->offset) {
				TRACE(TRACE_DEBUG,
				      "cur_offset while %d %d %d %p %p\n",
				      cur_offset, data_in_header->offset,
				      sg->length, cur_ptr, sg->address);
				if (cur_offset + sg->length
						> data_in_header->offset) {
					cur_ptr = cur_ptr
						  + data_in_header->offset
						  - cur_offset;
					cur_offset = data_in_header->offset;
				} else {
					cur_offset += sg->length;
					sg++;
					cur_ptr = sg->address;
				}
			}
#endif
			/* fill in the buffers till we receive amount of data
			 * corresponding to the datasegmentlength in the
			 * DataIn header
			 */
			for (received = 0; received < nbytes;
			     received += cur_size) {

/* jpd */
#ifdef K26
				cur_size = (__u32) sg_dma_address(sg) +
						sg_dma_len(sg) -
						(__u32) cur_ptr;

#else
				cur_size = (__u32) sg->address + sg->length
							- (__u32) cur_ptr;
#endif
				if (cur_size > nbytes - received) {
					cur_size = nbytes - received;
				}
				iov->iov_len = cur_size;
				iov->iov_base = cur_ptr;
				sg++;
/* jpd */
#ifdef K26
				cur_ptr = (__u8 *)sg_dma_address(sg);
#else
				cur_ptr = sg->address;
#endif
				iov++;
				count++;
			}
		}
	} /* end of  if (nbytes > 0) */

	nbytes += over_flow_data_size;

	/* may need to read padding and/or crc and/or extra data */
	padbytes = (-nbytes) & 3;
	if (unlikely(padbytes != 0)) {
		/* have to receive pad bytes */
		over_flow_data_size += padbytes;
		nbytes += padbytes;
	}

	if (unlikely(over_flow_data_size > 0)) {
		/* have to receive extra data and/or padding */
		if (over_flow_data_size <= sizeof(extra_data_buffer)) {
			/* use local buffer for small amounts of extra data */
			iov->iov_base = extra_data_buffer;
		} else {
			/* have to allocate a big buffer for extra data */
			*over_flow_buffer = (char *)my_kmalloc(
							over_flow_data_size,
							"local buf");
			if (unlikely(*over_flow_buffer == NULL)) {
				received = -ENOMEM;
				goto out_unlock;
			}
			iov->iov_base = *over_flow_buffer;
		}
		iov->iov_len = over_flow_data_size;
		iov++;
		count++;
	}

	if (current_connection->connection_flags & USE_DATADIGEST) {
		/* receive crc into local variable for checking */
		iov->iov_base = &recv_checksum;
		iov->iov_len = CRC_LEN;
		nbytes += CRC_LEN;
		iov++;
		count++;
		/* need to compute the checksum after reading,
		 * but recvmsg destroys the iovector during
		 * reading, so we need to make a copy now
		*/
		memcpy(current_connection->rx_iov_copy,
		       current_connection->rx_iov,
		       count * sizeof(struct iovec));
	}

	received = recv_iovector(current_connection, nbytes, count);
	if (unlikely(received <= 0))
		goto out;

	if (current_connection->connection_flags & USE_DATADIGEST) {
		/* need to go back to compute and check the checksum */
		checksum = 0;
		iov = current_connection->rx_iov_copy;
		for (count--; count > 0; count--) {
			do_crc(iov->iov_base, iov->iov_len, &checksum);
			iov++;
		}
		if (unlikely(checksum != recv_checksum)) {
			/* There is a Data Digest err,
			 * handle appropriately - SAI
			 */
			TRACE_ERROR("opcode 0x%02x, ITT %u, Data CRC wrong: "
				    "got 0x%08x, expected 0x%08x\n",
				    data_in_header->opcode,
				    related_command->init_task_tag,
				    ntohl(recv_checksum),
				    ntohl(checksum));
			data_in_header->length = 0;	/*already got all data*/
			received = recv_data_in_error(
						current_connection,
						related_command,
						data_in_header,
						size);
		}
	}

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave recv_data_in_data, received %d\n",
		received);
	return received;

out_unlock:
	UNH_UNLOCK(&current_connection->my_session->sess_lock,
		   current_connection->rx_lock_flags);
	goto out;
}

/*
 * the session->sess_lock MUST be held by the calling process/thread.
 * here to attach cmnds to the list after they have been properly
 * filled in.  The reason we needed to be a bit careful is that other
 * portions of the driver might be attempting to traverse the list
 * at the same time that we are here so we don't want to attach an
 * invalid entry.
 */
void
attach_pending_command(struct command *new_command,
		       struct connection *current_connection, int need_itt)
{
	struct iscsi_init_scsi_cmnd *iscsi_cmd;
	struct session *current_session;

	TRACE(TRACE_ENTER_LEAVE,
	      "Enter attach_pending_command, command %p, need_itt %d\n",
	      new_command, need_itt);

	/*  extract a copy of pdu's opcode and flags for use after sendmsg */
	/*  These bytes are in the "iscsi_cmd" field of the "struct command"
	   for all pdu types except DataOut when they are in the "data"
	   field of the "struct command"
	 */
	new_command->header_opcode =
	    ((__u8 *) (new_command->tx_iov[0].iov_base))[0];
	new_command->header_flags =
	    ((__u8 *) (new_command->tx_iov[0].iov_base))[1];

	current_session = current_connection->my_session;
	iscsi_cmd = &new_command->iscsi_cmd;

	/* now allocate a session-wide command number, and task tag if needed */
	if (need_itt >= 0) {
		switch (need_itt) {
		case 1:
			/* set ITT to unique value != 0xffffffff */
			new_command->init_task_tag
					= current_session->cur_task_tag++;
			if (unlikely(current_session->cur_task_tag == ALL_ONES))
				current_session->cur_task_tag++;
			iscsi_cmd->init_task_tag
					= htonl(new_command->init_task_tag);
			break;
		case 0:
			/* set ITT to 0xffffffff, and I bit to 1
			 *
			 * RFC 3721 Section 10.18.1 Initiator Task Tag (NOP-Out)
			 * "If the Initiator Task Tag contains 0xffffffff, the
			 * I bit MUST be set to 1 and the CmdSN is not advanced
			 * after this PDU is sent."
			 */
			iscsi_cmd->init_task_tag = new_command->init_task_tag
								= ALL_ONES;
			iscsi_cmd->opcode |= I_BIT;
			break;
		case 2:
			/* reusing old command struct, so ITT already set */
			list_del(&new_command->link);
			new_command->header_complete = 0;
			break;
		}	/* switch */

		/*
		 *  RFC 3720 Section 3.2.2.1 Command Numbering and Acknowledging
		 *
		 * "Commands meant for immediate delivery are marked with an
		 * immediate delivery flag; they MUST also carry the
		 * current CmdSN.  CmdSn does not advance after a command marked
		 * for immediate delivery is sent.  Command numbering starts
		 * with the first login request on the first connection of a
		 * session (the leading login on the leading connection) and
		 * command numbers are incremented by 1 for every non-immediate
		 * command issued afterwards.
		 * ...
		 * For the numbering mechanism, the initiator and target
		 * maintain the following three variables for each session:
		 *
		 * - CmdSN - the current command Sequence Number, advanced
		 * by 1 on each command shipped except for commands marked for
		 * immediate delivery.  CmdSN always contains the number to be
		 * assigned to the next Command PDU."
		 *
		 * RFC 3720 Section 10.2.1.1 I (bit in Basic Header Segment)
		 * "For request PDUs, the I bit set to 1 is an immediate
		 * delivery marker."
		 */
		iscsi_cmd->cmd_sn = htonl(current_session->cur_cmd_sn);
		if (likely(!(iscsi_cmd->opcode & I_BIT))) {
			/* not an immediate command, advance CmdSN by 1 */
			current_session->cur_cmd_sn++;
		} /* else I bit is set (immediate command), CmdSN unchanged */
	}

	TRACE(TRACE_DEBUG, "attach_pending command %p connection %p ###\n",
	      new_command, current_connection);

	/* fill in fields for tx thread */
	new_command->tx_sent_so_far = 0;       /* nothing sent so far */
	new_command->retransmit_tx_size = new_command->tx_size;

	/* first time command can be sent */
	new_command->tx_wait_to_send = jiffies;
	new_command->time_stamp = jiffies;

	/* add this new command to end of pending commands list */
	list_add_tail(&new_command->link,&current_connection->pending_commands);

	TRACE(TRACE_DEBUG, "Attach pending command, ITT %u\n",
	      new_command->init_task_tag);

	/* signal the tx thread for this session that is has work to do */
	if (atomic_read(&current_connection->tx_sem.count) <= 0) {
		TRACE(TRACE_ISCSI_FULL, "wake up tx_thread\n");
		up(&current_connection->tx_sem);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave attach_pending_command\n");
}
