/*	initiator/iscsi_initiator.c
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This file contains the core functions for iscsi initiator code that are
 *	responsible for connecting and maintaining sessions between initiator
 *	and target. The object file generated is the loadable kernel module.
 *
 *	Copyright (C) 2001-2004 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote 
 *	products derived from this software without specific prior 
 * 	written permission.
*/

#include "iscsi_initiator.h"
#include "initiator_utilities.h"
#include "initiator_error_rec.h"
#include "initiator_proc_iface.h"
#include "../common/lun_packing.h"
#include <linux/reboot.h>

/* max number of redirects to follow during login on any connection */
#define	MAX_REDIRECTION_RETRIES	2

/* Holds pointer to private data part of Scsi_Host structure.
 * This is set up once, by iscsi_detect()
 * which is called when the initiator module is loaded
 */
struct iscsi_hostdata *global_hostdata = NULL;

/* Holds pointer to the registered Scsi_Host */
struct Scsi_Host *global_host = NULL;

/* Used for all access to the shared structures, and during the operation
 * of any single thread to prevent things from disappearing from under
 * the thread.
*/

/* dummy that always returns false */
int
target_in_use(__u32 id)
{
return 0;
}

UNH_LOCK_TYPE host_data_lock = UNH_LOCK_UNLOCKED;

/* Mike Christie, mikenc@us.ibm.com */

static int iscsi_initiator_halt(struct notifier_block *nb, ulong event,
				void *buf);

/* Mike Christie, mikenc@us.ibm.com */
/* reboot notifier */
static struct notifier_block iscsi_initiator_notifier = {
		iscsi_initiator_halt, NULL, 0
};

/*
 * either the session->sess_lock MUST be held by the calling process/thread
 * or the session structure must not be active yet.
 * increase maximum number command structures session may use to:
 * 	(MAX_COMMANDS_PER_LUN + 1) * n_luns_in_session
 * (one extra in case of task-management and/or nopout commands)
 * Returns 1 on success, 0 on failure.
 */
static int __attribute__ ((no_instrument_function))
preallocate_commands(struct session *sess)
{
	__u32 limit;
	struct command *new_command;

	limit = (MAX_COMMANDS_PER_LUN + 1) * sess->n_luns_in_session;
	while (sess->n_commands_alloced < limit) {
		new_command = (struct command *)my_kmalloc(
					sizeof(struct command), "command");
		if (unlikely(new_command == NULL)) {
			return 0;
		}
		/* each new command goes at end of session's free list */
		endup_command(new_command, sess);
		sess->n_commands_alloced++;
	}
	TRACE(TRACE_DEBUG,"%u commands now allocated to session %p target %u\n",
		sess->n_commands_alloced, sess, sess->scsi_target_id);
	return 1;
}

/*
 * the session->sess_lock MUST be held by the calling process/thread.
 * reduce maximum number command structures session may use to:
 * 	(MAX_COMMANDS_PER_LUN + 1) * n_luns_in_session
 */
static void __attribute__ ((no_instrument_function))
unallocate_commands(struct session *sess)
{
	__u32 limit;
	struct command *old_command;
	struct list_head *lptr;

	limit = (MAX_COMMANDS_PER_LUN + 1) * sess->n_luns_in_session;
	while (sess->n_commands_alloced > limit
		&& !list_empty(&sess->free_commands)) {
		lptr = sess->free_commands.next;
		list_del(lptr);
		old_command = list_entry(lptr, struct command, link);
		my_kfree((void **)&old_command, "command");
		sess->n_commands_alloced--;
	}
	TRACE(TRACE_DEBUG,"%u commands now allocated to session %p target %u\n",
		sess->n_commands_alloced, sess, sess->scsi_target_id);
}

/* initialize all fields in struct iscsi_targetdata for one target */
void
init_targetdata(struct iscsi_targetdata *targ_data)
{
	/* Initialize all fields to 0 or NULL */
	memset(targ_data, 0, sizeof(struct iscsi_targetdata));

	/* now go back and initialize those that are not 0 or NULL by default */
	targ_data->isid_type = DEFAULT_ISID_TYPE;
	targ_data->isid_number = DEFAULT_ISID_NUMBER;
	targ_data->isid_qualifier = DEFAULT_ISID_QUALIFIER;

	targ_data->auth_parameter.chap_local_ctx = CHAP_InitializeContext();
	targ_data->auth_parameter.chap_peer_ctx = CHAP_InitializeContext();
	targ_data->auth_parameter.srp_ctx = SRP_InitializeContext();

	/* Initialize parameter list to a copy of default set of parameters */
	param_tbl_init(targ_data->param_tbl);
}

/* uninitialize fields in struct iscsi_targetdata that need freeing */
void
uninit_targetdata(struct iscsi_targetdata *targ_data)
{
	param_tbl_uncpy(targ_data->param_tbl);

        SRP_FinalizeContext(targ_data->auth_parameter.srp_ctx);
	CHAP_FinalizeContext(targ_data->auth_parameter.chap_peer_ctx);
        CHAP_FinalizeContext(targ_data->auth_parameter.chap_local_ctx);
}

/***************************************************************************
* Called from iscsi_initiator_detect() and unh_iscsi_probe().
* Initializes the hostdata structure.
* Creates the parameter table for this hostdata.
* Returns non-NULL pointer to parameter table if ok, NULL if error.
***************************************************************************/
static void __attribute__ ((no_instrument_function))
init_initiator(struct iscsi_hostdata *hostdata)
{
	int i;
	struct iscsi_targetdata *targ_data;

	TRACE(TRACE_ENTER_LEAVE, "Enter init_initiator, hostdata %p\n",
	      hostdata);

	/* Initialize all fields to 0 or NULL */
	memset(hostdata, 0, sizeof(struct iscsi_hostdata));

	/* now go back and initialize those that are not 0 or NULL by default */
	INIT_LIST_HEAD(&hostdata->session_list);

	for (i = 0, targ_data = hostdata->target_data;
		i < MAX_TARGET_IDS;
		targ_data++, i++) {
		init_targetdata(targ_data);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave init_initiator\n");
}

/***************************************************************************
* Called by the scsi subsystem when the iscsi_initiator module is loaded.
* (via the detect field in the Scsi_Host_Template driver_template)
* This allows the iscsi_initiator driver code to register itself with the
* SCSI Mid-level as being the driver for this type of host adapter.
* It registers space for iscsi_hostdata struct with the SCSI Mid-level.
* When called, io_request_lock is held by the midlevel caller.
* It returns:
* 	n > 0 if ok, where n is number of host adapters of this type
*		on this system (all will be driven by this code);
* 	n < 0 on error.
***************************************************************************/
static int
iscsi_initiator_detect(struct scsi_host_template * tmpt)
{
	int retval = 1;
	struct Scsi_Host *host = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_detect, tmpt %p\n",
	      tmpt);

	/* scsi_register() is defined in linux/drivers/scsi/hosts.c */
	host = scsi_register(tmpt, sizeof(struct iscsi_hostdata));

	TRACE(TRACE_DEBUG, "Back from scsi_register, host %p\n", host);

	if (unlikely(host == NULL)) {
		retval = -1;
		goto out;
	}

	TRACE(TRACE_ISCSI, "registered UNH iSCSI host number %d\n",
	      host->host_no);

	/* setup the security key hash table */
	setup_security_hash_table();

	/* three standard default values which can be overridden:
	 * host->max_channel = 0;
	 * host->max_id = 8;
	 * host->max_lun = 8;
	 */

	/* set the number of allowed "targets" this adaptor "card" can handle.
	*  the "target" numbers will be [0..max_id-1].
	*   each "target" will have LUN numbers in [0..max_lun-1].
	*/
	host->max_id = MAX_TARGET_IDS;
	host->max_lun = MAX_LUNS;

	/* set the maximum number of bytes we will accept in a scsi cdb */
	host->max_cmd_len = MAX_CDB_LEN;

	/* scsi_register() allocated space for our private structure at the end
	 *  of the host structure. This global variable is cheating!
	 *  Potentially each (physical) adaptor card can have its own hostdata.
	 *  We get away with having a single global variable here because
	 *  we know we will never have more than 1 adaptor of this type ever.
	*/

	global_hostdata = (struct iscsi_hostdata *) host->hostdata;
	global_host = host;

	/* Initialize the adapter's private data structure */
	init_initiator(global_hostdata);

	(void) initiator_register_device();

	/* Mike Christie, mikenc@us.ibm.com */
	register_reboot_notifier(&iscsi_initiator_notifier);

	TRACE(TRACE_DEBUG, "Register iSCSI host adapter \"%s\"\n",
	      tmpt->proc_name);

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_initiator_detect, retval %d\n",
	      retval);

	return retval;
}

/* called only when session->sched_scheme is CONN_SCHED_LUN.
 * the session->sess_lock MUST be held by the calling process/thread.
 * the host_data_lock MAY be held by the calling process/thread.
 * find the connection with the fewest number of LUNs assigned to it
 * and then to assign this lun to it
 */
static void __attribute__ ((no_instrument_function))
locked_add_lun_to_sess(__u32 lun, struct session *session)
{
	__u32 min_count;
	struct connection *conn, *sofar;

	/* find the connection with the fewest luns assigned to it */
	min_count = MAX_LUNS;
	sofar = NULL;
	for (conn = session->connection_head;
	     conn != NULL;
	     conn = conn->next) {
		if (conn->connection_state == CONNECTION_FULL_FEATURE_PHASE
			&& min_count > conn->assigned_lun_count) {
			min_count = conn->assigned_lun_count;
			sofar = conn;
		}
	}

	if (sofar) {
		session->lun_assignments[lun] = sofar;
		sofar->assigned_lun_count++;
		printk("Assign lun %u to connection %u, %u luns assigned\n",
			lun, sofar->connection_id,
			sofar->assigned_lun_count);
	}
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * find the connection with the fewest number of LUNs assigned to it
 * and then to assign this lun to it
 */
static void __attribute__ ((no_instrument_function))
add_lun_to_sess( __u32 lun, struct session *session )
{
	unsigned long lock_flags;

	UNH_LOCK(&session->sess_lock, lock_flags);

	/* find the connection with the fewest luns assigned to it */
	locked_add_lun_to_sess(lun, session);

	UNH_UNLOCK(&session->sess_lock, lock_flags);
}

/*
 * host_data_lock and sess->sess_lock MUST be held by the calling process/thread
 * called only when this lun is assigned to a specific connection
 * remove this lun from this session's table of assigned connections.
*/
static void __attribute__ ((no_instrument_function))
locked_drop_lun_from_sess(__u32 lun, struct session *sess, int free_inquiry)
{
	struct connection *conn;

	/* this lun was assigned to this connection, unassign it now */
	conn = sess->lun_assignments[lun];
	sess->lun_assignments[lun] = NULL;
	if (free_inquiry)
		free_inquiry_stuff(sess, lun);
	conn->assigned_lun_count--;
	printk("Unassign lun %u from connection %u, %u luns assigned\n",
		lun, conn->connection_id,
		conn->assigned_lun_count);
}

/*
 * the host_data_lock MUST be held by the calling process/thread.
 * remove this lun from this session's table of assigned connections.
 */
static void __attribute__ ((no_instrument_function))
drop_lun_from_sess(__u32 lun, struct session *session)
{
	unsigned long lock_flags;

	if (session->lun_assignments[lun]) {
		UNH_LOCK(&session->sess_lock, lock_flags);

		/* this lun was assigned to this connection, unassign it now */
		locked_drop_lun_from_sess(lun, session, 1);

		UNH_UNLOCK(&session->sess_lock, lock_flags);
	}
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * find a lun that is already in use in this session.
 * returns -1 on error, else lun number in range [0..31]
 */
int
find_used_lun(struct session *sess)
{
	int lun;

	if (unlikely(sess->lun_bits == 0)) {
		TRACE_ERROR
		    ("No LUNs currently in use by session %p target %u\n", sess,
		     sess->scsi_target_id);
		lun = -1;
	} else {
		/* at least 1 bit set in lun_bits, find the smallest
		 * numbered 1 bit
		 */
		for (lun = 0; !(sess->lun_bits & (1 << lun)); lun++) {
		}
	}

	return lun;
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Create a new recovery connection to keep the iSCSI session active.
 * Remember, iscsi requires at least one connection in a session to be active
 * for the session to be alive - SAI
 * Returns 1 on success, 0 on failure.
 */
int
create_rec_conn(struct connection *conn)
{
	__u32 connid;
	int lun, retval;
	struct connection *this_conn;
	struct session *sess = conn->my_session;
	struct iscsi_targetdata *targ_data;
	struct sockaddr *new_ip_address;
	int new_ip_length;
	unsigned long flags, lock_flags;

	TRACE(TRACE_ENTER_LEAVE, "Entering create_rec_conn\n");

	connid = conn->connection_id;
	TRACE(TRACE_DEBUG, "Only connection available is %p, id %u -"
	      "Create new connection for recovery\n", conn, connid);

	targ_data = &global_hostdata->target_data[sess->scsi_target_id];
	restore_global_data(targ_data);
	targ_data->nop_period = 0;
	targ_data->init_sess_flags = 0;
	targ_data->init_conn_flags = 0;
	global_hostdata->init_sessrec_flg = 0;

	/* find a connection id that is not yet used in this session */
      retry:
	connid++;
	for (this_conn = sess->connection_head; this_conn;
	     this_conn = this_conn->next) {
		if (this_conn->connection_id == connid) {
			/* this id already in use, try the next one */
			goto retry;
		}
	}

	/* find a lun that is already in use in this session */
	lun = find_used_lun(sess);
	if (unlikely(lun < 0)) {
		retval = 0;
		goto out;
	}

	/* got the lun, now create a new recovery session */
	if (dup_inet_struct(conn->ip_address, &new_ip_address,
							&new_ip_length) < 0
		|| create_session(sess->scsi_target_id, new_ip_address,
				new_ip_length, lun, connid, global_host) != 0) {
		TRACE_ERROR("Unable to create recovery connection\n");
		retval = 0;
		goto out;
	}

	/* if any luns were assigned to old connection, reassign them */
	if (conn->assigned_lun_count) {
		UNH_LOCK(&host_data_lock, flags);
		UNH_LOCK(&sess->sess_lock, lock_flags);
		for (lun = 0; lun < MAX_LUNS; lun++) {
			if (conn == sess->lun_assignments[lun]) {
				locked_drop_lun_from_sess(lun, sess, 0);
				locked_add_lun_to_sess(lun, sess);
			}
		}
		UNH_UNLOCK(&sess->sess_lock, lock_flags);
		UNH_UNLOCK(&host_data_lock, flags);
	}

	retval = 1;
out:
	TRACE(TRACE_ENTER_LEAVE, "Leaving create_rec_conn, retval = %d\n",
		retval);
	return retval;
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * release a socket and free the connection structure
 */
void __attribute__ ((no_instrument_function))
free_connection(struct connection *conn)
{
	char ip_string[INET6_ADDRSTRLEN+2], port_string[8];

	if (conn->sock != NULL) {
		TRACE(TRACE_NET, "Releasing socket %p\n", conn->sock);
		if (cnv_inet_to_string(conn->ip_address, ip_string,
					port_string) > 0) {
			printk("Disconnected from %s:%s\n", ip_string,
				port_string);
		}
		sock_release(conn->sock);
		TRACE(TRACE_NET, "Released socket %p\n", conn->sock);
		conn->sock = NULL;
	}
	TRACE(TRACE_NET, "Free ip_address %p\n", conn->ip_address);
	my_kfree((void **)&conn->ip_address, "ip_address");

	TRACE(TRACE_NET, "Free local_ip_address %p\n", conn->local_ip_address);
	my_kfree((void **)&conn->local_ip_address, "local_ip_address");

	TRACE(TRACE_DEBUG, "%s Free conn %p\n", current->comm, conn);
	my_kfree((void **)&conn, "connection");
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * clear disconnected and unused connections - SAI
 */
void __attribute__ ((no_instrument_function))
clear_free_conn(struct session *sess)
{
	struct connection *prev, *conn;

	if (!sess->connection_head) {
		TRACE_ERROR("No Valid connections available\n");
		return;
	}

	if (sess->nconnections <= 1
	    && sess->connection_head->connection_state==CONNECTION_RECOVERING) {
		/* No connections available - Need to create a new one - SAI */
		create_rec_conn(sess->connection_head);
	}

	for (prev = NULL, conn = sess->connection_head;
	     conn != NULL;
	     conn = conn->next) {
		if ((conn->connection_state == CONNECTION_NOT_PRESENT)) {
			TRACE(TRACE_ISCSI, "Releasing connection %p\n", conn);
			if (prev == NULL)
				sess->connection_head = conn->next;
			else
				prev->next = conn->next;

			/* Adjust the round robin connection head - SAI */
			if (conn == sess->rr_conn_head) {
				sess->rr_conn_head = prev;
			}

			free_connection(conn);
			break;
		}
		prev = conn;
	}

	return;
}

/*
 * executed by scsi midlevel process and rx thread
 * the session->sess_lock MUST be held by the calling process/thread.
 * performs round-robin scheduling by picking up at last connection we
 * scheduled and moving to next one that is in full feature phase.
 * Returns pointer to next connection to schedule, and leaves session's
 * rr_conn_head pointing to it.
 */
static struct connection * __attribute__ ((no_instrument_function))
do_round_robin(struct session *session)
{
	int nconn;
	struct connection *conn;

	if ((conn = session->rr_conn_head) == NULL
	   || (conn = conn->next) == NULL)
		conn = session->connection_head;

	for (nconn = 0; nconn < session->nconnections; nconn++) {
		if (conn->connection_state == CONNECTION_FULL_FEATURE_PHASE)
			break;

		if ((conn = conn->next) == NULL)
			conn = session->connection_head;
	}
	return session->rr_conn_head = conn;
}

/*
 * executed by scsi midlevel process and rx thread
 * the session->sess_lock MUST be held by the calling process/thread.
 * Multiple Connection Scheduling Algorithm - SAI
 */
struct connection * __attribute__ ((no_instrument_function))
get_connection(struct scsi_cmnd * Cmnd, struct session *sess)
{
	struct connection *conn;

	TRACE(TRACE_ENTER_LEAVE, "Entering get_connection\n");

	switch (sess->sched_scheme) {

	case CONN_SCHED_RR:
		/* get the next connection in the round-robin */

		if (sess->nconnections == 1)
			conn = sess->connection_head;
		else {
			conn = do_round_robin(sess);
		}
		break;

	case CONN_SCHED_LUN:
		/* get the connection assigned to this LUN */
#ifdef K26
		conn = sess->lun_assignments[Cmnd->device->lun];
#else
		conn = sess->lun_assignments[Cmnd->lun];
#endif
		break;

	default:
		/* use only one connection */
		conn = sess->connection_head;
		break;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving get_connection\n");
	return conn;
}

/*****************************************************************************
* When called, io_request_lock is held by the midlevel caller.
* This function is called by the SCSI Mid-level whenever a Read/Write Request
* has to be satisfied. This function will set up TCP buffers for sending SCSI
* Command and, if the negotiated ImmediateData is 1, ImmediateData. After
* setting up the TCP buffers, the tx thread is woken up which is waiting on
* tx_sem.
* Parameters: struct scsi_cmnd,
*	  done function which is called when any SCSI Command is satisfied.
* Return Value: 0, on success
* 		-1, on failure (any non-zero value seems to work)
*
* Note that the done() function should not be called until queuecommand returns!
*
*****************************************************************************/
static int
iscsi_initiator_queuecommand(struct scsi_cmnd * Cmnd,
			     void (*done) (struct scsi_cmnd *))
{
	int retval = 0;		/* assume this function will succeed */
	__u32 target_id;
	struct iscsi_hostdata *hostdata;
	struct session *current_session;
	unsigned long flags, lock_flags;

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	target_id = Cmnd->device->id;
#else
	target_id = Cmnd->target;
#endif

	TRACE(TRACE_ENTER_LEAVE,
	      "Enter iscsi_initiator_queuecommand, target id %u\n", target_id);

/* RDR */
#ifndef K26
	/* must not hold io_request_lock when trying to get our locks */
	spin_unlock(&io_request_lock);
#endif

	/* get exclusive access to the iscsi structures */
	UNH_LOCK(&host_data_lock, flags);

	if (unlikely(global_host == NULL || global_hostdata == NULL)) {
		TRACE_ERROR("queuecommand called after shutdown\n");
		retval = -1;
		goto out;
	}

	/* get pointer to hostdata struct registered to the SCSI Mid-level */

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	hostdata = (struct iscsi_hostdata *) Cmnd->device->host->hostdata;
#else
	hostdata = (struct iscsi_hostdata *) Cmnd->host->hostdata;
#endif
	if (unlikely(hostdata != global_hostdata)) {
		TRACE_ERROR
		    ("Hostdata struct %p not same as global_hostdata %p\n",
		     hostdata, global_hostdata);
		retval = -1;
		goto out;
	}

	/* tell Mid-level when I am done receiving the SCSI Cmnd info. */
	Cmnd->scsi_done = done;

	/* Find out the session corresponding to the target_id */
	current_session = find_session_by_id(target_id, hostdata);

	/* Check that we have finished the Login Phase
	 * with the concerned target and are now in the
	 * full feature phase of a normal session
	 */
	if (unlikely(current_session == NULL
	    || current_session->session_state != SESSION_FULL_FEATURE_PHASE
	    || current_session->connection_head == NULL
	    || current_session->connection_head->rx_thread == NULL
	    || current_session->connection_head->tx_thread == NULL
	    || current_session->oper_param->SessionType)) {

		/* Check to see if session recovery is active - SAI */
		if (global_hostdata && global_hostdata->init_sessrec_flg) {
			/* changes for session recovery new 18_04 - SAI */

			TRACE(TRACE_ISCSI, "iscsi_initiator_queuecommand : "
			      "Recovery in Progress for target %u\n",
			      target_id);

			/* create a new session recovery structure or
			 * get the already existing session recovery
			 * structure depending on whether one or
			 * multiple sessions participate in session
			 * recovery - new 18_04 SAI
			 */

			/* NOTE XXX jpd 7/18/2003
			 * this section has changed in ref20_17 compared
			 * to ref20_10 (RPM 1.2.0) .
			 * Needs investigations for lock spliting
			 * per-session. I don't know why these are commented
			 * out .
			 */
			/* create and store in a recovery session - SAI */
			set_rec_group(Cmnd);
			goto out;
		}

		printk("iscsi Target %u not logged in\n", target_id);

		Cmnd->result = DID_NO_CONNECT << 16;
		if (likely(done != NULL))
			/* should not do this before queuecommand returns! */
			done(Cmnd);
		goto out;
	}

	UNH_LOCK(&current_session->sess_lock, lock_flags);
	retval = scsi_to_iscsi(Cmnd, current_session);
	UNH_UNLOCK(&current_session->sess_lock, lock_flags);

out:
	/* release our exclusive access to the iscsi structures */
	UNH_UNLOCK(&host_data_lock, flags);

	TRACE(TRACE_ENTER_LEAVE,
	      "Leave iscsi_initiator_queuecommand, retval %d\n", retval);

/* RDR */
#ifndef K26
	/* get back midlevel caller's lock */
	spin_lock(&io_request_lock);
#endif

	return retval;
}

/*****************************************************************************
 * executed by scsi midlevel process and init_recovery thread
 * host_data_lock and sess->sess_lock MUST be held by the calling process/thread
 * This function takes the command and data from the scsi buffers, forms a new
 * iSCSI PDU with the corresponding command and data filled-in and attaches the
 * command to the end of the pending commands list for the session - SAI
 * scsi midlevel's SPIN_LOCK io_request_lock must NOT be held
*****************************************************************************/
int
scsi_to_iscsi(struct scsi_cmnd * Cmnd, struct session *current_session)
{
	int retval = 0;		/* assume this function succeeds */
	__u32 lun;
	struct iscsi_init_scsi_cmnd *iscsi_cmd;
	struct command *new_command;
	struct connection *conn;

	TRACE(TRACE_ENTER_LEAVE, "Enter scsi_to_iscsi\n");

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	lun = Cmnd->device->lun;
#else
	lun = Cmnd->lun;
#endif

	if (current_session->nconnections <= 1)
		conn = current_session->connection_head;
	else {
		/* multiple connections usage - SAI */
		conn = get_connection(Cmnd, current_session);
	}
	if (unlikely(conn == NULL)) {
		TRACE_ERROR("Session %u has no connections for lun %u\n",
			    current_session->scsi_target_id, lun);
		retval = -1;
		goto out;
	}

	/* make sure connection is still in FFP */
	if (unlikely(conn->connection_state != CONNECTION_FULL_FEATURE_PHASE)) {
		/* rx_thread gone -- peer probably reset the connection */
		Cmnd->result = DID_NO_CONNECT << 16;

		if (likely(Cmnd->scsi_done != NULL)) {
			/* can we do this before queuecommand returns?! */
			call_back_scsi_done(Cmnd);
		}
		goto out;
	}

	/* get a new struct command to send SCSI Request PDU to target */
	new_command = setup_command(Cmnd, conn);
	if (unlikely(new_command == NULL)) {
		retval = -1;
		goto out;
	}

	/* set up the SCSI_Cmnd header that is to be sent to the target */
	iscsi_cmd = &new_command->iscsi_cmd;
	iscsi_cmd->opcode = ISCSI_INIT_SCSI_CMND;

	/* SAM-2, section 4.12.2 LUN 0 address
	 * "To address the LUN 0 of a SCSI device the peripheral device
	 * address method shall be used."
	 *
	 * What this means is that whenever the LUN is 0, the full 8 bytes of
	 * the LUN field in the iscsi PDU will also be 0. Because we have
	 * already zeroed out the entire PDU header in setup_command() above,
	 * we don't need to do anything else here in this case.
	 */
	if (lun) {
		pack_lun(lun,
			 conn->connection_flags & USE_FLAT_SPACE_LUN,
			 (__u8 *)&iscsi_cmd->lun);
	}
	iscsi_cmd->xfer_len = htonl(Cmnd->request_bufflen);

	TRACE(TRACE_DEBUG, "Xfer len for queuecommand %d\n",
	      Cmnd->request_bufflen);

	/* copy the SCSI CDB into the PDU */
	memcpy(iscsi_cmd->cdb, Cmnd->cmnd, Cmnd->cmd_len);

	if (Cmnd->sc_data_direction == DMA_TO_DEVICE) {
		/* SCSI Command has write data to be sent.
		 * Mark the cmd to be built when xmit_thread wakes up in order
		 * to avoid any potential sleep hazards in
		 * build_write_command(), since we may be in an interrupt
		 * context here.
		 */
 		new_command->need_write_built |= CMD_NEEDS_WRITE_BUILT;
		TRACE(TRACE_DEBUG, "command %p marked for write\n",
			new_command);
	} else {
		/* not a write, probably a read, but check out of curiousity */

		/* always set the F_BIT if it is not a WRITE SCSI Commamnd */
		if (likely(Cmnd->sc_data_direction == DMA_FROM_DEVICE)) {
			iscsi_cmd->flags = F_BIT | R_BIT;
		} else {
			/* direction given as neither Read nor Write */
			TRACE(TRACE_DEBUG,
			      "sc_data_direction %u not READ or WRITE\n",
			      Cmnd->sc_data_direction);
			if (unlikely(Cmnd->request_bufflen != 0)) {
				TRACE_ERROR("request_bufflen %u not zero when "
					    "sc_data_direction %u not READ or "
					    "WRITE\n",
					    Cmnd->request_bufflen,
					    Cmnd->sc_data_direction);
				iscsi_cmd->flags = F_BIT | R_BIT;
			} else {
				/* no data expected to be transfered,
				 * not READ or WRITE, so only set the F bit
				 */
				iscsi_cmd->flags = F_BIT;
			}
		}
	}

	/* attach the new SCSI Request command to end of pending commands list
	 * this also sets the ITT and the CmdSN.
	 */
	attach_pending_command(new_command, conn, 1);
out:
	TRACE(TRACE_ENTER_LEAVE, "Leave scsi_to_iscsi, retval %d\n", retval);
	return retval;
}

/****************************************************************************
* This function is called by the system when the abort_timer set in
* the iscsi_initiator_abort function expires. This function wakes up
* the thread/process which is waiting on the task_mgt_sem semaphore.
****************************************************************************/
static void
abort_timer_function(unsigned long data)
{
	TRACE(TRACE_ENTER_LEAVE, "Enter abort_timer_function\n");
	up((struct semaphore *) data);
	TRACE(TRACE_ENTER_LEAVE, "Leave abort_timer_function\n");
}

static void
dump_command_info(struct command *related_command,
		  struct connection *related_connection,
		  struct session *related_session)
{
	int delta_t;
	struct r2t_cookie *cookie;

	delta_t = jiffies - related_command->time_stamp;
	printk
	    (" SCpnt %p, op 0x%02x, flags 0x%02x, time_stamp %u, delta_t %d\n",
	     related_command->SCpnt, related_command->header_opcode,
	     related_command->header_flags, related_command->time_stamp,
	     delta_t);
	printk(" recvd_length %u, data_offset %u, still_to_send %u, "
	       "wait_to_send %u, cmd_state 0x%02x\n",
	       related_command->recvd_length,
	       related_command->data_offset, related_command->still_to_send,
	       related_command->tx_wait_to_send, related_command->cmd_state);
	if (related_command->SCpnt) {
		printk( " request_bufflen %u, recv_so_far %u, "
			"data_in_sn %u, direction %c\n",
			related_command->SCpnt->request_bufflen,
			related_command->r2t_range_list.limit,
			related_command->data_in_sn,
			related_command->SCpnt->sc_data_direction
				== DMA_TO_DEVICE ? 'W' : 'R');
	}
	printk(" tx_size %u, tx_sent_so_far %u, r2t_sn %u, n_r2t_cookies %u, "
	       "msg_flags 0x%08x\n",
	       related_command->tx_size, related_command->tx_sent_so_far,
	       related_command->r2t_sn, related_command->n_r2t_cookies,
	       related_command->cmd_msg_flags);
	delta_t = related_session->max_cmd_sn
				- ntohl(related_command->iscsi_cmd.cmd_sn);
	printk(" short_command %p, max_cmd_sn %u, cmd_sn %u, delta %d\n",
	       related_connection->short_command, related_session->max_cmd_sn,
	       ntohl(related_command->iscsi_cmd.cmd_sn), delta_t);
	if (list_empty(&related_command->r2t_cookies)) {
		printk(" No R2T cookies\n");
	} else {
		list_for_each_entry(cookie,&related_command->r2t_cookies,link) {
			printk
			    ("cookie TTT %u, data_out_sn %u, r2t_xfer_length "
			     "%d, offset %u\n", ntohl(cookie->target_xfer_tag),
			     cookie->data_out_sn, (int) cookie->r2t_xfer_length,
			     cookie->offset);
		}
	}
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Searches iscsi structures for Cmnd
 * Returns 1 on success, 0 on failure to find this command
 * On success, it has filled in the last 3 parameters :
 *	session, connection, and related command
 *	with session->sess_lock LOCKED.
 * On failure, these last 3 parameters are unchanged /undefined
 * and session->sess_lock is NOT LOCKED.
 */
static int
find_aborted_command(struct scsi_cmnd * Cmnd,
		     struct session **found_session,
		     struct connection **found_connection,
		     struct command **found_command,
		     int print_on_success,
		     unsigned long *lock_flags)
{
	struct session *related_session;
	struct connection *related_connection = NULL;
	struct command *related_command;
	struct iscsi_hostdata *hostdata;
	int retval = 0;		/* assume failure */
	__u32 target_id;
	unsigned long flags;

	UNH_LOCK(&host_data_lock, flags);

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	target_id = Cmnd->device->id;
	hostdata = (struct iscsi_hostdata *) Cmnd->device->host->hostdata;
#else
	target_id = Cmnd->target;
	hostdata = (struct iscsi_hostdata *) Cmnd->host->hostdata;
#endif
	if (unlikely(hostdata != global_hostdata)) {
		TRACE_ERROR
		    ("Hostdata struct %p not same as global_hostdata %p\n",
		     hostdata, global_hostdata);
		UNH_UNLOCK(&host_data_lock, flags);
		goto out;
	}

	/* Find out the session that corresponds to this target ID */
	related_session = find_session_by_id(target_id, hostdata);

	if (unlikely(related_session == NULL)) {
		TRACE_ERROR("No session for target %u\n", target_id);
		UNH_UNLOCK(&host_data_lock, flags);
		goto out;
	}

	UNH_LOCK(&related_session->sess_lock, *lock_flags);
	UNH_UNLOCK(&host_data_lock, flags);

	/* check if the session is in full feature phase or not */
	if (unlikely(
	    related_session->session_state != SESSION_FULL_FEATURE_PHASE
	    || (related_connection = related_session->connection_head) == NULL
	    || related_connection->rx_thread == NULL
	    || related_connection->tx_thread == NULL
	    || related_connection->connection_state
	    	!= CONNECTION_FULL_FEATURE_PHASE)) {

		TRACE_ERROR("No connection to target %u\n", target_id);
		UNH_UNLOCK(&related_session->sess_lock, *lock_flags);
		goto out;
	}

	/* search for the PDU containing the SCSI command
	 * which has to be aborted
	 */
	for (; related_connection;
	     related_connection = related_connection->next) {
		list_for_each_entry(related_command,
				    &related_connection->pending_commands,
				    link) {
			if (related_command->SCpnt == Cmnd) {
				if (likely((related_command->iscsi_cmd.
						opcode & ISCSI_OPCODE)
							==ISCSI_INIT_SCSI_CMND))
					goto found_command_match;
				TRACE_ERROR
				    ("Related Cmnd %p for target %u has "
				     "opcode 0x%02x, expected 0x%02x\n", Cmnd,
				     target_id,
				     related_command->iscsi_cmd.opcode
							& ISCSI_OPCODE,
				     ISCSI_INIT_SCSI_CMND);
				UNH_UNLOCK(&related_session->sess_lock,
					   *lock_flags);
				goto out;
			}
		}
	}

	/* did not find the command to abort, dump pending cmnds */
	TRACE_ERROR("No related SCSI Cmnd %p for target %u\n", Cmnd,
		    target_id);

	for (related_connection = related_session->connection_head;
	     related_connection != NULL;
	     related_connection = related_connection->next) {
		printk
		    ("   pending cmds for connection %p, CID %u, jiffies %lu\n",
		     related_connection, related_connection->connection_id,
		     jiffies);
		list_for_each_entry(related_command,
				    &related_connection->pending_commands,
				    link) {
			dump_command_info(related_command, related_connection,
					  related_session);
		}
	}
	UNH_UNLOCK(&related_session->sess_lock, *lock_flags);
	goto out;

found_command_match:
	/* successfully found the matching command to be aborted
	 * session->sess_lock is LOCKED
	 */
	if (print_on_success) {
		TRACE_ERROR("Aborting SCSI Cmnd with ITT %u, CmndSN %u, "
			"jiffies %lu\n",
		    	related_command->init_task_tag,
			ntohl(related_command->iscsi_cmd.cmd_sn),
			jiffies);
		dump_command_info(related_command, related_connection,
				  related_session);
	}

	*found_session = related_session;
	*found_connection = related_connection;
	*found_command = related_command;
	retval = 1;

out:
	return retval;
}

/*
 * the session->sess_lock MUST be held by the calling process/thread.
 * Allocates, sets up, and sends a task management pdu to target,
 * then waits (up to a time limit) for a response from the target,
 * and finally frees the task management command structure.
 * Returns NULL in value of related_command on any error.
 */
static void
tm_send_and_wait(struct scsi_cmnd * Cmnd,
		 struct session *related_session,
		 struct connection *conn,
		 struct command **related_command,
		 unsigned long *lock_flags)
{
	struct timer_list abort_timer;
	struct command *task_mgt_command;
	struct iscsi_init_task_mgt_command *task_mgt_command_hdr;
	__u32 related_itt, lun;

	task_mgt_command = setup_command(NULL, conn);
	if (unlikely(task_mgt_command == NULL)) {
		*related_command = NULL;
		goto out;
	}

	related_itt = (*related_command)->init_task_tag;

	task_mgt_command->task_mgt_function = TMF_ABORT_TASK;

	task_mgt_command->task_mgt_response = FUNCTION_REJECTED;

	/* fill up the task_mgt_header info. -- command should be immediate */
	task_mgt_command_hdr = (struct iscsi_init_task_mgt_command *)
	    &task_mgt_command->iscsi_cmd;
	task_mgt_command_hdr->opcode = (ISCSI_INIT_TASK_MGMT_CMND | I_BIT);
	task_mgt_command_hdr->function = (TMF_ABORT_TASK | F_BIT);

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	lun = Cmnd->device->lun;
#else
	lun = Cmnd->lun;
#endif

	/* SAM-2, section 4.12.2 LUN 0 address
	 * "To address the LUN 0 of a SCSI device the peripheral device
	 * address method shall be used."
	 *
	 * What this means is that whenever the LUN is 0, the full 8 bytes of
	 * the LUN field in the iscsi PDU will also be 0. Because we have
	 * already zeroed out the entire PDU header in setup_command() above,
	 * we don't need to do anything else here in this case.
	 */
	if (lun) {
		pack_lun(lun,
			 conn->connection_flags & USE_FLAT_SPACE_LUN,
			 (__u8 *)&task_mgt_command_hdr->lun);
	}
	task_mgt_command_hdr->ref_task_tag = htonl(related_itt);

	task_mgt_command_hdr->ref_cmd_sn = (*related_command)->iscsi_cmd.cmd_sn;

	/* attach the new task mgt command to end of pending commands list
	 * this also assigns the ITT and sets the CmdSN.
	 */
	attach_pending_command(task_mgt_command, conn, 1);

	TRACE(TRACE_DEBUG, "Waiting to get task mgt. rsp from the target\n");

	/* set up timer and wait on task_mgt_sem until the timer expires */
	/* or we receive the task management response from the target */
	init_timer(&abort_timer);
	abort_timer.expires = jiffies + ABORT_TIMEOUT;
	abort_timer.data = (__u32) &conn->task_mgt_sem;
	abort_timer.function = abort_timer_function;
	add_timer(&abort_timer);

	/* release our exclusive access to the iscsi structures while waiting */
	UNH_UNLOCK(&related_session->sess_lock, *lock_flags);

	/* wait for our timer to expire */
	down(&conn->task_mgt_sem);

	del_timer_sync(&abort_timer);

	TRACE(TRACE_DEBUG, "Awake after down on task mgt. sem\n");
	TRACE(TRACE_DEBUG, "Task mgt response 0x%02x\n",
	      task_mgt_command->task_mgt_response);

	/*  we have to do all this all over again because we may have been
	   "away" for a long time and our operating environment could have
	   changed during that interval -- the session and/or connection
	   could have been closed (unlikely) or the command could have
	   completed  and been removed from the pending commands list
	   (more likely).
	 */
	if (!find_aborted_command(Cmnd, &related_session,
			&conn, related_command, 0, lock_flags)) {
		/* command must have "disappeared" while we were away! */
		/* get lock so we can free the tm command structure */
		UNH_LOCK(&related_session->sess_lock, *lock_flags);
		*related_command = NULL;
	} else {
		/* Successfully found matching command (again) to be aborted.
	 	 * Corresponding session, connection and command are valid.
	 	 * Session lock is held. ITT must be same as it was before.
	 	 */
		if (unlikely((*related_command)->init_task_tag != related_itt)){
			/* command changed while we were away! */
			TRACE_ERROR("Aborted command ITT now %u, used to be %u\n",
				    (*related_command)->init_task_tag,
				    related_itt);
			*related_command = NULL;
		}
	}

	/* free the task_mgt_command structure */
	task_mgt_command->cmd_state |= CMD_STATE_RXDONE | CMD_STATE_TXDONE;
	free_pending_command(task_mgt_command, conn);
out:
	return;
}

/***************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * This function is called by the SCSI Mid-level when it has to abort any
 * SCSI Command. This function will check if the command has been sent to the
 * target for satisfying the request. If it has been sent, then this function
 * will set up TCP buffers to send out a Task Mgt. Command.
 * When called by the SCSI Mid-level (from linux/drivers/scsi/scsi_error.c)
 * the io_request_lock (or equivalent) is held by the caller.
 *
 ****************************************************************************/
static int
iscsi_initiator_abort(struct scsi_cmnd * Cmnd)
{
	unsigned long lock_flags;
	/* pessimistic assumption */
	int retval = FAILED;
	struct session *related_session;
	struct connection *related_connection;
	struct command *related_command;
#if defined(CONFIG_ISCSI_DEBUG)
	__u32 save_trace;
#endif

	/* turn on (almost) all tracing while processing an abort */
	TRACE_GET(save_trace);
	TRACE_SET(TRACE_ALL
		    & ~(TRACE_ISCSI_FULL|TRACE_ENTER_LEAVE|TRACE_MY_MEMORY));

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_abort\n");

/* RDR */
#ifndef K26
	/* unlock the lock obtained by the scsi midlevel before calling us */
	spin_unlock(&io_request_lock);
#endif

	/* get exclusive access to the iscsi structures */

	if (!find_aborted_command(Cmnd, &related_session, &related_connection,
				  &related_command, 1, &lock_flags)) {
		goto out;
	}

	/* Successfully found the matching command to be aborted.
	 * Corresponding session, connection and command are valid.
	 * Session lock is held.  Mark this command as being aborted.
	 */

	related_command->cmd_state |= CMD_STATE_ABORTING;

	if (related_command->cmd_state & CMD_STATE_TXSTARTED) {
		/* this command was previously sent to target, so we now have
		   to send a task management function command to the target */
		tm_send_and_wait(Cmnd, related_session, related_connection,
				&related_command, &lock_flags);
		if (related_command == NULL )
			goto out_session_lock_held;
	}

	retval = SUCCESS;

	/* Free command struct for the SCSI command that had to be aborted */
	/* error recovery - SAI */
	Cmnd->result = DID_ABORT << 16;
	related_command->cmd_state |= CMD_STATE_RXDONE | CMD_STATE_TXDONE;
	free_pending_command(related_command, related_connection);
	TRACE(TRACE_DEBUG, "Freed aborted command, ITT %u\n",
			related_command->init_task_tag);

out_session_lock_held:

	UNH_UNLOCK(&related_session->sess_lock, lock_flags);

out:

/* RDR */
#ifndef K26
	/* get back lock so caller in scsi midlevel can release it! */
	spin_lock(&io_request_lock);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_initiator_abort, 0x%04x\n",
	      retval);

	/* restore tracing to what it was when we came in here */
	TRACE_SET(save_trace);

	return retval;
}

/*****************************************************************************
*	This function is called by the SCSI Mid-level.
*	(via the eh_device_reset_handler field in the
*	Scsi_Host_Template driver_template)
*	Still need to implement this function!!
*****************************************************************************/
static int
iscsi_initiator_reset(struct scsi_cmnd * Cmnd)
{
	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_reset\n");
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_initiator_reset, 0x%04x\n",
	      SUCCESS);
	return SUCCESS;
}

static int
scan_redirect_address(char *string, struct sockaddr **ip_address,
			int *ip_length)
{
	int retval;
	char *ptr;

	if (*string == '[') {
		/* start of a bracketed IPv6 address */
		if ((ptr = strchr(string, ']')) == NULL) {
			TRACE_ERROR("Illegal bracketed IPv6 address %s\n",
									string);
	    		retval = -ECONNABORTED;
		} else {
			/* see if a port was supplied or not */
			if (*(ptr+1) == ':') {
				ptr++;
			} else {
				ptr = ISCSI_WKP_STRING;
			}
			if ((retval = cnv_string_to_inet(string+1, ptr,
					ip_address, ip_length)) != AF_INET6) {
				if (retval >= 0) {
					TRACE_ERROR("Illegal bracketed IPv6 "
							"address %s\n", string);
		    			retval = -ECONNABORTED;
				}
			}
		}
	} else {
		/* must be start of an IPv4 address */
		if ((ptr = strchr(string, ':'))) {
			ptr++;
		} else {
			ptr = ISCSI_WKP_STRING;
		}
		if ((retval = cnv_string_to_inet(string, ptr, ip_address,
						ip_length)) != AF_INET) {
			if (retval >= 0) {
				TRACE_ERROR("Illegal IPv4 address %s\n",string);
				retval = -ECONNABORTED;
			}
		}
	}
	return retval;
}

/* set up the 6-byte isid from values configured for this target
 * configured values limited to 2 bits for type, 24 bits for number & qualifier
 *
 * RFC 3720 Section 10.12.5 ISID
 * T = 00b
 *	A&B - 22-bit OUI
 *	C&D - 24-bit qualifier
 * T = 01b
 *	A - 6-bit Reserved
 *	B&C - 24-bit IANA Enterprise Number
 *	D - 16-bit Qualifier
 * T = 10b
 *	A - 6-bit Reserved
 *	B&C - 24-bit Random number
 *	D - 16-bit Qualifier
 * T = 11b
 *	A, B, C, D - Reserved
 */
static void __attribute__ ((no_instrument_function))
setup_isid(struct iscsi_targetdata *targ_data, __u32 *isid)
{
	__u32 temp;
	__u16 qual, *ptr16 = (__u16 *)&isid[1];

	temp = targ_data->isid_type << 30;
	qual =	targ_data->isid_qualifier;

	switch (targ_data->isid_type) {
	case 0:	/* OUI Format */
		temp |= ((targ_data->isid_number & MASK_22_BITS)<<8)
			| ((targ_data->isid_qualifier >> 16) & MASK_8_BITS);
		break;
	case 1:	/* EN Format */
	case 2:	/* Random Format */
		temp |= targ_data->isid_number;
		break;
	case 3:	/* Reserved Format */
		qual = 0;
		break;
	}	/* switch */

	*isid = htonl(temp);
	*ptr16 = htons(qual);
}

/*****************************************************************************
*	Sets the session operation parameters and changes the session
*	state to LOGGED_IN, meaning the initiator is in done with login phase.
*	Called only from init_connection()
*	Return value:	=1 on redirection
*			=0 on any other success
*			<0 on failure
*****************************************************************************/
static int
iscsi_initiator_login(struct connection *current_connection,
		      __u32 when_called,
		      struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		      struct auth_parameter_type auth_param,
		      struct iscsi_targetdata *targ_data,
		      struct sockaddr **redirect_ip_address,
		      int *redirect_ip_length)
/* chap and srp support - CHONG */
{
	int retval = 0, cid;
	int login_task_tag;
	__u32 isid[2];
	struct session *current_session;
	__u32 pdu_version_number = 0;
	struct parameter_type *p;
#ifdef ISCSI_STATS
	unsigned long flags;
#endif

	TRACE(TRACE_ENTER_LEAVE,
	      "Enter iscsi_initiator_login, current connection %p\n",
	      current_connection);

	current_session = current_connection->my_session;

	/* set up the 6-byte isid */
	setup_isid(targ_data, isid);
	cid = current_connection->connection_id;	/* for offset 20 */

	/* initialize task tag for the whole login process */
	/* then update the session's task tag counter */
	login_task_tag = current_session->cur_task_tag++;

	TRACE(TRACE_DEBUG, "iscsi_initiator_login pdu_version_number %u\n",
	      pdu_version_number);

	/* In login phase always disable Header & Data Digests */
	current_connection->connection_flags &= ~USE_HEADERDIGEST;
	current_connection->connection_flags &= ~USE_DATADIGEST;
	current_connection->basic_hdr_len = ISCSI_HDR_LEN;

	retval = initiator_parameter_negotiate(current_connection->sock,
			p_param_tbl,
			auth_param,
			&current_session->cur_cmd_sn, login_task_tag, isid,
			&current_session->tsih,
			cid,
			&current_connection->connection_flags,
			pdu_version_number,
			when_called,
			&current_connection->exp_stat_sn,
			&current_connection->max_send_length,
			&current_connection->max_recv_length);

	if (unlikely(retval < 0)) {
			TRACE(TRACE_ISCSI, "parameter negotiation failed\n");
			/* software caused connection abort */
			retval = -ECONNABORTED;
			goto out;
	} else if (likely(retval == 0)) {
		/* new connection negotiation finished successfully */
		if (current_connection->connection_flags & USE_HEADERDIGEST)
			current_connection->basic_hdr_len += CRC_LEN;
		current_connection->connection_state = CONNECTION_LOGGED_IN;
#ifdef ISCSI_STATS
		UNH_LOCK(&host_data_lock, flags);
		global_hostdata->instance_stats.iscsi_intr_login_accept_rsp++;
		UNH_UNLOCK(&host_data_lock, flags);
#endif
	} else {
		/* have a redirection attempt */
#ifdef ISCSI_STATS
		UNH_LOCK(&host_data_lock, flags);
		global_hostdata->instance_stats.iscsi_intr_login_redirect_rsp++;
		UNH_UNLOCK(&host_data_lock, flags);
#endif
		p = find_parameter(TARGETADDRESS, p_param_tbl);
		if (unlikely(p == NULL)) {
			/* could not find TargetAddress key */
			TRACE_ERROR("Cannot find %s key\n", TARGETADDRESS);
			retval = -ECONNABORTED;
		} else if (!IS_KEY_GOT_FROM_OTHER_SIDE(p->neg_info)) {
			/* target never sent us this key */
			TRACE_ERROR
			    ("Target never sent TargetAddr key on Redirectn\n");
			retval = -ECONNABORTED;
		} else if (scan_redirect_address(p->str_value,
				redirect_ip_address, redirect_ip_length) < 0) {
			retval = -ECONNABORTED;
		}
	}

out:
#ifdef ISCSI_STATS
	if (retval == -ECONNABORTED) {
	    struct timeval tv;
	    /*
	     * Do do_gettimeofday() and find_parameter()
	     * before holding the lock to reduce the lock hold
	     * up time.
	     */
	    do_gettimeofday(&tv);
	    p = find_parameter(TARGETNAME, p_param_tbl);
	    UNH_LOCK(&host_data_lock, flags);
	    global_hostdata->instance_stats.iscsi_intr_login_failure++;
	    global_hostdata->instance_stats.iscsi_intr_last_failure_time=tv;
	    if (p != NULL && p->str_value) {
		strreplace(&(global_hostdata->instance_stats.iscsi_intr_last_tgt_failure_name),
			p->str_value);
	    }
	    if (current_connection->ip_address) {
	    	if (!global_hostdata->instance_stats.iscsi_intr_last_tgt_failure_address)
	    		global_hostdata->instance_stats.iscsi_intr_last_tgt_failure_address
				= my_kmalloc(sizeof(struct sockaddr), "failure address");
	    	if (global_hostdata->instance_stats.iscsi_intr_last_tgt_failure_address)
	    		memcpy(global_hostdata->instance_stats.iscsi_intr_last_tgt_failure_address,
			current_connection->ip_address,
			sizeof(struct sockaddr));
	    }
	    UNH_UNLOCK(&host_data_lock, flags);
	}
#endif
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_initiator_login, retval %d\n",
	      retval);

	return retval;
}

/****************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Initializes the fields of the first connection in a new session,
 * connects to a remote target, logs in to it,
 * and creates the rx_thread for that connection
 * if full feature phase has been attained successfully.
 * Called by create_session() when adding new connection to existing session
 * and by init_session() when this is first connection in a new session.
 * Return value: =1 for redirection,
 *		 =0 for any other success,
 *		 <0 for error
 ****************************************************************************/
static int
init_connection(struct connection *conn,
		__u32 when_called,
		struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		struct iscsi_targetdata *targ_data,
		struct sockaddr **redirect_ip_address,
		int *redirect_ip_length)
{
	int retval;
	struct auth_parameter_type new_authparam;
	char ip_string[INET6_ADDRSTRLEN+2], port_string[8];
	char local_ip_string[INET6_ADDRSTRLEN+2], local_port_string[8];
	int local_addr_len;

	/* extract address family into retval, ip address and port in strings */
	retval = cnv_inet_to_string(conn->ip_address, ip_string, port_string);
	if (unlikely(retval < 0)) {
		TRACE_ERROR("%s\n", ip_string);
		goto out0;
	}

	TRACE(TRACE_ENTER_LEAVE, "Enter init_connection, conn %p, sess %p\n",
	      			conn, conn->my_session);

	TRACE(TRACE_ISCSI, "Create connection %p to target %u CID %u "
			"at %s:%s family %d\n", conn,
	      		conn->my_session->scsi_target_id, conn->connection_id,
			ip_string, port_string, retval);

	/* finish allocating fields of new struct connection */

#ifdef FC2
	retval = sock_create(retval, SOCK_STREAM, 0, &conn->sock, 0);
#else
	retval = sock_create(retval, SOCK_STREAM, 0, &conn->sock);
#endif
	if (unlikely(retval < 0)) {
		TRACE_ERROR("Can't create socket, error %d\n", -retval);
		goto out0;
	}

	TRACE(TRACE_NET, "Created socket %p\n", conn->sock);

/* Ming Zhang, mingz@ele.uri.edu */
#ifndef K26
	/* to keep tcp from strangling due to lack of memory, force tcp's
	   calls to alloc_skb() to use atomic allocation, not GFP_KERNEL */
	conn->sock->sk->allocation = GFP_ATOMIC;
#endif

	/* auth support- CHONG */
	new_authparam.auth_flags = 0;
	new_authparam.chap_local_ctx = NULL;
	new_authparam.chap_peer_ctx = NULL;
	new_authparam.srp_ctx = NULL;

	retval = conn->sock->ops->connect(conn->sock, conn->ip_address,
			     		  conn->ip_length, 0);

	printk("%s to target %u cid %u at %s:%s conn %p\n",
	       retval < 0 ? "***ERROR*** can't connect" : "Connected",
	       conn->my_session->scsi_target_id, conn->connection_id,
	       ip_string, port_string, conn);

	if (unlikely(retval < 0)) {
		/* turn on (almost) all tracing after this call */
		if (TRACE_TEST(TRACE_DEBUG)) {
			TRACE_SET(TRACE_ALL
		    & ~(TRACE_ISCSI_FULL|TRACE_ENTER_LEAVE|TRACE_MY_MEMORY));
		}
		goto out1;
	}

	/* Turn off the Nagle Algorithm on this connection */
	tcp_nagle_off(conn->sock);

	/* 
	 * Get the socket details - local IP/port
	 * You should't fail here unless you have failed in
	 * connect() call in which case, you won't be here.
	 */
	local_addr_len = conn->ip_length;	/* value-result parameter */
	if (conn->local_ip_address == NULL)
		conn->local_ip_address = my_kmalloc(sizeof(struct sockaddr),
							"local_ip_address");
	if (conn->local_ip_address != NULL
		&& conn->sock->ops->getname(conn->sock, conn->local_ip_address,
						&local_addr_len, 0) == 0) {
		/* extract ip address and port into strings for printing */
		if (cnv_inet_to_string(conn->local_ip_address, local_ip_string,
							local_port_string) > 0)
	    		TRACE(TRACE_DEBUG,"iscsi local address %s:%s\n",
					local_ip_string, local_port_string);
	}

#ifdef ISCSI_STATS
	add_iscsi_proc_entry((void*)conn, CONNECTION);
#endif

	/*      Allocate buffer into which rx_thread will read
	 *	PDU headers from target.  For now, we ignore AHS,
	 *	so space is needed only for BHS and digest.
	 */
	conn->rx_buf = (__u8 *)my_kmalloc(ISCSI_HDR_LEN + CRC_LEN, "rx_buf");
	if (unlikely(conn->rx_buf == NULL)) {
		/* no memory for rx_buf */
		retval = -ENOMEM;
		goto out2;
	}

	/* we are connected to a target */
	conn->connection_state = CONNECTION_CONNECTED;

	/* chap and srp support - CHONG */
	new_authparam.auth_flags = targ_data->auth_parameter.auth_flags;
	new_authparam.chap_local_ctx
		= CHAP_CloneContext(targ_data->auth_parameter.chap_local_ctx);
	new_authparam.chap_peer_ctx
		= CHAP_CloneContext(targ_data->auth_parameter.chap_peer_ctx);
	new_authparam.srp_ctx
		= SRP_CloneContext(targ_data->auth_parameter.srp_ctx);

	/* now go do the login phase for this new connection */
	retval = iscsi_initiator_login(conn, when_called, p_param_tbl,
			new_authparam, targ_data, redirect_ip_address,
			redirect_ip_length);
	if (unlikely(retval != 0)) {
		goto out3;
	}

	/* Create connection's rx_thread only if get into Full Feature Phase */
	kernel_thread((int (*)(void *)) iscsi_initiator_rx_thread,
		      (void *) conn, 0);

	/* wait for the rx thread to set itself up */
	down_interruptible(&conn->rx_done_sem);

	if (conn->rx_thread == NULL) {
		/* the rx thread has already shut itself down */
		retval = -ECONNRESET;
		goto out;
	}

	/* connection is up and running */
	conn->connection_state = CONNECTION_FULL_FEATURE_PHASE;

	/* start up the tx thread for this new connection */
	kernel_thread((int (*)(void *)) iscsi_initiator_tx_thread,
		      (void *) conn, 0);

	/* wait for the tx thread to set itself up */
	down_interruptible(&conn->tx_done_sem);

	if (conn->rx_thread == NULL || conn->tx_thread == NULL) {
		/* the rx and/or tx thread have already shut themselves down */
		retval = -ECONNRESET;
	}

out:
	CHAP_FinalizeContext(new_authparam.chap_peer_ctx);
	CHAP_FinalizeContext(new_authparam.chap_local_ctx);
	SRP_FinalizeContext(new_authparam.srp_ctx);

out0:
	TRACE(TRACE_ENTER_LEAVE, "Leave init_connection, retval %d\n", retval);
	return retval;

out3:
	conn->connection_state = CONNECTION_NOT_PRESENT;
	my_kfree((void **)&conn->rx_buf, "rx_buf");

out2:
	printk("%s Disconnected from %s:%s\n", current->comm, ip_string,
		port_string);
#ifdef ISCSI_STATS
	remove_iscsi_proc_entry((void *)conn, CONNECTION);
#endif

out1:
	sock_release(conn->sock);
	conn->sock = NULL;
	goto out;
}

/****************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Called only by create_session() to
 * 1.	initialize a new session structure
 * 2.	create a new first connection for the session
 * 3.	connect to the target
 * 4.	conduct the complete login phase with the target
 *
 * Return value: =1 on redirection
 *		 =0 on all other successes,
 *		 <0 on failure
 ****************************************************************************/
static int
init_session(struct session *sess,
	     struct connection *conn,
	     __u32 lun,
	     struct iscsi_targetdata *targ_data,
	     struct sockaddr **redirect_ip_address,
	     int *redirect_ip_length)

{
	int retval = 0;

	TRACE(TRACE_ENTER_LEAVE, "Enter init_session, sess %p\n", sess);

	/* if not already allocated, allocate now - new 18_04 SAI */
	if (!sess->oper_param) {
		/*      finish allocating remaining structs
		 *	new struct session points to
		 */
		sess->oper_param = (struct session_operational_parameters *)
		    my_kmalloc(sizeof(struct session_operational_parameters),
			       "session oper_param");

		if (unlikely(sess->oper_param == NULL)) {
			retval = -ENOMEM;
			goto out;
		}
	}

	TRACE(TRACE_ISCSI, "Create session for target %u lun %u CID %u\n",
	      sess->scsi_target_id, lun, conn->connection_id);

#ifdef ISCSI_STATS
	add_iscsi_proc_entry((void*)sess, SESSION);
#endif

	/* Connect to the target and go through the complete login phase */
	/* chap and srp support - CHONG */
	retval = init_connection(conn, LEADING_ONLY | INITIAL_ONLY | ALL,
				 *sess->session_params,
				 targ_data,
				 redirect_ip_address, redirect_ip_length);
	if (unlikely(retval != 0))
		goto out2;

	sess->session_state = SESSION_LOGGED_IN;

	/* successfully created a new connection for a new session
	 * Set up the session-wide negotiated operational parameters for this
	 * new session
	 */
	set_session_parameters(sess->oper_param, *sess->session_params);

	TRACE(TRACE_DEBUG, "ErrorRecoveryLevel %u\n",
	      sess->oper_param->ErrorRecoveryLevel);

	/* force use of snack when error recovery is in effect - RDR
	 *      Draft 20, Section 10.16 SNACK Request
	 *      "If the implementation supports ErrorRecoveryLevel greater than
	 *      zero, it MUST support all SNACK types."
	 */
	if (sess->oper_param->ErrorRecoveryLevel > 0) {
		sess->init_snack_flag
		    |= (DATA_SNACK_ENABLE | STATUS_SNACK_ENABLE);
	}

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave init_session, retval %d\n", retval);

	return retval;

out2:
#ifdef ISCSI_STATS
	remove_iscsi_proc_entry((void*)sess, SESSION);
#endif
	my_kfree((void **)&sess->oper_param, "session oper_param");
	goto out;
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * free up the memory allocated to a session struct, but not any connection
 * structs it might point to, which must already have been freed up.
 * Except for those connection structs, undoes the allocations in
 * build_session_skeleton() and those that might come later,
 * in the reverse order
 */
static void
unbuild_session_skeleton( struct session *old_session )
{
	struct command *command;
	struct r2t_cookie *cookie;
	struct list_head *lptr, *next;

	/* done with unused r2t cookie structures */
	list_for_each_safe(lptr, next, &old_session->free_r2t_cookies) {
		list_del(lptr);
		cookie = list_entry(lptr, struct r2t_cookie, link);
		my_kfree((void **)&cookie, "r2t cookie");
	}

	/* done with unused command structures */
	list_for_each_safe(lptr, next, &old_session->free_commands) {
		list_del(lptr);
		command = list_entry(lptr, struct command, link);
		my_kfree((void **)&command, "command");
	}

#ifdef ISCSI_STATS
	my_kfree((void **)&old_session->sess_stats, "session_stats");
#endif

	if (old_session->session_params) {
		/* done with the session's parameter table */
		param_tbl_uncpy(*old_session->session_params);
		my_kfree((void **)&old_session->session_params,
			 "session param_tbl");
	}

	/* done with the session structure itself */
	my_kfree((void **)&old_session, "session");
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * create and initialize a session struct and its first connection struct
 */
static struct session *
build_session_skeleton(__u32 target, struct sockaddr *ip_address,
			int ip_length, __u32 lun, __u32 cid)
{
	struct session *new_session;
	struct connection *conn;
	struct iscsi_targetdata *targ_data;

	new_session
	    = (struct session *)my_kmalloc(sizeof(struct session), "session");
	if (unlikely(new_session == NULL)) {
		goto out1;
	}

	/* reset all fields in new struct session to 0 */
	memset(new_session, 0, sizeof(struct session));

	/* initialize most of the new struct session fields here */
	new_session->sess_lock = UNH_LOCK_UNLOCKED;
	INIT_LIST_HEAD(&new_session->free_commands);
	INIT_LIST_HEAD(&new_session->free_r2t_cookies);
	new_session->scsi_target_id = target;
	new_session->n_luns_in_session = 1;
	new_session->lun_bits = 1 << lun;
	new_session->session_state = SESSION_NOT_PRESENT;
	new_session->cur_task_tag = 88888;	/* something unique */
	new_session->cur_cmd_sn = 22222;	/* something unique */
	new_session->exp_cmd_sn = new_session->cur_cmd_sn + 1;
	new_session->max_cmd_sn = new_session->exp_cmd_sn;
	init_timer(&new_session->connrec_timer);

	targ_data = &global_hostdata->target_data[target];

	/* use the configured values to initialize following fields */
	new_session->session_flags = targ_data->init_sess_flags;
	new_session->init_conn_flags = targ_data->init_conn_flags;
	new_session->nop_period = targ_data->nop_period;

	TRACE(TRACE_TIMERS, "new session nop_period %u\n",
	      new_session->nop_period);

	/* Added data and status snack support - SAI */
	new_session->init_snack_flag = targ_data->init_snack_flag;

	/* Added the retransmit wait period - SAI */
	new_session->retran_period = targ_data->retran_period;

	/* Store the connection scheduling scheme - SAI */
	new_session->sched_scheme = targ_data->sched_scheme;

	/* allocate space for configured parameters */
	new_session->session_params
		= my_kmalloc(MAX_CONFIG_PARAMS * sizeof(struct parameter_type),
			     "session param_tbl");
	if (unlikely(new_session->session_params == NULL)) {
		goto out_err;
	}

	/* init session parameters with copy of configured parameter table */
	param_tbl_cpy(*new_session->session_params, targ_data->param_tbl);

#ifdef ISCSI_STATS
	/* get stats structure */
	new_session->sess_stats =
	    (struct iscsi_session_stats*)my_kmalloc(sizeof(struct iscsi_session_stats),
						      "session_stats");
	if (unlikely(new_session->sess_stats == NULL)) {
	    goto out_err;
	}
	/* reset all the stats field */
	memset(new_session->sess_stats, 0, sizeof(struct iscsi_session_stats));
#endif
	
	/* create the first connection on this session */
	conn = (struct connection *)my_kmalloc(sizeof(struct connection),
						"connection");
	if (unlikely(conn == NULL)) {
		/* no memory for new connection structure */
		goto out_err;
	}

	/* reset all fields in new struct connection */
	memset(conn, 0, sizeof(struct connection));

	conn->connection_id = cid;
	/* target's default MaxRecvDataSegmentLength */
	conn->max_send_length = 8192;
	/* initiator's MaxRecvDataSegmentLength */
	conn->max_recv_length = 8192;
	conn->nop_counter = 1;
	conn->my_session = new_session;
	conn->connection_flags = new_session->init_conn_flags;
	conn->connection_state = CONNECTION_NOT_PRESENT;
	init_MUTEX_LOCKED(&conn->rx_done_sem);
	init_MUTEX_LOCKED(&conn->tx_sem);
	init_MUTEX_LOCKED(&conn->tx_done_sem);
	init_MUTEX_LOCKED(&conn->task_mgt_sem);
	init_timer(&conn->tx_timer);
	INIT_LIST_HEAD(&conn->pending_commands);
	INIT_LIST_HEAD(&conn->r2t_list);
	conn->ip_address = ip_address;
	conn->ip_length = ip_length;

	/* point fields in new session at new connection */
	new_session->nconnections = 1;
	new_session->connection_head = conn;
	new_session->rr_conn_head = conn;

	/* pre-allocate maximum number command structures session may use
	 * (one extra in case of task-management and/or nopout commands)
	 */
	if (unlikely(preallocate_commands(new_session) == 0)) {
		goto out_err;
	}
out:
	return new_session;

out_err:
	unbuild_session_skeleton(new_session);
out1:
	TRACE(TRACE_NET, "Free ip_address %p\n", ip_address);
	my_kfree((void **)&ip_address, "ip_address");
	goto out;
}

static int
add_session_to_hostdata(struct connection *new_connection, __u32 lun,
			struct session *new_session, struct Scsi_Host *host);

/****************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * executed by rx thread or tx thread or init_recovery thread or command process
 * Creates and initializes a new session and its first connection struct by
 * calling corresponding functions. It then tries to connect to the target.
 * Called from iscsi_initiator_proc_info() when a line such as
 * 	"iscsi_initiator ip 1234abcd port 4000 target 1 lun 2 cid 3"
 * is written to /proc/scsi/iscsi_initiator/0
 * also called from create_rec_conn() to create a recovery connection
 * Return value:
 * 	=1 on success -- new session added, logged in, now in FFP
 * 	=2 on success -- new lun added to existing session
 * 	=0 on success -- new connection added, logged in, now in FFP
 *			 or new discovery session added, now in FFP
 * 	<0 on failure -- value is errno negated
 ****************************************************************************/
int
create_session(__u32 target, struct sockaddr *ip_address, int ip_length,
		__u32 lun, __u32 cid, struct Scsi_Host *host)
{
	unsigned long lock_flags;
	int i, retval, redirect_count = 0;
	struct sockaddr *redirect_ip_address;
	int redirect_ip_length;
	struct session *new_session;
	struct connection *conn;
	struct iscsi_hostdata *hostdata
	    = (struct iscsi_hostdata *) host->hostdata;

	if (unlikely(TRACE_TEST(TRACE_ENTER_LEAVE))) {
		char ip_string[INET6_ADDRSTRLEN+2];
		char port_string[8];

		if (cnv_inet_to_string(ip_address,ip_string,port_string) < 0) {
			/* should never happen */
			TRACE_ERROR("bad ip_address\n");
			TRACE(TRACE_NET, "Free ip_address %p\n", ip_address);
			my_kfree((void **)&ip_address, "ip_address");
			/* software caused connection abort */
			return -ECONNABORTED;
		}

		TRACE(TRACE_ENTER_LEAVE,
				"Enter create_session to %s:%s, "
				"target %u, lun %u, cid %u\n", ip_string,
				port_string, target, lun, cid);
	}

retry:	/* here to try again on a redirection */

	new_session = build_session_skeleton(target, ip_address, ip_length,
								lun, cid);
	if (unlikely(new_session == NULL)) {
		retval = -ENOMEM;
		goto out;
	}

	/* add the session and/or connection to the global hostdata structure */
	conn = new_session->connection_head;
	retval = add_session_to_hostdata(conn, lun, new_session, host);

	/* retval < 0: error, everything already exists or too many sessions
	 * retval = 0: adding a new connection (only) to an existing session
	 * retval = 1: adding complete new session, lun and connection
	 * retval = 2: adding a new lun (only) to an existing session
	 */
	switch (retval) {
	case 2:
		/* adding a new lun (only) to an existing session */
		/* fall through to default case */

	default:
		/* error, everything already exits or too many sessions
		 * free up newly created connection and session structures
		 */
		goto out3;

	case 0:
	 	/* adding a new connection (only) to an existing session */
		retval = init_connection(conn, INITIAL_ONLY | ALL,
				*new_session->session_params,
				&hostdata->target_data[target],
				&redirect_ip_address, &redirect_ip_length);
		if (unlikely(retval != 0)) {
			/* could not bring up the new connection,
			 * remove it from session
			 */
			goto out4;
		}

		/* connection is up and running
		 * don't need the session structure or the
		 * session parameter table
		 */
		goto out2;

	case 1:
		/* adding complete new session, lun and connection */
		retval = init_session(new_session, conn, lun,
				&hostdata->target_data[target],
				&redirect_ip_address,
				&redirect_ip_length);
		if (unlikely(retval != 0)) {
			/* initialization or login failed or got redirection */
			goto out4;
		}

		/* everything is now up and running in the new session */
		new_session->session_state = SESSION_FULL_FEATURE_PHASE;

		if (new_session->sched_scheme == CONN_SCHED_LUN) {
			add_lun_to_sess(lun, new_session);
		}

		if (new_session->oper_param->SessionType) {
			/* a discovery session, we have to drive it */
			TRACE(TRACE_ISCSI,
			      "%s Enter discovery session Full Feature Phase\n",
			      current->comm);
			UNH_LOCK(&new_session->sess_lock, lock_flags);
			/* gives back 1 on error, 0 on success */
			retval = drive_text_negotiate(conn, new_session, 0, 1);
			UNH_UNLOCK(&new_session->sess_lock, lock_flags);
		} else {
			TRACE(TRACE_ISCSI,
			      "%s Enter normal session Full Feature Phase\n",
			      current->comm);
			retval = 1;
		}
#ifdef K26 /* Hot-plug modifications for 2.6.0 kernel */
		scsi_add_device(host,0, target, lun); 
#endif
        
	} /* switch */

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave create_session, retval %d\n", retval);

	return retval;

out4:
	/* Here to recover from a late error, after host's list was updated. */
	if ((i = remove_session_from_hostdata(target,lun,cid,hostdata)) == 1) {
		/* memory for entire session already freed, just leave */
		goto out0;
	} else if (i == 0) {
		/* only memory for new connection freed, get rid
		 * of new session too
		 */
		goto out2;
	}

	/* should never happen */
	TRACE_ERROR("non-existent target %u LUN %u CID %u\n", target, lun, cid);

out3:
	/* here to free the new session struct and its connection struct */
	my_kfree((void **)&new_session->connection_head, "connection");
	TRACE(TRACE_NET, "Free ip_address %p\n", ip_address);
	my_kfree((void **)&ip_address, "ip_address");

out2:
	unbuild_session_skeleton(new_session);

out0:
	/* now check to see if failure was caused by request for redirection */
	if (retval == STAT_CLASS_REDIRECTION) {
		if (redirect_count++ < MAX_REDIRECTION_RETRIES) {
			TRACE(TRACE_ISCSI, "Attempting Redirection number %u\n",
			      redirect_count);
			ip_address = redirect_ip_address;
			ip_length = redirect_ip_length;
			goto retry;
		}
		TRACE_ERROR("Too many redirection attempts, max of %d\n",
			    MAX_REDIRECTION_RETRIES);
		TRACE(TRACE_NET, "Free ip_address %p\n", ip_address);
		my_kfree((void **)&redirect_ip_address, "ip_address");
		retval = -ECONNABORTED;	/* software caused connection abort */
	}
	goto out;
}

/*
 * host_data_lock and sess->sess_lock MUST be held by the calling process/thread
 * returns 0 if ok, <0 on error
 */
static int __attribute__ ((no_instrument_function))
add_conn_to_sess( struct connection *new_connection,
		  __u32 lun,
		  struct session *old_session)
{
	int retval;
	struct connection *conn, *prev;

	retval = 0;
	prev = NULL;
	for (conn = old_session->connection_head;
	     conn != NULL;
	     conn = conn->next) {
		if (unlikely(conn->connection_id
					== new_connection->connection_id)) {
			TRACE_ERROR("Already have target %u LUN %u CID %u\n",
					old_session->scsi_target_id, lun,
					new_connection->connection_id);
			retval = -EBUSY;
			goto out;
		}
		prev = conn;
	}

	if (unlikely(old_session->nconnections
				>= old_session->oper_param->MaxConnections)) {
		TRACE_ERROR("Already have maximum of %u connections"
				" for target %u\n",
				old_session->nconnections,
				old_session->scsi_target_id);
		retval = -EBUSY;
		goto out;
	}

	/* ok to add new connection at end of session's connection list */
	old_session->nconnections++;
	if (prev == NULL)
		old_session->connection_head = new_connection;
	else
		prev->next = new_connection;

	/* override initializations that depend on session */
	new_connection->my_session = old_session;
	new_connection->connection_flags = old_session->init_conn_flags;

out:
	return retval;
}

/****************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * executed by rx thread or tx thread or init_recovery thread or command process
 * Called only from create_session().
 * Return value:
 * 	=1 on success -- new_session added to end of hostdata's session_list ok
 * 	=2 on success -- adding new lun to existing session
 * 	=0 on success -- adding new connection to existing session
 * 	<0 on failure -- value is errno negated
 ****************************************************************************/
static int
add_session_to_hostdata(struct connection *new_connection, __u32 lun,
			struct session *new_session, struct Scsi_Host *host)
{
	unsigned long flags, lock_flags;
	struct session *old_session;
	int retval;
	struct iscsi_hostdata *hostdata
	    = (struct iscsi_hostdata *) host->hostdata;

	TRACE(TRACE_ENTER_LEAVE,
	      "Enter add_session_to_hostdata, new_session %p\n", new_session);

	UNH_LOCK(&host_data_lock, flags);

	/* see if a session with the same target id already exists */
	old_session = find_session_by_id(new_session->scsi_target_id, hostdata);

	if (old_session != NULL) {
		/* already have a session with this target id */
		UNH_LOCK(&old_session->sess_lock, lock_flags);
		if (!(old_session->lun_bits
				& new_session->lun_bits)) {
			/* adding new lun to existing session */
			old_session->lun_bits
				|= new_session->lun_bits;
			old_session->n_luns_in_session++;

			printk("Add LUN %u to existing session "
				"%p, target %u, now %u LUNS in session\n",
				lun, old_session,
				old_session->scsi_target_id,
				old_session->n_luns_in_session);

			preallocate_commands(old_session);

			if (old_session->sched_scheme
				 == CONN_SCHED_LUN) {
				locked_add_lun_to_sess(lun,
						old_session);
			}
			retval = 2;
		} else {
			/* this lun already part of session,
			 * must be adding a new connection
			 */
			retval = add_conn_to_sess(
						new_connection,
						lun,
						old_session);
		}
		UNH_UNLOCK(&old_session->sess_lock, lock_flags);
		goto out;
	}

	/* no session with this target id exists yet, add it */
	if (hostdata->nsessions < host->max_id) {
		/* ok to add another session */
		list_add_tail(&new_session->link, &hostdata->session_list);
		hostdata->nsessions++;
		retval = 1;
	} else {
		/* host already has maximum number of sessions allowed */
		TRACE_ERROR("Max of %u sessions, can't add session for "
			    "target %u LUN %u CID %u\n",
			    host->max_id, new_session->scsi_target_id, lun,
			    new_connection->connection_id);
		retval = -EBUSY;
	}

out:
	UNH_UNLOCK(&host_data_lock, flags);

	TRACE(TRACE_ENTER_LEAVE, "Leave add_session_to_hostdata, retval %d\n",
	      retval);

	return retval;
}

/***************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Called from close_session() and remove_session_from_hostdata().
 * Cleans up a connection struct whenever a connection is ended.
 * This does NOT close the socket or free the connection structure
 * and does NOT reduce the session's "nconnections" counter.
 ***************************************************************************/
static void
close_connection(struct connection *conn)
{
	void *temp;
	struct session *session;
	unsigned long lock_flags;

	TRACE(TRACE_ENTER_LEAVE, "Enter close_connection %p\n", conn);

	if (unlikely(conn == NULL
		|| conn->connection_state == CONNECTION_NOT_PRESENT)) {
		goto out;
	}

	TRACE(TRACE_ISCSI, "Closing connection %p for target %u CID %u\n",
	      conn, conn->my_session->scsi_target_id, conn->connection_id);

	session = conn->my_session;

	/* get exclusive access to the iscsi session structure */

	UNH_LOCK(&session->sess_lock, lock_flags);

	/* do not allow tx_timer to restart itself */
	conn->connection_flags |= TX_TIMER_OFF;

	/* now delete the tx_timer (del_timer checks for timer active or not) */
	del_timer_sync(&conn->tx_timer);

	if (conn->rx_thread != NULL) {
		/* rx_thread still up and running, kill it */
		TRACE(TRACE_ISCSI, "Killing rx-%u-%u\n",
		      conn->my_session->scsi_target_id, conn->connection_id);
		temp = conn->rx_thread;
		conn->rx_thread = NULL;
		/*      Mike Christie, mikenc@us.ibm.com */
		send_sig(ISCSI_SHUTDOWN_SIGNAL, temp, 1);

		/* release our exclusive access to the iscsi structures */

		UNH_UNLOCK(&session->sess_lock, lock_flags);

		TRACE(TRACE_SEM, "close_connection waiting on rx_done_sem\n");
		down_interruptible(&conn->rx_done_sem);
		TRACE(TRACE_SEM,
		      "close_connection done waiting on rx_done_sem\n");

		/* now get back exclusive access to the iscsi structures */
		UNH_LOCK(&session->sess_lock, lock_flags);
	}

	conn->connection_state = CONNECTION_NOT_PRESENT;
#ifdef ISCSI_STATS
	remove_iscsi_proc_entry((void*)conn, CONNECTION); 
#endif
	my_kfree((void **)&conn->rx_buf, "rx_buf");

	if (conn->tx_thread != NULL) {
		int ret;
		/* indicate that this thread is going to do the shut down */
		TRACE(TRACE_ISCSI, "Killing tx-%u-%u\n",
		      conn->my_session->scsi_target_id, conn->connection_id);
		temp = conn->tx_thread;
		conn->tx_thread = NULL;
		/*      Mike Christie, mikenc@us.ibm.com */
		ret = send_sig(ISCSI_SHUTDOWN_SIGNAL, temp, 1);

		/* then release our exclusive access to the iscsi structures */
		UNH_UNLOCK(&session->sess_lock, lock_flags);

		/* then wait for tx thread to actually die */
		TRACE(TRACE_SEM, "Waiting on tx_done_sem %x\n",ret);
		down_interruptible(&conn->tx_done_sem);
		TRACE(TRACE_SEM, "Done waiting on tx_done_sem \n");

	} else {
		/* some other thread is already shutting this session down */

		/* just release our exclusive access to the iscsi structures */
		UNH_UNLOCK(&session->sess_lock, lock_flags);
	}

	TRACE(TRACE_ISCSI, "Closed connection for target %u CID %u\n",
	      session->scsi_target_id, conn->connection_id);

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave close_connection\n");
}

/***************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Called only from remove_session_from_hostdata()
 * Cleans up session struct whenever a session has to be ended.
 * When called, this session struct must NOT be linked into the global list.
 * On exit, this session struct is NOT freed.
 ***************************************************************************/
static void
close_session(struct session *sess)
{
	struct connection *conn;
	struct connection *next_conn;
	__u32 lun;

	TRACE(TRACE_ENTER_LEAVE, "Enter close_session\n");

	TRACE(TRACE_ISCSI, "Closing session %p for target %u\n",
	      sess, sess->scsi_target_id);

	/* close all connections, but do not release them or their sockets yet*/
	for (conn = sess->connection_head; conn != NULL; conn = next_conn) {
		next_conn = conn->next;
		TRACE(TRACE_DEBUG,
		      "Going to close_connection(%p), next_conn %p\n", conn,
		      next_conn);
		/* close_connection is called without session lock held */
		close_connection(conn);
		TRACE(TRACE_DEBUG,
		      "Back from close_connection(%p), next_conn %p\n", conn,
		      next_conn);
	}

	TRACE(TRACE_ISCSI, "All connections closed in session for target %u\n",
	      sess->scsi_target_id);

	/* delete connection recovery timer - SAI */
	del_timer_sync(&sess->connrec_timer);

	/* go back and release all the connections' sockets and structures */
	while ((conn = sess->connection_head) != NULL) {
		sess->connection_head = conn->next;
		free_connection(conn);
	}

	my_kfree((void **)&sess->oper_param, "session oper_param");

#ifdef ISCSI_STATS
	remove_iscsi_proc_entry((void*)sess, SESSION); 
#endif

	for (lun = 0; lun < MAX_LUNS; lun++) {
		free_inquiry_stuff(sess, lun);
	}

	TRACE(TRACE_ISCSI, "Closed session for target %u\n",
	      sess->scsi_target_id);

	TRACE(TRACE_ENTER_LEAVE, "Leave close_session\n");
}

/*
 * host_data_lock and sess->sess_lock MUST be held by the calling process/thread
 * remove this connection from this session's list of connections.
 * return 1 if this is the last connection in the session,
 * 0 if more connections remain in the session.
 */
static int __attribute__ ((no_instrument_function))
drop_conn_from_session( struct connection *conn,
			struct connection *prev_conn,
			struct session *sess)
{
	__u32 lun;

	if (sess->nconnections > 1) {
		/* at least 1 more connection will remain in this session */
		sess->nconnections--;
		if (prev_conn == NULL) {
			sess->connection_head = conn->next;
		} else{
			prev_conn->next = conn->next;
		}

		/* Adjust the round robin connection head - SAI */
		if (conn == sess->rr_conn_head) {
			sess->rr_conn_head = prev_conn;
		}

		/* remove all lun assignments, if any */
		if (conn->assigned_lun_count) {
			for (lun = 0; lun < MAX_LUNS; lun++) {
				if (sess->lun_assignments[lun] == conn) {
					/* this lun was assigned to this
					 * connection, so assign it to the
					 * remaining connection that has
					 * the fewest luns assigned to it
					 */
					locked_drop_lun_from_sess(lun, sess, 0);
					locked_add_lun_to_sess(lun, sess);
				}
			}
		}
		return 0;
	} else {
		/* this is last connection, remove entire session */
		sess->lun_bits = 0;
		return 1;
	}
}

/****************************************************************************
* when called, NO LOCKS SHOULD BE LOCKED!
* Removes session, connection or lun for this target
*
* Return value:	=1 on success -- complete session removed from host list
* 		=2 on success -- old lun removed from old session
* 		=0 on success -- old connection removed from old session
* 		<0 on failure -- value is errno negated
*
*  	On entry: No locks held
*  	On exit: No locks held
*
* Called indirectly via create_session() from /proc interface
*		(scan_ip_stuff()) on a "up target lun cid" that fails
* Called directly from /proc interface (scan_ip_stuff())
*		on a "down target lun cid"
* Called indirectly via remove_luns() from /proc interface
*		(iscsi_initiator_release()) on module unload
* Called directly from rx thread (iscsi_initiator_rx_thread())
*		when rx thread terminates a discovery session
* Called indirectly via deal_with_connrec() from rx thread
*		to remove a failed connection after connection recovery
* Called indirectly via create_rec_conn() and create_session() from
*		rx thread to remove a connection after error on creation
* Called indirectly via create_session() from recovery_thread
*		(init_recovery_thread()) to remove a connection after error on
*		creation
* Called indirectly via remove_luns() from recovery_thread
*		(init_recovery_thread()) to remove a failed session
* Called indirectly via remove_luns() from tx thread (tx_shutdown_thread()
*		when tx thread terminates itself
* Called indirectly via create_rec_conn() and create_session() from
*		tx thread (iscsi_initiator_tx_thread()) on test 8 --
*		force session recovery logout
* Called indirectly via remove_luns() from reboot notifier callback
*		(iscsi_initiator_halt())
*
****************************************************************************/
int
remove_session_from_hostdata(__u32 target, __u32 lun, __u32 cid,
			     struct iscsi_hostdata *hostdata)
{
	unsigned long flags, lock_flags;
	int retval = -ENODEV;	/* assume failure */
	struct session *sess;
	struct connection *conn = NULL, *prev_conn;

	TRACE(TRACE_ENTER_LEAVE,
	      "Enter remove_session_from_hostdata, target %u lun %u cid %u\n",
	      target, lun, cid);

	UNH_LOCK(&host_data_lock, flags);

	sess = find_session_by_id(target, hostdata);
	if (likely(sess != NULL)) {
		/* this is the matching session structure, see if lun matches */
		if (sess->lun_bits & (1 << lun)) {
			/* lun matches, see if cid matches */
			/* We need to get the session lock here.
			 * As a rule, host_data_lock is used to protect
			 * global_hostdata and hence all the sessions
			 * attached to it. However, individual conns
			 * in a session are protected by sess_lock on
			 * each session structure. It is possible that
			 * queuecommand/get_connection hold this
			 * sess_lock and search the connection list
			 * when we are deleting/removing the same 
			 * connection -AJ
			 */
			UNH_LOCK(&sess->sess_lock, lock_flags);
			prev_conn = NULL;
			for (conn = sess->connection_head;
			     conn != NULL;
			     conn = conn->next) {
				if (conn->connection_id == cid) {
					if (drop_conn_from_session(conn,
							prev_conn,
							sess)) {
						/* last conn in sess */
						conn = NULL;
					} else {
						/* conn removed ok */
						retval = 0;
					}
					break;
				}
				prev_conn = conn;
			}
			UNH_UNLOCK(&sess->sess_lock, lock_flags);
			if (conn == NULL) {
				/* did not find matching connection,
				 * or matching connection is last in
				 * session, turn off bit for lun
				 */
				sess->lun_bits &= ~(1 << lun);
				sess->n_luns_in_session--;
				if (sess->lun_bits == 0) {
					/* no luns remain,
					 * remove entire session
					 */
					list_del(&sess->link);
					hostdata->nsessions--;
					retval = 1;
				} else {
					/* other luns remain, just
					 * remove the connection
					 * assignment for this lun
					 */
					drop_lun_from_sess(lun, sess);
					unallocate_commands(sess);
					retval = 2;
				}
			}
		} else {
			/* lun does not match, remove entire session */
			sess->lun_bits = 0;
			list_del(&sess->link);
			hostdata->nsessions--;
			retval = 1;
		}
	}

	UNH_UNLOCK(&host_data_lock, flags);

	switch (retval) {
	case 0:
		/* close only this connection (which has already been removed
		 * from session's connection list.  This has not
		 * released the socket yet.
		 */
		close_connection(conn);

		/* go back and release the socket and free the conn structure */
		free_connection(conn);
		break;

	case 1:
		/* close this entire session, including all
		 * remaining connections
		 */
		TRACE(TRACE_ISCSI, "Remove entire session for target %u\n",
		      target);

		/* free up any storage connected to this old session */
		if (sess->session_state == SESSION_NOT_PRESENT) {
			/* session's connection structure is empty,
			 * just free any connection it has
			 */
			if (sess->connection_head)
			      free_connection(sess->connection_head);
		} else {
			close_session(sess);
		}

		/* finally free up the old struct session itself */
		unbuild_session_skeleton(sess);
		break;

	case 2:
		/* just removed a lun, do nothing further here */
		TRACE(TRACE_ISCSI,
		      "Remove LUN from session for target %u lun %u\n", target,
		      lun);
		break;

	default:
		/* < 0, error somewhere */
		TRACE(TRACE_ISCSI, "non-existent target %u LUN %u CID %u\n",
		      target, lun, cid);
		break;

	} /* switch */

	TRACE(TRACE_ENTER_LEAVE,
	      "Leave remove_session_from_hostdata, retval %d\n", retval);
	return retval;
}

/*
 * executed only by command process.
 * the host_data_lock MUST be held by calling thread/process.
 * Returns	=0 on success (and sets old_connection, old_session),
 * 		<0 on failure (and leaves old_connection, old_session unchanged)
 */
int
find_connection(__u32 target,
		__u32 cid,
		struct iscsi_hostdata *hostdata,
		struct connection **old_connection,
		struct session **old_session)
{
	struct session *session;
	struct connection *conn;
	int retval = -ENODEV;	/* assume failure */

	TRACE(TRACE_ENTER_LEAVE, "Enter find_connection, target %u cid %u\n",
	      target, cid);

	session = find_session_by_id(target, hostdata);
	if (likely(session != NULL)) {
		/* this is the matching session structure,
		 * try to find matching cid
		 */
		for (conn = session->connection_head; conn != NULL;
		     conn = conn->next) {
			if (conn->connection_id == cid) {
				/* found a matching connection,
					pass back answers */
				*old_connection = conn;
				*old_session = session;
				retval = 0;
				break;
			}
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave find_connection, retval %d\n", retval);

	return retval;
}

/*	Mike Christie, mikenc@us.ibm.com
 *
 * when called, NO LOCKS SHOULD BE LOCKED!
 * remove all valid luns in session (last lun will take out session)
 * returns smallest lun number removed, -1 if none removed.
 */
int
remove_luns(struct session *sess, struct Scsi_Host *host,
	    struct iscsi_hostdata *hostdata)
{
	__u32 lun;
	int retval = -1;

	if (unlikely(!sess || sess->lun_bits == 0 || !host || !hostdata)) 
		goto out;

	for (lun = 0; lun < host->max_lun; lun++) {
		if (sess->lun_bits & (1 << lun)) {
			if (retval < 0)
				retval = lun;
			if (remove_session_from_hostdata
			   (sess->scsi_target_id, lun, ALL_ONES,
			   hostdata) == 1) {
				/* just removed the whole session structure */
				break;
			}
		}
	}
out:
	return retval;
}

/*	Mike Christie, mikenc@us.ibm.com
 *
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Called when the system is shutdown. Closes all sessions.
 * Does not care about memory menagement
*/
static int
iscsi_initiator_halt(struct notifier_block *nb, ulong event, void *buf)
{
	struct session *sess;
	struct list_head *lptr, *next;

	if (event != SYS_RESTART && event != SYS_HALT && event != SYS_POWER_OFF)
		return NOTIFY_DONE;

	if (global_host == NULL || global_hostdata == NULL)
		return NOTIFY_OK;

	/* Clear the session struct linked list */
	list_for_each_safe(lptr, next, &global_hostdata->session_list) {
		sess = list_entry(lptr, struct session, link);
		if (sess->lun_bits != 0)
			remove_luns(sess, global_host, global_hostdata);
	}
	global_host = NULL;
	global_hostdata = NULL;

	return NOTIFY_OK;
}

/****************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Called by the SCSI Mid-level when the iscsi_initiator module is removed.
 * (via the release field in the Scsi_Host_Template driver_template)
 * When called, the midlevel holds the big kernel lock, but it does
 * not hold the io_request_lock.
 * Parameter: Ptr. to Scsi_Host
 * Return Value: 0, on success
 * 		-1, on error
 ****************************************************************************/
static int
iscsi_initiator_release(struct Scsi_Host *host)
{
	int retval = 0, i;
	struct iscsi_hostdata *hostdata;
	struct session *sess;
	struct list_head *lptr, *next;

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_release, host %p\n",
	      host);

	TRACE(TRACE_ISCSI, "Releasing UNH iSCSI initiator module\n");

	/* If the host struct is NULL, it is an error and we want to return */
	if (unlikely(host == NULL || global_host == NULL
						|| global_hostdata == NULL)) {
		retval = -1;
		goto out;
	}

	/* If the hostdata struct is not what we expect, we want to return */
	if (unlikely((hostdata = (struct iscsi_hostdata *) host->hostdata)
							!= global_hostdata)) {
		retval = -1;
		goto out;
	}

	/* Clear the session struct linked list */
	list_for_each_safe(lptr, next, &hostdata->session_list) {
		sess = list_entry(lptr, struct session, link);
		if (sess->lun_bits == 0) {
			/* all luns already removed, just release
			 * this structure
			 */
			list_del(lptr);
			unbuild_session_skeleton(sess);
		} else {
			/* remove all valid luns in session
			 * (last lun will take out session)
			 */
			remove_luns(sess, host, hostdata);
		}
	}


	/* global pointer is no longer valid after unregister */
	if (global_hostdata) {
		for (i = 0; i < MAX_TARGET_IDS; i++) {
			param_tbl_uncpy(global_hostdata->target_data[i].
					param_tbl);
			CHAP_FinalizeContext(global_hostdata->target_data[i].
					     auth_parameter.chap_peer_ctx);
			CHAP_FinalizeContext(global_hostdata->target_data[i].
					     auth_parameter.chap_local_ctx);
			SRP_FinalizeContext(global_hostdata->target_data[i].
					    auth_parameter.srp_ctx);
		}
#ifdef ISCSI_STATS
		if (global_hostdata
			->instance_stats.iscsi_intr_last_tgt_failure_name) {
			kfree(global_hostdata
	   		->instance_stats.iscsi_intr_last_tgt_failure_name);
		}
		my_kfree((void **)&global_hostdata
			->instance_stats.iscsi_intr_last_tgt_failure_address,
			 "failure address");
#endif
	}

#ifdef K26
	printk("unh_iscsi_initiator removed\n");
#else
	scsi_unregister(host);

	initiator_unregister_device();
#endif

	global_hostdata = NULL;
	global_host = NULL;

	/* check for any unfreed memory */
	my_kempty();

	/* print statistics about io vector usage */
	endup_iovs();

	TRACE(TRACE_ISCSI, "unregister UNH iSCSI host number %d\n",
	      host->host_no);

	unregister_reboot_notifier(&iscsi_initiator_notifier);

	TRACE(TRACE_ISCSI, "Released UNH iSCSI module\n");

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_initiator_release, retval %d\n",
	      retval);
	return retval;
}

/*	Note the name "driver_template" is required for "scsi_module.c" to work! */
/*	Used by: "/usr/src/linux/drivers/scsi/scsi_module.c" */
/*	Wrap this in an ifdef because "scsi_module.c" says to! */
#ifdef MODULE

MODULE_AUTHOR("Bob Russell");
MODULE_DESCRIPTION(OUR_NAME);
MODULE_SUPPORTED_DEVICE("sd,sr,sg,st");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

/*
 * initializor for static Scsi_Host_Template struct as defined in
 *			/usr/src/linux/drivers/scsi/hosts.h
 *	which has good comments about each field.
 *
 *	There is one such struct for each different type of host adaptor,
 *	and since there is only one iscsi type, we only need 1 static instance.
 *	(This is normal for each different adaptor type.)
*/



#ifdef K26
/*
 * A single static instance of this structure for this adaptor type
*/
static struct scsi_host_template driver_template = {\
	proc_info:		iscsi_initiator_proc_info,\
	name:			OUR_NAME,\
        detect:                 iscsi_initiator_detect,\
        release:                iscsi_initiator_release,\
        info:                   NULL, \
	ioctl:			NULL,\
        queuecommand:           iscsi_initiator_queuecommand,\
        eh_abort_handler:       iscsi_initiator_abort,\
        eh_device_reset_handler:iscsi_initiator_reset,\
        eh_bus_reset_handler:	NULL,\
        eh_host_reset_handler:	NULL,\
/* use default scsicam_bios_param() in linux/drivers/scsi/scsicam.c */\
        bios_param:             NULL,\
/* max no. of simultaneously active SCSI commands driver can accept */\
        can_queue:              MAX_COMMANDS_PER_LUN * MAX_LUNS,\
        this_id:                -1, /* our host has no id on the SCSI bus */\
/*  TODO XXX K26 sg_tablesize:MAX_IOV_SLOTS-4, in a scatter-gather list */\
        sg_tablesize:           0,\
/* max no. of sectors driver can accept in 1 SCSI READ/WRITE command */\
	max_sectors:		4*(MAX_IOV_SLOTS-4),\
	/* max no. of simultaneously outstanding commands per LUN */\
        cmd_per_lun:            MAX_COMMANDS_PER_LUN,\
/* no. of cards of this type found on this machine */\
        present:                0,\
        unchecked_isa_dma:      0,\
/* midlevel should merge requests for adjacent blocks of memory */\
        use_clustering:         ENABLE_CLUSTERING,\
	emulated:		0,\
/* proc_name xxx becomes the name for the directory /proc/scsi/xxx */\
	proc_name:		"iscsi_initiator"\
	} ;

#else /* K26 */

static struct scsi_host_template driver_template = {\
	proc_info:		iscsi_initiator_proc_info,\
	name:			OUR_NAME,\
        detect:                 iscsi_initiator_detect,\
        release:                iscsi_initiator_release,\
        info:                   NULL, \
	ioctl:			NULL,\
        queuecommand:           iscsi_initiator_queuecommand,\
        eh_abort_handler:       iscsi_initiator_abort,\
        eh_device_reset_handler:iscsi_initiator_reset,\
        eh_bus_reset_handler:	NULL,\
        eh_host_reset_handler:	NULL,\
/* use default scsicam_bios_param() in linux/drivers/scsi/scsicam.c */\
        bios_param:             NULL,\
/* max no. of simultaneously active SCSI commands driver can accept */\
        can_queue:              MAX_COMMANDS_PER_LUN * MAX_LUNS,\
        this_id:                -1, /* our host has no id on the SCSI bus */\
/* max no. of slots driver can accept in a scatter-gather list */\
        sg_tablesize:           MAX_IOV_SLOTS-4,\
/* max no. of sectors driver can accept in 1 SCSI READ/WRITE command */\
	max_sectors:		4*(MAX_IOV_SLOTS-4),\
	/* max no. of simultaneously outstanding commands per LUN */\
        cmd_per_lun:            MAX_COMMANDS_PER_LUN,\
/* no. of cards of this type found on this machine */\
        present:                0,\
        unchecked_isa_dma:      0,\
/* midlevel should merge requests for adjacent blocks of memory */\
        use_clustering:         ENABLE_CLUSTERING,\
	use_new_eh_code:	1,\
	emulated:		0,\
/* proc_name xxx becomes the name for the directory /proc/scsi/xxx */\
	proc_name:		"iscsi_initiator"\
	} ;

/*
 * 2.4.x HBA module usage
 */

#include <scsi_module.c>

#endif  /* K26 define */

#ifdef K26

/* New 2.6.0 HotPlug modifications */

#define to_iscsi_hostdata(d)       \
        container_of(d, struct iscsi_hostdata, dev)

/* Function declaration. */
static int unh_iscsi_probe(struct device *);
static int add_HBA(void);
static int unh_iscsi_bus_match(struct device *, struct device_driver *);
static int unh_iscsi_remove(struct device *);
static void unh_iscsi_device_release(struct device *);
static void remove_HBA(void);

static char unh_iscsi_proc_name[] = "unh_iscsi_initiator";

/* Place holder for all the HBA */
static LIST_HEAD (unh_iscsi_host_list);

/*
 * This describes each of the pseudo HBA
 */
struct unh_iscsi_host {
	struct list_head host_list;
	struct Scsi_Host *shost;
	struct device dev;
	struct list_head dev_info_list;
};

/*
 * This describes each of the devices attached
 * this is equivalent of sessions
 */
struct unh_iscsi_dev {
	struct list_head dev_list;
	unsigned int channel;
	unsigned int target;
	unsigned int lun;
	struct unh_iscsi_host *iscsi_host;
};


/*
 * This describes the bus type
 */
static struct bus_type unh_iscsi_bus = {
	.name = "unh_iscsi_pseudo",
	.match = unh_iscsi_bus_match,
};

/*
 * The driver structure
 */
struct device_driver unh_iscsi_driver = {
	.name           = unh_iscsi_proc_name,
	.bus		= &unh_iscsi_bus,
	.probe          = unh_iscsi_probe,
	.remove         = unh_iscsi_remove,
};

static struct device unh_iscsi_pseudo = {
        .bus_id         = "unh_iscsi_pseudo_0",
	.release	= unh_iscsi_device_release,
};


/*
 * probe- does most of the stuff that was part of _detect().
 */
static
int unh_iscsi_probe( struct device *dev)
{
	struct iscsi_hostdata *iscsi_host;
	struct Scsi_Host *hostptr;
	int retval=0;

	TRACE(TRACE_ENTER_LEAVE, "Enter unh_iscsi_probe\n");

	iscsi_host = to_iscsi_hostdata(dev);

	hostptr = scsi_host_alloc(&driver_template, 
			sizeof(struct iscsi_hostdata));

	if (unlikely(hostptr == NULL)) {
		printk("scsi_register failure\n");
		return -ENODEV;
	}
	
	/* setup the security key hash table */
	setup_security_hash_table();

	/* set the number of allowed "targets" this adaptor "card" can handle.
	 * the "target" numbers will be [0..max_id-1].
	 * each "target" will have LUN numbers in [0..max_lun-1].
	 */
	hostptr->max_id = MAX_TARGET_IDS;
	hostptr->max_lun = MAX_LUNS;

	/* set the maximum number of bytes we will accept in a scsi cdb */
	hostptr->max_cmd_len = MAX_CDB_LEN;

	iscsi_host->shost = hostptr;
	*((struct iscsi_hostdata **)hostptr->hostdata) = iscsi_host;

	retval = scsi_add_host(hostptr, &iscsi_host->dev);
	if (unlikely(retval)) {
		printk("scsi_add_host failed for unh\n");
		retval = -ENODEV;
		scsi_host_put(hostptr);
	} else {
	 	/* This should go away in future: TODO -AJ */
	 	global_hostdata = (struct iscsi_hostdata *) hostptr->hostdata;
	 	global_host = hostptr;

	 	/* Initialize the adapter's private data structure */
		init_initiator(global_hostdata);

		/* Register reboot notifier */
		register_reboot_notifier(&iscsi_initiator_notifier);
	}
	
	TRACE(TRACE_ENTER_LEAVE, "Leave unh_iscsi_probe\n");
	return retval;
}

static
int unh_iscsi_remove(struct device *dev)
{
	struct iscsi_hostdata *iscsi_host;

	TRACE(TRACE_ENTER_LEAVE, "Enter unh_iscsi_remove\n");
	
	iscsi_host = to_iscsi_hostdata(dev);

	if (unlikely(iscsi_host == NULL)) {
		printk("iscsi_host is NULL\n");
		return -ENODEV;
	}

	if (unlikely(iscsi_host->shost == NULL)) {
		printk("iscsi_host->shost is NULL !!\n");
		return -ENODEV;
	}

	iscsi_initiator_release(iscsi_host->shost);
	scsi_remove_host(iscsi_host->shost);
	scsi_host_put(iscsi_host->shost);

	TRACE(TRACE_ENTER_LEAVE, "Leave unh_iscsi_remove\n");
	return 0;
}

/*
 * Just to silence the warnings
 */
static
void unh_iscsi_device_release(struct device *dev)
{
	TRACE(TRACE_ENTER_LEAVE, "unh_iscsi_device_release not implemented\n");
	return ;
}

static int unh_iscsi_bus_match(struct device *dev,
			struct device_driver *dev_driver)
{
	TRACE(TRACE_ENTER_LEAVE, "unh_iscsi_bus_match not implemented\n");
	return 1;
} 

static
void unh_iscsi_release (struct device *dev)
{
	struct iscsi_hostdata *iscsi_host;
	TRACE(TRACE_ENTER_LEAVE, "Enter unh_iscsi_release \n");
	/* Free all memory allocation */
	iscsi_host = to_iscsi_hostdata(dev);
	kfree(iscsi_host);
	return;
}


/*
 * add_HBA.
 * In our design, we add a new HBA on the pseudo
 * bus that we create. As of now only one HBA is
 * added, changes can be easily made to add more
 * As of now there is no pressing reason to add
 * more than one (pseudo)HBA.
 */
static
int add_HBA()
{
	struct iscsi_hostdata *iscsi_host;
	int retval=0;

	TRACE(TRACE_ENTER_LEAVE, "Enter add_HBA \n");

	iscsi_host = kmalloc(sizeof(struct iscsi_hostdata), GFP_KERNEL);
	if (unlikely(iscsi_host == NULL)) {
		printk("add_HBA: allocated failed !\n");
		return -ENOMEM;
	}
	
	memset(iscsi_host, 0, sizeof(struct iscsi_hostdata));
/* Add locking code here */
	list_add_tail(&iscsi_host->host_list, &unh_iscsi_host_list);
/* Add unlocking code here */

	iscsi_host->dev.bus = &unh_iscsi_bus;
	iscsi_host->dev.parent = &unh_iscsi_pseudo;
	iscsi_host->dev.release = &unh_iscsi_release;
	sprintf(iscsi_host->dev.bus_id, "unh_iscsi%d",0);

	retval = device_register(&iscsi_host->dev);
	if (unlikely(retval)) {
		printk("add_HBA:device_register () failed %d\n", retval);
		kfree(iscsi_host);
		return retval;
	}

	return retval;
}


/*
 * remove_HBA.
 * We remove the HBA that we had added.
 */
static
void remove_HBA()
{
	struct iscsi_hostdata *iscsi_host=NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter remove_HBA \n");
	if (!list_empty(&unh_iscsi_host_list)) {
		iscsi_host = list_entry(unh_iscsi_host_list.prev,
					struct iscsi_hostdata,
					host_list);
		list_del(&iscsi_host->host_list);
	}
	if (unlikely(iscsi_host == NULL)) {
		printk("iscsi_host is null!!!!, not unregistering\n");
		return;
	}
	device_unregister(&iscsi_host->dev);
}

static int
__init unh_iscsi_init(void)
{
	TRACE(TRACE_ENTER_LEAVE, "Enter unh_iscsi_init \n");

	device_register(&unh_iscsi_pseudo);
	bus_register(&unh_iscsi_bus);
	driver_register(&unh_iscsi_driver);

	add_HBA();
	return 0;
}

static void
__exit unh_iscsi_exit(void)
{
//iscsi_trace_mask = TRACE_ALL;
	TRACE(TRACE_ENTER_LEAVE, "Enter unh_iscsi_exit \n");
	remove_HBA();
	driver_unregister(&unh_iscsi_driver);
	bus_unregister(&unh_iscsi_bus);
	device_unregister(&unh_iscsi_pseudo);
}
	
device_initcall(unh_iscsi_init);
module_exit(unh_iscsi_exit);

#endif  /* K26 hot-plug modifications */

/* Ming Zhang, mingz@ele.uri.edu */
#ifndef K26
EXPORT_NO_SYMBOLS;
#endif

#endif /* ifdef MODULES */

