/*	initiator/iscsi_initiator.h
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This file contains all the required header files, struct
 *      definitions and constants for iscsi initiator code.
 *
 *	Copyright (C) 2001-2004 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 * 	written permission.
*/
#ifndef	_ISCSI_INITIATOR_H
#define	_ISCSI_INITIATOR_H

/* RDR */
#ifndef K26
#include <linux/blk.h>
#endif

#include <linux/version.h>
#include <linux/blkdev.h>
#include <linux/kernel.h>
#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/unistd.h>
#include <linux/smp_lock.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <net/sock.h>
#include <asm/uaccess.h>
#include <linux/init.h>
#include <scsi.h>

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi.h>
#else
#include <sd.h>
#endif

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 8)
#include <hosts.h>
#else
#include <scsi/scsi_host.h>
#endif

#include <linux/module.h>

#include "../common/iscsi_common.h"
#include "../common/debug.h"
#include "../common/my_memory.h"
#include "../common/range.h"
#include "../common/crc.h"
#include "../common/tcp_utilities.h"

/* chap and srp support - CHONG */
#include "../security/misc/misc_func.h"
#include "../security/chap/chap.h"
#include "../security/srp/srp.h"

#include "../common/text_param.h"
#include "../common/initiator_negotiate.h"

/*
 *	the name of our module (driver, HBA) (OUR_NAME)
 *	is built dynamically during the build phase
 */
#include "version.h"

/*	number of seconds to wait for a task management response to an abort */
#define ABORT_TIMEOUT		(5 * HZ)

/*	maximum number of iov slots iscsi_initiator can set up for I/O	*/
#define MAX_IOV_SLOTS		68

/*	maximum number of commands scsi midlevel can queue for 1 lun */
#define MAX_COMMANDS_PER_LUN	8

/*	maximum number of targets this iscsi_initiator host will register for */
#define MAX_TARGET_IDS		10

/*	maximum number of luns each target can handle */
#define MAX_LUNS		8

/*	maximum number of bytes we will accept in a scsi cdb */
#define MAX_CDB_LEN		16

/*  Multiple connections Round Robin Scheduling Scheme - SAI */
#define CONN_SCHED_RR		1
#define CONN_SCHED_LUN		2

/* Add Maximum Connections allowed per session - SAI */
#define MAX_CONNECTIONS		20

/* Default (initial) settings for initiator's 6-byte ISID
 * RFC 3720 Section 10.12.5 ISID
 */
#define DEFAULT_ISID_TYPE	2		/* random */
#define DEFAULT_ISID_NUMBER	0x123456	/* 24-bit B&C */
#define DEFAULT_ISID_QUALIFIER	0x789a		/* 16-bit D */

/*	Holds pointer to private data part of Scsi_Host structure.
 *	This is set up once, by iscsi_detect()
 *	which is called when the initiator module is loaded.
 */
extern struct iscsi_hostdata *global_hostdata;

/*	Holds pointer to the registered Scsi_Host. */
extern struct Scsi_Host *global_host;


#ifdef UNH_LOCK_DEBUG

/* generate debugging code for UNH locks */

#if !defined(CONFIG_SMP)

/* generate debugging code for UNH locks on uniprocessors */

typedef struct unh_lock_type { int lock; int line; void *flags; } unh_lock_t;
#define UNH_LOCK_TYPE	unh_lock_t

#define UNH_LOCK_UNLOCKED	(unh_lock_t) { 0, 0, 0 }

#define UNH_LOCK(l, f)	do { \
			    local_irq_save(f); \
			    if ((l)->lock != 0) { \
				printk("%s:%d:   LOCK value %d should be 0!\n",\
					__FUNCTION__, __LINE__, (l)->lock); \
			    	printk("last LOCK line %d\n", (l)->line); \
			    } \
			    if ((l)->flags) { \
				printk("%s:%d: non-zero flags %p\n", \
				__FUNCTION__, __LINE__, (l)->flags); \
			    	printk("last LOCK line %d\n", (l)->line); \
			    } \
			    (l)->lock += 1; \
			    (l)->line = __LINE__; \
			    (l)->flags = &f; \
			} while(0)

#define UNH_UNLOCK(l,f)	do { (l)->lock -= 1; \
			    if ((l)->lock != 0) { \
				printk("%s:%d: UNLOCK value %d should be 0!\n",\
					__FUNCTION__, __LINE__, (l)->lock); \
			    	printk("last LOCK line %d\n", (l)->line); \
			    } \
			    if ((l)->flags != &f) { \
				printk("%s:%d flags %p expected %p\n", \
				__FUNCTION__, __LINE__, &(f), (l)->flags); \
			    	printk("last LOCK line %d\n", (l)->line); \
			    } \
			    (l)->flags = 0; \
			    local_irq_restore(f); \
			} while(0)
#else

/* generate debugging code for UNH locks on multiprocessors */

typedef struct unh_lock_type { spinlock_t lock; int line; void * flags; } unh_lock_t;
#define UNH_LOCK_TYPE	unh_lock_t

#define UNH_LOCK_UNLOCKED	(unh_lock_t) { {1}, 0, 0 }

#define UNH_LOCK(l,f)   do { if(TRACE_TEST(TRACE_SEM)) { \
				printk("%s:%d: Try %p\n", \
				__FUNCTION__, __LINE__, (l)); \
				spin_lock_irqsave(&(l)->lock,f); \
				printk("%s:%d: Got %p\n", \
				__FUNCTION__, __LINE__, (l)); \
			    } else { \
				spin_lock_irqsave(&(l)->lock,f); \
			    } \
			    if ((l)->flags) { \
				printk("%s:%d: non-zero flags %p\n", \
				__FUNCTION__, __LINE__, (l)->flags); \
			    	printk("%s last LOCK line %d\n", \
				current->comm, (l)->line); \
				spin_unlock_irqrestore (&(l)->lock,f); \
				BUG(); \
			    } \
			    (l)->line = __LINE__; \
			    (l)->flags = &f; \
			} while(0)

#define UNH_UNLOCK(l,f) do { \
			    if(TRACE_TEST(TRACE_SEM)) { \
				printk("%s:%d: Out %p\n", \
				__FUNCTION__, __LINE__, l); \
			    } \
			    if ((l)->flags != &(f)) { \
				printk("%s:%d flags %p expected %p\n", \
				__FUNCTION__, __LINE__, &(f), (l)->flags); \
			    	printk("%s last LOCK line %d\n", \
				current->comm, (l)->line); \
				spin_unlock_irqrestore (&(l)->lock,f); \
				BUG(); \
			    } \
			    (l)->flags = 0; \
			    spin_unlock_irqrestore (&(l)->lock,f); \
			} while(0)
#endif
#else

/* do NOT generate debugging code for UNH locks */

#define UNH_LOCK_TYPE		spinlock_t
#define UNH_LOCK_UNLOCKED	SPIN_LOCK_UNLOCKED

#define UNH_LOCK(l,f)		spin_lock_irqsave(l,f)
#define UNH_UNLOCK(l,f)		spin_unlock_irqrestore (l,f)

#endif


/*	Used for all access to the shared structures global_host and during
 *	the operation of any single thread to prevent things from
 *	disappearing from under the thread.
 */
extern UNH_LOCK_TYPE host_data_lock;

/*	values for the session_state field of the session structure */
enum session_states {

	SESSION_NOT_PRESENT,	/* set in build_session_skeleton()
				   as initial value */
	SESSION_LOGGED_IN,	/* set in iscsi_initiator_login() after a
				   successful return from
				   initiator_parameter_negotiate() */
	SESSION_FULL_FEATURE_PHASE,	/* set in create_session() after a
					   successful return from init_session()
					   where tx and rx threads for first
					   connection in session are confirmed
					   to be up ok */
	SESSION_LOGGED_OUT,	/* set in rx_logout_rsp() after receiving a
				   successful logout response to a request of
				   type LOGOUT_CLOSE_SESSION */
	SESSION_CLOSED		/* set at end of iscsi_initiator_rx_thread()
				   by first rx thread to detect
				   SESSION_LOGGED_OUT, or by the only rx thread
				   in a discovery session */
};

/*	values for the connection_state field of the connection structure */
enum connection_states {

	CONNECTION_NOT_PRESENT,	/* set in build_session_skeleton()
				   as initial value.
				   Also set in close_connection(),
				   and init_connection() after
				   connection is closed */
	CONNECTION_CONNECTED,	/* set in init_connection() after
				   socket has been connected ok */
	CONNECTION_LOGGED_IN,	/* set in iscsi_initiator_login() after
				   successful return from
				   initiator_parameter_negotiate(), so
				   login phase has completed ok */
	CONNECTION_FULL_FEATURE_PHASE,	/* set in init_connection() after rx
					   thread for connection is confirmed
					   to be up ok */
	CONNECTION_RECOVERING,	/* set in deal_with_connrec() after
				   Time2Wait expires */
	CONNECTION_LOGGED_OUT,	/* set in rx_logout_rsp() after a successful
				   logout response has been received to a
				   request of type LOGOUT_CLOSE_CONNECTION */
	CONNECTION_DISCONNECTED	/* set at end of iscsi_initiator_rx_thread()
				   after rx thread shuts down */
};

/*	values for the cmd_error_state field of the command structure */
enum cmd_error_states {
	CMD_OK_SO_FAR,		/* no errors yet */
	CMD_NEEDS_SNACK,	/* need to send snack when remaining
				   data ins stop arriving */
	CMD_USING_SNACK		/* when receiving data ins in response
				   to a snack */
};

#define WHITE_SPACE " \t\v\f\n\r"

/*	Holds information extracted from a real or implied R2T from target.
 *	Used by initiator to control sending DataOut PDUs to target.
 */
struct r2t_cookie {
	struct list_head link;
	struct list_head order_link;	/* order of this r2t on a connection */
	__u32 target_xfer_tag;	/*opaque -- in network order (from target r2t)*/
	__u32 offset;		/* SCSI buffer offset of next InOrder Data Out*/
	/* but is limit of next Out-of-Order Data Out */
	__u32 r2t_xfer_length;	/* no. of bytes left to send in r2t sequence  */
	/* or no. received so far in DataIn sequence  */
	__u32 data_out_sn;	/* incremented from 0 by DataOuts for this r2t*/
};

struct command {
	struct list_head link;	/* for threaded, doubly-linked circular lists */
	__u8 cmd_state;		/* state of this cdb/cmd */
	__u8 header_opcode;	/* copy of opcode when sending a pdu */
	__u8 header_flags;	/* copy of flags when sending a pdu */
	__u8 activity_flg;	/* Stores Activity from target for command SAI*/
	__u8 need_write_built;	/* Create write infrastruct within xmit thread*/
	__u8 header_complete;	/* ExpStatSN and header CRC have been done */
	__u32 init_task_tag;
	__u32 data_offset;	/* offset to next data to receive or r2t */
	__u32 recvd_length;	/* bytes received so far in DataIn sequence */
	__u32 still_to_send;	/* total data bytes left to sent in write cmd */
	__u32 data_in_sn;	/* incremented from 0 by DataIns for this cmd */
	__u32 cmd_msg_flags;	/* MSG_MORE on writes when more data follows */
	struct order_range r2t_range_list;	/* for out-of-order r2ts */
	struct order_range pdu_range_list;	/* for out-of-order DataIn pdu*/
	__u32 n_r2t_cookies;	/* no. of real (not implied) r2ts outstanding */
	__u32 r2t_sn;		/* incremented from 0 by r2ts */
	void *buffer_to_free;	/* free this buffer after xmit (if not NULL) */
	__u8 snack_flags;	/* Command response and SNACK reject bits SAI */
	__u8 cmd_error_state;	/* CMD_OK_SO_FAR, ... */
	__u8 task_mgt_response;
	__u8 task_mgt_function;
	__u32 cmd_error_begrun;	/* first missing DataSn to ask for in SNACK */
	__u32 cmd_error_offset;	/* offset of first missing DataIn pdu */
	__u32 cmd_error_runlen;	/* number of missing DataIn pdus to ask for */
	__u32 cmd_error_size;	/* size of good DataIn pdus since last bad one*/
	__u32 tx_data_length;	/* length of attached data for tx stats */
	struct iscsi_init_scsi_cmnd iscsi_cmd;	/* SCSI Cmnd PDU hdr blt here */

	/* structures from this point on are NOT zeroed on command setup!
	 * Therefore, do NOT move the position of "SCpnt" unless you change
	 * the memset() in setup_command()!
	 */
	struct scsi_cmnd *SCpnt;		/* pointer to SCSI command */
	struct list_head r2t_cookies;		/* list of unsatisfied r2ts */
	__u32 abit_in_sn;		/* saved data_in_sn for A-bit rexmit */
	__u32 abit_ttt;			/* saved target_xfer_tag for A-bit rx */
	__u64 abit_lun;				/* saved lun for A-bit rexmit */
	__u32 abit_offset;			/* saved offset for A-bit rx */
	__u32 ref_task_tag;			/* TASK REASSIGN needs this */
	__u32 time_stamp;			/* time when command xmitted */
	__u32 tx_wait_to_send;		/* deadline til next send of this cmd */
	__u32 tx_size;			/* no. of bytes to send in sendmsg */
	__u32 tx_sent_so_far;		/* no. of bytes sent in prev sendmsgs */
	__u32 retransmit_tx_size;	/* original command tx_size - SAI */
	__u32 tx_iovlen;			/* no. of used tx_iov[]slots  */
	struct iovec tx_iov[MAX_IOV_SLOTS];	/* socket iovector built here */
	struct iovec tx_iov_copy[MAX_IOV_SLOTS];/* copy of tx_iov for sendmsg */
	struct msghdr tx_msghdr;		/* socket message hdr blt here*/
	__u32 data_checksum;			/* checksum for attached data */
	struct iscsi_init_scsi_data_out data;	/* DataOut PDU hdr built here */
#ifdef K26
	struct scatterlist sg[MAX_IOV_SLOTS];
	__u32 pindex;
	__u32 padding;
#endif
};

/*
 * cmd_state state flags
 */

#define CMD_STATE_TXDONE	0x01	/* tx_thread is done with cmd */
#define CMD_STATE_RXDONE	0x02	/* rx_thread is done with cmd */
#define CMD_STATE_REXMIT_ORIG	0x04	/* rebuild orig cmd in case of rexmit */
#define CMD_STATE_TXSTARTED	0x08	/* tx_thread sent something on cmd */
#define CMD_STATE_ABORTING	0x10	/* command is being aborted */


#define CMD_NEEDS_WRITE_BUILT 0x01  /* xmit thread needs to call build_write */

/*
 * The following structure holds the statistic value,
 * it is a per session structure.
 */
#ifdef ISCSI_STATS

/*
 * These are some session wide stats
 */

struct iscsi_session_stats {
    __u32 iscsi_ssn_noptx_pdus; /* #of nop PDUs txd */
    __u32 iscsi_ssn_scsicmd_pdus; /* #of command PDUs transferred */
    __u32 iscsi_ssn_tskmgmtcmd_pdus; /* #of tast mgmt PDUs */
    __u32 iscsi_ssn_login_pdus; /* #of login req pdus */
    __u32 iscsi_ssn_text_pdus; /* #of test req pdus */
    __u32 iscsi_ssn_dataout_pdus; /* #of data out -WRITE PDUs */
    __u32 iscsi_ssn_logout_pdus; /* #of logout pdus */
    __u32 iscsi_ssn_snack_pdus; /* #of SNACK pdus */

    __u32 iscsi_ssn_noprx_pdus; /* #of nop PDUs rxd */
    __u32 iscsi_ssn_rsp_pdus; /* #of rsp PDUs transferred */
    __u32 iscsi_ssn_tskmgmtrsp_pdus; /* #of task mgmt func resp */
    __u32 iscsi_ssn_textrsp_pdus; /* #of text resp */
    __u32 iscsi_ssn_datain_pdus; /* #of data in -READ PDUs */
    __u32 iscsi_ssn_logoutrsp_pdus; /* #of logout pdus */
    __u32 iscsi_ssn_r2t_pdus;	/* #of r2t rxd */
    __u32 iscsi_ssn_async_pdus; /* #of async pdu rxd */
    __u32 iscsi_ssn_rjt_pdus; /* #of reject pdus rxd */

    __u32 iscsi_ssn_digest_err; /* PDUs with data/hdr digest error */
    __u32 iscsi_ssn_timeout_err; /* connections terminated due to timeouts */

    __u64 iscsi_ssn_txdata_octets; /* Data octets transmitted */
    __u64 iscsi_ssn_rxdata_octets; /* Data octets received */
};

#define UPDATE_TX_SESS_STATS(opcode, command, sess_stats, connection)	\
  do {\
    __u64 txlength;\
    switch(opcode) {\
	case ISCSI_INIT_NOP_OUT:\
		sess_stats->iscsi_ssn_noptx_pdus++;\
		break;\
	case ISCSI_INIT_SCSI_CMND:\
		sess_stats->iscsi_ssn_scsicmd_pdus++;\
	 	txlength = command->tx_data_length;\
		sess_stats->iscsi_ssn_txdata_octets += txlength;\
		connection->iscsi_conn_txdata_octets += txlength;\
		break;\
	case ISCSI_INIT_TASK_MGMT_CMND:\
		sess_stats->iscsi_ssn_tskmgmtcmd_pdus++;\
		break;\
	case ISCSI_INIT_TEXT_CMND:\
		sess_stats->iscsi_ssn_text_pdus++;\
		break;\
	case ISCSI_INIT_SCSI_DATA_OUT:\
		sess_stats->iscsi_ssn_dataout_pdus++;\
	 	txlength = command->tx_data_length;\
		sess_stats->iscsi_ssn_txdata_octets += txlength;\
		connection->iscsi_conn_txdata_octets += txlength;\
		break;\
	case ISCSI_INIT_LOGOUT_CMND:\
		sess_stats->iscsi_ssn_logout_pdus++;\
		break;\
	case ISCSI_INIT_SNACK:\
		sess_stats->iscsi_ssn_snack_pdus++;\
		break;\
	default:\
		break;\
    }\
 } while(0)


#define UPDATE_RX_SESS_STATS(opcode, size, sess_stats, connection)	\
    switch(opcode) {\
	case ISCSI_TARG_NOP_IN:\
		sess_stats->iscsi_ssn_noprx_pdus++;\
		break;\
	case ISCSI_TARG_SCSI_RSP:\
		sess_stats->iscsi_ssn_rsp_pdus++;\
		break;\
	case ISCSI_TARG_TASK_MGMT_RSP:\
		sess_stats->iscsi_ssn_tskmgmtrsp_pdus++;\
		break;\
	case ISCSI_TARG_TEXT_RSP:\
		sess_stats->iscsi_ssn_textrsp_pdus++;\
		break;\
	case ISCSI_TARG_SCSI_DATA_IN:\
		sess_stats->iscsi_ssn_datain_pdus++;\
		sess_stats->iscsi_ssn_rxdata_octets += size;\
		connection->iscsi_conn_rxdata_octets += size;\
		break;\
	case ISCSI_TARG_R2T:\
		sess_stats->iscsi_ssn_r2t_pdus++;\
		break;\
	case ISCSI_TARG_ASYNC_MSG:\
		sess_stats->iscsi_ssn_async_pdus++;\
		break;\
	case ISCSI_TARG_RJT:\
		sess_stats->iscsi_ssn_rjt_pdus++;\
		break;\
	default:\
		break;\
    }


/*
 * These are what the SNMP Stats care about
 */

enum login_failure_type {
    UNKNOWN_FAILURE =0,
    ACCEPT_RESPONSE,
    REDIRECT_RESPONSE,
    AUTH_FAIL_RESPONSE,
    AUTHENTICATE_FAIL,
    NEGOTIATE_FAIL,
    OTHER_FAIL
};

enum session_failure_type {
    UNKNOWN_ERROR = 0,
    DIGEST_ERROR,
    ISCSI_TIMEOUT_ERROR,
    FORMAT_ERROR
};

/*
 * These are some of the Instance wide stats
 */

struct iscsi_instance_stats {
    __u32 iscsi_intr_ssn_failure; /* # time the sessions failed */
    enum session_failure_type iscsi_last_ssn_failure_type; /* Last session failure type */
    __u32 iscsi_ssn_digest_error; /* Sess which were failed due to receipt of a
				     PDU containing header/data digest error */
    __u32 iscsi_ssn_csn_timeout_error; /* Sess which were failed due to sequence
					  exceeding a time limit */
    __u32 iscsi_ssn_format_error; /* Sess which were failed due to receipt
				     of aPDU containing format error */

    __u32 iscsi_intr_login_failure; /* # login attempt from this initiator failed */

    enum login_failure_type iscsi_intr_last_failure_type; /* Last failure type */
    __u32 iscsi_intr_login_accept_rsp; /* Login rsp pdus with status 0x000 */
    __u32 iscsi_intr_login_redirect_rsp; /* Login rsp pdus with status class 0x01 */
    __u32 iscsi_intr_login_auth_fail_rsp; /* Login rsp pdus with status class 0x201 */
    __u32 iscsi_intr_login_authenticate_fails; /* Intr aborted because tgt couldn't be authenticated */
    __u32 iscsi_intr_login_negotiate_fails; /* parameter neg. with tgt failed */
    __u32 iscsi_intr_login_other_fail_rsp; /* login rsp pdus with status not defined below */

    __u32 iscsi_intr_logout_normals; /* logout cmd pdus  with reason code 0x0 */
    __u32 iscsi_intr_logout_others; /* logout cmd pdus with reason code other than 0 */

    struct timeval iscsi_intr_last_failure_time; /* timestamp of most recent login failure */
    char *iscsi_intr_last_tgt_failure_name; /* Target name */
    struct sockaddr *iscsi_intr_last_tgt_failure_address; /* Target address */
};

#endif


/*	one of these per connection */
struct connection {
	unsigned long rx_lock_flags;	/* irqsave when rx_thread holds lock */
	unsigned long tx_lock_flags;	/* irqsave when tx_thread holds lock */
	__u32 connection_id;		/* CID within session */
	__u32 max_send_length;		/* MaxRecvDataSegmentLength from targ */
	__u32 max_recv_length;		/* MaxRecvDataSegmentLength of init */
	__u32 nop_counter;		/* counter used in NopOut ping data */
	__u32 exp_stat_sn;		/* next StatSN expected from target */
	__u32 increment;		/* added to exp_stat_sn after recovery*/
	struct session *my_session;	/* ptr to sess this conn belongs to */
	__u32 connection_flags;		/* bits for hdr/data digests, activity*/
	__u32 connection_state;		/* CONNECTION_NOT_PRESENT, ... */
	struct semaphore rx_done_sem;	/* wait here for rx_thread start/stop */
	struct semaphore tx_sem;	/* tx_thread waits here for work */
	struct semaphore tx_done_sem;	/* wait here for tx_thread start/stop */
	struct semaphore task_mgt_sem;	/* wait here for task mgt resp/timeout*/

	__u32 nop_remaining;		/* no. of secs until next NopOut sent */
	atomic_t tx_timer_went_off;	/* no. of pending timer expirations */
	struct timer_list tx_timer;	/* timer for sending nops */

	/* the following field is manipulated by add_(remove_)session only */
	struct connection *next;	/* list of connections in this session*/

	struct socket *sock;	/* pointer to open socket to target */
	__u8 *rx_buf;		/* PDU headers from target read into here */

	/* the following 2 fields are only set up after login succeeds */
	struct task_struct *rx_thread;	/* pointer to TCB of rx_thread */
	struct task_struct *tx_thread;	/* pointer to TCB of tx_thread */

	int rec_tests;		/* Indicator to initiate conn recovery SAI */

	/* the following 5 fields are used by the rx_thread during operation */

	struct list_head pending_commands;/* SCSI commands active on this conn*/
	struct msghdr rx_msg;	/* header for sock_recvmsg on this conn */
	struct iovec rx_iov[MAX_IOV_SLOTS];	/* recvmsg iovec built here */
	struct iovec rx_iov_copy[MAX_IOV_SLOTS];/* copy of rx_iov if digest */
	struct list_head r2t_list;	/* threaded order to send r2ts */

	/* following 2 fields are used by tx_thread during operation */
	struct command *short_command;	/* Non-null if partially sent command */
	__u32 process_r2t_order;	/*next r2t to process in order on conn*/

	struct sockaddr	*ip_address;	/* holds target's ipv4/ipv6 address */
	int ip_length;			/* used no. of bytes of ip_address */
	struct sockaddr	*local_ip_address;/* holds init's ipv4/ipv6 address */

	/* the following field is only used for CONN_SCHED_LUN scheduling */
	__u32 assigned_lun_count;	/* no. of luns assigned to this conn */

	__u32 basic_hdr_len;		/* includes CRC_LEN as necessary */

#ifdef ISCSI_STATS
	struct proc_dir_entry *proc_dir; /* Proc dir entry */
    	__u64 iscsi_conn_txdata_octets; /* Data octets transmitted */
    	__u64 iscsi_conn_rxdata_octets; /* Data octets received */
#endif
};

/*	one of these structures for each initiator session */
struct session {
	struct list_head link;		/* for threaded-list of sessions */
	UNH_LOCK_TYPE sess_lock;	/* spin lock for each session */

	/* the following fields are used to store free command structures */
	struct list_head free_commands;
	__u16 n_commands_alloced;	/* no. of commands alloced to session */
	__u16 n_luns_in_session;	/* no. of active luns in session */

	/* the following field is used to store free r2t cookies */
	struct list_head free_r2t_cookies;

	__u32 scsi_target_id;	/* target in [host,bus,target,lun] */
	__u32 lun_bits;		/* bits for lun in [host,bus,target,lun] */
	__u32 session_state;	/* SESSION_NOT_PRESENT, ... */
	__u32 cur_task_tag;	/* generator for unique ITTs */
	__u32 cur_cmd_sn;	/* generator for sequential CmdSNs */
	__u32 exp_cmd_sn;	/* last ExpCmdSN from target */
	__u32 max_cmd_sn;	/* last MaxCmdSN from target */

	struct timer_list connrec_timer;   /* Time2wait timer - SAI */

	/* the following fields are initialized from configured hostdata */
	__u32 session_flags;	/* flags for session-wide control (none yet) */
	__u32 init_conn_flags;	/* initial connection-specific flags */
	__u32 nop_period;	/* no. of secs between sending NopOuts */
	__u32 init_snack_flag;	/* Store SNACK Related flags here - SAI */
	int retran_period;	/* retransmission threshold in sec - SAI */
	int sched_scheme;	/* connection scheduling scheme to use - SAI */

	struct parameter_type (*session_params)[MAX_CONFIG_PARAMS];

#ifdef ISCSI_STATS
	struct proc_dir_entry *proc_dir;	/* Proc dir entry */
	struct iscsi_session_stats *sess_stats;	/* per session Stats */
#endif

	/* the following 2 fields are manipulated by add_(remove_)session only*/
	__u32 nconnections;		/* no. of connections in next list */
	struct connection *connection_head;/* list of connections in this sess*/

	struct connection *rr_conn_head;   /* round robin connection head SAI */

	/* the following field is allocated late */
	struct session_operational_parameters *oper_param;

	/* the following field is returned by target if first login succeeds */
	__u16 tsih;		/* target's handle for this session */

	int n_rec_tasks;	/* stores no. of tasks that were transfered from
				 * a failing connection - SAI
				 */
	/* the following table is used only for CONN_SCHED_LUN scheduling */
	/* index by LUN to find assigned connection */
	struct connection *lun_assignments[MAX_LUNS];

	/* the following 2 arrays are used to keep track of inquiry data */
	/* index by LUN to get latest inquiry info */
	char *inquiry_buf[MAX_LUNS];	/* buffer to hold inquiry data */
	int inquiry_size[MAX_LUNS];	/* no of bytes alloced to inquiry_buf */

	/* the following 2 arrays are used to keep track of read capacity data*/
	/* index by LUN to get latest read capacity info */
	__u32 capacity_lba[MAX_LUNS];	/* returned logical block address */
	__u32 capacity_len[MAX_LUNS];	/* block length in bytes */

	/* the following 2 arrays are used to keep track of block limits data */
	/* index by LUN to get latest block limits info */
	__u32 limit_max[MAX_LUNS];	/* max block length limit */
	__u32 limit_min[MAX_LUNS];	/* min block length limit */
};

/*	Target specific configuration data */
struct iscsi_targetdata {
	__u32 nop_period;	/* timer value to force NopOuts */
	__u32 init_sess_flags;	/* initial session-wide flags */
	__u32 init_conn_flags;	/* initial connection-specific flags */
	__u8 init_snack_flag;	/* holds SNACK related flags - SAI */

	__u32 isid_type;	/* holds isid type, max of 0x03 */
	__u32 isid_number;	/* holds isid number, max of 0xffffff */
	__u32 isid_qualifier;	/* holds isid qualifier, max of 0xffffff */

	int retran_period;	/* holds rexmit threshold in sec SAI */
	int sched_scheme;	/* holds scheduling algorithm to use
				   with multiple connections - SAI */
	/* chap support -CHONG */
	struct auth_parameter_type auth_parameter;

	struct parameter_type param_tbl[MAX_CONFIG_PARAMS];
};

/*	One of these structures forms the private data part for each
 *	iscsi adaptor.  Since we have only 1 such adaptor, we will
 *	instantiate this structure only once, when the iscsi_initiator
 *	module is first loaded.  At that time the variable global_hostdata
 *	will be set to point to the instantiation, and the fields will be
 *	set up to the defaults (all 0 except for param_tbl, which gets a copy
 *	of config_params[]).  After that, whenever the user changes parameter
 *	values using iscsi_manage, those changes are made to the values in this
 *	structure.
 */
struct iscsi_hostdata {
	__u32 nsessions;		/* no. of targets/sessions so far */
	struct list_head session_list;	/* list of threaded struct sessions */

	/* recovery session list - SAI */
	struct sess_rec_group *sess_rec_head;

	int init_sessrec_flg;		/* 1 if sess recovery in progress SAI */
	/* puts scsi to sleep in sess recovery - SAI */
	struct semaphore init_sessrec_sem;
	struct iscsi_targetdata target_data[MAX_TARGET_IDS];
#ifdef K26   /* for 2.6.0 hot-plug HBA model */
        struct list_head host_list; 
        struct device dev;
        struct Scsi_Host *shost;
#endif
#ifdef ISCSI_STATS
	struct iscsi_instance_stats instance_stats; /* Login statistics */
#endif
};

/* Add the Error Recovery Structure - SAI */

struct init_error_rec {
	struct connection *curr_conn;
	struct command *related_cmd;
	__u8 err_type;
};

/*******************************************************************/
/*	the following functions are defined in initiator_proc_ioface.c */
/*******************************************************************/

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
int iscsi_initiator_proc_info(struct Scsi_Host *, char *, char **, off_t, int,
				int);
#else
int iscsi_initiator_proc_info(char *, char **, off_t, int, int, int);
#endif

/**********************************************************/
/*	the following functions are defined in initiator_rx.c */
/**********************************************************/

void iscsi_initiator_rx_thread(void *arg);

/**********************************************************/
/*	the following functions are defined in initiator_tx.c */
/**********************************************************/

/*	Called to send a NopOut to target as:
 *		a new nop (nopin_cmd == NULL, size == 0 )
 *	or
 *		a new ping (nopin_cmd == NULL, size != 0 )
 *	or
 *		a response to a NopIn (nopin_cmd != NULL).
 *	Buffer points to data to send, size is number of bytes of data to send.
 *	If size > 0 then better have buffer != NULL.
 *	Returns 1 if all ok, else 0.
 */
int tx_nopout(struct connection *current_connection,
	      __u32 i_bit,
	      struct iscsi_targ_nopin *nopin_cmd, char **buffer, int size);

/*	drives sending of a nop out to the target */
int drive_nopout(struct connection *conn,
		 struct session *current_session, __u32 i_bit, __u32 want_data);

/*	drives the discovery session from within the tx_thread
 *	returns 0 on success, 1 on failure
 */
int drive_text_negotiate(struct connection *conn,
			 struct session *current_session,
			 __u32 i_bit,
			 __u32 discovery);

/*	drives the logout of a session from within the tx_thread
 *	returns 1 on success, 0 on failure
 *	Modified to include connection logout - SAI
 */
int drive_logout(struct session *current_session, struct connection *conn,
		 int reason);

void iscsi_initiator_tx_thread(void *);

/*************************************************************
 *  the following functions are defined in iscsi_initiator.c *
 *************************************************************/

/* initialize all fields in struct iscsi_targetdata for one target */
void init_targetdata(struct iscsi_targetdata *targ_data);

/*	find a lun that is already in use in this session
 *	returns -1 on error, else lun number in range [0..31]
 */
int
find_used_lun(struct session *sess);

/* uninitialize fields in struct iscsi_targetdata that need freeing */
void uninit_targetdata(struct iscsi_targetdata *targ_data);

/* Create a new recovery connection to keep the iSCSI session active.
 * Remember, iscsi requires at least one connection in a session to be active
 * for the session to be alive - SAI
 * Returns 1 on success, 0 on failure.
 */
int
create_rec_conn(struct connection *conn);

void __attribute__ ((no_instrument_function))
clear_free_conn(struct session *sess);

/* Declaration needed as this function used by Error Recovery - SAI */
struct connection * __attribute__ ((no_instrument_function))
get_connection(struct scsi_cmnd * Cmnd, struct session *current_session);

int create_session(__u32 target, struct sockaddr *ip_address,
		   int ip_length, __u32 lun, __u32 cid,
		   struct Scsi_Host *host);

int remove_session_from_hostdata(__u32 target, __u32 lun, __u32 cid,
				 struct iscsi_hostdata *hostdata);

int find_connection(__u32 target,
		    __u32 cid,
		    struct iscsi_hostdata *hostdata,
		    struct connection **old_connection,
		    struct session **old_session);

/*	Mike Christie, mikenc@us.ibm.com
 *
 * remove all valid luns in session (last lun will take out session)
 * returns smallest lun number removed, -1 if none removed.
 */
int
remove_luns(struct session *sess, struct Scsi_Host *host,
	    struct iscsi_hostdata *hostdata);

/* SAI */
int scsi_to_iscsi(struct scsi_cmnd * Cmnd, struct session *current_session);

#ifdef CONFIG_ISCSI_DEBUG
int __get_cpu_number(void);
#endif

/* From iscsi_device.c */
extern int initiator_register_device(void);
extern void initiator_unregister_device(void);

#ifdef K26
static inline int __attribute__ ((no_instrument_function))
use_sendpage(struct connection *conn, struct command *cmnd)
{
	if (!(conn->connection_flags & USE_DATADIGEST) && cmnd->SCpnt != NULL
		&& cmnd->SCpnt->use_sg > 0)
		return 1;
	else
		return 0;
}
#endif

#if !defined(list_for_each_entry)

/* include/linux/list.h is out of date */

/**
 * list_for_each_entry	-	iterate over list of given type
 * @pos:	the type * to use as a loop counter.
 * @head:	the head for your list.
 * @member:	the name of the list_struct within the struct.
 */
#define list_for_each_entry(pos, head, member)				\
	for (pos = list_entry((head)->next, typeof(*pos), member),	\
		     prefetch(pos->member.next);			\
	     &pos->member != (head); 					\
	     pos = list_entry(pos->member.next, typeof(*pos), member),	\
		     prefetch(pos->member.next))

#endif

/* searches the session_list pointed to by hostdata to try to find
 * a session with id that matches target_id
 * Returns pointer to that session if found, NULL if not found
 */
static inline struct session * __attribute__ ((no_instrument_function))
find_session_by_id(__u32 target_id, struct iscsi_hostdata *hostdata)
{
	struct session * sess;

	list_for_each_entry(sess, &hostdata->session_list, link) {
		if (sess->scsi_target_id == target_id) {
			return sess;
		}
	}

	/* if loop finishes, no session with this id was found */
	return NULL;
}

#endif				/* _ISCSI_INITIATOR_H */
