/*	target/target_error_rec.c 

    vi: set autoindent tabstop=4 shiftwidth=4 :

*/

/*
	Copyright (C) 2001-2003 InterOperability Lab (IOL)
	University of New Hampshier (UNH)
	Durham, NH 03824

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2, or (at your option)
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
	USA.

	The name of IOL and/or UNH may not be used to endorse or promote products
	derived from this software without specific prior written permission.
*/

#include "iscsi_target.h"
#include "target_error_rec.h"

static struct semaphore targ_sess_recovery_sem;

/************************************************************************
 *  The Main Error Recovery Handler Called from iscsi_initiator when it *
 *  finds any error.                                                    *
 *  Returns 0 if recovery completed ok, -1 otherwise                    *
 ************************************************************************/

int
targ_do_error_recovery(struct targ_error_rec *err_rec)
{

	int retval = 0;
	struct iscsi_conn *curr_conn;

	TRACE(TRACE_ENTER_LEAVE, "Entering targ_do_error_recovery\n");

	if (err_rec == NULL) {
		TRACE_ERROR("%s Error Recovery structure is NULL\n",
					current->comm);
		retval = -1;
		goto out;
	}

	if (err_rec->curr_conn == NULL) {
		TRACE_ERROR("%s NULL Connection Pointer in Error Recovery structure\n",
					current->comm);
		retval = -1;
		goto out;
	}

	curr_conn = err_rec->curr_conn;

	switch (curr_conn->session->oper_param->ErrorRecoveryLevel) {
	case SESSION_RECOVERY:
		targ_session_recovery(curr_conn);
		break;
	case DIGEST_RECOVERY:
		retval = targ_digest_recovery(err_rec);
		break;
	case CONNECTION_RECOVERY:
		TRACE_ERROR("%s ErrorRecoveryLevel %u Not Implemented yet\n",
					current->comm, CONNECTION_RECOVERY);
		retval = -1;
		break;
	}

out:
	TRACE(TRACE_ENTER_LEAVE, "Leaving targ_do_error_recovery, retval %d\n",
		retval);
	return retval;
}

/************************************************************************
* This function handles the session recovery. It creates a recovery     *
* thread that will do session recovery.                                 *
************************************************************************/

void
targ_session_recovery(struct iscsi_conn *current_connection)
{
	TRACE(TRACE_ENTER_LEAVE, "Enter targ_session_recovery\n");

	TRACE(TRACE_ISCSI,
		  "Executing Target Session Recovery - cancelling Receive Thread\n");

	init_MUTEX_LOCKED(&targ_sess_recovery_sem);

	/* Create a recovery thread */
	if (kernel_thread(targ_recovery_thread,(void *)current_connection,0) < 0) {
		TRACE_ERROR("%s Unable to start targ_recovery_thread\n", current->comm);
	} else {
		/* Wait till the recovery thread completes processing */
		down_interruptible(&targ_sess_recovery_sem);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave targ_session_recovery\n");

	return;
}

/************************************************************************
* This thread handles the session recovery. Session Recovery will       *
* abort all pending tasks and closes all transport connection           *
* Section (6.12.4)                                                      *
************************************************************************/

int
targ_recovery_thread(void *arg)
{
	struct iscsi_conn *conn;
	struct iscsi_session *sess;
	struct iscsi_global *host;

	conn = (struct iscsi_conn *)arg;
	sess = conn->session;
	host = sess->devdata;

	lock_kernel();

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	daemonize(
#else
	daemonize();
	snprintf(current->comm, sizeof(current->comm),
#endif
			 "iscsi_reco_%d", host->ntsih);

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 18)
	/*
	 * Arne Redlich, agr1@users.sourceforge.net:
	 * This prevents the retran thread from becoming a zombie after it exits
	 */
	reparent_to_init();
#endif

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGBITS);

	unlock_kernel();

	printk("%s Starting pid %d\n", current->comm, current->pid);

	if (!down_interruptible(&host->session_sem)) {
		iscsi_release_session(sess);
		up(&host->session_sem);
	}

	up(&targ_sess_recovery_sem);
	printk("%s Exiting pid %d\n", current->comm, current->pid);
	return 0;
}

/************************************************************************
 * This function handles the digest errors and sequence errors that may *
 * have caused because of previous digest errors.                       *
 * Returns 0 if recovery is completed correctly, -1 otherwise           *
 ************************************************************************/

int
targ_digest_recovery(struct targ_error_rec *err_rec)
{
	int retval = 0, data_length = 0;
	int size = 0, err_type = 0, data_offset = 0;
	unsigned int opcode = 0;
	struct iscsi_conn *curr_conn;
	struct iscsi_init_scsi_data_out *data_out_hdr;
	struct iscsi_cmnd *cmd;

	curr_conn = err_rec->curr_conn;
	err_type = err_rec->err_type;
	cmd = err_rec->cmd;

	TRACE(TRACE_ENTER_LEAVE, "Entering targ_digest_recovery\n");

	if (err_rec->pdu_hdr == NULL) {
		TRACE_ERROR("%s iscsi NULL PDU Header\n", current->comm);
		retval = -1;
		goto digest_out;
	} else {
		opcode = err_rec->pdu_hdr->opcode;
		size = err_rec->pdu_hdr->length;
		if (size > MASK_24_BITS) {
			TRACE_ERROR("%s TotalAHSLength 0x%02x, expected 0x00\n",
						current->comm, size >> 24);
			retval = -1;
			goto digest_out;
		}
	}

	if (err_type == HEADER_DIGERR) {
		if (size > 0) {
			retval = targ_drop_pdu_data(curr_conn, size);
		}
		goto digest_out;
	}

	if (opcode == ISCSI_INIT_SCSI_DATA_OUT) {
		if (cmd == NULL) {
			retval = -1;
			goto digest_out;
		}

		if (err_type == PAYLOAD_DIGERR) {
			enqueue_reject(curr_conn, REASON_DATA_DIGEST_ERR);
			retval = send_recovery_r2t(cmd, cmd->data_done, NULL,
						   err_rec->pdu_hdr);
		} else if (err_type == SEQUENCE_ERR) {
			/* error recovery ver ref18_04 - SAI */
			/* if there is a sequence error and the arrived data 
			 * seq. indicates lost packets, queue the received 
			 * packets and request re-transmission of lost packets 
			 */
			queue_data(err_rec);

			data_length = cmd->data_length - cmd->data_done;
			data_offset = cmd->data_length - data_length;

			data_out_hdr = (struct iscsi_init_scsi_data_out *) err_rec->pdu_hdr;
			size = (data_out_hdr->offset - data_offset);

			retval = send_recovery_r2t(cmd, data_offset, NULL,
						   err_rec->pdu_hdr);

			//up(&curr_conn->tx_sem);

		} else {
			TRACE_ERROR("%s Unknown err_type %d for DataOut recovery\n",
						current->comm, err_type);
			retval = -1;
		}

	} else if (opcode == ISCSI_INIT_SCSI_CMND) {
		if (err_type == PAYLOAD_DIGERR) {
			/* digest error in immediate data attached to write */
			enqueue_reject(curr_conn, REASON_DATA_DIGEST_ERR);
		} else {
			TRACE_ERROR("%s Unknown err_type %d for SCSI_CMND recovery\n",
						current->comm, err_type);
			retval = -1;
		}
	} else {
		TRACE_ERROR("%s Unknown err_type %d for recovery\n", current->comm,
					err_type);
		retval = -1;
	}

digest_out:

	TRACE(TRACE_ENTER_LEAVE, "Leaving targ_digest_recovery, retval %d\n",
		retval);

	return retval;
}

/*******************************************************************
 * Reads the data portion of the dropped PDU due to digest errors. *
 * The data read is also ignored.                                  *
********************************************************************/

int
targ_drop_pdu_data(struct iscsi_conn *curr_conn, int size)
{
	int padding = 0, retval = 0, data_length = 0;
	int total_rx = 0, rx_loop = 0;
	char *buffer = NULL;
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t oldfs;

	TRACE(TRACE_ENTER_LEAVE, "Entering targ_drop_pdu_data\n");

	padding = -(size) & 3;
	size += padding;

	if (curr_conn->data_crc == 1) {
		size += CRC_LEN;
	}

	TRACE(TRACE_DEBUG, "targ_drop_pdu: data_read %d total_data %d\n",
		  data_length, size);

	data_length = size;

	while (data_length > 0) {
		if (data_length > MAX_MALLOC_SIZE)
			data_length = MAX_MALLOC_SIZE;

		buffer = kmalloc(data_length * sizeof(char), GFP_KERNEL);
		if (!buffer) {
			TRACE_ERROR("%s Malloc Error: Unable to drop the data PDU\n",
						current->comm);
			retval = -ENOMEM;
			goto end_drop_pdu;
		}

		for (total_rx = 0; total_rx < data_length;) {
			memset(&msg, 0, sizeof(struct msghdr));
			msg.msg_iovlen = 1;
			msg.msg_iov = &iov;

			iov.iov_len = data_length - total_rx;
			iov.iov_base = buffer + total_rx;

			oldfs = get_fs();
			set_fs(get_ds());
			rx_loop = sock_recvmsg(curr_conn->conn_socket, &msg,
								   (data_length - total_rx), MSG_WAITALL);
			set_fs(oldfs);

			if (rx_loop <= 0) {
				TRACE_ERROR("%s targ_drop_pdu: rx_loop %d total_rx %d\n",
							current->comm, rx_loop, total_rx);
				retval = -ECONNRESET;	/* Connection reset by peer */
				kfree(buffer);
				goto end_drop_pdu;
			}
			total_rx += rx_loop;

			TRACE(TRACE_DEBUG, "targ_drop_pdu: rx_loop %d total_rx %d\n",
				  rx_loop, total_rx);
		}

		size -= data_length;
		data_length = size;

		TRACE(TRACE_DEBUG,"targ_drop_pdu: data_read %d total_data %d\n",
			  data_length, size);

		kfree(buffer);
	}

end_drop_pdu:

	TRACE(TRACE_ENTER_LEAVE, "Leaving targ_drop_pdu_data\n");

	return retval;
}

void deal_with_r2t_timer(unsigned long data);

/*********************************************************************
 * (re)start R2T Timer is used for tracking the R2T sent. This timer
 * helps re-transmit the most recent R2T sent to the initiator.
 ********************************************************************/
/* (re)start the R2T retransmit timer */
void
restart_r2t_timer(struct iscsi_session *session)
{
	__u32 period = session->r2t_period;

	TRACE(TRACE_TIMERS, "Enter restart_r2t_timer period %u\n", period);

	if (period && session->r2t_timer && !timer_pending(session->r2t_timer)){
		TRACE(TRACE_TIMERS, "Start r2t timer for %u ticks\n", period);
		      session->r2t_timer->expires = jiffies + period;
		      session->r2t_timer->data = (unsigned long) session;
		      session->r2t_timer->function = deal_with_r2t_timer;
		      add_timer(session->r2t_timer);
	} else {
		TRACE_ERROR("%s R2T timer not started for %u ticks\n", current->comm,
					period);
	}

	TRACE(TRACE_TIMERS, "Leave restart_r2t_timer\n");
}

/*********************************************************************
* Called by kernel when R2T periodic session timer expires.
*********************************************************************/
void
deal_with_r2t_timer(unsigned long data)
{
	struct iscsi_session *session;
	struct iscsi_cmnd *cmnd = NULL;

	TRACE(TRACE_TIMERS, "Enter deal_with_r2t_timer\n");

	session = (struct iscsi_session *) data;

	for (cmnd = session->cmnd_list; cmnd != NULL; cmnd = cmnd->next) {
		if (cmnd->outstanding_r2t > 0 &&
			(jiffies > (cmnd->timestamp + session->r2t_period))) {
			up(&session->retran_sem);
			break;
		}
	}

	restart_r2t_timer(session);

	TRACE(TRACE_TIMERS, "Leave deal_with_r2t_timer\n");
}

/***************************************************************************
* create an R2T cookie and store the R2T details in the cookie. This cookie
* is esssential for R2T re-transmssions in error conditions.
***************************************************************************/
struct iscsi_cookie *
create_r2t_cookie(struct iscsi_cmnd *cmnd)
{
	struct iscsi_cookie *cookie = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter create_r2t_cookie\n");

	cookie = (struct iscsi_cookie *)kmalloc(sizeof(struct iscsi_cookie),
						GFP_KERNEL);
	if (cookie == NULL) {
		TRACE_ERROR("%s Malloc problem in creating R2T cookie\n",
					current->comm);
		goto end_iscsi_cookie;
	}
	cookie->next = NULL;

	if (cmnd->first_r2t_cookie == NULL) {
		cmnd->first_r2t_cookie = cmnd->last_r2t_cookie = cookie;
	} else {
		cmnd->last_r2t_cookie->next = cookie;
		cmnd->last_r2t_cookie = cookie;
	}

end_iscsi_cookie:

	TRACE(TRACE_ENTER_LEAVE, "Leave create_r2t_cookie\n");

	return cookie;
}

/* free the long pending R2T cookie list after the command is completed - SAI */
void
free_r2t_cookie(struct iscsi_cmnd *cmnd)
{
	struct iscsi_cookie *cookie = cmnd->first_r2t_cookie;
	struct iscsi_cookie *tmp_cookie = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter free_r2t_cookie\n");

	while (cookie) {
		tmp_cookie = cookie;
		cookie = cookie->next;
		kfree(tmp_cookie);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave free_r2t_cookie\n");

	return;

}


/*******************************************************************************
* This function uses the offset and data length parameter or cookie information
* to build a recovery R2T to be sent to the intiator for missing Dataouts. This
* function is called by SNACK mechanism and also by other places in the code
* that find a discrepency in Dataouts using data sequence numbering or offsets.
* with regard to R2TSN updating for recovery R2Ts:
*
* RFC 3720 Section 2.1. Definitions
* "- Recovery R2T: An R2T generated by a target upon detecting the loss
*    of one or more Data-Out PDUs through one of the following means: a
*    digest error, a sequence error, or a sequence reception timeout.  A
*    recovery R2T carries the next unused R2TSN, but requests all or
*    part of the data burst that an earlier R2T (with a lower R2TSN) had
*    already requested."
********************************************************************************/
int
send_recovery_r2t(struct iscsi_cmnd *cmnd, int data_offset,
		  struct iscsi_cookie *cookie, struct generic_pdu *hdr)
{
	int data_length;
	int max_burst_len;
	int err = 0;

	TRACE(TRACE_ENTER_LEAVE, "Enter send_recovery_r2t, recovery_r2t %u\n",
	      cmnd->recovery_r2t);

	if (cmnd->recovery_r2t) {
		/* don't send recovery_r2t for another error */
		if (hdr->flags & F_BIT) {
			/* last DataOut in sequence to be recovered */
			cmnd->recovery_r2t = 2;
			cmnd->next_burst_len = 0;
			cmnd->data_sn = 0;
			goto out_wakeup;
		}
		goto out;
	}

	max_burst_len = cmnd->conn->session->oper_param->MaxBurstLength;
	data_length = cmnd->data_length - data_offset;

	if (!cookie) {
		cmnd->r2t_data = data_length;
		cookie = create_r2t_cookie(cmnd);
		if (!cookie) {
			err = -1;
			goto out;
		}
		cmnd->startsn = cmnd->endsn = cmnd->r2t_sn;
		cookie->seq = cmnd->r2t_sn++;	/* recovery R2T stays in R2T sequence */
		cookie->offset = data_offset;
		cookie->xfer_len = (data_length <= max_burst_len)
					? data_length : max_burst_len;
	} else {
		data_offset = cookie->offset;
		cmnd->r2t_data = cmnd->data_length - data_offset;
		cmnd->r2t_sn = cookie->seq;
		cmnd->startsn = cmnd->endsn = cmnd->r2t_sn;
		cmnd->outstanding_r2t--;
	}

	TRACE_WARNING("%s Send recovery R2T, ITT %u R2TSN %u Buffer Offset %d\n",
				   current->comm, cmnd->init_task_tag, cmnd->r2t_sn,
				   data_offset);
	if (hdr->flags & F_BIT) {
		cmnd->recovery_r2t = 2;
		cmnd->next_burst_len = 0;
		cmnd->data_sn = 0;
	} else {
		cmnd->recovery_r2t = 1;
	}
	cmnd->state = ISCSI_BUFFER_RDY;

out_wakeup:
	if (atomic_read(&cmnd->conn->tx_sem.count) <= 0) {
		up(&cmnd->conn->tx_sem);
	}
out:
	TRACE(TRACE_ENTER_LEAVE, "Leave send_recovery_r2t, recovery_r2t %u, "
	      "err = %d\n", cmnd->recovery_r2t, err);
	return err;
}

/***********************************************************************
* This thread remains blocked on a semaphore for most of the time.
* When a r2t timeout happens this thread is awakened for re-transmitting
* any r2t requests that have not generated any activity in a long time.
***********************************************************************/

int
iscsi_retran_thread(void *param)
{
	struct iscsi_session *session = (struct iscsi_session *) param;
	struct iscsi_cmnd *cmnd;

	lock_kernel();

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	daemonize(
#else
	daemonize();
	snprintf(current->comm, sizeof(current->comm),
#endif
			 "iscsi_tran_%d", session->devdata->ntsih);

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 18)
	/*
	 * Arne Redlich, agr1@users.sourceforge.net:
	 * This prevents the retran thread from becoming a zombie after it exits
	 */
	reparent_to_init();
#endif

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGBITS);

	/* mark that this retransmit thread is alive */
	session->retran_thread = current;

	unlock_kernel();

	printk("%s Starting pid %d\n", current->comm, current->pid);

	while (1) {
		/* wait for periodic r2t timer to interrupt us */
		if (down_interruptible(&session->retran_sem))
			break;

		/* lock the session-wide list of commands */
		if (down_interruptible(&session->cmnd_sem))
			continue;

		for (cmnd = session->cmnd_list; cmnd; cmnd = cmnd->next) {
			if (cmnd->outstanding_r2t > 0
			   && jiffies > (cmnd->timestamp+session->r2t_period)
			   && cmnd->state == ISCSI_BUFFER_RDY
			   && !cmnd->retransmit_flg) {
				/* this is a WRITE command with outstanding
				 * R2Ts and no activity for a while.  Try
				 * retransmitting the last R2T.
				 */
				TRACE(TRACE_ERROR_RECOVERY,
				      "%s activity timeout, ITT %u, "
				      "recovery_r2t %u\n",
				      current->comm,
				      cmnd->init_task_tag,
				      cmnd->recovery_r2t);
				if (cmnd->recovery_r2t) {
					if (cmnd->recovery_r2t == 1)
						cmnd->recovery_r2t = 2;
				} else {
					cmnd->retransmit_flg = 1;
					cmnd->endsn = cmnd->startsn
							= cmnd->r2t_sn - 1;
				}
				/* signal the tx thread to do the rexmit */
				if (atomic_read(&cmnd->conn->tx_sem.count) <= 0) {
					up(&cmnd->conn->tx_sem);
				}
			}
		}
		up(&session->cmnd_sem);
	}

	session->retran_thread = NULL;
	up(&session->thr_kill_sem);
	printk("%s Exiting pid %d\n", current->comm, current->pid);
	return 0;
}

void
add_data_to_queue(struct iscsi_cmnd *cmd, struct iscsi_cookie *dataq)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering add_data_to_queue\n");

	if (!cmd->first_data_q) {
		cmd->first_data_q = dataq;
		cmd->last_data_q = dataq;
	} else {
		cmd->last_data_q->next = dataq;
		cmd->last_data_q = cmd->last_data_q->next;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving add_data_to_queue\n");

	return;
}

/********************************************************************
* This function drives the out-of-order dataouts received into the 
* appropriate scatter gather buffer location maintained by scsi. The
* information about the out-of-order dataout and also the corresponding
* scatter offsets are queued for future use.
**********************************************************************/
int
queue_data(struct targ_error_rec *err_rec)
{
	int offset = 0, size = 0, list_offset = 0;
	int count = 0;
	int niov, padding, err, retval = 0, data_crc_len = 0;
	__u32 data_sn, digest, pad_bytes;

	struct iscsi_conn *conn = NULL;
	struct iscsi_cmnd *cmd;
	struct iscsi_init_scsi_data_out *hdr;
	struct iscsi_cookie *dataq = NULL;
	struct scatterlist *st_list;
	struct iovec *iov = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Entering queue_data\n");

	cmd = err_rec->cmd;
	hdr = (struct iscsi_init_scsi_data_out *) err_rec->pdu_hdr;

	list_offset = offset = hdr->offset;
	data_sn = hdr->data_sn;
	size = hdr->length;
	conn = err_rec->curr_conn;

	st_list = (struct scatterlist *) cmd->cmnd->req->sr_buffer;

	while (st_list[count].length <= list_offset) {
		list_offset -= st_list[count].length;
		count++;
	}

	st_list += count;
	niov = find_iovec_needed(size, st_list, list_offset);

	if (niov < 0) {
		TRACE_ERROR("%s queue data: find_iovec_needed returned an error\n",
					current->comm);
		return -1;
	}

	padding = -(size) & 3;
	if (padding)
		niov++;

	if (conn->data_crc == 1)
		niov++;

	iov = (struct iovec *) kmalloc(niov * sizeof(struct iovec), GFP_KERNEL);

	if (!iov) {
		TRACE_ERROR("%s queue_data: Could not malloc iov\n", current->comm);
		return -1;
	}

	TRACE(TRACE_DEBUG, "queue_data: no. of iovec needed %d\n", niov);

	retval = fill_iovec(iov, 0, niov, st_list, &list_offset, size);

	if (padding) {
		iov[niov - 1 - conn->data_crc].iov_base = &pad_bytes;
		iov[niov - 1 - conn->data_crc].iov_len = padding;
	}

	if (conn->data_crc == 1) {
		data_crc_len = CRC_LEN;
		iov[niov - 1].iov_base = &digest;
		iov[niov - 1].iov_len = CRC_LEN;
	}

	err = iscsi_rx_data(cmd->conn, iov, niov, size + padding + data_crc_len);

	if (err != (size + padding + data_crc_len)) {
		TRACE_ERROR("%s queue_data: Trouble in iscsi_rx_data\n", current->comm);
		if (err == PAYLOAD_DIGERR)
			return err;

	}

	dataq = kmalloc(sizeof(struct iscsi_cookie), GFP_KERNEL);

	if (!dataq) {
		TRACE_ERROR("%s add_data_to_queue: Malloc returned an error\n",
					current->comm);
		return -1;
	}

	dataq->offset = offset;
	dataq->xfer_len = size;
	dataq->list_offset = list_offset;
	dataq->list_count = retval;
	dataq->next = NULL;

	add_data_to_queue(cmd, dataq);

	TRACE(TRACE_ENTER_LEAVE, "Leaving queue_data\n");

	return 0;
}

/***********************************************************************
* This function is called everytime a in-sequence dataout pdu is
* received. The function verifies if there was any previous
* out-of-order dataouts for the command and updates the state variables
* accordingly.
***********************************************************************/
void
search_data_q(struct iscsi_cmnd *cmd)
{
	int unhook_flg = 0;
	struct iscsi_cookie *dataq, *prev, *tmp_q;

	TRACE(TRACE_ENTER_LEAVE, "Entering search_data_q\n");

	if (!cmd->first_data_q)
		return;

	dataq = prev = cmd->first_data_q;

	while (dataq) {
		if (cmd->data_done > dataq->offset)
			unhook_flg = 1;
		else if (cmd->data_done == dataq->offset) {
			cmd->scatter_list_count += dataq->list_count;
			cmd->scatter_list_offset = dataq->list_offset;
			cmd->data_done += dataq->xfer_len;
			unhook_flg = 1;
		}

		if (unhook_flg) {
			/* unhook data node */
			if (prev == cmd->first_data_q) {
				cmd->first_data_q = dataq->next;
				prev = dataq->next;
			} else if (dataq == cmd->last_data_q)
				cmd->last_data_q = prev;
			else
				prev->next = dataq->next;

			tmp_q = dataq;
			dataq = dataq->next;
			kfree(tmp_q);
			unhook_flg = 0;
		} else {
			prev = dataq;
			dataq = dataq->next;
		}

		if (cmd->data_done >= cmd->data_length)
			break;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving search_data_q\n");
}
