/* Implementation file for kernel spec */
/* Zc030x     -- Driver --      Zc030x */
/* This file is under GPL              */
/* Copyright :
         Martin     Braun
         Arnaud     Patard
         Andre      Freitas
         Tommy
         Andrew     Birkett         
         Cyril      Russo
---------------------------------------*/

/* Include the kernel headers */
#include <linux/config.h>
#include <linux/version.h>

/* Declare module versions specific */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#endif

/* Module insertion + scheduler + usb */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/usb.h>
#include <linux/slab.h>
 
/* This module specific declarations */ 
/* Include kernel definitions */
#include "zc030x_kerneldef.h"
/* Jpeg decoding to RGB/YUV */
#include "zc030x_jpeg.h"
/* V4L specific */
#include "zc030x_v4l.h"
/* Proc fs specific */
/* #include "zc030x_proc.h" */
/* Memory management */
#include "zc030x_mm.h"
/* I2C register reading */
#include "zc030x_i2c.h"
/* Bridge chip registers */
#include "zc030x_reg.h"
/* Camera specific */
#include "zc030x_cameras.h"
/* Isochronous transfers */
#include "zc030x_isochron.h"
/* Matrix stuff */
#include "zc030x_matrix.h"

/* Module paramaters */
int debug = 5;
int width = 320;
int currentwidth = 0;
int height = 232;
int currentheight = 0;
int quality = 1;
int sensoraccept = 90;
char * sensor = (char*)"auto";

/* Debug level */
MODULE_PARM (debug, "i");
MODULE_PARM_DESC (debug, "Debug level 1:Release -> 5:All messages");
/* Default width */
MODULE_PARM (width, "i");
MODULE_PARM_DESC (width, "Default width of the frame");
/* Default height */
MODULE_PARM (height, "i");
MODULE_PARM_DESC (height, "Default height of the frame");
/* Default quality */
MODULE_PARM (quality, "i");
MODULE_PARM_DESC (quality, "Picture quality 1:Normal 5:Very high (CPU intensive)");
/* Sensor accept level */
MODULE_PARM (sensoraccept, "i");
MODULE_PARM_DESC (sensoraccept, "Sensor signature accept level (in percent) 0:Accept any unmatching sensors -> 100: Accept sensor only if its signature match");
/* Sensor accept level */
MODULE_PARM (sensor, "s");
MODULE_PARM_DESC (sensor, "Sensor name (to avoid detection)");

/* Needed declarations */
MODULE_AUTHOR("Martin Braun <martinbraun at gmx.net>, Arnaud Patard <thertp at users.sourceforge.net>, Andre Freitas <linux at uninet.com.br>, Cyril Russo<>, Andrew Birkett <andy@nobugs.org>");
MODULE_DESCRIPTION("ZC030X USB Camera Driver");
MODULE_LICENSE("GPL");
/* Device table */


/* USB device functions */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) 
static void *zc030x_probe (struct usb_device *dev, unsigned int ifnum,	const struct usb_device_id *id);
static void zc030x_disconnect (struct usb_device *dev, void *ptr);
#else
static int zc030x_probe (struct usb_interface *interface, const struct usb_device_id *id);
static void zc030x_disconnect (struct usb_interface *interface);
#endif

/* Exported read function */ 
ssize_t zc030x_read (struct file *file, char *buffer, size_t count, loff_t * ppos);
/* Exported ioctl function */ 
int zc030x_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
/* Exported mmap function */ 
int zc030x_mmap(struct file *file, struct vm_area_struct *vma);
/* Exported open function */ 
int zc030x_open (struct inode *inode, struct file *file);
/* Exported release function */ 
int zc030x_release (struct inode *inode, struct file *file);

/* File operations supported by this device */
struct file_operations zc030x_fops = 
{
    .owner      =	    THIS_MODULE,
    .open       =	    zc030x_open,
    .release    =	    zc030x_release,
    .read       =	    zc030x_read,
    .ioctl      =	    zc030x_ioctl,
    .mmap       =       zc030x_mmap, 
    .llseek     =       no_llseek,
};

/* Lock to protect the minor_table structure */
static DECLARE_MUTEX (disconnect_sem);

/* USB specific object needed to register this driver with the USB subsystem */
static struct usb_driver zc030x_driver = 
{
//    .owner =        THIS_MODULE,
    .name =	        "zc030x",
    .probe =	    zc030x_probe,
    .disconnect =   zc030x_disconnect,
    .id_table =     zc030x_table,
};



/* Probe for a device matching the ID given in the table */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
static int zc030x_probe(struct usb_interface *intf, const struct usb_device_id *id)
#else
static void *zc030x_probe (struct usb_device *udev, unsigned int ifnum,
	                   const struct usb_device_id *id)
#endif
{
    /* Zc030x device pointer */
    struct usb_zc030x *dev = NULL;
    /* Interface descriptor */
    struct usb_interface_descriptor *iface;
    /* Iterator */
    int i;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    /* Actual alternate settings */
    int actalt = 0;
#else
    struct usb_device *udev = interface_to_usbdev(intf);
#endif  

    /* Debug this probing */
    PDEBUG(1,"probe!!\n");
    
    /* Removed the following line */
    //    probe_calls++;
    
    /* Check for multiconfiguration devices */
    if (udev->descriptor.bNumConfigurations != 1)
    {
        PDEBUG(1,"ERROR: Can't handle camera: multiconfiguration device");
        goto nodevice;
    }
    
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    if (ifnum > 0)  goto nodevice;
    
    /* Skip probe for alternate settings (fixed)*/
    actalt = udev->actconfig->interface[ifnum].act_altsetting;
    if (actalt > 0) goto nodevice; 
#endif  
    
    /*** If we reach this point, we found our interface/altsetting */
    PDEBUG(3, ">> zc030x_probe");
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    MOD_INC_USE_COUNT;
#endif 
    
    /* Allocate memory for our device object and intialize it */
    dev = kmalloc (sizeof (struct usb_zc030x), GFP_KERNEL);
    
    /* Check allocation */
    if (dev == NULL) 
        { err ("Couldn't kalloc zc030x device struct!"); goto error; }
    /* Clear it */
    memset (dev, 0x00, sizeof (struct usb_zc030x));
    /* Start filling it */
    dev->udev = udev;

    /* Detect our camera ! */
    if ((zc030x_detect_camera (dev)) < 0)
    {
        PDEBUG(1,"ERROR: Device not supported !");
        goto error;
    }
    
   
    /* Init mutexes */
    init_MUTEX (&dev->sem);
    init_MUTEX (&dev->buf_lock);
    
    /* USB specific initialization... */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)   
    PDEBUG (3, "Set device configuration to %d\n", DEFAULT_CONFIGURATION);
    usb_set_configuration (udev, DEFAULT_CONFIGURATION);
#endif
    PDEBUG (3, "Set interface to %d\n", VIDEO_INTERFACE);
    PDEBUG (3, "Set alternate configuration %d\n", dev->alt);
    usb_set_interface (udev, VIDEO_INTERFACE, dev->alt);
    
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    iface = &udev->actconfig->interface[ifnum].altsetting[dev->alt];
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,6,6)
    iface = &intf->cur_altsetting->desc;
#else
    iface = &intf->altsetting[dev->alt].desc;
#endif
    
    dev->iface_desc = iface;
    
    /* Endpoint scan procedure was removed, think we can stick with fixed values */
    dev->isocend = 0x01;
    dev->intrend = 0x02;
    
    /* Create the Video For Linux device */
    if (zc030x_v4l_initdev(&zc030x_fops, dev, CamList[dev->type_id].Name, udev) == V4L_Error)
    {
        PDEBUG(1, "Error while registering the camera!");
        goto error;
    }
    
    /* Init device queue */
    init_waitqueue_head(&dev->wq);

    /* Init frame queue */
    for (i = 0; i < ZC030X_NUMFRAMES; i++)
    {
        init_waitqueue_head(&dev->frame[i].wq);
        /* Reset the frame sequence number */
        dev->frame[i].SeqNumber=0;
        /* And init its lock too */
        init_MUTEX(&dev->frame[i].Lock);
    }


    /* Set the default brightness contrast and gamma */
    dev->Brightness = DefaultBrightness;
    dev->Contrast   = DefaultContrast;
    dev->Gamma      = DefaultGamma;
    
    /* Create a gamma array and debug it */
    zc030x_gamma_init();

    /* Initialize the pseudo-JPEG decoder  */ 
    zc030x_jpeg_init();
    
    /* Configure the zc030x camera */
    if (zc030x_configure_camera(dev, width, height, quality) != 0 ) 
    {
        PDEBUG(1, "Error while configuring the camera!");
        goto error;
    }
    
    /* Allow device read, ioctl, mmap */
    dev->present = 1;
    
        
    //PDEBUG(1,"Probe was called %d times", probe_calls);

    /* Okay, return */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    MOD_DEC_USE_COUNT;
    return dev;
#else
    usb_set_intfdata(intf, dev);
    return 0;
#endif
    
nodevice:
    /* Quit */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    return NULL; 
#else
    return -ENODEV;
#endif 
    
error:
    /* Release video device */
    zc030x_v4l_releasedev(dev->pVideoDevice);
    /* Release the driver it self */
    if (dev != NULL)    kfree(dev);
    /* And quit */ 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    MOD_DEC_USE_COUNT;
    return NULL;
#else
    return -ENOMEM;
#endif  
}


/* Handle disconnected device  */
						 
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
static void zc030x_disconnect(struct usb_interface *interface)
{
    struct usb_zc030x *dev =  usb_get_intfdata (interface);
    usb_set_intfdata (interface, NULL);
#else  
static void zc030x_disconnect (struct usb_device *udev, void *ptr)
{
    struct usb_zc030x *dev = (struct usb_zc030x *) ptr;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    MOD_INC_USE_COUNT;
#endif

    PDEBUG(3,">> zc030x_disconnect()");
      
    /* prevent races with open */
    down (&disconnect_sem);
    
    /* prevent device read, write and ioctl */
    dev->present = 0;
    
    /* We don't want people trying to open up the device */
    zc030x_v4l_releasedev(dev->pVideoDevice); 
    
    /* Stop stream (if any) and unlink & free urbs */  
    zc030x_isoc_stop(dev);
    
    /* Dellocate buffers */
    zc030x_unallocate_buffers(dev); 
    
    /* Deregister interrupt urb */
    usb_unlink_urb(dev->inturb);
    usb_free_urb(dev->inturb);
    
    /* Take up the lock */
    down (&dev->sem);
    
    /* if the device is not opened, then we clean up right now */
    if (!dev->open) 
    {
        up (&dev->sem);
        kfree (dev);
    }
    else 
    {
        dev->udev = NULL;
        up (&dev->sem);
    }
    
    /* Release the lock */
    up (&disconnect_sem);
    
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    MOD_DEC_USE_COUNT;
#endif
}

/* Open the device from its /dev/videoX structure */
int zc030x_open (struct inode *inode, struct file *file)
{
    /* Get the video device */
    struct video_device *vdev = video_devdata(file);
    /* And the USB device too */
    struct usb_zc030x *dev;
    
    /* Debug this function */
    PDEBUG(3,">> zc030x_open()");

    /* Check argument */    
    if (vdev == NULL)
    {
        PDEBUG (2, "vdev is NULL, returning");
        return 0;
    }
    
    /* Get driver data */
    dev = video_get_drvdata(vdev);
    if (dev == NULL)
    {
        PDEBUG (2, "dev is NULL, returning");
        return 0;
    }
    
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    MOD_INC_USE_COUNT;
#endif
    
    /* Lock this device before modifying it */
    down (&dev->sem);
   
    /* Increment our usage count for the driver */
    ++dev->open;

    /* Save our object in the file's private structure */
    file->private_data = vdev;
    
    /* Check the current video format */
    if (!dev->VideoFormat)
        dev->VideoFormat = VIDEO_PALETTE_RGB24;
    /* Set the default brightness contrast and gamma */
    dev->Brightness = DefaultBrightness;
    dev->Contrast   = DefaultContrast;
    dev->Gamma      = DefaultGamma;
    
    /* Init the isochronous transfers */
    zc030x_isoc_start(dev);
    
    /* unlock this device */
    up (&dev->sem);
    
    /* Return */
    PDEBUG(3,"<< zc030x_open()");
    return 0;
}



/* Release the device through its /dev/videoX file */
int zc030x_release (struct inode *inode, struct file *file)
{
    /* Get the video device */
    struct video_device *vdev = video_devdata(file);
    /* And the USB device too */
    struct usb_zc030x *dev;
    /* Iterator */
    int i = 0;
    
    /* Debug this function */
    PDEBUG(3,">> zc030x_release()");

    /* Check parameter */
    if (vdev == NULL)
    {
        PDEBUG (2, "vdev is NULL, returning");
        return 0;
    }
    
    /* Get USB device too */
    dev = video_get_drvdata(vdev);
    if (dev == NULL)
    {
        PDEBUG (2, "dev is NULL, returning");
        return 0;
    }
    
    /* Lock our device */
    down (&dev->sem);
    
    /* Check if the device is opened */
    if (dev->open <= 0)
    {
        PDEBUG (5, "%s - device not opened", __FUNCTION__);
        up (&dev->sem);
        /* Return */
        return -ENODEV;
    }
    
    /* Stop isochronous transfers */
    zc030x_isoc_stop(dev);

    /* Wake up frame queue */
    for (i = 0; i < ZC030X_NUMFRAMES; i++)
    {
        if (waitqueue_active(&dev->frame[i].wq))
            wake_up_interruptible(&dev->frame[i].wq);
    }
    
    
    /* Check if the device is already unplugged */
    if(!dev->present && !dev->open)
    {
        /* the device was unplugged before the file was released */
        up (&dev->sem);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)      
        MOD_DEC_USE_COUNT;
#endif
        return 0;
    }
    
    /* Decrement our usage count for the device */
    --dev->open;
    
    /* How could this happen ? */
    if (dev->open <= 0) 
    {
        dev->open = 0;
    }
    
    /* decrement our usage count for the module */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
        MOD_DEC_USE_COUNT;
#endif
    
    /* Release the lock */
    up (&dev->sem);
    
    /* Okay exit */
    PDEBUG(3,"<< zc030x_release()");
    return 0;
}




/* Read a frame from the device */
ssize_t zc030x_read (struct file *file, char *buf, size_t count, loff_t * ppos)
{
    /* Iterator */
    int i = 0;
    /* Frame index */
    int frmx = -1;
    /* Check for non blocking reading */
    int noblock = file->f_flags&O_NONBLOCK;
    /* Need to have a temporary frame */ 
    zc030x_frame *frame;
    
    /* Get the video device */
    struct video_device *vdev = video_devdata(file);
    /* And the USB device too */
    struct usb_zc030x *dev;
    
    /* Debug this function */
    PDEBUG(3,">> zc030x_read()");

    /* Check parameter */
    if (vdev == NULL)
    {
        PDEBUG (2, "vdev is NULL, returning");
        return 0;
    }
    
    /* Get USB device too */
    dev = video_get_drvdata(vdev);
    if (dev == NULL)
    {
        PDEBUG (2, "dev is NULL, returning");
        return 0;
    }

    /* Check parameter */
    if (buf == NULL)
    {
        PDEBUG(2, "Bad NULL buffer to store the frame");
        return -EFAULT;
    }
  
    /* What is this ? */
    PDEBUG(4, "%d bytes, noblock=%d", count, noblock);
    
    /* Check if a signal is pending, so return interrupted */
    if (signal_pending(current))
        return -EINTR;
   
    /* Lock this object */
    down (&dev->sem);
    
    /* Verify that the device wasn't unplugged */
    if (!dev->present) 
    {
        up (&dev->sem);
        return -ENODEV;
    }

    /* Check the device was streaming */
    if(!dev->streaming) 
    {
        up(&dev->sem);
        return -EIO;
    }
    
    /* Release the device before sleeping */
    up(&dev->sem);
    
    /* Set the frame to wait on */
    if (down_trylock(&dev->frame[0].Lock))
    {
        return -EIO;
    }
    
    /* Change this state */
    /* Now this frame is ready to be grabbed into */
    dev->frame[0].GrabState = FRAME_READY;
    /* Release it */
    up(&dev->frame[0].Lock);
    
    /* Sleep until this frame is available */
    PDEBUG(4,"sleep interruptible!");
    interruptible_sleep_on(&dev->frame[0].wq);
    PDEBUG(4,"signal pending test!");
    /* And relock the device now */
    down (&dev->sem);
    
    /* Check if a signal is pending, so return interrupted */
    if (signal_pending(current)) 
    {
        up( &dev->sem);
        return -EINTR;
    }
    
    /* Now we should have a frame */
    if (dev->frame[0].GrabState == FRAME_DONE)
        frmx = i;
        
    /* Found at index */
    PDEBUG(4, "Frame number: %d", frmx);
    
    if (frmx < 0)
    {
        /* Done frame not found */ 
        PDEBUG(2, "frame not ready to be read %d", frmx);
        up (&dev->sem);
        return -EFAULT;
    }
    
    /* Get the frame */
    frame = &dev->frame[frmx];
    /* Lock the frame too */
    down (&frame->Lock);
    
    /* Need to check the user demand */
    i = (frame->HeaderWidth * frame->HeaderHeight * frame->Depth) >> 3;
    /* To avoid overflow (ie->reading kernel memory space) */
    if (copy_to_user(buf, frame->Data, count < i ? count : i)) 
    {
        PDEBUG(2, "Copy failed! %d bytes not copied", i);
        up (&frame->Lock);
        up (&dev->sem);
        return -EFAULT;
    }
    PDEBUG(2, "Maximum size is %dx%dx%d = %d asked %d", frame->HeaderWidth, frame->HeaderHeight, frame->Depth, i, count);
    
    /* Release the frame */
    frame->GrabState = FRAME_READY;
    /* Unlock it */
    up(&frame->Lock);
    
    /* Unlock the device */
    up (&dev->sem);
    
    /* Return */
    PDEBUG(3,"<< zc030x_read()");
    return count < i ? count : i;
}

/* Map a kernel memory page to user space */
int zc030x_mmap(struct file *file, struct vm_area_struct *vma)
{
    /* Start address */
    unsigned long start = vma->vm_start;
    /* Size of mapping */
    unsigned long size  = vma->vm_end - vma->vm_start;
    /* The page */
    unsigned long page;
    /* And the position */
    unsigned long pos;
    /* Get the video device */
    struct video_device *vdev = video_devdata(file);
    /* And the USB device too */
    struct usb_zc030x *dev;
    
    /* Debug this function */
    PDEBUG(3,">> zc030x_mmap()");

    /* Check parameter */
    if (vdev == NULL)
    {
        PDEBUG (2, "vdev is NULL, returning");
        return 0;
    }
    
    /* Get USB device too */
    dev = video_get_drvdata(vdev);
    if (dev == NULL)
    {
        PDEBUG (2, "dev is NULL, returning");
        return 0;
    }
    
    /* Check if the device is still here */
    if (!dev->present) 
        return -EIO;
    
    /* Debug here too */
    PDEBUG(4, "mmap: %ld (%lX) bytes", size, size);
    
    /* Check the asked size */
    if (size > (((ZC030X_NUMFRAMES * MAX_VIDEODATA_SIZE) + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1)))
        return -EINVAL;

    /* Convert the kernel address to userspace */
    pos = (unsigned long)dev->fbuf;
    /* And remap all concerned page to shared space */
    while (size > 0) 
    {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
        page = page_to_pfn(vmalloc_to_page((void *)pos));
        if (remap_pfn_range(vma, start, page, PAGE_SIZE, PAGE_SHARED))
#else        
        page = kvirt_to_pa(pos);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0) || defined (RH9_REMAP)
        if (remap_page_range(vma, start, page, PAGE_SIZE, PAGE_SHARED))
#else /* RH9_REMAP */
        if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED))
#endif /* RH9_REMAP */
#endif
            return -EAGAIN;

        start += PAGE_SIZE;
        pos += PAGE_SIZE;
        if (size > PAGE_SIZE)
            size -= PAGE_SIZE;
        else
            size = 0;
    }
    
    /* Okay return */
    return 0;
}



/* Perform IOCTLs now */
int zc030x_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    /* Return code */
	int rc ; 
    /* Get the video device */
    struct video_device *vdev = video_devdata(file);
    /* And the USB device too */
    struct usb_zc030x *dev;
    
    /* Debug this function */
    PDEBUG(3,">> zc030x_ioctl()");

    /* Check parameter */
    if (vdev == NULL)
    {
        PDEBUG (2, "vdev is NULL, returning");
        return 0;
    }
    
    /* Get USB device too */
    dev = video_get_drvdata(vdev);
    if (dev == NULL)
    {
        PDEBUG (2, "dev is NULL, returning");
        return 0;
    }
    
    /* Check if the device is still here */
    if (!dev->present) 
        return -EIO;
    
    
    /* Zc030x specific ioctl */
    switch (cmd)
    {
        /* Users asked register mapping */
        case ZCI2CREGISTERMAP:
        {
            /* Call the register mapping function */
            /*
            //unsigned char buf[1];
            unsigned char val[255];
            unsigned char prev = 0;
            int ret;
            int i;

            struct register_map * prm = (struct register_map*) arg;
            zc030x_i2c_init(dev->udev, 0x0c);
            // Write :
            PDEBUG(2, "HHHHHH ------------------------- Separator --------------");
            for (i = 0; i < 256; i ++)
            {
                // Save previous value (bug reported on ML)
                prev = zc030x_i2c_reg_read(dev->udev, i, 1);
                PDEBUG(2, "HHHHHH Read %02X at address %02X", prev, i);
                prm[i].addr = i;
                prm[i].value = 0xff;
                prm[i].mode = RM_ReadWrite;
                
                zc030x_i2c_reg_write(dev->udev, i, 0xff);
                val[i] = zc030x_i2c_reg_read(dev->udev, i, 1);
                if (val[i] != 0xff)
                {
                    PDEBUG(2, "REGISTERMAP: [FF] (%02X) contains some read only bit with %02X value", i, val[i]);

                    prm[i].value = val[i];

                    if (val[i] == 0)
                        prm[i].mode = RM_Unused;
                }
                zc030x_i2c_reg_write(dev->udev, i, 0x00);
                ret = zc030x_i2c_reg_read(dev->udev, i, 1);
                if (ret != 0x00)
                {
                    PDEBUG(2, "REGISTERMAP: [00] (%02X) contains some read only bit with %02X value", i, ret);
                }
                if (val[i] == ret && ret != 00)
                {
                    PDEBUG(2, "REGISTERMAP-OK: [ID] (%02X) is read only with %02X value", i, ret);
                    prm[i].addr = i;
                    prm[i].value = ret;
                    prm[i].mode = RM_ReadOnly;
                }
                // Restore previous value
                zc030x_i2c_reg_write(dev->udev, i, prev); 
            }
            */
            /* And return */
            rc = 0;
            break;
        }
        /* User asked sensor identification */
        case ZCIDSENSOR:
        {
            /* Call sensor identification function now */
            //unsigned char buf[1];
            int ret = 0;
            
            /* Currently not supported */
            ret = 0;
            /*
            
            struct sensor_list * psl = (struct sensor_list*) arg;
		    PDEBUG (2, ">>> ZCIDSENSOR asked for : %s, (%d)",psl->name, psl->id);
            
            // Let's try to read the i2c first register (I've no idea about how to read an I2C register but...)
            //buf[0] = 0x00;

            // Okay, try to find the sensor given its id
            usb_set_interface (dev->udev, 0, 0);
	    //fuck! as this has gone to main.c and is declared extern here, sizeof doesn't work!
            //ret = zc030x_find_sensor_by_id(dev->udev, slist, sizeof(slist) / sizeof(slist[0]));
	    ret = zc030x_find_sensor_by_id(dev->udev, slist, 16);
            usb_set_interface (dev->udev, 0, 7);
            if (ret != -1)
            {
                psl->id = slist[ret].id;
                sprintf(psl->name, "%s", slist[ret].name);
                psl->minwidth = slist[ret].minwidth;
                psl->maxwidth = slist[ret].maxwidth;
                psl->minheight = slist[ret].minheight;
                psl->maxheight = slist[ret].maxheight;
            } else
            {
                psl->id = 0;
                sprintf(psl->name, "Unknown");
                psl->minwidth = 0;
                psl->maxwidth = 0;
                psl->minheight = 0;
                psl->maxheight = 0;
            }
            psl->i2c_sig = NULL;
		    PDEBUG (2, "<<< ZCIDSENSOR asked for : %s, (%d)",psl->name, psl->id);
            */
            /* Okay return now */
            rc = 0;
            break;
        }
        /* Video for Linux specific IOCTLs */
        default:
            rc = video_usercopy(inode, file, cmd, arg, zc030x_v4l_ioctl);
            
    }

	/* Return now */
    PDEBUG (2, "<< zc030x_ioctl() arg = %08X", (unsigned int)arg);
	return rc;
}





/* Module loading */
static int __init   usb_zc030x_init (void)
{
    int result;
   
    PDEBUG(1,"Loading experimental zc030x... \n");

    /* register this driver with the USB subsystem */
    result = usb_register (&zc030x_driver);
    if (result < 0)
    {
        err ("usb_register failed for the " __FILE__ " driver. Error number %d",        result);
        return -1;
    }
    
    PDEBUG (5, DRIVER_DESC " " DRIVER_VERSION);

#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
//TODO: proc entry has been temorarily removed  zc030x_proc_create();
#endif

    return 0;
}

/* Module unloading */
static void __exit usb_zc030x_exit (void)
{

#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
//TODO: proc entry management  zc030x_proc_destroy();
#endif
    /* stop isoc stream (if streaming) ) */
    
    /* deregister this driver with the USB subsystem */
    PDEBUG (1,"Unloading experimental zc030x.\n");
    usb_deregister (&zc030x_driver);
}

module_init (usb_zc030x_init);
module_exit (usb_zc030x_exit);


