/*
 * Copyright (C) 2004 Red Hat UK Ltd.
 *
 * This file is released under the GPL.
 */

#ifndef DM_BIO_LIST_H
#define DM_BIO_LIST_H

#include <linux/bio.h>

struct bio_list {
	struct bio *head;
	struct bio *tail;
};

#define BIO_LIST_HEAD(bl)	struct bio_list bl = { NULL, NULL }

static inline void bio_list_init(struct bio_list *bl)
{
	bl->head = bl->tail = NULL;
}

#define bio_for_each(bio, bl) \
	for (bio = (bl)->head; bio; bio = bio->bi_next)

/* True if bio_list is empty */
static inline int bio_list_empty(struct bio_list *bl)
{
	return !bl->head;
}

/* Add a new bio to the end of a bio_list */
static inline void bio_list_add(struct bio_list *bl, struct bio *bio)
{
	bio->bi_next = NULL;

	if (bl->tail)
		bl->tail->bi_next = bio;
	else
		bl->head = bio;

	bl->tail = bio;
}

/* Merge bl2 at the tail of bl */
static inline void bio_list_merge(struct bio_list *bl, struct bio_list *bl2)
{
	if (bl->tail)
		bl->tail->bi_next = bl2->head;
	else
		bl->head = bl2->head;

	bl->tail = bl2->tail;
}

/* Join bl2 at the head of bl */
static inline void bio_list_join(struct bio_list *bl, struct bio_list *bl2)
{
	if (likely(!bio_list_empty(bl2))) {
		bl2->tail->bi_next = bl->head;
		bl->head = bl2->head;

		if (!bl->tail)
			bl->tail = bl2->tail;
	}
}

/* Join bl2 at the head of bl and initialize bl2 */
static inline void bio_list_join_init(struct bio_list *bl, struct bio_list *bl2)
{
	bio_list_join(bl, bl2);
	bio_list_init(bl2);
}

/* Merge bl2 at the tail of bl and initialize bl2 */
static inline void bio_list_merge_init(struct bio_list *bl,
				       struct bio_list *bl2)
{
	bio_list_merge(bl, bl2);
	bio_list_init(bl2);
}

/* Push bio onto the head of a bio_list */
static inline void bio_list_push(struct bio_list *bl, struct bio *bio)
{
	bio->bi_next = bl->head;
	bl->head = bio;

	if (!bl->tail)
		bl->tail = bio;
}

/* Pop bio off the head of a bio_list */
static inline struct bio *bio_list_pop(struct bio_list *bl)
{
	struct bio *bio = bl->head;

	if (bio) {
		bl->head = bl->head->bi_next;
		if (!bl->head)
			bl->tail = NULL;

		bio->bi_next = NULL;
	}

	return bio;
}

/* Get a bio from a bio_list and initialize the list */
static inline struct bio *bio_list_get(struct bio_list *bl)
{
	struct bio *bio = bl->head;

	if (bio)
		bio_list_init(bl);

	return bio;
}

#endif
