/*
 * Copyright (C) 2005,2006  Red Hat GmbH
 *
 * Module Author: Heinz Mauelshagen (Mauelshagen@RedHat.com)
 *
 * This file is released under the GPL.
 *
 * WARNING: this is Alpha software wich can corrupt your data!
 *
 *
 * Linux 2.6 Device Mapper RAID4 and RAID5 target.
 *
 * Supports:
 *	o RAID4 with dedicated parity device
 *	o RAID5 with rotating parity (Left+Right, Symmetric+Asymmetric)
 *
 *
 * Thanks to MD for:
 *    o the raid address calculation algorithm
 *    o the base of the biovec <-> page list copier.
 *
 *
 * Uses region hash to keep track of how many writes are in flight
 * to regions in order to use dirty log to keep state of regions to recover:
 *
 *    o clean regions (those which are synchronized
 * 	and don't have write io in flight)
 *    o dirty regions (those with write io in flight)
 *
 *
 * On startup, any dirty regions are migrated to the 'nosync' state
 * and are subject to recovery by the daemon.
 *
 * See raid_ctr() for table definition.
 *
 *
 * FIXME:
 * o remove instrumentation (REMOVEME:)
 *
 */

#define	VERSION	"v0.1442"

#include "dm.h"
#include "dm-bio-list.h"
#include "dm-io.h"
#include "dm-log.h"
#include "dm-mem-cache.h"
#include "dm-region_hash.h"

#include <linux/ctype.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/time.h>
#include <linux/vmalloc.h>
#include <linux/workqueue.h>

/*
 * Configurable parameters
 */

/* Default # of stripes if not set in constructor */
#define	STRIPES			64

/* Minimum/maximum # of selectable stripes */
#define	STRIPES_MIN		8
#define	STRIPES_MAX		16384

/* Default chunk size in sectors if not set in constructor */
#define	CHUNK_SIZE		64

/* Default io size in sectors if not set in constructor */
#define	IO_SIZE			16
#define	IO_SIZE_MIN		SECTORS_PER_PAGE

/* Maximum setable chunk size in sectors */
#define	CHUNK_SIZE_MAX		16384

/* Recover io size default in sectors */
#define	RECOVER_IO_SIZE		256

/* Default percentage recover io bandwidth */
#define	BANDWIDTH		10
#define	BANDWIDTH_MIN		1
#define	BANDWIDTH_MAX		100

/* Minimum/maximum ios queueable against any single raid set (throttling) */
#define	MIN_IO_COUNT		1024
#define	MAX_IO_COUNT		(16 * STRIPES_MAX)

/*
 * END Configurable parameters
 */


#define	TARGET	"dm-raid45"
#define	DAEMON	"kraid45d"
#define	PREFIX	TARGET ": "
#define	TST	PREFIX "(DEVEL) "	/* REMOVEME: */
#define	TI_ERR_RET(str, ret) \
	do { ti->error = PREFIX str; return ret; } while(0);
#define	TI_ERR(str)	TI_ERR_RET(str, -EINVAL)

#define	SECTORS_PER_PAGE	(PAGE_SIZE >> SECTOR_SHIFT)

/* Amount/size for __xor() */
#define	SECTORS_PER_XOR			8
#if SECTORS_PER_XOR > SECTORS_PER_PAGE
	#error dm-raid45: sectors per xor larger than sectors per page!!!
#endif

#define	XOR_SIZE	(SECTORS_PER_XOR << SECTOR_SHIFT)

/* Derive raid_set from stripe_cache member */
#define	RS(a)	\
	((struct raid_set*)((char *) a - offsetof(struct raid_set, sc)))


/* Check value in range */
#define	range_ok(i, min, max)	(i >= min && i <= max)

/* xor optimization */
typedef	unsigned long 	xor_t;

/*-----------------------------------------------------------------
 * Stripe cache
 *
 * Cache for all reads and writes to raid sets (operational or degraded)
 *
 * We need to run all data to and from a RAID set through this cache,
 * because parity chunks need to get calculated from data chunks
 * or, in the degraded/resynchronization case, missing chunks need
 * to be reconstructed using the other chunks of the stripe.
 *---------------------------------------------------------------*/
/* Protect kmem cache # counter */
static spinlock_t _stripe_sc_nr_lock = SPIN_LOCK_UNLOCKED;
static atomic_t _stripe_sc_nr = ATOMIC_INIT(0); /* kmem cache # counter */

/* A stripe set (holds bios hanging off) */
struct stripe_set {
	struct stripe *stripe;	/* Backpointer to stripe for endio() */

	struct bio_list reads;		/* Reads queued to stripe set */
	struct bio_list writes;		/* Writes queued to stripe set */
	struct bio_list writes_merged;	/* Writes merged to endio */
};

/* A stripe: the io object to handle all reads and writes to a RAID set */
struct stripe {
	struct stripe_cache *sc;	/* Backpointer to stripe cache */

	unsigned long flags;		/* flags (see above) */

	sector_t key;			/* Hash key */
	sector_t region;		/* Region stripe is mapped to */

	/* Reference count */
	atomic_t use;

	struct {
		/*
	 	* Pending ios in flight:
	 	*
	 	* used as a 'lock' to control move of stripe to endio list
	 	*/
		atomic_t pending;		/* Pending ios in flight */

		/* Sectors to read and write for multi page stripe sets */
		unsigned int size;
	} io;

	struct {
		struct list_head hash;	/* Stripe cache hash */

		/*
		 * Used to hang off caches io_list in order to
		 * initiate io in do_cache_flush().
		 */
		struct list_head io;

		/*
	 	 * Used to hang off stripes in endio() to end io in the daemon.
		 */
		struct list_head endio;

		/*
		 * If the reference count drops to zero, we put
		 * the stripe on the lru list in order to be able
		 * to do fast fetches in stripe_get().
		 */
		struct list_head lru;
	} lists;

	/* Keep context between stripe submission and endio processing */
	union {
		uint32_t pi;
	} context;

	/* This sets memory cache object (dm-mem-cache) */
	struct dm_mem_cache_object *obj;

	/* Array of stripe sets (dynamically allocated) */
	struct stripe_set ss[0];
};

/* States stripes can be in (flags field) */
enum stripe_states {
	/*
	 * The STRIPE_ACTIVE flag is Used as a fast 'lock' between
	 * the dm_raid() daemon and and the endio() function, which
	 * get's called asynchronuously by dm-io.
	 */
	STRIPE_ACTIVE,		/* Active io on stripe */
	STRIPE_ERROR,		/* io error on stripe */
	STRIPE_UPTODATE,	/* Stripe data uptodate */
	STRIPE_READ,		/* Read */
	STRIPE_RBW,		/* Read-before-write */
	STRIPE_RECONSTRUCT,	/* reconstruct of a missing chunk required */
	STRIPE_RECOVER,		/* Stripe used for RAID set recovery */
};

/* ... and macros to access them */
#define	STRIPE_BITOPS(name, flag) \
static inline void \
ClearStripe ## name(struct stripe *stripe) { clear_bit(flag, &stripe->flags); }\
static inline void \
SetStripe ## name(struct stripe *stripe) { set_bit(flag, &stripe->flags); } \
static inline int \
Stripe ## name(struct stripe *stripe) { return test_bit(flag, &stripe->flags); }

STRIPE_BITOPS(Active, STRIPE_ACTIVE)
STRIPE_BITOPS(Error, STRIPE_ERROR)
STRIPE_BITOPS(Uptodate, STRIPE_UPTODATE)
STRIPE_BITOPS(Read, STRIPE_READ)
STRIPE_BITOPS(RBW, STRIPE_RBW)
STRIPE_BITOPS(Reconstruct, STRIPE_RECONSTRUCT)
STRIPE_BITOPS(Recover, STRIPE_RECOVER)
#undef STRIPE_BITOPS

/* A stripe hash */
struct stripe_hash {
	struct list_head *hash;
	unsigned int buckets;
	unsigned int mask;
};

/* A stripe cache */
struct stripe_cache {
	/* Stripe hash */
	struct stripe_hash hash;

	struct {
		/*
		 * List of stripes with pages to io in do_cache_flush().
		 */
		struct list_head io;

		/*
		 * List of stripes to (copy data to pages and) end io in daemon.
		 */
		struct list_head endio;
		spinlock_t endio_lock;	/* Protect endio list */

		/*
		 * List of lru stripes (no io pending).
		 */
		struct list_head lru;
		spinlock_t lru_lock;	/* Protect lru list */
	} lists;

	/* Slab cache to allocate stripes from */
	struct {
		kmem_cache_t *cache; /* Cache itself */
		char name[32];	     /* Unique name */
	} kc;

	void *dm_io_client;	/* dm-io client resource context */

	/* dm-mem-cache client resource context */
	struct dm_mem_cache_client *dm_mem_cache_client;

	int stripes_parm;	/* # stripes parameter from constructor */
	atomic_t stripes;	/* actual # of stripes in cache */
	atomic_t stripes_to_shrink;/* # of stripes to shrink cache by */
	atomic_t stripes_last;	/* last # of stripes in cache */
	atomic_t active_stripes;/* actual # of active stripes in cache */

	/* REMOVEME: */
	atomic_t max_active_stripes;/* actual # of active stripes in cache */
};

/* Flag specs for raid_dev */;
enum raid_dev_flags { DEVICE_FAILED };

/* The raid device in a set */
struct raid_dev {
	struct dm_dev *dev;
	unsigned long flags;	/* raid_dev_flags */
	sector_t start;		/* offset to map to */
};

/* Flags spec for raid_set */
enum raid_set_flags {
	DEAD,		/* RAID set inoperational */
	DEVEL_STATS,	/* REMOVEME: display development status information */
	IO_ERROR,	/* io error on set */
	REGION_GET,	/* get a region to recover */
	SC_BUSY,	/* stripe cache busy -> send an event */
	CHECK_OVERWRITE,/* check stripe chunks for complete overwrites */
};

/* REMOVEME: devel stats counters */
struct devel_stats {
	atomic_t bandwidth;
	atomic_t bio_copy_pl_next;
	atomic_t bio_copy_redo;
	atomic_t check_overwrite;
	atomic_t degraded;
	atomic_t delayed_bios;
	atomic_t end_io_reads;
	atomic_t end_io_writes;
	atomic_t evict;
	atomic_t eperms;
	atomic_t fast_reads;
	atomic_t fast_active_reads;
	atomic_t flushs;
	atomic_t hits_1st;
	atomic_t ios_post;
	atomic_t ios_writes;
	atomic_t inscache;
	atomic_t ios_reads;
	atomic_t no_bandwidth;
	atomic_t nosync;
	atomic_t rc;
	atomic_t reads_added;
	atomic_t reconstruct_xor;
	atomic_t reconstruct_ei;
	atomic_t reconstruct_dev;
	atomic_t sum_delayed_bios;
	atomic_t wc;
	atomic_t writes_added;
	atomic_t wr_chunk_updated;
	atomic_t wr_stripe_updated;
	atomic_t xors;
};

/*
 * A RAID set
 */
struct raid_set {
	struct dm_target *ti;		/* Target pointer */

	unsigned long flags;		/* State flags */

	struct {
		spinlock_t lock;	/* Protects central input list below */
		struct bio_list in;	/* Pending ios (central input list) */
		struct bio_list work;	/* ios work set */
		struct semaphore max;	/* io throttling */
		wait_queue_head_t suspendq; /* suspend synchronization */
		atomic_t in_process;	/* counter of queued bios (suspendq) */

		/* io work */
		struct workqueue_struct *wq;
		struct work_struct ws;
	} io;

	struct stripe_cache sc;		/* Stripe cache for this set */

	/* Xor optimization */
	struct {
		void (*f)(unsigned int, xor_t **);
		const char *name;
		unsigned int chunks;
		unsigned int speed;
	} xor;

	/* Recovery parameters */
	struct recover {
		struct dirty_log *dl;		/* Dirty log */
		void *rh;			/* Region hash */

		region_t nr_regions;
		region_t nr_regions_to_recover;
		region_t nr_regions_recovered;
		unsigned long start_jiffies;
		unsigned long end_jiffies;

		unsigned int bandwidth;		/* Recovery bandwidth [%] */
		unsigned int bandwidth_work;	/* Recovery bandwidth [factor]*/
		unsigned int bandwidth_parm;	/*  " constructor parm */
		unsigned int io_size;		/* io size <= chunk size */
		unsigned int io_size_parm;	/* io size ctr parameter */

		/* recovery io throttling */
		atomic_t io_count[2];		/* counter recover/regular io */
		unsigned long last_jiffies;

		void *reg;		/* Actual region to recover */
		struct stripe *stripe;	/* Special stripe used for recovery */
		sector_t pos;		/* Position within region to recover */
		sector_t end;		/* End of region to recover */
	} recover;

	/* RAID set parameters */
	struct {
		struct raid_type *raid_type;  /* RAID type (eg, RAID4) */
		unsigned int raid_parms;      /* # variable raid parameters */

		unsigned int chunk_size;      /* Amount of sectors per chunk */
		unsigned int chunk_size_parm;
		unsigned int chunk_mask;      /* Mask for amount */
		unsigned int chunk_shift;     /* rsector chunk size shift */

		unsigned int io_size;	      /* Amount of sectors for io */
		unsigned int io_size_parm;
		unsigned int io_mask;	      /* Mask for amount */
		unsigned int io_shift_mask;   /* Mask for raid_address() */
		unsigned int io_shift;	      /* rsector io size shift */
		unsigned int pages_per_io;    /* Pages per io */

		sector_t sectors_per_dev;     /* Sectors per device */

		atomic_t failed_devs;	/* Amount of devices failed */

		/* Index of device to initialize */
		int dev_to_init;

		/* Raid devices dynamically allocated. */
		unsigned int raid_devs;		/* # of RAID devices below */
		unsigned int data_devs;		/* # of RAID data devices */

		int ei;		/* index of failed RAID device */

		/* index of dedicated parity device (i.e. RAID4) */
		int pi;
		int pi_parm;	/* constructor parm for status output */
	} set;

	/* REMOVEME: devel stats counters */
	struct devel_stats stats;

	/* Dynamically allocated temporary pointers for xor() */
	xor_t **data;

	/* Dynamically allocated RAID devices */
	struct raid_dev dev[0];
};

/*-----------------------------------------------------------------
 * Raid-4/5 set structures.
 *---------------------------------------------------------------*/
/* RAID level definitions */
enum raid_level {
	raid4,
	raid5,
};

/* Symmetric/Asymmetric, Left/Right parity rotating algorithms */
enum raid_algorithm {
	left_asym,
	right_asym,
	left_sym,
	right_sym,
};

struct raid_type {
	char *name;	/* Plain text name of RAID level (and algorithm) */
	char *descr;	/* Descriptor text for logging */
	unsigned int parity_devs;	/* # of parity devices */
	unsigned int minimal_devs;	/* minimal # of devices in set */
	enum raid_level level;		/* RAID level */
	enum raid_algorithm algorithm;	/* RAID algorithm */
};

/* Supported raid types and properties */
static struct raid_type raid_types[] = {
	{ "raid4",    "RAID4 (dedicated parity disk)", 1, 3, raid4, 0 },
	{ "raid5_la", "RAID5 (left asymmetric)", 1, 3, raid5, left_asym },
	{ "raid5_ra", "RAID5 (right asymmetric)", 1, 3, raid5, right_asym },
	{ "raid5_ls", "RAID5 (left symmetric)", 1, 3, raid5, left_sym },
	{ "raid5_rs", "RAID5 (right symmetric)", 1, 3, raid5, right_sym },
};

/* Address as calculated by raid_address() */
struct address {
	sector_t key;		/* Hash key (start address of stripe) */
	unsigned int di, pi;	/* Data and parity disk index */
};

/* REMOVEME: reset statistics counters */
static void stats_reset(struct raid_set *rs)
{
	struct devel_stats *s = &rs->stats;

	atomic_set(&s->bandwidth, 0);
	atomic_set(&s->bio_copy_pl_next, 0);
	atomic_set(&s->bio_copy_redo, 0);
	atomic_set(&s->check_overwrite, 0);
	atomic_set(&s->degraded, 0);
	atomic_set(&s->delayed_bios, 0);
	atomic_set(&s->end_io_reads, 0);
	atomic_set(&s->end_io_writes, 0);
	atomic_set(&s->eperms, 0);
	atomic_set(&s->evict, 0);
	atomic_set(&s->flushs, 0);
	atomic_set(&s->fast_reads, 0);
	atomic_set(&s->fast_active_reads, 0);
	atomic_set(&s->rc, 0);
	atomic_set(&s->hits_1st, 0);
	atomic_set(&s->inscache, 0);
	atomic_set(&s->ios_post, 0);
	atomic_set(&s->ios_reads, 0);
	atomic_set(&s->ios_writes, 0);
	atomic_set(&rs->sc.max_active_stripes, 0);
	atomic_set(&s->no_bandwidth, 0);
	atomic_set(&s->nosync, 0);
	atomic_set(&s->reads_added, 0);
	atomic_set(&s->reconstruct_xor, 0);
	atomic_set(&s->reconstruct_ei, 0);
	atomic_set(&s->reconstruct_dev, 0);
	atomic_set(&s->sum_delayed_bios, 0);
	atomic_set(&s->wc, 0);
	atomic_set(&s->writes_added, 0);
	atomic_set(&s->wr_chunk_updated, 0);
	atomic_set(&s->wr_stripe_updated, 0);
	atomic_set(&s->xors, 0);
}

/*----------------------------------------------------------------
 * RAID set management routines
 *--------------------------------------------------------------*/
/*
 * Begin small helper functions.
 */
/* Queue (optionally delayed) io work */
static void wake_do_raid_delayed(struct raid_set *rs, unsigned long delay)
{
	struct work_struct *ws = &rs->io.ws;

	cancel_delayed_work(ws);

	if (unlikely(delay))
		queue_delayed_work(rs->io.wq, ws, delay);
	else
		queue_work(rs->io.wq, ws);
}

/* Queue io work immediately (called from region hash too) */
static void wake_do_raid(void *context)
{
	wake_do_raid_delayed((struct raid_set*) context, 0);
}

/* Wait until all io has been processed */
static inline void wait_ios(struct raid_set *rs)
{
	wait_event(rs->io.suspendq, !atomic_read(&rs->io.in_process));
}

/* Get an io reference */
static inline void io_get(struct raid_set *rs)
{
	atomic_inc(&rs->io.in_process);
}

/* Put the io reference and conditionally wake io waiters */
static inline void io_put(struct raid_set *rs)
{
	if (atomic_dec_and_test(&rs->io.in_process))
		wake_up(&rs->io.suspendq);
}

/* Calculate device sector offset */
static inline sector_t _sector(struct raid_set *rs, struct bio *bio)
{
	sector_t r = bio->bi_sector;

	sector_div(r, rs->set.data_devs);
	return r;
}

/* Test device operational */
static inline int dev_operational(struct raid_set *rs, unsigned int p)
{
	return !test_bit(DEVICE_FAILED, &rs->dev[p].flags);
}

/* Are we recovering ? */
static inline int raid_set_recovering(struct raid_set *rs)
{
	return rs->recover.stripe ? 1 : 0;
}

/* Return # of active stripes in stripe cache */
static inline int sc_active(struct stripe_cache *sc)
{
	return atomic_read(&sc->active_stripes);
}

/* Test io pending on stripe */
static inline int stripe_io(struct stripe *stripe)
{
	return atomic_read(&stripe->io.pending);
}

static inline void stripe_io_inc(struct stripe *stripe)
{
	atomic_inc(&stripe->io.pending);
}

static inline int stripe_io_dec(struct stripe *stripe)
{
	return atomic_dec_and_test(&stripe->io.pending);
}

/* Wrapper needed by for_each_io_dev() */
static void _stripe_io_inc(struct stripe *stripe, unsigned int p)
{
	stripe_io_inc(stripe);
}

/* Error a stripe */
static inline void stripe_error(struct stripe *stripe, struct page *page)
{
	SetPageError(page);
	SetStripeError(stripe);
}

/* Return region state for a sector */
static inline int region_state(struct raid_set *rs, sector_t sector,
			       unsigned long state)
{
	void *rh = rs->recover.rh;

	return raid_set_recovering(rs) ?
	       (rh_state(rh, rh_sector_to_region(rh, sector), 1) & state) : 0;
}

/* Return true in case a page should be read/written */
static inline int page_io(struct dm_mem_cache_object *obj)
{
	struct page *page = obj->pl->page;

	return !PageUptodate(page) || PageDirty(page);
}

/* Call a function on each page list needing io */
static inline void for_each_io_dev(struct raid_set *rs, struct stripe *stripe,
				   void (*f_io)(struct stripe *stripe,
						unsigned int p))
{
	unsigned int p = rs->set.raid_devs;

	while (p--) {
		if (page_io(stripe->obj + p))
			f_io(stripe, p);
	}
}

/* Reconstruct a particular device ? */
static inline int dev_to_init(struct raid_set *rs)
{
	return rs->set.dev_to_init > -1;
}

/* Check maximum devices which may fail in a raid set */
static inline int raid_set_degraded(struct raid_set *rs)
{
	return atomic_read(&rs->set.failed_devs);
}

/* Check # of devices which may fail in a raid set */
static inline int raid_set_operational(struct raid_set *rs)
{
	/* Too many failed devices -> BAD */
	if (unlikely(raid_set_degraded(rs) > rs->set.raid_type->parity_devs))
		return 0;

	return 1;
}

/* Display RAID set dead message once */
static void raid_set_dead(struct raid_set *rs)
{
	if (!test_and_set_bit(DEAD, &rs->flags)) {
		unsigned int p;
		char buf[BDEVNAME_SIZE];

		DMERR(PREFIX "FATAL: too many devices failed -> RAID set dead");

		for (p = 0; p < rs->set.raid_devs; p++) {
			if (!dev_operational(rs, p))
				DMERR(PREFIX "device /dev/%s failed",
				      bdevname(rs->dev[p].dev->bdev, buf));
		}
	}
}

/* RAID set degrade check */
static int raid_set_check_and_degrade(struct raid_set *rs,
				      struct stripe *stripe, unsigned int p)
{
	if (test_and_set_bit(DEVICE_FAILED, &rs->dev[p].flags))
		return -EPERM;

	/* Through an event in case of errors */
	dm_table_event(rs->ti->table);
	atomic_inc(&rs->set.failed_devs);

	if (!test_and_set_bit(IO_ERROR, &rs->flags)) {
		char buf[BDEVNAME_SIZE];

		DMERR(PREFIX
		      "CRITICAL: %sio error on device /dev/%s "
		      "in region=%llu; DEGRADING RAID set",
		      stripe ? "" : "FAKED ",
		      bdevname(rs->dev[p].dev->bdev, buf),
		      stripe ? stripe->key : 0);
		DMERR(PREFIX "further device error messages suppressed");

		/* Store index for recovery */
		rs->set.ei = p;
	}

	return 0;
}

static int raid_set_check_degrade(struct raid_set *rs, struct stripe *stripe)
{
	unsigned int p = rs->set.raid_devs;
	int r = 0;
	struct page *page;

	while (p--) {
		page = stripe->obj[p].pl->page;

		if (PageError(page))
			raid_set_check_and_degrade(rs, stripe, p);
		else if (PageChecked(page)) {
			/*
			 * We've got an error on a stripe, but not on
			 * this chunk, because we optimized IO away ->
			 * io requeue is mandatory
			 */
			ClearPageUptodate(page);
			ClearPageChecked(page);
			r = 1;
		}
	}

	return r;
}

/* RAID set upgrade check */
static int raid_set_check_and_upgrade(struct raid_set *rs, unsigned int p)
{
	if (!test_and_clear_bit(DEVICE_FAILED, &rs->dev[p].flags))
		return -EPERM;

	if (atomic_dec_and_test(&rs->set.failed_devs)) {
		clear_bit(IO_ERROR, &rs->flags);
		rs->set.ei = -1;
	}

	return 0;
}

static char *basename(char *s)
{
	char *p = strrchr(s, '/');

	return p ? p + 1 : s;
}

/* Lookup a RAID device by name or by major:minor number */
union dev_lookup {
	const char *dev_name;
	struct raid_dev *dev;
};
enum lookup_type { byname, bymajmin, bynumber };
static int raid_dev_lookup(struct raid_set *rs, enum lookup_type by,
			   union dev_lookup *dl)
{
	unsigned int p;

	/*
	 * Must be an incremental loop, because the device array
	 * can have empty slots still on calls from raid_ctr()
	 */
	for (p = 0; p < rs->set.raid_devs; p++) {
		char buf[BDEVNAME_SIZE];
		struct raid_dev *dev = rs->dev + p;

		if (!dev->dev)
			break;

		/* Format dev string appropriately if necessary */
		switch (by) {
		case byname:
			bdevname(dev->dev->bdev, buf);
			break;

		case bymajmin:
			format_dev_t(buf, dev->dev->bdev->bd_dev);
			break;

		default:
			break;
		}

		/* Do the actual check */
		switch (by) {
		case byname:
		case bymajmin:
			if (!strcmp(dl->dev_name, buf))
				return p;

			break;

		case bynumber:
			if (dl->dev->dev->bdev->bd_dev == \
			    dev->dev->bdev->bd_dev)
				return p;
		}
	}

	return -ENODEV;
}

/*
 * End io wrapper to up maximum io count semaphore (throttling)
 */
static inline void _bio_endio(struct raid_set *rs, struct bio *bio, int error)
{
	/* REMOVEME:*/
	atomic_inc(bio_data_dir(bio) == WRITE ? &rs->stats.end_io_writes :
						&rs->stats.end_io_reads);

	bio_endio(bio, bio->bi_size, error);
	up(&rs->io.max);	/* Unthrottle io */
	io_put(rs);		/* Wake any suspend waiters */
}
/*
 * End small helper functions.
 */


/*
 * Stripe hash functions
 */

/* Initialize/destroy stripe hash */
static inline int hash_init(struct stripe_hash *hash, unsigned int stripes)
{
	unsigned int buckets = 2, max_buckets = stripes / 8;

	/* Calculate number of buckets (2^^n <= stripes / 8) */
	while (buckets < max_buckets)
		buckets <<= 1;

	/* Allocate stripe hash */
	hash->hash = vmalloc(buckets * sizeof(*hash->hash));
	if (!hash->hash)
		return -ENOMEM;

	hash->buckets = buckets;
	hash->mask = buckets - 1;

	/* Initialize buckets */
	while (buckets--)
		INIT_LIST_HEAD(hash->hash + buckets);

	return 0;
}

static inline void hash_exit(struct stripe_hash *hash)
{
	vfree(hash->hash);
}

#define	OFF(list) \
	static inline void stripe_ ## list ## _off(struct stripe *stripe) { \
		if (!list_empty(&stripe->lists.list)) \
			list_del_init(&stripe->lists.list); \
}

OFF(hash)
OFF(lru)
#undef OFF

#define POP(list) \
	if (list_empty(&sc->list)) \
		stripe = NULL; \
	else { \
		stripe = list_entry(sc->list.next, struct stripe, list); \
		list_del_init(&stripe->list); \
	}

/* Pop an available stripe off the lru list */
static struct stripe *stripe_lru_pop(struct stripe_cache *sc)
{
	struct stripe *stripe;

	spin_lock_irq(&sc->lists.lru_lock);
	POP(lists.lru);

	if (stripe)
		stripe_hash_off(stripe); /* Remove from hash before reuse */
	spin_unlock_irq(&sc->lists.lru_lock);

	return stripe;
}

static inline unsigned int stripe_hash(struct stripe_hash *hash, sector_t key)
{
	return (unsigned int)((key * 2654435387U) >> 12) & hash->mask;
}

static inline struct list_head *hash_bucket(struct stripe_hash *hash,
					    sector_t key)
{
	return hash->hash + stripe_hash(hash, key);
}

/* Insert an entry into a hash */
static inline void hash_insert(struct stripe_hash *hash, struct stripe *stripe)
{
	list_add_tail(&stripe->lists.hash, hash_bucket(hash, stripe->key));
}

/* Insert an entry into the stripe hash */
static inline void sc_insert(struct stripe_cache *sc, struct stripe *stripe)
{
	hash_insert(&sc->hash, stripe);
}

/* Lookup an entry in the stripe hash */
static inline struct stripe *stripe_lookup(struct stripe_cache *sc,
					   sector_t key)
{
	struct stripe *stripe;
	struct list_head *bucket = hash_bucket(&sc->hash, key);

	list_for_each_entry(stripe, bucket, lists.hash) {
		if (stripe->key == key)
			return stripe;
	}

	return NULL;
}

/* Resize the stripe cache hash on size changes */
static int hash_resize(struct stripe_cache *sc)
{
	if (atomic_read(&sc->stripes) > 2 * atomic_read(&sc->stripes_last) ||
	    atomic_read(&sc->stripes) < atomic_read(&sc->stripes_last) / 4) {
		int r;
		unsigned int b = sc->hash.buckets;
		struct list_head *pos, *tmp;
		struct stripe_hash hash, hash_tmp;

		r = hash_init(&hash, atomic_read(&sc->stripes));
		if (r)
			return r;

		memcpy(&hash_tmp, &sc->hash, sizeof(tmp));

		/* Walk old buckets and insert into new */
		spin_lock_irq(&sc->lists.lru_lock);
		while (b--) {
			list_for_each_safe(pos, tmp, sc->hash.hash + b)
				hash_insert(&hash,
					    list_entry(pos, struct stripe,
						       lists.hash));
		}

		memcpy(&sc->hash, &hash, sizeof(sc->hash));
		atomic_set(&sc->stripes_last, atomic_read(&sc->stripes));
		spin_unlock_irq(&sc->lists.lru_lock);

		hash_exit(&hash_tmp);
	}

	return 0;
}

/*
 * Stripe cache functions
 */

/*
 * Invalidate all page lists pages of a stripe
 *
 * I only keep state for the whole list in the first page
 */
static inline void stripe_pages_invalidate(struct raid_set *rs,
					   struct stripe *stripe)
{
	unsigned int p = rs->set.raid_devs;

	while (p--)
		stripe->obj[p].pl->page->flags = 0;
}

/* Prepare stripe for (re)use */
static inline void stripe_invalidate(struct raid_set *rs,
				     struct stripe *stripe)
{
	stripe->flags = 0;
	stripe_pages_invalidate(rs, stripe);
}

/* Initialize a stripe */
static void stripe_init(struct stripe_cache *sc, struct stripe *stripe,
			unsigned int io_size)
{
	unsigned int p = RS(sc)->set.raid_devs;

	/* Work all io chunks */
	while (p--) {
		struct stripe_set *ss = stripe->ss + p;

		stripe->obj[p].private = ss;
		ss->stripe = stripe;
		bio_list_init(&ss->reads);
		bio_list_init(&ss->writes);
		bio_list_init(&ss->writes_merged);
	}

	stripe_invalidate(RS(sc), stripe);
	stripe->sc = sc;
	INIT_LIST_HEAD(&stripe->lists.hash);
	INIT_LIST_HEAD(&stripe->lists.io);
	INIT_LIST_HEAD(&stripe->lists.endio);
	INIT_LIST_HEAD(&stripe->lists.lru);
	stripe->io.size = io_size;
	atomic_set(&stripe->use, 0);
	atomic_set(&stripe->io.pending, 0);
}

/* Number of pages per stripe */
static inline unsigned int stripe_pages(struct raid_set *rs,
					unsigned int io_size)
{
	return dm_div_up(io_size, SECTORS_PER_PAGE) * rs->set.raid_devs;
}

/* Allocate a stripe and its memory object */
enum grow { grow, nogrow };
static struct stripe *stripe_alloc(struct stripe_cache *sc,
				   unsigned int io_size, enum grow grow)
{
	struct stripe *stripe;
	int r;
	unsigned int pages_per_chunk = dm_div_up(io_size, SECTORS_PER_PAGE);

	stripe = kmem_cache_alloc(sc->kc.cache, SLAB_KERNEL);
	if (stripe) {
		memset(stripe, 0, sizeof(*stripe));

		/* Grow the dm-mem-cache on request */
		if (grow == grow) {
			r = dm_mem_cache_grow(sc->dm_mem_cache_client,
					      pages_per_chunk);
			if (r)
				goto err;
		}

		r = dm_mem_cache_alloc(sc->dm_mem_cache_client, pages_per_chunk,
				       &stripe->obj);
		if (r)
			goto err1;

		stripe_init(sc, stripe, io_size);
	}

	return stripe;

   err1:
	if (grow == grow)
		dm_mem_cache_shrink(sc->dm_mem_cache_client, pages_per_chunk);
   err:
	kmem_cache_free(sc->kc.cache, stripe);
	return NULL;
}

/*
 * Free a stripes memory object, shrink the
 * memory cache and free the stripe itself
 */
static void stripe_free(struct stripe *stripe)
{
	dm_mem_cache_free(stripe->sc->dm_mem_cache_client, stripe->obj);
	dm_mem_cache_shrink(stripe->sc->dm_mem_cache_client,
			    dm_div_up(stripe->io.size, SECTORS_PER_PAGE));
	kmem_cache_free(stripe->sc->kc.cache, stripe);
}

/* Free the recovery stripe */
static void stripe_recover_free(struct raid_set *rs)
{
	if (rs->recover.stripe) {
		stripe_free(rs->recover.stripe);
		rs->recover.stripe = NULL;
	}
}

/* Push a stripe safely onto the endio list to be handled by do_endios() */
static inline void stripe_endio_push(struct stripe *stripe)
{
	int wake;
	unsigned long flags;
	struct stripe_cache *sc = stripe->sc;

	spin_lock_irqsave(&sc->lists.endio_lock, flags);
	wake = list_empty(&sc->lists.endio);
	list_add_tail(&stripe->lists.endio, &sc->lists.endio);
	spin_unlock_irqrestore(&sc->lists.endio_lock, flags);

	if (wake)
		wake_do_raid(RS(sc));
}

/* Pop a stripe off safely off the endio list */
static inline struct stripe *stripe_endio_pop(struct stripe_cache *sc)
{
	struct stripe *stripe;

	/* This runs in parallel with endio() */
	spin_lock_irq(&sc->lists.endio_lock);
	POP(lists.endio)
	spin_unlock_irq(&sc->lists.endio_lock);
	return stripe;
}
#undef POP

/* Evict stripe from cache */
static void stripe_evict(struct raid_set *rs, struct stripe *stripe)
{
	stripe_hash_off(stripe); /* Take off hash */
	ClearStripeActive(stripe);

	if (list_empty(&stripe->lists.lru)) {
		spin_lock_irq(&stripe->sc->lists.lru_lock);
		list_add(&stripe->lists.lru, &rs->sc.lists.lru);
		spin_unlock_irq(&stripe->sc->lists.lru_lock);

		atomic_inc(&rs->stats.evict); /* REMOVEME: */
	}
}

/* Sort stripe into io_list */
static void stripe_to_flush_sorted(struct stripe *stripe)
{
	if (list_empty(&stripe->lists.io)) {
		struct stripe *s;
		struct list_head *io_list = &stripe->sc->lists.io;

		list_for_each_entry(s, io_list, lists.io) {
			if (stripe->key < s->key) {
				io_list = &s->lists.io;
				break;
			}
		}

		list_add_tail(&stripe->lists.io, io_list);
	}
}

/* Grow stripe cache */
static int sc_grow(struct stripe_cache *sc, unsigned int stripes,
		   enum grow grow)
{
	int r = 0;
	struct raid_set *rs = RS(sc);

	/* Try to allocate this many (additional) stripes */
	while (stripes--) {
		struct stripe *stripe = stripe_alloc(sc, rs->set.io_size, grow);

		if (likely(stripe)) {
			spin_lock_irq(&sc->lists.lru_lock);
			list_add_tail(&stripe->lists.lru, &sc->lists.lru);
			atomic_inc(&sc->stripes);
			spin_unlock_irq(&sc->lists.lru_lock);
		} else {
			r = -ENOMEM;
			break;
		}
	}

	clear_bit(SC_BUSY, &rs->flags);
	return r ? r : hash_resize(sc);
}

/* Shrink stripe cache */
static int sc_shrink(struct stripe_cache *sc, unsigned int stripes)
{
	int r = 0;

	/* Try to get unused stripe from LRU list */
	while (stripes--) {
		struct stripe *stripe;

		stripe = stripe_lru_pop(sc);
		if (stripe) {
			/* An lru stripe may never have ios pending! */
			BUG_ON(StripeActive(stripe) || stripe_io(stripe));
			stripe_free(stripe);
			atomic_dec(&sc->stripes);
		} else {
			r = -ENOENT;
			break;
		}
	}

	/* Check if stats are still sane */
	if (atomic_read(&sc->max_active_stripes) >
	    atomic_read(&sc->stripes))
		atomic_set(&sc->max_active_stripes, 0);

	if (r)
		return r;

	clear_bit(SC_BUSY, &(RS(sc)->flags));
	return hash_resize(sc);
}

/* Create stripe cache */
static int sc_init(struct raid_set *rs, unsigned int stripes)
{
	int nr, r;
	struct stripe_cache *sc = &rs->sc;
	struct stripe *stripe;

	r = hash_init(&sc->hash, stripes);
	if (r)
		return r;
	/* Initialize various lists and the lock */
	INIT_LIST_HEAD(&sc->lists.io);
	INIT_LIST_HEAD(&sc->lists.endio);
	INIT_LIST_HEAD(&sc->lists.lru);
	spin_lock_init(&sc->lists.endio_lock);
	spin_lock_init(&sc->lists.lru_lock);

	atomic_set(&sc->stripes, 0);
	atomic_set(&sc->stripes_last, stripes);
	atomic_set(&sc->stripes_to_shrink, 0);
	atomic_set(&sc->active_stripes, 0);
	atomic_set(&sc->max_active_stripes, 0); /* REMOVEME: */

	/*
	 * We need a runtime unique # to suffix the kmem cache name
	 * because we'll have one for each active RAID set
	 */
	spin_lock_irq(&_stripe_sc_nr_lock);
	nr = atomic_read(&_stripe_sc_nr);
	atomic_inc(&_stripe_sc_nr);
	spin_unlock_irq(&_stripe_sc_nr_lock);

	sprintf(sc->kc.name, "%s_%d", TARGET, nr);

	sc->kc.cache = kmem_cache_create(sc->kc.name,
					 sizeof(*stripe) +
					 rs->set.raid_devs *
					 sizeof(*stripe->ss),
					 0, 0, NULL, NULL);
	if (!sc->kc.cache)
		return -ENOMEM;


	r = dm_mem_cache_client_create(stripes * stripe_pages(rs, rs->set.io_size) + stripe_pages(rs, RECOVER_IO_SIZE), stripes + 1, rs->set.raid_devs, &sc->dm_mem_cache_client);
	if (r)
		return r;

	/* Allocate a stripe for set recovery */
	stripe = stripe_alloc(sc, RECOVER_IO_SIZE, nogrow);
	if (!stripe)
		return -ENOMEM;

	SetStripeRecover(stripe);
	rs->recover.stripe = stripe;
	return sc_grow(sc, stripes, nogrow); /* Grow the cache */
}

/* Destroy the stripe cache */
static void sc_exit(struct stripe_cache *sc)
{
	if (sc->hash.hash) {
		if (sc->kc.cache) {
			BUG_ON(sc_shrink(sc, atomic_read(&sc->stripes)));
			kmem_cache_destroy(sc->kc.cache);
		}

		if (sc->dm_mem_cache_client)
			dm_mem_cache_client_destroy(sc->dm_mem_cache_client);

		hash_exit(&sc->hash);
		sc->hash.hash = NULL;
	}
}

/*
 * Try getting a stripe either from the hash or from the lru list
 */
static inline void _stripe_get(struct stripe *stripe)
{
	atomic_inc(&stripe->use);
}

static int stripe_get(struct raid_set *rs, struct address *addr,
		      struct stripe **ret)
{
	struct stripe_cache *sc = &rs->sc;
	struct stripe *stripe;

	/* First try if stripe is already hashed */
	spin_lock_irq(&sc->lists.lru_lock);
	stripe = stripe_lookup(sc, addr->key);
	if (stripe) {
		/* Remove from the lru list if on */
		stripe_lru_off(stripe);
		spin_unlock_irq(&sc->lists.lru_lock);
		atomic_inc(&rs->stats.hits_1st); /* REMOVEME: */
		goto out;
	}

	spin_unlock_irq(&sc->lists.lru_lock);

	/* Second try to get an lru stripe */
	stripe = stripe_lru_pop(sc);
	if (!stripe)
		return -ENOENT;

	/* Invalidate before reinserting with changed key */
	stripe_invalidate(rs, stripe);
	stripe->key = addr->key;
	stripe->region = rh_sector_to_region(rs->recover.rh, addr->key);
	stripe->context.pi = addr->pi;
	sc_insert(sc, stripe);

	atomic_inc(&rs->stats.inscache); /* REMOVEME: */

   out:
	_stripe_get(stripe);
	*ret = stripe;
	return 0;
}

/*
 * Decrement reference count on a stripe.
 *
 * Move it to list of LRU stripes if zero and !RBW.
 */
static void stripe_put(struct stripe *stripe)
{
	if (atomic_dec_and_test(&stripe->use)) {
		/*
		 * Put stripe onto the lru list if
		 * not read-before-write, because
		 * those will be reclaimed directly
		 */
		if (list_empty(&stripe->lists.lru) &&
		    !StripeRBW(stripe)) {
			spin_lock_irq(&stripe->sc->lists.lru_lock);
			list_add_tail(&stripe->lists.lru,
				      &stripe->sc->lists.lru);
			spin_unlock_irq(&stripe->sc->lists.lru_lock);
		}
	}

	BUG_ON(atomic_read(&stripe->use) < 0);
}

/*
 * Calculate RAID address
 *
 * Delivers tuple with the index of the data disk holding the chunk
 * in the set, the parity disks index and the start of the stripe
 * within the address space of the set (used as the stripe cache hash key).
 */
/* thx MD */
static void raid_address(struct raid_set *rs, sector_t sector,
			 struct address *addr)
{
	unsigned int di, pi;
	sector_t stripe, tmp;

	/*
	 * chunk_number = sector / chunk_size
	 * stripe = chunk_number / data_devs
	 * di = stripe % data_devs;
	 */
	stripe = sector >> rs->set.chunk_shift;
	di = sector_div(stripe, rs->set.data_devs);

	switch (rs->set.raid_type->level) {
	case raid4:
		pi = rs->set.pi;
		break;

	case raid5:
		tmp = stripe;
		pi = sector_div(tmp, rs->set.raid_devs);

		switch (rs->set.raid_type->algorithm) {
		case left_asym:	/* Left asymmetric */
			pi = rs->set.data_devs - pi;
		case right_asym: /* Right asymmetric */
			if (di >= pi)
				di++;
			break;

		case left_sym: /* Left symmetric */
			pi = rs->set.data_devs - pi;
		case right_sym: /* Right symmetric */
			di = (pi + di + 1) % rs->set.raid_devs;
			break;

		default:
			DMERR(PREFIX "Unknown RAID algorithm %d",
			      rs->set.raid_type->algorithm);
			return;
		}

		break;

	default:
		DMERR(PREFIX "Unknown RAID level %d", rs->set.raid_type->level);
		return;
	}

	addr->di = di;
	addr->pi = pi;

	/*
	 * Hash key = start offset on any single device of the RAID set;
	 * adjusted in case io size differs from chunk size
	 */
	addr->key = (stripe << rs->set.chunk_shift) +
		    (sector & rs->set.io_shift_mask);
}

/*
 * Join reads and writes to a stripe by
 * hanging their bios off the pagesets lists
 */
static inline int _stripe_add_rw(struct raid_set *rs, struct stripe **ret,
				 struct bio *bio)
{
	int r;
	struct address addr;
	struct stripe *stripe;
	struct stripe_set *ss;

	raid_address(rs, bio->bi_sector, &addr);
	r = stripe_get(rs, &addr, &stripe);
	if (r)
		return r;

	/* Distinguish read and write cases */
	ss = stripe->ss + addr.di;
	if (bio_data_dir(bio) == READ) {
		SetStripeRead(stripe);
		bio_list_add(&ss->reads, bio);
		atomic_inc(&rs->stats.reads_added); /* REMOVEME: */
	} else {
		SetStripeRBW(stripe);
		bio_list_add(&ss->writes, bio);
		atomic_inc(&rs->stats.writes_added); /* REMOVEME: */
	}

	/*
	 * Sort stripe into stripe caches io_list,
	 * so that do_cache_flush() can belabour it
	 */
	stripe_to_flush_sorted(stripe);
	*ret = stripe;
	return 0;
}

/* Add a bio list to a stripe's stripe set */
static inline int stripe_add_rw(struct raid_set *rs, struct bio_list *process,
				struct bio_list *reject, atomic_t *stats)
{
	int r = 0;
	void *rh = rs->recover.rh;
	struct stripe *stripe;
	struct bio *bio;

	while ((bio = bio_list_pop(process))) {
		int ret = _stripe_add_rw(rs, &stripe, bio);

		if (ret) {
			/* Got no stripe from cache -> reject */
			bio_list_add(reject, bio);
			atomic_inc(stats); /* REMOVEME: */

		/* Add was ok */
		} else if (bio_data_dir(bio) == WRITE) {
			/* Inrement pending write count on region */
			rh_inc(rh, stripe->region);
			r = 1;	/* Region hash needs a flush */
		}
	}

	return r;
}

/*
 * Xor optimization macros
 */
#define	XOR_CHUNKS_MAX	5

#define DECLARE_2	xor_t *d0, *d1
#define DECLARE_3	DECLARE_2, *d2
#define DECLARE_4	DECLARE_3, *d3
#define DECLARE_5	DECLARE_4, *d4

#define INIT_2	d0 = data[0] + i; d1 = data[1] + i
#define INIT_3	INIT_2; d2 = data[2] + i
#define INIT_4	INIT_3; d3 = data[3] + i
#define INIT_5	INIT_4; d4 = data[4] + i

#define	PREFETCH_2(offset)	prefetchw(d0 + offset); prefetch(d1 + offset)
#define	PREFETCH_3(offset)	PREFETCH_2(offset); prefetch(d2 + offset)
#define	PREFETCH_4(offset)	PREFETCH_3(offset); prefetch(d3 + offset)
#define	PREFETCH_5(offset)	PREFETCH_4(offset); prefetch(d4 + offset)

#define D2(n)	d0[n] = d0[n] ^ d1[n]
#define D3(n)	D2(n) ^ d2[n]
#define D4(n)	D3(n) ^ d3[n]
#define D5(n)	D4(n) ^ d4[n]

#define	X_2(macro, offset)	macro(offset); macro(offset + 1)
#define	X_4(macro, offset)	X_2(macro, offset); X_2(macro, offset + 2)

#define	X_8(macro, offset)	X_4(macro, offset); X_4(macro, offset + 4)
#define	X_16(macro, offset)	X_8(macro, offset); X_8(macro, offset + 8)
#define	X_32(macro, offset)	X_16(macro, offset); X_16(macro, offset + 16)
#define	X_64(macro, offset)	X_32(macro, offset); X_32(macro, offset + 32)

#define	_XOR(chunks, xors_per_run) \
static void _xor ## chunks ## _ ## xors_per_run(xor_t **data) \
{ \
	unsigned int end = XOR_SIZE / sizeof(*data) - xors_per_run; \
	unsigned int i = 0; \
	DECLARE_ ## chunks; \
	INIT_ ## chunks; \
	PREFETCH_ ## chunks(i); \
\
	do { \
		PREFETCH_ ## chunks(i + xors_per_run); \
   again: \
		X_ ## xors_per_run(D ## chunks, i); \
	} while ((i += xors_per_run) < end); \
\
	if (i == end) \
		goto again; \
}

#define	MAKE_XOR_PER_RUN(xors_per_run) \
	_XOR(2, xors_per_run); _XOR(3, xors_per_run); \
	_XOR(4, xors_per_run); _XOR(5, xors_per_run);

MAKE_XOR_PER_RUN(8)	/* Define _xor_*_8() functions */
MAKE_XOR_PER_RUN(16)	/* Define _xor_*_16() functions */
MAKE_XOR_PER_RUN(32)	/* Define _xor_*_32() functions */
MAKE_XOR_PER_RUN(64)	/* Define _xor_*_64() functions */


#define MAKE_XOR(xors_per_run) \
static void xor_ ## xors_per_run(unsigned int n, xor_t **data) \
{ \
	struct { \
		void (*f)(xor_t**); \
	} xor_funcs ## xors_per_run[] = { \
		{ _xor2_ ## xors_per_run }, \
		{ _xor3_ ## xors_per_run }, \
		{ _xor4_ ## xors_per_run }, \
		{ _xor5_ ## xors_per_run }, \
	}; \
\
	xor_funcs ## xors_per_run[n - 2].f(data); \
}

MAKE_XOR(8)	/* Define xor_8() function */
MAKE_XOR(16)	/* Define xor_16() function */
MAKE_XOR(32)	/* Define xor_32() function */
MAKE_XOR(64)	/* Define xor_64() function */

#include <linux/raid/xor.h>
static void xor_asm(unsigned int n, xor_t **data)
{
 	xor_block(n, XOR_SIZE, (void**) data);
}

struct xor_func {
	void (*f)(unsigned int, xor_t **);
	const char *name;
};

static struct xor_func xor_funcs[] = {
 	{ xor_64,  "xor_64" },
 	{ xor_32,  "xor_32" },
 	{ xor_16,  "xor_16" },
	{ xor_8,   "xor_8" },
	{ xor_asm, "xor_asm" },
};

/*
 * Calculate crc
 *
 * This indexes into the page list of the stripe
 * xoring rs->set.raid_devs pages per run
 */
static void xor(struct raid_set *rs, struct dm_mem_cache_object *obj,
		unsigned int pi, unsigned int sector)
{
	unsigned int di = !pi, n = 0, o, p = rs->set.raid_devs;
	xor_t **data = rs->data;

#if SECTORS_PER_PAGE != SECTORS_PER_XOR
	/* Get offset into the page for SECTORS_PER_PAGE > SECTORS_PER_XOR */
	unsigned int offset =
		offset = ((sector % SECTORS_PER_PAGE) / SECTORS_PER_XOR) *
			 XOR_SIZE;
	#define	OFFSET	+ offset
#else
	#define OFFSET
#endif

	/* Get offset into the page_list */
	o = sector / SECTORS_PER_PAGE;

	/* Init parity page */
	data[0] = kmap_atomic(pl_elem(obj[pi].pl, o)->page, KM_USER0) OFFSET;
	data[1] = kmap_atomic(pl_elem(obj[di].pl, o)->page, KM_USER0) OFFSET;
	memcpy(data[0], data[1], XOR_SIZE);
	kunmap_atomic(data[1], KM_USER0);

	/* Preset pointers to data pages */
	while (p--) {
		if (p != di && p != pi)
			data[++n] = kmap_atomic(pl_elem(obj[p].pl, o)->page,
						KM_USER0) OFFSET;

		if (n && (n + 1 == rs->xor.chunks || !p)) {
			rs->xor.f(n + 1, data);

			for (; n; n--)
				kunmap_atomic(data[n], KM_USER0);
		}
	}

	kunmap_atomic(data[0], KM_USER0);

	/* Change state on first page in list */
	SetPageUptodate(obj[pi].pl->page);
}

static void common_xor(struct stripe *stripe, sector_t count,
		       unsigned int start, unsigned int p)
{
	struct raid_set *rs = RS(stripe->sc);
	unsigned int sector;

	for (sector = start; sector < count; sector += SECTORS_PER_XOR)
		xor(rs, stripe->obj, p, sector);

	atomic_inc(&rs->stats.xors); /* REMOVEME: */
}

/*
 * Degraded/reconstruction mode;
 *
 * rebuild a missing page list content using
 * the other page lists of the stripe
 */
static inline void error_xor(struct stripe *stripe, unsigned int p)
{
	common_xor(stripe, stripe->io.size, 0, p);
	ClearPageError(stripe->obj[p].pl->page);
}

/*
 * Calculate parity sectors on intact stripes
 *
 * Need to calculate raid address for recover stripe, because its
 * chunk sizes differs and is typically larger than io chunk size
 */
static void parity_xor(struct stripe *stripe)
{
	struct raid_set *rs = RS(stripe->sc);
	unsigned int chunk_size = rs->set.chunk_size,
		     io_size = stripe->io.size,
		     xor_size = chunk_size;
	sector_t off;

	if (chunk_size > io_size)
		xor_size = io_size;

	/* This can be the recover stripe with a larger io size */
	for (off = 0; off < io_size; off += chunk_size) {
		unsigned int pi;

		/*
		 * Recover stripe is bigger than regular io ones
		 * and has no precalculated parity disk index ->
		 * need to calculate RAID address
		 */
		if (unlikely(StripeRecover(stripe))) {
			struct address addr;

			raid_address(rs,
				     (stripe->key + off) * rs->set.data_devs,
				     &addr);
			pi = addr.pi;
		} else
			pi = stripe->context.pi;

		common_xor(stripe, xor_size, off, pi);
		dev_operational(rs, pi) ? SetPageDirty(stripe->obj[pi].pl->page) : ClearPageDirty(stripe->obj[pi].pl->page);
	}
}

/* Reconstruct missing pages */
static void reconstruct_xor(struct stripe *stripe)
{
	struct raid_set *rs = RS(stripe->sc);
	int idx = rs->set.ei;

	/* Degraded mode */
	if (idx > -1) {
		atomic_inc(&rs->stats.reconstruct_ei); /* REMOVEME: */

	/* Reconstruction mode */
	} else {
		/*
		 * This is a reconstruction because of resynchronization
		 * going on and region not in sync (yet) ->
		 * use either selected or stored parity disk index
		 */
		idx = dev_to_init(rs) ? rs->set.dev_to_init :
					stripe->context.pi;

		atomic_inc(&rs->stats.reconstruct_dev); /* REMOVEME: */
	}

	error_xor(stripe, idx);

	ClearStripeError(stripe);
	ClearStripeReconstruct(stripe);
	SetStripeUptodate(stripe);

	atomic_inc(&rs->stats.reconstruct_xor); /* REMOVEME: */
}

/*
 * Copy data across between stripe pages and bio vectors
 *
 * Pay attention to data alignment in stripe and bio pages
 */
static void bio_copy_page_list(int rw, struct stripe *stripe,
			       struct page_list *pl, struct bio *bio)
{
	unsigned int i, page_offset;
	void *page_addr;
	struct raid_set *rs = RS(stripe->sc);
	struct bio_vec *bv;

	/* Get start page in page list */
	i = (bio->bi_sector & rs->set.io_mask) / SECTORS_PER_PAGE;
	pl = pl_elem(pl, i);

	page_offset = to_bytes(bio->bi_sector & (SECTORS_PER_PAGE - 1));
	page_addr = kmap_atomic(pl->page, KM_USER0);

	/* Walk all segments and copy data across between bio_vecs and pages */
	bio_for_each_segment(bv, bio, i) {
		int len = bv->bv_len, size;
		unsigned bio_offset = 0;
		void *bio_addr = kmap_atomic(bv->bv_page, KM_USER0) +
					     bv->bv_offset;

   redo:
		size = (page_offset + len > PAGE_SIZE) ?
		       PAGE_SIZE - page_offset : len;

		if (rw == READ)
			memcpy(bio_addr + bio_offset,
			       page_addr + page_offset, size);
		else
			memcpy(page_addr + page_offset,
			       bio_addr + bio_offset, size);

		page_offset += size;
		if (page_offset == PAGE_SIZE) {
			/*
			 * We reached the end of the stripe page.
			 *
			 * Need to map the next one to copy mor data.
			 */
			/* REMOVEME: */
			atomic_inc(&RS(stripe->sc)->stats.bio_copy_pl_next);

			pl = pl->next;
			if (pl) { /* Could have been the last page */
				kunmap_atomic(page_addr, KM_USER0);
				page_addr = kmap_atomic(pl->page, KM_USER0);
				page_offset = 0;
				len -= size;
				if (len) {
					/* REMOVEME: */
					atomic_inc(&RS(stripe->sc)->stats.bio_copy_redo);
					bio_offset += size;
					goto redo;
				}
			}
		}

		__bio_kunmap_atomic(bio_addr, KM_USER0);
	}

	kunmap_atomic(page_addr, KM_USER0);
}

/*
 * Process end io
 *
 * I need to do it here because I can't in interrupt
 *
 * Read and write functions are split in order to avoid
 * conditionals in the main loop for performamce reasons.
 */
/* End io all read bios on a page list */
static inline int do_end_io_pl_reads(struct raid_set *rs, struct stripe *stripe,
				     struct page_list *pl, struct bio_list *bl,
				     int error)
{
	int r = 0;
	struct bio *bio;

	while ((bio = bio_list_pop(bl))) {
		/* Superfluous on error, but what the heck... */
		bio_copy_page_list(READ, stripe, pl, bio);
		_bio_endio(rs, bio, error);
		stripe_put(stripe);
		r = 1;
	}

	return r;
}

/* End io all write bios on a page list */
static inline int do_end_io_pl_writes(struct raid_set *rs,
				      struct stripe *stripe,
				      struct page_list *pl, struct bio_list *bl,
				      int error)
{
	int r = 0;
	struct bio *bio;

	while ((bio = bio_list_pop(bl))) {
		rh_dec(rs->recover.rh, stripe->region);
		_bio_endio(rs, bio, error);
		stripe_put(stripe);
		r = 1;
	}

	ClearPageDirty(pl->page);
	return r;
}

/* End io all bios on a page list */
static int do_end_io_pl(int rw, struct raid_set *rs, struct stripe *stripe,
			struct page_list *pl, struct stripe_set *ss)
{
	int error = PageError(pl->page) ? -EIO : 0;

	if (rw == READ)
		return do_end_io_pl_reads(rs, stripe, pl, &ss->reads, error);

	return do_end_io_pl_writes(rs, stripe, pl, &ss->writes_merged, error);
}

/* End io all reads/writes on a stripe */
static int do_end_io_rw(int rw, struct raid_set *rs, struct stripe *stripe)
{
	int r = 0;
	unsigned int p = rs->set.raid_devs;

	while (p--)
		r += do_end_io_pl(rw, rs, stripe, stripe->obj[p].pl,
				  stripe->ss + p);
	return r;
}


/* End io all reads on a stripe copyinf data across from stripe to bios */
static inline int do_end_io_reads(struct raid_set *rs, struct stripe *stripe)
{
	ClearStripeRead(stripe);
	return do_end_io_rw(READ, rs, stripe);
}

/* End io all writes on a stripe */
static inline void do_end_io_writes(struct raid_set *rs, struct stripe *stripe)
{
	do_end_io_rw(WRITE, rs, stripe);
}

/* Fail all ios on a bio list and return # of bios */
static unsigned int bio_list_fail(struct raid_set *rs, struct stripe *stripe,
				  struct bio_list *bl)
{
	unsigned int r;
	struct bio *bio;

	raid_set_dead(rs);

	/* Update region counters */
	if (stripe) {
		bio_for_each(bio, bl) {
			if (bio_data_dir(bio) == WRITE)
				rh_dec(rs->recover.rh, stripe->region);
		}
	}

	/* Error end io all bios */
	for (r = 0; (bio = bio_list_pop(bl)); r++)
		_bio_endio(rs, bio, -EIO);

	return r;
}

/* Fail all ios of a bio list of a stripe and drop io pending count */
static void stripe_bio_list_fail(struct raid_set *rs, struct stripe *stripe,
				 struct bio_list *bl)
{
	unsigned int put = bio_list_fail(rs, stripe, bl);

	while (put--)
		stripe_put(stripe);
}

/* Fail all ios hanging off all bio lists of a stripe */
static void stripe_ios_fail(struct raid_set *rs, struct stripe *stripe)
{
	unsigned int p = rs->set.raid_devs;

	stripe_evict(rs, stripe);

	while (p--) {
		struct stripe_set *ss = stripe->ss + p;

		stripe_bio_list_fail(rs, stripe, &ss->reads);
		stripe_bio_list_fail(rs, stripe, &ss->writes);
		stripe_bio_list_fail(rs, stripe, &ss->writes_merged);
	}
}

/*
 * Handle all stripes by handing them to the daemon, because we can't
 * map their pages to copy the data in interrupt context.
 *
 * We don't want to handle them here either, while interrupts are disabled.
 */

/* Read/write endio function for dm-io (interrupt context) */
static void endio(unsigned long error, void *context)
{
	struct dm_mem_cache_object *obj = context;
	struct stripe *stripe = ((struct stripe_set *)obj->private)->stripe;
	struct page *page = obj->pl->page;

	if (unlikely(error))
		stripe_error(stripe, page);
	else {
		ClearPageDirty(page);
		SetPageUptodate(page);
	}

	if (stripe_io_dec(stripe))
		/* Add stripe to endio list and wake daemon */
		stripe_endio_push(stripe);
}

/*
 * Recovery io throttling
 */
/* Conditionally reset io counters */
enum count_type { IO_WORK = 0, IO_RECOVER };
static int recover_io_reset(struct raid_set *rs)
{
	unsigned long j = jiffies;

	/* Pay attention to jiffies overflows */
	if (j > rs->recover.last_jiffies + 5 * HZ ||
	    j < rs->recover.last_jiffies) {
		rs->recover.last_jiffies = j;
		atomic_set(rs->recover.io_count + IO_WORK, 0);
		atomic_set(rs->recover.io_count + IO_RECOVER, 0);
		return 1;
	}

	return 0;
}

/* Count ios */
static inline void recover_io_count(struct raid_set *rs, struct stripe *stripe)
{
	if (raid_set_recovering(rs)) {
		recover_io_reset(rs);
		atomic_inc(rs->recover.io_count +
			   (stripe == rs->recover.stripe ?
			    IO_RECOVER : IO_WORK));
	}
}

/* Read/Write a page_list asynchronously */
static void page_list_rw(struct stripe *stripe, unsigned int p)
{
	struct raid_set *rs = RS(stripe->sc);
	struct dm_mem_cache_object *obj = stripe->obj + p;
	struct page_list *pl = obj->pl;
	struct page *page = pl->page;
	struct raid_dev *dev = rs->dev + p;
	struct io_region io = {
		.bdev = dev->dev->bdev,
		.sector = stripe->key,
		.count = stripe->io.size,
	};
	struct io_control control = {
		.rw = PageDirty(page) ? WRITE : READ,
		.mem.type = IO_PAGE_LIST,
		.mem.ptr.pl = pl,
		.mem.offset = 0,
		.notify.fn = endio,
		.notify.context = obj,
		.client = stripe->sc->dm_io_client,
	};

	/*
	 * Don't rw past end of device, which can happen, because
	 * typically sectors_per_dev isn't divisable by io_size
	 */
	if (unlikely(io.sector + io.count > rs->set.sectors_per_dev))
		io.count = rs->set.sectors_per_dev - io.sector;

	io.sector += dev->start; /* Add <offset> */
	recover_io_count(rs, stripe); /* Recovery io accounting */

	/* REMOVEME: */
	atomic_inc(PageDirty(page) ? &rs->stats.wc : &rs->stats.rc);

	BUG_ON(dm_io(&control, 1, &io));
}

/*
 * Write dirty / read not uptodate page lists of a stripe.
 */
static void stripe_page_lists_rw(struct raid_set *rs, struct stripe *stripe)
{
	/* Initially assume stripe will become active (io being performed) */
	SetStripeActive(stripe);

	/*
	 * Increment the pending count on the stripe
	 * first, so that we don't race in endio().
	 *
	 * An inc (IO) is needed for any page:
	 *
	 * o not uptodate
	 * o dirtied by writes merged
	 * o dirtied by parity calculations
	 */
	for_each_io_dev(rs, stripe, _stripe_io_inc);

	/* io needed ? (chunks are not uptodate/dirty) */
	if (likely(stripe_io(stripe))) {
		int max; /* REMOVEME: */

		/*
		 * Take off the lru list in case it got added
		 * there by degraded mode endios causing the
		 * stripe reference count to drop to zero
		 */
		stripe_lru_off(stripe);

		/* Submit actual io */
		for_each_io_dev(rs, stripe, page_list_rw);
		atomic_inc(&rs->sc.active_stripes);

		/* REMOVEME: */
		max = sc_active(&rs->sc);
		if (atomic_read(&rs->sc.max_active_stripes) < max)
			atomic_set(&rs->sc.max_active_stripes, max);

		atomic_inc(&rs->stats.flushs); /* REMOVEME: */
	} else
		ClearStripeActive(stripe); /* None -> reset active */
}

/* Work in all pending writes */
static inline void _writes_merge(struct raid_set *rs, struct stripe *stripe,
				 unsigned int p)

{
	struct bio *bio;
	struct dm_mem_cache_object *obj = stripe->obj + p;
	struct stripe_set *ss = obj->private;
	struct bio_list *writes = &ss->writes;

	if (bio_list_empty(writes))
		return;

	/*
	 * We can play with the lists without holding a lock,
	 * because it is just us accessing them anyway
	 */
	while ((bio = bio_list_pop(writes))) {
		bio_copy_page_list(WRITE, stripe, obj->pl, bio);
		bio_list_add(&ss->writes_merged, bio);
	}

	/*
	 * Only dirty chunks on intact devices
	 * in order to avoid LLD complaints
	 */
	dev_operational(rs, p) ? SetPageDirty(obj->pl->page) :
				 ClearPageDirty(obj->pl->page);
}

/*
 * Merge in all writes hence dirtying
 * respective pages and do parity calculation
 */
static inline void writes_merge(struct raid_set *rs, struct stripe *stripe)
{
	unsigned int p = rs->set.raid_devs;

	while (p--)
		_writes_merge(rs, stripe, p);

	/* Now that all writes got merged, calculate parity */
	parity_xor(stripe);
}

/*
 * Check, if a stripe has pages which will get completely
 * overwritten in order to avoid superfluous reads before writes.
 */
static inline void stripe_check_overwrite(struct raid_set *rs,
					  struct stripe *stripe)
{
	if (test_bit(CHECK_OVERWRITE, &rs->flags)) {
		unsigned int p = rs->set.raid_devs, stripe_sets = p, sectors;
		struct dm_mem_cache_object *obj;
		struct bio_list *bl;
		struct bio *bio;
	
		atomic_inc(&rs->stats.check_overwrite); /* REMOVEME: */
	
		/* Walk all stripe sets */
		while (p--) {
			obj = stripe->obj + p;
	
			/*
			 * I avoid to read the parity chunk,
			 * because it is going to be rewritten anyway
			 */
			if (p == stripe->context.pi)
				goto update;
	
			/* Sum up all sectors being written to the chunk */
			bl = &((struct stripe_set*) obj->private)->writes;
			sectors = 0;
			bio_for_each(bio, bl)
				sectors += bio_sectors(bio);
			/*
			 * If all pages of a chunk get written
			 * over, I avoid to read them
			 */
			if (sectors == rs->set.io_size) {
   update:
				SetPageUptodate(obj->pl->page);
				stripe_sets--;
				/* REMOVEME: */
				atomic_inc(&rs->stats.wr_chunk_updated);
			}
		}
	
		/* All pages will be overwritten -> set stripe uptodate */
		if (!stripe_sets) {
			SetStripeUptodate(stripe);
			/* REMOVEME: */
			atomic_inc(&rs->stats.wr_stripe_updated);
		}
	}
}

/*
 * Make sure all non-dirty pages are read in
 * in order to be able to reconstruct a missing
 * chunk during endio processing
 */
static inline void stripe_degrade(struct raid_set *rs, struct stripe *stripe,
				  unsigned int p)
{
	struct page *page = stripe->obj[p].pl->page;

	stripe_pages_invalidate(rs, stripe);

	/*
	 * In case the RAID set is operational and this is
	 * an optimization to avoid reading a chunk, mark as
	 * such using the checked flag
	 */
	if (likely(!raid_set_degraded(rs)))
		SetPageChecked(page);

	SetPageUptodate(page);
	SetStripeReconstruct(stripe);
}

/*
 * Degraded/reconstruction mode
 *
 * Return 0 to caller to tell, that no stripe overwrite
 * optimization may be carried out.
 *
 * If degraded and stripe not uptodate, (*) clear uptodate on
 * all stripe sets of a stripe in order to read them all in
 * for data/parity rebuilds.
 *
 * If resynchronizing and reading, do (*) on nosync regions in
 * order to avoid reading in broken (iow: not yet updated by
 * resynchronization) data.
 */
static inline int stripe_check_state(struct raid_set *rs, struct stripe *stripe)
{
	int r = 1; /* Allow overwrite optimization */

	if (StripeUptodate(stripe))
		return r;

	/*
	 * Degraded mode (device(s) failed) ->
	 * make sure all non-dirty pages are read in
	 */
	if (unlikely(raid_set_degraded(rs))) {
		stripe_degrade(rs, stripe, rs->set.ei);
		atomic_inc(&rs->stats.degraded); /* REMOVEME: */
		r = 0; /* Prohibit overwrite optimization! */

	/*
	 * Reconstruction mode (ie. a particular device or some (rotating)
         * parity chunk is being resynchronized) ->
	 *   o make sure all needed pages are read in
	 *   o writes are allowed to go through
	 */
	} else if (StripeRead(stripe) ||
		   StripeRBW(stripe)) {
		int rr = region_state(rs, stripe->key, RH_NOSYNC);

		if (rr) {
			stripe_degrade(rs, stripe, dev_to_init(rs) ?
				       rs->set.dev_to_init :
				       stripe->context.pi);
			atomic_inc(&rs->stats.nosync); /* REMOVEME: */
			r = 0; /* Prohibit overwrite optimization! */
		}
	}

	return r;
}

/*
 * Read/write a stripe.
 *
 * Return != 0 -> no ios queued.
 */
static int stripe_rw(struct raid_set *rs, struct stripe *stripe)
{
	int r;

	/*
	 * If new reads queued and stripe uptodate -> end io
	 *
	 * We can run this on an active stripe, because
	 * chunk content won't change under us.
	 */
	if (StripeUptodate(stripe)) {
		int r; /* REMOVEME: */

		r = do_end_io_reads(rs, stripe); /* REMOVEME: return value */

		/* REMOVEME: statistics */
		if (r) {
			atomic_inc(&rs->stats.fast_reads);

			if (StripeActive(stripe))
				atomic_inc(&rs->stats.fast_active_reads);
		}
	}

	/* Don't try initiating io on already active stripe */
	if (StripeActive(stripe)) {
		atomic_inc(&rs->stats.eperms); /* REMOVEME: */
		return -EPERM;
	}

	/*
	 * Once we get here, we're the only one accessing this stripe
	 * because there's no pending ios from/to it which will cause
	 * endios from dm-io and noone (including recovery) but
	 * stripe_rw() will ever initate any.
	 *
	 * Check the state of the RAID set and if degraded (or
	 * resynchronizing for reads), read in all other chunks but
	 * the one on the dead/resynchronizing device in order to be
	 * able to reconstruct the missing one.
	 *
	 * If fully operational (IOW: not degraded or resynchronizing):
	 * check if writes will completely overwrite the pages of a
	 * stripe and avoid reads before writes when possible.
	 *
	 * Merge all writes hanging off uptodate pages of the stripe.
	 */
	r = stripe_check_state(rs, stripe);

	if (StripeRBW(stripe)) {
		if (r && !StripeUptodate(stripe))
			stripe_check_overwrite(rs, stripe);

		if (StripeUptodate(stripe)) {
			writes_merge(rs, stripe); /* Merge writes in */
			ClearStripeRBW(stripe);	  /* Disable RBW */

	 		/*
			 * If we're degraded ->
			 * end io writes to the dead device here,
			 * because, they could be the only ones
			 * and no actual io would be submitted
			 * in stripe_page_lists_rw()
			 */
			if (unlikely(raid_set_degraded(rs)))
				do_end_io_pl(WRITE, rs, stripe,
					     stripe->obj[rs->set.ei].pl,
					     &stripe->ss[rs->set.ei]);
		}
	}

	/* Now submit any reads/writes */
	stripe_page_lists_rw(rs, stripe);
	return 0;
}

/*
 * Recovery functions
 */
/* Read a stripe off a raid set for recovery */
static void recover_read(struct raid_set *rs, struct stripe *stripe, int idx)
{
	/* Invalidate all pages so that they get read in */
	stripe_pages_invalidate(rs, stripe);

	/* Avoid reading the page(s) we will reconstruct anyway */
	if (idx > -1)
		SetPageUptodate(stripe->obj[idx].pl->page);

	stripe->key = rs->recover.pos;
	stripe_page_lists_rw(rs, stripe);
}

/* Write a stripe to a raid set for recovery */
static void recover_write(struct raid_set *rs, struct stripe *stripe, int idx)
{
	/*
	 * If this is a reconstruct of a particular device, then
	 * reconstruct the respective page(s), else create parity page(s).
	 */
	if (idx > -1) {
		error_xor(stripe, idx);
		SetPageDirty(stripe->obj[idx].pl->page);
	} else
		parity_xor(stripe);

	stripe_page_lists_rw(rs, stripe);
}

/* Recover bandwidth available ? */
static int recover_bandwidth(struct raid_set *rs)
{
	int r, work;

	/* On reset -> allow recovery */
	r = recover_io_reset(rs);
	if (r)
		goto out;

	work = atomic_read(rs->recover.io_count + IO_WORK);
	if (work) {
		/* Pay attention to larger recover stripe size */
		int recover = atomic_read(rs->recover.io_count + IO_RECOVER) *
			      rs->recover.stripe->io.size / rs->set.io_size;

		/*
		 * Don't use more than given bandwidth of
		 * the work io for recovery
		 */
		if (recover > work / rs->recover.bandwidth_work) {
			atomic_inc(&rs->stats.no_bandwidth); /* REMOVEME: */
			return 0;
		}
	}

   out:
	atomic_inc(&rs->stats.bandwidth); /* REMOVEME: */
	return 1;
}

/* Try to get a region to recover */
static int recover_get_region(struct raid_set *rs)
{
	int r;
	struct recover *rec = &rs->recover;
	void *rh = rec->rh;

	/* Start quiescing some regions */
	if (!test_bit(REGION_GET, &rs->flags)) {
		r = recover_bandwidth(rs); /* Enough bandwidth ? */
		if (r) {
			if (rh_recovery_prepare(rh) < 0) {
				DMINFO(PREFIX "No %sregions to recover",
				       rec->nr_regions_to_recover ?
				       "more " : "");
				return -ENOENT;
			}
		} else
			return -EAGAIN;

		set_bit(REGION_GET, &rs->flags);
	}

	if (!rec->reg) {
		rec->reg = rh_recovery_start(rh);
		if (rec->reg) {
			io_get(rs); /* One io reference for the region */
			rec->pos =
			   rh_region_to_sector(rh, rh_get_region_key(rec->reg));
			rec->end = rec->pos + rh_get_region_size(rh);
			ClearStripeUptodate(rec->stripe);
			ClearStripeRead(rec->stripe);
		} else
			return -EAGAIN;
	}

	return 0;
}

/* Return the index of the device to be recovered */
static int idx_get(struct raid_set *rs)
{
	/* Avoid to read in the pages to be reconstructed anyway */
	if (dev_to_init(rs))
		return rs->set.dev_to_init;
	else if (rs->set.raid_type->level == raid4)
		return rs->set.pi;

	return -1;
}

/* Read/write a recovery stripe */
static inline int recover_stripe_rw(struct raid_set *rs, struct stripe *stripe)
{
	/* Read/write flip-flop */
	if (StripeRead(stripe)) {
		ClearStripeRead(stripe);
		recover_write(rs, stripe, idx_get(rs));
		return 1;
	} else if (!StripeUptodate(stripe)) {
		SetStripeRead(stripe);
		recover_read(rs, stripe, idx_get(rs));
		return 1;
	}

	return 0;
}

/* Reset recovery variables */
static inline void recovery_reset(struct raid_set *rs)
{
	rs->recover.reg = 0;
	clear_bit(REGION_GET, &rs->flags);
}

/* Update region hash state */
static void recover_rh_update(struct raid_set *rs, int error)
{
	struct recover *rec = &rs->recover;
	void *rh = rs->recover.rh;
	void *reg = rec->reg;

	if (reg) {
		rh_recovery_end(reg, error);
		rec->nr_regions_recovered++;
		recovery_reset(rs);
	}

	rh_update_states(rh);
	rh_flush(rh);
	io_put(rs); /* Release the io reference for the region */
}

/* Called by main io daemon to recover regions */
static void do_recovery(struct raid_set *rs)
{
	/* FIXME: end_jiffies could be set on overflow and hence be zero */
	if (!rs->recover.end_jiffies) {
		int r;
		char buf[BDEVNAME_SIZE];
		struct recover *rec = &rs->recover;
		struct stripe *stripe = rec->stripe;

		if (raid_set_recovering(rs)) {
			/* If recovery is active -> return; */
			if (StripeActive(stripe))
				return;

			if (unlikely(StripeError(stripe)))
				goto err;

			/* Get a region to recover */
			r = recover_get_region(rs);
			switch (r) {
			case 0:		/* Got a region */
				break;

			case -EAGAIN:	/* No quiesced regions yet, try later */
				wake_do_raid_delayed(rs, HZ / 10);
				return;

			case -ENOENT:	/* No more regions */
				goto free;
			}

			/* Read/write a recover stripe */
			r = recover_stripe_rw(rs, stripe);

			/* Io initiated */
			if (r) {
				/* Another reference for the io on the region */
				io_get(rs);
				return;
			}

			ClearStripeUptodate(stripe);

			/* Update recovery position within region */
			rec->pos += stripe->io.size;
			if (rec->pos >= rec->end ||
			    rec->pos >= rs->set.sectors_per_dev)
				recover_rh_update(rs, 0);

			wake_do_raid(rs);
		}

		return;

   err:
		raid_set_check_degrade(rs, stripe);
		DMERR(PREFIX "stopping recovery due to "
		      "ERROR on /dev/%s %sing stripe at offset %llu",
		      bdevname(rs->dev[rs->set.ei].dev->bdev, buf),
		      StripeUptodate(stripe) ? "writ" : "read",
		      stripe->key);
	
		/* Make sure, that all quiesced regions get released */
		do {
			if (rec->reg)
				rh_recovery_end(rec->reg, -EIO);

			rec->reg = rh_recovery_start(rec->rh);
		} while (rec->reg);
	
		recover_rh_update(rs, -EIO);
   free:
		rs->recover.end_jiffies = jiffies;
		/* Check for jiffies overrun */
		rs->recover.end_jiffies =
			rs->recover.end_jiffies ? rs->recover.end_jiffies : ~0;
		stripe_recover_free(rs);
	}
}
/*
 * END recovery functions
 */

/* End io process all stripes handed in by endio() callback */
static void do_endios(struct raid_set *rs)
{
	int r;
	struct stripe_cache *sc = &rs->sc;
	struct stripe *stripe;

	while ((stripe = stripe_endio_pop(sc))) {
		/* Set stripe inactive, so that we can reuse it */
		ClearStripeActive(stripe);
		atomic_dec(&sc->active_stripes);

		/* Recovery stripe special case */
		if (unlikely(StripeRecover(stripe))) {
			io_put(rs); /* Release region io reference */
			SetStripeUptodate(stripe);
			continue;
		}

		/*
		 * If an error on a stripe occured, check it's pages
		 * in order to optionally requeue the stripe for io
		 * in case we avoided io for optimization in
		 * stripe_check_state() and are now missing data
		 * to reconstruct because another device failed on us.
		 */
		if (unlikely(StripeError(stripe))) {
			r = raid_set_check_degrade(rs, stripe);
			if (r) {
				/*
				 * We've a chunk too few because the
				 * failure happend on a different one
				 * -> requeue to be read in
				 */
				stripe_to_flush_sorted(stripe);
				continue;
			}
		}

		/*
		 * Check if the RAID set is inoperational and
		 * error ios or if a recovery rebuild is needed...
		 */
		if (unlikely(!raid_set_operational(rs))) {
			stripe_ios_fail(rs, stripe);
			BUG_ON(atomic_read(&stripe->use));
			goto out;
		}

		if (StripeError(stripe) || StripeReconstruct(stripe))
			reconstruct_xor(stripe);

		/*
		 * When we get here, the set is either operational
		 * or we were able to reconstruct a missing chunk.
		 * Hence we can set the stripe to uptodate
		 */
		SetStripeUptodate(stripe);

		/*
		 * Now that we've got an uptodate stripe,
		 * we can process end ios on reads
		 */
		do_end_io_reads(rs, stripe);

		/*
		 * Read-before-write stripes need to be
		 * flushed again in order to work the
		 * write data into the pages *after*
		 * they were read in
		 */
		if (StripeRBW(stripe))
			stripe_to_flush_sorted(stripe);
		else
			do_end_io_writes(rs, stripe);
   out:
		/*
		 * Conditionally put on lru list in case stripe got
		 * unreferenced before by degraded mode end ios and
		 * got taken off the lru list by parity writes
		 */
		_stripe_get(stripe);
		stripe_put(stripe);
	}
}

/*
 * Stripe cache shrinking
 */
static inline void do_sc_shrink(struct raid_set *rs)
{
	unsigned int shrink = atomic_read(&rs->sc.stripes_to_shrink);

	if (shrink) {
		unsigned int cur = atomic_read(&rs->sc.stripes);

		sc_shrink(&rs->sc, shrink);
		shrink -= cur - atomic_read(&rs->sc.stripes);
		atomic_set(&rs->sc.stripes_to_shrink, shrink);

		/*
		 * Wake daemon up in case we failed
		 * to shrink the requested amount
		 */
		if (shrink)
			wake_do_raid(rs);
	}
}


/*
 * Process all ios
 *
 * We do different things with the io depending on the
 * state of the region that it's in:
 *
 * o reads: hang off stripe cache or postpone if full
 *
 * o writes:
 *
 *  CLEAN/DIRTY/NOSYNC:	increment pending and hang io off stripe's stripe set.
 *			In case stripe cache is full or busy, postpone the io.
 *
 *  RECOVERING:		delay the io until recovery of the region completes.
 *
 */
static inline void do_ios(struct raid_set *rs)
{
	int r;
	void *rh;
	struct bio_list *ios = &rs->io.work;
	struct bio *bio;
	BIO_LIST_HEAD(delay);
	BIO_LIST_HEAD(process);
	BIO_LIST_HEAD(reject);

	if (bio_list_empty(ios))
		return;

	/*
	 * Classify each io:
	 *    o delay to recovering regions
	 *    o queue to all other regions
	 */
	rh = rs->recover.rh;
	while ((bio = bio_list_pop(ios))) {
		/*
		 * In case we get a barrier bio, push it back onto
		 * the input queue unless all work queues are empty
		 * and the stripe cache is inactive.
		 */
		if (unlikely(bio_barrier(bio))) {
			if (!bio_list_empty(&process) ||
			    !bio_list_empty(&delay) ||
			    sc_active(&rs->sc)) {
				bio_list_push(ios, bio);
				break;
			}
		}

		bio_list_add(region_state(rs, _sector(rs, bio), RH_RECOVERING) ?
 			     &delay : &process, bio);
	}

	/*
	 * Process ios to non-recovering regions by queueing
	 * them to stripes (does rh_inc()) for writes)
	 */
	r = stripe_add_rw(rs, &process, &reject, &rs->stats.ios_post);
	if (r)
		rh_flush(rh); /* In case ios got queued -> flush dirty log */

	bio_list_join(ios, &reject); /* Join rejected ios back to input list */

	/* Delay ios to regions which are recovering */
	while ((bio = bio_list_pop(&delay))) {
		atomic_inc(&rs->stats.delayed_bios); /* REMOVEME: */
		atomic_inc(&rs->stats.sum_delayed_bios); /* REMOVEME: */
		rh_delay_by_region(rh, bio,
				   rh_sector_to_region(rh, _sector(rs, bio)));
	}
}

/*
 * Flush the stripe cache by iterating
 * all stripes hanging off its io_list
 */
static inline void do_cache_flush(struct raid_set *rs)
{
	struct list_head *list = &rs->sc.lists.io, *pos, *tmp;

	list_for_each_safe(pos, tmp, list) {
		int r;
		struct stripe *stripe =
			list_entry(pos, struct stripe, lists.io);

		if (likely(raid_set_operational(rs)))
			r = stripe_rw(rs, stripe); /* Read/write stripe */
		else {
			/*
			 * Optimization:
			 * Fail early on failed sets and error stripe
			 */
			stripe_ios_fail(rs, stripe);
			r = 0;
		}

		/* Remove from the list only if the stripe got processed */
		if (!r)
			list_del_init(pos);
	}

	/* Send an event in case we're getting too busy */
	if ((sc_active(&rs->sc) > atomic_read(&rs->sc.stripes) * 4 / 5) &&
	    !test_and_set_bit(SC_BUSY, &rs->flags))
		dm_table_event(rs->ti->table);
}

/*-----------------------------------------------------------------
 * RAID daemon
 *---------------------------------------------------------------*/
/*
 * o belabour all end ios
 * o optionally shrink the stripe cache
 * o update the region hash states
 * o try to do recovery
 * o grab the input queue
 * o work an all requeued or new ios and perform stripe cache flushs
 *   unless the RAID set is inoperational (when we error ios)
 */
static void do_raid(void *context)
{
	struct raid_set *rs = context;

	/*
	 * We always need to end io, so that ios
	 * can get errored in case the set failed
	 * and the region counters get decremented
	 * before we update the region hash states
	 */
	do_endios(rs);

	/*
	 * Now that we've end io'd, which put stripes on the
	 * LRU list, we shrink the stripe cache if requested
	 */
	do_sc_shrink(rs);

	/* Update region hash states before we go any further */
	rh_update_states(rs->recover.rh);

	/* Try to recover regions */
	do_recovery(rs);

	/* Quickly grab all new ios queued and add them to the work list */
	spin_lock_irq(&rs->io.lock);
	bio_list_merge_init(&rs->io.work, &rs->io.in);
	spin_unlock_irq(&rs->io.lock);

	/* Let's assume we're operational most of the time ;-) */
	if (likely(raid_set_operational(rs))) {
		do_ios(rs);
		do_cache_flush(rs);
	} else
		/* No way to reconstruct data with too many devices failed */
		bio_list_fail(rs, NULL, &rs->io.work);
}

/*
 * Callback for region hash to dispatch
 * delayed bios queued to recovered regions
 * (Gets called via rh_update_states())
 */
static void dispatch_delayed_bios(void *context, struct bio_list *bl)
{
	struct raid_set *rs = context;
	struct bio *bio;

	 /* REMOVEME: decrement pending delayed bios counter */
	bio_for_each(bio, bl)
		atomic_dec(&rs->stats.delayed_bios);

	/* Merge region hash private list to work list */
	bio_list_join_init(&rs->io.work, bl);
}

/*************************************************************
 * Constructor helpers
 *************************************************************/
/*
 * Discover fastest xor algorithm and # of chunks combination
 */
/* Calculate speed for algorithm and # of chunks */
#define XOR_SPEED_SIZE	SECTORS_PER_XOR
#define	XOR_SPEED_RUNS	4
static unsigned inline int xor_speed(struct raid_set *rs)
{
	unsigned int ret = 0, runs = XOR_SPEED_RUNS;
	unsigned long j;

	/* Wait till next tick */
	for (j = jiffies; j == jiffies; );

	/* Do xors for a few ticks */
	while (runs--) {
		/* Do xors for a full tick */
		for (j = jiffies; j == jiffies; ) {
			mb();
			common_xor(rs->recover.stripe, XOR_SPEED_SIZE, 0, 0);
			mb();
			ret++;
			mb();
		}
	}

	return ret;
}

/* xor algorithm optimizer */
static void xor_optimize(struct raid_set *rs)
{
	unsigned int runs = 2, chunks, chunks_max = 2, speed_max = 0;
	struct xor_func *f_max = NULL;

	while (runs--) {
		/* We nee to vary rs->xor.chunks for xor() to work */
		for (chunks = 2; chunks <= XOR_CHUNKS_MAX; chunks++) {
			struct xor_func *f = ARRAY_END(xor_funcs);

			rs->xor.chunks = chunks;
			while (f-- > xor_funcs) {
				unsigned int speed;

				rs->xor.f = f->f;
				speed = xor_speed(rs);

				if (speed > speed_max) {
					speed_max = speed;
					chunks_max = rs->xor.chunks;
					f_max = f;
				}
			}
		}
	}

	rs->xor.f = f_max->f;
	rs->xor.name = f_max->name;
	rs->xor.chunks = chunks_max;
	rs->xor.speed = speed_max;
}

/*
 * Allocate a RAID context (a RAID set)
 */
static int context_alloc(struct raid_set **raid_set,
			 struct raid_type *raid_type,
			 unsigned int stripes,
			 unsigned int chunk_size, unsigned int io_size,
			 unsigned int recover_io_size,
			 unsigned int raid_devs, sector_t sectors_per_dev,
			 struct dm_target *ti,
			 unsigned int dl_parms, char **argv)
{
	int r;
	unsigned int max;
	size_t len;
	sector_t region_size, ti_len;
	struct raid_set *rs = NULL;
	struct dirty_log *dl;

	/*
	 * Create the dirty log
	 *
	 * We need to change length for the dirty log constructor,
	 * because we want an amount of regions for all stripes derived
	 * from the single device size, so that we can keep region
	 * size = 2^^n independant of the number of devices
	 */
	ti_len = ti->len;
	ti->len = sectors_per_dev;

	dl = dm_create_dirty_log(argv[0], ti, dl_parms, argv + 2);
	ti->len = ti_len;
	if (!dl)
		goto bad_dirty_log;

	/* Chunk size *must* be smaller than region size */
	region_size = dl->type->get_region_size(dl);
	if (chunk_size > region_size)
		goto bad_chunk_size;

	/* Recover io size *must* be smaller than region size as well */
	if (recover_io_size > region_size)
		goto bad_recover_io_size;

	/* Size and allocate the RAID set structure */
	len = sizeof(*rs->data) + sizeof(*rs->dev);
	if (array_too_big(sizeof(*rs), len, raid_devs))
		goto bad_array;

	len = sizeof(*rs) + raid_devs * len;
	rs = kzalloc(len, GFP_KERNEL);
	if (!rs)
		goto bad_alloc;

	atomic_set(&rs->io.in_process, 0);
	rs->recover.io_size = recover_io_size;

	rs->data = (xor_t**) ((void*) rs->dev + raid_devs * sizeof(*rs->dev));
	rs->recover.dl = dl;
	rs->set.raid_devs = raid_devs;
	rs->set.data_devs = raid_devs - raid_type->parity_devs;
	rs->set.raid_type = raid_type;

	/*
	 * Set chunk and io size and respective shifts
	 * (used to avoid divisions)
	 */
	rs->set.chunk_size = chunk_size;
	rs->set.chunk_mask = chunk_size - 1;
	rs->set.chunk_shift = ffs(chunk_size) - 1;

	rs->set.io_size = io_size;
	rs->set.io_mask = io_size - 1;
	rs->set.io_shift = ffs(io_size) - 1;
	rs->set.io_shift_mask = rs->set.chunk_mask & ~rs->set.io_mask;

	rs->set.pages_per_io = dm_div_up(io_size, SECTORS_PER_PAGE);
	rs->set.sectors_per_dev = sectors_per_dev;

	rs->set.ei = -1; /* Indicate no failed device */
	atomic_set(&rs->set.failed_devs, 0);

	rs->ti = ti;

	atomic_set(rs->recover.io_count + IO_WORK, 0);
	atomic_set(rs->recover.io_count + IO_RECOVER, 0);

	/* Initialize io lock and queues */
	spin_lock_init(&rs->io.lock);
	bio_list_init(&rs->io.in);
	bio_list_init(&rs->io.work);

	/* Io throttling: maximum ios queueable */
	max = 8 * stripes * rs->set.pages_per_io * rs->set.raid_devs;
	if (!range_ok(max, MIN_IO_COUNT, MAX_IO_COUNT))
		max = max > MAX_IO_COUNT ? MAX_IO_COUNT : MIN_IO_COUNT;

	sema_init(&rs->io.max, max);		/* IO throttling */
	init_waitqueue_head(&rs->io.suspendq);	/* Suspend waiters (dm-io) */

	rs->recover.nr_regions = dm_sector_div_up(sectors_per_dev, region_size);
	r = rh_init(&rs->recover.rh, 1, dispatch_delayed_bios, rs,
		    wake_do_raid, rs, dl, region_size, rs->recover.nr_regions);
	if (r)
		goto bad_rh;

	/* Initialize stripe cache */
	r = sc_init(rs, stripes);
	if (r)
		goto bad_sc;

	/* Create dm-io client context */
	r = dm_io_client_create(rs->set.raid_devs * rs->set.pages_per_io,
				&rs->sc.dm_io_client);
	if (r)
		goto bad_dm_io_client;

	/* Select best xor algorithm */
	xor_optimize(rs);

	/* REMOVEME: */
	stats_reset(rs);
	clear_bit(DEVEL_STATS, &rs->flags); /* Disable development status */

	*raid_set = rs;
        return 0;

   bad_dirty_log:
	TI_ERR_RET("Error creating dirty log", -ENOMEM);

   bad_chunk_size:
	dm_destroy_dirty_log(dl);
	TI_ERR("Chunk size larger than region size");

   bad_recover_io_size:
	dm_destroy_dirty_log(dl);
	TI_ERR("Recover stripe io size larger than region size");

   bad_array:
	dm_destroy_dirty_log(dl);
	TI_ERR("Arry too big");

   bad_alloc:
	dm_destroy_dirty_log(dl);
	TI_ERR_RET("Cannot allocate raid context", -ENOMEM);

   bad_rh:
	dm_destroy_dirty_log(dl);
	ti->error = PREFIX "Error creating dirty region hash";
	goto free_rs;

   bad_sc:
	ti->error = PREFIX "Error creating stripe cache";
	goto free;

   bad_dm_io_client:
	ti->error = PREFIX "Error allocating dm-io resources";
   free:
	sc_exit(&rs->sc);
	rh_exit(rs->recover.rh);
   free_rs:
	kfree(rs);
	return -ENOMEM;
}

/* Free a RAID context (a RAID set) */
static void context_free(struct raid_set *rs, struct dm_target *ti,
			 unsigned int r)
{
	while (r--)
		dm_put_device(ti, rs->dev[r].dev);

	dm_io_client_destroy(rs->sc.dm_io_client);
	stripe_recover_free(rs);
	sc_exit(&rs->sc);
	rh_exit(rs->recover.rh); /* Destroys dirty log as well */

	kfree(rs);
}

/* Create work queue and initialize work */
static int rs_workqueue_init(struct raid_set *rs)
{
	struct dm_target *ti = rs->ti;

	rs->io.wq = create_singlethread_workqueue(DAEMON);
	if (!rs->io.wq)
		TI_ERR_RET("failed to create " DAEMON, -ENOMEM);

	rs->recover.start_jiffies = jiffies;
	rs->recover.end_jiffies = 0;
	INIT_WORK(&rs->io.ws, do_raid, rs);
	return 0;
}

/* Return pointer to raid_type structure for raid name */
static struct raid_type *get_raid_type(char *name)
{
	struct raid_type *r = ARRAY_END(raid_types);

	while (r-- > raid_types) {
		if (!strnicmp(name, r->name, strlen(r->name)))
			return r;
	}

	return NULL;
}

/* FIXME: factor out to dm core */
static int multiple(sector_t a, sector_t b, sector_t *n)
{
	sector_t r = a;

	sector_div(r, b);
	*n = r;
	return a == r * b;
}

/* Calculate MB/sec */
static inline unsigned int mbpers(struct raid_set *rs)
{
	return to_bytes(rs->xor.speed * rs->set.data_devs * XOR_SPEED_SIZE * HZ / XOR_SPEED_RUNS >> 10) >> 10;
}

/* Log RAID set information to kernel log */
static void raid_set_log(struct raid_set *rs)
{
	unsigned int p;
	char buf[BDEVNAME_SIZE];

	for (p = 0; p < rs->set.raid_devs; p++)
		DMINFO(PREFIX "/dev/%s is raid disk %u",
		       bdevname(rs->dev[p].dev->bdev, buf), p);

	DMINFO(PREFIX "%d/%d sectors chunk/io size and %u stripes",
	       rs->set.chunk_size, rs->set.io_size,
	       atomic_read(&rs->sc.stripes));
	DMINFO(PREFIX "algorithm \"%s\", %u chunks with %uMB/s",
	       rs->xor.name, rs->xor.chunks, mbpers(rs));
	DMINFO(PREFIX "%s set with net %u/%u devices",
	       rs->set.raid_type->descr, rs->set.data_devs, rs->set.raid_devs);
}

/* Get all devices and offsets */
static int dev_parms(struct dm_target *ti, struct raid_set *rs,
		     char **argv, unsigned int *p)
{
	for (*p = 0; *p < rs->set.raid_devs; (*p)++, argv += 2) {
		int r;
		struct raid_dev *dev = rs->dev + *p;
		union dev_lookup dl = { .dev = dev };

		/* Get offset and device */
		r = sscanf(argv[1], "%llu", &dev->start);
		if (r != 1)
			TI_ERR("Invalid RAID device offset parameter");

		r = dm_get_device(ti, argv[0], dev->start,
				  rs->set.sectors_per_dev,
				  dm_table_get_mode(ti->table), &dev->dev);
		if (r)
			TI_ERR_RET("RAID device lookup failure", r);

		r = raid_dev_lookup(rs, bynumber, &dl);
		if (r != -ENODEV && r < *p) {
			(*p)++; /* Ensure dm_put_device() on actual device */
			TI_ERR_RET("Duplicate RAID device", -ENXIO);
		}
	}

	return 0;
}

/* Set recovery bandwidth */
static inline void recover_set_bandwidth(struct raid_set *rs,
					 unsigned int bandwidth)
{
	rs->recover.bandwidth = bandwidth;
	rs->recover.bandwidth_work = 100 / bandwidth;
}

/* Handle variable number of RAID parameters */
static int raid_variable_parms(struct dm_target *ti, char **argv,
			       unsigned int i, unsigned int *raid_parms,
			       int *chunk_size, int *chunk_size_parm,
			       int *stripes, int *stripes_parm,
			       int *io_size, int *io_size_parm,
			       int *recover_io_size, int *recover_io_size_parm,
			       int *bandwidth, int *bandwidth_parm)
{
	/* Fetch # of variable raid parameters */
	if (sscanf(argv[i++], "%d", raid_parms) != 1
	    || !range_ok(*raid_parms, 0, 5))
		TI_ERR("Bad variable raid parameters number");

	if (*raid_parms) {
		/*
		 * If we've got variable RAID parameters,
		 * chunk size is the first one
		 */
		if (sscanf(argv[i++], "%d", chunk_size) != 1 ||
		    (*chunk_size != -1 && (*chunk_size & (*chunk_size - 1) ||
	    	     !range_ok(*chunk_size, IO_SIZE_MIN, CHUNK_SIZE_MAX))))
			TI_ERR("Invalid chunk size; must be 2^^n and <= 16384");

		*chunk_size_parm = *chunk_size;
		if (*chunk_size == -1)
			*chunk_size = CHUNK_SIZE;

		/*
		 * In case we've got 2 or more variable raid
		 * parameters, the number of stripes is the second one
		 */
		if (*raid_parms > 1) {
			if (sscanf(argv[i++], "%d", stripes) != 1 ||
			    (*stripes != -1 &&
			     !range_ok(*stripes, STRIPES_MIN, STRIPES_MAX)))
				TI_ERR("Invalid number of stripes: must "
				       "be >= 8 and <= 8192");
		}

		*stripes_parm = *stripes;
		if (*stripes == -1)
			*stripes = STRIPES;

		/*
		 * In case we've got 3 or more variable raid
		 * parameters, the io size is the third one.
		 */
		if (*raid_parms > 2) {
			if (sscanf(argv[i++], "%d", io_size) != 1 ||
			    (*io_size != -1 && (*io_size & (*io_size - 1) ||
			     !range_ok(*io_size, IO_SIZE_MIN, *chunk_size))))
				TI_ERR("Invalid io size; must "
				       "be 2^^n and less equal chunk size");
		} else
			*io_size = *chunk_size;

		*io_size_parm = *io_size;
		if (*io_size == -1)
			*io_size = *chunk_size;

		/*
		 * In case we've got 4 variable raid parameters,
		 * the recovery stripe io_size is the fourth one
		 */
		if (*raid_parms > 3) {
			if (sscanf(argv[i++], "%d", recover_io_size) != 1 ||
			    (*recover_io_size != -1 &&
			     (*recover_io_size & (*recover_io_size - 1) ||
			      !range_ok(*recover_io_size, SECTORS_PER_PAGE,
					   CHUNK_SIZE_MAX))))
				TI_ERR("Invalid recovery io size; must "
				       "be 2^^n and less equal 16384");
		}

		*recover_io_size_parm = *recover_io_size;
		if (*recover_io_size == -1)
			*recover_io_size = RECOVER_IO_SIZE;

		/*
		 * In case we've got 5 variable raid parameters,
		 * the recovery IO bandwidth is the fifth one
		 */
		if (*raid_parms > 4) {
			if (sscanf(argv[i++], "%d", bandwidth) != 1 ||
			    (*bandwidth != -1 &&
			     !range_ok(*bandwidth,
					  BANDWIDTH_MIN, BANDWIDTH_MAX)))
				TI_ERR("Invalid recovery bandwidth "
				       "percentage; must be > 0 and <= 100");
		}

		*bandwidth_parm = *bandwidth;
		if (*bandwidth == -1)
			*bandwidth = BANDWIDTH;
	}

	return 0;
}

/*
 * Construct a RAID4/5 mapping:
 *
 * log_type #log_params <log_params> \
 * raid_type [#parity_dev] #raid_variable_params <raid_params> \
 * #raid_devs #dev_to_initialize [<dev_path> <offset>]{3,}
 *
 * log_type = "core"/"disk",
 * #log_params = 1-3 (1-2 for core dirty log type, 3 for disk dirty log only)
 * <log_params = [dirty_log_path] region_size [[no]sync])
 *
 * raid_type = "raid4", "raid5_la", "raid5_ra", "raid5_ls", "raid5_rs"
 *
 * #parity_dev = N if raid_type = "raid4"
 * o N = -1: pick default = last device
 * o N >= 0 and < #raid_devs: parity device index
 *
 * #raid_variable_params = 0-5; raid_params (-1 = default):
 *   [chunk_size [#stripes [io_size [recover_io_size [%recovery_bandwidth]]]]]
 *   o chunk_size (unit to calculate drive addresses; must be 2^^n, > 8
 *     and <= CHUNK_SIZE_MAX)
 *   o #stripes is number of stripes allocated to stripe cache
 *     (must be > 1 and < STRIPES_MAX)
 *   o io_size (io unit size per device in sectors; must be 2^^n and > 8)
 *   o recover_io_size (io unit size per device for recovery in sectors;
       must be 2^^n, > SECTORS_PER_PAGE and <= region_size)
 *   o %recovery_bandwith is the maximum amount spend for recovery during
 *     application IO (1-100%)
 * If raid_variable_params = 0, defaults will be used.
 * Any raid_variable_param can be set to -1 to apply a default
 *
 * #raid_devs = N (N >= 3)
 *
 * #dev_to_initialize = N
 * -1: initialize parity on all devices
 * >= 0 and < #raid_devs: initialize raid_path; used to force reconstruction
 * of a failed devices content after replacement
 *
 * <dev_path> = device_path (eg, /dev/sdd1)
 * <offset>   = begin at offset on <dev_path>
 *
 */
#define	MIN_PARMS	13
static int raid_ctr(struct dm_target *ti, unsigned int argc, char **argv)
{
	int bandwidth = BANDWIDTH, bandwidth_parm = -1,
	    chunk_size = CHUNK_SIZE, chunk_size_parm = -1,
	    dev_to_init, dl_parms, parity_parm, pi = -1,
	    i, io_size = IO_SIZE, io_size_parm = -1,
	    r, raid_devs, raid_parms,
	    recover_io_size = RECOVER_IO_SIZE, recover_io_size_parm = -1,
	    stripes = STRIPES, stripes_parm = -1;
	sector_t tmp, sectors_per_dev;
	struct raid_set *rs;
	struct raid_type *raid_type;

	/* Ensure minimum number of parameters */
	if (argc < MIN_PARMS)
		TI_ERR("Not enough parameters");

	/* Fetch # of dirty log parameters */
	if (sscanf(argv[1], "%d", &dl_parms) != 1 ||
	    !range_ok(dl_parms, 1, 4711))
		TI_ERR("Bad dirty log parameters number");

	/* Check raid_type */
	raid_type = get_raid_type(argv[dl_parms + 2]);
	if (!raid_type)
		TI_ERR("Bad raid type");

	/* In case of RAID4, parity drive is selectable */
	parity_parm = raid_type->level == raid4 ? 1 : 0;

	/* Handle variable number of RAID parameters */
	r = raid_variable_parms(ti, argv, dl_parms + parity_parm + 3,
				&raid_parms,
				&chunk_size, &chunk_size_parm,
				&stripes, &stripes_parm,
				&io_size, &io_size_parm,
				&recover_io_size, &recover_io_size_parm,
				&bandwidth, &bandwidth_parm);
	if (r)
		return r;

	/* # of raid devices */
	if (sscanf(argv[dl_parms + parity_parm + raid_parms + 4],
		   "%d", &raid_devs) != 1 ||
	    raid_devs < raid_type->minimal_devs)
		TI_ERR("Invalid number of raid devices");

	/* In case of RAID4, check parity drive index is in limits */
	if (raid_type->level == raid4) {
		/* Fetch index of parity device */
		if (sscanf(argv[dl_parms + 3], "%d", &pi) != 1 ||
		    !range_ok(pi, 0, raid_devs - 1))
			TI_ERR("Invalid RAID4 parity device index");
	}

	/*
	 * Index of device to initialize starts at 0
	 *
	 * o -1 -> don't initialize a particular device,
	 * o 0..raid_devs-1 -> initialize respective device
	 *   (used for reconstruction of a replaced device)
	 */
	if (sscanf(argv[dl_parms + parity_parm + raid_parms + 5],
		   "%d", &dev_to_init) != 1 ||
	    !range_ok(dev_to_init, -1, raid_devs - 1))
		TI_ERR("Invalid number for raid device to initialize");

	/* Check # of raid device arguments */
	if (argc - dl_parms - parity_parm - raid_parms - 6 != 2 * raid_devs)
		TI_ERR("Wrong number of raid device/offset arguments");

	/*
	 * Check that the table length is devisable
	 * w/o rest by (raid_devs - parity_devs)
	 */
	if (!multiple(ti->len, raid_devs - raid_type->parity_devs,
		      &sectors_per_dev))
		TI_ERR("Target length not divisable by number of data devices");

	/*
	 * Check that the device size is
	 * devisable w/o rest by chunk size
	 */
	if (!multiple(sectors_per_dev, chunk_size, &tmp))
		TI_ERR("Device length not divisable by chunk_size");

	/****************************************************************
	 * Now that we checked the constructor arguments ->
	 * let's allocate the RAID set
	 ****************************************************************/
	r = context_alloc(&rs, raid_type, stripes, chunk_size, io_size,
			  recover_io_size, raid_devs, sectors_per_dev,
			  ti, dl_parms, argv);
	if (r)
		return r;

	/*
	 * Set these here in order to avoid passing
	 * too many arguments to context_alloc()
	 */
	rs->set.dev_to_init = dev_to_init;
	rs->set.pi_parm = pi;
	rs->set.pi = (pi == -1) ? rs->set.data_devs : pi;
	rs->set.raid_parms = raid_parms;
	rs->set.chunk_size_parm = chunk_size_parm;
	rs->set.io_size_parm = io_size_parm;
	rs->sc.stripes_parm = stripes_parm;
	rs->recover.io_size_parm = recover_io_size_parm;
	rs->recover.bandwidth_parm = bandwidth_parm;
	recover_set_bandwidth(rs, bandwidth);

	/* Get the device/offset tupels */
	argv += dl_parms + 6 + parity_parm + raid_parms;
	r = dev_parms(ti, rs, argv, &i);
	if (r)
		goto err;

	/* Create work queue to handle this RAID set's io */
	r = rs_workqueue_init(rs);
	if (r)
		goto err;

	/*
	 * Make sure that dm core only hands maximum io size
	 * length down and pays attention to io boundaries
	 */
	ti->split_io = rs->set.io_size;
	ti->private = rs;

	/* Log information about RAID set */
	raid_set_log(rs);

	/* Enable overwrite optimization */
	set_bit(CHECK_OVERWRITE, &rs->flags);

	return 0;

   err:
	context_free(rs, ti, i);
	return r;
}

/*
 * Destruct a raid mapping
 */
static void raid_dtr(struct dm_target *ti)
{
	struct raid_set *rs = ti->private;

	/* Indicate recovery end */
	if (!rs->recover.end_jiffies)
		rs->recover.end_jiffies = jiffies;

	wake_do_raid(rs);	/* Wake daemon to free recovery stripe etc. */
	wait_ios(rs);		/* Wait for any io still being processed */
	destroy_workqueue(rs->io.wq);
	context_free(rs, ti, rs->set.raid_devs);
}

/* Queues ios to RAID sets */
static inline void queue_bio(struct raid_set *rs, struct bio *bio)
{
	int wake;
	struct bio_list *bl = &rs->io.in;

	spin_lock_irq(&rs->io.lock);
	wake = bio_list_empty(bl);
	bio_list_add(bl, bio);
	spin_unlock_irq(&rs->io.lock);

	if (wake)
		wake_do_raid(rs);
}

/* Raid mapping function */
static int raid_map(struct dm_target *ti, struct bio *bio,
		    union map_info *map_context)
{
	struct raid_set *rs = ti->private;

	/* I don't want to waste stripe cache capacity */
	if (bio_rw(bio) == READA)
		return -EIO;

	/*
	 * Get io reference to be waiting for to drop
	 * to zero on device suspension/destruction
	 */
	io_get(rs);
	down(&rs->io.max); /* Throttle io */

	/* FIXME: REMOVEME */
	atomic_inc(bio_data_dir(bio) == WRITE ? &rs->stats.ios_writes :
						&rs->stats.ios_reads);

	bio->bi_sector -= ti->begin; /* Remap sector */
	queue_bio(rs, bio);	/* Queue to the daemon */
	return 0;		/* Handle later */
}

/* Device suspend */
static void raid_postsuspend(struct dm_target *ti)
{
	struct raid_set *rs = ti->private;
	struct dirty_log *dl = rs->recover.dl;

	wake_do_raid(rs);
	if (raid_set_recovering(rs))
		rh_stop_recovery(rs->recover.rh); /* Wakes do_raid() */
	else
		wake_do_raid(rs);

	wait_ios(rs); /* Wait for completion of all ios being processed */
	if (dl->type->suspend && dl->type->suspend(dl)) /* Suspend dirty log */
		/* FIXME: need better error handling */
		DMWARN(PREFIX "log suspend failed");
}

/* Device resume */
static void raid_resume(struct dm_target *ti)
{
	struct raid_set *rs = ti->private;
	struct dirty_log *dl = rs->recover.dl;

	if (dl->type->resume && dl->type->resume(dl)) /* Resume dirty log */
		/* FIXME: need better error handling */
		DMWARN(PREFIX "log resume failed");

	rs->recover.nr_regions_to_recover =
		rs->recover.nr_regions - dl->type->get_sync_count(dl);

	/* Reset any unfinished recovery */
	if (raid_set_recovering(rs)) {
		recovery_reset(rs);
		rh_start_recovery(rs->recover.rh); /* Calls wake_do_raid() */
	} else
		wake_do_raid(rs);
}

static inline unsigned int sc_size(struct raid_set *rs)
{
	return to_sector(atomic_read(&rs->sc.stripes) *
			 (sizeof(struct stripe) +
			  (sizeof(struct stripe_set) +
			   (sizeof(struct page_list) +
			    to_bytes(rs->set.io_size) * rs->set.raid_devs)) +
			 (rs->recover.end_jiffies ? 0 : to_bytes(rs->set.raid_devs * rs->recover.io_size))));
}

/* REMOVEME: status output for development */
static void raid_devel_stats(struct dm_target *ti, char *result,
			     unsigned int *size, unsigned int maxlen)
{
	unsigned int sz = *size;
	unsigned long j;
	char buf[BDEVNAME_SIZE], *p;
	struct raid_set *rs = ti->private;
	struct recover *rec = &rs->recover;
	struct timeval tv;

	DMEMIT("%s r=%d/%d/%d w=%d/%d/%d ",
	       VERSION,
	       atomic_read(&rs->stats.ios_reads),
	       atomic_read(&rs->stats.reads_added),
	       atomic_read(&rs->stats.end_io_reads),
	       atomic_read(&rs->stats.ios_writes),
	       atomic_read(&rs->stats.writes_added),
	       atomic_read(&rs->stats.end_io_writes));
	DMEMIT("rc=%d wc=%d ",
	       atomic_read(&rs->stats.rc), atomic_read(&rs->stats.wc));
	DMEMIT("check_overwrite=%d ",
	       atomic_read(&rs->stats.check_overwrite));
	DMEMIT("wr_chunk_updated=%d wr_stripe_updated=%d ",
	       atomic_read(&rs->stats.wr_chunk_updated),
	       atomic_read(&rs->stats.wr_stripe_updated));
	DMEMIT("fast_reads=%d fast_active_reads=%d ",
	       atomic_read(&rs->stats.fast_reads),
	       atomic_read(&rs->stats.fast_active_reads));
	DMEMIT("h_1st=%d flushs=%d ",
	       atomic_read(&rs->stats.hits_1st),
	       atomic_read(&rs->stats.flushs));
	DMEMIT("active_stripes=%d ",
	       atomic_read(&rs->sc.active_stripes));
	DMEMIT("max_active_stripes=%d ",
	       atomic_read(&rs->sc.max_active_stripes));
	DMEMIT("inscache=%d ",
	       atomic_read(&rs->stats.inscache));
	DMEMIT("evict=%d ",
	       atomic_read(&rs->stats.evict));
	DMEMIT("io_inprocess=%d ",
	       atomic_read(&rs->io.in_process));
	DMEMIT("ios_post=%d ",
	       atomic_read(&rs->stats.ios_post));
	DMEMIT("eperms=%d ",
	       atomic_read(&rs->stats.eperms));
	DMEMIT("delayed_bios=%d sum_delayed_bios=%d ",
	       atomic_read(&rs->stats.delayed_bios),
	       atomic_read(&rs->stats.sum_delayed_bios));
	DMEMIT("degraded=%d ",
	       atomic_read(&rs->stats.degraded));
	DMEMIT("nosync=%d ",
	       atomic_read(&rs->stats.nosync));
	DMEMIT("xors=%d ",
	       atomic_read(&rs->stats.xors));
	DMEMIT("reconstruct_xor=%d ",
	       atomic_read(&rs->stats.reconstruct_xor));
	DMEMIT("reconstruct_ei=%d reconstruct_dev=%d ",
	       atomic_read(&rs->stats.reconstruct_ei),
	       atomic_read(&rs->stats.reconstruct_dev));
	DMEMIT("bio_copy_pl_next=%d bio_copy_redo=%d ",
	       atomic_read(&rs->stats.bio_copy_pl_next),
               atomic_read(&rs->stats.bio_copy_redo));
	DMEMIT("bandwidth=%d no_bandwidth=%d ",
	       atomic_read(&rs->stats.bandwidth),
	       atomic_read(&rs->stats.no_bandwidth));
	DMEMIT("ovr=%s ", test_bit(CHECK_OVERWRITE, &rs->flags) ? "on" : "off");
	DMEMIT("sc=%u/%u/%u/%u/%u ",
	       rs->set.chunk_size,
	       rs->set.io_size,
	       atomic_read(&rs->sc.stripes),
	       rs->sc.hash.buckets,
	       sc_size(rs));

	j = (rec->end_jiffies ? rec->end_jiffies : jiffies)
	    - rec->start_jiffies;
	jiffies_to_timeval(j, &tv);
	sprintf(buf, "%ld.%ld", tv.tv_sec, tv.tv_usec);
	p = strchr(buf, '.');
	p[3] = 0;

	DMEMIT("rg=%llu%s/%llu/%llu/%u %s ",
	       rec->nr_regions_recovered,
	       test_bit(REGION_GET, &rs->flags) ? "+" : "",
	       rec->nr_regions_to_recover,
	       rec->nr_regions, rec->bandwidth, buf);

	*size = sz;
}

static int raid_status(struct dm_target *ti, status_type_t type,
		       char *result, unsigned int maxlen)
{
	unsigned int i, sz = 0;
	char buf[BDEVNAME_SIZE];
	struct raid_set *rs = ti->private;

	switch (type) {
	case STATUSTYPE_INFO:
		/* REMOVEME: */
		if (test_bit(DEVEL_STATS, &rs->flags))
			raid_devel_stats(ti, result, &sz, maxlen);

		for (i = 0; i < rs->set.raid_devs; i++)
			DMEMIT("%s%s%s%s ",
			       format_dev_t(buf, rs->dev[i].dev->bdev->bd_dev),
			       dev_operational(rs, i) ? "o" : "F!",
			       rs->set.raid_type->level == raid4 &&
			       i == rs->set.pi ? "p" : "",
			       rs->set.dev_to_init == i ? "i" : "");

		break;

	case STATUSTYPE_TABLE:
		sz = rs->recover.dl->type->status(rs->recover.dl, type,
						  result, maxlen);
		DMEMIT("%s %u ", rs->set.raid_type->name, rs->set.raid_parms);

		if (rs->set.raid_type->level == raid4)
			DMEMIT("%d ", rs->set.pi_parm);

		if (rs->set.raid_parms)
			DMEMIT("%d ", rs->set.chunk_size_parm);

		if (rs->set.raid_parms > 1)
			DMEMIT("%d ", rs->sc.stripes_parm);

		if (rs->set.raid_parms > 2)
			DMEMIT("%d ", rs->set.io_size_parm);

		if (rs->set.raid_parms > 3)
			DMEMIT("%d ", rs->recover.io_size_parm);

		if (rs->set.raid_parms > 4)
			DMEMIT("%d ", rs->recover.bandwidth_parm);

		DMEMIT("%u %d ", rs->set.raid_devs, rs->set.dev_to_init);

		for (i = 0; i < rs->set.raid_devs; i++)
			DMEMIT("%s %llu ",
			       format_dev_t(buf, rs->dev[i].dev->bdev->bd_dev),
			       rs->dev[i].start);
	}

	return 0;
}

/****************************************************************************
 * Generic message parsing code
 */
/* FIXME: factor out to dm_message.h */

/* Message return bits */
enum dm_message_return {
	ret_ambiguous,	/* Action ambiguous */
	ret_inval,	/* Action invalid */
	ret_undef,	/* Action undefined */

	ret_option,	/* Option error */
	ret_arg,	/* Argument error */
	ret_argcount,	/* Argument count error */
};

/* Message argument type conversions */
enum dm_message_argument_type {
	basename_t,	/* Basename string */
	str_t,		/* String */
	d_t,		/* Signed int */
	u_t,		/* Unsigned int */
	u64_t,		/* Unsigned int 64 */
};

/* A message option */
struct dm_message_option {
	unsigned int num_options;
	char **options;
	unsigned long *actions;
};

/* Message arguments and types */
struct dm_message_argument {
	unsigned int num_args;
	unsigned long **ptr; /* FIXME */
	enum dm_message_argument_type *types;
};

/* Client message */
struct dm_msg {
	unsigned long action;	/* Identified action */
	unsigned long ret;	/* Return bits */
	unsigned int num_specs;	/* # of sepcifications */
	struct dm_msg_spec *cur;	/* current specification */
	struct dm_msg_spec *specs;	/* Specification list */
};

/* Secification of the message */
struct dm_msg_spec {
	const char *cmd;	/* Name of the command (i.e. 'bandwidth') */
	unsigned long action;
	unsigned int num_args;
	struct dm_message_option *options;
	struct dm_message_argument *args;
	unsigned long parm;
	int (*f)(void *context, struct dm_msg *msg); /* process for action */
};

/* Parse a message and its options and call a function */
int dm_message_parse(const char *caller, void *context,
		     int argc, char **argv, struct dm_msg *msg);


/*
 * FIXME: factor out to dm_message.c
 */
/* Get an argument depending on type */
static void message_arguments(int argc, char **argv, struct dm_msg *msg)
{

	if (argc) {
		int i;
		for (i = 0; i < msg->specs->num_args; i++) {
			int r;
			enum dm_message_argument_type type =
				msg->cur->args->types[i];

			switch (type) {
			case basename_t:
				*(char**) msg->cur->args->ptr[i] =
					basename(argv[i]);
				break;
		
			case str_t:
				*(char**) msg->cur->args->ptr[i] = argv[i];
				break;
	
			case u_t:
				r = sscanf(argv[i], "%u", (unsigned int*)
					   msg->cur->args->ptr[i]);
				goto check;
		
			case u64_t:
				r = sscanf(argv[i], "%llu",
					   (unsigned long long*)
					   msg->cur->args->ptr[i]);
				goto check;
		
			case d_t:
				r = sscanf(argv[i], "%d", (int*)
					   msg->cur->args->ptr[i]);
	   check:
				if (r != 1) {
					set_bit(ret_undef, &msg->ret);
					set_bit(ret_arg, &msg->ret);
					return;
				}
			}
		}
	}
}

/* Process function for a message */
static int message_process(const char *caller,
			   void *context, struct dm_msg *msg)
{
	int run = 1;
	const char *what = "", *prefix = "";

	if (test_bit(ret_ambiguous, &msg->ret))
		prefix = "ambiguous", run = 0;
	else if (test_bit(ret_undef, &msg->ret))
		prefix = "unrecognized", run = 0;

	if (run)
		return msg->cur->f(context, msg);

	if (test_bit(ret_inval, &msg->ret))
		prefix = "invalid";
	else if (test_bit(ret_arg, &msg->ret))
		what = "argument ";
	else if (test_bit(ret_option, &msg->ret))
		what = "option ";
	else if (test_bit(ret_argcount, &msg->ret))
		what = "argument count ";
	
	DMERR("%s%s message %sreceived", caller, prefix, what);
	return -EINVAL;
}

/* Check message arguments */
static void message_options_parse(int argc, char **argv, struct dm_msg *msg)
{
	int hit = 0;
	size_t l1 = strlen(*argv), l_hit = 0;
	char **option;
	unsigned long *action;
	struct dm_message_option *o;

	o = msg->cur->options;
	for (option = o->options, action = o->actions;
	     option < o->options + o->num_options;
	     option++, action++) {
		size_t l2 = strlen(*option);

		if (!strnicmp(*argv, *option, min(l1, l2))) {
			hit++;
			l_hit = l2;
			set_bit(*action, &msg->action);
		}
	}

	set_bit(ret_option, &msg->ret);

	if (!hit || l1 > l_hit)
		set_bit(ret_undef, &msg->ret);
	else if (hit > 1)
		set_bit(ret_ambiguous, &msg->ret);
	else {
		clear_bit(ret_option, &msg->ret);
		message_arguments(--argc, ++argv, msg);
	}
}

/* Parse a message action */
int dm_message_parse(const char *caller, void *context,
		     int argc, char **argv, struct dm_msg *msg)
{
	int hit = 0;
	size_t l1 = strlen(*argv), l_hit = 0;
	struct dm_msg_spec *s, *s_hit = NULL;

	if (argc < 2)
		return -EINVAL;

	for (s = msg->specs; s < msg->specs + msg->num_specs; s++) {
		size_t l2 = strlen(s->cmd);

		if (!strnicmp(*argv, s->cmd, min(l1, l2))) {
			hit++;
			l_hit = l2;
			s_hit = s;
			msg->action = s->action;
		}
	}

	if (!hit || l1 > l_hit) /* No hit or message string too long */
		set_bit(ret_undef, &msg->ret);
	else if (hit > 1) /* Ambiguous message */
		set_bit(ret_ambiguous, &msg->ret);
	else if (argc - 2 != s_hit->num_args) {
		set_bit(ret_undef, &msg->ret);
		set_bit(ret_argcount, &msg->ret);
	}

	if (msg->ret)
		return -EINVAL;

	msg->cur = s_hit;
	message_options_parse(--argc, ++argv, msg);
	return msg->ret ? -EINVAL : message_process(caller, context, msg);
}
/*
 * END Generic message parsing code
 ****************************************************************************/


/*
 * Message interface
 */
enum actions {
	act_bw,			/* Recovery bandwidth switch */
	act_dev,		/* Device failure switch */
	act_overwrite,		/* Stripe overwrite check */
	act_stats,		/* Development statistics switch */
	act_sc,			/* Stripe cache switch */

	act_on,			/* Set entity on */
	act_off,		/* Set entity off */
	act_reset,		/* Reset entity */

	act_set = act_on,	/* Set # absolute */
	act_grow = act_off,	/* Grow # by an amount */
	act_shrink  = act_reset,/* Shrink # by an amount */
};

/* Turn a delta to absolute */
static int _absolute(unsigned long action, int act, int r)
{
	/* Make delta absolute */
	if (test_bit(act_set, &action));
	else if (test_bit(act_grow, &action))
		r += act;
	else if (test_bit(act_shrink, &action))
		r = act - r;
	else
		r = -EINVAL;

	return r;
}

/* Change recovery io bandwidth */
static int bandwidth_change(void *context, struct dm_msg *msg)
{
	struct raid_set *rs = context;
	int act = rs->recover.bandwidth;
	int bandwidth = (int) *msg->cur->args->ptr[0];

	if (range_ok(bandwidth, BANDWIDTH_MIN, BANDWIDTH_MAX)) {
		/* Make delta bandwidth absolute */
		bandwidth = _absolute(msg->action, act, bandwidth);

		/* Check range */
		if (range_ok(bandwidth, BANDWIDTH_MIN, BANDWIDTH_MAX)) {
			recover_set_bandwidth(rs, bandwidth);
			return 0;
		}

		set_bit(ret_arg, &msg->ret);
		set_bit(ret_inval, &msg->ret);
	}

	return -EINVAL;
}

/* Change state of a device (running/offline) */
/* FIXME: this only works while recovering! */
static int device_state(void *context, struct dm_msg *msg)
{
	int r;
	const char *str = "is already ";
	union dev_lookup dl = { .dev_name = (char*) *msg->cur->args->ptr[0] };
	struct raid_set *rs = context;

	r = raid_dev_lookup(rs, strchr(dl.dev_name, ':') ?
				bymajmin : byname, &dl);
	if (r == -ENODEV) {
		DMERR(PREFIX "device %s is no member of this set", dl.dev_name);
		return r;
	}

	if (test_bit(act_off, &msg->action)){
 		if (dev_operational(rs, r))
			str = "";
	} else if (!dev_operational(rs, r))
			str = "";

	DMINFO(PREFIX "/dev/%s %s%s",
	       dl.dev_name, str,
	       test_bit(act_off, &msg->action) ? "offline" : "running");

	if (test_bit(act_off, &msg->action)) {
		if (raid_set_recovering(rs)) {
			struct stripe *stripe = rs->recover.stripe;

			/*
			 * Error stripe and wake the daemon to
			 * change the rest of the state by running
			 * the error path in do_recovery()
			 */
			stripe_error(stripe, stripe->obj[r].pl->page);
			wake_do_raid(rs);
		}

		return 0;
	} else
		return raid_set_check_and_upgrade(rs, r);
}

/* Set/reset development feature flags */
static int devel_flags(void *context, struct dm_msg *msg)
{
	struct raid_set *rs = context;

	if (test_bit(act_on, &msg->action))
		return test_and_set_bit(msg->cur->parm, &rs->flags) ?
		       -EPERM : 0;
	else if (test_bit(act_off, &msg->action))
		return test_and_clear_bit(msg->cur->parm, &rs->flags) ?
		       0 : -EPERM;
	else if (test_bit(act_reset, &msg->action)) {
		stats_reset(rs);
		return 0;
	}

	return -EINVAL;
}

/* Resize the stripe cache */
static int stripecache_resize(void *context, struct dm_msg *msg)
{
	int act, stripes;
	struct raid_set *rs = context;

	/* Deny permission in case the daemon is still shrinking! */
	if (atomic_read(&rs->sc.stripes_to_shrink))
		return -EPERM;

	stripes = (int) *msg->cur->args->ptr[0];
	if (stripes > 0) {
		act = atomic_read(&rs->sc.stripes);

		/* Make delta stripes absolute */
		stripes = _absolute(msg->action, act, stripes);

		/* Check range and that the # of stripes changes */
		if (range_ok(stripes, STRIPES_MIN, STRIPES_MAX)) {
			if (stripes > act)
				return sc_grow(&rs->sc, stripes - act, grow);
			else if (stripes < act) {
				atomic_set(&rs->sc.stripes_to_shrink,
					   act - stripes);
				wake_do_raid(rs);
			}

			return 0;
		}
	}

	set_bit(ret_arg, &msg->ret);
	set_bit(ret_inval, &msg->ret);
	return -EINVAL;
}

/* Parse the RAID message action */
static int raid_message_parse(struct raid_set *rs, int argc, char **argv, 
			      struct dm_msg *msg)
{
	/* Variables and types for the parameters */
	static int i;
	static char *p;
	static unsigned long *arg[] = { (unsigned long*) &i };
	static unsigned long *p_arg[] = { (unsigned long*) &p };
	static enum dm_message_argument_type type[] = { d_t };
	static enum dm_message_argument_type dev_type[] = { basename_t };

	/* Declare all message options */
	static char *str_sgs[] = { "set", "grow", "shrink" };
	static char *str_dev[] = { "running", "offline" };
	static char *str_oor[] = { "on", "off", "reset" };

	/* Declare all actions */
	static unsigned long act_sgs[] = { act_set, act_grow, act_shrink };
	static unsigned long act_oor[] = { act_on, act_off, act_reset };

	static struct dm_message_option bw_opt = { 3, str_sgs, act_sgs };
	static struct dm_message_argument bw_args = { 1, arg, type };
	static struct dm_message_option dev_opt = { 2, str_dev, act_oor };
	static struct dm_message_argument dev_args = { 1, p_arg, dev_type};
	static struct dm_message_option ovr_opt = { 2, str_oor, act_oor };
	static struct dm_message_argument null_args = { 0, NULL, NULL };
	static struct dm_message_option stats_opt = { 3, str_oor, act_oor };
	static struct dm_message_option stripe_opt = { 3, str_sgs, act_sgs };

	/* Declare our expected messages */
	static struct dm_msg_spec specs[] = {
		{ "bandwidth", act_bw, 1, &bw_opt, &bw_args,
		  0, bandwidth_change },
		{ "device", act_dev, 1, &dev_opt, &dev_args,
		  0, device_state },
		{ "overwrite", act_overwrite, 0, &ovr_opt, &null_args,
		  CHECK_OVERWRITE, devel_flags },
		{ "statistics", act_stats, 0, &stats_opt, &null_args,
		  DEVEL_STATS, devel_flags },
		{ "stripecache", act_sc, 1, &stripe_opt, &bw_args,
		  0, stripecache_resize },
	};

	msg->specs = specs;
	msg->num_specs = ARRAY_SIZE(specs);
	return dm_message_parse(PREFIX, rs, argc, argv, msg);
}

/*
 * send message:
 *
 * 'ba[ndwidth] {se[t],g[row],sh[rink]} #'	# i.e. 'ba se 50'
 * 'de{vice] o[ffline]/r[unning] DevName/maj:min' # i.e. 'device o /dev/sda'
 * 'o[verwrite] {on,of[f]}'			# i.e. 'over off'
 * 'sta[tistics] {on,of[f],r[eset]}'		# i.e. 'stat of'
 * 'str[ipecache] {se[t],g[row],sh[rink]} #'	# i.e. 'stripe set 1024'
 *
 */
static int raid_message(struct dm_target *ti, unsigned argc, char **argv)
{
	struct dm_msg msg;
	
	memset(&msg, 0, sizeof(msg));
	return raid_message_parse(ti->private, argc, argv, &msg);
}
/*
 * END message interface
 */

static struct target_type raid_target = {
	.name		= "raid45",
	.version	= {1, 0, 0},
	.module		= THIS_MODULE,
	.ctr		= raid_ctr,
	.dtr		= raid_dtr,
	.map		= raid_map,
	.postsuspend	= raid_postsuspend,
	.resume		= raid_resume,
	.status		= raid_status,
	.message	= raid_message,
};

static int __init dm_raid_init(void)
{
	int r;

	r = dm_register_target(&raid_target);
	if (r) {
		DMERR(PREFIX "Failed to register target [%d]", r);
		return r;
	}

	DMINFO(PREFIX "initialized %s", VERSION);
	return 0;
}

static void __exit dm_raid_exit(void)
{
	int r;

	r = dm_unregister_target(&raid_target);
	if (r)
		DMERR(PREFIX "unregister failed %d", r);
	else
		DMINFO(PREFIX "exit %s", VERSION);
}

/* Module hooks */
module_init(dm_raid_init);
module_exit(dm_raid_exit);

MODULE_DESCRIPTION(DM_NAME " raid4/5 target");
MODULE_AUTHOR("Heinz Mauelshagen <mauelshagen@redhat.com>");
MODULE_LICENSE("GPL");

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
