/*
 * Philips saa7145 controller driver, for the
 * Fast Multimedia AV Master.
 *
 * This part handles card-specific data and detection
 * 
 * Copyright (C) 2000 Serguei Miridonov <mirsev@cicese.mx>
 *
 * Currently maintained by:
 *   Luca Bonissi <lucabon@interfree.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/vmalloc.h>

#include <linux/proc_fs.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <linux/videodev.h>
#include <media/v4l2-common.h>

#include <linux/spinlock.h>
#include <linux/sem.h>
#include <linux/kmod.h>

#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/video_decoder.h>
#include <linux/video_encoder.h>

#include <asm/io.h>

#include "videocodec.h"
#include "avmaster.h"
#include "avmaster_card.h"
#include "avmaster_device.h"
#include "avmaster_procfs.h"

/* temp hack */
#define PCI_DEVICE_ID_PHILIPS_SAA7145 0x7145
#ifndef I2C_DRIVERID_SAA7187
#warning Using temporary hack for missing I2C driver-ID for saa7187
#define I2C_DRIVERID_SAA7187 0xF0
#endif
#ifndef I2C_HW_B_SAA7145
#warning Using temporary hack for missing I2C adapter-ID for saa7145
#define I2C_HW_B_SAA7145 0x20
#endif
/* /temp hack */

extern const struct avmaster_format avmaster_formats[];
extern const int avmaster_num_formats;

static int card[AVM_MAX] = { -1, -1, -1, -1 };
MODULE_PARM(card, "1-" __stringify(AVM_MAX) "i");
MODULE_PARM_DESC(card, "The type of card");

static int encoder[AVM_MAX] = { -1, -1, -1, -1 };
MODULE_PARM(encoder, "1-" __stringify(AVM_MAX) "i");
MODULE_PARM_DESC(encoder, "i2c TV encoder");

static int decoder[AVM_MAX] = { -1, -1, -1, -1 };
MODULE_PARM(decoder, "1-" __stringify(AVM_MAX) "i");
MODULE_PARM_DESC(decoder, "i2c TV decoder");

/*
   The video mem address of the video card.
   The driver has a little database for some videocards
   to determine it from there. If your video card is not in there
   you have either to give it to the driver as a parameter
   or set in in a VIDIOCSFBUF ioctl
 */

static unsigned long vidmem = 0;	/* Video memory base address */
MODULE_PARM(vidmem, "i");

/*
   Default input and video norm at startup of the driver.
*/

static int default_input = 0;	/* 0=Composite, 1=S-Video */
MODULE_PARM(default_input, "i");
MODULE_PARM_DESC(default_input,
		 "Default input (0=Composite, 1=S-Video, 2=Internal)");

static int default_norm = 0;	/* 0=PAL, 1=NTSC 2=SECAM */
MODULE_PARM(default_norm, "i");
MODULE_PARM_DESC(default_norm, "Default norm (0=PAL, 1=NTSC, 2=SECAM)");

static int video_nr = -1;	/* /dev/videoN, -1 for autodetect */
MODULE_PARM(video_nr, "i");
MODULE_PARM_DESC(video_nr, "video device number");

/*
   Number and size of grab buffers for Video 4 Linux
   The vast majority of applications should not need more than 2,
   the very popular BTTV driver actually does ONLY have 2.
   Time sensitive applications might need more, the maximum
   is VIDEO_MAX_FRAME (defined in <linux/videodev.h>).

   The size is set so that the maximum possible request
   can be satisfied. Decrease  it, if bigphys_area alloc'd
   memory is low. If you don't have the bigphys_area patch,
   set it to 128 KB. Will you allow only to grab small
   images with V4L, but that's better than nothing.

   v4l_bufsize has to be given in KB !

*/

int v4l_nbufs = 2;
int v4l_bufsize = 128;		/* Everybody should be able to work with this setting */
MODULE_PARM(v4l_nbufs, "i");
MODULE_PARM_DESC(v4l_nbufs, "Maximum number of V4L buffers to use");
MODULE_PARM(v4l_bufsize, "i");
MODULE_PARM_DESC(v4l_bufsize, "Maximum size per V4L buffer (in kB)");

int jpg_nbufs = 32;
int jpg_bufsize = 512;		/* max size for 100% quality full-PAL frame */
MODULE_PARM(jpg_nbufs, "i");
MODULE_PARM_DESC(jpg_nbufs, "Maximum number of JPG buffers to use");
MODULE_PARM(jpg_bufsize, "i");
MODULE_PARM_DESC(jpg_bufsize, "Maximum size per JPG buffer (in kB)");

int pass_through = 0;		/* 1=Pass through TV signal when device is not used */
				/* 0=Show color bar when device is not used (LML33: only if lml33dpath=1) */
MODULE_PARM(pass_through, "i");
MODULE_PARM_DESC(pass_through,
		 "Pass TV signal through to TV-out when idling");

int debug = 1;
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug level (0-4)");

MODULE_DESCRIPTION("Fast AV-Master JPEG codec driver");
MODULE_AUTHOR("Serguei Miridonov/Michael Stickel/Luca Bonissi");
MODULE_LICENSE("GPL");

static struct pci_device_id saa7145_pci_tbl[] = {
	{PCI_VENDOR_ID_PHILIPS, PCI_DEVICE_ID_PHILIPS_SAA7145,
	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{0}
};
MODULE_DEVICE_TABLE(pci, saa7145_pci_tbl);

#define dprintk(num, format, args...) \
	do { \
		if (debug >= num) \
			printk(format, ##args); \
	} while (0)

int avmaster_num;			/* number of Buzs in use */
struct avmaster avmaster[AVM_MAX];

/* helper function */
void my_wait(long ms)
{
        unsigned long timeout = jiffies + ((ms+9)/10);
        while(time_before(jiffies, timeout)) {
                schedule();
        }
}


#if 0
/* videocodec bus functions ZR36050 */
static u32
zr36050_read (struct videocodec *codec,
	      u16                reg)
{
	struct avmaster *avms = (struct avmaster *) codec->master_data->data;
	__u32 data;

	if (post_office_wait(avms)
	    || post_office_write(avms, 1, 0, reg >> 2)) {	// reg. HIGHBYTES
		return -1;
	}

	data = post_office_read(avms, 0, reg & 0x03) & 0xff;	// reg. LOWBYTES + read
	return data;
}

static void
zr36050_write (struct videocodec *codec,
	       u16                reg,
	       u32                val)
{
	struct avmaster *avms = (struct avmaster *) codec->master_data->data;

	if (post_office_wait(avms)
	    || post_office_write(avms, 1, 0, reg >> 2)) {	// reg. HIGHBYTES
		return;
	}

	post_office_write(avms, 0, reg & 0x03, val & 0xff);	// reg. LOWBYTES + wr. data
}

/* videocodec bus functions ZR36015 */
static u32
zr36015_read (struct videocodec *codec,
	      u16                reg)
{
	struct avmaster *avms = (struct avmaster *) codec->master_data->data;
	__u32 data;

	if (post_office_wait(avms)) {
		return -1;
	}

	data = post_office_read(avms, 2, reg & 0x03) & 0xff;	// read
	return data;
}
#endif

/* hack for in avmaster_device.c */
void
zr36015_write (struct videocodec *codec,
	       u16                reg,
	       u32                val)
{
	struct avmaster *avms = (struct avmaster *) codec->master_data->data;

	if (post_office_wait(avms)) {
		return;
	}

	post_office_write(avms, 2, reg & 0x03, val & 0x0ff);	// wr. data
}

/*
 * Board specific information
 */

static void
avmaster_init (struct avmaster *avms)
{
	dprintk(3, KERN_DEBUG "%s: avmaster_init()\n", avms->name);
}

static char *
i2cid_to_modulename (u16 i2c_id)
{
	char *name = NULL;

	switch (i2c_id) {
	case I2C_DRIVERID_SAA7110:
		name = "saa7110";
		break;
	case I2C_DRIVERID_SAA7187:
		name = "saa7187";
		break;
	default:
		break;
	}

	return name;
}

static char *
codecid_to_modulename (u16 codecid)
{
	char *name = NULL;

	switch (codecid) {
	case CODEC_TYPE_ZR36050:
		name = "zr36050";
		break;
	case CODEC_TYPE_ZR36015:
		name = "zr36015";
		break;
	default:
		break;
	}

	return name;
}

// struct tvnorm {
//      u16 Wt, Wa, HStart, HSyncStart, Ht, Ha, VStart;
// };

//static struct tvnorm f50sqpixel = { 944, 768, 174, 880, 625, 576, 17 };
//static struct tvnorm f60sqpixel = { 780, 640, 136, 716, 525, 480, 12 };
static struct tvnorm f50sqpixel = { 944, 768, 174, 880, 625, 576, 17 };
static struct tvnorm f60sqpixel = { 780, 640, 136, 716, 525, 480, 12 };

static struct card_info avmaster_cards[NUM_CARDS] __devinitdata = {
	{
		.type = AVMaster,
		.name = "AVMaster",
		.i2c_decoder = I2C_DRIVERID_SAA7110,
		.i2c_encoder = I2C_DRIVERID_SAA7187,
		.video_codec = CODEC_TYPE_ZR36050,
		.video_vfe = CODEC_TYPE_ZR36015,

		.inputs = 2,
		.input = {
				{ 1, "Composite" },
				{ 9, "S-Video" }
			},
		.norms = 3,
		.tvn = {
				&f50sqpixel,
				&f60sqpixel,
				&f50sqpixel},
		.vsync_int = ISRC_VBI,
		.gpio = { 2, 1, -1, 3, 7, 0, 4, 5 },
		.vfe_pol = { 0, 0, 0, 0, 0, 0, 0, 0 },
		.gpio_pol = { 0, 0, 0, 1, 0, 0, 0, 0 },
		.gpcs = { -1, 0 },
		.gws_not_connected = 0,
		.init = &avmaster_init,
	}
};

/*
 * I2C functions
 */
/* this function runs through the i2c-messages and prepares the data to be
   sent through the saa7145. have a look at the specifications p. 122 ff 
   to understand this. it returns the number of u32s to send, or -1
   in case of an error. */
int saa7145_i2c_msg_prepare(struct i2c_msg m[], int num, u32 *op)
{
	int h1, h2;
	int i, j, addr;
	int mem = 0, op_count = 0;

	/* first determine size of needed memory */
	for(i = 0; i < num; i++) {
		mem += m[i].len + 1;
	}

	/* worst case: we need one u32 for three bytes to be send
	   plus one extra byte to address the device */
	mem = 1 + ((mem-1) / 3);

	/* we assume that op points to a memory of at least SAA7145_I2C_MEM bytes
	   size. if we exceed this limit... */
	if ( (4*mem) > SAA7145_I2C_MEM ) {
		dprintk(1,KERN_ERR "saa7145: cannot prepare i2c-message..\n");
		return -1;
	}

	/* be careful: clear out the i2c-mem first */
	memset(op,0,sizeof(u32)*mem);

	/* loop through all messages */
	for(i = 0; i < num; i++) {

		/* insert the address of the i2c-slave.
		   note: we get 7-bit-i2c-addresses, so we have to perform a translation */
		addr = (m[i].addr*2) + ( (0 != (m[i].flags & I2C_M_RD)) ? 1 : 0); 
		h1 = op_count/3; h2 = op_count%3;
		op[h1] |= ( (u8)addr << ((3-h2)*8));
		op[h1] |= (IIC_START << ((3-h2)*2));
		op_count++;

		/* loop through all bytes of message i */
		for(j = 0; j < m[i].len; j++) {
		
			/* insert the data bytes */
			h1 = op_count/3; h2 = op_count%3;
			op[h1] |= ( (u32)((u8)m[i].buf[j]) << ((3-h2)*8));
			op[h1] |= (       	  IIC_CONT << ((3-h2)*2));
			op_count++;
		}
		
	}			

	/* have a look at the last byte inserted:
	  if it was: ...CONT change it to ...STOP */
	h1 = (op_count-1)/3; h2 = (op_count-1)%3;
	if ( IIC_CONT == (0x3 & (op[h1] >> ((3-h2)*2))) ) {
		//dprintk(1,KERN_ERR "stop inserted!\n");
		op[h1] &= ~(0x2 << ((3-h2)*2));
		op[h1] |= (IIC_STOP << ((3-h2)*2));
	} else {
//		printk("nothin!\n");
	}

	/* return the number of u32s to send */	
	dprintk(4,KERN_DEBUG "saa7145_i2c_msg_prepare: len = %d, mem = %d, value0=%08x\n",op_count,mem,op[0]);
	return mem;
}

/* this functions loops through all i2c-messages. normally, it should determine
   which bytes were read through the adapter and write them back to the corresponding
   i2c-message. but instead, we simply write back all bytes.
   
   fixme: this could be improved. */
int saa7145_i2c_msg_cleanup(struct i2c_msg m[], int num, u32 *op)
{
	int i, j;
	int op_count = 0;

	/* loop through all messages */
	for(i = 0; i < num; i++) {

		op_count++;

		/* loop throgh all bytes of message i */
		for(j = 0; j < m[i].len; j++) {
			/* write back all bytes that could have been read */
			m[i].buf[j] = (op[op_count/3] >> ((3-(op_count%3))*8));
			op_count++;
		}
	}
	
	return 0;
}

/* this functions returns the status-register of our i2c-device */
u32 saa7145_i2c_status(struct avmaster* saa) 
{
	u32 iicsta = saa7145_read(saa, IICSC);
	dprintk(5,KERN_DEBUG "saa7145: saa7145_i2c_status: 0x%08x\n",iicsta);
	return iicsta;
}

/* this functions resets the i2c-device and returns 0 if everything was fine, otherwise -1 */
int saa7145_i2c_reset(struct avmaster* saa) 
{
	/* get current status */
	u32 status = saa7145_i2c_status(saa);
	
	/* clear registers for sure */
	saa7145_write(saa, IICSC, saa->i2c_bitrate);

	/* check if any operation is still in progress */
	if ( 0 != ( status & IICSC_BUSY) ) {

		/* yes, kill ongoing operation */
		dprintk(2,KERN_INFO "%s: i2c_reset: busy_state detected.\n",saa->name);

		/* clear all error-bits pending */
		saa7145_write(saa, IICSC, saa->i2c_bitrate);
		my_wait(SAA7145_I2C_DELAY);

 	}

	/* check if any error is (still) present. (this can be necessary because p.123, note 1) */
	status = saa7145_i2c_status(saa);

	if ( saa->i2c_bitrate != status ) {

		dprintk(1,KERN_INFO "%s: i2c_reset: error_state detected. (0x%08x)\n",saa->name,status);

		/* Repeat the abort operation. This seems to be necessary
		   after serious protocol errors caused by e.g. the SAA7740 */
		//saa7145_write(saa, I2C_STATUS, (saa->i2c_bitrate | MASK_07));
		//saa7145_write(saa, MC2, (MASK_00 | MASK_16));
		//my_wait(saa,SAA7146_I2C_DELAY);

		/* clear all error-bits pending */
		//saa7145_write(saa, I2C_STATUS, saa->i2c_bitrate);
		//saa7145_write(saa, MC2, (MASK_00 | MASK_16));
		//my_wait(saa,SAA7146_I2C_DELAY);

		/* the data sheet says it might be necessary to clear the status
		   twice after an abort */
		//saa7145_write(saa, I2C_STATUS, saa->i2c_bitrate);
		//saa7145_write(saa, MC2, (MASK_00 | MASK_16));
		my_wait(SAA7145_I2C_DELAY);
     	}

	/* if any error is still present, a fatal error has occured ... */
	status = saa7145_i2c_status(saa);
	if ( saa->i2c_bitrate != status ) {
		dprintk(1,KERN_ERR "%s: i2c_reset: i2c_reset: fatal error, status:0x%08x\n",saa->name,status);
		return -1;
	}

	return 0;
}

/* this functions writes out the data-byte 'dword' to the i2c-device.
   it returns 0 if ok, -1 if the transfer failed, -2 if the transfer
   failed badly (e.g. address error) */
int saa7145_i2c_writeout(struct avmaster* saa, u32* dword)
{
	int i = 0;
	u32 status = 0;

	dprintk(5,KERN_DEBUG "%s: writeout: 0x%08x (before).\n",saa->name,*dword);
	
	/* write out i2c-command */
	//saa7145_write(saa, I2C_STATUS,	 saa->i2c_bitrate);
	saa7145_write(saa, IICTFR, *dword);

	/* wait until busy flag becomes inactive or we time out */
	for (i = 5; i > 0; i--) {
		status = saa7145_i2c_status(saa);
		/* check busy flag */
		if ( 0 == (status & IICSC_BUSY))
			break;
		/* check error flag */
		if ( 0 != (status & IICSC_ERR)) {
			break;
		}
		my_wait(SAA7145_I2C_DELAY);
	}
	if (0 == i) {
		dprintk(1,KERN_ERR "%s: saa7145_i2c_writeout: timeout error. #2\n",saa->name);
		return -1;
	}

	/* give a detailed status report */
	if ( 0 != (status & IICSC_ERR)) {

		if( 0 != (status & IICSC_DTERR) ) {
			dprintk(1,KERN_ERR "%s: saa7145_i2c_writeout: error in data transmission.\n",saa->name);
		}
		/* we handle address-errors here */
		if( 0 != (status & IICSC_APERR) ) {
			dprintk(1,KERN_ERR "%s: saa7145_i2c_writeout: error in address phase 0x%08x.\n",saa->name,*dword);
			return -2;
		}
		if( 0 == (status & (IICSC_DTERR | IICSC_APERR)) ) {
			dprintk(1,KERN_ERR "%s: saa7145_i2c_writeout: unknown error: 0x%08x.\n",saa->name,status);
		}
		return -1;
	}

	/* read back data, just in case we were reading ... */
	*dword = saa7145_read(saa, IICTFR);

	dprintk(5,KERN_DEBUG "%s: saa7145_i2c_writeout: 0x%08x (after).\n",saa->name,*dword);
	
//	saa7145_i2c_status(saa);
//	printk("saa7145_i2c_writeout() exit.\n");

	return 0;
}

int saa7145_i2c_transfer(struct avmaster* saa, struct i2c_msg msgs[], int num,  int retries)
{
	int i = 0, result = 0, count = 0;
	u32* buffer = saa->i2c_mem;

/*
	if(saa->dummy == 1) {
		int i = 0, j = 0;
		printk("num msgs: %d\n",num);
		for(i = 0; i < num; i++) {
		printk("%d: addr: 0x%02x, len: %d: ==> ",i,msgs[i].addr,msgs[i].len);
		for(j = 0; j<msgs[i].len;j++) {
			printk("0x%02x ",(u8)msgs[i].buf[j]);	
		}
		printk("\n");
		return -1;
		}
	}
*/
	/* prepare the message(s), get number of u32s to transfer */
	count = saa7145_i2c_msg_prepare(msgs, num, buffer);
	if ( 0 > count ) {
		dprintk(1,KERN_ERR "%s: saa7145_xfer: could not prepare i2c-message.\n",saa->name);
		return -1;
	}
	
	/* loop through number of retries ... */
	for(; retries >= 0; retries--) {
	
		/* reset the i2c-device if necessary */
		result = saa7145_i2c_reset(saa);
		if ( 0 > result ) {
			dprintk(1,KERN_ERR "%s: saa7145_i2c_transfer: could not reset i2c-device.\n",saa->name);
			return -1;
		}	

		/* write out the u32s one after another */
		for(i = 0; i < count; i++) {

//			printk("0x%08x\n",saa7145_read(saa, 0xfc ));
	
			result = saa7145_i2c_writeout( saa, &buffer[i] );
			if ( 0 != result) {
				/* if address-error occured, dont retry */
				if ( -2 == result ) {
					return -1;
				}
				dprintk(1,KERN_INFO "%s: saa7145_i2c_transfer: error while sending message(s). starting again.\n",saa->name);
				break;
			}
		}
		
		/* see if an error occured & the last retry failed */
		if( (0 != result) && (0 == retries) ) {
			dprintk(1,KERN_INFO "%s: saa7145_xfer: could not transfer i2c-message(s).\n",saa->name);
			return -1;
		}
		
		if( 0 == result )
			break;
	}
	
	/* if any things had to be read, get the results */
	if ( 0 != saa7145_i2c_msg_cleanup(msgs, num, buffer)) {
		dprintk(1,KERN_ERR "%s: saa7145_xfer: could not cleanup i2c-message.\n",saa->name);
		return -1;
	}

//	saa7145_i2c_status(saa);
//	printk("saa7145_i2c_transfer() exit.\n");
	return 0;
}

/* utility functions */
int saa7145_i2c_xfer(struct i2c_adapter* adapter, struct i2c_msg msg[], int num)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	struct avmaster *avms = (struct avmaster *) (adapter->data);
#else
	struct avmaster *avms = (struct avmaster *) i2c_get_adapdata(adapter);
#endif
					
	dprintk(4,KERN_DEBUG "%s: saa7145_i2c_xfer for adapter %s called.\n", avms->name, adapter->name);

	/* use helper function to transfer data */
	if( 0 != saa7145_i2c_transfer(avms, msg, num, adapter->retries)) {
		return -1;
	}

	/* return the number of delivered messages */
	return num;	
}

int saa7145_i2c_ioctl(struct i2c_adapter *adapter, unsigned int cmd, unsigned long arg)
{
	dprintk(5,KERN_DEBUG "saa7145_i2c_ioctl for adapter %s called.\n", adapter->name);
	return 0;
}

u32 saa7145_i2c_func(struct i2c_adapter *adapter)
{
	dprintk(5,KERN_DEBUG "saa7145_i2c_func for adapter %s called.\n", adapter->name);

	return	  I2C_FUNC_I2C
		| I2C_FUNC_SMBUS_QUICK
		| I2C_FUNC_SMBUS_READ_BYTE	| I2C_FUNC_SMBUS_WRITE_BYTE
		| I2C_FUNC_SMBUS_READ_BYTE_DATA | I2C_FUNC_SMBUS_WRITE_BYTE_DATA;
}


/* exported algorithm data */
struct i2c_algorithm saa7145_algo = {
	saa7145_i2c_xfer,
	NULL,
	NULL,					/* slave_xmit		*/
	NULL,					/* slave_recv		*/
	saa7145_i2c_ioctl,			/* ioctl		*/
	saa7145_i2c_func,			/* functionality	*/
};



static int
avmaster_i2c_client_register (struct i2c_client *client)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	struct avmaster *avms = (struct avmaster *) (client->adapter->data);
#else
	struct avmaster *avms = (struct avmaster *) i2c_get_adapdata(client->adapter);
#endif
	int res = 0;

	dprintk(2,
		KERN_DEBUG "%s: i2c_client_register() - driver id = %d\n",
		avms->name, client->driver->id);

	down(&avms->resource_lock);

	if (avms->user > 0) {
		res = -EBUSY;
		goto clientreg_unlock_and_return;
	}

	if (client->driver->id == avms->card.i2c_decoder)
		avms->decoder = client;
	else if (client->driver->id == avms->card.i2c_encoder)
		avms->encoder = client;
	else
		return -ENODEV;

clientreg_unlock_and_return:
	up(&avms->resource_lock);

	return res;
}

static int
avmaster_i2c_client_unregister (struct i2c_client *client)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	struct avmaster *avms = (struct avmaster *) (client->adapter->data);
#else
	struct avmaster *avms = (struct avmaster *) i2c_get_adapdata(client->adapter);
#endif
	int res = 0;

	dprintk(2, KERN_DEBUG "%s: i2c_client_unregister()\n", avms->name);

	down(&avms->resource_lock);

	if (avms->user > 0) {
		res = -EBUSY;
		goto clientunreg_unlock_and_return;
	}

	/* try to locate it */
	if (client == avms->encoder) {
		avms->encoder = NULL;
	} else if (client == avms->decoder) {
		avms->decoder = NULL;
		snprintf(avms->name, sizeof(avms->name), "MJPEG[%d]", avms->id);
	}
clientunreg_unlock_and_return:
	up(&avms->resource_lock);
	return res;
}

static int
avmaster_register_i2c (struct avmaster *avms)
{
	dprintk(3, KERN_DEBUG "%s: avmaster_register_i2c()\n", avms->name);
	strcpy(avms->i2c_adapter.name, avms->name);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	avms->i2c_adapter.data = avms;
#else
	i2c_set_adapdata(&avms->i2c_adapter,avms);
#endif
	avms->i2c_adapter.algo = &saa7145_algo;
	avms->i2c_adapter.algo_data = NULL;
	avms->i2c_adapter.id = I2C_HW_B_SAA7145;
	avms->i2c_adapter.client_register = avmaster_i2c_client_register;
	avms->i2c_adapter.client_unregister = avmaster_i2c_client_unregister;
	avms->i2c_adapter.timeout = SAA7145_I2C_TIMEOUT;
	avms->i2c_adapter.retries = SAA7145_I2C_RETRIES;
	avms->i2c_mem = (u32*)kmalloc(SAA7145_I2C_MEM, GFP_KERNEL);
	if( !avms->i2c_mem ) {
		dprintk(1, KERN_ERR "%s: avmaster_register_i2c() - kmalloc (i2c_mem) failed\n",
			avms->name);
		return -ENOMEM;
	}
	avms->i2c_bitrate=IICCC_DIV480;
	memset(avms->i2c_mem, 0x00, SAA7145_I2C_MEM);
	i2c_add_adapter(&avms->i2c_adapter);
	return 0;
}

static void
avmaster_unregister_i2c (struct avmaster *avms)
{
	dprintk(3, KERN_DEBUG "%s: avmaster_unregister_i2c()\n", avms->name);
	i2c_del_adapter(&avms->i2c_adapter);
	kfree(avms->i2c_mem);
}

/* Check a avmaster_params struct for correctness, insert default params */

int
avmaster_check_jpg_settings (struct avmaster              *avms,
			  struct avmaster_jpg_settings *settings)
{
	int err = 0, err0 = 0;
	dprintk(4,
		KERN_DEBUG
		"%s: check_jpg_settings() - dec: %d, Hdcm: %d, Vdcm: %d, Tdcm: %d\n",
		avms->name, settings->decimation, settings->HorDcm,
		settings->VerDcm, settings->TmpDcm);
	dprintk(4,
		KERN_DEBUG
		"%s: check_jpg_settings() - x: %d, y: %d, w: %d, y: %d\n",
		avms->name, settings->img_x, settings->img_y,
		settings->img_width, settings->img_height);
	/* Check decimation, set default values for decimation = 1, 2, 4 */
	switch (settings->decimation) {
	case 1:

		settings->HorDcm = 1;
		settings->VerDcm = 1;
		settings->TmpDcm = 1;
		settings->field_per_buff = 2;
		settings->img_x = 0;
		settings->img_y = 0;
		settings->img_width = AVM_MAX_WIDTH;
		settings->img_height = AVM_MAX_HEIGHT / 2;
		break;
	case 2:

		settings->HorDcm = 2;
		settings->VerDcm = 1;
		settings->TmpDcm = 2;
		settings->field_per_buff = 1;
		settings->img_x = (AVM_MAX_WIDTH == 720) ? 8 : 0;
		settings->img_y = 0;
		settings->img_width =
		    (AVM_MAX_WIDTH == 720) ? 704 : AVM_MAX_WIDTH;
		settings->img_height = AVM_MAX_HEIGHT / 2;
		break;
	case 4:

		settings->HorDcm = 4;
		settings->VerDcm = 2;
		settings->TmpDcm = 2;
		settings->field_per_buff = 1;
		settings->img_x = (AVM_MAX_WIDTH == 720) ? 8 : 0;
		settings->img_y = 0;
		settings->img_width =
		    (AVM_MAX_WIDTH == 720) ? 704 : AVM_MAX_WIDTH;
		settings->img_height = AVM_MAX_HEIGHT / 2;
		break;
	case 0:

		/* We have to check the data the user has set */

		if (settings->HorDcm != 1 && settings->HorDcm != 2)
			err0++;
		if (settings->VerDcm != 1 && settings->VerDcm != 2)
			err0++;
		if (settings->TmpDcm != 1 && settings->TmpDcm != 2)
			err0++;
		if (settings->field_per_buff != 1 &&
		    settings->field_per_buff != 2)
			err0++;
		if (settings->img_x < 0)
			err0++;
		if (settings->img_y < 0)
			err0++;
		if (settings->img_width < 0)
			err0++;
		if (settings->img_height < 0)
			err0++;
		if (settings->img_x + settings->img_width > AVM_MAX_WIDTH)
			err0++;
		if (settings->img_y + settings->img_height >
		    AVM_MAX_HEIGHT / 2)
			err0++;
		if (settings->HorDcm && settings->VerDcm) {
			if (settings->img_width %
			    (16 * settings->HorDcm) != 0)
				err0++;
			if (settings->img_height %
			    (8 * settings->VerDcm) != 0)
				err0++;
		}

		if (err0) {
			dprintk(1,
				KERN_ERR
				"%s: check_jpg_settings() - error in params for decimation = 0\n",
				avms->name);
			err++;
		}
		break;
	default:
		dprintk(1,
			KERN_ERR
			"%s: check_jpg_settings() - decimation = %d, must be 0, 1, 2 or 4\n",
			avms->name, settings->decimation);
		err++;
		break;
	}

	if (settings->jpg_comp.quality > 100)
		settings->jpg_comp.quality = 100;
	if (settings->jpg_comp.quality < 5)
		settings->jpg_comp.quality = 5;
	if (settings->jpg_comp.APPn < 0)
		settings->jpg_comp.APPn = 0;
	if (settings->jpg_comp.APPn > 15)
		settings->jpg_comp.APPn = 15;
	if (settings->jpg_comp.APP_len < 0)
		settings->jpg_comp.APP_len = 0;
	if (settings->jpg_comp.APP_len > 60)
		settings->jpg_comp.APP_len = 60;
	if (settings->jpg_comp.COM_len < 0)
		settings->jpg_comp.COM_len = 0;
	if (settings->jpg_comp.COM_len > 60)
		settings->jpg_comp.COM_len = 60;
	if (err)
		return -EINVAL;
	return 0;
}

void
avmaster_open_init_params (struct avmaster *avms)
{
	int i;

	/* Shadowed register and IRQ count */
	avms->vdmac=0;
	avms->cdmac=0;
	memset(avms->irq_count,0,sizeof(avms->irq_count));

	/* User must explicitly set a window */
	avms->overlay_settings.is_set = 0;
	avms->overlay_mask = NULL;
	avms->overlay_active = AVMASTER_FREE;

	avms->v4l_memgrab_active = 0;
	avms->v4l_overlay_active = 0;
	avms->v4l_grab_frame = NO_GRAB_ACTIVE;
	avms->v4l_grab_seq = 0;
	avms->v4l_settings.width = 192;
	avms->v4l_settings.height = 144;
	avms->v4l_settings.format = &avmaster_formats[4];	/* YUY2 - YUV-4:2:2 packed */
	avms->v4l_settings.bytesperline =
	    avms->v4l_settings.width *
	    ((avms->v4l_settings.format->depth + 7) / 8);

	/* DMA ring stuff for V4L */
	avms->v4l_pend_tail = 0;
	avms->v4l_pend_head = 0;
	avms->v4l_buffers.active = AVMASTER_FREE;
	for (i = 0; i < VIDEO_MAX_FRAME; i++) {
		avms->v4l_buffers.buffer[i].state = AVM_STATE_USER;	/* nothing going on */
	}
	avms->v4l_buffers.allocated = 0;

	for (i = 0; i < AVM_MAX_FRAME; i++) {
		avms->jpg_buffers.buffer[i].state = AVM_STATE_USER;	/* nothing going on */
	}
	avms->jpg_buffers.active = AVMASTER_FREE;
	avms->jpg_buffers.allocated = 0;
	/* Set necessary params and call avmaster_check_jpg_settings to set the defaults */
	avms->jpg_settings.decimation = 1;
	avms->jpg_settings.jpg_comp.quality = 50;	/* default compression factor 8 */
	avms->jpg_settings.odd_even = 1;
	avms->jpg_settings.jpg_comp.APPn = 0;
	avms->jpg_settings.jpg_comp.APP_len = 0;	/* No APPn marker */
	memset(avms->jpg_settings.jpg_comp.APP_data, 0,
	       sizeof(avms->jpg_settings.jpg_comp.APP_data));
	avms->jpg_settings.jpg_comp.COM_len = 0;	/* No COM marker */
	memset(avms->jpg_settings.jpg_comp.COM_data, 0,
	       sizeof(avms->jpg_settings.jpg_comp.COM_data));
	avms->jpg_settings.jpg_comp.jpeg_markers =
	    JPEG_MARKER_DHT | JPEG_MARKER_DQT;
	i = avmaster_check_jpg_settings(avms, &avms->jpg_settings);
	if (i)
		dprintk(1,
			KERN_ERR
			"%s: avmaster_open_init_params() internal error\n",
			avms->name);

	clear_interrupt_counters(avms);
	avms->testing = 0;
}

static void __devinit
test_interrupts (struct avmaster *avms)
{
#if 0
	int timeout, icr;

	clear_interrupt_counters(avms);

	avms->testing = 1;
	icr = btread(ZR36057_ICR);
	btwrite(0x78000000 | ZR36057_ICR_IntPinEn, ZR36057_ICR);
	timeout = interruptible_sleep_on_timeout(&avms->test_q, 1 * HZ);
	btwrite(0, ZR36057_ICR);
	btwrite(0x78000000, ZR36057_ISR);
	avms->testing = 0;
	dprintk(5, KERN_INFO "%s: Testing interrupts...\n", avms->name);
	if (timeout) {
		dprintk(1, ": time spent: %d\n", 1 * HZ - timeout);
	}
	if (debug > 1)
		print_interrupts(avms);
	btwrite(icr, ZR36057_ICR);
#endif
}

static int __devinit
saa7145_init (struct avmaster *avms)
{
	int two = 2;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	void *vdev;
#endif
	int zero = 0;

	dprintk(1,
		KERN_INFO
		"%s: saa7145_init() - initializing card[%d], avms=%p\n",
		avms->name, avms->id, avms);

	/* default setup of all parameters which will persist between opens */
	avms->user = 0;

	init_waitqueue_head(&avms->v4l_capq);
	init_waitqueue_head(&avms->jpg_capq);
	init_waitqueue_head(&avms->test_q);
	avms->jpg_buffers.allocated = 0;
	avms->v4l_buffers.allocated = 0;

	avms->buffer.base = (void *) vidmem;
	avms->buffer.width = 0;
	avms->buffer.height = 0;
	avms->buffer.depth = 0;
	avms->buffer.bytesperline = 0;

	/* Avoid nonsense settings from user for default input/norm */
	if (default_norm < VIDEO_MODE_PAL &&
	    default_norm > VIDEO_MODE_SECAM)
		default_norm = VIDEO_MODE_PAL;
	avms->norm = default_norm;
	if (!(avms->timing = avms->card.tvn[avms->norm])) {
		dprintk(1,
			KERN_WARNING
			"%s: saa7145_init() - default TV standard not supported by hardware. PAL will be used.\n",
			avms->name);
		avms->norm = VIDEO_MODE_PAL;
		avms->timing = avms->card.tvn[avms->norm];
	}

	avms->input = default_input = (default_input ? 1 : 0);

	/* Should the following be reset at every open ? */
	avms->hue = 32768;
	avms->contrast = 32768;
	avms->saturation = 32768;
	avms->brightness = 32768;

	/* default setup (will be repeated at every open) */
	avmaster_open_init_params(avms);

	/* allocate memory *before* doing anything to the hardware
	 * in case allocation fails */
	avms->rps_mem = (u32*)kmalloc(SAA7145_RPS_MEM, GFP_KERNEL);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	vdev = (void *) kmalloc(sizeof(struct video_device), GFP_KERNEL);
	if( !avms->rps_mem || !vdev)
#else
	if( !avms->rps_mem )
#endif
	{
		dprintk(1, KERN_ERR "%s: saa7145_init() - kmalloc (rps_mem) failed\n",
			avms->name);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		if(vdev)
			kfree(vdev);
#endif
		if(avms->rps_mem)
			kfree(avms->rps_mem);
		return -ENOMEM;
	}
	avms->rps_bus = virt_to_bus(avms->rps_mem);

	/*
	 *   Now add the template and register the device unit.
	 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	avms->video_dev=&avms->video_dev_buf;
#else
	avms->video_dev=vdev;
#endif
	memcpy(avms->video_dev, &avmaster_template, sizeof(avmaster_template));
	strcpy(avms->video_dev->name, avms->name);
	if (video_register_device (avms->video_dev, VFL_TYPE_GRABBER, video_nr) < 0)
	{
		avmaster_unregister_i2c(avms);
		kfree((void *) avms->rps_mem);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		kfree(vdev);
#endif
		return -1;
	}

	avmaster_init_hardware(avms);
	if (debug > 2)
		detect_guest_activity(avms);
	test_interrupts(avms);
	if (!pass_through) {
		decoder_command(avms, DECODER_ENABLE_OUTPUT, &zero);
		encoder_command(avms, ENCODER_SET_INPUT, &two);
	}

	avms->avmaster_proc = NULL;
	avms->initialized = 1;
	return 0;
}

static void
avmaster_release (struct avmaster *avms)
{
	if (!avms->initialized)
		return;
	/* unregister videocodec bus */
	if (avms->codec) {
		struct videocodec_master *master = avms->codec->master_data;
		videocodec_detach(avms->codec);
		if (master)
			kfree(master);
	}
	if (avms->vfe) {
		struct videocodec_master *master = avms->vfe->master_data;
		videocodec_detach(avms->vfe);
		if (master)
			kfree(master);
	}

	/* unregister i2c bus */
	avmaster_unregister_i2c(avms);
	/* disable PCI bus-mastering */
	avmaster_set_pci_master(avms, 0);
	/* put chip into reset */
	free_irq(avms->pci_dev->irq, avms);
	/* unmap and free memory */
	kfree((void *) avms->rps_mem);
	avmaster_proc_cleanup(avms);
	iounmap(avms->saa7145_mem);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	pci_disable_device(avms->pci_dev);
#endif
	video_unregister_device(avms->video_dev);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
void
avmaster_vdev_release (struct video_device *vdev)
{
        kfree(vdev);
}	
#endif


#if 0
static struct videocodec_master * __devinit
avmaster_setup_videocodec (struct avmaster *avms,
			int           type)
{
	struct videocodec_master *m = NULL;

	m = kmalloc(sizeof(struct videocodec_master), GFP_KERNEL);
	if (!m) {
		dprintk(1,
			KERN_ERR
			"%s: avmaster_setup_videocodec() - no memory\n",
			avms->name);
		return m;
	}

	m->magic = 0L; /* magic not used */
	m->type = VID_HARDWARE_SAA7145;
	m->flags = CODEC_FLAG_ENCODER | CODEC_FLAG_DECODER;
	strncpy(m->name, avms->name, sizeof(m->name));
	m->data = avms;

	switch (type)
	{
	case CODEC_TYPE_ZR36050:
		m->readreg = zr36050_read;
		m->writereg = zr36050_write;
		m->flags |= CODEC_FLAG_JPEG;
		break;
	case CODEC_TYPE_ZR36016:
		m->readreg = zr36015_read;
		m->writereg = zr36015_write;
		m->flags |= CODEC_FLAG_VFE;
		break;
	}

	return m;
}
#endif

/*
 *   Scan for a Buz card (actually for the PCI contoler ZR36057),
 *   request the irq and map the io memory
 */
static int __devinit
find_saa7145 (void)
{
	unsigned char latency, need_latency;
	struct avmaster *avms;
	struct pci_dev *dev = NULL;
	int result;
	//struct videocodec_master *master_vfe = NULL;
	//struct videocodec_master *master_codec = NULL;
	int card_num;
	char *i2c_enc_name, *i2c_dec_name, *codec_name, *vfe_name;

	avmaster_num = 0;
	while (avmaster_num < AVM_MAX &&
	       (dev =
		pci_find_device(PCI_VENDOR_ID_PHILIPS,
				PCI_DEVICE_ID_PHILIPS_SAA7145, dev)) != NULL) {
		card_num = card[avmaster_num];
		avms = &avmaster[avmaster_num];
		memset(avms, 0, sizeof(struct avmaster));	// Just in case if previous cycle failed
		avms->pci_dev = dev;
		//avms->saa7145_mem = NULL;
		avms->id = avmaster_num;
		snprintf(avms->name, sizeof(avms->name), "MJPEG[%u]", avms->id);
		spin_lock_init(&avms->spinlock);
		init_MUTEX(&avms->resource_lock);
		if (pci_enable_device(dev))
			continue;
		avms->saa7145_adr = pci_resource_start(avms->pci_dev, 0);
		pci_read_config_byte(avms->pci_dev, PCI_CLASS_REVISION,
				     &avms->revision);
		int i;
		unsigned short ss_vendor, ss_device;
		ss_vendor = avms->pci_dev->subsystem_vendor;
		ss_device = avms->pci_dev->subsystem_device;
		dprintk(1,
			KERN_INFO
			"%s: Philips SAA7145 (rev %d) irq: %d, memory: 0x%08x\n",
			avms->name, avms->revision, avms->pci_dev->irq,
			avms->saa7145_adr);
		dprintk(1,
			KERN_INFO
			"%s: subsystem vendor=0x%04x id=0x%04x\n",
			avms->name, ss_vendor, ss_device);
		if (card_num == -1) {
			dprintk(3,
				KERN_DEBUG
				"%s: find_saa7145() - trying to autodetect card type\n",
				avms->name);
			for (i=0;i<NUM_CARDS;i++) {
				if (ss_vendor == avmaster_cards[i].vendor_id &&
				    ss_device == avmaster_cards[i].device_id) {
					dprintk(3,
						KERN_DEBUG
						"%s: find_saa7145() - card %s detected\n",
						avms->name,
						avmaster_cards[i].name);
					card_num = i;
					break;
				}
			}
			if (i == NUM_CARDS) {
				dprintk(1,
					KERN_ERR
					"%s: find_saa7145() - unknown card\n",
					avms->name);
				continue;
			}
		}

		if (card_num < 0 || card_num >= NUM_CARDS) {
			dprintk(2,
				KERN_ERR
				"%s: find_saa7145() - invalid cardnum %d\n",
				avms->name, card_num);
			continue;
		}

		/* even though we make this a non pointer and thus
		 * theoretically allow for making changes to this struct
		 * on a per-individual card basis at runtime, this is
		 * strongly discouraged. This structure is intended to
		 * keep general card information, no settings or anything */
		avms->card = avmaster_cards[card_num];
		snprintf(avms->name, sizeof(avms->name),
			 "%s[%u]", avms->card.name, avms->id);

		avms->saa7145_mem = ioremap_nocache(avms->saa7145_adr, 0x1000);
		if (!avms->saa7145_mem) {
			dprintk(1,
				KERN_ERR
				"%s: find_saa7145() - ioremap failed\n",
				avms->name);
			continue;
		}
		avms->saa7145_reg = (unsigned long*)avms->saa7145_mem;

		result =
		    request_irq(avms->pci_dev->irq, avmaster_irq,
				SA_SHIRQ | SA_INTERRUPT, avms->name,
				(void *) avms);
		if (result < 0) {
			if (result == -EINVAL) {
				dprintk(1,
					KERN_ERR
					"%s: find_saa7145() - bad irq number or handler\n",
					avms->name);
			} else if (result == -EBUSY) {
				dprintk(1,
					KERN_ERR
					"%s: find_saa7145() - IRQ %d busy, change your PnP config in BIOS\n",
					avms->name, avms->pci_dev->irq);
			} else {
				dprintk(1,
					KERN_ERR
					"%s: find_saa7145() - can't assign irq, error code %d\n",
					avms->name, result);
			}
			goto avms_unmap;
		}

		/* set PCI latency timer */
		pci_read_config_byte(avms->pci_dev, PCI_LATENCY_TIMER,
				     &latency);
		need_latency = avms->revision > 1 ? 32 : 48;
		if (latency != need_latency) {
			dprintk(2,
				KERN_INFO
				"%s: Changing PCI latency from %d to %d.\n",
				avms->name, latency, need_latency);
			pci_write_config_byte(avms->pci_dev,
					      PCI_LATENCY_TIMER,
					      need_latency);
		}

		saa7145_restart(avms);
		/* i2c */
		dprintk(2, KERN_INFO "%s: Initializing i2c bus...\n",
			avms->name);

		/* i2c decoder */
		if (decoder[avms->id] != -1) {
			i2c_dec_name = i2cid_to_modulename(decoder[avms->id]);
			avms->card.i2c_decoder = decoder[avms->id];
		} else if (avms->card.i2c_decoder != 0) {
			i2c_dec_name =
				i2cid_to_modulename(avms->card.i2c_decoder);
		} else {
			i2c_dec_name = NULL;
		}

		if (i2c_dec_name) {
			if ((result = request_module(i2c_dec_name)) < 0) {
				dprintk(1,
					KERN_ERR
					"%s: failed to load module %s: %d\n",
					avms->name, i2c_dec_name, result);
			}
		}

		/* i2c encoder */
		if (encoder[avms->id] != -1) {
			i2c_enc_name = i2cid_to_modulename(encoder[avms->id]);
			avms->card.i2c_encoder = encoder[avms->id];
		} else if (avms->card.i2c_encoder != 0) {
			i2c_enc_name =
				i2cid_to_modulename(avms->card.i2c_encoder);
		} else {
			i2c_enc_name = NULL;
		}

		if (i2c_enc_name) {
			if ((result = request_module(i2c_enc_name)) < 0) {
				dprintk(1,
					KERN_ERR
					"%s: failed to load module %s: %d\n",
					avms->name, i2c_enc_name, result);
			}
		}

		if (avmaster_register_i2c(avms) < 0) {
			dprintk(1,
				KERN_ERR
				"%s: find_saa7145() - can't initialize i2c bus\n",
				avms->name);
			goto avms_free_irq;
		}

		dprintk(2,
			KERN_INFO "%s: Initializing videocodec bus...\n",
			avms->name);

		if (avms->card.video_codec != 0 &&
		    (codec_name =
		     codecid_to_modulename(avms->card.video_codec)) != NULL) {
			if ((result = request_module(codec_name)) < 0) {
				dprintk(1,
					KERN_ERR
					"%s: failed to load modules %s: %d\n",
					avms->name, codec_name, result);
			}
		}
		if (avms->card.video_vfe != 0 &&
		    (vfe_name =
		     codecid_to_modulename(avms->card.video_vfe)) != NULL) {
			if ((result = request_module(vfe_name)) < 0) {
				dprintk(1,
					KERN_ERR
					"%s: failed to load modules %s: %d\n",
					avms->name, vfe_name, result);
			}
		}

#if 0
		/* reset JPEG codec */
		jpeg_codec_sleep(avms, 1);
		jpeg_codec_reset(avms);
		/* video bus enabled */
		/* display codec revision */
		if (avms->card.video_codec != 0) {
			master_codec = avmaster_setup_videocodec(avms,
							      avms->card.video_codec);
			if (!master_codec)
				goto avms_unreg_i2c;
			avms->codec = videocodec_attach(master_codec);
			if (!avms->codec) {
				dprintk(1,
					KERN_ERR
					"%s: find_saa7145() - no codec found\n",
					avms->name);
				goto avms_free_codec;
			}
			if (avms->codec->type != avms->card.video_codec) {
				dprintk(1,
					KERN_ERR
					"%s: find_saa7145() - wrong codec\n",
					avms->name);
				goto avms_detach_codec;
			}
		}
		if (avms->card.video_vfe != 0) {
			master_vfe = avmaster_setup_videocodec(avms,
							    avms->card.video_vfe);
			if (!master_vfe)
				goto avms_detach_codec;
			avms->vfe = videocodec_attach(master_vfe);
			if (!avms->vfe) {
				dprintk(1,
					KERN_ERR
					"%s: find_saa7145() - no VFE found\n",
					avms->name);
				goto avms_free_vfe;
			}
			if (avms->vfe->type != avms->card.video_vfe) {
				dprintk(1,
					KERN_ERR
					"%s: find_saa7145() = wrong VFE\n",
					avms->name);
				goto avms_detach_vfe;
			}
		}
#endif
		avmaster_num++;
		continue;

		// Init errors
/*
	      avms_detach_vfe:
		videocodec_detach(avms->vfe);
	      avms_free_vfe:
		kfree(master_vfe);
	      avms_detach_codec:
		videocodec_detach(avms->codec);
	      avms_free_codec:
		kfree(master_codec);
	      avms_unreg_i2c:
		avmaster_unregister_i2c(avms);
*/
	      avms_free_irq:
		free_irq(avms->pci_dev->irq, avms);
	      avms_unmap:
		iounmap(avms->saa7145_mem);
		continue;
	}
	if (avmaster_num == 0) {
		dprintk(1, KERN_INFO "No known MJPEG cards found.\n");
	}
	return avmaster_num;
}

static int __init
init_avmaster_cards (void)
{
	int i;
	memset(avmaster, 0, sizeof(avmaster));
	printk(KERN_INFO "Fast AV-Master MJPEG board driver version %d.%d.%d\n",
	       MAJOR_VERSION, MINOR_VERSION, RELEASE_VERSION);

	/* Look for cards */
	if (find_saa7145() < 0) {
		return -EIO;
	}
	if (avmaster_num == 0)
		return -ENODEV;
	dprintk(1, KERN_INFO "%s: %d card(s) found\n", AVMASTER_NAME,
		avmaster_num);
	/* check the parameters we have been given, adjust if necessary */
	if (v4l_nbufs < 2)
		v4l_nbufs = 2;
	if (v4l_nbufs > VIDEO_MAX_FRAME)
		v4l_nbufs = VIDEO_MAX_FRAME;
	/* The user specfies the in KB, we want them in byte
	 * (and page aligned) */
	v4l_bufsize = PAGE_ALIGN(v4l_bufsize * 1024);
	if (v4l_bufsize < 32768)
		v4l_bufsize = 32768;
	/* 2 MB is arbitrary but sufficient for the maximum possible images */
	if (v4l_bufsize > 2048 * 1024)
		v4l_bufsize = 2048 * 1024;
	if (jpg_nbufs < 4)
		jpg_nbufs = 4;
	if (jpg_nbufs > AVM_MAX_FRAME)
		jpg_nbufs = AVM_MAX_FRAME;
	jpg_bufsize = PAGE_ALIGN(jpg_bufsize * 1024);
	if (jpg_bufsize < 8192)
		jpg_bufsize = 8192;
	if (jpg_bufsize > (512 * 1024))
		jpg_bufsize = 512 * 1024;
	/* Use parameter for vidmem or try to find a video card */
	if (vidmem) {
		dprintk(1,
			KERN_INFO
			"%s: Using supplied video memory base address @ 0x%lx\n",
			AVMASTER_NAME, vidmem);
	}

	/* random nonsense */
	dprintk(5, KERN_DEBUG "Jotti is een held!\n");

	/* some mainboards might not do PCI-PCI data transfer well */
	if (pci_pci_problems & PCIPCI_FAIL) {
		dprintk(1,
			KERN_WARNING
			"%s: chipset may not support reliable PCI-PCI DMA\n",
			AVMASTER_NAME);
	}

	/* take care of Natoma chipset and a revision 1 saa7145 */
	for (i = 0; i < avmaster_num; i++) {
		struct avmaster *avms = &avmaster[i];

		if (saa7145_init(avms) < 0) {
			for (i = 0; i < avmaster_num; i++)
				avmaster_release(&avmaster[i]);
			return -EIO;
		}
		avmaster_proc_init(avms);
	}

	return 0;
}

static void __exit
unload_avmaster_cards (void)
{
	int i;
	for (i = 0; i < avmaster_num; i++)
		avmaster_release(&avmaster[i]);
}

module_init(init_avmaster_cards);
module_exit(unload_avmaster_cards);
