/*
 * Philips saa7145 controller driver, for the
 * Fast Multimedia AV Master.
 *
 * This part handles device access (PCI/I2C/codec/...)
 * 
 * Copyright (C) 2000 Serguei Miridonov <mirsev@cicese.mx>
 *
 * Currently maintained by:
 *   Luca Bonissi <lucabon@interfree.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/vmalloc.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/interrupt.h>
#endif

#include <linux/proc_fs.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <linux/videodev.h>
#include <linux/spinlock.h>
#include <linux/sem.h>

#include <linux/pci.h>
#include <linux/video_decoder.h>
#include <linux/video_encoder.h>
#include <linux/delay.h>

#include "videocodec.h"
#include "avmaster.h"
#include "avmaster_device.h"

#define IRQ_MASK1 ( 0xffffffff & \
		   ~(ISRC_IIC_D	| \
		     ISRC_FIDN	| \
		     ISRC_FID	| \
		     ISRC_VFO	| \
		     ISRC_VPE 	  \
		   ))

#define IRQ_MASK (0xffffffff)

const char *irq_name[]={
	"MASTER",
	"Event Counter 1",
	"Event Counter 2",
	"Reserved",
	"GPIO-0",
	"GPIO-1",
	"GPIO-2",
	"GPIO-3",
	"Video FIFO overflow",
	"Video DMA Address protection",
	"Clipper FIFO Underflow",
	"Audio FIFO overflow",
	"Field ID",
	"Field ID Not",
	"Horizontal Blank Interval",
	"Vertical Blank Interval",
	"Video FIFO empty",
	"RPS Page Error",
	"RPS Late Error",
	"RPS Timeout",
	"RPS Interrupt command",
	"RPS Semaphore",
	"RPS Upload",
	"Audio Range",
	"I2C done",
	"I2C error",
	"DEBI timeout",
	"DEBI XIRQ",
	"DEBI done",
	"PCI parity error during write",
	"PCI access abort",
	"PCI parity error during read"
};

extern const struct avmaster_format avmaster_formats[];
extern const int avmaster_num_formats;

/* helper functioni - defined in avmaster_card.c */
extern void my_wait(long ms);

extern int debug;

int upload_done=0;

#define dprintk(num, format, args...) \
	do { \
		if (debug >= num) \
			printk(format, ##args); \
	} while (0)


/*
 * General Purpose I/O and Guest bus access
 */

/*
 * This is a bit tricky. When a board lacks a GPIO function, the corresponding
 * GPIO bit number in the card_info structure is set to 0.
 */

void
GPIO (struct avmaster *avms,
      int           bit,
      unsigned int  value)
{
	//u32 reg;
	//u32 mask;

	/* Make sure the bit number is legal
	 * A bit number of -1 (lacking) gives a mask of 0,
	 * making it harmless */
	/*
	mask = (1 << (24 + bit)) & 0xff000000;
	reg = btread(GPIOC) & ~mask;
	if (value) {
		reg |= mask;
	}
	btwrite(reg, GPIOC);
	udelay(1);
	*/
}

/*
 * Wait til post office is no longer busy
 */

int
post_office_wait (struct avmaster *avms)
{
#if 0
	u32 por;

//      while (((por = btread(ZR36057_POR)) & (ZR36057_POR_POPen | ZR36057_POR_POTime)) == ZR36057_POR_POPen) {
	while ((por = btread(ZR36057_POR)) & ZR36057_POR_POPen) {
		/* wait for something to happen */
	}
	if ((por & ZR36057_POR_POTime) && !avms->card.gws_not_connected) {
		/* In LML33/AVM \GWS line is not connected, so it has always timeout set */
		dprintk(1, KERN_INFO "%s: pop timeout %08x\n", avms->name,
			por);
		return -1;
	}

#endif
	return 0;
}

int
post_office_write (struct avmaster *avms,
		   unsigned int  guest,
		   unsigned int  reg,
		   unsigned int  value)
{
#if 0
	u32 por;

	por =
	    ZR36057_POR_PODir | ZR36057_POR_POTime | ((guest & 7) << 20) |
	    ((reg & 7) << 16) | (value & 0xFF);
	btwrite(por, ZR36057_POR);

#endif
	return post_office_wait(avms);
}

int
post_office_read (struct avmaster *avms,
		  unsigned int  guest,
		  unsigned int  reg)
{
#if 0
	u32 por;

	por = ZR36057_POR_POTime | ((guest & 7) << 20) | ((reg & 7) << 16);
	btwrite(por, ZR36057_POR);
	if (post_office_wait(avms) < 0) {
		return -1;
	}

	return btread(ZR36057_POR) & 0xFF;
#endif
	return 0;
}

/*
 * detect guests
 */

#if 0
static void
dump_guests (struct avmaster *avms)
{
	if (debug > 2) {
		int i, guest[8];

		for (i = 1; i < 8; i++) {	// Don't read jpeg codec here
			guest[i] = post_office_read(avms, i, 0);
		}

		printk(KERN_INFO "%s: Guests:", avms->name);

		for (i = 1; i < 8; i++) {
			printk(" 0x%02x", guest[i]);
		}
		printk("\n");
	}
}
#endif

static inline unsigned long
get_time (void)
{
	struct timeval tv;
	do_gettimeofday(&tv);
	return (1000000 * tv.tv_sec + tv.tv_usec);
}

void
detect_guest_activity (struct avmaster *avms)
{
#if 0
	int timeout, i, j, res, guest[8], guest0[8], change[8][3];
	unsigned long t0, t1;

	dump_guests(avms);
	printk(KERN_INFO "%s: Detecting guests activity, please wait...\n",
	       avms->name);
	for (i = 1; i < 8; i++) {	// Don't read jpeg codec here
		guest0[i] = guest[i] = post_office_read(avms, i, 0);
	}

	timeout = 0;
	j = 0;
	t0 = get_time();
	while (timeout < 10000) {
		udelay(10);
		timeout++;
		for (i = 1; (i < 8) && (j < 8); i++) {
			res = post_office_read(avms, i, 0);
			if (res != guest[i]) {
				t1 = get_time();
				change[j][0] = (t1 - t0);
				t0 = t1;
				change[j][1] = i;
				change[j][2] = res;
				j++;
				guest[i] = res;
			}
		}
		if (j >= 8)
			break;
	}
	printk(KERN_INFO "%s: Guests:", avms->name);

	for (i = 1; i < 8; i++) {
		printk(" 0x%02x", guest0[i]);
	}
	printk("\n");
	if (j == 0) {
		printk(KERN_INFO "%s: No activity detected.\n", avms->name);
		return;
	}
	for (i = 0; i < j; i++) {
		printk(KERN_INFO "%s: %6d: %d => 0x%02x\n", avms->name,
		       change[i][0], change[i][1], change[i][2]);
	}
#endif
}

/*
 * JPEG Codec access
 */

void
jpeg_codec_sleep (struct avmaster *avms,
		  int           sleep)
{
	GPIO(avms, avms->card.gpio[GPIO_JPEG_SLEEP], !sleep);
#if 0
	if (!sleep) {
		dprintk(3,
			KERN_DEBUG
			"%s: jpeg_codec_sleep() - wake GPIO=0x%08x\n",
			avms->name, btread(ZR36057_GPPGCR1));
		udelay(500);
	} else {
		dprintk(3,
			KERN_DEBUG
			"%s: jpeg_codec_sleep() - sleep GPIO=0x%08x\n",
			avms->name, btread(ZR36057_GPPGCR1));
		udelay(2);
	}
#endif
}

int
jpeg_codec_reset (struct avmaster *avms)
{
	/* Take the codec out of sleep */
	jpeg_codec_sleep(avms, 0);

	if (avms->card.gpcs[GPCS_JPEG_RESET] != 0xff) {
		post_office_write(avms, avms->card.gpcs[GPCS_JPEG_RESET], 0,
				  0);
		udelay(2);
	} else {
		GPIO(avms, avms->card.gpio[GPIO_JPEG_RESET], 0);
		udelay(2);
		GPIO(avms, avms->card.gpio[GPIO_JPEG_RESET], 1);
		udelay(2);
	}
	return 0;
}

/*
 *   Set the registers for the size we have specified. Don't bother
 *   trying to understand this without the ZR36057 manual in front of
 *   you [AC].
 *
 *   PS: The manual is free for download in .pdf format from
 *   www.avmaster.com - nicely done those folks.
 */

static void
saa7145_adjust_vfe (struct avmaster          *avms,
		    enum avmaster_codec_mode  mode)
{
#if 0
	u32 reg;

	switch (mode) {
	case AVM_MODE_MOTION_DECOMPRESS:
		btand(~ZR36057_VFESPFR_ExtFl, ZR36057_VFESPFR);
		reg = btread(ZR36057_VFEHCR);
		if ((reg & (1 << 10)) && avms->card.type != LML33R10) {
			reg += ((1 << 10) | 1);
		}
		btwrite(reg, ZR36057_VFEHCR);
		break;
	case AVM_MODE_MOTION_COMPRESS:
	case AVM_MODE_IDLE:
	default:
		if (avms->norm == VIDEO_MODE_NTSC ||
		    (avms->card.type == LML33R10 &&
		     avms->norm == VIDEO_MODE_PAL))
			btand(~ZR36057_VFESPFR_ExtFl, ZR36057_VFESPFR);
		else
			btor(ZR36057_VFESPFR_ExtFl, ZR36057_VFESPFR);
		reg = btread(ZR36057_VFEHCR);
		if (!(reg & (1 << 10)) && avms->card.type != LML33R10) {
			reg -= ((1 << 10) | 1);
		}
		btwrite(reg, ZR36057_VFEHCR);
		break;
	}
#endif
}

/* Helper logaritm base-2, max 6+1 */
int log2_6plus1(int value)
{
	int i=0;
	if(i>=64) return 7;
	while((value >>= 1)>0) i++;
	return(i+1);
}
/*
 * set geometry
 */

static void
saa7145_set_vfe (struct avmaster              *avms,
		 int                        video_width,
		 int                        video_height,
		 const struct avmaster_format *format)
{
	struct tvnorm *tvn;
	unsigned Wa, Ha;
	u32 reg;
	int mask_line_size;
	int first_dword_offset;
	int ilo;

	tvn = avms->timing;

	Wa = tvn->Wa;
	Ha = tvn->Ha;

	dprintk(2, KERN_INFO "%s: set_vfe() - width = %d, height = %d\n",
		avms->name, video_width, video_height);

	if (avms->norm != VIDEO_MODE_PAL &&
	    avms->norm != VIDEO_MODE_NTSC &&
	    avms->norm != VIDEO_MODE_SECAM) {
		dprintk(1,
			KERN_ERR "%s: set_vfe() - norm = %d not valid\n",
			avms->name, avms->norm);
		return;
	}
	if (video_width < AVM_MIN_WIDTH ||
	    video_height < AVM_MIN_HEIGHT ||
	    video_width > Wa || video_height > Ha) {
		dprintk(1, KERN_ERR "%s: set_vfe: w=%d h=%d not valid\n",
			avms->name, video_width, video_height);
		return;
	}

	/**** saa7145 ****/

	/* horizontal */
        reg = ((tvn->HStart & XWC_XWS_M) << XWC_XWS_O)
            | ((tvn->Wa     & XWC_XWZ_M) << XWC_XWZ_O);
        btwrite(reg, XWC);
        /* Vertical */
        reg = ((tvn->VStart & YWC_YWS_M) << YWC_YWS_O)
            | (((tvn->Ha/2)     & YWC_YWZ_M) << YWC_YWZ_O);
        btwrite(reg, YWC);
        /* scaler and pixel format */
        ilo = (video_height > (AVM_MAX_HEIGHT / 2) ? 1: 0);
        //reg = (ilo ? FSCTL_ILO : 0);
        reg = FSCTL_ILO;
        reg |= log2_6plus1(tvn->Wa/video_width) << FSCTL_FMODE_O;
        reg |= ((1024*video_width/tvn->Wa) & FSCTL_XSCI_M ) << FSCTL_XSCI_O;
        reg |= ((1024*video_height/tvn->Ha) & FSCTL_YSCI_M ) << FSCTL_YSCI_O;

	btwrite(reg, FSCTL);


	/* Writing YPO and YPE, both 512 for line pair scaling (interlaced) 
	   or not interlaced */
	btwrite((512 << YPR_YPO_O) | (512 << YPR_YPE_O) , YPR);
	btwrite( VPCTL_XSS|VPCTL_VPZ, VPCTL);

	/* Writing Line Counter Threshold */
	reg = (((tvn->VStart+(tvn->Ha/2)) & LCR_SLCT_M) << LCR_SLCT_O) |
	      (((tvn->Ha/2) & LCR_TLCT_M) << LCR_TLCT_O);
	btwrite(reg, LCR);

	reg=0;
	switch (format->palette) {
	case VIDEO_PALETTE_YUV422:
		reg |= VDMAC_PFMT_YUV16;
		break;
	case VIDEO_PALETTE_RGB555:
		reg |= VDMAC_PFMT_RGB15;
		break;
	case VIDEO_PALETTE_RGB565:
		reg |= VDMAC_PFMT_RGB16;
		break;
	case VIDEO_PALETTE_RGB24:
		reg |= VDMAC_PFMT_RGB24;
		break;
	case VIDEO_PALETTE_RGB32:
		reg |= VDMAC_PFMT_RGB32;
		break;
	default:
		dprintk(1,
			KERN_INFO "%s: set_vfe() - unknown color_fmt=%x\n",
			avms->name, format->palette);
		return;
	}
	reg |= VDMAC_VEND_NONE;
	btwritemask(avms->vdmac, reg, VDMAC_VEND | VDMAC_PFMT, VDMAC);

	/* (Ronald) don't write this if overlay_mask = NULL */
	if (avms->overlay_mask) {
		/* Write overlay clipping mask data, but don't enable overlay clipping */
		/* RJ: since this makes only sense on the screen, we use 
		 * avms->overlay_settings.width instead of video_width */
		mask_line_size = (avms->overlay_settings.width + 31 ) / 32;
		first_dword_offset = mask_line_size*32 - avms->overlay_settings.width;
		reg = virt_to_bus(avms->overlay_mask);
		btwrite(reg, CDBA);
		btwritemask(avms->cdmac,(mask_line_size)&CDMAC_CPCH,CDMAC_CPCH,CDMAC);
		reg  = (mask_line_size & CDF_CPIX_M) << CDF_CPIX_O;
		reg |= (first_dword_offset & CDF_CPO_M) << CDF_CPO_O;
		reg |= (avms->overlay_settings.height & CDF_CLINE_M) << CDF_CLINE_O;
		btwrite(reg, CDF);
	}
	saa7145_adjust_vfe(avms, avms->codec_mode);
}

/*
 * Switch overlay on or off
 */

void saa7145_upload(struct avmaster *avms)
{
	int i=0;
	upload_done=0;
	btwrite(ISRC_UPLD,IMR);
	// Wait maximum 10 msec
	while(i<10 && !upload_done)
	{
	  my_wait(1);
	  i++;
	}
}

void saa7145_safeupload(struct avmaster *avms)
{
	/* Safe upload with RPS after the End of Active Window.
	   First of all, check if the Video Y parameter is set and stop the RPS */

	if((btread(YWC) & YWC_YWZ_M)==0) {
	  /* Video Parameters not loaded, force the upload */
	  saa7145_upload(avms);
	  return;
	}

	btwrite(0,RPSR);

	int i;
	u32 *ptr=avms->rps_mem;
	RPS_CLR(ptr,RPS_FLG_EAW);
	RPS_CHECK(ptr,RPS_FLG_EAW);
	RPS_STOP(ptr);
	if (avms->rps_bus & 3)
		dprintk (1,KERN_ERR "%s: RPS - program is not DWORD aligned\n",avms->name);
	else {
		upload_done=0;
		btwrite(RPSTO_TOSEL_BOTH | 
		        ((2 & RPSTO_TOV_M)<<RPSTO_TOV_O) |
			(PCI_CLOCK/25*2)<<RPSTO_TOP_O,
			RPSTO);
		btwrite(avms->rps_bus | RPSR_ERPSP ,RPSR);
		btwrite(avms->rps_bus | RPSPTR_ERPS ,RPSPTR);
		i=0;
		// Wait maximum 80 msec (about 2 PAL frames)
		while(i<80 && !upload_done)
		{
		  my_wait(1);
		  i++;
		}
		if(!upload_done) {
			dprintk (1,KERN_ERR "%s: Time Out doing upload, forcing\n",avms->name);
			saa7145_upload(avms);
		}
		else {
			dprintk(4,KERN_DEBUG "%s: RPS Upload in %d msec\n",avms->name,i);
		}
	}
}

void
saa7145_overlay (struct avmaster *avms,
		 int           on)
{
	u32 reg;

	if (on) {
		/* do the necessary settings ... */
		if(btread(VDMAC) & VDMAC_EVID ) {
			btwritemask(avms->vdmac, 0, VDMAC_EVID, VDMAC);	/* switch it off first */
			saa7145_safeupload(avms);
		}

		saa7145_set_vfe(avms,
				avms->overlay_settings.width,
				avms->overlay_settings.height,
				avms->overlay_settings.format);

		/* Start and length of each line MUST be 4-byte aligned.
		 * This should be allready checked before the call to this routine.
		 * All error messages are internal driver checking only! */

		/* video display top and bottom registers */
		reg = (u32) avms->buffer.base +
		    avms->overlay_settings.x *
		    ((avms->overlay_settings.format->depth + 7) / 8) +
		    avms->overlay_settings.y *
		    avms->buffer.bytesperline;
		dprintk(4,KERN_DEBUG "%s: Overlay Base: %08x\n",avms->name,(u32)avms->buffer.base );
		dprintk(4,KERN_DEBUG "%s: Overlay X   : %08x\n",avms->name,avms->overlay_settings.x );
		dprintk(4,KERN_DEBUG "%s: Overlay Y   : %08x\n",avms->name,avms->overlay_settings.y );
		dprintk(4,KERN_DEBUG "%s: Overlay FMT : %08x\n",avms->name,avms->overlay_settings.format->depth );
		dprintk(4,KERN_DEBUG "%s: Overlay BPL : %08x\n",avms->name,avms->buffer.bytesperline );
		dprintk(4,KERN_DEBUG "%s: writing Video DMA Base address: %08x\n",avms->name,reg);
		btwrite(reg, VDBA);
		btwrite(((reg+avms->buffer.bytesperline*(avms->overlay_settings.height+2)+4) | 1),VDPA);

		/* Set overlay clipping */
		if (avms->overlay_settings.clipcount > 0) 
			btwritemask(avms->cdmac,CDMAC_ECLIP|CDMAC_CEND_NONE_BE, CDMAC_ECLIP|CDMAC_CEND, CDMAC)
		else 
			btwritemask(avms->cdmac,0,CDMAC_ECLIP,CDMAC);
		//btwritemask(avms->cdmac,0,CDMAC_ECLIP,CDMAC);

		/* video line pitch, data burst, FIFO Threshold, and.. switch it on */
		reg = (avms->buffer.bytesperline & VDMAC_VPCH) |
		      VDMAC_VBS_64 |
		      VDMAC_VTHR_64 |
		      VDMAC_VEND_NONE;
		dprintk(3,KERN_DEBUG "%s: Switch on overlay: %08x\n",avms->name,reg);
		btwritemask(avms->vdmac,reg,
			    VDMAC_VBS | VDMAC_VTHR | VDMAC_VFLIP | VDMAC_VPCH,
			    VDMAC);
		saa7145_upload(avms);
		btwritemask(avms->vdmac,VDMAC_EVID,VDMAC_EVID,VDMAC);
	} else {
		/* Switch it off */
		dprintk(3,KERN_DEBUG "%s: Switch off overlay\n",avms->name);
		btwritemask(avms->vdmac,0, VDMAC_EVID, VDMAC);
	}
	saa7145_safeupload(avms);

	reg = btread(VDBA);
	dprintk(4,KERN_DEBUG "%s: reading Video DMA Base address: %08x\n",avms->name,reg);
	reg = btread(VDMAC);
	dprintk(4,KERN_DEBUG "%s: reading Video DMA Control     : %08x\n",avms->name,reg);
}

/*
 * The overlay mask has one bit for each pixel on a scan line,
 *  and the maximum window size is AVM_MAX_WIDTH * AVM_MAX_HEIGHT pixels.
 */

void
write_overlay_mask (struct file       *file,
		    struct video_clip *vp,
		    int                count)
{
	struct avmaster_fh *fh = file->private_data;
	struct avmaster *avms = fh->avms;
	int mask_line_size;
	int first_dword_offset;
	u32 *mask;
	int x, y, width, height, bitmask;
	unsigned i, j, k;
	u32 reg;

	dprintk(1,KERN_INFO "%s: Writing Overlay Mask\n",
			avms->name);
	mask_line_size = (fh->overlay_settings.width + 31 ) / 32;
	first_dword_offset = mask_line_size*32 - fh->overlay_settings.width;
	/* fill mask with one bits */
	//memset(fh->overlay_mask, ~0, mask_line_size * 4 * fh->overlay_settings.height);
	memset(fh->overlay_mask, ~0, (AVM_MAX_WIDTH +31 ) /32 * 4 * AVM_MAX_HEIGHT);
	reg = 0;

	for (i = 0; i < count; i++) {
		/* pick up local copy of clip */
		x = vp[i].x;
		y = vp[i].y;
		width = vp[i].width;
		height = vp[i].height;

		/* trim clips that extend beyond the window */
		if (x < 0) {
			width += x;
			x = 0;
		}
		if (y < 0) {
			height += y;
			y = 0;
		}
		if ((x + width) > fh->overlay_settings.width) {
			width = fh->overlay_settings.width - x;
		}
		if ((y + height) > fh->overlay_settings.height) {
			height = fh->overlay_settings.height - y;
		}

		dprintk(3,KERN_DEBUG "%s: %d vp, w= %d (%d), h= %d (%d), x = %d (%d), y = %d (%d)\n",
			avms->name,i,vp[i].width,width,vp[i].height,height,vp[i].x,x,vp[i].y,y);
		/* ignore degenerate clips */
		if (height <= 0) {
			continue;
		}
		if (width <= 0) {
			continue;
		}

		/* apply clip for each scan line */
		for (j = 0; j < height; j++) {
			/* reset bit for each pixel */
			/* this can be optimized later if need be */
			mask = fh->overlay_mask + (y + j) * mask_line_size;
			bitmask = x + first_dword_offset;
			for (k = 0; k < width; k++) {
				mask[bitmask / 32] &=
				    ~((u32) 1 << (31-(bitmask % 32)));
				bitmask++;
			}
		}
	}
}

/* Enable/Disable uncompressed memory grabbing of the 36057 */

void
saa7145_set_memgrab (struct avmaster *avms,
		     int           mode)
{
	if (mode) {
#if 0
		if (btread(ZR36057_VSSFGR) &
		    (ZR36057_VSSFGR_SnapShot | ZR36057_VSSFGR_FrameGrab))
			dprintk(1,
				KERN_WARNING
				"%s: saa7145_set_memgrab(1) with SnapShot or FrameGrab on!?\n",
				avms->name);

		/* switch on VSync interrupts */
		btwrite(IRQ_MASK, ZR36057_ISR);	// Clear Interrupts
		btor(avms->card.vsync_int, ZR36057_ICR);	// SW

		/* enable SnapShot */
		btor(ZR36057_VSSFGR_SnapShot, ZR36057_VSSFGR);

		/* Set saa7145 video front end  and enable video */
		saa7145_set_vfe(avms, avms->v4l_settings.width,
				avms->v4l_settings.height,
				avms->v4l_settings.format);

#endif
		avms->v4l_memgrab_active = 1;
	} else {
		avms->v4l_memgrab_active = 0;

		/* switch off VSync interrupts */
		//btand(~avms->card.vsync_int, ZR36057_ICR);	// SW

		/* reenable grabbing to screen if it was running */
		if (avms->v4l_overlay_active) {
			saa7145_overlay(avms, 1);
		} else {
			//btand(~ZR36057_VDCR_VidEn, ZR36057_VDCR);
			//btand(~ZR36057_VSSFGR_SnapShot, ZR36057_VSSFGR);
		}
	}
}

int
wait_grab_pending (struct avmaster *avms)
{
#if 0
	unsigned long flags;

	/* wait until all pending grabs are finished */

	if (!avms->v4l_memgrab_active)
		return 0;

	while (avms->v4l_pend_tail != avms->v4l_pend_head) {
		interruptible_sleep_on(&avms->v4l_capq);
		if (signal_pending(current))
			return -ERESTARTSYS;
	}

	spin_lock_irqsave(&avms->spinlock, flags);
	saa7145_set_memgrab(avms, 0);
	spin_unlock_irqrestore(&avms->spinlock, flags);
#endif
	return 0;
}

/*****************************************************************************
 *                                                                           *
 *  Set up the Buz-specific MJPEG part                                       *
 *                                                                           *
 *****************************************************************************/

static inline void
set_frame (struct avmaster *avms,
	   int           val)
{
	GPIO(avms, avms->card.gpio[GPIO_JPEG_FRAME], val);
}

static void
set_videobus_dir (struct avmaster *avms,
		  int           val)
{
		GPIO(avms, avms->card.gpio[GPIO_VID_DIR],
		     avms->card.gpio_pol[GPIO_VID_DIR] ? !val : val);
}

static void
init_jpeg_queue (struct avmaster *avms)
{
	int i;

	/* re-initialize DMA ring stuff */
	avms->jpg_que_head = 0;
	avms->jpg_dma_head = 0;
	avms->jpg_dma_tail = 0;
	avms->jpg_que_tail = 0;
	avms->jpg_seq_num = 0;
	avms->JPEG_error = 0;
	avms->num_errors = 0;
	avms->jpg_err_seq = 0;
	avms->jpg_err_shift = 0;
	avms->jpg_queued_num = 0;
	for (i = 0; i < avms->jpg_buffers.num_buffers; i++) {
		avms->jpg_buffers.buffer[i].state = AVM_STATE_USER;	/* nothing going on */
	}
}

static void
saa7145_set_jpg (struct avmaster          *avms,
		 enum avmaster_codec_mode  mode)
{
#if 0
	struct tvnorm *tvn;
	u32 reg;

	tvn = avms->timing;

	/* assert P_Reset, disable code transfer, deassert Active */
	btwrite(0, ZR36057_JPC);

	/* MJPEG compression mode */
	switch (mode) {

	case AVM_MODE_MOTION_COMPRESS:
	default:
		reg = ZR36057_JMC_MJPGCmpMode;
		break;

	case AVM_MODE_MOTION_DECOMPRESS:
		reg = ZR36057_JMC_MJPGExpMode;
		reg |= ZR36057_JMC_SyncMstr;
		/* RJ: The following is experimental - improves the output to screen */
		//if(avms->jpg_settings.VFIFO_FB) reg |= ZR36057_JMC_VFIFO_FB; // No, it doesn't. SM
		break;

	case AVM_MODE_STILL_COMPRESS:
		reg = ZR36057_JMC_JPGCmpMode;
		break;

	case AVM_MODE_STILL_DECOMPRESS:
		reg = ZR36057_JMC_JPGExpMode;
		break;

	}
	reg |= ZR36057_JMC_JPG;
	if (avms->jpg_settings.field_per_buff == 1)
		reg |= ZR36057_JMC_Fld_per_buff;
	btwrite(reg, ZR36057_JMC);

	/* vertical */
	btor(ZR36057_VFEVCR_VSPol, ZR36057_VFEVCR);
	reg = (6 << ZR36057_VSP_VsyncSize) |
	      (tvn->Ht << ZR36057_VSP_FrmTot);
	btwrite(reg, ZR36057_VSP);
	reg = ((avms->jpg_settings.img_y + tvn->VStart) << ZR36057_FVAP_NAY) |
	      (avms->jpg_settings.img_height << ZR36057_FVAP_PAY);
	btwrite(reg, ZR36057_FVAP);

	/* horizontal */
	if (avms->card.vfe_pol.hsync_pol)
		btor(ZR36057_VFEHCR_HSPol, ZR36057_VFEHCR);
	else
		btand(~ZR36057_VFEHCR_HSPol, ZR36057_VFEHCR);		
	reg = ((tvn->HSyncStart) << ZR36057_HSP_HsyncStart) |
	      (tvn->Wt << ZR36057_HSP_LineTot);
	btwrite(reg, ZR36057_HSP);
	reg = ((avms->jpg_settings.img_x +
	        tvn->HStart + 4) << ZR36057_FHAP_NAX) |
	      (avms->jpg_settings.img_width << ZR36057_FHAP_PAX);
	btwrite(reg, ZR36057_FHAP);

	/* field process parameters */
	if (avms->jpg_settings.odd_even)
		reg = ZR36057_FPP_Odd_Even;
	else
		reg = 0;

	btwrite(reg, ZR36057_FPP);

	/* Set proper VCLK Polarity, else colors will be wrong during playback */
	//btor(ZR36057_VFESPFR_VCLKPol, ZR36057_VFESPFR);

	/* code base address */
	reg = virt_to_bus(avms->stat_com);
	btwrite(reg, ZR36057_JCBA);

	/* FIFO threshold (FIFO is 160. double words) */
	/* NOTE: decimal values here */
	switch (mode) {

	case AVM_MODE_STILL_COMPRESS:
	case AVM_MODE_MOTION_COMPRESS:
		if (avms->card.type != AVM)
			reg = 140;
		else
			reg = 60;
		break;

	case AVM_MODE_STILL_DECOMPRESS:
	case AVM_MODE_MOTION_DECOMPRESS:
		reg = 20;
		break;

	default:
		reg = 80;
		break;

	}
	btwrite(reg, ZR36057_JCFT);
	saa7145_adjust_vfe(avms, mode);
#endif

}

u32
print_interrupts (struct avmaster *avms)
{
    int i;
    u32 isr = btread(ISR);
    u32 isrtmp=isr;
    //if (isr&ISRC_MASTER)
    if (isr)
    {
	/*
        if (isr&ISRC_VBI)
	  {
            avm.frame_count++;
	  }
        if (isr&ISRC_HBI)
            avm.frame_count++;
	*/
	for(i=0;i<32;i++) {
		if(isrtmp & 1) {
			if(i>0)
        		dprintk (4,KERN_INFO "%s Interrupt: %s\n",avms->name,irq_name[i]);
			avms->irq_count[i]++;
		}
		isrtmp >>=1;
	}
	/*
        if (isr&ISRC_VFE) dprintk (4,KERN_INFO "%s Interrupt: Video FIFO empty\n",avms->name);
        if (isr&ISRC_VFO) dprintk (4,KERN_INFO "%s Interrupt: Video FIFO overflow\n",avms->name);
        if (isr&ISRC_AFO) dprintk (4,KERN_INFO "%s Interrupt: Audio FIFO overflow\n",avms->name);
        if (isr&ISRC_CFU) dprintk (4,KERN_INFO "%s Interrupt: Clipper FIFO Underflow\n",avms->name);
        if (isr&ISRC_IIC_D) dprintk (5,KERN_DEBUG "%s Interrupt: I2C done raised\n",avms->name);
        if (isr&ISRC_IIC_E) dprintk (4,KERN_INFO "%s Interrupt: I2C error raised\n",avms->name);
        if (isr&ISRC_DEBI_I)  dprintk(4,KERN_INFO "%s Interrupt: DEBI XIRQ raised\n",avms->name);
        if (isr&ISRC_DEBI_D)  dprintk(4,KERN_INFO "%s Interrupt: DEBI done\n",avms->name);
        if (isr&ISRC_DEBI_TO) dprintk(4,KERN_INFO "%s Interrupt: DEBI timeout\n",avms->name);
        if (isr&ISRC_PPER) dprintk (4,KERN_INFO "%s Interrupt: PCI parity error during read\n",avms->name);
        if (isr&ISRC_PPEW) dprintk (4,KERN_INFO "%s Interrupt: PCI parity error during write\n",avms->name);
        if (isr&ISRC_PABO) dprintk (4,KERN_INFO "%s Interrupt: PCI access abort\n",avms->name);
	if (isr&ISRC_ARI)    dprintk (4,KERN_INFO "%s Interrupt: Audio Range\n",avms->name);
	if (isr&ISRC_UPLD)   dprintk (4,KERN_INFO "%s Interrupt: RPS Upload\n",avms->name);
	//	if (isr&ISRC_RPS_S)  printk ("\n"); 
	if (isr&ISRC_RPS_I)  dprintk (4,KERN_INFO "%s Interrupt: RPS Interrupt command\n",avms->name);
	if (isr&ISRC_RPS_TO) dprintk (4,KERN_INFO "%s Interrupt: RPS Timeout\n",avms->name);
	if (isr&ISRC_RPS_L)  dprintk (4,KERN_INFO "%s Interrupt: RPS Late Error\n",avms->name);
	if (isr&ISRC_RPS_PE) dprintk (4,KERN_INFO "%s Interrupt: RPS Page Error\n",avms->name);
	if (isr&ISRC_FIDN)   dprintk (4,KERN_INFO "%s Interrupt: Field ID Not\n",avms->name);
	if (isr&ISRC_FID)    dprintk (4,KERN_INFO "%s Interrupt: Field ID\n",avms->name);
	if (isr&ISRC_GP3)    dprintk (4,KERN_INFO "%s Interrupt: GPIO-0\n",avms->name);
	if (isr&ISRC_GP2)    dprintk (4,KERN_INFO "%s Interrupt: GPIO-1\n",avms->name);
	if (isr&ISRC_GP1)    dprintk (4,KERN_INFO "%s Interrupt: GPIO-2\n",avms->name);
	if (isr&ISRC_GP0)    dprintk (4,KERN_INFO "%s Interrupt: GPIO-3\n",avms->name);
	if (isr&ISRC_EC2)    dprintk (4,KERN_INFO "%s Interrupt: Event Counter 2\n",avms->name);
	if (isr&ISRC_EC1)    dprintk (4,KERN_INFO "%s Interrupt: Event Counter 1\n",avms->name);
	*/
	if (isr & (ISRC_RPS_I | ISRC_RPS_TO | ISRC_RPS_L |ISRC_RPS_PE))
	  {
	    //avm.saa7145_reg[0x3C/4] &= ~1;
	    dprintk (4,KERN_INFO "%s Interrupt: RPS stopped\n",avms->name);
	  }
    }
    return isr;
}

void
clear_interrupt_counters (struct avmaster *avms)
{
	avms->intr_counter_GIRQ1 = 0;
	avms->intr_counter_GIRQ0 = 0;
	avms->intr_counter_CodRepIRQ = 0;
	avms->intr_counter_JPEGRepIRQ = 0;
	avms->field_counter = 0;
	avms->IRQ1_in = 0;
	avms->IRQ1_out = 0;
	avms->JPEG_in = 0;
	avms->JPEG_out = 0;
	avms->JPEG_0 = 0;
	avms->JPEG_1 = 0;
	avms->END_event_missed = 0;
	avms->JPEG_missed = 0;
	avms->JPEG_max_missed = 0;
	avms->JPEG_min_missed = 0x7fffffff;
}

#if 0
static u32
count_reset_interrupt (struct avmaster *avms)
{

	u32 isr;
	isr=btread(ISR);
#if 0
	if ((isr = btread(ZR36057_ISR) & 0x78000000)) {
		if (isr & ZR36057_ISR_GIRQ1) {
			btwrite(ZR36057_ISR_GIRQ1, ZR36057_ISR);
			avms->intr_counter_GIRQ1++;
		}
		if (isr & ZR36057_ISR_GIRQ0) {
			btwrite(ZR36057_ISR_GIRQ0, ZR36057_ISR);
			avms->intr_counter_GIRQ0++;
		}
		if (isr & ZR36057_ISR_CodRepIRQ) {
			btwrite(ZR36057_ISR_CodRepIRQ, ZR36057_ISR);
			avms->intr_counter_CodRepIRQ++;
		}
		if (isr & ZR36057_ISR_JPEGRepIRQ) {
			btwrite(ZR36057_ISR_JPEGRepIRQ, ZR36057_ISR);
			avms->intr_counter_JPEGRepIRQ++;
		}
	}
#endif
	return isr;
}
#endif

/* hack */
extern void zr36015_write (struct videocodec *codec,
			   u16                reg,
			   u32                val);

void
jpeg_start (struct avmaster *avms)
{
#if 0
	int reg;
	avms->frame_num = 0;

	/* deassert P_reset, disable code transfer, deassert Active */
	btwrite(ZR36057_JPC_P_Reset, ZR36057_JPC);
	/* stop flushing the internal code buffer */
	btand(~ZR36057_MCTCR_CFlush, ZR36057_MCTCR);
	/* enable code transfer */
	btor(ZR36057_JPC_CodTrnsEn, ZR36057_JPC);

	/* clear IRQs */
	btwrite(IRQ_MASK, ZR36057_ISR);
	/* enable the JPEG IRQs */
	btwrite(avms->card.jpeg_int |
			ZR36057_ICR_JPEGRepIRQ |
			ZR36057_ICR_IntPinEn,
		ZR36057_ICR);

	set_frame(avms, 0);	// \FRAME

	/* set the JPEG codec guest ID */
	reg = (avms->card.gpcs[1] << ZR36057_JCGI_JPEGuestID) |
	       (0 << ZR36057_JCGI_JPEGuestReg);
	btwrite(reg, ZR36057_JCGI);

	if (avms->card.video_vfe == CODEC_TYPE_ZR36015 &&
	    avms->card.video_codec == CODEC_TYPE_ZR36050) {
		/* Enable processing on the ZR36015 */
		if (avms->vfe)
			zr360136015te(avms->vfe, 0, 1);

		/* load the address of the GO register in the ZR36050 latch */
		post_office_write(avms, 0, 0, 0);
	}

	/* assert Active */
	btor(ZR36057_JPC_Active, ZR36057_JPC);

	/* enable the Go generation */
	btor(ZR36057_JMC_Go_en, ZR36057_JMC);
	udelay(30);

	set_frame(avms, 1);	// /FRAME

#endif
	dprintk(3, KERN_DEBUG "%s: jpeg_start\n", avms->name);
}

void
saa7145_enable_jpg (struct avmaster          *avms,
		    enum avmaster_codec_mode  mode)
{
	static int zero = 0;
	static int one = 1;
	struct vfe_settings cap;
	int field_size =
	    avms->jpg_buffers.buffer_size / avms->jpg_settings.field_per_buff;

	avms->codec_mode = mode;

	cap.x = avms->jpg_settings.img_x;
	cap.y = avms->jpg_settings.img_y;
	cap.width = avms->jpg_settings.img_width;
	cap.height = avms->jpg_settings.img_height;
	cap.decimation =
	    avms->jpg_settings.HorDcm | (avms->jpg_settings.VerDcm << 8);
	cap.quality = avms->jpg_settings.jpg_comp.quality;

	switch (mode) {

	case AVM_MODE_MOTION_COMPRESS:
		/* In motion compress mode, the decoder output must be enabled, and
		 * the video bus direction set to input.
		 */
		set_videobus_dir(avms, 0);
		decoder_command(avms, DECODER_ENABLE_OUTPUT, &one);
		encoder_command(avms, ENCODER_SET_INPUT, &zero);

		/* Take the JPEG codec and the VFE out of sleep */
		jpeg_codec_sleep(avms, 0);
		/* Setup the VFE */
		if (avms->vfe) {
			avms->vfe->control(avms->vfe, CODEC_S_JPEG_TDS_BYTE,
					 sizeof(int), &field_size);
			avms->vfe->set_video(avms->vfe, avms->timing, &cap,
					   &avms->card.vfe_pol);
			avms->vfe->set_mode(avms->vfe, CODEC_DO_COMPRESSION);
		}
		/* Setup the JPEG codec */
		avms->codec->control(avms->codec, CODEC_S_JPEG_TDS_BYTE,
				   sizeof(int), &field_size);
		avms->codec->set_video(avms->codec, avms->timing, &cap,
				     &avms->card.vfe_pol);
		avms->codec->set_mode(avms->codec, CODEC_DO_COMPRESSION);

		init_jpeg_queue(avms);
		saa7145_set_jpg(avms, mode);	// \P_Reset, ... Video param, FIFO

		clear_interrupt_counters(avms);
		dprintk(2, KERN_INFO "%s: enable_jpg(MOTION_COMPRESS)\n",
			avms->name);
		break;

	case AVM_MODE_MOTION_DECOMPRESS:
		/* In motion decompression mode, the decoder output must be disabled, and
		 * the video bus direction set to output.
		 */
		decoder_command(avms, DECODER_ENABLE_OUTPUT, &zero);
		set_videobus_dir(avms, 1);
		encoder_command(avms, ENCODER_SET_INPUT, &one);

		/* Take the JPEG codec and the VFE out of sleep */
		jpeg_codec_sleep(avms, 0);
		/* Setup the VFE */
		if (avms->vfe) {
			avms->vfe->set_video(avms->vfe, avms->timing, &cap,
					   &avms->card.vfe_pol);
			avms->vfe->set_mode(avms->vfe, CODEC_DO_EXPANSION);
		}
		/* Setup the JPEG codec */
		avms->codec->set_video(avms->codec, avms->timing, &cap,
				     &avms->card.vfe_pol);
		avms->codec->set_mode(avms->codec, CODEC_DO_EXPANSION);

		init_jpeg_queue(avms);
		saa7145_set_jpg(avms, mode);	// \P_Reset, ... Video param, FIFO

		clear_interrupt_counters(avms);
		dprintk(2, KERN_INFO "%s: enable_jpg(MOTION_DECOMPRESS)\n",
			avms->name);
		break;

	case AVM_MODE_IDLE:
	default:
		/* shut down processing */
		/*
		btand(~(avms->card.jpeg_int | ZR36057_ICR_JPEGRepIRQ),
		      ZR36057_ICR);
		btwrite(avms->card.jpeg_int | ZR36057_ICR_JPEGRepIRQ,
			ZR36057_ISR);
		btand(~ZR36057_JMC_Go_en, ZR36057_JMC);	// \Go_en
		*/

		current->state = TASK_UNINTERRUPTIBLE;
		schedule_timeout(HZ / 20);

		set_videobus_dir(avms, 0);
		set_frame(avms, 1);	// /FRAME
		/*
		btor(ZR36057_MCTCR_CFlush, ZR36057_MCTCR);	// /CFlush
		btwrite(0, ZR36057_JPC);	// \P_Reset,\CodTrnsEn,\Active
		btand(~ZR36057_JMC_VFIFO_FB, ZR36057_JMC);
		btand(~ZR36057_JMC_SyncMstr, ZR36057_JMC);
		*/
		jpeg_codec_reset(avms);
		jpeg_codec_sleep(avms, 1);
		saa7145_adjust_vfe(avms, mode);

		decoder_command(avms, DECODER_ENABLE_OUTPUT, &one);
		encoder_command(avms, ENCODER_SET_INPUT, &zero);

		dprintk(2, KERN_INFO "%s: enable_jpg(IDLE)\n", avms->name);
		break;

	}
}

#if 0
static void
error_handler (struct avmaster *avms,
	       u32           astat,
	       u32           stat)
{
	/* This is JPEG error handling part */
	if ((avms->codec_mode != AVM_MODE_MOTION_COMPRESS) &&
	    (avms->codec_mode != AVM_MODE_MOTION_DECOMPRESS)) {
		//dprintk(1, KERN_ERR "%s: Internal error: error handling request in mode %d\n", avms->name, avms->codec_mode);
		return;
	}

	if ((stat & 1) == 0 &&
	    avms->codec_mode == AVM_MODE_MOTION_COMPRESS &&
	    avms->jpg_dma_tail - avms->jpg_que_tail >=
	     avms->jpg_buffers.num_buffers) {
		/* No free buffers... */
		avmaster_reap_stat_com(avms);
		avmaster_feed_stat_com(avms);
		wake_up_interruptible(&avms->jpg_capq);
		avms->JPEG_missed = 0;
		return;
	}

	if (avms->JPEG_error != 1) {
		/*
		 * First entry: error just happened during normal operation
		 * 
		 * In AVM_MODE_MOTION_COMPRESS:
		 * 
		 * Possible glitch in TV signal. In this case we should
		 * stop the codec and wait for good quality signal before
		 * restarting it to avoid further problems
		 * 
		 * In AVM_MODE_MOTION_DECOMPRESS:
		 * 
		 * Bad JPEG frame: we have to mark it as processed (codec crashed
		 * and was not able to do it itself), and to remove it from queue.
		 */
		btand(~ZR36057_JMC_Go_en, ZR36057_JMC);
		udelay(1);
		stat = stat | (post_office_read(avms, 7, 0) & 3) << 8;
		btwrite(0, ZR36057_JPC);
		btor(ZR36057_MCTCR_CFlush, ZR36057_MCTCR);
		jpeg_codec_reset(avms);
		jpeg_codec_sleep(avms, 1);
		avms->JPEG_error = 1;
		avms->num_errors++;

		/* Report error */
		if (debug > 1 && avms->num_errors <= 8) {
			long frame;
			frame =
			    avms->jpg_pend[avms->jpg_dma_tail & AVM_MASK_FRAME];
			printk(KERN_ERR
			       "%s: JPEG error stat=0x%08x(0x%08x) queue_state=%ld/%ld/%ld/%ld seq=%ld frame=%ld. Codec stopped. ",
			       avms->name, stat, avms->last_isr,
			       avms->jpg_que_tail, avms->jpg_dma_tail,
			       avms->jpg_dma_head, avms->jpg_que_head,
			       avms->jpg_seq_num, frame);
			printk("stat_com frames:");
			{
				int i, j;
				for (j = 0; j < AVM_NUM_STAT_COM; j++) {
					for (i = 0;
					     i < avms->jpg_buffers.num_buffers;
					     i++) {
						if (avms->stat_com[j] ==
						    avms->jpg_buffers.
						    buffer[i].
						    frag_tab_bus) {
							printk("% d->%d",
							       j, i);
						}
					}
				}
				printk("\n");
			}
		}

		/* Find an entry in stat_com and rotate contents */
		{
			int i;

			if (avms->jpg_settings.TmpDcm == 1)
				i = (avms->jpg_dma_tail -
				     avms->jpg_err_shift) & AVM_MASK_STAT_COM;
			else
				i = ((avms->jpg_dma_tail -
				      avms->jpg_err_shift) & 1) * 2;
			if (avms->codec_mode == AVM_MODE_MOTION_DECOMPRESS) {
				/* Mimic zr36067 operation */
				avms->stat_com[i] |= 1;
				if (avms->jpg_settings.TmpDcm != 1)
					avms->stat_com[i + 1] |= 1;
				/* Refill */
				avmaster_reap_stat_com(avms);
				avmaster_feed_stat_com(avms);
				wake_up_interruptible(&avms->jpg_capq);
				/* Find an entry in stat_com again after refill */
				if (avms->jpg_settings.TmpDcm == 1)
					i = (avms->jpg_dma_tail -
					     avms->jpg_err_shift) &
					    AVM_MASK_STAT_COM;
				else
					i = ((avms->jpg_dma_tail -
					      avms->jpg_err_shift) & 1) * 2;
			}
			if (i) {
				/* Rotate stat_comm entries to make current entry first */
				int j;
				u32 bus_addr[AVM_NUM_STAT_COM];

				memcpy(bus_addr, avms->stat_com,
				       sizeof(bus_addr));
				for (j = 0; j < AVM_NUM_STAT_COM; j++) {
					avms->stat_com[j] =
					    bus_addr[(i + j) &
						     AVM_MASK_STAT_COM];
				}
				avms->jpg_err_shift += i;
				avms->jpg_err_shift &= AVM_MASK_STAT_COM;
			}
			if (avms->codec_mode == AVM_MODE_MOTION_COMPRESS)
				avms->jpg_err_seq = avms->jpg_seq_num;	/* + 1; */
		}
	}

	/* Now the stat_comm buffer is ready for restart */
	do {
		int status, mode;

		if (avms->codec_mode == AVM_MODE_MOTION_COMPRESS) {
			decoder_command(avms, DECODER_GET_STATUS, &status);
			mode = CODEC_DO_COMPRESSION;
		} else {
			status = 0;
			mode = CODEC_DO_EXPANSION;
		}
		if (avms->codec_mode == AVM_MODE_MOTION_DECOMPRESS ||
		    (status & DECODER_STATUS_GOOD)) {
			/********** RESTART code *************/
			jpeg_codec_reset(avms);
			avms->codec->set_mode(avms->codec, mode);
			saa7145_set_jpg(avms, avms->codec_mode);
			jpeg_start(avms);

			if (avms->num_errors <= 8)
				dprintk(2, KERN_INFO "%s: Restart\n",
					avms->name);

			avms->JPEG_missed = 0;
			avms->JPEG_error = 2;
			/********** End RESTART code ***********/
		}
	} while (0);
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
void
#else
irqreturn_t
#endif
avmaster_irq (int             irq,
	   void           *dev_id,
	   struct pt_regs *regs)
{
	//u32 stat, astat;
	int count;
	struct avmaster *avms;
	//unsigned long flags;
	avms = (struct avmaster *) dev_id;
	u32 isr;
	isr=print_interrupts(avms);
	if(isr) {
		// Check for RPS Upload and set the flag
		if(isr & ISRC_UPLD) upload_done=1;
        	/* all should be cleared to prevent infinite loops */
		btwrite(ISRC_MASTER | isr,ISR);
        }
	count = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	return;
#else
	return IRQ_HANDLED;
#endif
#if 0

	u32 stat, astat;
	int count;
	struct avmaster *avms;
	unsigned long flags;

	avms = (struct avmaster *) dev_id;
	count = 0;

	if (avms->testing) {
		/* Testing interrupts */
		spin_lock_irqsave(&avms->spinlock, flags);
		while ((stat = count_reset_interrupt(avms))) {
			if (count++ > 100) {
				btand(~ZR36057_ICR_IntPinEn, ZR36057_ICR);
				dprintk(1,
					KERN_ERR
					"%s: IRQ lockup while testing, isr=0x%08x, cleared int mask\n",
					avms->name, stat);
				wake_up_interruptible(&avms->test_q);
			}
		}
		avms->last_isr = stat;
		spin_unlock_irqrestore(&avms->spinlock, flags);
		return;
	}

	spin_lock_irqsave(&avms->spinlock, flags);
	while (1) {
		/* get/clear interrupt status bits */
		stat = count_reset_interrupt(avms);
		astat = stat & IRQ_MASK;
		if (!astat) {
			break;
		}
		dprintk(4,
			KERN_DEBUG
			"avmaster_irq: astat: 0x%08x, mask: 0x%08x\n",
			astat, btread(ZR36057_ICR));
		if (astat & avms->card.vsync_int) {	// SW

			if (avms->codec_mode == AVM_MODE_MOTION_DECOMPRESS ||
			    avms->codec_mode == AVM_MODE_MOTION_COMPRESS) {
				/* count missed interrupts */
				avms->JPEG_missed++;
			}
			//post_office_read(avms,1,0);
			/* Interrupts may still happen when
			 * avms->v4l_memgrab_active is switched off.
			 * We simply ignore them */

			if (avms->v4l_memgrab_active) {

				/* A lot more checks should be here ... */
				if ((btread(ZR36057_VSSFGR) &
				     ZR36057_VSSFGR_SnapShot) == 0)
					dprintk(1,
						KERN_WARNING
						"%s: BuzIRQ with SnapShot off ???\n",
						avms->name);

				if (avms->v4l_grab_frame != NO_GRAB_ACTIVE) {
					/* There is a grab on a frame going on, check if it has finished */

					if ((btread(ZR36057_VSSFGR) &
					     ZR36057_VSSFGR_FrameGrab) ==
					    0) {
						/* it is finished, notify the user */

						avms->v4l_buffers.buffer[avms->
								       v4l_grab_frame].
						    state = AVM_STATE_DONE;
						avms->v4l_buffers.buffer[avms->
								       v4l_grab_frame].
						    bs.seq =
						    avms->v4l_grab_seq;
						do_gettimeofday(&avms->
								v4l_buffers.
								buffer[avms->
								       v4l_grab_frame].
								bs.
								timestamp);
						avms->v4l_grab_frame =
						    NO_GRAB_ACTIVE;
						avms->v4l_pend_tail++;
					}
				}

				if (avms->v4l_grab_frame == NO_GRAB_ACTIVE)
					wake_up_interruptible(&avms->
							      v4l_capq);

				/* Check if there is another grab queued */

				if (avms->v4l_grab_frame == NO_GRAB_ACTIVE &&
				    avms->v4l_pend_tail !=
				    avms->v4l_pend_head) {

					int frame =
					    avms->v4l_pend[avms->
							 v4l_pend_tail &
							 V4L_MASK_FRAME];
					u32 reg;

					avms->v4l_grab_frame = frame;

					/* Set saa7145 video front end and enable video */

					/* Buffer address */

					reg =
					    avms->v4l_buffers.buffer[frame].
					    fbuffer_bus;
					btwrite(reg, ZR36057_VDTR);
					if (avms->v4l_settings.height >
					    AVM_MAX_HEIGHT / 2)
						reg +=
						    avms->v4l_settings.
						    bytesperline;
					btwrite(reg, ZR36057_VDBR);

					/* video stride, status, and frame grab register */
					reg = 0;
					if (avms->v4l_settings.height >
					    AVM_MAX_HEIGHT / 2)
						reg +=
						    avms->v4l_settings.
						    bytesperline;
					reg =
					    (reg <<
					     ZR36057_VSSFGR_DispStride);
					reg |= ZR36057_VSSFGR_VidOvf;
					reg |= ZR36057_VSSFGR_SnapShot;
					reg |= ZR36057_VSSFGR_FrameGrab;
					btwrite(reg, ZR36057_VSSFGR);

					btor(ZR36057_VDCR_VidEn,
					     ZR36057_VDCR);
				}
			}

			/* even if we don't grab, we do want to increment
			 * the sequence counter to see lost frames */
			avms->v4l_grab_seq++;
		}
#if (IRQ_MASK & ZR36057_ISR_CodRepIRQ)
		if (astat & ZR36057_ISR_CodRepIRQ) {
			avms->intr_counter_CodRepIRQ++;
			IDEBUG(printk
			       (KERN_DEBUG "%s: ZR36057_ISR_CodRepIRQ\n",
				avms->name));
			btand(~ZR36057_ICR_CodRepIRQ, ZR36057_ICR);
		}
#endif				/* (IRQ_MASK & ZR36057_ISR_CodRepIRQ) */

#if (IRQ_MASK & ZR36057_ISR_JPEGRepIRQ)
		if (astat & ZR36057_ISR_JPEGRepIRQ) {

			if (avms->codec_mode == AVM_MODE_MOTION_DECOMPRESS ||
			    avms->codec_mode == AVM_MODE_MOTION_COMPRESS) {
				if (debug > 1 &&
				    (!avms->frame_num || avms->JPEG_error)) {
					printk(KERN_INFO
					       "%s: first frame ready: state=0x%08x odd_even=%d field_per_buff=%d delay=%d\n",
					       avms->name, stat,
					       avms->jpg_settings.odd_even,
					       avms->jpg_settings.
					       field_per_buff,
					       avms->JPEG_missed);
					{
						char sc[] = "0000";
						char sv[5];
						int i;
						strcpy(sv, sc);
						for (i = 0; i < 4; i++) {
							if (avms->
							    stat_com[i] &
							    1)
								sv[i] =
								    '1';
						}
						sv[4] = 0;
						printk(KERN_INFO
						       "%s: stat_com=%s queue_state=%ld/%ld/%ld/%ld\n",
						       avms->name, sv,
						       avms->jpg_que_tail,
						       avms->jpg_dma_tail,
						       avms->jpg_dma_head,
						       avms->jpg_que_head);
					}
				} else {
					if (avms->JPEG_missed > avms->JPEG_max_missed)	// Get statistics
						avms->JPEG_max_missed =
						    avms->JPEG_missed;
					if (avms->JPEG_missed <
					    avms->JPEG_min_missed)
						avms->JPEG_min_missed =
						    avms->JPEG_missed;
				}

				if (debug > 2 && avms->frame_num < 6) {
					int i;
					printk("%s: seq=%ld stat_com:",
					       avms->name, avms->jpg_seq_num);
					for (i = 0; i < 4; i++) {
						printk(" %08x",
						       avms->stat_com[i]);
					}
					printk("\n");
				}
				avms->frame_num++;
				avms->JPEG_missed = 0;
				avms->JPEG_error = 0;
				avmaster_reap_stat_com(avms);
				avmaster_feed_stat_com(avms);
				wake_up_interruptible(&avms->jpg_capq);
			} /*else {
			      dprintk(1,
					KERN_ERR
					"%s: JPEG interrupt while not in motion (de)compress mode!\n",
					avms->name);
			}*/
		}
#endif				/* (IRQ_MASK & ZR36057_ISR_JPEGRepIRQ) */

		/* DATERR, too many fields missed, error processing */
		if ((astat & avms->card.jpeg_int) ||
		    avms->JPEG_missed > 25 ||
		    avms->JPEG_error == 1	||
		    ((avms->codec_mode == AVM_MODE_MOTION_DECOMPRESS) &&
		     (avms->frame_num & (avms->JPEG_missed >
				       avms->jpg_settings.field_per_buff)))) {
			error_handler(avms, astat, stat);
		}

		count++;
		if (count > 10) {
			dprintk(2, KERN_WARNING "%s: irq loop %d\n",
				avms->name, count);
			if (count > 20) {
				btand(~ZR36057_ICR_IntPinEn, ZR36057_ICR);
				dprintk(2,
					KERN_ERR
					"%s: IRQ lockup, cleared int mask\n",
					avms->name);
				break;
			}
		}
		avms->last_isr = stat;
	}
	spin_unlock_irqrestore(&avms->spinlock, flags);
#endif
}

void
avmaster_set_pci_master (struct avmaster *avms,
		      int           set_master)
{
	u16 command;
	if (set_master) {
		pci_set_master(avms->pci_dev);
	} else {
		pci_read_config_word(avms->pci_dev, PCI_COMMAND, &command);
		command &= ~PCI_COMMAND_MASTER;
		pci_write_config_word(avms->pci_dev, PCI_COMMAND, command);
	}
}

void
avmaster_init_hardware (struct avmaster *avms)
{
	int j, zero = 0;

	/* Enable bus-mastering */
	avmaster_set_pci_master(avms, 1);

	/* Initialize the board */
	if (avms->card.init) {
		avms->card.init(avms);
	}

	j = avms->card.input[avms->input].muxsel;

	decoder_command(avms, 0, NULL);
	decoder_command(avms, DECODER_SET_NORM, &avms->norm);
	decoder_command(avms, DECODER_SET_INPUT, &j);

	encoder_command(avms, 0, NULL);
	encoder_command(avms, ENCODER_SET_NORM, &avms->norm);
	encoder_command(avms, ENCODER_SET_INPUT, &zero);

	/* toggle JPEG codec sleep to sync PLL */
	jpeg_codec_sleep(avms, 1);
	jpeg_codec_sleep(avms, 0);

	/* set individual interrupt enables (without GIRQ1)
	 * but don't global enable until avmaster_open() */

	//btwrite(IRQ_MASK & ~ZR36057_ISR_GIRQ1, ZR36057_ICR);  // SW
	// It looks like using only JPEGRepIRQEn is not always reliable,
	// may be when JPEG codec crashes it won't generate IRQ? So,
	 /*CP*/			//        btwrite(IRQ_MASK, ZR36057_ICR); // Enable Vsync interrupts too. SM    WHY ? LP

	saa7145_init_vfe(avms);

	saa7145_enable_jpg(avms, AVM_MODE_IDLE);

	btwrite(btread(ISR), ISR);	// Clears interrupts
	btwrite(IRQ_MASK, IER);	// Enable interrupts
	//btwrite(0xffffffff, IMR);	// Testing interrupts
}

void
saa7145_restart (struct avmaster *avms)
{
	/* Disable Video DMA, Audio DMA, Clipper, RPS, DEBI */
	btwritemask(avms->vdmac,0,VDMAC_EVID,VDMAC);
	saa7145_upload(avms);

	btwrite(0,ADMAC);
	btwrite(0,RPSPTR);
	btwrite(DEBICFG_INMO|DEBICFG_XIRQ_WAIT|DEBICFG_SLAVE16,DEBICFG);
	btwritemask(avms->cdmac,0,CDMAC_ECLIP,CDMAC);
	saa7145_upload(avms);
#if 0
	btwrite(0, ZR36057_SPGPPCR);
	mdelay(1);
	btor(ZR36057_SPGPPCR_SoftReset, ZR36057_SPGPPCR);
	mdelay(1);

	/* assert P_Reset */
	btwrite(0, ZR36057_JPC);
	/* set up GPIO direction - all output */
	btwrite(ZR36057_SPGPPCR_SoftReset | 0, ZR36057_SPGPPCR);

	/* set up GPIO pins and guest bus timing */
	btwrite((0x81 << 24) | 0x8888, ZR36057_GPPGCR1);
#endif
}

/*
 * initialize video front end
 */

void
saa7145_init_vfe (struct avmaster *avms)
{
#if 0
	u32 reg;
	reg = btread(ZR36057_VFESPFR);
	reg |= ZR36057_VFESPFR_LittleEndian;
	reg &= ~ZR36057_VFESPFR_VCLKPol;
	reg |= ZR36057_VFESPFR_ExtFl;
	reg |= ZR36057_VFESPFR_TopField;
	btwrite(reg, ZR36057_VFESPFR);
	reg = btread(ZR36057_VDCR);
	if (pci_pci_problems & PCIPCI_TRITON)
		// || avms->revision < 1) // Revision 1 has also Triton support
		reg &= ~ZR36057_VDCR_Triton;
	else
		reg |= ZR36057_VDCR_Triton;
	btwrite(reg, ZR36057_VDCR);
#endif
}

/*
 * Interface to decoder and encoder chips using i2c bus
 */

int
decoder_command (struct avmaster *avms,
		 int           cmd,
		 void         *data)
{
	if (avms->decoder == NULL)
		return -EIO;

	/*
	if (avms->card.type == AVMaster &&
	    (cmd == DECODER_SET_NORM || DECODER_SET_INPUT)) {
		int res;
		// Bt819 needs to reset its FIFO buffer using #FRST pin and
		// LML33 card uses GPIO(7) for that.
		GPIO(avms, 7, 0);
		res = avms->decoder->driver->command(avms->decoder, cmd, data);
		// Pull #FRST high.
		GPIO(avms, 7, 1);
		return res;
	} else
	*/
		return avms->decoder->driver->command(avms->decoder, cmd,
						    data);
}

int
encoder_command (struct avmaster *avms,
		 int           cmd,
		 void         *data)
{
	if (avms->encoder == NULL)
		return -1;

	return avms->encoder->driver->command(avms->encoder, cmd, data);
}
