/*
 * SAA7145 PCI controller driver, for the Fast AV Master
 *
 * This part handles the procFS entries (/proc/avms[%d])
 * 
 * Copyright (C) 2000 Serguei Miridonov <mirsev@cicese.mx>
 *
 * Currently maintained by:
 *   Luca Bonissi <lucabon@interfree.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/vmalloc.h>

#include <linux/proc_fs.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <linux/videodev.h>
#include <linux/spinlock.h>
#include <linux/sem.h>

#include <linux/ctype.h>
#include <asm/io.h>

#include "videocodec.h"
#include "avmaster.h"
#include "avmaster_procfs.h"

extern int debug;

#define dprintk(num, format, args...) \
	do { \
		if (debug >= num) \
			printk(format, ##args); \
	} while (0)

#ifdef CONFIG_PROC_FS
struct procfs_params_avmaster {
	char *name;
	short reg;
	u32 mask;
	short bit;
};

static const struct procfs_params_avmaster av[] = {
	{"HSPol", 0x000, 1, 30},
	{"HStart", 0x000, 0x3ff, 10},
	{"HEnd", 0x000, 0x3ff, 0},

	{"VSPol", 0x004, 1, 30},
	{"VStart", 0x004, 0x3ff, 10},
	{"VEnd", 0x004, 0x3ff, 0},

	{"ExtFl", 0x008, 1, 26},
	{"TopField", 0x008, 1, 25},
	{"VCLKPol", 0x008, 1, 24},
	{"DupFld", 0x008, 1, 20},
	{"LittleEndian", 0x008, 1, 0},

	{"HsyncStart", 0x10c, 0xffff, 16},
	{"LineTot", 0x10c, 0xffff, 0},

	{"NAX", 0x110, 0xffff, 16},
	{"PAX", 0x110, 0xffff, 0},

	{"NAY", 0x114, 0xffff, 16},
	{"PAY", 0x114, 0xffff, 0},

	/* {"",,,}, */

	{NULL, 0, 0, 0},
};

struct procfs_io {
	char *buffer;
	char *end;
	int neof;
	int count;
	int count_current;
};

static void
setparam (struct avmaster *avms,
	  char         *name,
	  char         *sval)
{
	int i = 0, reg0, reg, val;

	while (av[i].name != NULL) {
		if (!strncmp(name, av[i].name, strlen(av[i].name))) {
			reg = reg0 = btread(av[i].reg);
			reg &= ~(av[i].mask << av[i].bit);
			if (!isdigit(sval[0]))
				break;
			val = simple_strtoul(sval, NULL, 0);
			if ((val & ~av[i].mask))
				break;
			reg |= (val & av[i].mask) << av[i].bit;
			dprintk(4,
				KERN_INFO
				"%s: setparam: setting SAA7145 register 0x%03x: 0x%08x=>0x%08x %s=%d\n",
				avms->name, av[i].reg, reg0, reg,
				av[i].name, val);
			btwrite(reg, av[i].reg);
			break;
		}
		i++;
	}
}

static int
print_procfs (struct procfs_io *io,
	      const char       *fmt,
	      ...)
{
	va_list args;
	int i;

	if (io->buffer >= io->end) {
		io->neof++;
		return 0;
	}
	if (io->count > io->count_current++)
		return 0;
	va_start(args, fmt);
	i = vsprintf(io->buffer, fmt, args);
	io->buffer += i;
	va_end(args);
	return i;
}

char *regname[] = {"Video DMA Base Address",
		   "Video DMA Prot Address",
		   "Video DMA Control",
		   "Video Port Control",
		   "Horizontal Window",
		   "Vertical   Window",
		   "Filter and Scaler",
		   "Vertical Phase",
		   "Clip Mask DMA Address",
		   "Clip Data Format",
		   "Clipper DMA Control",
		   "Line Counter Register",
		   "Audio DMA Base Address",
		   "Audio DMA Prot Address",
		   "Audio DMA Control",
		   "RPS Pointer Address",
		   "RPS Page Register",
		   "RPS Time Out",
		   "Interrupt Status",
		   "Interrupt Enable",
		   "Interrupt Monitor",
		   "Event Counter",
		   "Event Counter Thresh.",
		   "Event Counter Control",
		   "GPIO Control",
		   "DEBI Configuration",
		   "DEBI Data/Address",
		   "DEBI Command",
		   "I2C Transfer Control",
		   "I2C Status and Clock",
		   "Utility Register",
		   "Reserved"};

static void
avmaster_procfs_output (struct procfs_io *io,
		     void             *data)
{
	int i;
	struct avmaster *avms;
	avms = (struct avmaster *) data;

	print_procfs(io, "SAA7145 registers:\n");
	for (i = 0; i <= (0x7C>>2); i++) {
		print_procfs(io, "%03X %22s = %08X\n", i, regname[i], btread((i<<2)));
	};
}

static int
avmaster_read_proc (char  *buffer,
		 char **start,
		 off_t  offset,
		 int    size,
		 int   *eof,
		 void  *data)
{
	struct procfs_io io;
	int nbytes;

	io.buffer = buffer;
	io.end = buffer + size - 128;	// Just to make it a little bit safer
	io.count = offset;
	io.count_current = 0;
	io.neof = 0;
	avmaster_procfs_output(&io, data);
	*start = (char *) (io.count_current - io.count);
	nbytes = (int) (io.buffer - buffer);
	*eof = !io.neof;
	return nbytes;

	return 0;
}

static int
avmaster_write_proc (struct file   *file,
		  const char    *buffer,
		  unsigned long  count,
		  void          *data)
{
	char *string, *sp;
	char *line, *ldelim, *varname, *svar, *tdelim;
	struct avmaster *avms;

	avms = (struct avmaster *) data;

	string = sp = vmalloc(count + 1);
	if (!string) {
		dprintk(1,
			KERN_ERR
			"%s: write_proc: can not allocate memory\n",
			avms->name);
		return -ENOMEM;
	}
	memcpy(string, buffer, count);
	string[count] = 0;
	dprintk(4, KERN_INFO "%s: write_proc: name=%s count=%lu data=%x\n",
		avms->name, file->f_dentry->d_name.name, count, (int) data);
	ldelim = " \t\n";
	tdelim = "=";
	line = strpbrk(sp, ldelim);
	while (line) {
		*line = 0;
		svar = strpbrk(sp, tdelim);
		if (svar) {
			*svar = 0;
			varname = sp;
			svar++;
			setparam(avms, varname, svar);
		}
		sp = line + 1;
		line = strpbrk(sp, ldelim);
	}
	vfree(string);

	return count;
}
#endif

int
avmaster_proc_init (struct avmaster *avms)
{
#ifdef CONFIG_PROC_FS
	char name[8];
	snprintf(name, 7, "avms%d", avms->id);
	if ((avms->avmaster_proc = create_proc_entry(name, 0, 0))) {
		avms->avmaster_proc->read_proc = avmaster_read_proc;
		avms->avmaster_proc->write_proc = avmaster_write_proc;
		avms->avmaster_proc->data = avms;
		avms->avmaster_proc->owner = THIS_MODULE;
		dprintk(2,
			KERN_INFO
			"%s: procfs entry /proc/%s allocated. data=%p\n",
			avms->name, name, avms->avmaster_proc->data);
	} else {
		dprintk(1, KERN_ERR "%s: Unable to initialise /proc/%s\n",
			avms->name, name);
		return 1;
	}
#endif
	return 0;
}

void
avmaster_proc_cleanup (struct avmaster *avms)
{
#ifdef CONFIG_PROC_FS
	char name[8];
	snprintf(name, 7, "avms%d", avms->id);
	if (avms->avmaster_proc) {
		remove_proc_entry(name, 0);
	}
	avms->avmaster_proc = NULL;
#endif
}
