/*
 * Zoran ZR36015 basic configuration functions
 *
 * Copyright (C) 2001 Wolfgang Scherr <scherr@net4you.at>
 * Copyright (C) 2004 Luca Bonissi <lucabon@interfree.it>
 *
 * $Id: zr36015.c,v 1.1.2.8 2003/03/26 20:30:21 rbultje Exp $
 *
 * ------------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ------------------------------------------------------------------------
 */

#define ZR015_VERSION "v0.7"

#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/delay.h>

#include <linux/types.h>
#include <linux/wait.h>

/* includes for structures and defines regarding video 
   #include<linux/videodev.h> */

/* I/O commands, error codes */
#include<asm/io.h>
//#include<errno.h>

/* v4l  API */
#include<linux/videodev.h>

/* headerfile of this module */
#include"zr36015.h"

/* codec io API */
#include"videocodec.h"

/* it doesn't make sense to have more than 20 or so,
  just to prevent some unwanted loops */
#define MAX_CODECS 20

/* amount of chips attached via this driver */
static int zr36015_codecs = 0;

/* debugging is available via module parameter */

static int debug = 1;
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug level (0-4)");

#define dprintk(num, format, args...) \
	do { \
		if (debug >= num) \
			printk(format, ##args); \
	} while (0)

/* =========================================================================
   Local hardware I/O functions:

   read/write via codec layer (registers are located in the master device)
   ========================================================================= */

/* read and write functions */
static u8
zr36015_read (struct zr36015 *ptr,
	      u16             reg)
{
	u8 value = 0;

	// just in case something is wrong...
	if (ptr->codec->master_data->readreg)
		value =
		    (ptr->codec->master_data->
		     readreg(ptr->codec, reg)) & 0xFF;
	else
		dprintk(1,
			KERN_ERR "%s: invalid I/O setup, nothing read!\n",
			ptr->name);

	dprintk(4, "%s: reading from 0x%04x: %02x\n", ptr->name, reg,
		value);

	return value;
}

static void
zr36015_write (struct zr36015 *ptr,
	       u16             reg,
	       u8              value)
{
	dprintk(4, "%s: writing 0x%02x to 0x%04x\n", ptr->name, value,
		reg);

	// just in case something is wrong...
	if (ptr->codec->master_data->writereg) {
		ptr->codec->master_data->writereg(ptr->codec, reg, value);
	} else
		dprintk(1,
			KERN_ERR
			"%s: invalid I/O setup, nothing written!\n",
			ptr->name);
}

/* indirect read and write functions */
static u8
zr36015_readi (struct zr36015 *ptr,
	       u16             reg)
{
	u8 value = 0;

	// just in case something is wrong...
	if ((ptr->codec->master_data->writereg) &&
	    (ptr->codec->master_data->readreg)) {
		ptr->codec->master_data->writereg(ptr->codec, ZR015_IADDR, reg & 0x0F);	// ADDR
		value = (ptr->codec->master_data->readreg(ptr->codec, ZR015_IDATA)) & 0xFF;	// DATA
	} else
		dprintk(1,
			KERN_ERR
			"%s: invalid I/O setup, nothing read (i)!\n",
			ptr->name);

	dprintk(4, "%s: reading indirect from 0x%04x: %02x\n", ptr->name,
		reg, value);
	return value;
}

static void
zr36015_writei (struct zr36015 *ptr,
	        u16             reg,
	        u8              value)
{
	dprintk(4, "%s: writing indirect 0x%02x to 0x%04x\n", ptr->name,
		value, reg);

	// just in case something is wrong...
	if (ptr->codec->master_data->writereg) {
		ptr->codec->master_data->writereg(ptr->codec, ZR015_IADDR, reg & 0x0F);	// ADDR
		ptr->codec->master_data->writereg(ptr->codec, ZR015_IDATA, value & 0x0FF);	// DATA
	} else
		dprintk(1,
			KERN_ERR
			"%s: invalid I/O setup, nothing written (i)!\n",
			ptr->name);
}

/* =========================================================================
   Local helper function:

   basic test of "connectivity", writes/reads to/from HW-Lo register
   ========================================================================= */

static int
zr36015_basic_test (struct zr36015 *ptr)
{
	if (debug) {
		int i;
		zr36015_writei(ptr, ZR015I_HW_LO, 0x55);
		dprintk(1, KERN_INFO "%s: registers: ", ptr->name);
		for (i = 0; i <= 0x09; i++)
			dprintk(1, "%02x ", zr36015_readi(ptr, i));
		dprintk(1, "\n");
	}
	// for testing just write 0, then the default value to a register and read
	// it back in both cases 
	zr36015_writei(ptr, ZR015I_HW_LO, 0x00);
	if (zr36015_readi(ptr, ZR015I_HW_LO) != 0x0) {
		dprintk(1,
			KERN_ERR
			"%s: attach failed, can't connect to vfe processor!\n",
			ptr->name);
		return -ENXIO;
	}
	zr36015_writei(ptr, ZR015I_HW_LO, 0x0d0);
	if (zr36015_readi(ptr, ZR015I_HW_LO) != 0x0d0) {
		dprintk(1,
			KERN_ERR
			"%s: attach failed, can't connect to vfe processor!\n",
			ptr->name);
		return -ENXIO;
	}

	return 0;		/* looks good! */
}

/* =========================================================================
   Local helper function:

   simple loop for pushing the init datasets - NO USE --
   ========================================================================= */

#if 0
static int zr36015_pushit (struct zr36015 *ptr,
			   u16             startreg,
			   u16             len,
                           const char     *data)
{
        int i=0;

        dprintk(4, "%s: write data block to 0x%04x (len=%d)\n",
		ptr->name, startreg,len);
        while (i<len) {
                zr36015_writei(ptr, startreg++,  data[i++]);
        }

        return i;
}
#endif

/* =========================================================================
   Basic datasets & init:

   //TODO//
   ========================================================================= */

// needed offset values          PAL NTSC SECAM
static const int zr015_xoff[] = { 20, 20, 20 };
static const int zr015_yoff[] = { 8, 9, 7 };

static void
zr36015_init (struct zr36015 *ptr)
{
	// stop any processing
	zr36015_write(ptr, ZR015_RESET, 0);

	// mode setup (yuv422 in and out, compression/expansion due to mode,
	//             decimation due to xdec)
	zr36015_write(ptr, ZR015_MODE,
			ZR015_YUV422_YUV422 | (ptr->xdec ? ZR015_DECIMATE : 0) |
			(ptr->mode == CODEC_DO_COMPRESSION ?
			ZR015_COMPRESSION : ZR015_EXPANSION)
		     );
	// Window setup
	// (no extra offset for now, norm defines offset, default width height)
	zr36015_writei(ptr, ZR015I_HW_HI, ptr->width >> 8);
	zr36015_writei(ptr, ZR015I_HW_LO, ptr->width & 0xFF);
	zr36015_writei(ptr, ZR015I_VW_HI, ptr->height >> 8);
	zr36015_writei(ptr, ZR015I_VW_LO, ptr->height & 0xFF);
	zr36015_writei(ptr, ZR015I_HD_HI, ptr->xoff >> 8);
	zr36015_writei(ptr, ZR015I_HD_LO, ptr->xoff & 0xFF);
	zr36015_writei(ptr, ZR015I_VD_HI, ptr->yoff >> 8);
	zr36015_writei(ptr, ZR015I_VD_LO, ptr->yoff & 0xFF);

}

/* =========================================================================
   CODEC API FUNCTIONS

   this functions are accessed by the master via the API structure
   ========================================================================= */

/* set compression/expansion mode and launches codec -
   this should be the last call from the master before starting processing */
static int
zr36015_set_mode (struct videocodec *codec,
		  int                mode)
{
	struct zr36015 *ptr = (struct zr36015 *) codec->data;

	dprintk(2, "%s: set_mode %d call\n", ptr->name, mode);

	if ((mode != CODEC_DO_EXPANSION) && (mode != CODEC_DO_COMPRESSION))
		return -EINVAL;

	ptr->mode = mode;
	zr36015_init(ptr);

	return 0;
}

/* set picture size */
static int
zr36015_set_video (struct videocodec   *codec,
		   struct tvnorm       *norm,
		   struct vfe_settings *cap,
		   struct vfe_polarity *pol)
{
	struct zr36015 *ptr = (struct zr36015 *) codec->data;

	dprintk(2, "%s: set_video %d.%d, %d/%d-%dx%d call\n", ptr->name,
		norm->HStart, norm->VStart, cap->x, cap->y, cap->width,
		cap->height);

	/* if () return -EINVAL;
	 * trust the master driver that it knows what it does - so
	 * we allow invalid startx/y for now ... */
	ptr->width = cap->width;
	ptr->height = cap->height;
	ptr->xoff = norm->HStart;
	ptr->yoff = norm->VStart;
	zr36015_init(ptr);

	return 0;
}

/* additional control functions */
static int
zr36015_control (struct videocodec *codec,
		 int                type,
		 int                size,
		 void              *data)
{
	struct zr36015 *ptr = (struct zr36015 *) codec->data;
	int *ival = (int *) data;

	dprintk(2, "%s: control %d call with %d byte\n", ptr->name, type,
		size);

	switch (type) {
	case CODEC_G_STATUS:	/* get last status - we don't know it ... */
		if (size != sizeof(int))
			return -EFAULT;
		*ival = 0;
		break;

	case CODEC_G_CODEC_MODE:
		if (size != sizeof(int))
			return -EFAULT;
		*ival = 0;
		break;

	case CODEC_S_CODEC_MODE:
		if (size != sizeof(int))
			return -EFAULT;
		if (*ival != 0)
			return -EINVAL;
		/* not needed, do nothing */
		return 0;

	case CODEC_G_VFE:
	case CODEC_S_VFE:
		return 0;

	case CODEC_S_MMAP:
		/* not available, give an error */
		return -ENXIO;

	default:
		return -EINVAL;
	}

	return size;
}

/* =========================================================================
   Exit and unregister function:

   Deinitializes Zoran's raster to block converter
   ========================================================================= */

static int
zr36015_unset (struct videocodec *codec)
{
	struct zr36015 *ptr = codec->data;

	if (ptr) {
		/* do wee need some codec deinit here, too ???? */

		dprintk(1, "%s: finished codec #%d\n", ptr->name,
			ptr->num);
		kfree(ptr);
		codec->data = NULL;

		zr36015_codecs--;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
		MOD_DEC_USE_COUNT;
#else
		module_put(THIS_MODULE);
#endif

		return 0;
	}

	return -EFAULT;
}

/* =========================================================================
   Setup and registry function:

   Initializes Zoran's raster to block converter

   Also sets pixel size, average code size, mode (compr./decompr.)
   (the given size is determined by the processor with the video interface)
   ========================================================================= */

static int
zr36015_setup (struct videocodec *codec)
{
	struct zr36015 *ptr;
	int res;

	dprintk(2, "zr36015: initializing VFE subsystem #%d.\n",
		zr36015_codecs);

	if (zr36015_codecs == MAX_CODECS) {
		dprintk(1,
			KERN_ERR "zr36015: Can't attach more codecs!\n");
		return -ENOSPC;
	}
	//mem structure init
	codec->data = ptr = kmalloc(sizeof(struct zr36015), GFP_KERNEL);
	if (NULL == ptr) {
		dprintk(1, KERN_ERR "zr36015: Can't get enough memory!\n");
		return -ENOMEM;
	}
	memset(ptr, 0, sizeof(struct zr36015));

	snprintf(ptr->name, sizeof(ptr->name), "zr36015[%d]",
		 zr36015_codecs);
	ptr->num = zr36015_codecs++;
	ptr->codec = codec;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_INC_USE_COUNT;
#else
	if ((res = try_module_get(THIS_MODULE)) != 0) {
		dprintk(1,
			KERN_ERR
			"zr36015: failed to increase module use count\n");
		kfree(ptr);
		zr36015_codecs--;
		return res;
	}
#endif

	//testing
	res = zr36015_basic_test(ptr);
	if (res < 0) {
		zr36015_unset(codec);
		return res;
	}
	//final setup
	ptr->mode = CODEC_DO_COMPRESSION;
	ptr->width = 768;
	ptr->height = 576;
	//zr36015_init(ptr);

	dprintk(1, KERN_INFO "%s: codec  attached and running\n",
		ptr->name);

	return 0;
}

static const struct videocodec zr36015_codec = {
	.name = "zr36015",
	.magic = 0L,		// magic not used
	.flags =
	    CODEC_FLAG_HARDWARE | CODEC_FLAG_VFE | CODEC_FLAG_ENCODER |
	    CODEC_FLAG_DECODER,
	.type = CODEC_TYPE_ZR36015,
	.setup = zr36015_setup,	// functionality
	.unset = zr36015_unset,
	.set_mode = zr36015_set_mode,
	.set_video = zr36015_set_video,
	.control = zr36015_control,
	// others are not used
};

/* =========================================================================
   HOOK IN DRIVER AS KERNEL MODULE
   ========================================================================= */

static int __init
zr36015_init_module (void)
{
	//dprintk(1, "ZR36015 driver %s\n",ZR015_VERSION);
	zr36015_codecs = 0;
	videocodec_register(&zr36015_codec);

	return 0;
}

static void __exit
zr36015_cleanup_module (void)
{
	if (zr36015_codecs) {
		dprintk(1,
			"zr36015: something's wrong - %d codecs left somehow.\n",
			zr36015_codecs);
	} else {
		videocodec_unregister(&zr36015_codec);
	}
}

module_init(zr36015_init_module);
module_exit(zr36015_cleanup_module);

MODULE_AUTHOR("Luca Bonissi <lucabon@interfree.it>");
MODULE_DESCRIPTION("Driver module for ZR36015 video frontends "
		   ZR015_VERSION);
MODULE_LICENSE("GPL");
