/*
 * kernel/power/proc.c
 *
 * Copyright (C) 2002-2006 Nigel Cunningham <nigel@suspend2.net>
 *
 * This file is released under the GPLv2.
 *
 * This file contains support for proc entries for tuning Suspend2.
 *
 * We have a generic handler that deals with the most common cases, and
 * hooks for special handlers to use.
 */

#include <linux/suspend.h>
#include <linux/module.h>
#include <asm/uaccess.h>

#include "proc.h"
#include "suspend2.h"
#include "storage.h"

static int suspend_proc_initialised = 0;

static struct list_head suspend_proc_entries;
static struct proc_dir_entry *suspend_dir;
static struct suspend_proc_data proc_params[];

extern void __suspend_try_resume(void);
extern void suspend_main(void);

extern struct subsystem power_subsys;

#define to_proc_data(_attr) container_of(_attr, struct suspend_proc_data, attr)

static int __suspend_read_proc(struct suspend_proc_data *proc_data, char *page,
		int count)
{
	int len = 0;

	if (suspend_start_anything(0))
		return -EBUSY;

	if (proc_data->needs_storage_manager & PROC_NEEDS_FOR_READ)
		suspend_prepare_usm();
	
	switch (proc_data->type) {
		case SUSPEND_PROC_DATA_CUSTOM:
			len = (proc_data->data.special.read_proc) ?
				(proc_data->data.special.read_proc)(page, count)
				: 0;
			break;
		case SUSPEND_PROC_DATA_BIT:
			len = sprintf(page, "%d\n", 
				-test_bit(proc_data->data.bit.bit,
					proc_data->data.bit.bit_vector));
			break;
		case SUSPEND_PROC_DATA_INTEGER:
			len = sprintf(page, "%d\n",
				*(proc_data->data.integer.variable));
			break;
		case SUSPEND_PROC_DATA_LONG:
			len = sprintf(page, "%ld\n",
				*(proc_data->data.a_long.variable));
			break;
		case SUSPEND_PROC_DATA_UL:
			len = sprintf(page, "%lu\n",
				*(proc_data->data.ul.variable));
			break;
		case SUSPEND_PROC_DATA_STRING:
			len = sprintf(page, "%s\n",
				proc_data->data.string.variable);
			break;
	}
	/* Side effect routine? */
	if (proc_data->read_side_effect)
		proc_data->read_side_effect();

	if (proc_data->needs_storage_manager & PROC_NEEDS_FOR_READ)
		suspend_cleanup_usm();

	suspend_finish_anything(0);

	return len;
}

/* suspend_read_proc
 *
 * Generic handling for reading the contents of bits, integers,
 * unsigned longs and strings.
 */
static int suspend_read_proc(char *page, char **start, off_t off, int count,
		int *eof, void *data)
{
	int len = 0, copy_len;
	struct suspend_proc_data *proc_data = (struct suspend_proc_data *) data;
	char *all = (char *) get_zeroed_page(GFP_ATOMIC);

	if (!all)
		return -ENOMEM;

	len = __suspend_read_proc(proc_data, all, PAGE_SIZE);
	
	copy_len = min(len - (int) off, count);
	if (copy_len < 0)
		copy_len = 0;

	if (copy_len) {
		memcpy(page, all + off, copy_len);
		*start = page;
	} 

	if (copy_len + off == len)
		*eof = 1;

	if (len <= count)
		*eof = 1;
	
	free_page((unsigned long) all);

	return copy_len;
}

#define BOUND(_variable, _type) \
	if (proc_data->data._type.maximum) { \
		if (*_variable < proc_data->data._type.minimum) \
			*_variable = proc_data->data._type.minimum; \
		else if (*_variable > proc_data->data._type.maximum) \
			*_variable = proc_data->data._type.maximum; \
	}

static int __suspend_write_proc(struct suspend_proc_data *proc_data,
		char *my_buf, int count)
{
	int assigned_temp_buffer = 0, result = count;
	
	if (suspend_start_anything(proc_data == &proc_params[0]))
		return -EBUSY;

	my_buf[count] = 0;

	if (proc_data->needs_storage_manager & PROC_NEEDS_FOR_WRITE)
		suspend_prepare_usm();

	switch (proc_data->type) {
		case SUSPEND_PROC_DATA_CUSTOM:
			if (proc_data->data.special.write_proc)
				result = (proc_data->data.special.write_proc)
					(my_buf, count);
			break;
		case SUSPEND_PROC_DATA_BIT:
			{
			int value = simple_strtoul(my_buf, NULL, 0);
			if (value)
				set_bit(proc_data->data.bit.bit, 
					(proc_data->data.bit.bit_vector));
			else
				clear_bit(proc_data->data.bit.bit,
					(proc_data->data.bit.bit_vector));
			}
			break;
		case SUSPEND_PROC_DATA_INTEGER:
			{
				int *variable = proc_data->data.integer.variable;
				*variable = simple_strtol(my_buf, NULL, 0);
				BOUND(variable, integer);
				break;
			}
		case SUSPEND_PROC_DATA_LONG:
			{
				long *variable = proc_data->data.a_long.variable;
				*variable = simple_strtol(my_buf, NULL, 0);
				BOUND(variable, a_long);
				break;
			}
		case SUSPEND_PROC_DATA_UL:
			{
				unsigned long *variable = proc_data->data.ul.variable;
				*variable = simple_strtoul(my_buf, NULL, 0);
				BOUND(variable, ul);
				break;
			}
			break;
		case SUSPEND_PROC_DATA_STRING:
			{
				int copy_len = count;
				char *variable =
					proc_data->data.string.variable;

				if (proc_data->data.string.max_length &&
				    (copy_len > proc_data->data.string.max_length))
					copy_len = proc_data->data.string.max_length;

				if (!variable) {
					proc_data->data.string.variable =
						variable = (char *) get_zeroed_page(GFP_ATOMIC);
					assigned_temp_buffer = 1;
				}
				strncpy(variable, my_buf, copy_len);
				if ((copy_len) &&
					 (my_buf[copy_len - 1] == '\n'))
					variable[count - 1] = 0;
				variable[count] = 0;
			}
			break;
	}

	/* Side effect routine? */
	if (proc_data->write_side_effect)
		proc_data->write_side_effect();

	/* Free temporary buffers */
	if (assigned_temp_buffer) {
		free_page((unsigned long) proc_data->data.string.variable);
		proc_data->data.string.variable = NULL;
	}

	if (proc_data->needs_storage_manager & PROC_NEEDS_FOR_WRITE)
		suspend_cleanup_usm();

	suspend_finish_anything(proc_data == &proc_params[0]);

	return result;
}

/* suspend_write_proc
 *
 * Generic routine for handling writing to files representing
 * bits, integers and unsigned longs.
 */

static int suspend_write_proc(struct file *file, const char *buffer,
		unsigned long count, void *data)
{
	struct suspend_proc_data *proc_data = (struct suspend_proc_data *) data;
	char *my_buf = (char *) get_zeroed_page(GFP_ATOMIC);
	int result = count;

	if (!my_buf)
		return -ENOMEM;

	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	if (copy_from_user(my_buf, buffer, count)) {
		free_page((unsigned long) my_buf);
		return -EFAULT;
	}
	
	result = __suspend_write_proc(proc_data, my_buf, count);

	free_page((unsigned long) my_buf);

	return result;
}

ssize_t suspend2_attr_show(struct kobject *kobj, struct attribute *attr,
			      char *page)
{
	struct suspend_proc_data *proc_data;
	proc_data = to_proc_data(attr);
	return __suspend_read_proc(proc_data, page, PAGE_SIZE);
}

ssize_t suspend2_attr_store(struct kobject *kobj, struct attribute *attr,
		const char *page, size_t count)
{
	struct suspend_proc_data *proc_data = to_proc_data(attr);
	return __suspend_write_proc(proc_data, (char *) page, count);
}

static struct sysfs_ops suspend2_sysfs_ops = {
	.show	= &suspend2_attr_show,
	.store	= &suspend2_attr_store,
};

static struct kobj_type suspend2_ktype = {
	.sysfs_ops	= &suspend2_sysfs_ops,
};

decl_subsys(suspend2, &suspend2_ktype, NULL);

/* Non-module proc entries.
 *
 * This array contains entries that are automatically registered at
 * boot. Modules and the console code register their own entries separately.
 *
 * NB: If you move do_suspend, change suspend_write_proc's test so that
 * suspend_start_anything still gets a 1 when the user echos > do_suspend!
 */

static struct suspend_proc_data proc_params[] = {
	{ SUSPEND2_ATTR("do_suspend", PROC_WRITEONLY),
	  PROC_CUSTOM(NULL, NULL, PROC_NEEDS_FOR_WRITE),
	  .write_side_effect = suspend_main
	},

	{ SUSPEND2_ATTR("do_resume", PROC_WRITEONLY),
	  PROC_CUSTOM(NULL, NULL, PROC_NEEDS_FOR_WRITE),
	  .write_side_effect = __suspend_try_resume
	},
};

struct kobject *make_suspend2_sysdir(char *name)
{
	struct kobject *kobj = kzalloc(sizeof(struct kobject), GFP_KERNEL);
	int err;

	BUG_ON(!kobj);

	err = kobject_set_name(kobj, "%s", name);

	if (err) {
		kfree(kobj);
		return NULL;
	}

	kobj->kset = &suspend2_subsys.kset;

	err = kobject_register(kobj);

	return err ? NULL : kobj;
}

/* suspend_initialise_proc
 *
 * Initialise the /proc/suspend2 directory.
 */

static void suspend_initialise_proc(void)
{
	int i, error;
	int numfiles = sizeof(proc_params) / sizeof(struct suspend_proc_data);
	
	if (suspend_proc_initialised)
		return;

	suspend_dir = proc_mkdir("suspend2", NULL);
	
	BUG_ON(!suspend_dir);

	INIT_LIST_HEAD(&suspend_proc_entries);

	/* Make our suspend2 directory a child of /sys/power */
	kobj_set_kset_s(&suspend2_subsys.kset, power_subsys);
	error = subsystem_register(&suspend2_subsys);

	if (error)
		return;

	/* Make it use the .store and .show routines above */
	kobj_set_kset_s(&suspend2_subsys.kset, suspend2_subsys);

	suspend_proc_initialised = 1;

	for (i=0; i< numfiles; i++)
		suspend_register_procfile(&suspend2_subsys.kset.kobj,
				&proc_params[i]);
}

/* suspend_register_procfile
 *
 * Helper for registering a new /proc/suspend2 entry.
 */

struct proc_dir_entry *suspend_register_procfile(
		struct kobject *kobj,
		struct suspend_proc_data *suspend_proc_data)
{
	struct proc_dir_entry *new_entry;
	if (!suspend_proc_initialised)
		suspend_initialise_proc();

	new_entry = create_proc_entry(suspend_proc_data->attr.name,
			suspend_proc_data->attr.mode, suspend_dir);
	
	//sysfs_create_file(&suspend2_subsys.kset.kobj, &suspend_proc_data->attr);
	sysfs_create_file(kobj, &suspend_proc_data->attr);
	
	if (new_entry) {
		list_add_tail(&suspend_proc_data->proc_data_list,
				&suspend_proc_entries);
		new_entry->read_proc = suspend_read_proc;
		new_entry->write_proc = suspend_write_proc;
		new_entry->data = suspend_proc_data;
	} else {
		printk("Error! create_proc_entry returned NULL.\n");
		INIT_LIST_HEAD(&suspend_proc_data->proc_data_list);
	}
	return new_entry;
}

/* suspend_unregister_procfile
 *
 * Helper for removing unwanted /proc/suspend2 entries.
 *
 */
void suspend_unregister_procfile(struct suspend_proc_data *suspend_proc_data)
{
	if (list_empty(&suspend_proc_data->proc_data_list))
		return;

	remove_proc_entry(suspend_proc_data->attr.name, suspend_dir);
	
	sysfs_remove_file(&suspend2_subsys.kset.kobj, &suspend_proc_data->attr);
	
	list_del(&suspend_proc_data->proc_data_list);
}
