/*	target/iscsi_target.c

	vi: set autoindent tabstop=4 shiftwidth=4 :

 	This is the iscsi target front-end which interfaces with the STML.
	The front-end has been written to the specifications of Draft 20 of the
	iSCSI spec.

*/
/*
	Copyright (C) 2001-2004 InterOperability Lab (IOL)
							University of New Hampshier (UNH)
							Durham, NH 03824

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2, or (at your option)
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
	USA.

	The name IOL and/or UNH may not be used to endorse or promote products
	derived from this software without specific prior written permission.
*/

#if !defined(MANGLE_INQUIRY_DATA)
#define MANGLE_INQUIRY_DATA
#endif

#include "iscsi_target.h"
#include "target_error_rec.h"
#include <net/sock.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <net/if_inet6.h>
#include <net/ipv6.h>
#include <net/addrconf.h>

/* added by RDR when rcu_read_lock(), rcu_read_ulock() came in with 2.6.10
 * These macros define it the way it was in 2.4 and in earlier versions of 2.6
 */
#ifndef rcu_read_lock
#define rcu_read_lock()	read_lock(&inetdev_lock)
#endif
#ifndef rcu_read_unlock
#define rcu_read_unlock()	read_unlock(&inetdev_lock)
#endif

static void clean_bad_stuff(void);
static int iscsi_release_connection(struct iscsi_conn *conn);

static int handle_login(struct iscsi_conn *conn, __u8 *buffer);
static int handle_text_request(struct iscsi_conn *conn,
							   struct iscsi_session *session,
							   __u8 *buffer);
static int handle_logout(struct iscsi_conn *conn,
						 struct iscsi_session *session,
						 __u8 *buffer);
static int handle_nopout(struct iscsi_conn *conn,
						 struct iscsi_session *session,
						 __u8 *buffer);
static int handle_cmnd(struct iscsi_conn *conn,
					   struct iscsi_session *session,
					   __u8 *buffer);
static int handle_task_mgt_command(struct iscsi_conn *conn,
								   struct iscsi_session *session,
								   __u8 *buffer);
static int handle_data(struct iscsi_conn *conn,
					   struct iscsi_session *session,
					   __u8 *buffer);

static int send_unsolicited_data(struct iscsi_cmnd *cmnd,
								 struct iscsi_conn *conn,
								 struct iscsi_session *session);
static void check_queued_cmnd(struct iscsi_session *session);
static void free_data_list(struct iscsi_cmnd *cmnd);

static struct iscsi_cmnd * __attribute__ ((no_instrument_function))
search_iscsi_cmnd(Target_Scsi_Cmnd * cmnd, struct iscsi_session **result_sess);
static struct iscsi_cmnd *search_task_mgt_command(Target_Scsi_Message *
												  message);
static struct iscsi_cmnd * __attribute__ ((no_instrument_function))
search_tags(struct iscsi_conn *conn, __u32 init_task_tag,
									 __u32 target_xfer_tag,
									 int dumpall);

static int handle_discovery_rsp(struct iscsi_cmnd *cmnd,
								struct iscsi_conn *conn,
								struct iscsi_session *session);
static int ask_for_more_text(struct iscsi_cmnd *cmnd,
							 struct iscsi_conn *conn,
							 struct iscsi_session *session);
static int handle_logout_rsp(struct iscsi_cmnd *cmnd,
							 struct iscsi_conn *conn,
							 struct iscsi_session *session);
static int handle_nopin(struct iscsi_cmnd *cmnd,
						struct iscsi_conn *conn,
						struct iscsi_session *session);
static int generate_nopin(struct iscsi_conn *conn,
						  struct iscsi_session *session);
static int handle_iscsi_done(struct iscsi_cmnd *cmnd,
							 struct iscsi_conn *conn,
							 struct iscsi_session *session);
static int handle_iscsi_mgt_fn_done(struct iscsi_cmnd *cmnd,
									struct iscsi_conn *conn,
									struct iscsi_session *session);
static void iscsi_dequeue(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn);
static int iscsi_tx_login_reject(struct iscsi_conn *conn,
								 struct iscsi_init_login_cmnd *pdu,
								 __u8 status_class,
								 __u8 status_detail);
static int iscsi_tx_r2t(struct iscsi_cmnd *cmnd,
						struct iscsi_conn *conn,
						struct iscsi_session *session);

static int
dequeue_reject(struct iscsi_conn *conn, int sendit);

/* Added for handling SNACK requests - SAI */
static int handle_snack(struct iscsi_conn *conn,
						struct iscsi_session *session,
						__u8 *buffer);
static int send_iscsi_response(struct iscsi_cmnd *cmnd,
							   struct iscsi_conn *conn,
							   struct iscsi_session *session);

/* Daren Hayward, darenh@4bridgeworks.com */
#if defined(MANGLE_INQUIRY_DATA)
static int
mangle_inquiry_data(struct iscsi_cmnd *iscsi_command, struct iovec *iov,
					int start_iov, int limit_iov);
#endif

struct iscsi_global *devdata;

/* 
 * initialize stuff in devdata.
 */
static int
init_target(struct iscsi_global *devdata)
{
	memset(devdata, 0, sizeof(struct iscsi_global));

	INIT_LIST_HEAD(&devdata->session_list);
	INIT_LIST_HEAD(&devdata->bad_session_list);

	init_MUTEX(&devdata->session_sem);
	init_MUTEX(&devdata->session_read_mutex);
	init_MUTEX_LOCKED(&devdata->server_sem);

	devdata->param_tbl = my_kmalloc(MAX_CONFIG_PARAMS
							* sizeof(struct parameter_type), "param_tbl");
	if (!(devdata->param_tbl)) {
		return -1;
	}

	/* Copy the default parameters */
	param_tbl_init(*devdata->param_tbl);

	/* chap and srp support - CHONG */
	devdata->auth_parameter.chap_local_ctx = CHAP_InitializeContext();
	devdata->auth_parameter.chap_peer_ctx = CHAP_InitializeContext();
	devdata->auth_parameter.srp_ctx = SRP_InitializeContext();

	return 0;
}

void
bring_down_portals( void )
{
	int i;
	struct portal_group *ptr;

	for (i = 0, ptr = iscsi_portal_groups; i < MAX_PORTAL; i++, ptr++) {
		if (ptr->in_use == 0)
			continue;
	
		if (devdata->server_thr[i]) {
			/*  Mike Christie mikenc@us.ibm.com */
			send_sig(ISCSI_SHUTDOWN_SIGNAL, devdata->server_thr[i], 1);
			down_interruptible(&devdata->server_sem);
		}
	
		if (devdata->server_socket[i]) {
			sock_release(devdata->server_socket[i]);
			devdata->server_socket[i] = NULL;
		}

		my_kfree((void **)&ptr->ip_address, "ip_address");
		if (ptr->in_use == 2)
		{
			my_kfree((void **)&ptr->ip_string, "ip_string");
			my_kfree((void **)&ptr->port_string, "port_string");
		}
		
		ptr->in_use = 0;
	}

	if (devdata->param_tbl) {
		param_tbl_uncpy(*devdata->param_tbl);
		my_kfree((void**)&devdata->param_tbl, "param_tbl");
	}

	/* chap and srp support - CHONG */
	CHAP_FinalizeContext(devdata->auth_parameter.chap_local_ctx);
	CHAP_FinalizeContext(devdata->auth_parameter.chap_peer_ctx);
	SRP_FinalizeContext(devdata->auth_parameter.srp_ctx);

	my_kfree((void **)&devdata, "devdata");

	/* check for any unfreed memory */
	my_kempty();
}

int stop_server_thread (char *ip_string, char *port_string)
{
	int i;
	
	for (i = 0; i < MAX_PORTAL; i++)
	{
		if (iscsi_portal_groups[i].in_use == 0)
			continue;
	
		if ((strcmp(iscsi_portal_groups[i].ip_string, ip_string) == 0)
			&& (strcmp(iscsi_portal_groups[i].port_string, port_string) == 0))
		{
			if (devdata->server_thr[i]) {
				/*  Mike Christie mikenc@us.ibm.com */
				send_sig(ISCSI_SHUTDOWN_SIGNAL, devdata->server_thr[i], 1);
				down_interruptible(&devdata->server_sem);
			}
	
			if (devdata->server_socket[i]) {
				sock_release(devdata->server_socket[i]);
				devdata->server_socket[i] = NULL;
			}

			my_kfree((void **)&iscsi_portal_groups[i].ip_address, "ip_address");
			if (iscsi_portal_groups[i].in_use == 2)
			{
				my_kfree((void **)&iscsi_portal_groups[i].ip_string, "ip_string");
				my_kfree((void **)&iscsi_portal_groups[i].port_string, "port_string");
			}

			iscsi_portal_groups[i].in_use = 0;
			return 0;
		}
	}

	TRACE_ERROR("%s Can't find match with ip %s, port %s\n", current->comm,
				ip_string, port_string);
	return -1;
}

/* returns 0 on success, -1 on failure */
int
bring_up_portal( struct portal_group *ptr )
{
	int backlog = 5;
	struct socket *sockptr;

	TRACE(TRACE_DEBUG, "start_server_thread: ip_string %s, port_string %s\n",
									ptr->ip_string,
									ptr->port_string);
	
	ptr->family = cnv_string_to_inet(
									ptr->ip_string,
									ptr->port_string,
									&ptr->ip_address,
									&ptr->ip_length);

	if (ptr->family < 0) {
		goto out;
	}

	/* create a socket */
#ifdef FC2
	if (sock_create(ptr->family, SOCK_STREAM, 0, &sockptr, 0) < 0) {
#else
	if (sock_create(ptr->family, SOCK_STREAM, 0, &sockptr) < 0) {
#endif
		TRACE_ERROR("%s Could not create socket on %s:%s\n", current->comm,
					ptr->ip_string, ptr->port_string);
		goto out1;
	}

	TRACE(TRACE_NET, "Socket %p created\n", sockptr);

	/* bind */
	if (sockptr->ops->bind(sockptr, ptr->ip_address, ptr->ip_length)) {
		TRACE_ERROR("%s Could not bind socket on %s:%s\n", current->comm,
					ptr->ip_string, ptr->port_string);
		goto out2;
	}

	TRACE(TRACE_NET, "Socket %p bound\n", sockptr);

	/*  turn on the option to reuse this socket's port quickly */
	tcp_reuse_port(sockptr);

	/* listen */
	if ((sockptr->ops->listen(sockptr, backlog))) {
		TRACE_ERROR("%s Could not listen with socket on %s:%s\n", current->comm,
					ptr->ip_string, ptr->port_string);
		goto out2;
	}

	devdata->server_socket[ptr - iscsi_portal_groups] = sockptr;

	/* create a server_thread that can accept connections */
	if (kernel_thread(iscsi_server_thread, (void *)ptr, 0) < 0) {
		TRACE_ERROR("%s Unable to create server thread\n", current->comm);
		devdata->server_socket[ptr - iscsi_portal_groups] = NULL;
		goto out2;
	}

	TRACE(TRACE_NET, "Server_thread spawned for socket %p\n", sockptr);

	/* wait for that server_thread to come up before continuing */
	if (!down_interruptible(&devdata->server_sem))
		return 0;
	
out2:
	sock_release(sockptr);

out1:
	my_kfree((void **)&ptr->ip_address, "ip_address");

out:
	return -1;
}


int start_server_thread (char *ip_string, char *port_string, int tag)
{
	int i;
	struct portal_group *ptr;

	if (!devdata) {
		TRACE_ERROR("%s No device available\n", current->comm);
		goto out;
	}

	for (i = 0, ptr = iscsi_portal_groups; i < MAX_PORTAL; i++, ptr++)
	{
		if (!ptr->in_use)
			break;
	}

	if (i >= MAX_PORTAL)
	{
		TRACE_ERROR("%s No more portals permitted\n", current->comm);
		goto out;
	}

	/* mark this portal in use with dynamic ip strings */
	ptr->in_use = 2;
	
	ptr->ip_string = (char *)my_kmalloc(strlen(ip_string)+1, "ip_string");
	if (ptr->ip_string == NULL) {
		goto out0;
	}
	strcpy(ptr->ip_string, ip_string);

	ptr->port_string = (char *)my_kmalloc(strlen(port_string)+1,"port_string");
	if (ptr->port_string == NULL) {
		goto out1;
	}
	strcpy(ptr->port_string, port_string);

	ptr->tag = tag;
	
	if (bring_up_portal(ptr)) {
		goto out2;
	}

	return 0;

out2:
	my_kfree((void **)&ptr->port_string, "port_string");
out1:
	my_kfree((void **)&ptr->ip_string, "ip_string");
out0:
	ptr->in_use = 0;
out:
	return -1;
}


/*
 * iscsi_detect: this function sets up the server so that it can accept 
 * multiple connections.
 */
int
iscsi_detect(Scsi_Target_Template * tmpt)
{
	long i;
	struct portal_group *ptr;

	TRACE(TRACE_DEBUG, "Entering iSCSI detect\n");

	devdata = (struct iscsi_global *)my_kmalloc(sizeof(struct iscsi_global),
						     					"devdata");

	if (!devdata) {
		return -1;
	}

	/* Initialize devdata */
	if (init_target(devdata) < 0) {
		TRACE_ERROR("%s Initialize devdata failed\n", current->comm);
		my_kfree((void **)&devdata, "devdata");
		return -1;
	}

	/* setup the security key hash table */
	setup_security_hash_table();

	TRACE(TRACE_DEBUG, "iSCSI initialization completed\n");

	for (i = 0, ptr = iscsi_portal_groups; i < MAX_PORTAL && ptr->tag != 0;
																i++, ptr++) {
		ptr->in_use = 1;
		if (bring_up_portal(ptr)) {
			ptr->in_use = 0;
			goto out;
		}
	}

	/* register the front end now */
	devdata->device = register_target_front_end(tmpt);

	if (!devdata->device) {
		TRACE_ERROR("%s Device registration failed\n", current->comm);
		goto out;
	}

	devdata->device->dev_specific = (void *) devdata;

	TRACE(TRACE_DEBUG, "Registration complete\n");

	return 1;					/* one device detected */

out:
	bring_down_portals();

	return -2;
}

/*
 * iscsi_release: function to release the iSCSI device as required by
 * the Mid-Level
 */
int
iscsi_release(Scsi_Target_Device * device)
{
	struct iscsi_session *session;
	struct list_head *list_ptr, *list_temp;
	int err = 0;

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_release\n");

	if ((struct iscsi_global *) device->dev_specific != devdata) {
		TRACE_ERROR("%s This is not an iscsi device\n", current->comm);
		err = -1;
		goto out;
	}

	/* destructive access to session lists */
	if (!down_interruptible(&devdata->session_sem)) {
		list_for_each_safe(list_ptr, list_temp, &devdata->session_list) {
			session = list_entry(list_ptr, struct iscsi_session, sess_link);
			iscsi_release_session(session);
		}
		list_for_each_safe(list_ptr, list_temp, &devdata->bad_session_list) {
			session = list_entry(list_ptr, struct iscsi_session, sess_link);
			TRACE(TRACE_DEBUG, "iscsi%d: release bad session %p, tsih %u\n",
			  	(int) devdata->device->id, session, session->tsih);
			iscsi_release_session(session);
		}
		up(&devdata->session_sem);
	}
	
	bring_down_portals();

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_release, err %d\n", err);

	return err;
}

/* print socket connect/disconnect message */
void
report_peer_up_down(struct sockaddr *ip_address, char *message)
{
	char ip_string[INET6_ADDRSTRLEN+2], port_string[8];

	if (cnv_inet_to_string(ip_address, ip_string, port_string) > 0) {
		printk("%s %s %s:%s\n", current->comm, message,
													ip_string, port_string);
	}
}

/*
 * generate the next TTT in a session
 * must be called with session->cmnd_sem lock held
 */
static inline __u32 __attribute__ ((no_instrument_function))
generate_next_ttt(struct iscsi_session* session)
{
	__u32 retval;

	retval = session->cmnd_id++;
	if (session->cmnd_id == 0) {
		/* just assigned -1 to retval, which is reserved
		 * so do it again, since 0 is not reserved
		 */
		retval = session->cmnd_id++;
	}
	return retval;
}

/*
 * allocates all the structures necessary for a new connection and a
 * new session.  If later we find out that this connection belongs to
 * an existing session, the session structure alloced here will be
 * freed.  Its just easier to do everything here in one place, and it
 * makes it easier to recover from errors if everything is all set up
 * (i.e., there are no partial structures)
 * returns NULL if anything fails, in which case everything is freed
 *		   pointer to new iscsi_conn, which points to new iscsi_sess, if all ok
 */
static struct iscsi_conn *
build_conn_sess(struct socket *sock, struct portal_group *ptr)
{
	struct iscsi_conn *conn;
	struct iscsi_session *session;
	int addr_len;

	conn = (struct iscsi_conn *)my_kmalloc(sizeof(struct iscsi_conn),
										   "iscsi_conn");
	if (!conn) {
		goto out1;
	}

	TRACE(TRACE_DEBUG, "new conn %p for sock %p\n", conn, sock);

	memset(conn, 0, sizeof(struct iscsi_conn));

	INIT_LIST_HEAD(&conn->conn_link);

	INIT_LIST_HEAD(&conn->reject_list);
	init_MUTEX(&conn->reject_sem);

	conn->active = 1;
	conn->conn_id = ++devdata->conn_id;
	conn->conn_socket = sock;
	conn->dev = devdata;
	conn->max_send_length = 8192;
	conn->max_recv_length = 8192;
	conn->portal_group_tag = ptr->tag;
	conn->connection_flags = devdata->force;

	conn->nop_period = devdata->nop_period;
	init_timer(&conn->nop_timer);

	init_MUTEX_LOCKED(&conn->tx_sem);
	init_MUTEX_LOCKED(&conn->kill_rx_sem);
	init_MUTEX_LOCKED(&conn->kill_tx_sem);
	init_MUTEX(&conn->text_in_progress_sem);

	/* get ip addressing info about both ends of this connection */
	if ((conn->ip_address = my_kmalloc(sizeof(struct sockaddr), "ip address"))
								== NULL) {
		goto out2;
	}
	addr_len = ptr->ip_length;		/* value-result parameter */
	if (sock->ops->getname(sock, conn->ip_address, &addr_len, 1) < 0) {
		TRACE_ERROR("%s Could not get peer name for socket %p\n", current->comm,
					sock);
		goto out3;
	}

	if ((conn->local_ip_address = my_kmalloc(sizeof(struct sockaddr),
											 "local ip address")) == NULL) {
		goto out3;
	}
	addr_len = ptr->ip_length;		/* value-result parameter */
	if (sock->ops->getname(sock, conn->local_ip_address, &addr_len, 0) < 0) {
		TRACE_ERROR("%s Could not get local name for socket %p\n",
					current->comm, sock);
		goto out4;
	}

	session = (struct iscsi_session *)my_kmalloc(sizeof(struct iscsi_session),
												 "session");
	if (!session) {
			goto out4;
	}

	memset(session, 0, sizeof(struct iscsi_session));

	INIT_LIST_HEAD(&session->sess_link);
	INIT_LIST_HEAD(&session->conn_list);
	list_add_tail(&conn->conn_link, &session->conn_list);

	conn->session = session;
	session->nconn = 1;
	session->devdata = devdata;
	session->portal_group_tag = ptr->tag;
	session->version_max = ISCSI_MAX_VERSION;
	session->version_min = ISCSI_MIN_VERSION;

	if ((session->r2t_timer = (struct timer_list *)
				my_kmalloc(sizeof(struct timer_list), "r2t timer")) == NULL) {
				goto out5;
	}
	init_timer(session->r2t_timer);
	TRACE(TRACE_DEBUG, "Allocated R2T timer %p for session %p\n",
		  session->r2t_timer, session);

	if (!(session->session_params = my_kmalloc(MAX_CONFIG_PARAMS *
											   sizeof(struct parameter_type),
											   "session_params"))) {
		goto out6;
	}

	if (!(session->oper_param = my_kmalloc(MAX_CONFIG_PARAMS *
								sizeof(struct session_operational_parameters),
								"oper_param"))) {
		goto out7;
	}

	/* copy the parameters information from the global structure */
	param_tbl_cpy(*session->session_params, *devdata->param_tbl);
	session->r2t_period = devdata->r2t_period;

	/* Store SNACK flags as part of Session - SAI */
	session->targ_snack_flg = devdata->targ_snack_flg;

	init_MUTEX(&session->cmnd_sem);
	init_MUTEX_LOCKED(&session->retran_sem);
	init_MUTEX_LOCKED(&session->thr_kill_sem);

	/* all ok */
	goto out;

out7:
	my_kfree((void**)&session->session_params, "session_params");

out6:
	TRACE(TRACE_DEBUG, "Releasing R2T timer %p for session %p\n",
		  session->r2t_timer, session);
	my_kfree((void**)&session->r2t_timer, "r2t timer");

out5:
	my_kfree((void**)&session, "session");

out4:
	my_kfree((void**)&conn->local_ip_address, "local ip address");
	
out3:
	my_kfree((void**)&conn->ip_address, "ip address");

out2:
	my_kfree((void**)&conn, "iscsi_conn");
	
out1:
	sock_release(sock);

out:
	return conn;
}

/*
 * iscsi_server_thread: This thread is responsible for accepting
 * connections
 */
int
iscsi_server_thread(void *param)
{
	struct iscsi_conn *new_conn;
	struct socket *newsock, *sock;
	struct portal_group *ptr = param;
	long i;

	i = ptr - iscsi_portal_groups;
	
	lock_kernel();

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	daemonize(
#else
	daemonize();
	snprintf(current->comm, sizeof(current->comm),
#endif
			 "iscsi_thr_%ld", i);

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 18)
	/*
	 * Arne Redlich, agr1@users.sourceforge.net:
	 * This prevents the server thread from becoming a zombie after it exits.
	 */
	reparent_to_init();
#endif

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGBITS);

	/* mark that this server thread is alive */
	devdata->server_thr[i] = current;

	unlock_kernel();

	printk("%s Starting pid %d\n", current->comm, current->pid);

	sock = devdata->server_socket[i];

	printk("%s Listening on %s:%s\n", current->comm,
											ptr->ip_string,
											ptr->port_string);

	/* notify our parent that this thread is up */
	up(&devdata->server_sem);

	/* loop forever accepting new connections from initiators */
	for ( ; ; ) {

	/* fix for 2.6.10 contributed by Charles Coffing, ccoffing@novell.com */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 9)
		newsock = sock_alloc();
		if (!newsock)
#else
		int ret;

		ret = sock_create_kern(sock->sk->sk_family, sock->type,
							   sock->sk->sk_protocol, &newsock);
		if (ret)
#endif
		{
			TRACE_ERROR("%s Could not allocate socket for accept\n",
						current->comm);
			goto iscsi_server_thread_out;
		}

		newsock->type = sock->type;
		newsock->ops = sock->ops;

		TRACE(TRACE_NET, "before accept: sock %p newsock %p\n", sock, newsock);

		if (sock->ops->accept(sock, newsock, 0) < 0) {
			goto out1;
		}

		TRACE(TRACE_NET, "after accept: sock %p newsock %p\n", sock, newsock);

		/* turn off the Nagle Algorithm on this socket */
		tcp_nagle_off(newsock);

		/* clear bad sessions and connections here */
		clean_bad_stuff();

		/* allocate and fill in connection details */
		if ((new_conn = build_conn_sess(newsock, ptr)) == NULL)
			goto iscsi_server_thread_out;

		TRACE(TRACE_ISCSI, "Connect socket %p on conn %p, cid %d, session %p\n",
			  newsock, new_conn, new_conn->conn_id, new_conn->session);
		/* 
		 * NOTE THAT WE CANNOT DO MUCH OF ANYTHING WITH THIS
		 * CONNECTION AS WE DO NOT KNOW WHAT SESSION IT BELONGS
		 * TO. THIS WILL HAVE TO BE A PART OF THE RX_THREAD
		 */

		/* start the tx thread */
		if (kernel_thread(iscsi_tx_thread, (void *)new_conn, 0) < 0) {
			TRACE_ERROR("%s unable to create tx_thread\n", current->comm);
			goto out1;
		}

		/* start the rx thread */
		if (kernel_thread(iscsi_rx_thread, (void *)new_conn, 0) < 0) {
			TRACE_ERROR("%s unable to create rx_thread\n", current->comm);
			if (new_conn->tx_thread) {
				/*  Mike Christie mikenc@us.ibm.com */
				send_sig(ISCSI_SHUTDOWN_SIGNAL, new_conn->tx_thread, 1);
			}
			goto out1;
		}
	}

out1:
	sock_release(newsock);

iscsi_server_thread_out:

	/* mark that this server thread is no longer alive */
	devdata->server_thr[i] = NULL;

	printk("%s closed %s:%s\n", current->comm,
							ptr->ip_string,
							ptr->port_string);

	/* exit this thread after notifying whoever might be waiting */
	up(&devdata->server_sem);
	printk("%s Exiting pid %d\n", current->comm, current->pid);
	return 0;
}

/*
 * for each session on the list,
 * release inactive connections and the session itself if no conns remain.
 * devdata->session_sem MUST be locked before this routine is called.
 */
static void __attribute__ ((no_instrument_function))
clean_session_list(struct list_head *list)
{
	struct iscsi_session *session;
	struct iscsi_conn *conn;
	struct list_head *list_ptr1, *list_temp1;
	struct list_head *list_ptr2, *list_temp2;

	list_for_each_safe(list_ptr1, list_temp1, list) {
		session = list_entry(list_ptr1, struct iscsi_session, sess_link);

		list_for_each_safe(list_ptr2, list_temp2, &session->conn_list) {
			conn = list_entry(list_ptr2, struct iscsi_conn, conn_link);

			if (conn->active == 0) {
				if (iscsi_release_connection(conn) < 0) {
					TRACE_ERROR("%s Error releasing connection\n",
								current->comm);
				}
			}
		}

		if (list_empty(&session->conn_list)) {
			iscsi_release_session(session);
		}
	}
}

static void __attribute__ ((no_instrument_function))
clean_bad_stuff(void)
{
	/* destructive access to session lists */
	if (!down_interruptible(&devdata->session_sem)) {
		clean_session_list(&devdata->bad_session_list);
		clean_session_list(&devdata->session_list);
		up(&devdata->session_sem);
	}
}

/*
 * iscsi_release_session: This function is responsible for closing out
 * a session and removing it from whatever list it is on.
 * host->session_sem MUST be locked before this routine is called.
 * INPUT: session to be released
 * OUTPUT: 0 if success, < 0 if there is trouble
 */
int
iscsi_release_session(struct iscsi_session *session)
{
	struct iscsi_cmnd *cmnd;
	struct iscsi_conn *conn;
	struct list_head *list_ptr, *list_temp;

	if (!session) {
		TRACE_ERROR("%s Cannot release a NULL session\n", current->comm);
		return -1;
	}

	if (TRACE_TEST(TRACE_ISCSI)) {
		print_isid_tsih_message(session, "Release session with ");
	}

	TRACE(TRACE_DEBUG, "Releasing R2T timer %p for session %p\n",
		  session->r2t_timer, session);

	/* Delete r2t timer - SAI */
	if (session->r2t_timer) {
		TRACE(TRACE_DEBUG, "Deleting r2t timer %p\n", session->r2t_timer);
		del_timer_sync(session->r2t_timer);
		TRACE(TRACE_DEBUG, "Deleted r2t timer\n");
		my_kfree((void**)&session->r2t_timer, "r2t timer");
	}

	/* free commands */
	while ((cmnd = session->cmnd_list)) {
		session->cmnd_list = cmnd->next;

		if (cmnd->cmnd != NULL) {
			if (scsi_release(cmnd->cmnd) < 0) {
				TRACE_ERROR("%s Trouble releasing command, opcode 0x%02x, "
							"ITT %u, state 0x%x\n", current->comm,
							cmnd->opcode_byte, cmnd->init_task_tag,
							cmnd->state);
			}
		}
		/* free data_list if any, cdeng */
		free_data_list(cmnd);
		my_kfree((void**)&cmnd->ping_data, "data_buf");
		my_kfree((void**)&cmnd, "iscsi_cmnd");
	}

	/* free connections */
	list_for_each_safe(list_ptr, list_temp, &session->conn_list) {
		conn = list_entry(list_ptr, struct iscsi_conn, conn_link);
		TRACE(TRACE_ISCSI, "iscsi%d: releasing connection %d\n", (int)
			  session->devdata->device->id, conn->conn_id);

		if (iscsi_release_connection(conn) < 0) {
			TRACE_ERROR("%s Trouble releasing connection\n", current->comm);
		}
	}

	/* dequeue session if it is linked into some list */
	if (!list_empty(&session->sess_link)) {
		list_del(&session->sess_link);

		/* error recovery ver new 18_04 - SAI */
		if (session->retran_thread) {
			/*  Mike Christie mikenc@us.ibm.com */
			send_sig(ISCSI_SHUTDOWN_SIGNAL, session->retran_thread, 1);
			down_interruptible(&session->thr_kill_sem);
		}
	}

	/* free session structures */
	my_kfree((void**)&session->session_params, "session_params");

	my_kfree((void**)&session->oper_param, "oper_param");

	my_kfree((void**)&session, "session");

	return 0;
}

/*
 * called by both rx and tx threads, but only after the rx thread has been
 * started on a connection (so the connected message has been given).
 * if socket for this conn has not already been released, do it now and
 * print a message that says we disconnected.
 * Also NULL out the conn's socket so we don't try to do this again later.
 */
static void
iscsi_release_socket(struct iscsi_conn *conn)
{
	if (conn->conn_socket != NULL) {
		/* Release socket */
		TRACE(TRACE_ISCSI, "Release socket %p on conn %p, cid %d, session %p\n",
			  conn->conn_socket, conn, conn->conn_id, conn->session);

		report_peer_up_down(conn->ip_address, "disconnected from");

		sock_release(conn->conn_socket);
		conn->conn_socket = NULL;
	}
}

/*
 * iscsi_release_connection: releases all the stuff related to a conn.
 * Right now, it does not do anything for releasing commands specific
 * to a connection since commands are more specific to a session as
 * against a connection. It also dequeues the connection from the list
 */
static int
iscsi_release_connection(struct iscsi_conn *conn)
{
	if (!conn) {
		TRACE_ERROR("%s Cannot release NULL connection/n", current->comm);
		return -1;
	}

	/* dequeue any unsent reject messages (do not send them) */
	dequeue_reject(conn, 0);

	/* Kill the rx and tx thread */

	if (conn->rx_thread) {
		/*  Mike Christie mikenc@us.ibm.com */
		TRACE(TRACE_ISCSI, "Releasing iscsi_rx_%d\n", conn->conn_id);
		send_sig(ISCSI_SHUTDOWN_SIGNAL, conn->rx_thread, 1);
		down_interruptible(&conn->kill_rx_sem);
	}

	if (conn->tx_thread) {
		TRACE(TRACE_ISCSI, "Releasing iscsi_tx_%d\n", conn->conn_id);
		/*  Mike Christie mikenc@us.ibm.com */
		send_sig(ISCSI_SHUTDOWN_SIGNAL, conn->tx_thread, 1);
		down_interruptible(&conn->kill_tx_sem);
	}

	/* Release socket */
	iscsi_release_socket(conn);

	TRACE(TRACE_ISCSI_FULL, "Dequeue connection conn->cid %u\n", conn->conn_id);

	/* Dequeue connection */
	list_del(&conn->conn_link);
	conn->session->nconn--;

	/* Free connection */
	my_kfree((void **)&conn->local_ip_address, "local ip address");
	my_kfree((void **)&conn->ip_address, "ip address");
	del_timer_sync(&conn->nop_timer);
	my_kfree((void**)&conn, "conn");

	return 0;
}

/*
 * executed only by the rx thread.
 * Allocates new reject item, fills it in with header of rejected pdu,
 * and enqueues it for tx thread on reject_list of this connection.
 */
int
enqueue_reject(struct iscsi_conn *conn, __u8 reason)
{
	int err = 0;
	struct reject_item	*new_item;

	TRACE(TRACE_ENTER_LEAVE, "Enter enqueue_reject, reason %u\n", reason);

	if ((new_item = (struct reject_item *)my_kmalloc(sizeof(struct reject_item),
													 "reject item")) == NULL) {
		err = -1;
		goto out;
	}
	memcpy(new_item->bad_header, conn->bad_hdr, ISCSI_HDR_LEN);
	new_item->reason = reason;
	if (down_interruptible(&conn->reject_sem)) {
		my_kfree((void **)&new_item, "reject item");
		err = -1;
		goto out;
	}
	list_add_tail(&new_item->reject_link, &conn->reject_list);
	up(&conn->reject_sem);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		up(&conn->tx_sem);
	}
out:
	TRACE(TRACE_ENTER_LEAVE, "Leave enqueue_reject, err %d\n", err);

	return err;
}

/*
 * executed only by the tx thread.
 * locks connection's queue of pending rejects, then takes each item off
 * queue in order and uses it to generate and send a Reject pdu to initiator.
 */
static int
dequeue_reject(struct iscsi_conn *conn, int sendit)
{
	int err = 0;
	struct reject_item	*item;
	struct list_head *list_ptr, *list_temp;

	if (down_interruptible(&conn->reject_sem)) {
		err = -1;
		goto out;
	}

	list_for_each_safe(list_ptr, list_temp, &conn->reject_list) {
		item = list_entry(list_ptr, struct reject_item, reject_link);
		if (sendit)
			err = iscsi_tx_rjt(conn, item->bad_header, item->reason);
		list_del(list_ptr);
		my_kfree((void **)&item, "reject item");
		if (err)
			break;
	}
	up(&conn->reject_sem);
out:
	return err;
}

/* 
 * iscsi_rx_thread: This thread is responsible for receiving iSCSI PDUs
 * and messages from the Initiator and then figuring out what to do with
 * them
 */
int
iscsi_rx_thread(void *param)
{
	struct iscsi_conn *conn = (struct iscsi_conn *) param;
	struct iovec iov;				/* for reading header and header digest */
	struct targ_error_rec err_rec;
	__u8 buffer[ISCSI_HDR_LEN];		/* read 48-byte PDU header in here */
	__u32 digest;					/* read 4-byte header digest in here */
	__u32 hdr_crc;					/* compute PDU header digest in here */
	__u32 opcode;					/* extract 6-bit PDU opcode into here */
	__u32 local_itt;				/* convert 4-byte PDU ITT into here */
	int err;

	lock_kernel();

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	daemonize(
#else
	daemonize();
	sprintf(current->comm,
#endif

			"iscsi_rx_%d", conn->conn_id);

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 18)
	/*
	 * Arne Redlich, agr1@users.sourceforge.net:
	 * This prevents the rx_thread from becoming a zombie after it exits.
	 */
	reparent_to_init();
#endif

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGBITS);

	/* mark that this rx thread is alive */
	conn->rx_thread = current;

	unlock_kernel();

	printk("%s Starting pid %d\n", current->comm, current->pid);

	report_peer_up_down(conn->ip_address, "connected to");

	/* receive loop */
	while (1) {
		iov.iov_len = ISCSI_HDR_LEN;
		iov.iov_base = buffer;

		/* receive iSCSI header */
		err = iscsi_rx_data(conn, &iov, 1, ISCSI_HDR_LEN);
		if (err != ISCSI_HDR_LEN) {
			goto iscsi_rx_thread_out;
		}

		TRACE_BUFFER(TRACE_BUF, buffer, ISCSI_HDR_LEN,
					 "%s Got PDU header\n", current->comm);

		opcode = buffer[0] & ISCSI_OPCODE;

		local_itt = ntohl(((struct generic_pdu *)buffer)->init_task_tag);

		if (!(conn->session->tsih)) {
			/* not in full feature phase yet, only accept Login request pdus */
			if (opcode != ISCSI_INIT_LOGIN_CMND) {
				TRACE_ERROR("%s ITT %u has opcode 0x%02x, expected 0x%02x\n",
							current->comm, local_itt, opcode,
							ISCSI_INIT_LOGIN_CMND);
				/* MUST immediately terminate the connection */
				goto iscsi_rx_thread_out;
			}

			TRACE(TRACE_ISCSI, "%s Got login request, ITT %u\n",
				  current->comm, local_itt);

			if (handle_login(conn, buffer) < 0) {
				TRACE_ERROR("%s Trouble in handle_login\n", current->comm);
				goto iscsi_rx_thread_out;
			}
			continue;
		}

		/* keep a copy of this header in case we need it in a later reject */
		memcpy(conn->bad_hdr, buffer, ISCSI_HDR_LEN);

		/* connection belongs to session that is in full feature phase */
		if (conn->hdr_crc) {
			/* read and check header digest on this pdu */
			TRACE(TRACE_DEBUG, "iscsi_rx_thread: Header digest check\n");
			hdr_crc = 0;
			do_crc(buffer, ISCSI_HDR_LEN, &hdr_crc);

			iov.iov_len = CRC_LEN;
			iov.iov_base = &digest;

			if ((err = iscsi_rx_data(conn, &iov, 1, CRC_LEN)) != CRC_LEN) {
				goto iscsi_rx_thread_out;
			}

			if (hdr_crc != digest) {
				TRACE_ERROR("%s Expected header crc 0x%08x, got 0x%08x\n",
							current->comm, ntohl(hdr_crc), ntohl(digest));
				/* Error Recovery Code Added by SAI */
				TRACE(TRACE_ERROR_RECOVERY,
					  "Start header digest error recovery\n");
				err_rec.curr_conn = conn;
				err_rec.pdu_hdr = (struct generic_pdu *)buffer;
				err_rec.pdu_hdr->length = be32_to_cpu(err_rec.pdu_hdr->length);
				err_rec.err_type = HEADER_DIGERR;
				if ((err = targ_do_error_recovery(&err_rec)) < 0)
					goto iscsi_rx_thread_out;
				continue;
			} else {
				TRACE(TRACE_ISCSI_FULL, "Got header crc 0x%08x\n",
					  ntohl(digest));
			}
		}

		switch (opcode) {

		case ISCSI_INIT_LOGIN_CMND:
			{
				/* got Login request while in full feature phase */
				TRACE_ERROR("%s Got login request ITT %u in full feature "
							"phase\n", current->comm, local_itt);
				goto iscsi_rx_thread_out;
			}

		case ISCSI_INIT_TEXT_CMND:
			{
				TRACE(TRACE_ISCSI, "%s Got text request, ITT %u\n",
					  current->comm, local_itt);

				if (handle_text_request(conn, conn->session, buffer) < 0) {
					TRACE_ERROR("%s Trouble in handle_text_request, ITT %u\n",
								current->comm, local_itt);
					goto iscsi_rx_thread_out;
				}

				break;
			}

		case ISCSI_INIT_SCSI_CMND:
			{
				TRACE(TRACE_ISCSI, "%s Got SCSI command, CmdSN %u, ITT %u\n",
					  current->comm,
					  ntohl(((struct generic_pdu *)buffer)->cmd_sn),
					  local_itt);

				if (handle_cmnd(conn, conn->session, buffer) < 0) {
					TRACE_ERROR("%s Trouble in handle_cmnd, ITT %u\n",
								current->comm, local_itt);
					goto iscsi_rx_thread_out;
				}

				break;
			}

		case ISCSI_INIT_SCSI_DATA_OUT:
			{
				TRACE(TRACE_ISCSI, "%s Got data-out, ITT %u, offset %u\n",
					  current->comm, local_itt,
					  ntohl(((struct generic_pdu *) buffer)->offset));

				if (handle_data(conn, conn->session, buffer) < 0) {
					TRACE_ERROR("%s Trouble in handle_data, ITT %u\n",
								current->comm, local_itt);
					goto iscsi_rx_thread_out;
				}

				break;
			}

		case ISCSI_INIT_TASK_MGMT_CMND:
			{
				TRACE(TRACE_ISCSI,
					  "%s Got task mgt command, ITT %u\n",
					  current->comm, local_itt);

				if (handle_task_mgt_command(conn, conn->session, buffer) < 0) {
					TRACE_ERROR("%s Trouble in handle_task_mgt_cmnd, ITT %u\n",
								current->comm, local_itt);
					goto iscsi_rx_thread_out;
				}

				break;
			}

		case ISCSI_INIT_LOGOUT_CMND:
			{
				TRACE(TRACE_ISCSI, "%s Got logout request, ITT %u\n",
					  current->comm, local_itt);

				if (handle_logout(conn, conn->session, buffer) < 0) {
					TRACE_ERROR("%s Trouble in handle_logout, ITT %u\n",
								current->comm, local_itt);
					goto iscsi_rx_thread_out;
				}

				break;
			}

		case ISCSI_INIT_NOP_OUT:
			{
				TRACE(TRACE_ISCSI, "%s Got NOP_OUT, ITT %u\n",
					  current->comm, local_itt);

				if (handle_nopout(conn, conn->session, buffer) < 0) {
					TRACE_ERROR("%s Trouble in handle_nopout, ITT %u\n",
								current->comm, local_itt);
					goto iscsi_rx_thread_out;
				}

				break;
			}

			/* SNACK Handling by Target - SAI */
		case ISCSI_INIT_SNACK:
			{
				TRACE(TRACE_ERROR_RECOVERY, "%s Got SNACK Request\n",
					  current->comm);

				if (handle_snack(conn, conn->session, buffer) < 0) {
					TRACE_ERROR("%s Trouble in handle_snack\n", current->comm);
					goto iscsi_rx_thread_out;
				}

				break;
			}

		case ISCSI_TARG_NOP_IN:
		case ISCSI_TARG_SCSI_RSP:
		case ISCSI_TARG_TASK_MGMT_RSP:
		case ISCSI_TARG_LOGIN_RSP:
		case ISCSI_TARG_TEXT_RSP:
		case ISCSI_TARG_SCSI_DATA_IN:
		case ISCSI_TARG_LOGOUT_RSP:
		case ISCSI_TARG_R2T:
		case ISCSI_TARG_ASYNC_MSG:
		case ISCSI_TARG_RJT:
			{
				TRACE_ERROR("%s Initiator sent a target opcode %.2x, ITT %u\n",
							current->comm, opcode, local_itt);
				/*****
				TRACE_SET(TRACE_DEBUG | TRACE_ISCSI | TRACE_ERROR_RECOVERY
							  | TRACE_ISCSI_FULL | TRACE_ENTER_LEAVE);
				*****/

				/* Send a Reject PDU and escalate to session recovery */
				enqueue_reject(conn, REASON_PROTOCOL_ERR);
				targ_session_recovery(conn);
				goto iscsi_rx_thread_out;
			}

		default:
			{
				TRACE_ERROR("%s Got unknown opcode %.2x, ITT %u\n",
							current->comm, opcode, local_itt);
				/*****
				TRACE_SET(TRACE_DEBUG | TRACE_ISCSI | TRACE_ERROR_RECOVERY
							  | TRACE_ISCSI_FULL | TRACE_ENTER_LEAVE);
				*****/

				/* Send a Reject PDU and escalate to session recovery */
				enqueue_reject(conn, REASON_COMMAND_NOT_SUPPORTED);
				targ_session_recovery(conn);
				goto iscsi_rx_thread_out;
			}
		} /* switch */

		/* in case there are any out-of-order commands that are now in-order */
		check_queued_cmnd(conn->session);

	} /* while(1) */

iscsi_rx_thread_out:
	conn->active = 0;
	conn->rx_thread = NULL;
	if (conn->tx_thread) {
		/* tx thread still going, shut it down */
		send_sig(ISCSI_SHUTDOWN_SIGNAL, conn->tx_thread, 1);
	} else {
		/* tx thread has shut down, if socket is still open, release it */
		iscsi_release_socket(conn);
	}
	up(&conn->kill_rx_sem);
	printk("%s Exiting pid %d\n", current->comm, current->pid);
	return 0;
}

static void
deal_with_nop_timer(unsigned long data);

static void
restart_nop_timer(struct iscsi_conn *conn)
{
	conn->nop_timer.expires = jiffies + conn->nop_period;
	conn->nop_timer.data = (unsigned long) conn;
	conn->nop_timer.function = deal_with_nop_timer;
	add_timer(&conn->nop_timer);
}

/* called by kernel when nop periodic timer expires on a connection
 * Always turn on the silence bit,
 * and also turn on the need nopin bit if silence bit was on before the set.
*/
static void
deal_with_nop_timer(unsigned long data)
{
	struct iscsi_conn *conn;

	conn = (struct iscsi_conn *)data;
	if (test_and_set_bit(SILENCE_BIT, &conn->control)) {
		set_bit(NEED_NOPIN_BIT, &conn->control);
	}
	restart_nop_timer(conn);
	if (atomic_read(&conn->tx_sem.count) <= 0) {
		up(&conn->tx_sem);
	}
}

/*
 * iscsi_tx_thread: This thread is responsible for the transmission of
 * responses and/or status information to the initiator
 */
int
iscsi_tx_thread(void *param)
{
	struct iscsi_conn *conn = (struct iscsi_conn *) param;
	struct iscsi_session *session;
	struct iscsi_cmnd *cmnd, *prev_cmnd;
	int count, skipover;
#if defined(CONFIG_ISCSI_DEBUG)
	__u32 save_mask;
#endif

	lock_kernel();

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	daemonize(
#else
	daemonize();
	snprintf(current->comm, sizeof(current->comm),
#endif
			 "iscsi_tx_%d", conn->conn_id);

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 18)
	/*
	 * Arne Redlich, agr1@users.sourceforge.net:
	 * This prevents the tx_thread from becoming a zombie after it exits.
	 */
	reparent_to_init();
#endif

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGBITS);

	/* mark that this tx thread is alive */
	conn->tx_thread = current;

	unlock_kernel();

	printk("%s Starting pid %d\n", current->comm, current->pid);

	if (conn->nop_period) {
		/* set up to start the nopin timer first time through the main loop */
		set_bit(NEED_NOPIN_BIT, &conn->control);
		set_bit(SILENCE_BIT, &conn->control);
	}

	TRACE_GET(save_mask);

	while (1) {
		/* block until somebody adds/changes a command for this conn */
		if (down_interruptible(&conn->tx_sem))
			goto iscsi_tx_thread_out;

		/* can't set session sooner because it is not set up in conn until
		 * AFTER login is complete on this connection.  But nothing will wake
		 * up this tx thread until AFTER that login is complete, so we are safe.
		 */
		session = conn->session;

		TRACE(TRACE_DEBUG, "%s awake, session %p, conn %p\n", current->comm,
			  session, conn);

		skipover = 0;				/* don't skip anything yet */

		TRACE_GET(save_mask);

		if (test_and_clear_bit(NEED_NOPIN_BIT, &conn->control)) {
			/* bit was set, either first time through
			 * or timer expired after an interval of silence from initiator
			 */
			if (!timer_pending(&conn->nop_timer)) {
				/* first time thru here -- timer not going yet, start it up */
				restart_nop_timer(conn);
			} else {
				/* not first time thru here, timer must have expired */
				if ((count = atomic_read(&conn->outstanding_nopins))) {
					TRACE_ERROR("%s No reply to %u outstanding NopIns\n",
								current->comm, count);
					if (count >= MAX_OUTSTANDING_NOPINS) {
						TRACE_ERROR("%s Max outstanding NopIns exceeded\n",
									current->comm);
						goto iscsi_tx_thread_out;
					}
					TRACE_SET(save_mask | TRACE_DEBUG | TRACE_ISCSI
							  | TRACE_ISCSI_FULL | TRACE_ENTER_LEAVE);
				}
				if (generate_nopin(conn, session) < 0)
						goto iscsi_tx_thread_out;
				atomic_inc(&conn->outstanding_nopins);
			}
		}

restart_after_dequeue:	/* back here after dequeueing a command from the list,
						 * at which time skipover will be set so we can skip
						 * ahead to our place in the list (which may have been
						 * changed by other tx threads while we were busy).
						 */
		if (!list_empty(&conn->reject_list)) {
			/* process connection's queue of pending rejects */
			if (dequeue_reject(conn, 1))
				goto iscsi_tx_thread_out;
		}

		/* lock the session-wide list of commands */
		if (down_interruptible(&session->cmnd_sem))
			goto iscsi_tx_thread_out;

		/* look at and count each command belonging to this session */
		count = 0;
		prev_cmnd = NULL;
		cmnd = session->cmnd_list;
		while (cmnd != NULL) {
			TRACE(TRACE_DEBUG, "%s pick up cmnd %p\n", current->comm, cmnd);
			if (cmnd->conn == conn && ++count >= skipover) {
			  /* this command is for this connection, handle it */
			  if (!(conn->connection_flags & CONN_LOGGED_OUT)) {
				TRACE(TRACE_DEBUG, "%s handle cmnd no. %d, ITT %u, "
					  "opcode 0x%02x, state %d\n",
					  current->comm, count, cmnd->init_task_tag,
					  cmnd->opcode_byte, cmnd->state);
				TRACE(TRACE_DEBUG, "%s ImmData %u, UnsolData %u, data_len %u, "
					  "data_done %u, r2t_data %d\n",
					  current->comm,
					  cmnd->immediate_data_present,
					  cmnd->unsolicited_data_present,
					  cmnd->data_length,
					  cmnd->data_done,
					  cmnd->r2t_data);
				switch (cmnd->state) {
				case ISCSI_SEND_TEXT_RESPONSE:
					{
						up(&session->cmnd_sem);
						if (handle_discovery_rsp(cmnd, conn, session) < 0) {
							TRACE_ERROR("%s Trouble in handle_discovery_rsp\n",
										current->comm);
							goto iscsi_tx_thread_out;
						}
						if (down_interruptible(&session->cmnd_sem))
							goto iscsi_tx_thread_out;
						break;
					}

				case ISCSI_ASK_FOR_MORE_TEXT:
					{
						up(&session->cmnd_sem);
						if (ask_for_more_text(cmnd, conn, session) < 0) {
							TRACE_ERROR("%s Trouble in ask_for_more_text\n",
										current->comm);
							goto iscsi_tx_thread_out;
						}
						if (down_interruptible(&session->cmnd_sem))
							goto iscsi_tx_thread_out;
						break;
					}

				case ISCSI_LOGOUT:
					{
						/* take this command out of the session-wide list */
						if (prev_cmnd == NULL)
							session->cmnd_list = cmnd->next;
						else
							prev_cmnd->next = cmnd->next;
						up(&session->cmnd_sem);
						if (handle_logout_rsp(cmnd, conn, session) < 0) {
							TRACE_ERROR("%s Trouble in handle_logout_rsp\n",
										current->comm);
						}
						/* always exit the tx thread after a logout response */
						iscsi_dequeue(cmnd, conn);
						goto iscsi_tx_thread_out;
					}

				case ISCSI_PING:
					{
						up(&session->cmnd_sem);
						if (handle_nopin(cmnd, conn, session) < 0) {
							TRACE_ERROR("%s Trouble in handle_nopin\n",
										current->comm);
							goto iscsi_tx_thread_out;
						}
						if (down_interruptible(&session->cmnd_sem))
							goto iscsi_tx_thread_out;
						break;
					}

				case ISCSI_DONE:
					{
						up(&session->cmnd_sem);
						if (handle_iscsi_done(cmnd, conn, session) < 0) {
							TRACE_ERROR("%s Trouble in handle_iscsi_done\n",
										current->comm);
							goto iscsi_tx_thread_out;
						}
						if (down_interruptible(&session->cmnd_sem))
							goto iscsi_tx_thread_out;
						break;
					}

				case ISCSI_RESEND_STATUS:
					{
						up(&session->cmnd_sem);
						if (send_iscsi_response(cmnd, conn, session) < 0) {
							TRACE_ERROR("%s Trouble in send_iscsi_response\n",
										current->comm);
							goto iscsi_tx_thread_out;
						}
						if (down_interruptible(&session->cmnd_sem))
							goto iscsi_tx_thread_out;
						break;
					}

				case ISCSI_MGT_FN_DONE:
					{
						up(&session->cmnd_sem);
						if (handle_iscsi_mgt_fn_done(cmnd, conn, session) < 0) {
							TRACE_ERROR("%s Trouble in iscsi_mgt_fn_done\n",
										current->comm);
							goto iscsi_tx_thread_out;
						}
						if (down_interruptible(&session->cmnd_sem))
							goto iscsi_tx_thread_out;
						break;
					}

				case ISCSI_BUFFER_RDY:
					{
						up(&session->cmnd_sem);
						if (iscsi_tx_r2t(cmnd, conn, session) < 0) {
							TRACE_ERROR("%s Trouble in iscsi_tx_r2t\n",
										current->comm);
							goto iscsi_tx_thread_out;
						}
						if (down_interruptible(&session->cmnd_sem))
							goto iscsi_tx_thread_out;
						break;
					}

				case ISCSI_DEQUEUE:
					{
						TRACE(TRACE_DEBUG, "dequeue command, ITT %u, CmndSN %u,"
							  " count %u, skipover %u\n", cmnd->init_task_tag,
							  cmnd->cmd_sn, count, skipover);
						/* take this command out of the session-wide list */
						if (prev_cmnd == NULL)
							session->cmnd_list = cmnd->next;
						else
							prev_cmnd->next = cmnd->next;
						up(&session->cmnd_sem);
						iscsi_dequeue(cmnd, conn);
						skipover = count;
						goto restart_after_dequeue;
					}

				case ISCSI_QUEUE_CMND_RDY:
					{
						if (send_unsolicited_data(cmnd, conn, session) < 0) {
							TRACE_ERROR("%s Trouble in send_unsolicited_data\n",
										current->comm);
							up(&session->cmnd_sem);
							goto iscsi_tx_thread_out;
						}
						break;
					}

				case ISCSI_QUEUE_CMND:
				case ISCSI_QUEUE_OTHER:
				case ISCSI_CMND_RECEIVED:
				case ISCSI_NEW_CMND:
				case ISCSI_SENT:
				case ISCSI_NOPIN_SENT:
				case ISCSI_ALL_R2TS_SENT:
				case ISCSI_IMMEDIATE_DATA_IN:
				case ISCSI_UNSOLICITED_DATA_IN:
				case ISCSI_DATA_IN:
				case ISCSI_BLOCKED_SENDING_TEXT:
				case ISCSI_AWAIT_MORE_TEXT:
					{
						/* Not much to do */
						break;
					}

				default:		/* NEVER HERE */
					{
						up(&session->cmnd_sem);
						TRACE_ERROR("%s Unknown command state %u\n",
									current->comm, cmnd->state);
						goto iscsi_tx_thread_out;
					}
				} /* switch */
			  }
			} /* if */
			prev_cmnd = cmnd;
			cmnd = cmnd->next;
		} /* while */

		/* unlock the session-wide list of commands */
		up(&session->cmnd_sem);

		TRACE(TRACE_DEBUG, "%s handled %d commands\n", current->comm, count);

		TRACE_SET(save_mask);

		/*
		 * Moved check to see if connection is active here so that the
		 * tx_thread can transmit one last packet, i.e., the logout response
		 */
		if (!conn->active)
			goto iscsi_tx_thread_out;
	} /* while */

iscsi_tx_thread_out:
	TRACE_SET(save_mask);
	conn->active = 0;
	conn->tx_thread = NULL;
	del_timer_sync(&conn->nop_timer);
	if (conn->rx_thread) {
		/* rx thread still going, shut it down */
		send_sig(ISCSI_SHUTDOWN_SIGNAL, conn->rx_thread, 1);
	} else {
		/* rx thread has shut down, if socket is still open, release it */
		iscsi_release_socket(conn);
	}
	up(&conn->kill_tx_sem);
	printk("%s Exiting pid %d\n", current->comm, current->pid);
	return 0;
}

/* called after reading unsolicited data attached to a WRITE SCSI command pdu,
 * or solicited data attached to a DataOut pdu.
 */
static void __attribute__ ((no_instrument_function))
update_after_read(struct generic_pdu *hdr, struct iscsi_cmnd *cmnd)
{
	cmnd->data_done += hdr->length;
	cmnd->immediate_data_present = 0;
	if (hdr->flags & F_BIT) {
		/* end of this (solicited or unsolicited) sequence,
		 * reset the counters for the next solicited burst (if any)
		 */
		cmnd->data_sn = 0;
		cmnd->unsolicited_data_present = 0;
	}
}

/*
 * iscsi_rx_data: will receive a fixed amount of data.
 * INPUT: struct iscsi_conn (what connection), struct iovec, number of iovecs,
 * total amount of data in bytes
 * OUTPUT: > 0 total bytes read if everything is okay
 * 	       < 0 if trouble (-ECONNRESET means end-of-file)
 *		   = 0 for PAYLOAD_DIGERR
 */
int
iscsi_rx_data(struct iscsi_conn *conn, struct iovec *iov, int niov, int data)
{
	struct msghdr msg;
	int msg_iovlen;
	int rx_loop, total_rx;
	__u32 checksum, data_crc;
	int i;
	struct iovec *iov_copy, *iov_ptr;
	mm_segment_t oldfs;

	total_rx = 0;

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_rx_data, niov %d, data %d\n",
		  niov, data);

	if (!conn->conn_socket) {
		TRACE_ERROR("%s Transport endpoint is not connected\n", current->comm);
		total_rx = -ENOTCONN;		/* Transport endpoint is not connected */
		goto out;
	}

	iov_copy =
		(struct iovec *)my_kmalloc(niov * sizeof(struct iovec), "iovec");
	if (iov_copy == NULL) {
		total_rx = -ENOMEM;
		goto out;
	}

	while (total_rx < data) {
		/* get a clean copy of the original io vector to work with */
		memcpy(iov_copy, iov, niov * sizeof(struct iovec));
		msg_iovlen = niov;
		iov_ptr = iov_copy;

		if ((rx_loop = total_rx)) {
			/* have done only partial read, recalculate iov and niov -- cdeng */
			TRACE(TRACE_ISCSI,
				  "iscsi_rx_data: data %d, received so far %d, recompute iov\n",
				  data, total_rx);
			do {
				if (iov_ptr->iov_len <= rx_loop) {
					rx_loop -= iov_ptr->iov_len;
					iov_ptr++;
					msg_iovlen--;
				} else {
					iov_ptr->iov_base += rx_loop;
					iov_ptr->iov_len -= rx_loop;
					rx_loop = 0;
				}
			} while (rx_loop);
		}

		memset(&msg, 0, sizeof(struct msghdr));
		msg.msg_iovlen = msg_iovlen;
		msg.msg_iov = iov_ptr;

		oldfs = get_fs();
		set_fs(get_ds());
		rx_loop = sock_recvmsg(conn->conn_socket, &msg, (data - total_rx),
							   MSG_WAITALL);
		set_fs(oldfs);

		/* this receive from initiator broke the silence */
		clear_bit(SILENCE_BIT, &conn->control);

		if (rx_loop <= 0) {
			my_kfree((void **)&iov_copy, "iovec");
			total_rx = -ECONNRESET;		/* Connection reset by peer */
			goto out;
		}

		total_rx += rx_loop;
		TRACE(TRACE_DEBUG, "iscsi_rx_data: rx_loop %d total_rx %d\n", rx_loop,
			  total_rx);
	}

	my_kfree((void**)&iov_copy, "iovec");

	if (niov > 1 && conn->data_crc) {
		/* this really is for a data segment, and data digests are in effect */
		data_crc = 0;
		for (i = 0; i < niov - 1; i++)
			do_crc(iov[i].iov_base, iov[i].iov_len, &data_crc);
		checksum = *(__u32 *) (iov[niov - 1].iov_base);
		if (checksum != data_crc) {
			TRACE_ERROR("%s Got data crc 0x%08x, expected 0x%08x\n",
						current->comm, ntohl(checksum), ntohl(data_crc));
			/* Data Digest Error Recovery - SAI */
			total_rx = PAYLOAD_DIGERR;
			goto out;
		} else {
			TRACE(TRACE_ISCSI_FULL, "Got data crc 0x%08x\n", ntohl(checksum));
		}
	}
out:
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_rx_data, total_rx %d\n", total_rx);

	return total_rx;
}

/* read data segment from cmd's connection into a single buffer 
 * that is newly allocated to hold bufsize bytes of data
 * returns > 0 total number of bytes read (including pad & crc) if all ok,
 *		   = 0 if recovered from payload digest error ok
 *		   < 0 error
 * if ok, sets result to newly allocated buffer, else leaves result unchanged
 */
static int
read_single_data_seg(__u8 *buffer, struct iscsi_cmnd *cmd,
					 int bufsize, char **result)
{
	struct iovec iov[3];
	int size, niov = 1, padding, err;
	__u32 digest, pad_bytes;
	char *data_buf;
	struct targ_error_rec err_rec;

	size = bufsize;

	padding = (-size) & 3;
	if (padding) {
		iov[niov].iov_base = &pad_bytes;
		iov[niov].iov_len = padding;
		size += padding;
		niov++;
	}

	if (cmd->conn->data_crc) {
		iov[niov].iov_base = &digest;
		iov[niov].iov_len = CRC_LEN;
		size += CRC_LEN;
		niov++;
	}

	if ((data_buf = (char *)my_kmalloc(bufsize, "data_buf"))) {
		iov[0].iov_base = data_buf;
		iov[0].iov_len = bufsize;

		err = iscsi_rx_data(cmd->conn, iov, niov, size);

		if (err != size) {
			/* Payload Digest Error Recovery - SAI */
			if (err == PAYLOAD_DIGERR) {
				TRACE(TRACE_ERROR_RECOVERY,
					  "Start payload digest error recovery\n");
				err_rec.curr_conn = cmd->conn;
				err_rec.pdu_hdr = (struct generic_pdu *)buffer;
				err_rec.cmd = cmd;
				err_rec.err_type = PAYLOAD_DIGERR;
				err = targ_do_error_recovery(&err_rec);
			}
			my_kfree((void**)&data_buf, "data_buf");
		} else {
			*result = data_buf;
		}
	} else {
		err = -1;
	}
	return err;
}

/*
 * executed only by the rx thread.
 * read data segment for command into list of buffers at given offset
 * this can be either immediate data in a write, or the payload of a DataIn
 * returns > 0 if ok, < 0 if error, = 0 if error recovery completed ok
 */
static int
read_list_data_seg(struct generic_pdu *hdr, struct iscsi_cmnd *cmd,
				   struct scatterlist *st_list, int offset)
{
	struct targ_error_rec err_rec;
	struct iovec *iov;
	int size, niov, padding, err, sgindex;
	__u32 digest, pad_bytes;

	size = hdr->length;
	niov = find_iovec_needed(size, st_list, offset);
	if (niov <= 0) {
		TRACE_ERROR("%s Trouble in find_iovec_needed\n", current->comm);
		return -1;
	}

	/* allocate 2 extra iov slots for possible padding and crc */
	iov = (struct iovec *)my_kmalloc((niov + 2) * sizeof(struct iovec),
									 "iovec");
	if (!iov) {
		return -1;
	}

	/* gives back number of st_list elements used */
	sgindex = fill_iovec(iov, 0, niov, st_list, &offset, size);

	padding = (-size) & 3;
	if (padding) {
		iov[niov].iov_base = &pad_bytes;
		iov[niov].iov_len = padding;
		niov++;
		size += padding;
	}

	if (cmd->conn->data_crc) {
		iov[niov].iov_base = &digest;
		iov[niov].iov_len = CRC_LEN;
		niov++;
		size += CRC_LEN;
	}

	err = iscsi_rx_data(cmd->conn, iov, niov, size);

	if (err == size) {
		/* we received everything we expected to receive
		 * Store scatter list count and offset for recovery purposes - SAI
		 */
		cmd->scatter_list_count += sgindex;
		cmd->scatter_list_offset = offset;
		update_after_read(hdr, cmd);
	} else {
		/* Payload Digest Error Recovery - SAI */
		if (err == PAYLOAD_DIGERR) {
			TRACE(TRACE_ERROR_RECOVERY,
				  "Start payload digest error recovery\n");
			err_rec.curr_conn = cmd->conn;
			err_rec.pdu_hdr = hdr;
			err_rec.cmd = cmd;
			err_rec.err_type = PAYLOAD_DIGERR;
			err = targ_do_error_recovery(&err_rec);
		}
	}
	my_kfree((void**)&iov, "iovec");
	return err;
}

/*
 * executed only by the tx thread.
 * iscsi_tx_data: will transmit a fixed-size PDU of any type.
 * INPUT: struct iscsi_conn (what connection), struct iovec, number of iovecs,
 * total amount of data in bytes,
 * iovec[0] must be the 48-byte PDU header.
 * iovec[1] (if header digests are in use, calculate it in this function)
 * iovec[...] the data
 * iovec[niov-1] (if data digests are in use, calculate it in this function)
 * OUTPUT: int total bytes written if everything is okay
 * 	       < 0 if trouble
 */
int
iscsi_tx_data(struct iscsi_conn *conn, struct iovec *iov, int niov, int data)
{

# ifdef DEBUG_DATA
	int i, j;
	struct iovec *debug_iov;
	__u8 *to_print;
# endif

	struct msghdr msg;
	int msg_iovlen;
	int total_tx, tx_loop;
	__u32 hdr_crc, data_crc;
	int data_len, k;
	mm_segment_t oldfs;
	struct iovec *iov_copy, *iov_ptr;
	struct generic_pdu *pdu;

	if (!conn->conn_socket) {
		TRACE_ERROR("%s NULL conn_socket\n", current->comm);
		return -1;
	}
# ifdef DEBUG_DATA
	TRACE(TRACE_DEBUG, "iscsi_tx_data: iovlen %d\n", niov);

	debug_iov = iov;

	for (i = 0; i < niov; i++) {
		to_print = (__u8 *) debug_iov->iov_base;

		for (j = 0; ((j < debug_iov->iov_len) && (j < 64)); j++) {
			TRACE(TRACE_DEBUG, "%.2x ", to_print[j]);

			if (((j + 1) % 16) == 0)
				TRACE(TRACE_DEBUG, "\n");
			else if (((j + 1) % 4) == 0)
				TRACE(TRACE_DEBUG, "    ");
		}

		TRACE(TRACE_DEBUG, "\n");

		debug_iov++;
	}
# endif

	/* compute optional header digest */
	if (conn->hdr_crc) {
		hdr_crc = 0;
		do_crc(iov[0].iov_base, ISCSI_HDR_LEN, &hdr_crc);

		iov[1].iov_base = &hdr_crc;
		iov[1].iov_len = CRC_LEN;
		TRACE(TRACE_ISCSI_FULL, "Send header crc 0x%08x\n", ntohl(hdr_crc));
	}

	/* compute optional data digest */
	if (conn->data_crc && niov > conn->hdr_crc + 2) {
		data_len = 0;
		data_crc = 0;
		for (k = conn->hdr_crc + 1; k < niov - 1; k++) {
			do_crc(iov[k].iov_base, iov[k].iov_len, &data_crc);
			data_len += iov[k].iov_len;
		}

		iov[niov - 1].iov_base = &data_crc;
		iov[niov - 1].iov_len = CRC_LEN;
		TRACE(TRACE_ISCSI_FULL, "Send data len %d, data crc 0x%08x\n",
			  data_len, ntohl(data_crc));
	}

	iov_copy =
		(struct iovec *)my_kmalloc(niov * sizeof(struct iovec), "iovec");
	if (iov_copy == NULL)
		return -1;

	total_tx = 0;
	while (total_tx < data) {
		/* get a clean copy of the original io vector to work with */
		memcpy(iov_copy, iov, niov * sizeof(struct iovec));
		msg_iovlen = niov;
		iov_ptr = iov_copy;

		if ((tx_loop = total_tx)) {
			TRACE(TRACE_ISCSI,
				  "iscsi_tx_data: data %d not completed, recompute iov\n",data);
			do {
				if (iov_ptr->iov_len <= tx_loop) {
					tx_loop -= iov_ptr->iov_len;
					iov_ptr++;
					msg_iovlen--;
				} else {
					iov_ptr->iov_base += tx_loop;
					iov_ptr->iov_len -= tx_loop;
					tx_loop = 0;
				}
			} while (tx_loop);
			TRACE(TRACE_ISCSI, "sock_sendmsg total_tx %d, data %d, niov %d, "
							   "msg_iovlen %d\n",
							   total_tx, data, niov, msg_iovlen);
		}

		memset(&msg, 0, sizeof(struct msghdr));
		msg.msg_iov = iov_ptr;
		msg.msg_iovlen = msg_iovlen;
		msg.msg_flags = MSG_NOSIGNAL;

		TRACE(TRACE_DEBUG, "iscsi_tx_data: niov %d, data %d, total_tx %d\n",
			  niov, data, total_tx);

		oldfs = get_fs();
		set_fs(get_ds());
		tx_loop = sock_sendmsg(conn->conn_socket, &msg, (data - total_tx));
		set_fs(oldfs);

		if (tx_loop <= 0) {
			pdu = (struct generic_pdu *)iov[0].iov_base;
			TRACE_ERROR("%s sock_sendmsg error %d, total_tx %d, data %d, niov "
						"%d, msg_iovlen %d, op 0x%02x, flags 0x%02x, ITT %u\n",
						current->comm, tx_loop, total_tx, data, niov,
						msg_iovlen, pdu->opcode, pdu->flags,
						ntohl(pdu->init_task_tag));
			my_kfree((void**)&iov_copy, "iovec");
			return tx_loop;
		}

		total_tx += tx_loop;
		TRACE(TRACE_DEBUG, "iscsi_tx_data: tx_loop %d total_tx %d\n", tx_loop,
			  total_tx);
	}

	my_kfree((void**)&iov_copy, "iovec");
	return total_tx;
}

/*
 * executed only by the tx thread.
 * RDR
 */
static int
send_hdr_plus_1_data(struct iscsi_conn *conn, void *iscsi_hdr,
					 void *data_buf, int data_len)
{
	struct iovec iov[5];
	int niov, total_size, padding, retval;
	int pad_bytes = 0;
	
	/* set up the header in the first iov slot */
	iov[0].iov_base = iscsi_hdr;
	iov[0].iov_len = ISCSI_HDR_LEN;
	total_size = ISCSI_HDR_LEN;
	niov = 1;					/* one for the header */

	if (conn->hdr_crc) {
		/* set up the header digest in the second iov slot */
		iov[niov].iov_len = CRC_LEN;
		total_size += CRC_LEN;
		niov++;					/* one for header digest */
	}

	if (data_len) {
		/* set up the data in the next iov slot */
		iov[niov].iov_base = data_buf;
		iov[niov].iov_len = data_len;
		total_size += data_len;
		niov++;						/* one for data */

		padding = (-data_len) & 3;
		if (padding) {
			/* set up the data padding in the next iov slot */
			iov[niov].iov_base = &pad_bytes;
			iov[niov].iov_len = padding;
			total_size += padding;
			niov++;					/* one for padding */
			TRACE(TRACE_DEBUG, "padding attached: %d bytes\n", padding);
		}

		if (conn->data_crc) {
			/* set up the data digest in the next iov slot */
			iov[niov].iov_len = CRC_LEN;
			total_size += CRC_LEN;
			niov++;					/* one for data digest */
		}
	}

	retval = iscsi_tx_data(conn, iov, niov, total_size);

	if (retval != total_size) {
		TRACE_ERROR("%s Trouble in iscsi_tx_data, expected %d bytes, got %d\n",
					current->comm, total_size, retval);
		retval = -1;
	}

	return retval;
}

/*
 * executed only by the tx thread.
 * RDR
 */
static inline int
send_hdr_only(struct iscsi_conn *conn, void *iscsi_hdr)
{
	return send_hdr_plus_1_data(conn, iscsi_hdr, NULL, 0);
}

/*	Draft 20, Section 3.2.2.1 Command Numbering and Acknowledging
 *	...
 *	For non-immediate commands, the CmdSN field can take any value from
 *	ExpCmdSN to MaxCmdSN inclusive.  The target MUST silently ignore
 *	any non-immediate command outside of this range or non-immediate
 *	duplicates within the range.  The CmdSN carried by immediate commands
 *	may lie outside the ExpCmdSN to MaxCmdSN range.
 */
/*	RDR: check that cmd_sn is within the range [ExpCmdSN .. MaxCmdSN]
 *	returns 0 if equal to exp_cmd_sn (and exp_cmd_sn is updated)
 *			1 if greater than exp_cmd_sn (exp_cmd_sn not updated)
 *			-1 if outside the range (and hence should be ignored)
 *			0 for all immediate commands, regardless of cmd_sn value
 *  when called, session->cmnd_sem is locked so access to session-wide
 *  max_cmd_sn and exp_cmd_sn fields is atomic.
 */
static int
check_cmd_sn(struct iscsi_cmnd *cmnd, void *ptr, struct iscsi_session *session,
			 __u32 increment)
{
	struct generic_pdu *pdu = ptr;
	int delta;

	if (!(pdu->opcode & I_BIT)) {
		/* this is an non-immediate command */
		delta = session->max_cmd_sn - pdu->cmd_sn;
		if (delta < 0) {		/* cmd_sn is greater than MaxCmdSN */
			return -1;
		}

		delta = pdu->cmd_sn - session->exp_cmd_sn;
		if (delta < 0) {		/* cmd_sn is less than ExpCmdSN */
			return -1;
		}

		if (delta > 0) {		/* cmd_sn is in range, but out of order */
			return 1;
		}

		/* non-immediate command is in expected order */
		if (increment) {
			session->exp_cmd_sn++;
		} else {
			/* exp_cmd_sn is not incremented.
			 * This only happens on in-order, non-immediate SCSI command pdus,
			 * which don't bump exp_cmd_sn until after midlevel confirms
			 * command was delivered.  Set cmd_sn_increment to tell
			 * check_queued_cmnd() not to process this command once it has been
			 * queued!
			 */
			cmnd->cmd_sn_increment = 1;			/* increment exp_cmd_sn later */
		}
	}

	return 0;
}

/*
 * executed only by the rx thread.
 * find commands for this connection already in the queue that have already
 * sent everything and are being ACKed by this command in order to mark them
 * for dequeueing.
 * Then adds this command to end of queue if flag is set.
 * Finally, if any command's state was changed, or if this command was added
 * to the queue, wake up the tx thread.
 * NOTE: cmnd MAY be NULL, and if it is, add_cmnd_to_queue MUST be 0!
 * returns 0 on success, -1 on error.
 */
static int
ack_sent_cmnds(struct iscsi_conn *conn,
				  struct iscsi_cmnd *cmnd,
				  __u32 exp_stat_sn,
				  int add_cmnd_to_queue)
{
	struct iscsi_cmnd *temp;
	int delta, changed_something, count;

	changed_something = add_cmnd_to_queue;

	if (down_interruptible(&conn->session->cmnd_sem))
		return -1;

	count = 0;
	for (temp = conn->session->cmnd_list; temp != NULL; temp = temp->next) {
		if (temp->conn == conn) {
			count++;
			if (temp->state == ISCSI_SENT) {
				delta = temp->stat_sn - exp_stat_sn;
				if (delta < 0) {
					TRACE(TRACE_DEBUG, "set dequeue command statsn %d, "
						  "received exp_stat_sn %d, command state %d\n",
						  temp->stat_sn, exp_stat_sn, temp->state);
					temp->state = ISCSI_DEQUEUE;
					changed_something = 1;
				}
			}
		}

		if (temp->next == NULL)
			break;
	}

	if (add_cmnd_to_queue) {
		/* Add this command to the queue */
		TRACE(TRACE_DEBUG, "%s add command %p to queue, ITT %u, CmdSN %u, "
			  "state %d, count %d\n", current->comm, cmnd, cmnd->init_task_tag,
			  cmnd->cmd_sn, cmnd->state, count);
		count++;
		if (temp)
			temp->next = cmnd;
		else
			conn->session->cmnd_list = cmnd;
	}

	up(&conn->session->cmnd_sem);

	/* tell the tx thread it has something to do */
	if (changed_something) {
		if (atomic_read(&conn->tx_sem.count) <= 0) {
			TRACE(TRACE_DEBUG, "%s up(tx_sem), %d cmnds in queue\n",
				  current->comm, count);
			up(&conn->tx_sem);
		}
	}

	return 0;
}

/*
 * RDR
 * allocate and zero out a new struct iscsi_cmnd.
 * Returns pointer to the new struct if all ok, else NULL after error message
 */
struct iscsi_cmnd *
get_new_cmnd( void )
{
	struct iscsi_cmnd *cmnd;

	cmnd =
		(struct iscsi_cmnd *)my_kmalloc(sizeof(struct iscsi_cmnd), "cmnd");

	if (cmnd) {
		memset(cmnd, 0, sizeof(struct iscsi_cmnd));
		cmnd->state = ISCSI_NEW_CMND;
	}

	return cmnd;
}

/**************************************************************************/
/*                          handle income pdu                             */
/**************************************************************************/

/*
 * called only from rx thread.
 * handle_login: This function is responsible for interpreting the login
 * message and then making sure the necessary details get filled in, a
 * new session is opened if required by the login and the response is returned
 *
 * INPUT: Details of the connection, buffer containing the login
 * OUTPUT: 0 if everything is okay, < 0 if there is a problem
 * Note a 0 value returned may still mean that a connection is rejected
 */
static int
handle_login(struct iscsi_conn *conn, __u8 *buffer)
{
	struct iscsi_init_login_cmnd *pdu = (struct iscsi_init_login_cmnd *) buffer;
	struct iscsi_session *session, *temp;
	struct iscsi_conn *temp_conn;
	__u32 when_called;
	int found, retval = -1;
	struct auth_parameter_type auth_param;	/*chap and srp support - CHONG */
	struct parameter_type *p;
	struct parameter_type (*this_param_tbl)[MAX_CONFIG_PARAMS];
	struct parameter_type (*temp_params)[MAX_CONFIG_PARAMS];
	struct list_head *list_ptr;
	struct iscsi_global *host;

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_login_cmnd(pdu);

	/* this will be non-NULL if the parameter table should be freed when done */
	temp_params = NULL;

	if ((session = conn->session) == NULL) {
		/* should never happen */
		goto out;
	}

	if ((host = session->devdata) == NULL) {
		/* should never happen */
		goto out;
	}

	/* always use clean copy of configured parameter table for negotiations */
	this_param_tbl = session->session_params;

	pdu->length = be32_to_cpu(pdu->length);
	pdu->tsih = be16_to_cpu(pdu->tsih);
	pdu->init_task_tag = be32_to_cpu(pdu->init_task_tag);
	pdu->cid = be16_to_cpu(pdu->cid);
	pdu->cmd_sn = be32_to_cpu(pdu->cmd_sn);
	pdu->exp_stat_sn = be32_to_cpu(pdu->exp_stat_sn);

	/* destructive access to session lists */
	if (down_interruptible(&host->session_sem))
		goto err_conn_out;

	if (pdu->tsih == 0) {
		/* a new session, the iscsi_session structure is already set up */
		conn->cid = pdu->cid;
		conn->stat_sn = pdu->exp_stat_sn;
		session->cmd_sn = pdu->cmd_sn;
		session->exp_cmd_sn = pdu->cmd_sn;
		session->max_cmd_sn = pdu->cmd_sn + QUEUE_DEPTH_ALLOWED;

		memcpy(session->isid, pdu->isid, 6);

		/* set up the operational parameters from the global structure */
		set_session_parameters(session->oper_param, *session->session_params);

		/* add new session to the end of the global session list
		 * with tsih = 0 to indicate login not finished
		 */
		list_add_tail(&session->sess_link, &host->session_list);

		/* set flags to allow keys for new session (leading) login */
		when_called = LEADING_ONLY | INITIAL_ONLY | ALL;

		TRACE(TRACE_ISCSI, "%s Start new session %p\n",
			  current->comm, session);
	} else {
		/* existing session, check through the session list to find it */
		found = 0;
		list_for_each(list_ptr, &conn->dev->session_list) {
			session = list_entry(list_ptr, struct iscsi_session, sess_link);
			if (session->tsih == pdu->tsih) {
				found = 1;
				break;
			}
		}

		if (!found) {
			TRACE_ERROR("%s No existing session with TSIH %u, "
						"terminate this connection\n", current->comm,
						pdu->tsih);
			goto err_conn_out;
		}

		if (conn->portal_group_tag != session->portal_group_tag) {
			TRACE_ERROR("%s Portal group tag %u for new connection does "
						"not match portal group tag %u of session\n",
						current->comm, conn->portal_group_tag,
						session->portal_group_tag);

			iscsi_tx_login_reject(conn, pdu, STAT_CLASS_INITIATOR,
								  STAT_DETAIL_NOT_INCLUDED);
			goto err_conn_out;
		}

		/* check isid */
		if (memcmp(pdu->isid, session->isid, 6) != 0) {
			TRACE_ERROR("%s The session has a different ISID, "
						"terminate the connection\n", current->comm);

			iscsi_tx_login_reject(conn, pdu, STAT_CLASS_INITIATOR,
								  STAT_DETAIL_ERR);
			goto err_conn_out;
		}

		conn->cid = pdu->cid;
		conn->stat_sn = pdu->exp_stat_sn;

		/* check cid, and if it already exists then release old connection */
		list_for_each(list_ptr, &session->conn_list) {
			temp_conn = list_entry(list_ptr, struct iscsi_conn, conn_link);
			if (temp_conn->cid == conn->cid) {
				TRACE(TRACE_ISCSI, "connection reinstatement with cid %u\n",
					  conn->cid);

				if (iscsi_release_connection(temp_conn) < 0) {
					TRACE_ERROR("%s Error releasing connection\n",
								current->comm);
				}
				break;
			}
		}

		TRACE(TRACE_ISCSI, "new connection cid %u attached to "
				"existing session tsih %u, IP 0x%08x\n",
			  conn->cid, pdu->tsih,
			  ((struct sockaddr_in *)conn->ip_address)->sin_addr.s_addr);

		/* add new connection to end of connection list for existing session */
		temp = conn->session;
		conn->session = session;
		list_del(&conn->conn_link);
		temp->nconn = 0;
		list_add_tail(&conn->conn_link, &session->conn_list);
		session->nconn++;

		/* use clean parameter table for negotiations, free it later */
		temp_params = this_param_tbl;
		temp->session_params = NULL;

		/* free up the no-longer-needed session structure */
		iscsi_release_session(temp);

		/* set back the leading-only keys if these keys were set to
		 * " KEY_TO_BE_NEGOTIATED" in the leading connection negotation.*/
		reset_parameter_table(*this_param_tbl);

		/* set flags to allow keys for new connection (only) login */
		when_called = INITIAL_ONLY | ALL;
	}

	/* by the time we get here, the connection is attached to the correct
	 * session and that session is threaded onto the devdata->session_list.
	 */

	up(&host->session_sem);

	/* chap and srp support - CHONG */
	auth_param.auth_flags = 0;
	auth_param.chap_local_ctx =
		CHAP_CloneContext(conn->dev->auth_parameter.chap_local_ctx);
	auth_param.chap_peer_ctx =
		CHAP_CloneContext(conn->dev->auth_parameter.chap_peer_ctx);
	auth_param.srp_ctx = SRP_CloneContext(conn->dev->auth_parameter.srp_ctx);

	if ((p = find_flag_parameter(TARGETPORTALGROUPTAG_FLAG,
								 *conn->session->session_params))) {
		/* force the TargetPortalGroupTag key to have as its value the
		   portal group tag for this session */
		p->int_value = conn->session->portal_group_tag;
	}

	if (parameter_negotiate(conn, *this_param_tbl, pdu, when_called, auth_param)
							< 0) {
		TRACE_ERROR("%s Parameter negotiation failed\n", current->comm);
		/* chap and srp support - CHONG */
		CHAP_FinalizeContext(auth_param.chap_local_ctx);
		CHAP_FinalizeContext(auth_param.chap_peer_ctx);
		SRP_FinalizeContext(auth_param.srp_ctx);
		goto out;
	}

	/* Bjorn Thordarson, 10-May-2004 -- start -- */
	if (TRACE_TEST(TRACE_ISCSI_FULL)) {
		int k;

		for (k = 0; k < MAX_CONFIG_PARAMS ; k++) {
			if ((*this_param_tbl)[k].parameter_name != NULL &&
				(*this_param_tbl)[k].str_value != NULL)
				printk("PARAM: %s = %s\n", (*this_param_tbl)[k].parameter_name,
					   (*this_param_tbl)[k].str_value);
		}
	}
	/* Bjorn Thordarson, 10-May-2004 -- end -- */

	/* chap and srp support - CHONG */
	CHAP_FinalizeContext(auth_param.chap_local_ctx);
	CHAP_FinalizeContext(auth_param.chap_peer_ctx);
	SRP_FinalizeContext(auth_param.srp_ctx);

	conn->stat_sn++;

	if (conn->stat_sn < 0) {
		TRACE_ERROR("%s Parameter negotiation with error stat_sn\n",
					current->comm);
		goto out;
	}

	/* set the operational parameters to the negotiated value */
	if (pdu->tsih == 0) {
		/* this is a new connection in a new session */
		set_session_parameters(session->oper_param, *session->session_params);

		/* Initialize R2T timer - ref18_09c SAI */
		if (session->oper_param->ErrorRecoveryLevel > 0
			&& session->r2t_period > 0) {
			/* error recovery ver ref18_04 */

			/* create a retransmit_thread for handling error recovery */
			if (kernel_thread(iscsi_retran_thread, (void *)session, 0) < 0) {
				TRACE_ERROR("%s Unable to create retran_thread\n",
							current->comm);
				session->r2t_period = 0;
			} else {
				restart_r2t_timer(session);
				TRACE(TRACE_ISCSI, "R2T timer %p started for session %p\n",
					  session->r2t_timer, session);
			}
		}
	}

	/* we are now in Full Feature Phase */
	conn->hdr_crc = (conn->connection_flags & USE_HEADERDIGEST) ? 1 : 0;
	conn->data_crc = (conn->connection_flags & USE_DATADIGEST) ? 1 : 0;

	/* all ok */
	retval = 0;

out:
	/* here when nothing further can or needs to be done
	 * to the session or connection structs
	 */
	if (temp_params) {
		param_tbl_uncpy(*temp_params);
		my_kfree((void**)&temp_params, "temp_params");
	}

	return retval;

err_conn_out:
	/* here on a fatal error detected before session is linked into devdata */
	iscsi_release_socket(conn);

	/* put this session onto end of "bad-session" list for cleanup later */
	session = conn->session;
	TRACE(TRACE_DEBUG, "add to list bad session %p, conn %p\n",
		  session, conn);

	list_add_tail(&session->sess_link, &host->bad_session_list);

	up(&host->session_sem);

	goto out;
}

/* helper routine for print_expanded_address_any() */
static int
convert_and_print_ip_stuff(char *ptr, struct sockaddr *real_ip_address,
							struct portal_group *pg_ptr)
{
	char ip_string[INET6_ADDRSTRLEN+2], port_string[8];
	int k = 0;

	if (cnv_inet_to_string(real_ip_address, ip_string, port_string) >= 0 ) {
		k = 1 + sprintf(ptr, "TargetAddress=%s:%s,%u", ip_string, port_string,
						pg_ptr->tag);
		TRACE(TRACE_DEBUG, "Expand %s to %s\n", pg_ptr->ip_string, ptr);
	}
	return k;
}

/* called only from do_text_request()
 * when we find a portal_group entry with IP address INADDR_ANY
 * in order to expand it to all IP addresses on this host and print them
 * as values of TargetAddress== keys.
 *
 * this code is modelled on the functions inet_select_addr() and
 * inet_dump_ifaddr() in the file net/ipv4/devinet.c for IPV4; and on
 * the function ipv6_get_saddr() in the file net/ipv6/addrconf.c for IPV6.
 */
static int
print_expanded_address_any(char *ptr, struct portal_group *pg_ptr)
{
	struct net_device *dev;
	struct in_device *in_dev;
	struct in_ifaddr *ifa;
	struct inet6_dev *idev;
	struct inet6_ifaddr *ifp;
	struct sockaddr real_ip_address;
	struct sockaddr_in *iptr = (struct sockaddr_in *)&real_ip_address;
	struct sockaddr_in6 *i6ptr = (struct sockaddr_in6 *)&real_ip_address;
	int n = 0;

	if (pg_ptr->family != AF_INET && pg_ptr->family != AF_INET6)
		return n;

	/* search the list of all known network devices */
	read_lock(&dev_base_lock);

#ifdef CONFIG_IPV6
	if (pg_ptr->family == AF_INET6)
		read_lock(&addrconf_lock);
	else
#endif
		rcu_read_lock();

	for (dev = dev_base; dev != NULL; dev = dev->next) {
		memset(&real_ip_address, 0, pg_ptr->ip_length);
		real_ip_address.sa_family = pg_ptr->family;
		if (pg_ptr->family == AF_INET) {
			/* this is an IPv4 address */
			if ((in_dev = in_dev_get(dev)) == NULL)
				continue;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
			read_lock(&in_dev->lock);
#endif

			for (ifa = in_dev->ifa_list; ifa != NULL; ifa = ifa->ifa_next) {
				/* do not advertize the loopback address */
				if (ifa->ifa_local == htonl(INADDR_LOOPBACK))
					continue;
				iptr->sin_addr.s_addr = ifa->ifa_local;
				iptr->sin_port
						= ((struct sockaddr_in *)pg_ptr->ip_address)->sin_port;
				n += convert_and_print_ip_stuff(ptr + n, &real_ip_address,
												pg_ptr);
			}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
			read_unlock(&in_dev->lock);
#endif

		} else {
			/* this is an IPv6 address */
			if ((idev = __in6_dev_get(dev)) == NULL)
				continue;
			read_lock(&idev->lock);
			for (ifp = idev->addr_list; ifp != NULL; ifp = ifp->if_next) {
				/* do not advertize link-local or loopback addresses */
#ifdef CONFIG_IPV6
				if (ipv6_addr_type(&ifp->addr)
								& (IPV6_ADDR_LINKLOCAL | IPV6_ADDR_LOOPBACK))
					continue;
#endif
				ipv6_addr_copy(&i6ptr->sin6_addr, &ifp->addr);
				i6ptr->sin6_port
					= ((struct sockaddr_in6 *)pg_ptr->ip_address)->sin6_port;
				n += convert_and_print_ip_stuff(ptr + n, &real_ip_address,
												pg_ptr);
			}
			read_unlock(&idev->lock);
		}
		/* else ignore this address family */
	}

#ifdef CONFIG_IPV6
	if (pg_ptr->family == AF_INET6)
		read_unlock(&addrconf_lock);
	else
#endif
		rcu_read_unlock();

	read_unlock(&dev_base_lock);
	return n;
}

enum text_types {TEXT_EMPTY, TEXT_ST_ALL, TEXT_ST_TN, TEXT_ST_NULL, TEXT_OTHER};
/*
 *	parses data segment sent in a Text Request
 *	Returns 0 if ok, else reason code for reject
 *  If ok, sets result to
 *			TEXT_EMPTY		if no data segment
 *			TEXT_ST_ALL		if only "SendTargets=ALL" key
 *			TEXT_ST_TN		if only "SendTargets=<target-name>" key
 *			TEXT_ST_NULL	if only "SendTargets=" key
 *			TEXT_OTHER		if only other keys
 */
static int __attribute__ ((no_instrument_function))
parse_text_buffer(struct iscsi_cmnd *cmnd, __u8 discovery_session,
				  enum text_types *result)
{
	char *ptr, *ptr2, *equal, *which;
	int size, reason = 0;
	enum text_types text_type = TEXT_EMPTY;

	if (cmnd->ping_data) {
		/* size bytes of text data were read into data_buf */
		if ((size = cmnd->data_length) > 0 ) {
			for (ptr = cmnd->ping_data; size > 0; ptr = ptr2) {
				equal = NULL;
				for (ptr2 = ptr;
					*ptr2 != '\0' && size > 0;
					ptr2++, size--) {
					/* skip over key itself */
					if (*ptr2 == '=' && equal == NULL)
						equal = ptr2;
				}

				if (*ptr2 != '\0') {
					*ptr2 = '\0';
					TRACE_ERROR("%s Unterminated key \"%s\"\n", current->comm,
								ptr);
					reason = REASON_PROTOCOL_ERR;
					break;
				}

				if (equal == NULL) {
					TRACE_ERROR("%s Missing '=' in key \"%s\"\n", current->comm,
								ptr);
					reason = REASON_PROTOCOL_ERR;
					break;
				}

				if (discovery_session) {
					which = "discovery";
				} else {
					which = "text request";
				}
				printk("iscsi %s ITT %u %s\n", which, cmnd->init_task_tag, ptr);

				equal++;
				if (strncmp(ptr, "SendTargets=", 12) == 0) {
					if (text_type != TEXT_EMPTY) {
						TRACE_ERROR("SendTargets key not only key in text\n");
						reason = REASON_NEGOTIATION_RESET;
						break;
					}
					if (strcmp(equal, "All") == 0) {
						if (!discovery_session) {
							TRACE_ERROR("%s %s not allowed in Normal Session\n",
										current->comm, ptr);
							reason = REASON_NEGOTIATION_RESET;
							break;
						}
						text_type = TEXT_ST_ALL;
					} else if (*equal != '\0')
						text_type = TEXT_ST_TN;
					else
						text_type = TEXT_ST_NULL;
				} else {
					if (discovery_session) {
						TRACE_ERROR("%s %s not allowed in Discovery Session\n",
									current->comm, ptr);
						reason = REASON_NEGOTIATION_RESET;
						break;
					}
					text_type = TEXT_OTHER;
				}

				for (; *ptr2 == '\0' && size > 0; ptr2++, size--) {	
					/* skip over nulls between keys */
				}
			}
		}
	}
	*result = text_type;
	return reason;
}

/*
 * executed only by the rx thread.
 * connection's text_in_progress_sem MUST be locked by caller.
 * returns 0 if ok, != 0 is reason for reject
 */
static int
accumulate_text_input(struct iscsi_cmnd *cmnd, struct iscsi_cmnd *in_progress)
{
	__u32 size;

	if ((size = cmnd->first_burst_len)) {
		/* command has some data attached, must accumulate it */
		if (in_progress->in_progress_buffer == NULL) {
			in_progress->in_progress_buffer
						= (char *)my_kmalloc(MAX_TEXT_LEN, "text buf");
			if (in_progress->in_progress_buffer == NULL) {
				return REASON_OUT_OF_RESOURCES;
			}
		}

		size += in_progress->data_length;
		if (size > MAX_TEXT_LEN) {
			TRACE_ERROR("%s too many total bytes %u in text request, max %u\n",
						current->comm, size, MAX_TEXT_LEN);
			return REASON_OUT_OF_RESOURCES;
		}

		memcpy(in_progress->in_progress_buffer + in_progress->data_length,
			   cmnd->ping_data, cmnd->first_burst_len);
		in_progress->data_length = size;
		my_kfree((void**)&cmnd->ping_data, "data_buf");
	}
	return 0;
}

/*
 * executed only by the rx thread.
 * called when the Response to a Text Request should be generated.
 * connection's text_in_progress_sem MUST be locked by caller.
 */
static void
generate_text_response(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
					   struct iscsi_session *session)
{
	enum text_types text_type;
	char *buffer = NULL, *ptr, *ip_ptr, *equal;
	int reason, size = 0, i, j, n, maxt;
	struct portal_group *pg_ptr;

	reason = parse_text_buffer(cmnd, session->oper_param->SessionType,
							   &text_type);
	if (reason != 0) {
		goto outbad;
	}

	if (text_type == TEXT_EMPTY)
		goto outok;

	if (text_type == TEXT_OTHER) {
		TRACE_ERROR("%s Text Negotiation in Normal Session not implemented\n",
					current->comm);
		reason = REASON_NEGOTIATION_RESET;
		goto outbad;
	}

	/* now generate the SendTargets response in a new, "big" buffer */
	if ((buffer = (char *)my_kmalloc(MAX_TEXT_LEN, "discovery buf")) == NULL) {
		reason = REASON_OUT_OF_RESOURCES;
		goto outbad;
	}

	/* send the '\0' at the end of each key=value string too */
	ptr = buffer;
	maxt = MAX_TARGETS;
	if (maxt < target_count)
		maxt = target_count;
	for (i = 0; i < maxt; i++) {
		if (!target_in_use(i) || (text_type == TEXT_ST_NULL
									&& i != session->oper_param->TargetName))
			continue;
		n = 1 + sprintf(ptr, "TargetName=%s%d", TARGETNAME_HEADER, i);
		if (text_type == TEXT_ST_TN) {
			/* need to match the name in SendTargets=<target-name> */
			if ((equal = strchr(cmnd->ping_data, '=')) == NULL
								|| strcmp(equal, strchr(ptr, '=')) != 0) {
			continue;
			}
		}
		ptr += n;
		size += n;
		
		for (j = 0; j < MAX_PORTAL; j++) {
			
			pg_ptr = &iscsi_portal_groups[j];
			if (!pg_ptr->in_use)
				continue;
			
			if (size > (MAX_TEXT_LEN - TEXT_FUDGE_LEN)) {
				TRACE_ERROR("%s Text Response to SendTargets is %d bytes, "
							"max %d\n", current->comm, size,
							MAX_TEXT_LEN - TEXT_FUDGE_LEN);
				my_kfree((void **)&buffer, "discovery buf");
				reason = REASON_OUT_OF_RESOURCES;
				goto outbad;
			}
			/* expand the address "any" to all IP address on this host */
			ip_ptr = pg_ptr->ip_string;
			if (strcmp(ip_ptr, INADDR_ANY_STRING) == 0
								|| strcmp(ip_ptr, IN6ADDR_ANY_STRING) == 0) {
				n = print_expanded_address_any(ptr, pg_ptr);
			} else {
				n = 1 + sprintf(ptr, "TargetAddress=%s:%s,%u", ip_ptr,
								pg_ptr->port_string, pg_ptr->tag);
			}
			ptr += n;
			size += n;
		}
	}

outok:
	my_kfree((void**)&cmnd->ping_data, "data_buf");
	cmnd->ping_data = buffer;
	cmnd->data_length = size;
	cmnd->data_done = 0;
	cmnd->state = ISCSI_SEND_TEXT_RESPONSE;
	return;

outbad:
	cmnd->state = ISCSI_DEQUEUE;
	enqueue_reject(conn, reason);
	return;
}

static inline void __attribute__ ((no_instrument_function))
copy_in_progress_stuff(struct iscsi_cmnd *cmnd, struct iscsi_cmnd *in_progress)
{
	cmnd->state = ISCSI_DEQUEUE;
	in_progress->command_flags = cmnd->command_flags;
	in_progress->cmd_sn = cmnd->cmd_sn;
	in_progress->stat_sn = cmnd->stat_sn;
}

/*
 * executed only by the rx thread.
 * called when a Text Request is the next command to be processed.
 */
static void
do_text_request(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
				struct iscsi_session *session)
{
	int reason;
	struct iscsi_cmnd *in_progress;
	char *which;

	if (down_interruptible(&conn->text_in_progress_sem)) {
		return;
	}

	if (cmnd->init_task_tag == ALL_ONES) {
		TRACE_ERROR("%s Text Request with reserved ITT=0x%08x\n",
					current->comm, ALL_ONES);
		goto outbadprotocol;
	}

	/*	RFC 3720 Section 10.10.2 C (Continue) Bit
	 *	"A Text Request with the C bit set to 1 MUST have the F bit set to 0."
	 */
	if ((cmnd->command_flags & (C_BIT | F_BIT)) == (C_BIT | F_BIT)) {
		TRACE_ERROR("%s Text Request with ITT %u has C=1 and F=1\n",
					current->comm, cmnd->init_task_tag);
		goto outbadprotocol;
	}

	in_progress = (struct iscsi_cmnd *)conn->text_in_progress;
	if (in_progress) {
		/* Text negotiation already in progress, must be same task:
		 *
		 * RFC 3720 Section 10.10.3 Initiator Task Tag
		 * "If the command is sent as part of a sequence of text requests
		 * and responses, the Initiator Task Tag MUST be the same for all the
		 * requests within the sequence (similar to linked SCSI commands).
		 * The I bit for all requests in a sequence also MUST be the same.
		 */
		if (cmnd->init_task_tag != in_progress->init_task_tag) {
			TRACE_ERROR("%s Text Request has ITT=%u, expected %u\n",
					current->comm, cmnd->init_task_tag,
					in_progress->init_task_tag);
			goto outbadprotocol;
		}

		if (cmnd->opcode_byte != in_progress->opcode_byte) {
			TRACE_ERROR("%s Text Request with ITT %u has I=%u, expected %u\n",
					current->comm, cmnd->init_task_tag,
					(cmnd->opcode_byte & I_BIT) == I_BIT,
					(in_progress->opcode_byte & I_BIT) == I_BIT);
			goto outbadprotocol;
		}

		/*
		 * RFC 3720 Section 10.10.4 Target Transfer Tag
		 * "When the Target Transfer Tag is set to the reserved value
		 * 0xffffffff, it tells the target that this is a new request and
		 * the target resets any internal state associated with the
		 * Initiator Task Tag (resets the current negotiation state)
		 *
		 * The target sets the Target Transfer Tag in a text response to
		 * a value other than the reserved value 0xffffffff whenever it
		 * indicates that it has more data to send or more operations to
		 * perform that are associated with the specified Initiator Task
		 * Tag.  It MUST do so whenever it sets the F bit to 0 in the
		 * response.  By copying the Target Transfer Tag from the response
		 * to the next Text Request, the initiator tells the target to
		 * continue the operation for the specific Initiator Task Tag.
		 */
		if (cmnd->target_xfer_tag == ALL_ONES) {
			/* initiator is resetting an in-progress text exchange
			 * just forget everything that happened up to this point
			 */
			if (session->oper_param->SessionType)
				which = "discovery";
			else
				which = "text request";
			printk("iscsi %s ITT %u reset\n", which,in_progress->init_task_tag);
			in_progress->state = ISCSI_DEQUEUE;
			in_progress->init_task_tag = ALL_ONES;
			conn->text_in_progress = in_progress = NULL;
			/* now fall thru to start again as if a new command */
		} else {
			/* initiator is continuing an in-progress text exchange
			 * the Target Transfer Tag must match the one we sent earlier
			 */
			if (cmnd->target_xfer_tag != in_progress->target_xfer_tag) {
				TRACE_ERROR("%s Text Request with ITT %u has TTT=%u, "
							"expected %u\n", current->comm,
							cmnd->init_task_tag, cmnd->target_xfer_tag,
							in_progress->target_xfer_tag);
				goto outbadprotocol;
			}
			
			if (in_progress->state == ISCSI_BLOCKED_SENDING_TEXT) {
				/* The last Text Response we sent had C = 1 or this connection
				 * has USE_ONE_KEY_PER_TEXT set.  We expect this Text Request
				 * should not contain any data and should have its C = 0
				 */
				if (cmnd->first_burst_len) {
					TRACE_ERROR("%s Text Request with ITT %u TTT %u has DSL=%u,"
								" expected 0\n", current->comm,
								cmnd->init_task_tag, cmnd->target_xfer_tag,
								cmnd->first_burst_len);
					goto outbadprotocol;
				}
				if (cmnd->command_flags & C_BIT) {
					TRACE_ERROR("%s Text Request with TTT %u has C=1, "
								"expected 0\n", current->comm,
								cmnd->target_xfer_tag);
					goto outbadprotocol;
				}
				/* ok to send our next text response pdu (done with this pdu) */
				in_progress->state = ISCSI_SEND_TEXT_RESPONSE;
				copy_in_progress_stuff(cmnd, in_progress);
				goto out;
			}
			/* Our last Text Response must have had C = 0 (empty or complete) */
			if (in_progress->state != ISCSI_AWAIT_MORE_TEXT) {
				/* this should never happen */
				TRACE_ERROR("%s Text in_progress with ITT %u has state=%u, "
							"expected %u\n", current->comm,
							in_progress->init_task_tag,
							in_progress->state, ISCSI_AWAIT_MORE_TEXT);
				reason = REASON_OUT_OF_RESOURCES;
				goto outbad;
			}
		}
	} else if (cmnd->target_xfer_tag != ALL_ONES) {
		TRACE_ERROR("%s Text Request with ITT %u has TTT=%u, "
					"expected 0xffffffff\n", current->comm,
					cmnd->init_task_tag, cmnd->target_xfer_tag);
		goto outbadprotocol;
	}

	if (!in_progress) {
		in_progress = cmnd;
	}
	if ((reason = accumulate_text_input(cmnd, in_progress)))
		goto outbad;

	if (in_progress == cmnd) {
		conn->text_in_progress = in_progress;
	} else {
		copy_in_progress_stuff(cmnd, in_progress);
	}
	if (cmnd->command_flags & C_BIT) {
		in_progress->state = ISCSI_ASK_FOR_MORE_TEXT;
	} else {
		/* have accumulated a complete Text Request, generate response to it */
		in_progress->ping_data = in_progress->in_progress_buffer;
		in_progress->in_progress_buffer = NULL;
		generate_text_response(in_progress, conn, session);
	}

out:
	up(&conn->text_in_progress_sem);
	return;

outbadprotocol:
	reason = REASON_PROTOCOL_ERR;
outbad:
	cmnd->state = ISCSI_DEQUEUE;
	cmnd->init_task_tag = ALL_ONES;
	enqueue_reject(conn, reason);
	goto out;
}

/*
 * executed only by the rx thread.
 * Called when a Text Request pdu has just been received.
 * Currently it only supports the SendTarget command.
 * INPUT: connection, session and the buffer containing the text pdu.
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble.
 */
static int
handle_text_request(struct iscsi_conn *conn,
				 struct iscsi_session *session,
				 __u8 *buffer)
{
	struct iscsi_init_text_cmnd *pdu = (struct iscsi_init_text_cmnd *)buffer;
	struct iscsi_cmnd *cmnd;
	int err;

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_text_cmnd(pdu);

	pdu->length = be32_to_cpu(pdu->length);
	pdu->init_task_tag = be32_to_cpu(pdu->init_task_tag);
	pdu->target_xfer_tag = be32_to_cpu(pdu->target_xfer_tag);
	pdu->cmd_sn = be32_to_cpu(pdu->cmd_sn);
	pdu->exp_stat_sn = be32_to_cpu(pdu->exp_stat_sn);

	if( (cmnd = get_new_cmnd()) == NULL) {
		return -1;
	}

	cmnd->conn = conn;
	cmnd->session = session;
	cmnd->opcode_byte = pdu->opcode;
	cmnd->command_flags = pdu->flags;
	cmnd->first_burst_len = pdu->length;
	cmnd->init_task_tag = pdu->init_task_tag;
	cmnd->target_xfer_tag = pdu->target_xfer_tag;
	cmnd->cmd_sn = pdu->cmd_sn;
	cmnd->stat_sn = pdu->exp_stat_sn;

	if (pdu->length > 0) {
		/* read length bytes of text data into a newly-allocated buffer */
		err = read_single_data_seg(buffer, cmnd, pdu->length, &cmnd->ping_data);
		if (err <= 0) {
			my_kfree((void**)&cmnd, "cmnd");
			return err;
		}
	} else if (conn->text_in_progress == NULL ||
						 ((struct iscsi_cmnd *)conn->text_in_progress)->state
												!= ISCSI_BLOCKED_SENDING_TEXT) {
		/* this should probably be removed after testing */
		TRACE_WARNING("%s Empty Text Request, CmdSN %u, ExpCmdSN "
					  "%u, ITT %u, opcode 0x%02x\n", current->comm,
					  cmnd->cmd_sn, session->exp_cmd_sn,
					  cmnd->init_task_tag, cmnd->opcode_byte);
	}

	if (down_interruptible(&session->cmnd_sem)) {
		my_kfree((void *)&cmnd->ping_data, "data_buf");
		my_kfree((void **)&cmnd, "cmnd");
		return -1;
	}

	err = check_cmd_sn(cmnd, pdu, session, 1);

	up(&session->cmnd_sem);

	if (err < 0) {
		/* out of range, silently ignore it */
		TRACE_ERROR("%s ignoring out of range CmdSN %u, ExpCmdSN %u, ITT %u, "
					"opcode 0x%02x\n", current->comm,
					cmnd->cmd_sn, session->exp_cmd_sn,
					cmnd->init_task_tag, cmnd->opcode_byte);
		ack_sent_cmnds(conn, cmnd, pdu->exp_stat_sn, 0);
		my_kfree((void *)&cmnd->ping_data, "data_buf");
		my_kfree((void**)&cmnd, "cmnd");
	} else {
		if (err == 0) {
			/* do immediate command or expected non-immediate command now */
			do_text_request(cmnd, conn, session);
		} else {
			/* within range but out of order, queue it for later */
			cmnd->state = ISCSI_QUEUE_OTHER;
		}

		ack_sent_cmnds(conn, cmnd, pdu->exp_stat_sn, 1);
	}
	return 0;
}

/*
 * executed only by the rx thread.
 * called when a Logout Request has just been received.
 * INPUT: connection, session and the buffer containing the logout pdu
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
handle_logout(struct iscsi_conn *conn,
			  struct iscsi_session *session,
			  __u8 *buffer)
{
	struct iscsi_init_logout_cmnd *pdu =
		(struct iscsi_init_logout_cmnd *) buffer;
	struct iscsi_cmnd *cmnd;
	int err;

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_logout_cmnd(pdu);

	pdu->length = be32_to_cpu(pdu->length);
	pdu->init_task_tag = be32_to_cpu(pdu->init_task_tag);
	pdu->cid = be16_to_cpu(pdu->cid);
	pdu->cmd_sn = be32_to_cpu(pdu->cmd_sn);
	pdu->exp_stat_sn = be32_to_cpu(pdu->exp_stat_sn);

	printk("Logout ITT %u, CmdSN %u, reason %u, cid %u\n",
			pdu->init_task_tag, pdu->cmd_sn, pdu->flags & LOGOUT_REASON,
			pdu->cid);

	if ((cmnd = get_new_cmnd()) == NULL) {
		return -1;
	}

	cmnd->conn = conn;
	cmnd->session = session;
	cmnd->opcode_byte = pdu->opcode;
	cmnd->data_length = pdu->length;
	cmnd->init_task_tag = pdu->init_task_tag;
	cmnd->cmd_sn = pdu->cmd_sn;
	cmnd->stat_sn = pdu->exp_stat_sn;

	if (pdu->length > 0) {
		/* read length bytes of text data into a newly-allocated buffer */
		TRACE_ERROR("%s Data attached to logout request, CmdSN %u, ExpCmdSN "
					"%u, ITT %u, opcode 0x%02x\n", current->comm,
					cmnd->cmd_sn, session->exp_cmd_sn,
					cmnd->init_task_tag, cmnd->opcode_byte);
		err = read_single_data_seg(buffer, cmnd, pdu->length, &cmnd->ping_data);
		if (err <= 0) {
			my_kfree((void**)&cmnd, "cmnd");
			return err;
		}
	}

	if (down_interruptible(&session->cmnd_sem)) {
		my_kfree((void *)&cmnd->ping_data, "data_buf");
		my_kfree((void**)&cmnd, "cmnd");
		return -1;
	}

	err = check_cmd_sn(cmnd, pdu, session, 1);
	up(&session->cmnd_sem);

	if (err < 0) {	
		/* out of range, silently ignore it */
		TRACE_ERROR("%s ignoring out of range CmdSN %u, ExpCmdSN %u, ITT %u, "
					"opcode 0x%02x\n", current->comm,
					cmnd->cmd_sn, session->exp_cmd_sn,
					cmnd->init_task_tag, cmnd->opcode_byte);
		ack_sent_cmnds(conn, cmnd, pdu->exp_stat_sn, 0);
		my_kfree((void *)&cmnd->ping_data, "data_buf");
		my_kfree((void**)&cmnd, "cmnd");
	} else {
		if (err == 0) {		
			/* do immediate command or expected non-immediate command now */
			cmnd->state = ISCSI_LOGOUT;
		} else {
			/* within range but out of order, queue it for later */
			cmnd->state = ISCSI_QUEUE_OTHER;
		}

		ack_sent_cmnds(conn, cmnd, pdu->exp_stat_sn, 1);
	}
	return 0;
}

/* 
 * executed only by the rx thread.
 * called when a NopIn has just been received.
 * INPUT: connection, session and the buffer containing the nopin pdu
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
handle_nopout(struct iscsi_conn *conn,
			  struct iscsi_session *session,
			  __u8 *buffer)
{
	struct iscsi_init_nopout *pdu = (struct iscsi_init_nopout *) buffer;
	struct iscsi_cmnd *cmnd;
	int err;

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_nopout(pdu);

	pdu->length = be32_to_cpu(pdu->length);
	pdu->init_task_tag = be32_to_cpu(pdu->init_task_tag);
	pdu->target_xfer_tag = be32_to_cpu(pdu->target_xfer_tag);
	pdu->cmd_sn = be32_to_cpu(pdu->cmd_sn);
	pdu->exp_stat_sn = be32_to_cpu(pdu->exp_stat_sn);

	if (pdu->init_task_tag == ALL_ONES) {
		/* Draft 20, Section 10.18.1 Initiator Task Tag
		 * "If the Initiator Task Tag contains 0xffffffff, the I bit MUST be
		 * set to 1 and the CmdSN is not advanced after this PDU is sent."
		 */
		if (!(pdu->opcode & I_BIT)) {
			TRACE_ERROR("%s NopIn with ITT 0x%08x but I bit not set\n",
						current->comm, ALL_ONES);
			pdu->opcode |= I_BIT;	/* so CmdSN is not advanced below */
		}
		if (pdu->length) {
			/* should not have ping data if not expecting a response */
			TRACE_ERROR("%s NopIn with ITT 0x%08x but DSL %u\n",
						current->comm, ALL_ONES, pdu->length);
		}
	}

	if (pdu->target_xfer_tag != ALL_ONES) {
		/* this is a reply to earlier NopIn sent from target */
		/* should now find that existing command in our queue */
		cmnd = search_tags(conn, pdu->init_task_tag, pdu->target_xfer_tag, 0);
		if (!cmnd) {
			TRACE_ERROR("%s No command found for NopIn with TTT 0x%08x\n",
						current->comm, pdu->target_xfer_tag);
		} else {
			/* this is the reply we expected.  Check the LUN and then free
			 * the cmnd (which should not have any ping data attached to it).
			 */
			 cmnd->state = ISCSI_DEQUEUE;
			 atomic_dec(&conn->outstanding_nopins);
		}
		TRACE(TRACE_ISCSI, "Got NopIn ping reply, TTT %u, %u nopins left\n",
			  pdu->target_xfer_tag, atomic_read(&conn->outstanding_nopins));
	}

	if ((cmnd = get_new_cmnd()) == NULL) {
		return -1;
	}

	cmnd->state = ISCSI_PING;
	cmnd->conn = conn;
	cmnd->session = session;
	cmnd->opcode_byte = pdu->opcode;
	cmnd->data_length = pdu->length;
	cmnd->init_task_tag = pdu->init_task_tag;
	cmnd->target_xfer_tag = ALL_ONES;
	cmnd->cmd_sn = pdu->cmd_sn;
	cmnd->stat_sn = pdu->exp_stat_sn;

	if (pdu->length > 0 ) {
		/* read length bytes of nopout data into a newly-allocated buffer */
		err = read_single_data_seg(buffer, cmnd, pdu->length, &cmnd->ping_data);
		if (err <= 0) {
			my_kfree((void**)&cmnd, "cmnd");
			return err;
		}
	}

	if (down_interruptible(&session->cmnd_sem)) {
		my_kfree((void *)&cmnd->ping_data, "data_buf");
		my_kfree((void**)&cmnd, "cmnd");
		return -1;
	}

	err = check_cmd_sn(cmnd, pdu, session, 1);
	up(&session->cmnd_sem);

	if (err < 0 || (err == 0 && pdu->init_task_tag == ALL_ONES)) {
		/* nopout is out of range, or is in order but is not used as a
		 * ping request (so no reply NopIn expected), silently ignore it
		 */
		ack_sent_cmnds(conn, cmnd, pdu->exp_stat_sn, 0);
		if (err < 0) {
			/* out of range, silently ignore it */
			TRACE_ERROR("%s ignoring out of range CmdSN %u, ExpCmdSN %u, "
						"ITT %u, opcode 0x%02x\n", current->comm,
						cmnd->cmd_sn, session->exp_cmd_sn,
						cmnd->init_task_tag, cmnd->opcode_byte);
		} else if (!(cmnd->opcode_byte & I_BIT)) {
			/* the NopOut pdu was not immediate, CmdSN was advanced */
			session->max_cmd_sn++;
		}
		TRACE(TRACE_DEBUG, "tossing CmdSN %u, ExpCmdSN %u, ITT %u "
							"opcode 0x%02x\n",
							cmnd->cmd_sn, session->exp_cmd_sn,
							cmnd->init_task_tag, cmnd->opcode_byte);
		my_kfree((void *)&cmnd->ping_data, "data_buf");
		my_kfree((void**)&cmnd, "cmnd");
	} else {
		if (err > 0) {
			/* within range but out of order, queue it for later */
			cmnd->state = ISCSI_QUEUE_OTHER;
		}
		TRACE(TRACE_DEBUG, "queueing CmdSN %u, ExpCmdSN %u, ITT %u "
							"opcode 0x%02x, state %u, data_length %u\n",
							cmnd->cmd_sn, session->exp_cmd_sn,
							cmnd->init_task_tag, cmnd->opcode_byte, cmnd->state,
							cmnd->data_length);
		ack_sent_cmnds(conn, cmnd, pdu->exp_stat_sn, 1);
	}
	return 0;
}

/* computes the abort response when the referenced task is not found:
 * RFC 3720 section 10.6.1 Response
 * For the ABORT TASK function,
 * ...
 * "b) If the Referenced Task Tag does not identify an existing task,
 * but if the CmdSN indicated by the RefCmdSN field in the Task
 * Management function request is within the valid CmdSN window and less
 * than the CmdSN of the Task Management function request itself,
 * then targets must consider the CmdSN received and return the
 * 'Function complete' response.
 * "c) If the Referenced Task Tag does not identify an existing task
 * and if the CmdSN indicated by the RefCmdSN field in the Task
 * Management function request is outside the valid CmdSN window,
 * then targets must return the 'Task does not exist' response.
 */
static int __attribute__ ((no_instrument_function))
get_abort_response( struct iscsi_session *session,
					struct iscsi_cmnd *cmnd)
{
	int delta, retval = TASK_DOES_NOT_EXIST;

	if (down_interruptible(&session->cmnd_sem))
		goto out;

	delta = session->max_cmd_sn - cmnd->ref_cmd_sn;
	if (delta < 0) {
		/* ref_cmd_sn is greater than MaxCmdSN, so outside valid CmdSN window */
		goto out_unlock;
	}

	delta = cmnd->ref_cmd_sn - session->exp_cmd_sn;
	if (delta < 0) {
		/* ref_cmd_sn is less than ExpCmdSN, so outside valid CmdSN window */
		goto out_unlock;
	}

	delta = cmnd->cmd_sn - cmnd->ref_cmd_sn;
	if (delta <= 0) {
		/* ref_cmd_sn is NOT less than cmd_sn */
		goto out_unlock;
	}

	/* target must consider the CmdSN received, and return FUNCTION_COMPLETE */
	session->exp_cmd_sn++;
	retval = FUNCTION_COMPLETE;

out_unlock:
	up(&session->cmnd_sem);

out:
	return retval;
}

static void __attribute__ ((no_instrument_function))
do_task_mgt(struct iscsi_conn *conn,
			struct iscsi_cmnd *cmnd)
{
	struct iscsi_cmnd *ref_command;

	if (cmnd->ref_function == TMF_ABORT_TASK) {
		printk("ABORT TASK with ITT %u\n", cmnd->ref_task_tag);
		ref_command = search_tags(conn, cmnd->ref_task_tag, ALL_ONES, 1);
		if (ref_command == NULL) {
			/* can't find command with this ITT, check the RefCmdSN field */
			TRACE_ERROR("%s No command with ITT %u\n", current->comm,
						cmnd->ref_task_tag);
			cmnd->response = get_abort_response(conn->session, cmnd);
			cmnd->state = ISCSI_MGT_FN_DONE;
		} else {
			TRACE_ERROR("%s Aborting opcode 0x%02x, ITT %u, xfer_len %u, "
						"data_done %u, r2t_data %d, r2t_sn %u, state %u\n",
						current->comm,
						ref_command->opcode_byte,
						ref_command->init_task_tag,
						ref_command->data_length,
						ref_command->data_done,
						ref_command->r2t_data,
						ref_command->r2t_sn,
						ref_command->state);
			cmnd->message = rx_task_mgmt_fn(conn->dev->device,
									cmnd->ref_function, ref_command->cmnd);
			if (!cmnd->message) {
				cmnd->response = FUNCTION_REJECTED;
				cmnd->state = ISCSI_MGT_FN_DONE;
			}
		}
	} else {
		/* we don't deal with this task management function (yet) */
		cmnd->response = TASK_MANAGEMENT_FUNCTION_NOT_SUPPORTED;
		cmnd->state = ISCSI_MGT_FN_DONE;
	}
}

/*
 * executed only by the rx thread.
 * called when a Task Management Request has just been received.
 * INPUT: connection, session and the buffer containing the TM pdu
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
handle_task_mgt_command(struct iscsi_conn *conn,
						struct iscsi_session *session,
						__u8 *buffer)
{
	struct iscsi_init_task_mgt_command *pdu
						= (struct iscsi_init_task_mgt_command *)buffer;
	struct iscsi_cmnd *cmnd;
	int err = 0;
	__u32 save_trace;

	/* turn on (almost) all tracing while processing a TM command */
	TRACE_GET(save_trace);
	TRACE_SET(TRACE_ALL
				& ~(TRACE_NET | TRACE_BUF | TRACE_MY_MEMORY | TRACE_TIMERS));

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_task_mgt_command(pdu);

	pdu->init_task_tag = be32_to_cpu(pdu->init_task_tag);
	pdu->ref_task_tag = be32_to_cpu(pdu->ref_task_tag);
	pdu->cmd_sn = be32_to_cpu(pdu->cmd_sn);
	pdu->exp_stat_sn = be32_to_cpu(pdu->exp_stat_sn);
	pdu->ref_cmd_sn = be32_to_cpu(pdu->ref_cmd_sn);
	pdu->exp_data_sn = be32_to_cpu(pdu->exp_data_sn);

	TRACE_ERROR("%s Got TM Req, ITT %u, RTT %u, CmdSN %u, ref CmdSN %u, "
				"ExpCmdSN %u, ExpStatSN %u\n",
				current->comm, pdu->init_task_tag, pdu->ref_task_tag,
				pdu->cmd_sn, pdu->ref_cmd_sn, session->exp_cmd_sn,
				pdu->exp_stat_sn);

	if (!(pdu->function & F_BIT)) {
		TRACE_ERROR("%s task mgt request with F_BIT 0\n", current->comm);
	}

	pdu->function = pdu->function & 0x7f;

	/* RFC 3720 Section 10.5.4 Referenced Task Tag
	 * "The Initiator (sic -- should be Referenced) Task Tag of the task to be
	 * aborted for the ABORT TASK function or reassigned for the TASK REASSIGN
	 * function.  For all the other functions this field MUST be set to the
	 * reserved value 0xffffffff."
	 */
	if ((pdu->function != TMF_ABORT_TASK)
		&& (pdu->function != TMF_TASK_REASSIGN)
		&& (pdu->ref_task_tag != ALL_ONES)) {
		TRACE_ERROR("%s task mgt request with RTT "
					"%u, should be 0x%08x for function %d\n",
					current->comm, pdu->ref_task_tag, ALL_ONES, pdu->function);
	}

	if( (cmnd = get_new_cmnd()) == NULL) {
		err = -1;
		goto out;
	}

	cmnd->conn = conn;
	cmnd->session = session;
	cmnd->opcode_byte = pdu->opcode;
	cmnd->init_task_tag = pdu->init_task_tag;
	cmnd->ref_task_tag = pdu->ref_task_tag;
	cmnd->ref_cmd_sn = pdu->ref_cmd_sn;
	cmnd->ref_function = pdu->function;
	cmnd->cmd_sn = pdu->cmd_sn;
	cmnd->stat_sn = pdu->exp_stat_sn;
	cmnd->response = FUNCTION_COMPLETE;

	if (pdu->length > 0) {
		/* read length bytes of text data into a newly-allocated buffer */
		TRACE_ERROR("%s Data attached to task mgt request, CmdSN %u, ExpCmdSN "
					"%u, ITT %u, opcode 0x%02x\n", current->comm,
					cmnd->cmd_sn, session->exp_cmd_sn,
					cmnd->init_task_tag, cmnd->opcode_byte);
		err = read_single_data_seg(buffer, cmnd, pdu->length, &cmnd->ping_data);
		if (err <= 0) {
			my_kfree((void**)&cmnd, "cmnd");
			goto out;
		}
	}

	if (down_interruptible(&session->cmnd_sem)) {
		my_kfree((void *)&cmnd->ping_data, "data_buf");
		my_kfree((void**)&cmnd, "cmnd");
		err = -1;
		goto out;
	} else {
		err = check_cmd_sn(cmnd, pdu, session, 1);
		up(&session->cmnd_sem);
	}

	if (err < 0) {	
		/* out of range, silently ignore it */
		TRACE_ERROR("%s Dropping out of range task mgt request CmdSN %u, "
					"ExpCmdSN %u\n", current->comm,
					pdu->cmd_sn, session->exp_cmd_sn);
		ack_sent_cmnds(conn, cmnd, pdu->exp_stat_sn, 0);
		my_kfree((void *)&cmnd->ping_data, "data_buf");
		my_kfree((void**)&cmnd, "cmnd");
	} else {
		if (err > 0) {		
			/* within range but out of order, queue it for later */
			cmnd->state = ISCSI_QUEUE_OTHER;
		} else {
			/* is immediate command or expected non-immediate command */
			do_task_mgt(conn, cmnd);
		}

		ack_sent_cmnds(conn, cmnd, pdu->exp_stat_sn, 1);
	}
	err = 0;
out:
	/* restore tracing to what it was when we came in here */
	TRACE_SET(save_trace);

	return err;
}

/* called when out-of-order queued commands (except for SCSI Commands)
 * become in-order and have to be "delivered"
 */
static void __attribute__ ((no_instrument_function))
deliver_queue_other(struct iscsi_cmnd *cmnd, struct iscsi_session *session)
{
	__u8 opcode;

	opcode = cmnd->opcode_byte & ISCSI_OPCODE;
	if (opcode == ISCSI_INIT_NOP_OUT) {
		/* this was an out-of-order NopIn */
		if (cmnd->init_task_tag == ALL_ONES) {
			/* NopIn not used as a ping request, done with it now */
			TRACE(TRACE_DEBUG, "Freeing NopIn, ITT 0x%08x, CmdSN %u\n",
					cmnd->init_task_tag, cmnd->cmd_sn);
			/* NopOut pdu was not immediate, CmdSN was advanced */
			session->max_cmd_sn++;
			my_kfree((void *)&cmnd->ping_data, "data_buf");
			cmnd->state = ISCSI_DEQUEUE;
		} else {
			cmnd->state = ISCSI_PING;
			TRACE(TRACE_DEBUG, "ping back CmdSN %u, ExpCmdSN %u, ITT %u"
							   " opcode 0x%02x, state %u, data_length "
							   "%u\n",
								cmnd->cmd_sn, session->exp_cmd_sn,
								cmnd->init_task_tag, cmnd->opcode_byte,
								cmnd->state, cmnd->data_length);
		}
	} else if (opcode == ISCSI_INIT_TEXT_CMND) {
		/* this was an out-of-order Text Request */
		do_text_request(cmnd, cmnd->conn, session);
	} else if (opcode == ISCSI_INIT_LOGOUT_CMND) {
		/* this was an out-of-order Logout Request */
		cmnd->state = ISCSI_LOGOUT;
	} else if (opcode == ISCSI_INIT_TASK_MGMT_CMND) {
		/* this was an out-of-order Task Management Request */
		do_task_mgt(cmnd->conn, cmnd);
	} else {
		/* should never happen */
		TRACE_ERROR("%s Unexpected queue other ITT 0x%08x, CmdSN %u, "
					"opcode 0x%02x\n", current->comm,
					cmnd->init_task_tag, cmnd->cmd_sn, cmnd->opcode_byte);
	}
}

/*	cdeng: called by any rx_thread or rx_thread.
 *	Checks session's command queue to see if any out-of-order command is now
 *	in order.  If so, and it is a SCSI command,
 *	call rx_cmnd for it as the same as the handle_cmnd does
 */
static void
check_queued_cmnd(struct iscsi_session *session)
{
	struct iscsi_init_scsi_cmnd *pdu;
	struct iscsi_cmnd *temp, *prev = NULL;

	/* need to restart search loop whenever ExpCmdSN is incremented,
	 * which happens when we find and deliver a previously out-of-order
	 * command because it is now in order.
	 */
restart:
	/* lock the session-wide list of commands while searching it */
	if (down_interruptible(&session->cmnd_sem))
		goto out;

	for (temp = session->cmnd_list; temp != NULL; temp = temp->next) {
		if (temp->cmd_sn == session->exp_cmd_sn
			&& !(temp->opcode_byte & I_BIT)
			&& temp->state != ISCSI_DEQUEUE) {
			TRACE(TRACE_ISCSI, "CmdSN %u ITT %u now in order\n",
				  temp->cmd_sn, temp->init_task_tag);

			/* ok now to deliver the command to the target in proper order */
			if (temp->state == ISCSI_QUEUE_OTHER) {
				/* this is non-SCSI command, deliver it to us, not target */
				session->exp_cmd_sn++;
				up(&session->cmnd_sem);
				deliver_queue_other(temp, session);
				goto restart;				/* exp_cmd_sn has been bumped */
			} else if (!temp->cmd_sn_increment) {
				/* this was an out-of-order (non-immediate) SCSI Command that
				 * is now in-order, deliver it to target and mark it
				 * until delivery is confirmed */
				temp->cmd_sn_increment = 1;
				up(&session->cmnd_sem);
				TRACE(TRACE_ISCSI, "delivering CmdSN %u ITT %u to target\n",
					  temp->cmd_sn, temp->init_task_tag);
				pdu = (struct iscsi_init_scsi_cmnd *)temp->hdr;
				rx_cmnd(temp->conn->dev->device,
						session->oper_param->TargetName, pdu->lun, pdu->cdb,
						ISCSI_CDB_LEN, pdu->xfer_len, pdu->flags, &temp->cmnd);

				if (temp->cmnd)					/* rx_cmnd worked ok */
					goto restart;				/* exp_cmd_sn may be bumped */

				TRACE_ERROR("%s rx_cmnd returned NULL, ITT %u\n",
							current->comm, pdu->init_task_tag);

				if (down_interruptible(&session->cmnd_sem)) {
					goto out;
				}
				/* have to bump this on error */
				session->exp_cmd_sn += temp->cmd_sn_increment;
				temp->cmd_sn_increment = 0;
				goto unlock;
			}
		}
		prev = temp;
	}
unlock:
	up(&session->cmnd_sem);
out:
	return;
}

static void
free_data_list(struct iscsi_cmnd *cmnd)
{
	struct data_list *data;

	while ((data = cmnd->unsolicited_data_head)) {
		cmnd->unsolicited_data_head = data->next;
		my_kfree((void **)&data->buffer, "data_buf");
		my_kfree((void **)&data, "data_list");
	}
}

/* 
 * executed only by the rx thread.
 * cdeng: called by handle_cmnd and handle_data if the 
 * command is out of order.  Store the data into data_list in 
 * iSCSI layer rather than the st_list in SCSI layer.
 * Returns 0 if all ok, -1 on any error (no memory for buffers).
 */

static int
save_unsolicited_data(struct iscsi_cmnd *cmnd, __u32 offset,
					  struct generic_pdu *hdr)
{
	int err;
	struct data_list *data;
	__u32 total_length, length;

	TRACE(TRACE_ENTER_LEAVE, "Enter save_unsolicited_data\n");

	if ((total_length = hdr->length) == 0)
		return 0;

	TRACE(TRACE_ISCSI, "Save_unsolicited_data: offset %u, length %u\n", offset,
		  total_length);

	while ((length = total_length)) {
		if (length > MAX_KMALLOC_SIZE)
			length = MAX_KMALLOC_SIZE;

		/* receive length bytes of data */
		data = my_kmalloc(sizeof(struct data_list), "data_list");
		if (data == NULL)
			return -1;

		data->offset = offset;
		data->length = length;

		/* receive unsolicited data into a newly-allocated buffer */
		err = read_single_data_seg((__u8 *)hdr, cmnd, length, &data->buffer);
		if (err <= 0) {
			my_kfree((void **)&data, "data_list");
			return err;
		}

		/* add the new buffer to end of list of unsolicted data buffers */
		data->next = NULL;
		if (cmnd->unsolicited_data_head)
			cmnd->unsolicited_data_tail->next = data;
		else
			cmnd->unsolicited_data_head = data;
		cmnd->unsolicited_data_tail = data;

		offset += length;
		total_length -= length;
		}

	update_after_read(hdr, cmnd);

	if (atomic_read(&cmnd->conn->tx_sem.count) <= 0) {
		up(&cmnd->conn->tx_sem);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving save_unsolicited_data\n");
	return 0;
}

static inline __u32 __attribute__ ((no_instrument_function))
get_sglen(struct scatterlist *st_list)
{
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
		return sg_dma_len(st_list);
#else
		return st_list->length;
#endif
}

static inline __u8 * __attribute__ ((no_instrument_function))
get_sgbuf(struct scatterlist *st_list)
{
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
			return (__u8 *)sg_dma_address(st_list);
#else
			return st_list->address;
#endif
}

/* skip thru sg list to appropriate offset for data buffer
 * i is index to sg list item
 * offset is total number of bytes to skip
 * returns number of bytes remaining to be skipped in ith sg list item
 */
static __u32 __attribute__ ((no_instrument_function))
skip_thru_sg_list(struct scatterlist *st_list, __u32 *i, __u32 offset)
{
	__u32 sglen;

	*i = 0;							/* index to current sg list item */
	while (offset > 0) {
		sglen = get_sglen(st_list);
		if (offset < sglen) {
			/* copying starts at "offset" bytes in ith sg list item */
			break;
		}
		/* need to skip over this entire sg list item */
		offset -= sglen;
		st_list++;
		*i += 1;
	}
	return offset;
}

/* 
 * executed only by the tx thread when state is ISCSI_QUEUE_CMND_RDY.
 * copy the data in data_list in iSCSI layer to st_list in SCSI middle level.
 * if all the data has been read at this time,
 * then change state to ISCSI_DATA_IN and call scsi_rx_data()
 * else change state to ISCSI_BUFFER_RDY
 * this thread is called with session->cmnd_sem lock held by caller (tx thread)
 */
static int
send_unsolicited_data(struct iscsi_cmnd *cmd,
					  struct iscsi_conn *conn,
					  struct iscsi_session *session)
{
	struct scatterlist *st_list;
	struct data_list *data;
	int err;
	__u32 expected_data_offset, offset, length, sglen, i, n;
	__u8 *buffer, *sgbuf;

	if (cmd->unsolicited_data_present | cmd->immediate_data_present) {
		/* more unsolicited data has to be read, just stay in this state */
		return 0;
	}

	/* all unsolicited data has been read */
	TRACE(TRACE_ISCSI, "Send_unsolicited_data: cmd_sn %d\n", cmd->cmd_sn);

	if (!cmd->cmnd) {
		TRACE_ERROR("%s no cmnd found\n", current->comm);
		return -1;
	}

	st_list = (struct scatterlist *)cmd->cmnd->req->sr_buffer;

	/* copy each buffer in data list to appropriate place in sg list */
	expected_data_offset = offset = i = 0;
	data = cmd->unsolicited_data_head;
	while (data) {
		/* if items in data list are in correct order by offset,
		 * then "offset" and "i" are already positioned correctly in the sg list
		 */
		if (data->offset != expected_data_offset) {
			offset = skip_thru_sg_list(st_list, &i, data->offset);
		}

		/* copy all "length" bytes of data from data list "buffer" to sg list,
		 * starting at "offset" bytes in ith sg list item */
		buffer = data->buffer;
		length = data->length;
		expected_data_offset = data->offset + length;
		while (length > 0) {
			sglen = get_sglen(st_list + i);
			sgbuf = get_sgbuf(st_list + i);
			n = sglen - offset;		/* no of bytes left in sg list item's buf */
			if (n > length)
				n = length;			/* only this many data bytes left to copy */
			memcpy(sgbuf + offset, buffer, n);
			length -= n;
			buffer += n;
			offset += n;
			if (offset >= sglen) {
				/* copied all bytes in this sg list item's buf, move to next */
				offset = 0;
				i++;
			}
		}

		/* "i" and "offset" positioned to fill next sg list slot in order */
		cmd->scatter_list_count = i;
		cmd->scatter_list_offset = offset;
		data = data->next;
	}

	if (cmd->data_done >= cmd->data_length) {
		/* already read all the data, which was all unsolicited */
		cmd->state = ISCSI_DATA_IN;
		if ((err = scsi_rx_data(cmd->cmnd)) < 0) {
			TRACE_ERROR("%s scsi_rx_data returned an error\n", current->comm);
		}
	} else {
		/* more data left to read, which is all solicited */
		cmd->state = ISCSI_BUFFER_RDY;
		err = 0;
	}

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		up(&conn->tx_sem);
	}

	return err;
}

/*
 * executed only by the rx thread.
 * called only by handle_cmnd()
 * when called, err is +1 if cmnd is in range but out of order
 *				err is -1 if cmnd is out of range and should be ignored
 * returns	0 if command is queued ok or freed because it was out of range
 *			-1 if there was an error (no memory to save immediate data)
 */
static int
out_of_order_cmnd(struct iscsi_conn *conn,
				  struct iscsi_session *session,
				  __u8 *buffer,
				  struct iscsi_cmnd *cmnd, int err)
{
	struct iscsi_init_scsi_cmnd *hdr = (struct iscsi_init_scsi_cmnd *) buffer;
	int retval;

	/* store "length" bytes of immediate data attached to this command */
	if ((retval = save_unsolicited_data(cmnd, 0, (struct generic_pdu *)hdr))) {
		my_kfree((void**)&cmnd, "cmnd");
		goto out;
	}

	memcpy(cmnd->hdr, buffer, ISCSI_HDR_LEN);
	cmnd->state = ISCSI_QUEUE_CMND;

	/* last param is 0 (don't queue) if cmnd is out of range (err == -1) */
	ack_sent_cmnds(conn, cmnd, hdr->exp_stat_sn, err+1);

	if (err < 0) {	
		/* out of range, silently ignore it */
		TRACE_ERROR("%s ignoring out of range CmdSN %u, ExpCmdSN %u, ITT %u, "
					"opcode 0x%02x\n", current->comm,
					cmnd->cmd_sn, session->exp_cmd_sn,
					cmnd->init_task_tag, cmnd->opcode_byte);
		/* command was not put into queue */
		free_data_list(cmnd);
		my_kfree((void**)&cmnd, "cmnd");
	} else {		
		/* within range but out of order, we queued it for later (above) */
		TRACE(TRACE_ISCSI, "out of order CmdSN %u bigger than ExpCmdSN %u\n",
			  cmnd->cmd_sn, cmnd->session->exp_cmd_sn);
	}
out:
	return retval;
}

/*
 * executed only by the rx thread.
 * handle_cmnd: this function is responsible for processing a SCSI
 * command received from an initiator, making the mid-level aware of the
 * command and then queueing up the command in the command queue for the
 * session
 * INPUT: connection and the buffer containing the just-received PDU header
 * OUTPUT: >= 0 if everything is okay, < 0 if there is trouble
 */
static int
handle_cmnd(struct iscsi_conn *conn,
			struct iscsi_session *session,
			__u8 *buffer)
{
	struct iscsi_init_scsi_cmnd *pdu = (struct iscsi_init_scsi_cmnd *) buffer;
	struct iscsi_cmnd *cmnd;
	int err;

	TRACE(TRACE_ENTER_LEAVE, "Enter handle_cmnd\n");

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_scsi_cmnd(pdu);

	pdu->length = be32_to_cpu(pdu->length);
	pdu->init_task_tag = be32_to_cpu(pdu->init_task_tag);
	pdu->xfer_len = be32_to_cpu(pdu->xfer_len);
	pdu->cmd_sn = be32_to_cpu(pdu->cmd_sn);
	pdu->exp_stat_sn = be32_to_cpu(pdu->exp_stat_sn);

	/* check data length, immediate data should not bigger than 
	 * MaxRecvPDULength or FirstBurstLength
	 */
	if (pdu->length > conn->max_recv_length
		|| pdu->length > session->oper_param->FirstBurstLength) {
		printk("max_recv_length: %d\n", conn->max_recv_length);
		printk("FirstBurstLength: %d\n", session->oper_param->FirstBurstLength);
		TRACE_WARNING("%s DSL %u too big\n", current->comm, pdu->length);
	}

	if( (cmnd = get_new_cmnd()) == NULL) {
		err = -1;
		goto out;
	}

	cmnd->conn = conn;
	cmnd->session = session;
	cmnd->opcode_byte = pdu->opcode;
	cmnd->init_task_tag = pdu->init_task_tag;
	cmnd->data_length = pdu->xfer_len;
	cmnd->lun = pdu->lun;
	cmnd->cdb[0] = pdu->cdb[0];
	cmnd->cdb[4] = pdu->cdb[4];
	cmnd->cmd_sn = pdu->cmd_sn;
	cmnd->first_burst_len = pdu->length;
	cmnd->stat_sn = pdu->exp_stat_sn;

	if (pdu->length)
		/* there is data attached to this PDU, it must be immediate data */
		cmnd->immediate_data_present = 1;

	if (!(pdu->flags & F_BIT)) {
		/* F bit not set, so unsolicited DataIn PDUs must follow this PDU */
		cmnd->unsolicited_data_present = 1;
		/* we solicit all data that might follow a full first burst
		 * NOTE: this subtraction can produce a negative value!
		 */
		cmnd->r2t_data_total = cmnd->data_length
								- session->oper_param->FirstBurstLength;
	} else {
		/* no unsolicited DataIn PDUs, we solicit all data except immediate */
		cmnd->r2t_data_total = cmnd->data_length - pdu->length;
	}

	/* Retransmit Flag for Error Recovery - SAI */

	/* error recovery ver ref18_04 - SAI */

	init_MUTEX_LOCKED(&cmnd->unsolicited_data_sem);

	/* check cmd_sn for command ordering */
	if (down_interruptible(&session->cmnd_sem)) {
		my_kfree((void**)&cmnd, "cmnd");
		err = -1;
		goto out;
	}

	/* Last parameter is 0 because we cannot bump the exp_cmd_sn for the session
	 * until this command has been delivered to the mid-level in proper order.
	 * (bumping exp_cmd_sn might let other commands jump ahead of this)
	 * Also causes command list to become blocked if this command is in order.
	 */
	err = check_cmd_sn(cmnd, pdu, session, 0);

	/* generate the next TTT, in case we need it */
	cmnd->target_xfer_tag = generate_next_ttt(session);
	up(&session->cmnd_sem);

	if (err) {
		/* command received is not in expected CmdSN order */
		err = out_of_order_cmnd(conn, session, buffer, cmnd, err);
		/* cmnd has now been queued or freed, as appropriate */
		goto out;
	}

	/* else command is immediate or (is in-order and cmnd_list is blocked) */

	TRACE(TRACE_DEBUG, "unsolicited_data_present %d, err %d, flags "
		  "0x%02x\n", cmnd->unsolicited_data_present, err, pdu->flags);

	ack_sent_cmnds(conn, cmnd, pdu->exp_stat_sn, 1);

	/* pass this command on to the midlevel so it can set up the buffers */
	rx_cmnd(conn->dev->device, session->oper_param->TargetName,
			pdu->lun, pdu->cdb, ISCSI_CDB_LEN, pdu->xfer_len,
			pdu->flags, &cmnd->cmnd);

	if (!(cmnd->cmnd)) {
		TRACE_ERROR("%s rx_cmnd returned NULL, ITT %u\n", current->comm,
					cmnd->init_task_tag);
		err = -1;
		if (!down_interruptible(&session->cmnd_sem)) {
			/* bump exp_cmd_sn as needed */
			session->exp_cmd_sn += cmnd->cmd_sn_increment;
			cmnd->cmd_sn_increment = 0;
			up(&session->cmnd_sem);
		}
		goto out2;
	}

	if (cmnd->immediate_data_present) {
		/* this WRITE pdu has immediate data attached,
		 * wait for midlevel to set up buffers so we can
		 * read it directly into midlevel buffers
		 */
		TRACE(TRACE_SEM, "Blocked on unsolicited_data_sem\n");
		if (down_interruptible(&cmnd->unsolicited_data_sem)) {
			err = -1;
			goto out2;
		}
		TRACE(TRACE_SEM, "Unblocked on unsolicited_data_sem\n");

		/* state MUST now be ISCSI_BUFFER_RDY !*/
		if (cmnd->state != ISCSI_BUFFER_RDY) {
			TRACE_ERROR("%s got cmnd->state %u, expected %u\n",
						current->comm, cmnd->state, ISCSI_BUFFER_RDY);
		}

		/* now read immediate data directly into midlevel buffers */
		err = read_list_data_seg((struct generic_pdu *) pdu, cmnd,
				(struct scatterlist *) cmnd->cmnd->req->sr_buffer, 0);
		if (err <= 0) {
			if (err) {
				TRACE_ERROR("%s read_list_data_seg returned error %d\n",
							current->comm, err);
			} /* else recovered error */
			goto out2;
		}

		if (!session->oper_param->DataPDUInOrder) {
			/* immediate data should always begin from offset 0 */
			cmnd->seq_range_list.offset = 0;
			cmnd->pdu_range_list.offset = 0;
			cmnd->pdu_range_list.limit = pdu->length;
			merge_offset_length(&cmnd->pdu_range_list, 0, pdu->length);
		}

		if (!session->oper_param->DataSequenceInOrder
			&& !cmnd->unsolicited_data_present
			&& cmnd->data_done < cmnd->data_length) {
			merge_offset_length(&cmnd->seq_range_list, 0, pdu->length);
		}

		if (cmnd->data_done >= cmnd->data_length) {
			/* all the data was sent as immediate, there will be no
			 * unsolicited or solicited data to follow, therefore no need
			 * to generate any R2Ts.
			 * the WRITE is finished
			 */
			if (!session->oper_param->DataPDUInOrder) {
				check_range_list_complete(&cmnd->pdu_range_list);
				free_range_list(&cmnd->pdu_range_list);
			}

			TRACE(TRACE_DEBUG, "%d received for cmnd %p\n", pdu->length, cmnd);

			if (down_interruptible(&session->cmnd_sem)) {
				err = -1;
			} else {
				cmnd->state = ISCSI_DATA_IN;
				err = scsi_rx_data(cmnd->cmnd);
				up(&session->cmnd_sem);
			}

			if (err < 0) {
				TRACE_ERROR("%s scsi_rx_data returned an error\n",
							current->comm);
				goto out2;
			}
			if (atomic_read(&cmnd->conn->tx_sem.count) <= 0) {
				up(&cmnd->conn->tx_sem);
			}
		} else if (!(cmnd->unsolicited_data_present)
					&& (pdu->length < cmnd->data_length)) {
			/* is this next wakeup ever really needed? */
			if (atomic_read(&cmnd->conn->tx_sem.count) <= 0) {
				up(&cmnd->conn->tx_sem);
			}
		}
	}

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave handle_cmnd, err = %d\n", err);
	return err;

out2:
	/* here for errors that are detected after cmnd is added to queue */
	cmnd->state = ISCSI_DEQUEUE;
	goto out;
}

/* until the last data_out received, it's ready for recovery r2t */
void
check_r2t_done(struct iscsi_cmnd *cmd, struct iscsi_init_scsi_data_out *hdr)
{
	struct iscsi_cookie *next;

	/* in case we lost the last data_out of unsolicited data */
	if (hdr->offset > cmd->session->oper_param->FirstBurstLength)
		cmd->unsolicited_data_present = 0;

	if (cmd->unsolicited_data_present) {
		if (hdr->flags & F_BIT) {
			cmd->data_sn = 0;
			cmd->unsolicited_data_present = 0;
		}
	} else {
		if (hdr->flags & F_BIT) {
			cmd->next_burst_len = 0;
			cmd->data_sn = 0;
			/* unhook data_q now */
			while (cmd->first_data_q) {
				next = cmd->first_data_q->next;
				my_kfree((void**)&cmd->first_data_q, "first_data_q");
				cmd->first_data_q = next;
			}
			cmd->last_data_q = NULL;
			if (!cmd->recovery_r2t)
				cmd->outstanding_r2t--;
			if (atomic_read(&cmd->conn->tx_sem.count) <= 0) {
				up(&cmd->conn->tx_sem);
			}
		}
	}
}

static void __attribute__ ((no_instrument_function))
merge_out_of_order( struct iscsi_init_scsi_data_out *hdr,
					struct iscsi_cmnd *cmd)
{
	__u32 limit;

	if (cmd->pdu_range_list.offset > hdr->offset)
		cmd->pdu_range_list.offset = hdr->offset;
	limit = cmd->pdu_range_list.offset + hdr->length;
	if (cmd->pdu_range_list.limit < limit)
		cmd->pdu_range_list.limit = limit;
	merge_offset_length(&cmd->pdu_range_list, hdr->offset, hdr->length);
}

/*
 * executed only by the rx thread.
 * handle_data: This function is responsible for receiving DataOut pdus that
 * correspond to a WRITE command. It receives the data in the buffers that
 * are allocated by the STML. It then has to change the state of the
 * command and inform the STML about the data received
 * INPUT: connection and buffer containing the ISCSI_DATA (write) PDU
 * OUTPUT: 0, if everything is okay, < 0 otherwise
 */
static int
handle_data(struct iscsi_conn *conn,
			struct iscsi_session *session,
			__u8 *buffer)
{
	struct iscsi_init_scsi_data_out *hdr =
		(struct iscsi_init_scsi_data_out *) buffer;
	struct iscsi_cmnd *cmd;
	struct targ_error_rec err_rec;
	struct scatterlist *st_list;
	int offset, err = 0, giveback = 0;
	int i;
	__u32 data_sn;

	TRACE(TRACE_ENTER_LEAVE, "Entered handle_data\n");

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_scsi_data_out(hdr);

	hdr->length = be32_to_cpu(hdr->length);
	hdr->init_task_tag = be32_to_cpu(hdr->init_task_tag);
	hdr->target_xfer_tag = be32_to_cpu(hdr->target_xfer_tag);
	hdr->exp_stat_sn = be32_to_cpu(hdr->exp_stat_sn);
	hdr->offset = be32_to_cpu(hdr->offset);
	data_sn = be32_to_cpu(hdr->data_sn);

	/* check data length, should not be bigger than MaxRecvDataSegmentLength */
	if (hdr->length > conn->max_recv_length) {
		TRACE_WARNING("%s DataOut ITT %u, DataSN %u, DSL %u exceeds "
					  "MaxRecvDataSegmentLength %u\n", current->comm,
					  hdr->init_task_tag, data_sn, hdr->length,
					  conn->max_recv_length);
		/* now ignore this, since we can deal with it */
	}

	/* search command queue to match command */
	cmd = search_tags(conn, hdr->init_task_tag, hdr->target_xfer_tag, 0);

	if (!cmd) {
		TRACE_ERROR("%s DataOut ITT %u, DataSN %u, TTT %u, No matching "
					"command\n", current->comm, hdr->init_task_tag, data_sn,
					hdr->target_xfer_tag);
		TRACE(TRACE_ISCSI, "Probably SCSI cmnd PDU lost - drop it\n");
		targ_drop_pdu_data(conn, hdr->length);
		goto end_handle_data;
	}

	/* make sure we use the status information, but don't queue this cmd */
	ack_sent_cmnds(conn, cmd, hdr->exp_stat_sn, 0);

	/* update last time this command got some data back from initiator */
	cmd->timestamp = jiffies;

	if (session->oper_param->DataPDUInOrder
		&& session->oper_param->DataSequenceInOrder) {
		/* error recovery ver ref18_04 - SAI */
		if (cmd->data_done > hdr->offset) {
			TRACE_WARNING("%s Dropping duplicate DataOut ITT %u, DataSN %u, "
						  "Offset %u\n", current->comm,
						  hdr->init_task_tag, data_sn, hdr->offset);
			targ_drop_pdu_data(conn, hdr->length);
			goto update_handle_data;
		} else if (cmd->data_done < hdr->offset) {
			/* offset error - expected packets might have been dropped */
			TRACE_ERROR("%s DataOut ITT %u, DataSN %u, Offset %u bigger than "
						"expected %u\n", current->comm, hdr->init_task_tag,
						data_sn, hdr->offset, cmd->data_done);

			TRACE(TRACE_ERROR_RECOVERY, "Start sequence error recovery\n");
			err_rec.curr_conn = conn;
			err_rec.pdu_hdr = (struct generic_pdu *) hdr;
			err_rec.cmd = cmd;
			err_rec.err_type = SEQUENCE_ERR;
			if ((err = targ_do_error_recovery(&err_rec)) < 0)
				goto end_handle_data;

			if (session->oper_param->ErrorRecoveryLevel != SESSION_RECOVERY) {
				check_r2t_done(cmd, hdr);
			}
			goto end_handle_data;
		} else if (data_sn != cmd->data_sn) {
			TRACE_WARNING("%s DataOut ITT %u, Got DataSN %u, expected %u\n",
						  current->comm, hdr->init_task_tag, data_sn,
						  cmd->data_sn);
			/* now ignore this, since we can deal with it */
		}
	}

	cmd->recovery_r2t = 0;
	cmd->data_sn = data_sn + 1;

	/* check FirstBurstLength and MaxBurstLength */
	if (cmd->unsolicited_data_present) {
		/* this DataOut PDU is part of the unsolicited first burst */
		cmd->seq_range_list.offset = 0;

		if (cmd->first_burst_len == 0
			&& !session->oper_param->DataPDUInOrder) {
			cmd->pdu_range_list.offset = ALL_ONES;
			cmd->pdu_range_list.limit = 0;
		}

		if (!session->oper_param->DataPDUInOrder) {
			/* DataPDUInOrder==No, so PDU can be out-of-order */
			merge_out_of_order(hdr, cmd);
		}

		/* number of bytes read so far in this unsolicited first burst */
		cmd->first_burst_len += hdr->length;

		if (cmd->first_burst_len > session->oper_param->FirstBurstLength) {
			TRACE_WARNING("%s DataOut ITT %u, DataSN %u, data length %u "
						  "exceeds FirstBurstLength %u\n", current->comm,
						  hdr->init_task_tag, data_sn, cmd->first_burst_len,
						  session->oper_param->FirstBurstLength);
			/* now ignore this, since we can deal with it */
		}

		if (hdr->flags & F_BIT) {
			/* end of the unsolicited first burst */
			if (!session->oper_param->DataPDUInOrder) {	
				/* DataPDUInOrder==No, PDUs can be out-of-order */
				check_range_list_complete(&cmd->pdu_range_list);
				free_range_list(&cmd->pdu_range_list);
			}

			/* unsolicited data should always be the first sequence */
			if (!session->oper_param->DataSequenceInOrder) {	
				/* DataSequenceInOrder==No, sequence can be out-of-order */
				merge_offset_length(&cmd->seq_range_list, 0,
									cmd->first_burst_len);
			}
		}
	} else {
		/* this DataOut PDU is part of a solicited burst */
		if (cmd->next_burst_len == 0) {
			/* this is the first PDU in a solicited burst */
			cmd->seq_range_list.offset = hdr->offset;
			if (!session->oper_param->DataPDUInOrder) {
				cmd->pdu_range_list.offset = ALL_ONES;
				cmd->pdu_range_list.limit = 0;
			}
		}

		if (!session->oper_param->DataPDUInOrder) {
			/* DataPDUInOrder==No, so PDU can be out-of-order */
			merge_out_of_order(hdr, cmd);
		}

		/* number of bytes read so far in this solicited burst */
		cmd->next_burst_len += hdr->length;

		if (cmd->next_burst_len > session->oper_param->MaxBurstLength) {
			TRACE_WARNING("%s DataOut ITT %u, DataSN %u, data length %u "
						  "exceeds MaxBurstLength %u\n", current->comm,
						  hdr->init_task_tag, data_sn, cmd->next_burst_len,
						  session->oper_param->MaxBurstLength);
			/* now ignore this, since we can deal with it */
		}

		if (hdr->flags & F_BIT) {
			/* this is the end of the current solicited burst */
			if (!session->oper_param->DataPDUInOrder) {
				/* DataPDUInOrder==No, PDUs can be out-of-order */
				check_range_list_complete(&cmd->pdu_range_list);
				cmd->seq_range_list.offset = cmd->pdu_range_list.offset;
				free_range_list(&cmd->pdu_range_list);
			}

			if (!session->oper_param->DataSequenceInOrder) {
				/* DataSequenceInOrder==No, sequence can be out-of-order */
				merge_offset_length(&cmd->seq_range_list,
									cmd->seq_range_list.offset,
									cmd->next_burst_len);
			}

			/* reset counters for start of next solicited burst (if any) */
			cmd->next_burst_len = 0;
			cmd->data_sn = 0;

			/* this is the end of a solicited burst,
			 * so current R2T is no longer outstanding
			 * we can now generate another R2T
			 * to ask for remaining solicited data.
			 * (but not until after this data has actually been read)
			 */
			giveback = 1;
		}
	}

	if ((cmd->state == ISCSI_QUEUE_CMND)
		|| (cmd->state == ISCSI_QUEUE_CMND_RDY)) {
		/* cdeng, store unsolicited data for queued command */
		err = save_unsolicited_data(cmd,hdr->offset,(struct generic_pdu *)hdr);
		goto end_handle_data;
	} else if ((cmd->state != ISCSI_BUFFER_RDY)
			   && (cmd->state != ISCSI_ALL_R2TS_SENT)) {
		TRACE(TRACE_SEM, "handle_data: Blocked on unsolicited_data_sem\n");
		if (down_interruptible(&cmd->unsolicited_data_sem)) {
			err = -1;
			goto end_handle_data;
		}
		TRACE(TRACE_SEM, "handle_data: Unblocked on unsolicited_data_sem\n");
	}

	/* receive this data */
	st_list = (struct scatterlist *) cmd->cmnd->req->sr_buffer;

	if (session->oper_param->DataPDUInOrder
		&& session->oper_param->DataSequenceInOrder) {
		/* everything is in order, pick up where we left off after reading data
		 * from DataOut pdu (or from data segment of a WRITE containing
		 * immediate data)
		 */
		offset = cmd->scatter_list_offset;
		st_list += cmd->scatter_list_count;
		TRACE(TRACE_DEBUG, "scatter list offset %d, count %d\n",
			  cmd->scatter_list_offset, cmd->scatter_list_count);
	} else {
		offset = skip_thru_sg_list(st_list, &i, hdr->offset);
		st_list += i;
		TRACE(TRACE_DEBUG, "scatter list offset %d, index %d\n", offset, i);
	}

	TRACE(TRACE_DEBUG, "handle_data: receiving data for cmd_sn %.8x "
		  "init_task_tag %.8x target_xfer_tag %.8x\n",
		  cmd->cmd_sn, cmd->init_task_tag, cmd->target_xfer_tag);

	/* read data directly into midlevel buffers */
	if ((err = read_list_data_seg((struct generic_pdu *) hdr, cmd, st_list,
																offset)) <= 0)
		goto end_handle_data;

	/* error recovery ver ref18_04 - SAI */
	if (cmd->first_data_q)
		search_data_q(cmd);

	if (cmd->data_done >= cmd->data_length) {
		/* we have just finished reading everything the command expected */
		if (!(hdr->flags & F_BIT)) {
			TRACE_ERROR("%s DataOut ITT %u, DataSN %u, F bit = 0 but data done "
						"%u >= data length %u\n", current->comm,
						hdr->init_task_tag, data_sn, cmd->data_done,
						cmd->data_length);
			hdr->flags |= F_BIT;	/* so we wakeup tx thread below */
		}
		if (!session->oper_param->DataSequenceInOrder) {
			cmd->seq_range_list.offset = 0;
			cmd->seq_range_list.limit = cmd->data_done;
			check_range_list_complete(&cmd->seq_range_list);
			free_range_list(&cmd->seq_range_list);
		}

		/* tell the scsi midlevel that all the expected data has arrived */
		if (down_interruptible(&session->cmnd_sem)) {
			err = -1;
		} else {
			cmd->state = ISCSI_DATA_IN;
			err = scsi_rx_data(cmd->cmnd);
			up(&session->cmnd_sem);
		}
	}

update_handle_data:
	if (hdr->flags & F_BIT) {
		/* end of DataOut pdu sequence, tx thread may need to do something */
		cmd->outstanding_r2t -= giveback;
		if (atomic_read(&conn->tx_sem.count) <= 0) {
			up(&conn->tx_sem);
		}
	}

/* Added for Error Recovery - SAI */
end_handle_data:

	/* update last time this command got some data back from initiator */
	if (cmd)
		cmd->timestamp = jiffies;

	TRACE(TRACE_ENTER_LEAVE, "Leave handle_data, err = %d\n", err);
	return err;
}

/*
 * executed only by the rx thread.
 * Added by SAI for Handling SNACK requests from Initiator as 
 * part of Error Recovery 
 * For details, see Draft 20, Section 10.16 SNACK Request
 */

static int
handle_snack(struct iscsi_conn *conn,
			 struct iscsi_session *session,
			 __u8 *buffer)
{
	__u32 runlen, begrun, reason = REASON_DATA_SNACK;
	int delta1, delta2;
	struct iscsi_init_snack *pdu = (struct iscsi_init_snack *) buffer;
	struct iscsi_cmnd *cmd = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter handle_snack\n");

	if (TRACE_TEST(TRACE_ERROR_RECOVERY))
		print_init_snack(pdu);

	pdu->length = be32_to_cpu(pdu->length);
	pdu->init_task_tag = be32_to_cpu(pdu->init_task_tag);
	pdu->target_xfer_tag = be32_to_cpu(pdu->target_xfer_tag);
	pdu->exp_stat_sn = be32_to_cpu(pdu->exp_stat_sn);
	pdu->begrun = be32_to_cpu(pdu->begrun);
	pdu->runlen = be32_to_cpu(pdu->runlen);

	runlen = pdu->runlen;
	begrun = pdu->begrun;

	/* SNACK has no CmdSN but does have an ExpStatSN so use that now */
	ack_sent_cmnds(conn, NULL, pdu->exp_stat_sn, 0);

	if (!session->oper_param->ErrorRecoveryLevel) {
		TRACE_ERROR("%s Got SNACK type %u when ErrorRecoveryLevel=0, reject "
					"SNACK\n", current->comm, pdu->flags & SNACK_TYPE);
		goto out_reject;
	}

	switch (pdu->flags & SNACK_TYPE) {

	case DATA_R2T_SNACK:
		{
			if (pdu->init_task_tag == ALL_ONES) {
				TRACE_ERROR("%s Data/R2T SNACK with ITT = 0x%08x, reject "
							"SNACK\n", current->comm, pdu->init_task_tag);
				goto out_reject;
			} else if (pdu->target_xfer_tag != ALL_ONES) {
				TRACE_ERROR("%s Data/R2T SNACK with TTT = %u, reject SNACK\n",
							current->comm, pdu->target_xfer_tag);
				goto out_reject;
			} else if (!(session->targ_snack_flg & DATA_SNACK_ENABLE)) {
				TRACE_ERROR("%s Data/R2T SNACK not enabled, reject SNACK\n",
							current->comm);
				goto out_reject;
			} else if (session->targ_snack_flg & DATA_SNACK_REJECT) {
				TRACE(TRACE_ERROR_RECOVERY,
					  "handle_snack: Send Data-SNACK Reject for SNACK\n");
				goto out_reject;
			} else {
				TRACE_WARNING("%s Got Data/R2T-SNACK for ITT %u\n",
							  current->comm, pdu->init_task_tag);
			}
		}
		break;

	case STATUS_SNACK:
		{
			if (pdu->init_task_tag != ALL_ONES) {
				TRACE_ERROR("%s Status SNACK with ITT = %u, reject SNACK\n",
							current->comm, pdu->init_task_tag);
				goto out_reject;
			} else if (pdu->target_xfer_tag != ALL_ONES) {
				TRACE_ERROR("%s Status SNACK with TTT = %u, reject SNACK\n",
							current->comm, pdu->target_xfer_tag);
				goto out_reject;
			} else if (!(session->targ_snack_flg & STATUS_SNACK_ENABLE)) {
				TRACE_ERROR("%s Status SNACK not enabled, reject SNACK\n",
							current->comm);
				goto out_reject;
			} else {
				TRACE_WARNING("%s Got Status-SNACK\n", current->comm);
			}
		}
		break;

	case DATACK_SNACK:
		{
			if (pdu->init_task_tag != ALL_ONES) {
				TRACE_ERROR("%s Status SNACK with ITT = %u, reject SNACK\n",
							current->comm, pdu->init_task_tag);
				goto out_reject;
			} else if (pdu->target_xfer_tag == ALL_ONES) {
				TRACE_ERROR("%s DataACK SNACK with TTT = 0x%08x, reject "
							"SNACK\n", current->comm, pdu->target_xfer_tag);
				goto out_reject;
			} else if (runlen != 0) {
				TRACE_ERROR("%s DataACK SNACK with RunLength=%u, reject "
							"SNACK\n", current->comm, runlen);
				goto out_reject;
			} else if (!(session->targ_snack_flg & DATACK_SNACK_ENABLE)) {
				TRACE_ERROR("%s DataACK SNACK not enabled, reject SNACK\n",
							current->comm);
				goto out_reject;
			} else {
				TRACE(TRACE_ERROR_RECOVERY, "Got DataACK SNACK, TTT = %u\n",
						pdu->target_xfer_tag);
			}
			/* we have nothing more to do with this type of snack */
			goto out;
		}

	case R_DATA_SNACK:
		{
			if (pdu->init_task_tag == ALL_ONES) {
				TRACE_ERROR(" %s R-Data SNACK with ITT = 0x%08x, reject "
							"SNACK\n", current->comm, pdu->init_task_tag);
				goto out_reject;
			} else if (pdu->target_xfer_tag == 0
						|| pdu->target_xfer_tag == ALL_ONES) {
				TRACE_ERROR("%s R-Data SNACK with TTT = 0x%08x, reject "
							"SNACK\n", current->comm, pdu->target_xfer_tag);
				goto out_reject;
			} else if (begrun != 0) {
				TRACE_ERROR("%s R-Data SNACK with BegRun=%u, reject SNACK\n",
							current->comm, begrun);
				goto out_reject;
			} else if (runlen != 0) {
				TRACE_ERROR("%s R-Data SNACK with RunLength=%u, reject "
							"SNACK\n", current->comm, runlen);
				goto out_reject;
			}
			TRACE_ERROR("%s R-Data SNACK, type %u, ITT = %u, not implemented, "
						"reject SNACK\n", current->comm,
						pdu->flags & SNACK_TYPE, pdu->init_task_tag);
			goto out_reject;
		}

	default:
		{
			TRACE_ERROR("%s Invalid SNACK type %u, ITT = %u, reject SNACK\n",
						current->comm, pdu->flags & SNACK_TYPE,
						pdu->init_task_tag);
			goto out_reject;
		}
	}	/* switch */


	/* here only if SNACK type is DATA_R2T_SNACK or STATUS_SNACK */
	if (down_interruptible(&session->cmnd_sem)) {
			goto out;
		}
	for (cmd = session->cmnd_list; cmd != NULL; cmd = cmd->next) {
		if (cmd->init_task_tag == pdu->init_task_tag
			&& (pdu->flags & SNACK_TYPE) == DATA_R2T_SNACK
			&& !cmd->retransmit_flg) {
			if (cmd->state == ISCSI_SENT) {
				/* Command was a READ, SNACK must be to retransmit DataIn pdus*/
				if (cmd->data_sn >= pdu->begrun) {
					/* have at least 1 DataIn pdu to retransmit for this READ */
					TRACE(TRACE_ERROR_RECOVERY,
						  "handle_snack: Re-Transmit Data on SNACK Request\n");
					cmd->retransmit_flg = 1;
					cmd->scatter_list_count = 0;
					cmd->scatter_list_offset = 0;
					/* set limits for which DataIn pdus should be resent */
					cmd->startsn = pdu->begrun;
					if (!pdu->runlen)
						cmd->endsn = cmd->data_sn;
					else
						cmd->endsn = pdu->begrun + pdu->runlen - 1;
					/* signal the tx thread to do the retransmission */
					cmd->state = ISCSI_DONE;
					if (atomic_read(&cmd->conn->tx_sem.count) <= 0) {
						up(&cmd->conn->tx_sem);
					}
				}
			} else if (cmd->state == ISCSI_BUFFER_RDY) {
				/* Command was a WRITE, SNACK must be to retransmit R2T pdus */
				if (cmd->r2t_sn >= pdu->begrun) {
					/* Found the corresponding R2T PDU - re-transmit it */
					TRACE(TRACE_ERROR_RECOVERY,
						  "handle_snack: Re-Transmit R2T on SNACK Request\n");

					cmd->retransmit_flg = 1;
					cmd->startsn = begrun;
					if (!pdu->runlen) {
						cmd->endsn = cmd->r2t_sn;
					} else
						cmd->endsn = pdu->begrun + pdu->runlen - 1;
					/* signal the tx thread to do the retransmission */
					if (atomic_read(&cmd->conn->tx_sem.count) <= 0) {
						up(&cmd->conn->tx_sem);
					}
				}
			} else {
				/* RFC 3720 Section 10.16. SNACK Request
				 * "Any SNACK that requests a numbered-resonse, Data, or R2T
				 * that was not sent by the target or was already acknowledged
				 * by the initiator, MUST be rejected with a reason code of
				 * 'Protocol error'."
				 */
				reason = REASON_PROTOCOL_ERR;
				up(&session->cmnd_sem);
				goto out_reject;
			}
		} else if ((pdu->flags & SNACK_TYPE) == STATUS_SNACK) {
			if (cmd->state == ISCSI_SENT) {
				delta1 = cmd->stat_sn - begrun;
				delta2 = delta1 - runlen + 1;
				if (delta1 >= 0 && (runlen == 0 || delta2 <= 0)) {
					/* Found the corresponding Status PDU - re-transmit it*/
					TRACE(TRACE_ERROR_RECOVERY,
						  "handle_snack: Re-Transmit iscsi Response "
						  "on SNACK Request\n");

					cmd->retransmit_flg = 1;
					cmd->state = ISCSI_RESEND_STATUS;
					if (atomic_read(&cmd->conn->tx_sem.count) <= 0) {
						up(&cmd->conn->tx_sem);
					}
				}
			}
		} else {
				/* RFC 3720 Section 10.16. SNACK Request
				 * "Any SNACK that requests a numbered-resonse, Data, or R2T
				 * that was not sent by the target or was already acknowledged
				 * by the initiator, MUST be rejected with a reason code of
				 * 'Protocol error'."
				 */
				reason = REASON_PROTOCOL_ERR;
				up(&session->cmnd_sem);
				goto out_reject;
		}
	}
	up(&session->cmnd_sem);

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave handle_snack \n");

	return 0;

out_reject:
	enqueue_reject(conn, reason);
	goto out;
}

/*
 * executed only by midlevel's target_process_thread as a call-back.
 * iscsi_xmit_response: responsible for transmitting data and scsi
 * response
 */
int
iscsi_xmit_response(Target_Scsi_Cmnd * cmnd)
{
	struct iscsi_cmnd *cmd;
	struct iscsi_session *session;

	cmd = search_iscsi_cmnd(cmnd, &session);

	if (!cmd) {
		TRACE_ERROR("%s iscsi_xmit_response could not find cmnd %d\n",
					current->comm, cmnd->id);
		return -1;
	}

	/* bump next expected cmd sequence number if necessary */
	session->exp_cmd_sn += cmd->cmd_sn_increment;
	cmd->cmd_sn_increment = 0;

	cmd->state = ISCSI_DONE;

	TRACE(TRACE_ISCSI, "CmdSN %u ITT %u done by target, ExpCmdSN %u\n",
					  cmd->cmd_sn, cmd->init_task_tag, session->exp_cmd_sn);
	up(&session->cmnd_sem);

	/* wake up the transmit thread */
	if (atomic_read(&cmd->conn->tx_sem.count) <= 0) {
		TRACE(TRACE_DEBUG, "%s up(tx_sem) for cmnd %p, state %d\n",
			  current->comm, cmd, cmd->state);
		up(&cmd->conn->tx_sem);
	}

	return 0;
}

/*
 * executed only by midlevel's target_process_thread as a call-back.
 * iscsi_rdy_to_xfer: function changes the state of the command
 * and wakes up the tx thread to send an R2T to the Initiator if needed
 */
int
iscsi_rdy_to_xfer(Target_Scsi_Cmnd * cmnd)
{
	struct iscsi_cmnd *cmd;
	struct iscsi_session *session;

	cmd = search_iscsi_cmnd(cmnd, &session);

	if (!cmd) {
		TRACE_ERROR("%s iscsi_rdy_to_xfer could not find cmnd %d\n",
					current->comm, cmnd->id);
		return -1;
	}

	/* bump next expected cmd sequence number if necessary */
	session->exp_cmd_sn += cmd->cmd_sn_increment;
	cmd->cmd_sn_increment = 0;

	/* allow r2t generation (if any) for this command */
	cmd->r2t_data = cmd->r2t_data_total;

	if (cmd->data_length == 0)
		/* this command expects no data from initiator, so it is done */
		cmd->state = ISCSI_DONE;
	else if (cmd->state == ISCSI_QUEUE_CMND)
		/* this command is queued, now it is queued and ready to receive */
		cmd->state = ISCSI_QUEUE_CMND_RDY;
	else {
		/* this command is ready to receive data from initiator */
		if (cmd->state != ISCSI_NEW_CMND) {
			TRACE_ERROR("%s iscsi_rdy_to_xfer for CmdSN %u ITT %u opcode 0x%02x"
						" expected state %u, got state %u, setting state %u\n",
						current->comm, cmd->cmd_sn, cmd->init_task_tag,
						cmd->opcode_byte, ISCSI_NEW_CMND, cmd->state,
						ISCSI_BUFFER_RDY);
		}
		cmd->state = ISCSI_BUFFER_RDY;
	}

	up(&session->cmnd_sem);

	/* tell iscsi rx thread that midlevel buffers are now ready */
	TRACE(TRACE_SEM, "iscsi_rdy_to_xfer: unblocking unsolicited_data_sem\n");
	up(&cmd->unsolicited_data_sem);

	if (atomic_read(&cmd->conn->tx_sem.count) <= 0) {
		up(&cmd->conn->tx_sem);
	}

	return 0;
}

/*
 * iscsi_task_mgt_fn_done:
 *
 */
void
iscsi_task_mgt_fn_done(Target_Scsi_Message * msg)
{
	struct iscsi_cmnd *related_command;

	related_command = search_task_mgt_command(msg);

	if (!related_command) {
		TRACE_ERROR("%s Could not find msg %d\n", current->comm, msg->message);
		return;
	}

	related_command->state = ISCSI_MGT_FN_DONE;

	/* wake up the transmit thread */
	if (atomic_read(&related_command->conn->tx_sem.count) <= 0) {
		up(&related_command->conn->tx_sem);
	}

	return;
}

/*
 * executed only by midlevel's target_process_thread as part of a call-back
 * via iscsi_xmit_response() or iscsi_rdy_to_xfer().
 * search_iscsi_cmnd: This goes through individual sessions and then
 * through the command queues for that session to look for a match to
 * a given command.
 * INPUT: Target_Scsi_Cmnd
 * OUTPUT: struct iscsi_cmnd if it finds one, NULL otherwise
 *		   if return value is not NULL, result_sess has been filled in with
 *							the session, and the session->cmnd_sem is locked
 */
static struct iscsi_cmnd * __attribute__ ((no_instrument_function))
search_iscsi_cmnd(Target_Scsi_Cmnd * cmnd, struct iscsi_session **result_sess)
{
	struct iscsi_cmnd *cmd = NULL;
	struct iscsi_session *session;
	struct iscsi_global *host;
	struct list_head *list_ptr;

	if (!cmnd) {
		TRACE_ERROR("%s Cannot search for a NULL command\n", current->comm);
		return NULL;
	}

	host = (struct iscsi_global *) cmnd->device->dev_specific;

	/* non-destructive access to session lists */
	if (down_interruptible(&host->session_read_mutex))
		return NULL;
	if (host->session_readers++ == 0) {
		/* first reader, must get access for all readers */
		if (down_interruptible(&host->session_sem)) {
			up(&host->session_read_mutex);
			return NULL;
		}
	}
	up(&host->session_read_mutex);

	list_for_each(list_ptr, &host->session_list) {
		session = list_entry(list_ptr, struct iscsi_session, sess_link);
		if (down_interruptible(&session->cmnd_sem)) {
			cmd = NULL;
			goto search_iscsi_cmnd_out;
		}

		for (cmd = session->cmnd_list; cmd != NULL; cmd = cmd->next) {
			if (cmd->cmnd == cmnd) {
				*result_sess = session;
				goto search_iscsi_cmnd_out;
			}
		}

		up(&session->cmnd_sem);
	}

search_iscsi_cmnd_out:

	down_interruptible(&host->session_read_mutex);
	if (--host->session_readers == 0) {
		/* last reader, must release access for all readers */
		up(&host->session_sem);
	}
	up(&host->session_read_mutex);

	return cmd;
}

/*
 * search_iscsi_message: This goes through individual sessions and then
 * through the command queues for that session to look for a match to
 * a given message.
 * INPUT: Target_Scsi_Message
 * OUTPUT: struct iscsi_cmnd if it finds one, NULL otherwise
 */
static struct iscsi_cmnd *search_task_mgt_command(Target_Scsi_Message * message) {
	struct iscsi_cmnd *related_task_mgt_command = NULL;
	struct iscsi_session *related_session;
	struct iscsi_global *host;
	struct list_head *list_ptr;

	if (!message) {
		TRACE_ERROR("%s Cannot search for a NULL command\n", current->comm);
		return NULL;
	}

	host = (struct iscsi_global *) message->device->dev_specific;

	/* non-destructive access to session lists */
	if (down_interruptible(&host->session_read_mutex))
		return NULL;
	if (host->session_readers++ == 0) {
		/* first reader, must get access for all readers */
		if (down_interruptible(&host->session_sem)) {
			up(&host->session_read_mutex);
			return NULL;
		}
	}
	up(&host->session_read_mutex);

	list_for_each(list_ptr, &host->session_list) {
		related_session = list_entry(list_ptr, struct iscsi_session, sess_link);
		if (down_interruptible(&related_session->cmnd_sem)) {
			related_task_mgt_command = NULL;
			goto out;
		}

		for (related_task_mgt_command = related_session->cmnd_list;
			 related_task_mgt_command != NULL;
			 related_task_mgt_command = related_task_mgt_command->next) {
			if (related_task_mgt_command->message == message) {
				up(&related_session->cmnd_sem);
				goto out;
			}
		}

		up(&related_session->cmnd_sem);
	}

out:

	down_interruptible(&host->session_read_mutex);
	if (--host->session_readers == 0) {
		/* last reader, must release access for all readers */
		up(&host->session_sem);
	}
	up(&host->session_read_mutex);

	return related_task_mgt_command;
}

/*
 * search_tags: This goes through individual sessions and then
 * through the command queues for that session to look for a match to
 * a given set of tags.
 * INPUT: connection, initiator and target tag
 * OUTPUT: struct iscsi_cmnd if it finds one, NULL otherwise
 */
static struct iscsi_cmnd * __attribute__ ((no_instrument_function))
search_tags(struct iscsi_conn *conn, __u32 init_task_tag, 
				__u32 target_xfer_tag, int dumpall)
{
	struct iscsi_cmnd *cmd = NULL;
	struct iscsi_session *session;

	if (!conn) {
		TRACE_ERROR("%s Cannot search a NULL connection\n", current->comm);
		return NULL;
	}

	/* get session from conn structure directly to support multiple sessions */
	session = conn->session;

	if (down_interruptible(&session->cmnd_sem))
		goto search_tags_out;

	if (dumpall || TRACE_TEST(TRACE_DEBUG)) {
		for (cmd = session->cmnd_list; cmd != NULL; cmd = cmd->next) {
			printk("scsi cmnd %p opcode 0x%02x init_task_tag %d target_xfer_tag"
				   " %d data_done %d xfer length %d stat_sn %u state %u\n",
				   cmd->cmnd, cmd->opcode_byte,
				   cmd->init_task_tag, cmd->target_xfer_tag,
				   cmd->data_done, cmd->data_length, cmd->stat_sn, cmd->state);
		}
	}

	for (cmd = session->cmnd_list; cmd != NULL; cmd = cmd->next) {
		if ((cmd->init_task_tag == init_task_tag)
			&& ((cmd->target_xfer_tag == target_xfer_tag)
					  || (target_xfer_tag == ALL_ONES))) {
			up(&session->cmnd_sem);

			TRACE(TRACE_DEBUG, "Search found the command\n");
			TRACE(TRACE_DEBUG,
				  "scsi cmnd %p, init_task_tag %d target_xfer_tag %d"
				  " data_done %d xfer length %d\n", cmd->cmnd,
				  cmd->init_task_tag, cmd->target_xfer_tag, cmd->data_done,
				  cmd->data_length);

			goto search_tags_out;
		}
	}

	up(&session->cmnd_sem);

search_tags_out:
	return cmd;
}

/**************************************************************************/
/*                            handle response                             */
/**************************************************************************/

/*
 * executed only by the tx thread.
 * called when it is ok to send a Text Response pdu for this command.
 * (i.e., when cmnd state is ISCSI_SEND_TEXT_RESPONSE)
 */
static int
handle_discovery_rsp(struct iscsi_cmnd *cmnd,
					 struct iscsi_conn *conn,
					 struct iscsi_session *session)
{
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	struct iscsi_targ_text_rsp *hdr;
	int size, offset, retval = 0;
	int next_state = ISCSI_SENT;
	struct iscsi_cmnd *next_in_progress = NULL;
	char *ptr;


	if (down_interruptible(&conn->text_in_progress_sem)) {
		return -1;
	}

	/* under protection of this lock, make sure command has not been reset */
	if (cmnd->state != ISCSI_SEND_TEXT_RESPONSE
		|| conn->text_in_progress != cmnd)
		goto out;

	/* get offset to, and amount of, data remaining to send back to initiator */
	offset = cmnd->data_done;
	size = cmnd->data_length - offset;
	ptr = cmnd->ping_data + offset;

	memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_text_rsp *)iscsi_hdr;

	hdr->opcode = ISCSI_TARG_TEXT_RSP;

	if (size > 0 && conn->connection_flags & USE_ONE_KEY_PER_TEXT) {
		/* send just one key per text response (for testing) */
		size = strlen(ptr) + 1;
		cmnd->data_done += size;
		next_state = ISCSI_BLOCKED_SENDING_TEXT;
		next_in_progress = cmnd;
	} else if (size > conn->max_send_length) {
		/* reply too big for one pdu, more text reply pdus follow this one */
		size = conn->max_send_length;
		cmnd->data_done += size;
		hdr->flags |= C_BIT;
		next_state = ISCSI_BLOCKED_SENDING_TEXT;
		next_in_progress = cmnd;
	} else if (cmnd->command_flags & F_BIT) {
		/* this is our last reply and initiator has no more to send */
		hdr->flags |= F_BIT;
		cmnd->target_xfer_tag = hdr->target_xfer_tag = ALL_ONES;
		next_state = ISCSI_SENT;
		next_in_progress = NULL;
	} else {
		/* this is our last reply but initiator has more to send */
		cmnd->data_length = 0;			/* reset for next Request PDU */
		cmnd->in_progress_buffer = cmnd->ping_data;
		cmnd->ping_data = NULL;
		next_state = ISCSI_AWAIT_MORE_TEXT;
		next_in_progress = cmnd;
	}
	hdr->length = cpu_to_be32(size);
	hdr->init_task_tag = cpu_to_be32(cmnd->init_task_tag);
	hdr->stat_sn = cpu_to_be32(conn->stat_sn++);

	if (down_interruptible(&session->cmnd_sem)) {
		goto out1;
	}
	/* generate next TTT if we expect initiator to send another text request */
	if (next_in_progress) {
		cmnd->target_xfer_tag = generate_next_ttt(session);
		hdr->target_xfer_tag = cpu_to_be32(cmnd->target_xfer_tag);
	}

	hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);
	if (!(cmnd->opcode_byte & I_BIT)) {
		/* the text command pdu was not immediate, CmdSN advances */
		session->max_cmd_sn++;
	}
	cmnd->state = next_state;
	up(&session->cmnd_sem);

	conn->text_in_progress = next_in_progress;

	if (send_hdr_plus_1_data(conn, iscsi_hdr, ptr, size) < 0) {
		goto out1;
	}

	TRACE(TRACE_ISCSI, "%s text response sent, ITT %u\n", current->comm,
		  cmnd->init_task_tag);
	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_text_rsp(hdr);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		up(&conn->tx_sem);
	}

out:
	up(&conn->text_in_progress_sem);
	return retval;

out1:
	cmnd->state = ISCSI_DEQUEUE;
	cmnd->init_task_tag = ALL_ONES;
	conn->text_in_progress = NULL;
	retval = -1;
	goto out;
}

/*
 * executed only by the tx thread.
 * Called only when cmnd state is ISCSI_ASK_FOR_MORE_TEXT.
 * Sends back an empty text response so initiator can send more text data.
 * This text response pdu has C=0 F=0 DSL=0 ITT=from request TTT=newly generated
 */
static int
ask_for_more_text(struct iscsi_cmnd *cmnd,
				  struct iscsi_conn *conn,
				  struct iscsi_session *session)
{
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	struct iscsi_targ_text_rsp *hdr;
	int retval = 0;


	if (down_interruptible(&conn->text_in_progress_sem)) {
		return -1;
	}

	/* under protection of this lock, make sure command has not been reset */
	if (cmnd->state != ISCSI_ASK_FOR_MORE_TEXT
		|| conn->text_in_progress != cmnd)
		goto out1;

	memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_text_rsp *)iscsi_hdr;

	hdr->opcode = ISCSI_TARG_TEXT_RSP;

	hdr->init_task_tag = cpu_to_be32(cmnd->init_task_tag);
	hdr->stat_sn = cpu_to_be32(conn->stat_sn++);

	if (down_interruptible(&session->cmnd_sem)) {
		retval = -1;
		goto out1;
	}
	/* generate the next TTT */
	cmnd->target_xfer_tag = generate_next_ttt(session);
	hdr->target_xfer_tag = cpu_to_be32(cmnd->target_xfer_tag);

	hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);
	if (!(cmnd->opcode_byte & I_BIT)) {
		/* the text command pdu was not immediate, CmdSN advances */
		session->max_cmd_sn++;
	}
	up(&session->cmnd_sem);

	cmnd->state = ISCSI_AWAIT_MORE_TEXT;

	if (send_hdr_plus_1_data(conn, iscsi_hdr, NULL, 0) < 0) {
		retval = -1;
		goto out1;
	}

	TRACE(TRACE_ISCSI, "%s text response sent, ITT %u\n", current->comm,
		  cmnd->init_task_tag);
	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_text_rsp(hdr);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		up(&conn->tx_sem);
	}

out1:
	up(&conn->text_in_progress_sem);
	return retval;
}

/*
 * executed only by the tx thread.
*/
static int
handle_logout_rsp(struct iscsi_cmnd *cmnd,
				  struct iscsi_conn *conn,
				  struct iscsi_session *session)
{
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	struct iscsi_targ_logout_rsp *hdr;

	memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_logout_rsp *) iscsi_hdr;

	hdr->opcode = ISCSI_TARG_LOGOUT_RSP;

	hdr->flags |= F_BIT;

	/* for now, we always send response = 0, connection or session closed ok */

	hdr->init_task_tag = cpu_to_be32(cmnd->init_task_tag);
	hdr->stat_sn = cpu_to_be32(conn->stat_sn++);

	if (down_interruptible(&session->cmnd_sem))
		return -1;
	hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);
	if (!(cmnd->opcode_byte & I_BIT)) {
		/* the logout command pdu was not immediate, CmdSN advances */
		session->max_cmd_sn++;
	}
	up(&session->cmnd_sem);

	cmnd->state = ISCSI_SENT;

	/* connection now logged out, do not send any more commands after this */
	conn->connection_flags |= CONN_LOGGED_OUT;

	if (send_hdr_only(conn, iscsi_hdr) < 0) {
		return -1;
	}

	TRACE(TRACE_ISCSI, "%s logout response sent\n", current->comm);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_logout_rsp(hdr);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		up(&conn->tx_sem);
	}

	return 0;
}

/*
 * executed only by the tx thread.
 * sends a NopIn PDU to initiator
 * either in response to an earlier NopOut ping from initiator
 * or as a new NopIn ping generated by the target
 */
static int
handle_nopin(struct iscsi_cmnd *cmnd,
			 struct iscsi_conn *conn,
			 struct iscsi_session *session)
{
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	struct iscsi_targ_nopin *hdr;

	memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_nopin *) iscsi_hdr;

	hdr->opcode = ISCSI_TARG_NOP_IN;

	hdr->flags |= F_BIT;

	hdr->length = cpu_to_be32(cmnd->data_length);
	hdr->init_task_tag = cpu_to_be32(cmnd->init_task_tag);
	hdr->target_xfer_tag = cpu_to_be32(cmnd->target_xfer_tag);

	/* Draft 20, Section 10.19.2 StatSN
	 * "The StatSN field will always contain the next StatSN. However,
	 * when the Initiator Task Tag is set to 0xffffffff StatSN for the
	 * connection is not advanced after this PDU is sent."
	 */
	hdr->stat_sn = cpu_to_be32(conn->stat_sn);
	if (cmnd->init_task_tag != ALL_ONES)
		conn->stat_sn++;

	if (down_interruptible(&session->cmnd_sem))
		return -1;
	hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);
	if (!(cmnd->opcode_byte & I_BIT)) {
		/* the nopout command pdu was not immediate, CmdSN advances */
		session->max_cmd_sn++;
	}
	up(&session->cmnd_sem);

	if (cmnd->target_xfer_tag == ALL_ONES) {
		/* target not expecting a reply from initiator */
		cmnd->state = ISCSI_SENT;
	}

	if (send_hdr_plus_1_data(conn, iscsi_hdr, cmnd->ping_data,
													cmnd->data_length) < 0) {
		return -1;
	}

	TRACE(TRACE_DEBUG, "sent NopIn CmdSN %u, ExpCmdSN %u, ITT %u "
						"opcode 0x%02x, state %u\n",
						cmnd->cmd_sn, session->exp_cmd_sn,
						cmnd->init_task_tag, cmnd->opcode_byte, cmnd->state);

	TRACE(TRACE_ISCSI, "%s nopin sent\n", current->comm);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_nopin(hdr);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		up(&conn->tx_sem);
	}

	my_kfree((void *)&cmnd->ping_data, "data_buf");

	return 0;
}

/*
 * executed only by the tx thread.
 * generates and sends a NopIn "ping" PDU to initiator
 */
static int
generate_nopin(struct iscsi_conn *conn,
			   struct iscsi_session *session)
{
	struct iscsi_cmnd *cmnd;
	struct iscsi_cmnd *temp;

	if ((cmnd = get_new_cmnd()) == NULL) {
		return -1;
	}

	/* set up new command to look like an immediate NopOut sent by initiator */
	cmnd->conn = conn;
	cmnd->session = session;
	cmnd->opcode_byte = ISCSI_INIT_NOP_OUT | I_BIT;
	cmnd->init_task_tag = ALL_ONES;
	cmnd->state = ISCSI_NOPIN_SENT;

	if (down_interruptible(&session->cmnd_sem)) {
		my_kfree((void**)&cmnd, "cmnd");
		return -1;
	}
	cmnd->target_xfer_tag = generate_next_ttt(session);

	/* Add this command to end of the queue */
	for (temp = session->cmnd_list; temp != NULL; temp = temp->next) {
		if (temp->next == NULL)
			break;
	}

	if (temp)
		temp->next = cmnd;
	else
		session->cmnd_list = cmnd;
	up(&session->cmnd_sem);

	TRACE(TRACE_ISCSI, "Send NopIn ping, TTT %u\n", cmnd->target_xfer_tag);

	return handle_nopin(cmnd, conn, session);
}

#define UNDERFLOW_FLAG	0x0001
#define OVERFLOW_FLAG	0x0002
#define SEND_SENSE_FLAG	0x0004
#define LAST_SEQ_FLAG	0x0010

static __u32
do_command_status(struct iscsi_cmnd *cmnd, Scsi_Request *req,
				  int *data_left, int *residual_count)
{
	int transfer = 0;
	int data_length_left;
	__u32 flags = 0;

	data_length_left = req->sr_bufflen;

	TRACE(TRACE_DEBUG, "Sense: %02x %02x %02x %02x %02x %02x %02x %02x\n",
		  req->sr_sense_buffer[0], req->sr_sense_buffer[1],
		  req->sr_sense_buffer[2], req->sr_sense_buffer[3],
		  req->sr_sense_buffer[4], req->sr_sense_buffer[5],
		  req->sr_sense_buffer[6], req->sr_sense_buffer[7]);

	if ((req->sr_sense_buffer[0] & 0x7e) == 0x70) {
		/* current or deferred error -- else, we don't send the sense buffer */
		flags |= SEND_SENSE_FLAG;

		if (req->sr_sense_buffer[0] & 0x80) {
			/* bytes 3 thru 6 contain valid data */
			transfer = (req->sr_sense_buffer[3] << 24) |
				(req->sr_sense_buffer[4] << 16) |
				(req->sr_sense_buffer[5] << 8) | req->sr_sense_buffer[6];
			TRACE(TRACE_DEBUG, "information in sense data: %d\n", transfer);
		}

		if ((req->sr_sense_buffer[2] & 0x20) != 0) {		/* ILI bit set */
				data_length_left -= transfer;
		}
	} else if (req->sr_command) {
		/* ensure correct DataSegmentLength and ResidualCount */
		TRACE(TRACE_DEBUG, "data_length_left %d, sr_command->resid %d\n",
				data_length_left, req->sr_command->resid);
		data_length_left -= req->sr_command->resid;
	}

	/* check overflow or underflow problem, cdeng 3/24/02 */
	if (data_length_left > cmnd->data_length) {
		*residual_count = data_length_left - cmnd->data_length;
		data_length_left = cmnd->data_length;
		flags |= OVERFLOW_FLAG;
	} else if (data_length_left < cmnd->data_length) {
		*residual_count = cmnd->data_length - data_length_left;
		flags |= UNDERFLOW_FLAG;
	}

	TRACE(TRACE_DEBUG,"data_length_left %d, residual_count %d, flags 0x%08x\n",
		  data_length_left, *residual_count, flags);

	*data_left = data_length_left;
	return flags;
}

/* 
 * executed only by the tx thread
 * called directly when state is ISCSI_RESEND_STATUS
 * and called indirectly via handle_iscsi_done() when state is ISCSI_DONE
 * to send ISCSI Response PDU, with status information if needed.
 * Separated Data-In from Status Response PDU sending for Status 
 * Error Recovery - SAI 
 */
static int
send_iscsi_response(struct iscsi_cmnd *cmnd,
					struct iscsi_conn *conn,
					struct iscsi_session *session)
{
	struct iscsi_targ_scsi_rsp *rsp;
	Scsi_Request *req;
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	int data_length_left = 0;
	struct sense_data_buf {
		__u16	len;						/* added by iscsi standard */
		__u8	data[SCSI_SENSE_BUFFERSIZE];/* direct from scsi */
		} sense_data = {0};
	__u32 flags = 0;
	int residual_count = 0;

	TRACE(TRACE_DEBUG, "send_scsi_response\n");

	memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	rsp = (struct iscsi_targ_scsi_rsp *)iscsi_hdr;
	rsp->opcode = ISCSI_TARG_SCSI_RSP;
	rsp->flags |= F_BIT;
	rsp->init_task_tag = cpu_to_be32(cmnd->init_task_tag);

	req = cmnd->cmnd->req;
	if ((req->sr_data_direction == DMA_FROM_DEVICE)
									&& host_byte(req->sr_result) == DID_OK) {
		flags = do_command_status(cmnd,req,&data_length_left,&residual_count);
		rsp->exp_data_sn = cpu_to_be32(cmnd->data_sn);
	}

	/* cdeng 3/24/02 */
	if (flags & OVERFLOW_FLAG) {
		rsp->flags |= O_BIT;
		rsp->resid = cpu_to_be32(residual_count);
	} else if (flags & UNDERFLOW_FLAG) {
		rsp->flags |= U_BIT;
		rsp->resid = cpu_to_be32(residual_count);
	}

	/* RDR - do NOT assign and increment stat_sn on a retransmission */
	if (!cmnd->retransmit_flg) {
		cmnd->stat_sn = conn->stat_sn++;
	}
	rsp->stat_sn = cpu_to_be32(cmnd->stat_sn);
	rsp->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	rsp->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);

	if (flags & SEND_SENSE_FLAG) {
		/* sense data has to be sent as part of SCSI Response pdu */
		rsp->status = CHECK_CONDITION << 1;	/* why does Linux do this shift? */
		sense_data.len = SCSI_SENSE_BUFFERSIZE;
		if (flags & UNDERFLOW_FLAG) {
			TRACE(TRACE_DEBUG, "underflow is found\n");
			memset(sense_data.data, 0x0, SCSI_SENSE_BUFFERSIZE);
			sense_data.data[0] = 0xf0;		/* scsi valid bit, code 70h */
			sense_data.data[2] = 0x20;		/* scsi ILI bit */
			sense_data.data[7] = 0x07;		/* scsi additional length = 7 */
			memcpy(sense_data.data + 3, req->sr_sense_buffer + 3, 4);
		} else {
			TRACE(TRACE_DEBUG, "sense key 0x%x\n",
				  req->sr_sense_buffer[2] & 0xf);
			memcpy(sense_data.data,req->sr_sense_buffer,SCSI_SENSE_BUFFERSIZE);
		}
		sense_data.len += 2;	/* for the extra iscsi 2-byte length header */
	}
	rsp->length = cpu_to_be32(sense_data.len);

	TRACE(TRACE_DEBUG, "send_iscsi_response: sending status for cmnd_rn %.8x "
		  "init_task_tag %.8x target_xfer_tag %.8x\n", cmnd->cmd_sn,
		  cmnd->init_task_tag, cmnd->target_xfer_tag);

	cmnd->retransmit_flg = 0;
	cmnd->state = ISCSI_SENT;
	if (send_hdr_plus_1_data(conn, iscsi_hdr, &sense_data, sense_data.len) < 0){
		return -1;
	}

	TRACE(TRACE_ISCSI, "%s scsi response sent, ITT %u\n",
		  current->comm, cmnd->init_task_tag);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_scsi_rsp(rsp);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		up(&conn->tx_sem);
	}
	return 0;
}

/*
 * executed only by the tx thread when state is ISCSI_DONE.
 * called only from handle_iscsi_done() on a READ-type command.
 * sends all the DataIn pdus for this command to the initiator. 
 *
 * During this sending:
 *	data_length_left	number of bytes left to send in following sequences.
 *	seq_length			number of bytes left to send in this sequence.
 *	seq_low_byte		offset to lowest byte in this sequence.
 *	seq_limit_byte		offset to highest byte plus one in this sequence.
 *	pdu_offset			offset to first byte in this pdu.
 *	data_payload_length	number of data bytes to send in this pdu.
 *	total_data_length	total number of all bytes to send in this pdu
 */
static int
send_read_data(struct iscsi_cmnd *cmnd,
			   struct iscsi_conn *conn,
			   struct iscsi_session *session,
			   int *phase_collapse)
{
	int err = 0;
	__u32 flags;
	struct iscsi_targ_scsi_data_in *hdr;
	struct scatterlist *st_list;
	int residual_count;
	int prevsn;
	struct iovec *iov;
	int niov;
	Scsi_Request *req;
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	__u32 pad_bytes;
	int padding;
	int data_length_left;
	int seq_length;
	int seq_low_byte, seq_limit_byte;
	int pdu_offset;
	int data_payload_length;
	int total_data_length;

/* Daren Hayward, darenh@4bridgeworks.com */
#if defined(MANGLE_INQUIRY_DATA)
	int miov, siov;
#endif

	TRACE(TRACE_DEBUG, "sending sense data for cmnd_sn "
		  "%u, ITT %u, TTT %u\n", cmnd->cmd_sn, cmnd->init_task_tag,
		  cmnd->target_xfer_tag);

	req = cmnd->cmnd->req;
	flags = do_command_status(cmnd, req, &data_length_left, &residual_count);

	/* Store the previous data sequence for error recovery - SAI */
	prevsn = cmnd->prev_data_sn;
	if (!cmnd->retransmit_flg)
		cmnd->prev_data_sn = cmnd->data_sn;

	/* one of these is changed before use, depending on DataSequenceInOrder */
	seq_low_byte = 0;
	seq_limit_byte = data_length_left;

	/* once around this loop for each input sequence sent to initiator */
	while (data_length_left > 0) {
		if (data_length_left <= session->oper_param->MaxBurstLength) {
			/* all remaining data fits in this last sequence */
			seq_length = data_length_left;
			flags |= LAST_SEQ_FLAG;
		} else {
			/* another sequence will follow this one */
			seq_length = session->oper_param->MaxBurstLength;
		}
		data_length_left -= seq_length;

		if (session->oper_param->DataSequenceInOrder)
			seq_limit_byte = seq_low_byte + seq_length;	/* in order */
		else
			seq_low_byte = seq_limit_byte - seq_length;	/* reverse order */

		if (session->oper_param->DataPDUInOrder)
			pdu_offset = seq_low_byte;					/* in order */
		else
			pdu_offset = seq_limit_byte;				/* reverse order */

		/* once around this loop for each pdu in this sequence */
		while (seq_length > 0) {
			TRACE(TRACE_DEBUG, "data_length_left: %d, seq_length: %d\n",
				  data_length_left, seq_length);

			memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
			hdr = (struct iscsi_targ_scsi_data_in *) iscsi_hdr;

			hdr->opcode = ISCSI_TARG_SCSI_DATA_IN;
			hdr->target_xfer_tag = ALL_ONES;

			if (seq_length > conn->max_send_length) {
				data_payload_length = conn->max_send_length;
			} else {
				/* this is the last data-in pdu in this sequence */
				data_payload_length = seq_length;
				hdr->flags |= F_BIT;

				if ((session->targ_snack_flg & DATACK_SNACK_ENABLE)
							&& session->oper_param->ErrorRecoveryLevel > 0) {
					/* set A bit on each data-in pdu that ends a sequence.
					 * (only legal at ErrorRecoveryLevel 1 or 2)
					 *
					 * Draft 20, Section 10.7.4 Target Transfer Tag and LUN
					 *
					 * On incoming data, the Target Transfer Tag and LUN
					 * MUST be provided by the target if the A bit is set
					 * to 1; otherwise they are reserved.
					 */
					hdr->flags |=A_BIT;
					hdr->target_xfer_tag = cpu_to_be32(cmnd->target_xfer_tag);
					hdr->lun = cmnd->lun;
					TRACE(TRACE_ERROR_RECOVERY,
						  "Send DataIn, A=1, ITT %u, TTT %u, dataSN %u\n",
						  cmnd->init_task_tag, cmnd->target_xfer_tag,
						  cmnd->data_sn);
				}

				if ((flags & LAST_SEQ_FLAG) && !(flags & SEND_SENSE_FLAG)) {
					/* this is also the last data-in pdu for this command and
					 * we are not sending any sense data, try phase collapse.
					 * 0 never collapse, 1 always collapse, 
					 * 2 collapse this time, -1 do not collapse this time
					 */
					if (devdata->phase_collapse > 0) {
						hdr->flags |= S_BIT;
						*phase_collapse = 1;

						if (devdata->phase_collapse == 2)
							devdata->phase_collapse = -1;
					} else if (devdata->phase_collapse == -1)
						devdata->phase_collapse = 2;
				}
			}

			if (!session->oper_param->DataPDUInOrder)
				pdu_offset -= data_payload_length;		/* reverse order */

			hdr->length = cpu_to_be32(data_payload_length);
			hdr->init_task_tag = cpu_to_be32(cmnd->init_task_tag);

			if (hdr->flags & S_BIT) {
				/* RDR - already assigned and incremented stat_sn before 
				 * a retransmission
				 */
				if (!cmnd->retransmit_flg) {
					cmnd->stat_sn = conn->stat_sn++;
				}
				hdr->stat_sn = cpu_to_be32(cmnd->stat_sn);
			}

			hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
			hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);

			/* re-transmit only the requested Data PDU - SAI */
			if (!cmnd->retransmit_flg)
				hdr->data_sn = cpu_to_be32(cmnd->data_sn++);
			else
				hdr->data_sn = cpu_to_be32(prevsn);

			hdr->offset = cpu_to_be32(pdu_offset);

			st_list = (struct scatterlist *)cmnd->cmnd->req->sr_buffer;

			/* if anything out of order, need to skip through command's
			 * scatter-gather list to find start of this pdu
			 */
			if (!session->oper_param->DataPDUInOrder
								|| !session->oper_param->DataSequenceInOrder) {
				cmnd->scatter_list_offset = skip_thru_sg_list(st_list,
													&cmnd->scatter_list_count,
													pdu_offset);
			}

			st_list += cmnd->scatter_list_count;

			TRACE(TRACE_DEBUG,
				  "offset into sg entry %d, offset into sg list" " %d\n",
				  cmnd->scatter_list_offset, cmnd->scatter_list_count);

			/* get number of iovec slots needed to send this DataIn payload */
			niov = find_iovec_needed(data_payload_length, st_list,
									 cmnd->scatter_list_offset);
			if (niov <= 0) {
				TRACE_ERROR("%s Trouble in find_iovec_needed\n",current->comm);
				err = -1;
				goto out;
			}

			if (session->oper_param->DataPDUInOrder)
				pdu_offset += data_payload_length;	/* in order */

			/* now add on number of additional iovec slots this pdu needs */
			niov++;					/* add one for header */
			total_data_length = data_payload_length + ISCSI_HDR_LEN;
			/* Daren Hayward, darenh@4bridgeworks.com */
			#if defined(MANGLE_INQUIRY_DATA)
				siov = 1;
			#endif

			if (conn->hdr_crc) {
				niov++;				/* add one for hdr_crc */
				total_data_length += CRC_LEN;
				/* Daren Hayward, darenh@4bridgeworks.com */
				#if defined(MANGLE_INQUIRY_DATA)
					siov++;
				#endif
			}

			/* Daren Hayward, darenh@4bridgeworks.com */
			#if defined(MANGLE_INQUIRY_DATA)
				miov = niov;
			#endif

			padding = (-data_payload_length) & 3;
			if (padding) {
				niov++;				/* add one for padding */
				total_data_length += padding;
			}

			if (conn->data_crc) {
				niov++;				/* add one for data_crc */
				total_data_length += CRC_LEN;
			}

			iov = (struct iovec *)my_kmalloc(niov * sizeof(struct iovec),
											 "iovec");
			if (!iov) {
				TRACE(TRACE_DEBUG, "handle_iscsi_done: No space for iov\n");
				err = -1;
				goto out;
			}

			/* now go back and fill in the iovec slots */
			iov[0].iov_base = iscsi_hdr;
			iov[0].iov_len = ISCSI_HDR_LEN;

			cmnd->scatter_list_count +=
						fill_iovec(iov, 1 + conn->hdr_crc, niov, st_list,
								   &cmnd->scatter_list_offset,
								   data_payload_length);

/* Daren Hayward, darenh@4bridgeworks.com */
#if defined(MANGLE_INQUIRY_DATA)
			if (cmnd->cdb[0] == INQUIRY
								&& mangle_inquiry_data(cmnd, iov, siov, miov)) {
				err = -1;
				goto out;
			}
#endif

			if (padding) {
				iov[niov - 1 - conn->data_crc].iov_base = &pad_bytes;
				iov[niov - 1 - conn->data_crc].iov_len = padding;
				TRACE(TRACE_DEBUG, "padding attached: %d bytes\n", padding);
			}

			if (conn->data_crc) {
					iov[niov - 1].iov_len = CRC_LEN;
			}

			/* skip data PDU re-transmissions that are not asked for - SAI*/
			if (cmnd->retransmit_flg) {
				if (prevsn < cmnd->startsn || prevsn > cmnd->endsn)
					goto skip_transmission;
				TRACE_WARNING("%s Retransmitting DataIn, ITT %u, DataSN %u,"
							  " Buffer Offset %u\n", current->comm,
							  cmnd->init_task_tag,
							  prevsn, be32_to_cpu(hdr->offset));
			}

			/* stop sending more data_in pdus in case command was aborted */
			if (cmnd->state == ISCSI_DEQUEUE) {
				err = 0;
				goto out1;
			}

			err = iscsi_tx_data(conn, iov, niov, total_data_length);

			if (err != total_data_length) {
				TRACE_ERROR("%s Trouble in iscsi_tx_data, expected %d "
							"bytes, got %d\n", current->comm,
					 		total_data_length, err);
				err = -1;
				goto out1;
			}

			if (TRACE_TEST(TRACE_ISCSI_FULL)) {
				TRACE(TRACE_ISCSI_FULL, "%s DataIn sent, offset %u\n",
					  current->comm, be32_to_cpu(hdr->offset));
				print_targ_scsi_data_in(hdr);
			}

			/* Added for error recovery - SAI */
skip_transmission:

			my_kfree((void**)&iov, "iovec");

			/* Increment the retransmitted data sequence number - SAI */
			prevsn++;

			seq_length -= data_payload_length;

			TRACE(TRACE_DEBUG,
				  "data sent %d data left in seq %d sg_list_offset %d "
				  "sg_entry_offset %d\n", err, seq_length,
				  cmnd->scatter_list_count, cmnd->scatter_list_offset);
		}

		/* end of this sequence of DataIn pdus, set up for next seq (if any) */
		if (session->oper_param->DataSequenceInOrder)
			seq_low_byte = seq_limit_byte;		/* in order */
		else
			seq_limit_byte = seq_low_byte;		/* reverse order */
	}

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave send_read_data, err = %d\n", err);
	return err;
out1:
	my_kfree((void**)&iov, "iovec");
	goto out;
}

/*
 * executed only by the tx thread when state is ISCSI_DONE.
 * handle_iscsi_done: This is responsible for building the data (if any)
 * and response headers and transmitting them to the correct initiator.
 * The command state is set to DEQUEUE to reflect that it is done
 * executing
 * INPUT: command that is done.
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
handle_iscsi_done(struct iscsi_cmnd *cmnd,
				  struct iscsi_conn *conn,
				  struct iscsi_session *session)
{
	Scsi_Request *req;
	int err = 0;
	int phase_collapse = 0;

	TRACE(TRACE_ENTER_LEAVE, "Enter handle_iscsi_done, cmnd %p, state %d\n",
		  cmnd, cmnd->state);

	if (cmnd->cmnd == NULL) {
		TRACE_ERROR("%s cmnd is NULL for cmnd %p\n", current->comm, cmnd);
		err = -1;
		goto out;
	}

	req = cmnd->cmnd->req;

	if (req == NULL) {
		TRACE_ERROR("%s req is NULL for cmnd %p\n", current->comm, cmnd);
		err = -1;
		goto out;
	}

	if ((req->sr_data_direction == DMA_FROM_DEVICE)
									&& host_byte(req->sr_result) == DID_OK) {
			err = send_read_data(cmnd, conn, session, &phase_collapse);
			if (err < 0)
				goto out;
	}

	if (!(cmnd->opcode_byte & I_BIT)) {
		/* the scsi command pdu was not immediate, CmdSN advances */
		if (down_interruptible(&session->cmnd_sem)) {
			err = -1;
			goto out;
		}
		session->max_cmd_sn++;
		up(&session->cmnd_sem);
	}

	err = 0;

	/* Separated Response for Error Recovery - SAI */
	if (cmnd->retransmit_flg) {
		cmnd->retransmit_flg = 0;
		cmnd->state = ISCSI_SENT;
	} else {
		if (phase_collapse == 0)
			err = send_iscsi_response(cmnd, conn, session);
		else
			cmnd->state = ISCSI_SENT;
	}

	/* in case there are any out-of-order commands that are now in-order */
	check_queued_cmnd(session);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		up(&conn->tx_sem);
	}

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave handle_iscsi_done, err = %d\n", err);
	return err;
}

/*
 * executed only by the tx thread.
 * when state is ISCSI_MGT_FN_DONE:
 * handle_iscsi_mgt_fn_done: This is responsible for building the task mgt
 * response header and transmitting it to the correct initiator.
 * The command state is set to DEQUEUE to reflect that it is finished.
 * INPUT: command that is done.
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
handle_iscsi_mgt_fn_done(struct iscsi_cmnd *cmnd,
						 struct iscsi_conn *conn,
						 struct iscsi_session *session)
{
	struct iscsi_targ_task_mgt_response rsp;
	struct iscsi_cmnd *aborted_command = NULL;

	/* most fields in task management function response pdu go back as 0 */
	memset(&rsp, 0x0, ISCSI_HDR_LEN);

	rsp.opcode = ISCSI_TARG_TASK_MGMT_RSP;
	rsp.flags |= F_BIT;
	rsp.response = cmnd->response;
	rsp.init_task_tag = htonl(cmnd->init_task_tag);
	rsp.stat_sn = htonl(conn->stat_sn++);

	if (down_interruptible(&session->cmnd_sem)) {
		return -1;
	}
	rsp.exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	rsp.max_cmd_sn = cpu_to_be32(session->max_cmd_sn);
	if (!(cmnd->opcode_byte & I_BIT)) {
		/* the task management function request pdu was not immediate,
		 * CmdSN advances
		 */
		session->max_cmd_sn++;
	}
	up(&session->cmnd_sem);

	cmnd->state = ISCSI_DEQUEUE;

	if (send_hdr_only(conn, &rsp) < 0) {
		return -1;
	}

	TRACE(TRACE_ISCSI, "%s task mgt response sent\n", current->comm);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_task_mgt_response(&rsp);

	aborted_command = search_tags(conn, cmnd->ref_task_tag, ALL_ONES, 0);

	if (down_interruptible(&session->cmnd_sem))
		return -1;

	if (cmnd->ref_cmd_sn == session->exp_cmd_sn) {
		session->exp_cmd_sn++;
		if (aborted_command != NULL)
			aborted_command->cmd_sn_increment = 0;
	}

	/* clear the aborted command struct entry by dequeue routine */
	if (aborted_command != NULL)
		aborted_command->state = ISCSI_DEQUEUE;

	up(&session->cmnd_sem);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		up(&conn->tx_sem);
	}
	return 0;
}

/*
 * executed only by the tx thread.
 * iscsi_dequeue: this function frees up a command's allocated memory
 * after it has been removed from the session list (before the call)
 * INPUT: command which has to be freed
 */
static void
iscsi_dequeue(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn)
{
	TRACE(TRACE_DEBUG, "free cmnd with ITT %u\n", cmnd->init_task_tag);

	if (cmnd->cmnd != NULL) {
		if (scsi_target_done(cmnd->cmnd) < 0) {
			TRACE_ERROR("%s scsi_target_done returned an error\n",
						current->comm);
		}
	}

	/* if this command was the text command currently in progress on this
	 * connection, unblock the connection so it can process more text commands.
	 */
	if (!down_interruptible(&conn->text_in_progress_sem)) {
		if (conn->text_in_progress == cmnd)
			conn->text_in_progress = NULL;
		up(&conn->text_in_progress_sem);
	}

	/* Free the R2T cookie if any - SAI */
	free_r2t_cookie(cmnd);
	free_data_list(cmnd);
	my_kfree((void **)&cmnd->in_progress_buffer, "text buf");
	my_kfree((void **)&cmnd->ping_data, "data_buf");
	my_kfree((void **)&cmnd, "cmnd");
}

/*
 * executed only by the tx thread.
 * iscsi_tx_reject: this function transmits a reject to the Initiator
 * the attached 48-byte data segment is the header of the rejected PDU
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
int
iscsi_tx_rjt(struct iscsi_conn *conn, __u8 *bad_hdr, __u8 reason)
{
	int err = 0;
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	struct iscsi_targ_rjt *hdr;

	memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_rjt *) iscsi_hdr;

	hdr->opcode = ISCSI_TARG_RJT;
	hdr->flags |= F_BIT;
	hdr->reason = reason;
	hdr->length = cpu_to_be32(ISCSI_HDR_LEN);
	/* init task tag must be all ones - new 18_04 SAI */
	hdr->init_task_tag = ALL_ONES;

	hdr->stat_sn = cpu_to_be32(conn->stat_sn++);

	if (conn->session) {
		hdr->exp_cmd_sn = cpu_to_be32(conn->session->exp_cmd_sn);
		hdr->max_cmd_sn = cpu_to_be32(conn->session->max_cmd_sn);
	}

	if (send_hdr_plus_1_data(conn, iscsi_hdr, bad_hdr, ISCSI_HDR_LEN) < 0) {
		err = -1;
		goto out;
	}

	TRACE_WARNING("%s Send Reject\n", current->comm);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_rjt(hdr);

out:
	return err;
}

/*
 * This is executed only by the rx thread, NOT the tx thread as name implies!!!
 * During a login, this is called by the rx thread to send a login reject pdu.
 * The status_class parameter MUST not be zero.
 * Therefore, conn->session will not be valid when called by rx thread.
 * relevant parts of the standard are:
 *
 * Draft 20 section 5.3.1 Login Phase Start
 *	"-Login Response with Login reject. This is an immediate rejection from
 *	the target that causes the connection to terminate and the session to
 *	terminate if this is the first (or only) connection of a new session.
 *	The T bit and the CSG and NSG fields are reserved."
 *
 * Draft 20 section 10.13.1 Version-max
 *	"All Login responses within the Login Phase MUST carry the same
 *	Version-max."
 *
 * Draft 20 section 10.13.2 Version-active
 *	"All Login responses within the Login Phase MUST carry the same
 *	Version-active."
 *
 * Draft 20 section 10.13.3 TSIH
 *	"With the exception of the Login Final-Response in a new session,
 *	this field should be set to the TSIH provided by the initiator in
 *	the Login Request."
 *
 * Draft 20 section 10.13.4 StatSN
 *	"This field is only valid if the Status-Class is 0."
 *
 * Draft 20 section 10.13.5 Status-Class and Status-Detail
 *	"0 - Success - indicates that the iSCSI target successfully received,
 *	"understood, and accepted the request. The numbering fields (StatSN,
 *	ExpCmdSN, MaxCmdSN) are only valid if Status-Class is 0."
 */
static int
iscsi_tx_login_reject(struct iscsi_conn *conn,
				   struct iscsi_init_login_cmnd *pdu,
				   __u8 status_class, __u8 status_detail)
{
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	struct iscsi_targ_login_rsp *hdr;

	/* most fields in the Login reject header are zero (reserved) */
	memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_login_rsp *) iscsi_hdr;

	hdr->opcode = ISCSI_TARG_LOGIN_RSP;

	/* the T bit and CSG and NSG fields are reserved on a Login reject */

	hdr->version_max = ISCSI_MAX_VERSION;
	hdr->version_active = ISCSI_MIN_VERSION;

	/* there is no data attached to this login reject */

	memcpy(hdr->isid, pdu->isid, 6);
	hdr->tsih = cpu_to_be16(pdu->tsih);
	hdr->init_task_tag = cpu_to_be32(pdu->init_task_tag);

	/* The numbering fields (StatSN, ExpCmdSn, MaxCmdSN) are only valid if
	 * Status-Class is 0, which it is not on a Login reject */

	hdr->status_class = status_class;
	hdr->status_detail = status_detail;

	if (send_hdr_only(conn, iscsi_hdr) < 0) {
		return -1;
	}

	TRACE(TRACE_ISCSI, "%s login response sent\n", current->comm);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_login_rsp(hdr);

	return 0;
}

/*
 * executed only by the tx thread when state is ISCSI_BUFFER_RDY.
 * iscsi_tx_r2t: this function transmits a ready-to-transfer to the
 * Initiator so that it can transmit the data necessary for a WRITE
 * INPUT: command for which R2T needs to be transferred
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 *
 * during a WRITE, the following fields in struct iscsi_cmnd maintain state:
 *
 *	data_length		The total number of bytes to be received in this WRITE
 *					(includes bytes in immediate and unsolicited pdus)
 *					Set from EDTL in iSCSI command and never changed
 *
 *	data_done		Number of bytes correctly received so far in this WRITE
 *					(includes bytes in immediate and unsolicited pdus)
 *					When sequences and PDUs are being sent in order,
 *					this value will be the expected offset of the next
 *					DataOut PDU we should receive from the initiator.
 *					WRITE should be finished when this equals data_length
 *
 *	first_burst_len	Number of unsolicited bytes received so far in this WRITE
 *					(counts only bytes in immediate and unsolicited pdus)
 *
 *	immediate_data_present		1 until all immediate data have been read
 *
 *	unsolicited_data_present	1 until all unsolicited pdus have been read
 *
 *	next_burst_len	Number of solicited bytes received so far in this sequence
 *					reset to 0 on receiving last DataOut PDU in a sequence
 *					which is used to detect first DataOut PDU in next sequence
 *
 *	r2t_data_total	The total number of bytes to be solicited by R2Ts
 *					Set during iSCSI command processing and never changed
 *					Used only to initialize r2t_data on midlevel call back
 *					(can be negative)
 *
 *	r2t_data		The remaining number of bytes left to be solicited by R2Ts
 *					set to 0 until midlevel calls back iscsi_rdy_to_xfer()
 *					to indicate target's receive buffers have been set up
 *					updated as R2Ts are sent
 *					(can be negative)
 *
 *	retransmit_flg	1 when retransmitting a previously sent R2T in response to
 *					a timeout on target or an explicit SNACK from initiator
 *					Not set on a timeout when recovery_r2t is also non-zero.
 *
 *	recovery_r2t	non-zero when recovering from target-detected error in a
 *					DataOut received from initiator
 *					1	waiting to get DataOut with F=1 in current burst
 *					2	got DataOut with F=1 or got timeout, need to send
 *						the recovery R2T to initiator
 *					0	after recovery R2T has been sent
 *
 *	startsn			used only when retransmit_flg or recovery_r2t are non-zero
 *					holds R2TSN of next cookie to be used to send next R2T
 *
 *	endsn			used only when retransmit_flg or recovery_r2t are non-zero
 *					holds R2TSN of last cookie to be used to send R2Ts
 *
 *	prev_data_sn	used only during READ commands
 *
 *	timestamp		set to jiffies upon receipt of iSCSI command and DataOut
 *					pdus, also on sending of an R2T.  Used to check for
 *					initiator inactivity which provokes a timeout R2T retransmit
 *
 *	outstanding_r2t	keeps track of number of R2Ts sent to initiator but for
 *					which the end of the corresponding solicited sequence has
 *					not been detected (i.e., by receiving a DataOut pdu with
 *					the F bit set).  Initialized to 0, incremented for every
 *					new or recovery (i.e., not retransmitted) R2T sent,
 *					decremented for every DataOut pdu received with F bit set.
 *
 *	r2t_sn			counts number of new and recovery (i.e., not retransmitted)
 *					R2Ts sent so far to initiator for this command.  This value
 *					becomes the R2TSN in the next new or recovery R2T sent.
 *					Initialized to 0, incremented for every new or recovery R2T
 *					sent.
 *
 *	data_sn			number expected in DataSN field of next DataOut pdu received
 *					in current sequence.  Initialized to 0 and reset to 0 at the
 *					end of every sequence (i.e., upon receiving a DataOut with
 *					F bit set).  Incremented by one for every DataOut pdu
 *					received.
 */
static int
iscsi_tx_r2t(struct iscsi_cmnd *cmnd,
			 struct iscsi_conn *conn,
			 struct iscsi_session *session)
{
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	int err = 0;
	struct iscsi_targ_r2t *hdr;
	int data_length_left;
	int max_burst_len = 0;
	struct iscsi_cookie *cookie = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_tx_r2t, r2t_data %d, "
		  "retransmit_flg %u, outstanding_r2t %d, recovery_r2t %u\n",
		  cmnd->r2t_data, cmnd->retransmit_flg, cmnd->outstanding_r2t,
		  cmnd->recovery_r2t);

	if ((data_length_left = cmnd->r2t_data) <= 0 && !cmnd->retransmit_flg
		&& cmnd->recovery_r2t != 2) {
		/* no need to send any (more) R2Ts (yet) */
		goto r2t_out;
	}

	/* try to send all the R2T's at once, but limited by MaxOutstandingR2T */
	do {
		if (session->oper_param->MaxOutstandingR2T <= cmnd->outstanding_r2t
			&& !cmnd->retransmit_flg && cmnd->recovery_r2t != 2) {
			/* we have hit the limit on number of outstanding R2Ts allowed */
			goto r2t_out;
		}

		memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
		hdr = (struct iscsi_targ_r2t *) iscsi_hdr;

		hdr->opcode = ISCSI_TARG_R2T;
		hdr->flags |= F_BIT;
		hdr->init_task_tag = cpu_to_be32(cmnd->init_task_tag);
		hdr->target_xfer_tag = cpu_to_be32(cmnd->target_xfer_tag);
		hdr->stat_sn = cpu_to_be32(conn->stat_sn);
		hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
		hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);
		if ((cmnd->retransmit_flg || cmnd->recovery_r2t == 2)) {
			/* sending a recovery R2T or retransmitting R2Ts in response to
			 * an R2T SNACK from initiator
			 */
			for (cookie=cmnd->first_r2t_cookie; cookie; cookie=cookie->next) {
				if (cookie->seq == cmnd->startsn) {
					/* this cookie lets us regenerate the missing r2t */
					if (cmnd->retransmit_flg) {
						TRACE_WARNING("%s Retransmit R2T, ITT %u R2TSN %u "
									  "Buffer Offset %u\n", current->comm,
									  cmnd->init_task_tag,
									  cookie->seq, cookie->offset);
					}
					hdr->r2t_sn = cpu_to_be32(cookie->seq);
					hdr->offset = cpu_to_be32(cookie->offset);
					hdr->xfer_len = cpu_to_be32(cookie->xfer_len);
					if (cmnd->recovery_r2t == 2) {
						cmnd->r2t_data = data_length_left = cmnd->data_length
										- (cookie->offset + cookie->xfer_len);
					} else if (++cmnd->startsn <= cmnd->endsn && cookie->next) {
						/* more R2Ts to retransmit */
						data_length_left = cookie->next->xfer_len;
					} else {
						/* last R2T to retransmit, go back to normal r2ts */
						data_length_left = cmnd->r2t_data;
						cmnd->retransmit_flg = 0;
					}
					break;
				}
			}
			if (cmnd->recovery_r2t == 2) {
				/* done with recovery if get here */
				cmnd->recovery_r2t = 0;
			}
			if (!cookie) {
				/* no cookies to retransmit, go back to normal r2t processing */
				data_length_left = cmnd->r2t_data;
				cmnd->retransmit_flg = 0;
				continue;
			}
		} else {

			hdr->r2t_sn = cpu_to_be32(cmnd->r2t_sn++);
			hdr->offset = cpu_to_be32(cmnd->data_length - data_length_left);

			max_burst_len = session->oper_param->MaxBurstLength;

			/* form an R2T cookie to be used for re-transmissions - SAI */
			if (session->oper_param->ErrorRecoveryLevel > 0) {
				cookie = create_r2t_cookie(cmnd);

				if (cookie) {
					cookie->seq = cmnd->r2t_sn - 1;
					cookie->offset = cmnd->data_length - data_length_left;
					cookie->xfer_len = (data_length_left <= max_burst_len)
											? data_length_left : max_burst_len;
				}
			}

			if (data_length_left <= max_burst_len) {
				/* this is the last R2T we need to send for this command */
				hdr->xfer_len = cpu_to_be32(data_length_left);
				data_length_left = 0;
			} else {
				hdr->xfer_len = cpu_to_be32(max_burst_len);
				data_length_left -= max_burst_len;
			}
			cmnd->r2t_data = data_length_left;

			/* if DataSequenceInOrder is no, send r2t offset in reverse order */
			if (!session->oper_param->DataSequenceInOrder)
				hdr->offset = cpu_to_be32(data_length_left
								+ cmnd->first_burst_len);

			/* Increment the outstanding R2T for every R2T sent - SAI */
			cmnd->outstanding_r2t++;
		}

		if (send_hdr_only(conn, iscsi_hdr) < 0) {
			err = -1;
			goto r2t_out;
		}

		TRACE(TRACE_ISCSI, "%s r2t sent, ITT %u, offset %d\n",
			  current->comm, cmnd->init_task_tag,
			  be32_to_cpu(hdr->offset));

		if (TRACE_TEST(TRACE_ISCSI_FULL))
			print_targ_r2t(hdr);

		/* store the r2t time stamp - SAI */
		/* error recovery ver ref18_04 */
		cmnd->timestamp = jiffies;
	} while (data_length_left > 0);

r2t_out:
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_tx_r2t, r2t_data %d, "
		  "retransmit_flg %u, outstanding_r2t %d, recovery_r2t %u, err %d\n",
		  cmnd->r2t_data, cmnd->retransmit_flg, cmnd->outstanding_r2t,
		  cmnd->recovery_r2t, err);
	return err;
}

/*
 * find_iovec_needed: this function returns the number of iovecs that
 * are needed to receive some data
 * INPUT: amount of data needed, pointer to scattergather list (from the
 * 	  point where the data is to start for this round, offset within
 * 	  the first scattergather element - in case it was not used up
 * 	  completely
 * OUTPUT: number of iovectors needed, else < 0 if there is trouble
 */
int __attribute__ ((no_instrument_function))
find_iovec_needed(__u32 data_len, struct scatterlist *st_list, __u32 offset)
{
	int i = 0;
	__u32 sglen;

	while (data_len) {
		sglen = get_sglen(st_list);
		if (data_len > (sglen - offset))
			data_len -= (sglen - offset);
		else
			data_len = 0;

		i++;
		st_list++;
		offset = 0;
	}

	return (i);
}

/*
 * fill_iovec: This function fills in a given iovec structure so as to
 * reflect a given scattergather list upto the length that is required
 * No checks are made on the length on the iovec given except that it
 * needs to be greater than the size needed. Also, the scatterlist is
 * also not checked. The integer offset upon return reflects the posn of
 * the offset in the last scatterlist that was used for this function
 * INPUT: struct iovec, no. of iovecs, scatterlist, ptr to offset within
 * 	  the first element, and the total length required
 * OUTPUT: how many st_list elements are used, < 0 if there is trouble
 */
int __attribute__ ((no_instrument_function))
fill_iovec(struct iovec *iov, int p, int niov,
		   struct scatterlist *st_list, int *offset, __u32 data)
{
	int count = 0;
	__u32 sglen;

	TRACE(TRACE_DEBUG, "offset: %d, data_len: %d\n", *offset, data);

	iov += p;
	while (data > 0 && p < niov) {
		iov->iov_base = (void *)(get_sgbuf(st_list) + (*offset));
		sglen = get_sglen(st_list);
		if ((sglen - *offset) > data) {
			/* more space left in this st list item than iovector can use */
			iov->iov_len = data;
			*offset += data;
		} else {
			/* iovector uses all the space in this st list item */
			iov->iov_len = sglen - *offset;
			*offset = 0;
			count++;
		}

		data -= iov->iov_len;

		TRACE(TRACE_DEBUG, "iov %p, p %d, iov_base %p, iov_len %d\n", iov,
			  p, iov->iov_base, iov->iov_len);

		p++;
		iov++;
		st_list++;
	}

	return count;
}

/*  called when a file in the /proc/scsi_target/iscsi/ directory is read or written */
int
iscsi_proc_info(char *buffer, char **start, off_t offset,
				int length, int inode, int inout)
{
	int result;
	int int_val;
	int param_neg_info;
	int tag, action;
	char *ptr;
	char *ptr_to_keytext;
	struct iscsi_session *session;
	struct iscsi_conn *conn;
	int i, count;
	char ip_string[INET6_ADDRSTRLEN+2], port_string[8];
	struct list_head *list_ptr1, *list_ptr2;

	TRACE(TRACE_DEBUG, "Entering iscsi_proc_info\n");

	if (inout == 1) {
		/* write to this file, buffer should be "iscsi_initiator ip 1234abcd 
		   target 6 lun 0" */
		if (buffer == NULL || length > PAGE_SIZE)
			result = -EINVAL;
		else {
			result = length;

			if (length == PAGE_SIZE)
				length--;

			buffer[length] = '\0';	/* make the buffer into a string */

			TRACE(TRACE_DEBUG, "string got: %s\n", buffer);

			if (length >= 16
				&& strncmp(buffer, "iscsi_target ip ", 16) == 0) {
				/* this buffer has expected preface, skip over ip address */
				char *ip_string, *port_string, *tag_string;
				
				ip_string = buffer + 16;
				ptr = strchr(ip_string, ' ');
				
				if (ptr == NULL || strncmp(ptr, " port ", 6) != 0) {
					result = -EINVAL;
					goto out;
				}
				*ptr = '\0';
				
				/* this buffer has expected preface, scan port number */
				port_string = ptr + 6;
				
				ptr = strchr(port_string, ' ');
				if (strncmp(ptr, " tag ", 5) != 0) {
					result = -EINVAL;
					goto out;
				}
				*ptr = '\0';
			
				/* this buffer has expected preface, scan target number */
				tag_string = ptr + 5;
				
				tag = simple_strtoul(tag_string, &ptr, 0);
				
				ptr = strchr(tag_string, ' ');
				if (strncmp(ptr, " action ", 8) != 0) {
					result = -EINVAL;
					goto out;
				}
			
				/* this buffer has expected preface, scan action number */
				action = simple_strtoul(ptr + 8, &ptr, 0);
				
				if (action == 0) {
					if (start_server_thread(ip_string, port_string, tag) < 0)
					{
						result = -EINVAL;
						goto out;
					}
				}
				else {
					if (stop_server_thread(ip_string, port_string) < 0)
					{
						result = -EINVAL;
						goto out;
					}
				}
			}
			else if (length >= 20
				&& strncmp(buffer, "iscsi_target manage ", 20) == 0) {
				/* this is for configuring the login parameters */

				/* this buffer has expected interface, scan neg_info flags */
				param_neg_info =
					simple_strtoul(buffer + 20, &ptr_to_keytext, 0);

				/* Move a position to ignore space */
				ptr_to_keytext++;

				/* Now configure the parameters */
				if (param_neg_info == FORCE) {
					/* special to force login subphases */
					if (strcmp(ptr_to_keytext, "s") == 0)
						devdata->force |= USE_SECURITY_PHASE;
					else if (strcmp(ptr_to_keytext, "o") == 0)
						devdata->force |= USE_OPERATIONAL_PHASE;
					else if (strcmp(ptr_to_keytext, "r") == 0)
						devdata->force |= USE_FULL_REPLIES;
					else if (strcmp(ptr_to_keytext, "xok") == 0)
						devdata->force |= USE_REFLECT_XKEYS;
					else if (strcmp(ptr_to_keytext, "tk1") == 0)
						devdata->force |= USE_ONE_KEY_PER_TEXT;
					else if (strncmp(ptr_to_keytext, "n=", 2) == 0)
						devdata->nop_period =
							simple_strtoul(ptr_to_keytext + 2, NULL, 0) * HZ;
					else if (strncmp(ptr_to_keytext, "v=", 2) == 0) {	
						/* formerly set draft number to enforce */
						int_val = simple_strtoul(ptr_to_keytext + 2, NULL, 0);
						if (int_val != DRAFT20) {
							TRACE(TRACE_ISCSI, "Draft number %u ignored\n",
								  int_val / DRAFT_MULTIPLIER);
						}
					}
					else if (strncmp(ptr_to_keytext, "V=", 2) == 0) {
							char  *ptr;
							int trace_info;
							TRACE(TRACE_ISCSI, "Debug Verbose set to \"%s\"\n",
									ptr_to_keytext + 2);
							trace_info = simple_strtoul(ptr_to_keytext + 2,
									&ptr, 0);
							printk("trace_info = %x\n",trace_info);
							TRACE_SET(trace_info);
					}

					/* set phase-collapse choice */
					else if (strncmp(ptr_to_keytext, "p=", 2) == 0) {
						int_val = simple_strtoul(ptr_to_keytext + 2, NULL, 0);
						if (int_val > 2 || int_val < 0) {
							TRACE_ERROR("%s Bad value for phase-collapse "
										"setting: %d", current->comm, int_val);
						} else
							devdata->phase_collapse = int_val;
					}
					/* set the r2t period - SAI */
					else if (strncmp(ptr_to_keytext, "r2tp=", 5) == 0) {
						devdata->r2t_period =
							simple_strtoul(ptr_to_keytext + 5, NULL, 0) * HZ;
					}
					/* chap and srp support - CHONG */
					else if (strcmp(ptr_to_keytext, "t") == 0) {
						TRACE(TRACE_ISCSI, "target confirmation enabled\n");
						devdata->auth_parameter.auth_flags |=
							USE_TARGET_CONFIRMATION;
					} else if (strcmp(ptr_to_keytext, "b") == 0) {
						TRACE(TRACE_ISCSI, "base64 number enabled\n");
						devdata->auth_parameter.auth_flags |= USE_BASE64;
						devdata->auth_parameter.chap_local_ctx->number_format =
							BASE64_FORMAT;
						devdata->auth_parameter.chap_peer_ctx->number_format =
							BASE64_FORMAT;
					} else if (strncmp(ptr_to_keytext, "px=", 3) == 0) {
						TRACE(TRACE_ISCSI, "CHAP peer secret set to %s\n",
							  ptr_to_keytext + 3);
						CHAP_SetSecret(ptr_to_keytext + 3,
									   devdata->auth_parameter.chap_peer_ctx);
					} else if (strncmp(ptr_to_keytext, "pn=", 3) == 0) {
						TRACE(TRACE_ISCSI, "CHAP peer name set to %s\n",
							  ptr_to_keytext + 3);
						CHAP_SetName(ptr_to_keytext + 3,
									 devdata->auth_parameter.chap_peer_ctx);
					} else if (strncmp(ptr_to_keytext, "lx=", 3) == 0) {
						TRACE(TRACE_ISCSI, "CHAP local secret set to %s\n",
							  ptr_to_keytext + 3);
						CHAP_SetSecret(ptr_to_keytext + 3,
									   devdata->auth_parameter.chap_local_ctx);
					} else if (strncmp(ptr_to_keytext, "ln=", 3) == 0) {
						TRACE(TRACE_ISCSI, "CHAP local name set to %s\n",
							  ptr_to_keytext + 3);
						CHAP_SetName(ptr_to_keytext + 3,
									 devdata->auth_parameter.chap_local_ctx);
					} else if (strncmp(ptr_to_keytext, "cl=", 3) == 0) {
						int_val = simple_strtoul(ptr_to_keytext + 3, NULL, 0);
						if (int_val <= 0 || int_val > MAX_CHAP_BINARY_LENGTH) {
							TRACE_ERROR("%s invalid CHAP challenge length %d\n",
										current->comm, int_val);
						}
						TRACE(TRACE_ISCSI, "challenge length set to %d\n",
							  int_val);
						CHAP_SetChallengeLength(int_val,
												devdata->auth_parameter.
												chap_local_ctx);
					} else if (strncmp(ptr_to_keytext, "sx=", 3) == 0) {
						TRACE(TRACE_ISCSI, "SRP secret set to \"%s\"\n",
							  ptr_to_keytext + 3);
						SRP_SetSecret(ptr_to_keytext + 3,
									  devdata->auth_parameter.srp_ctx);
					} else if (strncmp(ptr_to_keytext, "sn=", 3) == 0) {
						TRACE(TRACE_ISCSI, "SRP name set to \"%s\"\n",
							  ptr_to_keytext + 3);
						SRP_SetName(ptr_to_keytext + 3,
									devdata->auth_parameter.srp_ctx);
					} else {
						TRACE_ERROR("%s unknown force \"%s\"\n", current->comm,
									ptr_to_keytext);
					}
				} else if (param_neg_info == RESTORE) {
					/* restore all settings to default values */

					/* reset parameter list to copy of defaults */
					param_tbl_uncpy(*devdata->param_tbl);
					param_tbl_init(*devdata->param_tbl);

					/* reset all other fields to 0 */
					devdata->force = 0;
					devdata->nop_period = 0;
					devdata->r2t_period = 0;
					devdata->targ_snack_flg = 0;
				} else if (param_neg_info == SNACK) {
					/* Error Recovery options to enable or disable SNACK - SAI*/

					if (strncmp(ptr_to_keytext, "a=", 2) == 0) {
						if ((strcmp(ptr_to_keytext + 2, "y") == 0)
							|| strcmp(ptr_to_keytext + 2, "Y") == 0) {
							devdata->targ_snack_flg |= DATACK_SNACK_ENABLE;
						} else {
							devdata->targ_snack_flg &= ~DATACK_SNACK_ENABLE;
						}
					} else if (strncmp(ptr_to_keytext, "d=", 2) == 0) {
						if ((strcmp(ptr_to_keytext + 2, "y") == 0)
							|| strcmp(ptr_to_keytext + 2, "Y") == 0) {
							devdata->targ_snack_flg |= DATA_SNACK_ENABLE;
						} else {
							devdata->targ_snack_flg &= ~DATA_SNACK_ENABLE;
						}
					} else if (strncmp(ptr_to_keytext, "s=", 2) == 0) {
						if ((strcmp(ptr_to_keytext + 2, "y") == 0)
							|| strcmp(ptr_to_keytext + 2, "Y") == 0) {
							devdata->targ_snack_flg |= STATUS_SNACK_ENABLE;
						} else {
							devdata->targ_snack_flg &= ~STATUS_SNACK_ENABLE;
						}
					} else if (strncmp(ptr_to_keytext, "h=", 2) == 0) {
						if (simple_strtoul(ptr_to_keytext + 2, NULL, 0) == 1)
							devdata->targ_snack_flg |= DATA_SNACK_RETRANSMIT;
						else if (simple_strtoul(ptr_to_keytext + 2, NULL, 0) ==
								 2)
							devdata->targ_snack_flg |= DATA_SNACK_REJECT;

						TRACE(TRACE_DEBUG,
							  "SNACK Flags : datasnack %d statsnack %d DataSnackRexmit %d  DataSnackReject %d\n",
							  (devdata->targ_snack_flg & DATA_SNACK_ENABLE),
							  ((devdata->
								targ_snack_flg & STATUS_SNACK_ENABLE) >> 1),
							  ((devdata->
								targ_snack_flg & DATA_SNACK_RETRANSMIT) >> 7),
							  ((devdata->
								targ_snack_flg & DATA_SNACK_REJECT) >> 6));
					} else {
						TRACE_ERROR("%s unknown snack \"%s\"\n", current->comm,
									ptr_to_keytext);
					}
				} else {
					configure_parameter(param_neg_info, ptr_to_keytext,
										*devdata->param_tbl);
				}
			} else
				result = -EINVAL;
		}
	} else {
		result = sprintf(buffer, "UNH-IOL iSCSI Target ");
#ifdef MEMORYIO
		result += sprintf(buffer + result, "in Memory Mode ");
#else
#ifdef DISKIO
		result += sprintf(buffer + result, "in Disk Mode ");
#else
#ifdef FILEIO
		result += sprintf(buffer + result, "in File Mode ");
#else
		result += sprintf(buffer + result, "in Generic Mode ");
#endif
#endif
#endif

		result += sprintf(buffer + result, "with BLOCKSIZE %d\n", BLOCKSIZE);

		result += sprintf(buffer + result, "Total of %d target devices\n",
						  target_count);

		count = 0;
		for (i = 0; i < MAX_PORTAL; i++)
		{
			if (iscsi_portal_groups[i].in_use)
				count++;
		}
		result += sprintf(buffer + result, "Total of %d server threads\n", count);

		for (i=0; i<MAX_PORTAL; i++)
		{
			if (!iscsi_portal_groups[i].in_use)
				continue;

			result += 
				sprintf(buffer + result, "    server_thread_%d at %s:%s,%d\n",
											i, 
											iscsi_portal_groups[i].ip_string, 
											iscsi_portal_groups[i].port_string,
											iscsi_portal_groups[i].tag);
		}	

		/* clean bad sessions to get current information */
		clean_bad_stuff();

		/* non-destructive access to session lists */
		if (down_interruptible(&devdata->session_read_mutex))
			goto out;

		if (devdata->session_readers++ == 0) {
			/* first reader, must get access for all readers */
			if (down_interruptible(&devdata->session_sem)) {
				up(&devdata->session_read_mutex);
				goto out;
			}
		}
		up(&devdata->session_read_mutex);

		count = 0;
		list_for_each(list_ptr1, &devdata->session_list) {
			count++;
		}
		result += sprintf(buffer + result, "Total of %d sessions\n", count);
		if (count == 0) {
			result += print_config_info(*devdata->param_tbl, buffer + result);
		}
		list_for_each(list_ptr1, &devdata->session_list) {
			session = list_entry(list_ptr1, struct iscsi_session, sess_link);
			result +=
				sprintf(buffer + result, "    session for target %u is %p\n",
						session->oper_param->TargetName, session);
			result +=
				sprintf(buffer + result, "    session has %u connections\n",
						session->nconn);
			list_for_each(list_ptr2, &session->conn_list) {
				conn = list_entry(list_ptr2, struct iscsi_conn, conn_link);
				result +=
					sprintf(buffer + result,
							"        connection with CID %u is %p on socket %p\n",
							conn->cid, conn, conn->conn_socket);
				if (cnv_inet_to_string(conn->ip_address, ip_string, port_string)
																		> 0) {
					result +=
						sprintf(buffer + result, "%30s  %s:%s\n",
							"Initiator Address", ip_string, port_string);
				}
				if (cnv_inet_to_string(conn->local_ip_address, ip_string,
															port_string) > 0) {
					result +=
						sprintf(buffer + result, "%30s  %s:%s\n",
							"Local Address", ip_string, port_string);
				}
			}
			/* if target has lots of sessions, printing the session parameters
			 * for all of them here may overflow the proc 4K buffer.
			 */
			result +=
				print_session_params(session->oper_param,
									 *session->session_params,
									 buffer + result);
			result += sprintf(buffer + result, "\n");
		}

		down_interruptible(&devdata->session_read_mutex);
		if (--devdata->session_readers == 0) {
			/* last reader, must release access for all readers */
			up(&devdata->session_sem);
		}
		up(&devdata->session_read_mutex);

	}

out:
	TRACE(TRACE_DEBUG, "Leaving iscsi_proc_info\n");
	return result;
}

/* Daren Hayward, darenh@4bridgeworks.com */
#if defined(MANGLE_INQUIRY_DATA)

/*
 * offsets to first byte and limit byte of version descriptors
 * section of standard INQUIRY data layout
 */
#define VER_DESC_START	58
#define VER_DESC_LIMIT	74

/*
 * Mangle the returned INQUIRY data to ensure compliance
 * INPUT: struct iscsi_cmnd, struct iovec,
 * number of first data iovec, number of limit data iovec
 * OUTPUT returns 0 if successfull, -1 on error
 * 22nd July 2004 Daren Hayward, modified by rdr
 */
static int
mangle_inquiry_data(struct iscsi_cmnd *iscsi_command, struct iovec *iov,
			int start_iov, int limit_iov)
{
	Scsi_Request *req;
	int data_size;
	int i, j, offset;
	int ver_desc_start;
	int ver_desc_limit;
	unsigned char *vdb;		/* Version Descriptor byte pointer */

	if (iscsi_command->cmnd == NULL || (req = iscsi_command->cmnd->req) == NULL
		|| req->sr_command == NULL)
		return 0;

	if (req->sr_cmnd[0] != INQUIRY || (req->sr_cmnd[1] & 0x3) != 0) {
		/* Not INQUIRY, or not Standard INQUIRY data format */
		return 0;
	}

	TRACE(TRACE_ISCSI, "INQUIRY of STANDARD type\n");
	TRACE(TRACE_DEBUG, "start_iov=%d, limit_iov=%d\n", start_iov, limit_iov);

	/* Is it possible for data_size to be negative?? */
	data_size = req->sr_bufflen - req->sr_command->resid;
	TRACE(TRACE_DEBUG, "mangle_inquiry_data: sr_bufflen %d, resid %d, "
		  "data_size %d\n", req->sr_bufflen, req->sr_command->resid, data_size);
	if (data_size <= VER_DESC_START) {
		/* Version Descriptors start 58 bytes in */
		return 0;
	}

	/* ensure iSCSI(0960) is in Version Descriptors */
	offset = 0;
	iov += start_iov;
	for (i = start_iov; i < limit_iov; iov++, i++) {
		if (offset >= VER_DESC_LIMIT) {
			/* we've gone past end of descriptor area */
			break;
		}
		if (offset + iov->iov_len <= VER_DESC_START) {
			/* not yet up to start of descriptor area */
			offset += iov->iov_len;
			continue;
		}
		if (offset < VER_DESC_START) {
			/* start part way into this item */
			ver_desc_start = VER_DESC_START - offset;
		} else {
			/* start at beginning of this item */
			ver_desc_start = 0;
		}
		offset += iov->iov_len;
		if (offset <= VER_DESC_LIMIT) {
			/* use this entire item */
			ver_desc_limit = iov->iov_len;
		} else {
			/* use only first part of this item */
			ver_desc_limit = VER_DESC_LIMIT - (offset - iov->iov_len);
		}
		TRACE(TRACE_DEBUG, "i=%d, iov_len=%d, start=%d, limit=%d\n", i,
			  iov->iov_len, ver_desc_start, ver_desc_limit);

		/* find first all-zero vector and change it to ISCSI */
		for (j = ver_desc_start; j < ver_desc_limit; j += 2) {
			vdb = (unsigned char *)iov->iov_base;
			TRACE(TRACE_ISCSI, "Version Descriptor %d = %.2x %.2x\n",
				  j, vdb[j], vdb[j + 1]);
			/* Check to see if anything has already been set */
			if (vdb[j] == 0x09 && vdb[j + 1] == 0x60) {
				/* ISCSI code already set, leave now */
				TRACE(TRACE_ISCSI, "ISCSI Version Descriptor found\n");
				return 0;
			}
			if (vdb[j] == 0 && vdb[j + 1] == 0) {
				/* Set empty Version Descriptor to be ISCSI */
				vdb[j] = 0x09;
				vdb[j + 1] = 0x60;
				TRACE(TRACE_ISCSI, "ISCSI Version Descriptor added\n");
				return 0;
			}
		}
	}

	/* Success */
	return 0;
}
#endif

/*      
 * Note the name "driver_template" is required for "scsi_module.c" to work! 
 *      Used by: "/usr/src/linux/drivers/scsi/scsi_module.c" 
 *      Wrap this in an ifdef because "scsi_module.c" says to! 
 */
#ifdef MODULE

MODULE_AUTHOR("UNH IOL ISCSI");
MODULE_DESCRIPTION("UNH-IOL iSCSI reference target");
MODULE_SUPPORTED_DEVICE("sd");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

/* initializor for static Scsi_Target_Template */

#define ISCSI {\
  name: "iscsi_target", \
  detect:iscsi_detect, \
  release:iscsi_release,\
  xmit_response:iscsi_xmit_response, \
  rdy_to_xfer:iscsi_rdy_to_xfer,\
  proc_info:iscsi_proc_info, \
  task_mgmt_fn_done:iscsi_task_mgt_fn_done \
};

Scsi_Target_Template my_template = ISCSI;

#include "scsi_target_module.c"

#endif
